#!/bin/sh



#
# parseArg - see if arg 1 exists if so then this is the directory where the auto.sh
#            script lives.
#

function parseArg () {
   if [ -f options ] ; then
       export M2OPTIONS="`cat options`"
   else
       export M2OPTIONS="-O -g -bounds"
   fi
#  echo set options to $M2OPTIONS
   if [ "$1" = "" ] ; then
       echo "testing gm2-i386-gas from " `type gm2-i386-gas`
       AUTODIR="."
   else
       AUTODIR="$1"
   fi
}


#
# checkRun - check whether we should link module $1 and test it's execution.
#

function checkRun () {
   if [ -f mustrunandfail ] ; then
       m2f -M ". $AUTODIR/../../comp/libs" $M2OPTIONS $1 > /dev/null
       m2f -M ". $AUTODIR/../../comp/libs" -l $M2OPTIONS $1
       if ./a.out >& /dev/null ; then
           echo "failed run"
           exit 1
       else
           echo -n "[r]"
       fi
   fi
   if [ -f mustrunandpass ] ; then
       m2f -M ". $AUTODIR/../../comp/libs" $M2OPTIONS $1 > /dev/null
       m2f -M ". $AUTODIR/../../comp/libs" -l $M2OPTIONS $1
       if ./a.out >& /dev/null ; then
           echo -n "[r]"
       else
           echo "failed run"
           exit 1
       fi
   fi
}


#
# compilePass - runs the compiler on all modules and expects it to succeed on all.
#

function compilePass () {
   j=0
   for i in *.mod ; do
        if [ -f $i ] ; then
           echo -n "$i"
           if gm2-i386-gas $M2OPTIONS -quiet -M ". $AUTODIR/../../comp/libs" -O `basename $i .mod` ; then
              checkRun $i
              echo -n "	"
              j=`expr $j + 1`
              if [ "`expr $j % 4`" = "0" ] ; then
                   echo ""
              fi
           else
              echo " x"
              echo "failed to compile $i" ;
              exit 1
           fi
        fi
    done
}


#
# compileWarning - runs the compiler on all modules and expects it to generate warnings on all.
#

function compileWarning () {
   j=0
   echo "------------------------------------------------------------"
   echo "     compiling warning code tests                           "
   echo "------------------------------------------------------------"
   for i in *.mod ; do
        if [ -f $i ] ; then
           echo -n "$i"
           gm2-i386-gas $M2OPTIONS -quiet -M ". $AUTODIR/../../comp/libs" -O `basename $i .mod` > messages
           if grep warn messages > /dev/null ; then
              echo -n "	"
              j=`expr $j + 1`
              if [ "`expr $j % 4`" = "0" ] ; then
                   echo ""
              fi
           else
              echo " x"
              echo "failed to detect warning in $i" ;
              exit 1
           fi
        fi
    done
    rm -f messages
    echo " "
    echo "------------------------------------------------------------"
}


#
# compileFail - runs the compiler on all modules and expects it to fail on all.
#

function compileFail () {
   j=0
   echo "------------------------------------------------------------"
   echo "     compiling errant code tests                            "
   echo "------------------------------------------------------------"
   for i in *.mod ; do
        if [ -f $i ] ; then
           echo -n "$i"
           if gm2-i386-gas $M2OPTIONS -quiet -M ". $AUTODIR/../../comp/libs" -O `basename $i .mod` > /dev/null ; then
              echo " x"
              echo "failed to detect error in $i" ;
              exit 1
           else
              echo -n "	"
              j=`expr $j + 1`
              if [ "`expr $j % 4`" = "0" ] ; then
                   echo ""
              fi
           fi
        fi
    done
    echo "------------------------------------------------------------"
}


#
# testModules - test all the modules
#

function testModules () {
   unset DEBUGCSE
   if [ -f warning ] ; then
       compileWarning
   elif [ -f mustfail ] ; then
       compileFail
   else
       compilePass
   fi
}


#
# runDirs - traverse subdirectories and execute the test programs.
#

function runDirs () {
   for i in * ; do
      if [ -d $i ] ; then
          if [ "$i" != "CVS" ] ; then
              echo " "
              echo "-------------------------------------------------"
              echo "Regression tests running in: `pwd`/$i"
              echo "-------------------------------------------------"
              cd $i ; ../$AUTODIR/auto.sh "../$AUTODIR" ; result=$? ; cd ..
              echo " "
              if [ "$result" = "0" ] ; then
                  echo "================================================="
                  echo "Directory `pwd`/$i has PASSED"
                  echo "================================================="
              else
                  echo "Directory `pwd`/$i has FAILED"
              fi
          fi
      fi
   done
}


#
# main
#

parseArg $1
testModules
runDirs
echo ""
