MODULE Parser ;


FROM DiskIO IMPORT GetNextSymbol, CurrentSymbol, OpenSource,
                   CloseSourceFile, WriteError ;

FROM Args IMPORT GetArg, Narg ;

FROM StrLib IMPORT StrEqual, StrLen ;

FROM StrIO IMPORT WriteString, WriteLn ;


VAR
   CharNo: CARDINAL ;  (* Index into the CurrentSymbol *)
   High  : CARDINAL ;  (* StrLen(CurrentSymbol)        *)


(* InitChar - resets the CharNo. *)

PROCEDURE InitChar ;
BEGIN
   CharNo := 0 ;
   High := StrLen(CurrentSymbol)
END InitChar ;


(* Returns TRUE if current character is ch *)
(*         FALSE otherwise.                *)
(* and increments the CharNo if TRUE       *)

PROCEDURE CharIs (ch: CHAR) : BOOLEAN ;
BEGIN
   IF (CharNo<High) AND (ch=CurrentSymbol[CharNo])
   THEN
      INC( CharNo ) ;
      RETURN( TRUE )
   ELSE
      RETURN( FALSE )
   END
END CharIs ;


(* IsChar - tests for current character = ch *)
(*          no side effects upon CharNo.     *)

PROCEDURE IsChar (ch: CHAR) : BOOLEAN ;
BEGIN
   RETURN (CharNo<High) AND (CurrentSymbol[CharNo]=ch)
END IsChar ;


(* Built in predicate for Character *)

PROCEDURE Character () : BOOLEAN ;
BEGIN
   IF CharNo<High
   THEN
      INC(CharNo) ;
      RETURN( TRUE )
   ELSE
      RETURN( FALSE )
   END
END Character ;


(* Built in predicate for Letter *)

PROCEDURE Letter () : BOOLEAN ;
BEGIN
   IF (CharNo<High) AND
      (((CurrentSymbol[CharNo]>='a') AND (CurrentSymbol[CharNo]<='z')) OR
       ((CurrentSymbol[CharNo]>='A') AND (CurrentSymbol[CharNo]<='Z')))
   THEN
      INC(CharNo) ;
      RETURN( TRUE )
   ELSE
      RETURN( FALSE )
   END
END Letter ;


PROCEDURE SymIs (a: ARRAY OF CHAR) : BOOLEAN ;
BEGIN
   IF StrEqual(a, CurrentSymbol)
   THEN
      GetNextSymbol ;
      RETURN( TRUE )
   ELSE
      RETURN( FALSE )
   END
END SymIs ;


PROCEDURE IsSym (a: ARRAY OF CHAR) : BOOLEAN ;
BEGIN
   RETURN( StrEqual(a, CurrentSymbol) )
END IsSym ;


PROCEDURE Ident() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Letter()
   THEN
      REPEAT
         IF Letter()
         THEN
            Success := TRUE
         ELSIF Digit()
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      UNTIL NOT Success ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Ident ;

PROCEDURE Number() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Integer()
   THEN
      Success := TRUE
   ELSIF Real()
   THEN
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Number ;

PROCEDURE Integer() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Digit()
   THEN
      WHILE Digit() DO
      END ;
      Success := TRUE
   ELSIF OctalDigit()
   THEN
      WHILE OctalDigit() DO
      END ;
      IF SymIs('B')
      THEN
         Success := TRUE
      ELSIF SymIs('C')
      THEN
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSIF Digit()
   THEN
      WHILE HexDigit() DO
      END ;
      IF SymIs('H')
      THEN
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Integer ;

PROCEDURE Real() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Digit()
   THEN
      WHILE Digit() DO
      END ;
      IF SymIs('.')
      THEN
         WHILE Digit() DO
         END ;
         IF ScaleFactor()
         THEN
         END ;
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Real ;

PROCEDURE ScaleFactor() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('E')
   THEN
      IF SymIs('+')
      THEN
      ELSIF SymIs('-')
      THEN
      END ;
      IF Digit()
      THEN
         WHILE Digit() DO
         END ;
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ScaleFactor ;

PROCEDURE HexDigit() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Digit()
   THEN
      Success := TRUE
   ELSIF SymIs('A')
   THEN
      Success := TRUE
   ELSIF SymIs('B')
   THEN
      Success := TRUE
   ELSIF SymIs('C')
   THEN
      Success := TRUE
   ELSIF SymIs('D')
   THEN
      Success := TRUE
   ELSIF SymIs('E')
   THEN
      Success := TRUE
   ELSIF SymIs('F')
   THEN
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END HexDigit ;

PROCEDURE Digit() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF OctalDigit()
   THEN
      Success := TRUE
   ELSIF SymIs('8')
   THEN
      Success := TRUE
   ELSIF SymIs('9')
   THEN
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Digit ;

PROCEDURE OctalDigit() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('0')
   THEN
      Success := TRUE
   ELSIF SymIs('1')
   THEN
      Success := TRUE
   ELSIF SymIs('2')
   THEN
      Success := TRUE
   ELSIF SymIs('3')
   THEN
      Success := TRUE
   ELSIF SymIs('4')
   THEN
      Success := TRUE
   ELSIF SymIs('5')
   THEN
      Success := TRUE
   ELSIF SymIs('6')
   THEN
      Success := TRUE
   ELSIF SymIs('7')
   THEN
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END OctalDigit ;

PROCEDURE String() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs("'")
   THEN
      WHILE Character() DO
      END ;
      IF SymIs("'")
      THEN
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSIF SymIs('"')
   THEN
      WHILE Character() DO
      END ;
      IF SymIs('"')
      THEN
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END String ;

PROCEDURE Qualident() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Ident()
   THEN
      WHILE SymIs('.') DO
         IF Ident()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Qualident ;

PROCEDURE ConstantDeclaration() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Ident()
   THEN
      IF SymIs('=')
      THEN
         IF ConstExpression()
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ConstantDeclaration ;

PROCEDURE ConstExpression() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SimpleConstExpr()
   THEN
      IF Relation()
      THEN
         IF SimpleConstExpr()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ConstExpression ;

PROCEDURE Relation() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('=')
   THEN
      Success := TRUE
   ELSIF SymIs('#')
   THEN
      Success := TRUE
   ELSIF SymIs('<>')
   THEN
      Success := TRUE
   ELSIF SymIs('<')
   THEN
      Success := TRUE
   ELSIF SymIs('<=')
   THEN
      Success := TRUE
   ELSIF SymIs('>')
   THEN
      Success := TRUE
   ELSIF SymIs('>=')
   THEN
      Success := TRUE
   ELSIF SymIs('IN')
   THEN
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Relation ;

PROCEDURE SimpleConstExpr() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('+')
   THEN
   ELSIF SymIs('-')
   THEN
   END ;
   IF ConstTerm()
   THEN
      WHILE AddOperator() DO
         IF ConstTerm()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END SimpleConstExpr ;

PROCEDURE AddOperator() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('+')
   THEN
      Success := TRUE
   ELSIF SymIs('-')
   THEN
      Success := TRUE
   ELSIF SymIs('OR')
   THEN
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END AddOperator ;

PROCEDURE ConstTerm() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF ConstFactor()
   THEN
      WHILE MulOperator() DO
         IF ConstFactor()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ConstTerm ;

PROCEDURE MulOperator() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('*')
   THEN
      Success := TRUE
   ELSIF SymIs('/')
   THEN
      Success := TRUE
   ELSIF SymIs('DIV')
   THEN
      Success := TRUE
   ELSIF SymIs('MOD')
   THEN
      Success := TRUE
   ELSIF SymIs('AND')
   THEN
      Success := TRUE
   ELSIF SymIs('&')
   THEN
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END MulOperator ;

PROCEDURE ConstFactor() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Qualident()
   THEN
      Success := TRUE
   ELSIF Number()
   THEN
      Success := TRUE
   ELSIF String()
   THEN
      Success := TRUE
   ELSIF Set()
   THEN
      Success := TRUE
   ELSIF SymIs('(')
   THEN
      IF ConstExpression()
      THEN
         IF SymIs(')')
         THEN
            Success := TRUE
         ELSIF SymIs('NOT')
         THEN
            IF ConstFactor()
            THEN
               Success := TRUE
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ConstFactor ;

PROCEDURE Set() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Qualident()
   THEN
   END ;
   IF SymIs('{')
   THEN
      IF Element()
      THEN
         WHILE SymIs(',') DO
            IF Element()
            THEN
            END
         END
      END ;
      IF SymIs('}')
      THEN
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Set ;

PROCEDURE Element() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF ConstExpression()
   THEN
      IF SymIs('..')
      THEN
         IF ConstExpression()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Element ;

PROCEDURE TypeDeclaration() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Ident()
   THEN
      IF SymIs('=')
      THEN
         IF Type()
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END TypeDeclaration ;

PROCEDURE Type() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SimpleType()
   THEN
      Success := TRUE
   ELSIF ArrayType()
   THEN
      Success := TRUE
   ELSIF RecordType()
   THEN
      Success := TRUE
   ELSIF SetType()
   THEN
      Success := TRUE
   ELSIF PointerType()
   THEN
      Success := TRUE
   ELSIF ProcedureType()
   THEN
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Type ;

PROCEDURE SimpleType() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Qualident()
   THEN
      Success := TRUE
   ELSIF Enumeration()
   THEN
      Success := TRUE
   ELSIF SubrangeType()
   THEN
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END SimpleType ;

PROCEDURE Enumeration() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('(')
   THEN
      IF IdentList()
      THEN
         IF SymIs(')')
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Enumeration ;

PROCEDURE IdentList() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Ident()
   THEN
      WHILE SymIs(',') DO
         IF Ident()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END IdentList ;

PROCEDURE SubrangeType() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('[')
   THEN
      IF ConstExpression()
      THEN
         IF SymIs('..')
         THEN
            IF ConstExpression()
            THEN
               IF SymIs(']')
               THEN
                  Success := TRUE
               ELSE
                  Success := FALSE
               END
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END SubrangeType ;

PROCEDURE ArrayType() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('ARRAY')
   THEN
      IF SimpleType()
      THEN
         WHILE SymIs(',') DO
            IF SimpleType()
            THEN
            END
         END ;
         IF SymIs('OF')
         THEN
            IF Type()
            THEN
               Success := TRUE
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ArrayType ;

PROCEDURE RecordType() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('RECORD')
   THEN
      IF FieldListSequence()
      THEN
         IF SymIs('END')
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END RecordType ;

PROCEDURE FieldListSequence() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF FieldList()
   THEN
      WHILE SymIs(';') DO
         IF FieldList()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END FieldListSequence ;

PROCEDURE FieldList() : BOOLEAN ;
BEGIN
   IF IdentList()
   THEN
      IF SymIs(':')
      THEN
         IF Type()
         THEN
         END
      END
   ELSE
      IF SymIs('CASE')
      THEN
         IF Ident()
         THEN
            IF SymIs(':')
            THEN
            END
         END ;
         IF Qualident()
         THEN
            IF SymIs('OF')
            THEN
               IF Varient()
               THEN
                  WHILE SymIs('|') DO
                     IF Varient()
                     THEN
                     END
                  END ;
                  IF SymIs('ELSE')
                  THEN
                     IF FieldListSequence()
                     THEN
                     END
                  END ;
                  IF SymIs('END')
                  THEN
                  END
               END
            END
         END
      END
   END ;
   RETURN( TRUE )
END FieldList ;

PROCEDURE Varient() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF CaseLabelList()
   THEN
      IF SymIs(':')
      THEN
         IF FieldListSequence()
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Varient ;

PROCEDURE CaseLabelList() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF CaseLabels()
   THEN
      WHILE SymIs(',') DO
         IF CaseLabels()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END CaseLabelList ;

PROCEDURE CaseLabels() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF ConstExpression()
   THEN
      IF SymIs('..')
      THEN
         IF ConstExpression()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END CaseLabels ;

PROCEDURE SetType() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('SET')
   THEN
      IF SymIs('OF')
      THEN
         IF SimpleType()
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END SetType ;

PROCEDURE PointerType() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('POINTER')
   THEN
      IF SymIs('TO')
      THEN
         IF Type()
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END PointerType ;

PROCEDURE ProcedureType() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('PROCEDURE')
   THEN
      IF FormalTypeList()
      THEN
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ProcedureType ;

PROCEDURE FormalTypeList() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('(')
   THEN
      IF ProcedureParameters()
      THEN
      END ;
      IF SymIs(')')
      THEN
         IF SymIs(':')
         THEN
            IF Qualident()
            THEN
            END
         END ;
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END FormalTypeList ;

PROCEDURE ProcedureParameters() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('VAR')
   THEN
   END ;
   IF FormalType()
   THEN
      WHILE SymIs(',') DO
         IF SymIs('VAR')
         THEN
         END ;
         IF FormalType()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ProcedureParameters ;

PROCEDURE VariableDeclaration() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF IdentList()
   THEN
      IF SymIs(':')
      THEN
         IF Type()
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END VariableDeclaration ;

PROCEDURE Designator() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Qualident()
   THEN
      REPEAT
         IF SymIs('.')
         THEN
            IF Ident()
            THEN
               Success := TRUE
            ELSE
               Success := FALSE
            END
         ELSIF SymIs('[')
         THEN
            IF ExpList()
            THEN
               IF SymIs(']')
               THEN
                  Success := TRUE
               ELSE
                  Success := FALSE
               END
            ELSE
               Success := FALSE
            END
         ELSIF SymIs('^')
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      UNTIL NOT Success ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Designator ;

PROCEDURE ExpList() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Expression()
   THEN
      WHILE SymIs(',') DO
         IF Expression()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ExpList ;

PROCEDURE Expression() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SimpleExpression()
   THEN
      IF Relation()
      THEN
         IF SimpleExpression()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Expression ;

PROCEDURE SimpleExpression() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('+')
   THEN
   ELSIF SymIs('-')
   THEN
   END ;
   IF Term()
   THEN
      WHILE AddOperator() DO
         IF Term()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END SimpleExpression ;

PROCEDURE Term() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Factor()
   THEN
      WHILE MulOperator() DO
         IF Factor()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Term ;

PROCEDURE Factor() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Number()
   THEN
      Success := TRUE
   ELSIF String()
   THEN
      Success := TRUE
   ELSIF Set()
   THEN
      Success := TRUE
   ELSIF Designator()
   THEN
      IF ActualParameters()
      THEN
      END ;
      Success := TRUE
   ELSIF SymIs('(')
   THEN
      IF Expression()
      THEN
         IF SymIs(')')
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSIF SymIs('NOT')
   THEN
      IF Factor()
      THEN
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Factor ;

PROCEDURE ActualParameters() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('(')
   THEN
      IF ExpList()
      THEN
      END ;
      IF SymIs(')')
      THEN
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ActualParameters ;

PROCEDURE Statement() : BOOLEAN ;
BEGIN
   IF Assignment()
   THEN
   ELSIF ProcedureCall()
   THEN
   ELSIF IfStatement()
   THEN
   ELSIF CaseStatement()
   THEN
   ELSIF WhileStatement()
   THEN
   ELSIF RepeatStatement()
   THEN
   ELSIF LoopStatement()
   THEN
   ELSIF ForStatement()
   THEN
   ELSIF WithStatement()
   THEN
   ELSIF SymIs('EXIT')
   THEN
   ELSIF SymIs('RETURN')
   THEN
      IF Expression()
      THEN
      END
   END ;
   RETURN( TRUE )
END Statement ;

PROCEDURE Assignment() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Designator()
   THEN
      IF SymIs(':=')
      THEN
         IF Expression()
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Assignment ;

PROCEDURE ProcedureCall() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Designator()
   THEN
      IF ActualParameters()
      THEN
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ProcedureCall ;

PROCEDURE StatementSequence() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Statement()
   THEN
      WHILE SymIs(';') DO
         IF Statement()
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END StatementSequence ;

PROCEDURE IfStatement() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('IF')
   THEN
      IF Expression()
      THEN
         IF SymIs('THEN')
         THEN
            IF StatementSequence()
            THEN
               WHILE SymIs('ELSIF') DO
                  IF Expression()
                  THEN
                     IF SymIs('THEN')
                     THEN
                        IF StatementSequence()
                        THEN
                        END
                     END
                  END
               END ;
               IF SymIs('ELSE')
               THEN
                  IF StatementSequence()
                  THEN
                  END
               END ;
               IF SymIs('END')
               THEN
                  Success := TRUE
               ELSE
                  Success := FALSE
               END
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END IfStatement ;

PROCEDURE CaseStatement() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('CASE')
   THEN
      IF Expression()
      THEN
         IF SymIs('OF')
         THEN
            IF Case()
            THEN
               WHILE SymIs('|') DO
                  IF Case()
                  THEN
                  END
               END ;
               IF SymIs('ELSE')
               THEN
                  IF StatementSequence()
                  THEN
                  END
               END ;
               IF SymIs('END')
               THEN
                  Success := TRUE
               ELSE
                  Success := FALSE
               END
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END CaseStatement ;

PROCEDURE Case() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF CaseLabelList()
   THEN
      IF SymIs(':')
      THEN
         IF StatementSequence()
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Case ;

PROCEDURE WhileStatement() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('WHILE')
   THEN
      IF Expression()
      THEN
         IF SymIs('DO')
         THEN
            IF StatementSequence()
            THEN
               IF SymIs('END')
               THEN
                  Success := TRUE
               ELSE
                  Success := FALSE
               END
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END WhileStatement ;

PROCEDURE RepeatStatement() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('REPEAT')
   THEN
      IF StatementSequence()
      THEN
         IF SymIs('UNTIL')
         THEN
            IF Expression()
            THEN
               Success := TRUE
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END RepeatStatement ;

PROCEDURE ForStatement() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('FOR')
   THEN
      IF Ident()
      THEN
         IF SymIs(':=')
         THEN
            IF Expression()
            THEN
               IF SymIs('TO')
               THEN
                  IF Expression()
                  THEN
                     IF SymIs('BY')
                     THEN
                        IF ConstExpression()
                        THEN
                        END
                     END ;
                     IF SymIs('DO')
                     THEN
                        IF StatementSequence()
                        THEN
                           IF SymIs('END')
                           THEN
                              Success := TRUE
                           ELSE
                              Success := FALSE
                           END
                        ELSE
                           Success := FALSE
                        END
                     ELSE
                        Success := FALSE
                     END
                  ELSE
                     Success := FALSE
                  END
               ELSE
                  Success := FALSE
               END
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ForStatement ;

PROCEDURE LoopStatement() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('LOOP')
   THEN
      IF StatementSequence()
      THEN
         IF SymIs('END')
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END LoopStatement ;

PROCEDURE WithStatement() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('WITH')
   THEN
      IF Designator()
      THEN
         IF SymIs('DO')
         THEN
            IF StatementSequence()
            THEN
               IF SymIs('END')
               THEN
                  Success := TRUE
               ELSE
                  Success := FALSE
               END
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END WithStatement ;

PROCEDURE ProcedureDeclaration() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF ProcedureHeading()
   THEN
      IF SymIs(';')
      THEN
         IF Block()
         THEN
            IF Ident()
            THEN
               Success := TRUE
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ProcedureDeclaration ;

PROCEDURE ProcedureHeading() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('PROCEDURE')
   THEN
      IF Ident()
      THEN
         IF FormalParameters()
         THEN
         END ;
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ProcedureHeading ;

PROCEDURE Block() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   WHILE Declaration() DO
   END ;
   IF SymIs('BEGIN')
   THEN
      IF StatementSequence()
      THEN
      END
   END ;
   IF SymIs('END')
   THEN
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Block ;

PROCEDURE Declaration() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('CONST')
   THEN
      WHILE ConstantDeclaration() DO
         IF SymIs(';')
         THEN
         END
      END ;
      Success := TRUE
   ELSIF SymIs('TYPE')
   THEN
      WHILE TypeDeclaration() DO
         IF SymIs(';')
         THEN
         END
      END ;
      Success := TRUE
   ELSIF SymIs('VAR')
   THEN
      WHILE VariableDeclaration() DO
         IF SymIs(';')
         THEN
         END
      END ;
      Success := TRUE
   ELSIF ProcedureDeclaration()
   THEN
      IF SymIs(';')
      THEN
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSIF ModuleDeclaration()
   THEN
      IF SymIs(';')
      THEN
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Declaration ;

PROCEDURE FormalParameters() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('(')
   THEN
      IF FPSection()
      THEN
         WHILE SymIs(';') DO
            IF FPSection()
            THEN
            END
         END
      END ;
      IF SymIs(')')
      THEN
         IF SymIs(':')
         THEN
            IF Qualident()
            THEN
            END
         END ;
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END FormalParameters ;

PROCEDURE FPSection() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('VAR')
   THEN
   END ;
   IF IdentList()
   THEN
      IF SymIs(':')
      THEN
         IF FormalType()
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END FPSection ;

PROCEDURE FormalType() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('ARRAY')
   THEN
      IF SymIs('OF')
      THEN
      END
   END ;
   IF Qualident()
   THEN
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END FormalType ;

PROCEDURE ModuleDeclaration() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('MODULE')
   THEN
      IF Ident()
      THEN
         IF Priority()
         THEN
         END ;
         IF SymIs(';')
         THEN
            WHILE Import() DO
            END ;
            IF Export()
            THEN
            END ;
            IF Block()
            THEN
               IF Ident()
               THEN
                  Success := TRUE
               ELSE
                  Success := FALSE
               END
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ModuleDeclaration ;

PROCEDURE Priority() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('[')
   THEN
      IF ConstExpression()
      THEN
         IF SymIs(']')
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Priority ;

PROCEDURE Export() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('EXPORT')
   THEN
      IF SymIs('QUALIFIED')
      THEN
      END ;
      IF IdentList()
      THEN
         IF SymIs(';')
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Export ;

PROCEDURE Import() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('FROM')
   THEN
      IF Ident()
      THEN
      END
   END ;
   IF SymIs('IMPORT')
   THEN
      IF IdentList()
      THEN
         IF SymIs(';')
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Import ;

PROCEDURE DefinitionModule() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('DEFINITION')
   THEN
      IF SymIs('MODULE')
      THEN
         IF Ident()
         THEN
            IF SymIs(';')
            THEN
               WHILE Import() DO
               END ;
               IF Export()
               THEN
               END ;
               WHILE Definition() DO
               END ;
               IF SymIs('END')
               THEN
                  IF Ident()
                  THEN
                     IF SymIs('.')
                     THEN
                        Success := TRUE
                     ELSE
                        Success := FALSE
                     END
                  ELSE
                     Success := FALSE
                  END
               ELSE
                  Success := FALSE
               END
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END DefinitionModule ;

PROCEDURE Definition() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('CONST')
   THEN
      WHILE ConstantDeclaration() DO
         IF SymIs(';')
         THEN
         END
      END ;
      Success := TRUE
   ELSIF SymIs('TYPE')
   THEN
      WHILE Ident() DO
         IF SymIs('=')
         THEN
            IF Type()
            THEN
            END
         END ;
         IF SymIs(';')
         THEN
         END
      END ;
      Success := TRUE
   ELSIF SymIs('VAR')
   THEN
      WHILE VariableDeclaration() DO
         IF SymIs(';')
         THEN
         END
      END ;
      Success := TRUE
   ELSIF ProcedureHeading()
   THEN
      IF SymIs(';')
      THEN
         Success := TRUE
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Definition ;

PROCEDURE ProgramModule() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF SymIs('MODULE')
   THEN
      IF Ident()
      THEN
         IF Priority()
         THEN
         END ;
         IF SymIs(';')
         THEN
            WHILE Import() DO
            END ;
            IF Block()
            THEN
               IF Ident()
               THEN
                  IF SymIs('.')
                  THEN
                     Success := TRUE
                  ELSE
                     Success := FALSE
                  END
               ELSE
                  Success := FALSE
               END
            ELSE
               Success := FALSE
            END
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END ProgramModule ;

PROCEDURE CompilationUnit() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF DefinitionModule()
   THEN
      Success := TRUE
   ELSE
      IF SymIs('IMPLEMENTATION')
      THEN
      END ;
      IF ProgramModule()
      THEN
         Success := TRUE
      ELSE
         Success := FALSE
      END
   END ;
   RETURN( Success )
END CompilationUnit ;

PROCEDURE StartParsing ;
VAR
   Name: ARRAY [0..20] OF CHAR ;
BEGIN
   IF GetArg(Name, 1)
   THEN
      OpenSource(Name) ;
      IF NOT CompilationUnit()
      THEN
         WriteError('Parser Failed')
      END ;
      CloseSourceFile
   END
END StartParsing ;


BEGIN
   StartParsing
END .
