IMPLEMENTATION MODULE M2AS ;


FROM SYSTEM IMPORT ADR ;

FROM M2AssemIO IMPORT OpenAssemblyFile, CloseAssemblyFile ;

FROM i486 IMPORT DWORDSIZE, WORDSIZE, BYTESIZE,
                 PentiumDataAlignmentSize,
                 GetScope, GetOffset, GetSize, GetUnitPrefix, GetUnitOfTransfer,
                 MaxDatum ;

FROM M2System IMPORT IsPseudoSystemFunction ;

FROM M2GDB IMPORT FileInfo, TypeInfo,
                  VariableInfo, BeginInfo, EndInfo, MainInfo ;

FROM M2Configure IMPORT ReturnArgumentOnStack, SystemIsXenix, AlignmentSize ;
FROM M2AsmUtil IMPORT WriteAsmName, UnderScoreString, DotifyLabel, GetModuleInitName ;
FROM M2FileName IMPORT CalculateFileName ;
FROM M2ALU IMPORT PopInt ;
FROM NameKey IMPORT WriteKey, GetKey ;

FROM SymbolTable IMPORT NulSym,
                        ModeOfAddr,
                        GetMainModule,
                        IsConstString, GetString, GetStringLength,
                        IsProcedure, IsVar, IsModule, IsExported,
                        GetSymName, GetNth, GetMode, GetType, GetVarFather,
                        PushValue, ForeachInnerModuleDo ;

FROM M2Debug IMPORT Assert ;
FROM M2Lexical IMPORT WriteError, NearToken, InternalError ;
FROM M2Error IMPORT BeginError, EndError ;
FROM StdIO IMPORT Write, PushOutput, PopOutput ;
FROM FIO IMPORT OpenToWrite, Close, WriteChar, IsNoError, File, WriteNBytes ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM NumberIO IMPORT WriteCard, CardToStr, WriteInt, WriteHex ;
FROM M2GDB IMPORT LineInfo ;

FROM M2Options IMPORT Optimizing, GenerateDebugging, Profiling,
                      WordLength, Alignment ;

FROM StrLib IMPORT StrLen, StrEqual, StrConCat, StrCopy ;
FROM M2Quads IMPORT QuadOperator, GetQuad, GetNextQuad ;


CONST
   MaxString                = 255 ;
   nul                      =  0C ;  (* normally imported from ASCII but p2c complains *)

VAR
   ProfileCount      : CARDINAL ; (* Gives a unique identifier for profile  *)
                                  (* count buckets.                         *)
   IsLineFeedOn      : BOOLEAN ;  (* automatic line feeds required?         *)


(* %%%FORWARD%%%
PROCEDURE CheckLineNumber (CurrentQuad: CARDINAL) ; FORWARD ;
PROCEDURE CheckProfiling ; FORWARD ;
PROCEDURE InstStr (Inst: Instruction; String: ARRAY OF CHAR) ; FORWARD ;
PROCEDURE WritePrefix (Size: DatumPrefix) ; FORWARD ;
PROCEDURE WriteScope (Scope: KindOfScope; o1, go: INTEGER; Sym: CARDINAL) ; FORWARD ;
PROCEDURE WritePositiveOffset (o: INTEGER) ; FORWARD ;
PROCEDURE WriteOffset (o: INTEGER) ; FORWARD ;
PROCEDURE WriteConstant (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteInstruction (Inst: Instruction; Unit: DatumPrefix) ; FORWARD ;
PROCEDURE WriteInstPrefix (Unit: DatumPrefix) ; FORWARD ;
PROCEDURE WriteRegister (Reg: Register; Unit: DatumPrefix) ; FORWARD ;
PROCEDURE GenerateQuadLabel (QuadNo: CARDINAL) ; FORWARD ;
PROCEDURE DeclareGlobalVariables (ModSym: CARDINAL) ; FORWARD ;
PROCEDURE DeclareLocalVariables (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteSize (Size: CARDINAL) ; FORWARD ;
PROCEDURE DisplayString (a: ARRAY OF CHAR; n: CARDINAL) ; FORWARD ;
   %%%FORWARD%%% *)


(*
   TurnLineFeedOn - automatically generates a line feed when an
                    instruction is generated.
*)

PROCEDURE TurnLineFeedOn ;
BEGIN
   IsLineFeedOn := TRUE
END TurnLineFeedOn ;


(*
   TurnLineFeedOff - do not automatically generate a line feed when
                     an instruction is generated.
*)

PROCEDURE TurnLineFeedOff ;
BEGIN
   IsLineFeedOn := FALSE
END TurnLineFeedOff ;


(*
   AutoWriteLn - generates a WriteLn if IsLineFeedOn
*)

PROCEDURE AutoWriteLn ;
BEGIN
   IF IsLineFeedOn
   THEN
      WriteLn
   END
END AutoWriteLn ;


(*
   CheckLineNumber - writes out the next line number if debugging
      	       	     has been enabled.
*)

PROCEDURE CheckLineNumber (CurrentQuad: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   QuadNo  : CARDINAL ;
BEGIN
   IF GenerateDebugging
   THEN
      QuadNo := CurrentQuad ;
      REPEAT
      	 GetQuad(QuadNo, Operator, Operand1, Operand2, Operand3) ;
      	 QuadNo := GetNextQuad(QuadNo)
      UNTIL (Operator=LineNumberOp) OR (Operator=KillLocalVarOp) OR (Operator=EndOp) ;
      IF Operator=LineNumberOp
      THEN
      	 LineInfo(Operand3)            (* Generate GDB information *)
      END
   END
END CheckLineNumber ;


(*
   CheckProfiling - checks to see whether profiling is turned on
                    and if so it emits a call to _mcount.
*)

PROCEDURE CheckProfiling ;
BEGIN
   IF Profiling
   THEN
      WriteString('.data') ; WriteLn ;
      WriteString('     .align ') ; WriteCard(AlignmentSize, 0) ; WriteLn ;
      WriteString('LP') ; WriteCard(ProfileCount, 0) ; Write(':') ; WriteLn ;
      WriteString('     .long 0') ; WriteLn ;
      WriteString('.text') ; WriteLn ;
      WriteInstruction(mov, MaxDatum()) ;
      WriteString(' $LP') ; WriteCard(ProfileCount, 0) ;
      Write(',') ; WriteRegister(RegEDX, MaxDatum()) ; WriteLn ;
      WriteString('    call ') ; UnderScoreString('mcount') ; WriteLn ;
      INC(ProfileCount)
   END
END CheckProfiling ;


(*
   InstRegConstant - instruction constant and memory.  (Reg <- inst <- Constant)
                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

                     Example:

                     mov   12, Reg
                     add   12, Reg
*)

PROCEDURE InstRegConstant (Inst: Instruction;
                           Unit: DatumPrefix;
                           Reg : Register;
                           Sym : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteConstant(Sym) ; WriteString(', ') ;
   WriteRegister(Reg, Unit) ;
   AutoWriteLn
END InstRegConstant ;


(*
   InstRegConstant - instruction register and constant.
      	       	     	   (cmp Constant, Reg)

                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

                     Example:

                     mov   Reg, 12
                     add   Reg, 12
*)

PROCEDURE InstConstantReg (Inst: Instruction;
                           Unit: DatumPrefix;
                           Sym : CARDINAL;
                           Reg : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString('  ') ;
   WriteRegister(Reg, Unit) ;
   WriteString(', $') ;
   WriteConstant(Sym) ;
   AutoWriteLn
END InstConstantReg ;


(*
   InstMDConstant - instruction constant and memory.  (Mem <- inst <- Constant)
                    Inst   - determines the instruction.
                    Unit   - determines the DatumPrefix of transfer.
                    o1, go - determine the offset into element1.
                    Scope  - determines the scope of element1.
                    Sym1   - element1 where the constant is to be placed.
                             Symbol requiring the memory reference.
                    Sym2   - constant element2.

                    Example:

                    mov   12, (ebp+15)
*)

PROCEDURE InstMDConstant (Inst  : Instruction;
                          Unit  : DatumPrefix;
                          o1, go: INTEGER ;
                          Scope : KindOfScope;
                          Sym1,
                          Sym2  : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteConstant(Sym2) ;
   WriteString(' ,') ;
   WriteScope(Scope, o1, go, Sym1) ;
   AutoWriteLn
END InstMDConstant ;


(*
   InstMRIConstant - instruction constant and memory indexed by a register.
                     (Mem[Reg1] <- inst <- Constant)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     o1, go - determine the offset into element1.
                     Sym2   - constant element2.

                     Example:

                     mov   12, (edi+15)
*)

PROCEDURE InstMRIConstant (Inst  : Instruction;
                           Unit  : DatumPrefix;
                           o1, go: INTEGER ;
                           Reg1  : Register;
                           Sym2  : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteConstant(Sym2) ;
   WriteString(', ') ;
   WritePositiveOffset(o1) ;
   WriteOffset(go) ;
   Write('(') ;
   WriteRegister(Reg1, MaxDatum()) ;
   Write(')') ;
   AutoWriteLn
END InstMRIConstant ;


(*
   InstMRISymConstant - instruction constant and memory indexed by a register.
                        (Mem[Reg1] <- inst <- Constant)
                        Inst   - determines the instruction.
                        Unit   - determines the DatumPrefix of transfer.
                        Sym1   - global variable.
                        go     - determine the offset into element1.
                        Sym2   - constant element2.

                        Example:

                        mov   12, myglobal+3(edi)
*)

PROCEDURE InstMRISymConstant (Inst  : Instruction;
                              Unit  : DatumPrefix;
                              Sym1  : CARDINAL;
                              go    : INTEGER ;
                              Reg1  : Register;
                              Sym2  : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteConstant(Sym2) ;
   WriteString(', ') ;
   WriteScope(Global, 0, go, Sym1) ;
   Write('(') ;
   WriteRegister(Reg1, MaxDatum()) ;
   Write(')') ;
   AutoWriteLn
END InstMRISymConstant ;


(*
   InstMRICardinal - instruction constant and memory indexed by a register.
                     (Mem[Reg1] <- inst <- Constant)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     o1, go - determine the offset into element1.
                     Card   - cardinal constant.

                     Example:

                     mov   12H, (edi+15)
*)

PROCEDURE InstMRICardinal (Inst  : Instruction;
                           Unit  : DatumPrefix;
                           o1, go: INTEGER ;
                           Reg1  : Register;
                           Card  : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteString('$0x') ; WriteHex(Card, 0) ;
   WriteString(', ') ;
   WritePositiveOffset(o1) ;
   WriteOffset(go) ;
   Write('(') ;
   WriteRegister(Reg1, MaxDatum()) ;
   Write(')') ;
   AutoWriteLn
END InstMRICardinal ;


(*
   InstMRIReg - instruction constant and memory indexed by a register.
                (Mem[Reg1] <- inst <- Reg2)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                o1, go - determine the offset into element1.
                Reg1   - indexes memory.
                Reg2   - source of data.

                Example:

                mov   eax, (edi+15)
*)

PROCEDURE InstMRIReg (Inst  : Instruction;
                      Unit  : DatumPrefix;
                      o1, go: INTEGER ;
                      Reg1  ,
                      Reg2  : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteRegister(Reg2, Unit) ;
   WriteString(', ') ;
   WritePositiveOffset(o1) ;
   WriteOffset(go) ;
   Write('(') ;
   WriteRegister(Reg1, MaxDatum()) ;
   Write(')') ;
   AutoWriteLn
END InstMRIReg ;


(*
   InstMDReg - inst reg into memory. (Mem1 <- inst <- Reg)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o1, go - determine the offset into element1.
               Reg    - determine the register.
               Scope1 - determines the scope of element1.
               Sym    - symbol requiring the memory reference.

               Example:

               mov   eax, (ebp+15)
*)

PROCEDURE InstMDReg (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     o1, go: INTEGER ;
                     Reg   : Register;
                     Scope1: KindOfScope;
                     Sym   : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteRegister(Reg, Unit) ;
   WriteString(', ') ;
   WriteScope(Scope1, o1, go, Sym) ;
   AutoWriteLn
END InstMDReg ;


(*
   InstRegMD - inst memory into reg. (Reg <- inst <- Mem1)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o2, go - determine the offset into element2.
               Reg    - determine the register.
               Scope2 - determines the scope of element2.
               Sym    - symbol requiring the memory reference.

               Example:

               ; inst  Mem{scope}+o1+go, reg
               mov   (ebp+15), eax
*)

PROCEDURE InstRegMD (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg   : Register;
                     o1, go: INTEGER ;
                     Scope1: KindOfScope;
                     Sym   : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteScope(Scope1, o1, go, Sym) ;
   WriteString(', ') ;
   WriteRegister(Reg, Unit) ;
   AutoWriteLn
END InstRegMD ;


(*
   InstRegMIO - inst memory into reg. (Reg1 <- inst <- Mem[Reg2])
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg1   - determine the destination register.
                Reg2   - determine the index register.
                Offset - offset to be added to the index register.

               Example:

               mov   (edi+4), eax
               ; mov   (reg2+offset), reg1
*)

PROCEDURE InstRegMIO (Inst  : Instruction;
                      Unit  : DatumPrefix;
                      Reg1,
                      Reg2  : Register;
                      Offset: INTEGER) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WritePositiveOffset(Offset) ;
   Write('(') ;
   WriteRegister(Reg2, MaxDatum()) ;
   WriteString('), ') ;
   WriteRegister(Reg1, Unit) ;
   AutoWriteLn
END InstRegMIO ;


(*
   InstRegMISym - inst memory into reg. (Reg1 <- inst <- global[Reg2])
                  Inst   - determines the instruction.
                  Unit   - determines the DatumPrefix of transfer.
                  Reg1   - determine the destination register.
                  Reg2   - determine the index register.
                  Offset - offset to be added to the index register.
                  Sym    - the global variable.

                  Example:

                  mov   myglobal+4(edi), eax
                ; mov   Sym+offset(reg2), reg1
*)

PROCEDURE InstRegMISym (Inst  : Instruction;
                        Unit  : DatumPrefix;
                        Reg1,
                        Reg2  : Register;
                        Offset: INTEGER;
                        Sym   : CARDINAL) ;
BEGIN
   Assert(NOT IsProcedure(GetVarFather(Sym))) ;
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteScope(Global, 0, Offset, Sym) ;
   Write('(') ;
   WriteRegister(Reg2, MaxDatum()) ;
   WriteString('), ') ;
   WriteRegister(Reg1, Unit) ;
   AutoWriteLn
END InstRegMISym ;


(*
   InstRegMI - inst memory into reg. (Reg1 <- inst <- Mem[Reg2])
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               Reg1   - determine the destination register.
               Reg2   - determine the index register.

               Example:

               mov   (edi), eax
               ; mov   (reg2), reg1
*)

PROCEDURE InstRegMI (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg1,
                     Reg2  : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' (') ;
   WriteRegister(Reg2, MaxDatum()) ;
   WriteString('), ') ;
   WriteRegister(Reg1, Unit) ;
   AutoWriteLn
END InstRegMI ;


(*
   InstMIOReg - inst memory into reg. (Mem[Reg1] <- inst <- Reg2)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg1   - determine the index register.
                Offset - determines the offset to be added to the index
                         register.
                Reg2   - determine the source register.

               Example:

               mov   eax, (edi)
               ; mov   reg2, (reg1)
*)

PROCEDURE InstMIOReg (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg1  : Register;
                     Offset: INTEGER;
                     Reg2  : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteRegister(Reg2, Unit) ;
   WriteString(', ') ;
   WritePositiveOffset(Offset) ;
   Write('(') ;
   WriteRegister(Reg1, MaxDatum()) ;
   Write(')') ;
   AutoWriteLn
END InstMIOReg ;


(*
   InstMISymReg - inst reg into memory. (global[Reg2] <- inst <- Reg1)
                  Inst   - determines the instruction.
                  Unit   - determines the DatumPrefix of transfer.
                  Reg1   - determine the destination register.
                  Reg2   - determine the index register.
                  Offset - offset to be added to the index register.
                  Sym    - the global variable.

                  Example:

                  mov   %eax, myglobal+4(edi),
                ; mov   reg1, Sym+offset(reg2)
*)

PROCEDURE InstMISymReg (Inst  : Instruction;
                        Unit  : DatumPrefix;
                        Reg1,
                        Reg2  : Register;
                        Offset: INTEGER;
                        Sym   : CARDINAL) ;
BEGIN
   Assert(NOT IsProcedure(GetVarFather(Sym))) ;
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteRegister(Reg1, Unit) ;
   WriteString(', ') ;
   WriteScope(Global, 0, Offset, Sym) ;
   Write('(') ;
   WriteRegister(Reg2, MaxDatum()) ;
   Write(')') ;
   AutoWriteLn
END InstMISymReg ;


(*
   InstMIReg - inst memory into reg. (Mem[Reg1] <- inst <- Reg2)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               Reg1   - determine the index register.
               Reg2   - determine the source register.

               Example:

               mov   eax, (edi)
               ; mov   reg2, (reg1)
*)

PROCEDURE InstMIReg (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg1,
                     Reg2  : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteRegister(Reg2, Unit) ;
   WriteString(', (') ;
   WriteRegister(Reg1, MaxDatum()) ;
   Write(')') ;
   AutoWriteLn
END InstMIReg ;


(*
   InstRegReg - inst reg into reg. (Reg1 <- inst <- Reg2)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg1   - determine the destination register.
                Reg2   - determine the index register.

                Example:

                mov   ebp, eax
               ; mov   reg2, reg1
*)

PROCEDURE InstRegReg (Inst: Instruction; Unit: DatumPrefix;
                      Reg1, Reg2: Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   IF Inst=sal
   THEN
      (* only makes sense to use an 8 bit operand with sal *)
      WriteRegister(Reg2, PByte)
   ELSE
      WriteRegister(Reg2, Unit)
   END ;
   WriteString(', ') ;
   WriteRegister(Reg1, Unit) ;
   AutoWriteLn
END InstRegReg ;


(*
   InstRegInt - inst Reg with an Int. (Reg <- inst <- Int)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg    - determines the destination register.
                Int    - the integer constant as the second operand.

                Example:

                mov   15, eax
               ; mov   Int, Reg
*)

PROCEDURE InstRegInt (Inst: Instruction; Unit: DatumPrefix;
                      Reg: Register; Int: INTEGER) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString('  $') ;
   WriteInt(Int, 0) ;
   WriteString(', ') ;
   WriteRegister(Reg, Unit) ;
   AutoWriteLn
END InstRegInt ;


(*
   InstRegHexCard - inst Reg with an cardinal. (Reg <- inst <- Card)
                    Inst   - determines the instruction.
                    Unit   - determines the DatumPrefix of transfer.
                    Reg    - determines the destination register.
                    Card   - the cardinal constant as the second operand.

                    Example:

                    and   $0FFFF0000H, eax
                   ; mov   Int, Reg
*)

PROCEDURE InstRegHexCard (Inst: Instruction; Unit: DatumPrefix;
                          Reg: Register; Card: CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $0x') ;
   WriteHex(Card, 0) ;
   WriteString(', ') ;
   WriteRegister(Reg, Unit) ;
   AutoWriteLn
END InstRegHexCard ;


(*
   InstMDOffset - inst Mem with an offset Int. (Mem[Sym1] <- inst <- OFFSET Sym2)
                  Offset is the address of symbol Sym2.
                  Inst   - determines the instruction.
                  Unit   - determines the DatumPrefix of transfer.
                  Sym1   - indicates the memory location which will
                           contain the address of Sym2.
                  Sym2   - symbol requiring the memory reference.

                  Example:

                  mov   $_varname, (ebp-15)
                  ; mov  $_varname, {unit prefix} {scope}
*)

PROCEDURE InstMDOffset (Inst: Instruction; Unit: DatumPrefix;
                        Sym1, Sym2: CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteAsmName(Sym2) ;
   WriteString(', ') ;
   WriteScope(GetScope(Sym1), GetOffset(Sym1), 0, Sym1) ;
   AutoWriteLn
END InstMDOffset ;


(*
   InstRegOffset - inst reg with an offset Int. (Reg <- inst <- OFFSET Sym)
                   Offset is the address of symbol Sym.
                   Inst   - determines the instruction.
                   Unit   - determines the DatumPrefix of transfer.
                   Reg    - register which contain the address of Sym2.
                   Sym    - symbol requiring the memory reference.

                   Example:

                   mov   $_varname, $eax
                  ; mov  $_varname, {unit prefix} {reg}
*)

PROCEDURE InstRegOffset (Inst: Instruction; Unit: DatumPrefix;
                         Reg: Register; Sym: CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteAsmName(Sym) ;
   WriteString(', ') ;
   WriteRegister(Reg, Unit) ;
   AutoWriteLn
END InstRegOffset ;


(*
   InstMDInt - inst Mem with an Int. (Mem[Sym1] <- inst <- Int)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o1, go - determine the offset into element1.
               Int    - the integer constant as the second operand.
               Scope  - determines the scope of element1.
               Sym    - symbol requiring the memory reference.

               Example:

               mov   123, (ebp-15)
               ; mov   int, {unit prefix} {scope}
*)

PROCEDURE InstMDInt (Inst: Instruction; Unit: DatumPrefix;
                     o1, go: INTEGER; Int: INTEGER;
                     Scope: KindOfScope; Sym: CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteInt(Int, 0) ;
   WriteString(', ') ;
   WriteScope(Scope, o1, go, Sym) ;
   AutoWriteLn
END InstMDInt ;


(*
   InstStr - inst followed by a string:
             inst String
             Inst   - determines the instruction.
             String - string that follows the Inst

             Example:

             Inst  String
             ; call  String
*)

PROCEDURE InstStr (Inst: Instruction; String: ARRAY OF CHAR) ;
BEGIN
   WriteInstruction(Inst, MaxDatum()) ;
   WriteString('  ') ;
   WriteString(String) ;
   AutoWriteLn
END InstStr ;


(*
   BinInst - write a sole inst.
             Inst   - determines the instruction.

             Example:

             ret
*)

PROCEDURE BinInst (Inst: Instruction) ;
BEGIN
   IF Inst=repmovs
   THEN
      InternalError('should use BinInstUnit - as you must supply the unit of move with repmovs', __FILE__, __LINE__)
   END ;
   WriteInstruction(Inst, MaxDatum()) ;
   AutoWriteLn
END BinInst ;


(*
   BinInstUnit - write a sole inst.
                 Inst   - determines the instruction.

                 Example:

                 rep
                 movsl %esp:$0, %esp:$0
*)

PROCEDURE BinInstUnit (Inst: Instruction; Unit: DatumPrefix; Length: CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   AutoWriteLn
END BinInstUnit ;


(*
   BinInstReg - inst operating on a reg. (inst <-> Reg)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg    - determine the operand register.

                Example:

                push   eax
*)

PROCEDURE BinInstReg (Inst: Instruction;
                      Unit: DatumPrefix;
                      Reg : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteRegister(Reg, Unit) ;
   AutoWriteLn
END BinInstReg ;


(*
   BinInstRegI - inst operating on a reg. ( inst <-> (Reg) )
                 Inst   - determines the instruction.
                 Unit   - determines the DatumPrefix of transfer.
                 Reg    - determine the operand register.

                 Example:

                 call   (eax)
*)

PROCEDURE BinInstRegI (Inst: Instruction;
                       Unit: DatumPrefix;
                       Reg : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' *') ;
   WriteRegister(Reg, Unit) ;
   AutoWriteLn
END BinInstRegI ;


(*
   BinInstConstant - inst operating on a reg. (inst <-> Reg)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     Sym    - determine the constant.

                     Example:

                     push   15
*)

PROCEDURE BinInstConstant (Inst: Instruction;
                           Unit: DatumPrefix;
                           Sym : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteConstant(Sym) ;
   AutoWriteLn
END BinInstConstant ;


(*
   BinInstMD - inst operating on a memory location. (inst <-> Mem)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o1, go - determine the operand memory location.
               Scope1 - determines the scope of the operand.
               Sym    - symbol requiring the memory reference.

               Example:

               push   (ebp+15)
*)

PROCEDURE BinInstMD (Inst: Instruction; Unit: DatumPrefix;
                     o1, go  : INTEGER; Scope1: KindOfScope; Sym: CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteScope(Scope1, o1, go, Sym) ;
   AutoWriteLn
END BinInstMD ;


(*
   BinInstHex - inst operating on a number. (inst <-> hex)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                c      - number.

                Example:

                push   0x1234
*)

PROCEDURE BinInstHex (Inst: Instruction;
                      Unit: DatumPrefix;
                      c   : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   IF (Unit=PByte) AND (Inst=push)
   THEN
      WriteError('cannot push a byte on the 8088, 8086, 80386 or 80486')
   END ;
   WriteString(' $0x') ;
   WriteHex(c, 0) ; AutoWriteLn ;
END BinInstHex ;


(*
   Comment - places string, a, into the assembly file.
*)

PROCEDURE Comment (a: ARRAY OF CHAR) ;
BEGIN
   WriteString('# ') ; WriteString(a) ; WriteLn
END Comment ;


(*
   WritePrefix - writes the prefix associated with transfering Size
                 bytes of data.
*)

PROCEDURE WritePrefix (Size: DatumPrefix) ;
BEGIN
   CASE Size OF

   PByte :  WriteString('BYTE PTR') |
   PWord :  WriteString('WORD PTR') |
   PDWord:  WriteString('DWORD PTR')

   ELSE
      WriteError('Expecting PByte, PWord, PDWord')
   END
END WritePrefix ;


(*
   WriteScope - writes the referencing information to a particular scope.
                Scope is either Global or Local.
                o1 is the offset of the element and go is the Delta
                offset. It is useful to contain both as separate
                entities as it makes human debugging easier!
                Sym - is the variable requiring memory access.
                      If global then the name Sym is written
                      otherwise the activation record offset in the form of
                      o1+go is written.
*)

PROCEDURE WriteScope (Scope: KindOfScope; o1, go: INTEGER; Sym: CARDINAL) ;
VAR
   Proc : CARDINAL ;
BEGIN
   IF Scope=Local
   THEN
      WritePositiveOffset(o1) ; WriteOffset(go) ;
      WriteString('(%ebp)')
   ELSE
      WriteAsmName(Sym) ; WriteOffset(go)
   END
END WriteScope ;


(*
   WritePositiveOffset - writes an integer offset. It only has a sign if
                         the offset is negative.
*)

PROCEDURE WritePositiveOffset (o: INTEGER) ;
BEGIN
   WriteInt(o, 0)
END WritePositiveOffset ;


(*
   WriteOffset - writes an integer offset with a sign beforehand.
*)

PROCEDURE WriteOffset (o: INTEGER) ;
BEGIN
   IF o#0
   THEN
      IF o>=0
      THEN
         Write('+')
      END ;
      WriteInt(o, 0)
   END
END WriteOffset ;


(*
   WriteConstant - writes a constant from a symbol.
                   writes out the following consts:

                   (1)  Integer, Cardinal value constants.
                   (2)  Character constants.
                   (3)  String constants.
*)

PROCEDURE WriteConstant (Sym: CARDINAL) ;
VAR
   Size: CARDINAL ;
   Str : ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IF IsConstString(Sym)
   THEN
      Size := GetStringLength(Sym) ;
      GetKey(GetString(Sym), Str) ;
      IF Size=1
      THEN
         Write("'") ;
         IF Str[0]='\'   (* \ must be escaped by \ *)
         THEN
            Write('\')
         END ;
         Write(Str[0])
      ELSIF Size=0
      THEN
         WriteString(' 0x00') ;  (* Empty string *)
      ELSE
         WriteString(Str)
      END
   ELSE
      PushValue(Sym) ;
      WriteInt(PopInt(), 0)
   END
END WriteConstant ;


(*
   WriteInstruction - writes an instruction.
*)

PROCEDURE WriteInstruction (Inst: Instruction; Unit: DatumPrefix) ;
BEGIN
   WriteString('    ') ;
   CASE Inst OF

   mov    : WriteString('mov') ; WriteInstPrefix(Unit) |
   add    : WriteString('add') ; WriteInstPrefix(Unit) |
   sub    : WriteString('sub') ; WriteInstPrefix(Unit) |
   mult   : WriteString('imul') ; WriteInstPrefix(Unit) |
   idiv   : WriteString('idiv') ; WriteInstPrefix(Unit) |
   div    : WriteString('div') ; WriteInstPrefix(Unit) |
   cdq    : WriteString('cdq') | (* no prefix must be given for GAS *)
   mod    : WriteString('mod') ; WriteInstPrefix(Unit) |
   neg    : WriteString('neg') ; WriteInstPrefix(Unit) |
   push   : WriteString('push') ; WriteInstPrefix(Unit) |
   pop    : WriteString('pop') ; WriteInstPrefix(Unit) |
   jmp    : WriteString('jmp') |
   je     : WriteString('je') |
   jne    : WriteString('jne') |
   jl     : WriteString('jl') |
   jg     : WriteString('jg') |
   jle    : WriteString('jle') |
   jge    : WriteString('jge') |
   jae    : WriteString('jae') |
   cmp    : WriteString('cmp') ; WriteInstPrefix(Unit) |
   not    : WriteString('not') ; WriteInstPrefix(Unit) |
   repmovs: WriteString('rep') ; WriteLn ;
            WriteString('    movs') ; WriteInstPrefix(Unit) ; WriteString(' %esp:$0, %esp:$0') |
   and    : WriteString('and') ; WriteInstPrefix(Unit) |
   or     : WriteString('or') ; WriteInstPrefix(Unit) |
   xor    : WriteString('xor') ; WriteInstPrefix(Unit) |
   test   : WriteString('test') ; WriteInstPrefix(Unit) |
   sal    : WriteString('sal') ; WriteInstPrefix(Unit) |
   sar    : WriteString('sar') ; WriteInstPrefix(Unit) |
   shr    : WriteString('shr') ; WriteInstPrefix(Unit) |
   shl    : WriteString('shl') ; WriteInstPrefix(Unit) |
   ret    : WriteString('ret') |
   call   : WriteString('call') |
   leave  : WriteString('leave') |
   inc    : WriteString('inc') |
   dec    : WriteString('dec') |
   nop    : WriteString('nop') |
   lea    : WriteString('lea')

   ELSE
      WriteError('Instruction not known - expand case clause')
   END
END WriteInstruction ;


(*
   WriteInstPrefix - writes the instruction prefix, which depends
                     upon the Unit of data to be processed by this
                     instruction.
*)

PROCEDURE WriteInstPrefix (Unit: DatumPrefix) ;
BEGIN
   CASE Unit OF

   PByte : Write('b') |
   PWord : Write('w') |
   PDWord: Write('l')

   ELSE
      WriteError('Unrecognised prefix value in case clause')
   END
END WriteInstPrefix ;


(*
   WriteRegister - writes a register.
*)

PROCEDURE WriteRegister (Reg: Register; Unit: DatumPrefix) ;
BEGIN
   Write('%') ;
   CASE Unit OF

   PDWord: CASE Reg OF

           RegEAX: WriteString('eax') |
           RegEBX: WriteString('ebx') |
           RegECX: WriteString('ecx') |
           RegEDX: WriteString('edx') |
           RegEBP: WriteString('ebp') |
           RegESP: WriteString('esp') |
           RegESI: WriteString('esi') |
           RegEDI: WriteString('edi')

           ELSE
              InternalError('register not known - expand case clause', __FILE__, __LINE__)
           END |

   PWord : CASE Reg OF

           RegEAX: WriteString('ax') |
           RegEBX: WriteString('bx') |
           RegECX: WriteString('cx') |
           RegEDX: WriteString('dx') |
           RegEBP: WriteString('bp') |
           RegESP: WriteString('sp') |
           RegESI: WriteString('si') |
           RegEDI: WriteString('di')

           ELSE
              InternalError('register not known - expand case clause', __FILE__, __LINE__)
           END |

   PByte : CASE Reg OF

           RegEAX: WriteString('al') |
           RegEBX: WriteString('bl') |
           RegECX: WriteString('cl') |
           RegEDX: WriteString('dl') |
           RegEBP: InternalError('cannot get byte access ebp', __FILE__, __LINE__) |
           RegESP: InternalError('cannot get byte access esp', __FILE__, __LINE__) |
           RegESI: InternalError('cannot get byte access esi', __FILE__, __LINE__) |
           RegEDI: InternalError('cannot get byte access edi', __FILE__, __LINE__)

           ELSE
              WriteError('Register not known - expand case clause')
           END

   ELSE
      InternalError('DatumPrefix value not known', __FILE__, __LINE__)
   END
END WriteRegister ;


(*
   GenerateQuadLabel - writes a label refering to quad, QuadNo.
*)

PROCEDURE GenerateQuadLabel (QuadNo: CARDINAL) ;
VAR
   Str: ARRAY [0..20] OF CHAR ;
BEGIN
   CardToStr(QuadNo, 0, Str) ;
   StrConCat('L', Str, Str) ;
   StrConCat(Str, ':', Str) ;
   DotifyLabel(Str) ;
   WriteString(Str) ; WriteLn
END GenerateQuadLabel ;


(*
   OpenAssemblyFileAndInitialize - displays the initialization code.
*)

PROCEDURE OpenAssemblyFileAndInitialize ;
VAR
   ModuleName,
   FileName  : ARRAY [0..80] OF CHAR ;
BEGIN
   OpenAssemblyFile(GetSymName(GetMainModule())) ;
   WriteLn ;
   WriteLn ;
   Comment('       M-2 -> assembly language') ;
   Comment('');
   WriteLn ;
   WriteString('        .file   "') ;
   GetKey(GetSymName(GetMainModule()), ModuleName) ;
   CalculateFileName(ModuleName, 'mod', FileName) ;
   WriteString(FileName) ;
   Write('"') ; WriteLn ;
   IF SystemIsXenix
   THEN
      WriteString('        .stabs "__fltused",0,0,0,0') ; WriteLn ;
   END ;
   WriteString('gcc2_compiled.:') ; WriteLn ;
   FileInfo(GetMainModule()) ;
   WriteLn ;
   IF Alignment
   THEN
      WriteString('           .bss') ; WriteLn ;
      WriteString('           .align ') ; WriteCard(PentiumDataAlignmentSize, 0) ; WriteLn
   END ;

   TypeInfo ;  (* Display the GDB info for the types used in this MODULE *)
   WriteString('.text') ; WriteLn ;
   CodeAlignment ;
(*
   ForeachExportedDo(GetMainModule(), DisplayExportedProcedure) ;
   WriteString('PUBLIC  ') ; WriteString('__M2_') ;
   WriteKey(GetSymName(GetMainModule())) ; WriteString('_init') ; WriteLn ;
*)
   DeclareGlobalVariables(GetMainModule()) ;
(*
   ForeachImportedDo(GetMainModule(), DisplayImportedVar) ;
   ForeachProcedureDo(GetMainModule(), DisplayProcedureName) ;
   ForeachImportedDo(GetMainModule(), DisplayImportedProcedure) ;
   IF GetSymName(GetMainModule())#MakeKey('M2RTS')
   THEN
      WriteString('EXTRN    _M2RTS_HALT:NEAR') ; WriteLn ;
      WriteString('EXTRN    _M2RTS_Terminate:NEAR') ; WriteLn ;
   END ;
*)
   WriteLn ;
END OpenAssemblyFileAndInitialize ;


(*
   DisplayImportedProcedure - displays a symbol, Sym, (if a procedure)
                              in a MASM imported form.
*)

PROCEDURE DisplayImportedProcedure (Sym: CARDINAL) ;
BEGIN
   IF IsProcedure(Sym) AND (NOT IsPseudoSystemFunction(Sym))
   THEN
      WriteString('EXTRN   ') ; WriteAsmName(Sym) ; WriteString(':NEAR') ;
      WriteLn
   END
END DisplayImportedProcedure ;


(*
   DisplayExportedProcedure - displays a symbol, Sym, (if a procedure)
                              in a MASM exported form.
*)

PROCEDURE DisplayExportedProcedure (Sym: CARDINAL) ;
BEGIN
   IF IsProcedure(Sym) AND (NOT IsPseudoSystemFunction(Sym))
   THEN
      WriteString('PUBLIC   ') ; WriteAsmName(Sym) ;
      WriteLn
   END
END DisplayExportedProcedure ;


(*
   DisplayProcedureName - displays the procedure name in terms of byte strings.
*)

PROCEDURE DisplayProcedureName (Sym: CARDINAL) ;
BEGIN
   IF IsProcedure(Sym)
   THEN
      WriteString(';;;   ALIGN 4') ; WriteLn ;
      WriteString(';;;$S') ; WriteCard(Sym, 0) ;
      WriteString("   DB   '") ;
      WriteAsmName(Sym) ;
      WriteString("', 0aH, 00H") ; WriteLn
   END
END DisplayProcedureName ;


(*
   DisplayExportedVar - displays a symbol, Sym, (if a variable)
                        in a GDB exported form.
*)

PROCEDURE DisplayExportedVar (Sym: CARDINAL) ;
BEGIN
   IF IsVar(Sym)
   THEN
      WriteString('GLOBAL   ') ; WriteAsmName(Sym) ; WriteLn
   END
END DisplayExportedVar ;


(*
   DisplayImportedVar - displays a symbol, Sym, (if a variable)
                        in a MASM exported form.
*)

PROCEDURE DisplayImportedVar (Sym: CARDINAL) ;
BEGIN
   IF IsVar(Sym)
   THEN
      WriteString('EXTRN   ') ; WriteAsmName(Sym) ;
      WriteString(':NEAR') ; (* WriteSize(GetSize(Sym)) ; *)
      WriteLn
   END
END DisplayImportedVar ;


(*
   WritePrologue - writes the prologue code before the
                   initialization code of a module, ModSym.
*)

PROCEDURE WritePrologue (ModSym: CARDINAL;
                         CurrentQuad: CARDINAL; Line: CARDINAL) ;
BEGIN
   WriteString('.text') ; WriteLn ;
   MainInfo(ModSym, Line) ;
   CodeAlignment ;
   WriteKey(GetModuleInitName(ModSym)) ; Write(':') ; WriteLn ;
   IF GetMainModule()=ModSym
   THEN
      (* main module exports its initialization code *)
      WriteString('.globl  ') ;
      WriteKey(GetModuleInitName(ModSym)) ; WriteLn
   END ;
   LineInfo(Line) ;
   CheckLineNumber(CurrentQuad) ;
   BinInstReg(push, MaxDatum(), RegEBP) ;
   InstRegReg(mov, MaxDatum(), RegEBP, RegESP) ;
   CheckProfiling ;
   BinInstReg(push, MaxDatum(), RegEBX) ;
   BinInstReg(push, MaxDatum(), RegEDI) ;
   BinInstReg(push, MaxDatum(), RegESI) ;
   BeginInfo
END WritePrologue ;


(*
   WriteEpilogue - writes the termination code after the initialization
                   of a module.
*)

PROCEDURE WriteEpilogue (ModSym: CARDINAL) ;
BEGIN
   EndInfo ;
   IF IsModule(GetMainModule()) AND (ModSym=GetMainModule())
   THEN
      Comment(' clean up using Modula-2 M2RTS termination') ;
      WriteInstruction(call, MaxDatum()) ;
      WriteString('  ') ;
      UnderScoreString('M2RTS_Terminate') ;
      WriteLn
   END ;
   BinInstReg(pop, MaxDatum(), RegESI) ; 
   BinInstReg(pop, MaxDatum(), RegEDI) ; 
   BinInstReg(pop, MaxDatum(), RegEBX) ;
   BinInst(leave) ;
   BinInst(ret)
END WriteEpilogue ;


(*
   CheckAlignment - checks the alignment of data to ensure that we match the offset calculated
                    inside M2EvalSym.
*)

PROCEDURE CheckAlignment (Son: CARDINAL; RunningOffset: CARDINAL) ;
VAR
   offset: CARDINAL ;
BEGIN
   offset := GetOffset(Son) ;
   IF Alignment AND (offset#RunningOffset)
   THEN
      (* offset does not equal RunningOffset, thus alignment is required *)
      IF (offset MOD PentiumDataAlignmentSize=0) AND
         ((RunningOffset+PentiumDataAlignmentSize) DIV PentiumDataAlignmentSize=(offset DIV PentiumDataAlignmentSize))
      THEN
         WriteString('           .align ') ; WriteCard(PentiumDataAlignmentSize, 0) ; WriteLn
      ELSIF (offset MOD DWORDSIZE=0) AND
         ((RunningOffset+DWORDSIZE) DIV DWORDSIZE=(offset DIV DWORDSIZE))
      THEN
         WriteString('           .align ') ; WriteCard(DWORDSIZE, 0) ; WriteLn
      ELSE
         Assert(offset=RunningOffset)
      END
   ELSE
      Assert(offset=RunningOffset)
   END
END CheckAlignment ;


(*
   CodeAlignment - emits a code alignment assembly statement.
*)

PROCEDURE CodeAlignment ;
BEGIN
   IF Alignment
   THEN
      WriteString('     .align 32') ; (* WriteCard(AlignmentSize, 0) *) ; WriteLn
   ELSE
      WriteString('     .align 32') ; (* WriteCard(AlignmentSize, 0) *) ; WriteLn
   END
END CodeAlignment ;


(*
   DeclareGlobalVariables - lists the Global variables for Module ModSym
                            together with their offset.
*)

PROCEDURE DeclareGlobalVariables (ModSym: CARDINAL) ;
VAR
   o, s,
   n, Son: CARDINAL ;
BEGIN
   IF Alignment
   THEN
      WriteString('           .bss') ; WriteLn ;
      WriteString('           .align ') ; WriteCard(PentiumDataAlignmentSize, 0) ; WriteLn
   END ;
   n := 1 ;
   Son := GetNth(ModSym, n) ;
   o := 0 ;
   WHILE Son#NulSym DO
      VariableInfo(ModSym, Son) ;   (* Generate GDB debugging info *)
      CheckAlignment(Son, o) ;
      IF IsExported(ModSym, Son)
      THEN
         (* Actually exported variables are defined as COMM ... *)
         (* was - WriteString('EXTRN  ') *)
         WriteString('.comm  ')
      ELSE
         WriteString('.lcomm   ')
      END ;
      WriteAsmName(Son) ; WriteString('  ') ;
      s := GetSize(Son) ;
      Write(',') ; WriteSize(s) ;
      WriteString(' # offset') ;
      WriteInt(GetOffset(Son), 6) ;
      IF GetMode(Son)=LeftValue
      THEN
         WriteString(' (Pointer To ') ; WriteKey(GetSymName(GetType(Son))) ;
         Write(')') ;
      END ;
      o := GetOffset(Son)+s ;
      WriteLn ;
      INC(n) ;
      Son := GetNth(ModSym, n)
   END ;
   ForeachInnerModuleDo(ModSym, DeclareGlobalVariables)
END DeclareGlobalVariables ;


(*
   DeclareLocalVariables - lists the Local variables for Procedure Sym
                           together with their offset.
*)

PROCEDURE DeclareLocalVariables (Sym: CARDINAL) ;
VAR
   s,
   n,
   Son: CARDINAL ;
BEGIN
   n := 1 ;   (* Later collect function return param *)
   Son := GetNth(Sym, n) ;
   WHILE Son#NulSym DO
      WriteString('#   ') ;
      WriteAsmName(Son) ; Write(':') ;
      s := GetSize(Son) ;
      WriteSize(s) ;
      WriteString(' ; offset') ;
      WriteInt(GetOffset(Son), 6) ;
      IF GetMode(Son)=LeftValue
      THEN
         WriteString(' (Pointer To ') ; WriteKey(GetSymName(GetType(Son))) ;
         Write(')') ;
      END ;
      WriteLn ;
      INC(n) ;
      Son := GetNth(Sym, n)
   END
END DeclareLocalVariables ;


(*
   WriteSize - writes the Size in bytes, but in a human form.
*)

PROCEDURE WriteSize (Size: CARDINAL) ;
BEGIN
   WriteCard(Size, 0)
END WriteSize ;


(*
   Init - initialize global variables for this module.
*)

PROCEDURE Init ;
BEGIN
   ProfileCount := 1 ;
   TurnLineFeedOn
END Init ;


(*
   InstRegImmName -  instruction immediate name into register.  (Reg <- inst <- Constant)
                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

                     Example:

                     mov   $foo, Reg
*)

PROCEDURE InstRegImmName (Inst: Instruction;
                          Unit: DatumPrefix;
                          Reg : Register;
                          Sym : CARDINAL) ;
BEGIN
   WriteInstruction(mov, Unit) ;
   WriteString('  $') ; (* $ means immediate in gas *)
   WriteAsmName(Sym) ;
   WriteString(', ') ;
   WriteRegister(Reg, Unit) ;
   AutoWriteLn
END InstRegImmName ;


(*
   GenerateEntryProcedureLabel - generates a procedure label for procedure, Sym.
                                 The procedure is defined at CurrentQuad.
*)

PROCEDURE GenerateEntryProcedureLabel (CurrentQuad: CARDINAL; Sym: CARDINAL) ;
BEGIN
   WriteString('.text') ; WriteLn ;
   WriteString('        .align ') ; WriteCard(AlignmentSize, 0) ; WriteLn ;
   IF IsExported(GetMainModule(), Sym)
   THEN
      WriteString('.globl  ') ;    (* Exported procedures defined as glob *)
      WriteAsmName(Sym) ;
      WriteLn ;
   END ;
   WriteAsmName(Sym) ; Write(':') ; WriteLn ;
   CheckLineNumber(CurrentQuad)
END GenerateEntryProcedureLabel ;


(*
   GenerateExitProcedureLabel - generates a procedure label for procedure, Sym.
                                The procedure is defined at CurrentQuad.
*)

PROCEDURE GenerateExitProcedureLabel (Sym: CARDINAL) ;
BEGIN
END GenerateExitProcedureLabel ;


(*
   InstName - instruction to operate on a symbol name.
              Inst  - determines the instruction.
              Unit  - determines the DatumPrefix of manipulation
              Sym   - symbol whose name is used as the operand

              Example:

              call  $foo
*)

PROCEDURE InstName (Inst: Instruction;
                    Unit: DatumPrefix;
                    Sym : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString('  ') ;
   WriteAsmName(Sym) ; AutoWriteLn
END InstName ;


(*
   InstKey - inst followed by a string:
             inst String
             Inst   - determines the instruction.
             String - string that follows the Inst

             Example:

             Inst  String
             ; call  String
*)

PROCEDURE InstKey (Inst: Instruction; string: CARDINAL) ;
BEGIN
   WriteInstruction(Inst, PDWord) ; WriteString('  ') ; WriteKey(string) ;
   AutoWriteLn
END InstKey ;


(*
   MovePackedStringTo - moves a packed string b into the memory location
                        indicated by DestSym.
                        GeneralOffset determines the number of bytes offset
                        from DestSym the packed string is to be placed.
                        Pack determines the number of packed characters
                        must be stored at location DestSym.
                        If the string, b, contains less characters than
                        Pack then the remaining characters must be filled
                        in as nul.
*)

PROCEDURE MovePackedStringTo (DestSym: CARDINAL; GeneralOffset: INTEGER;
                              b: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   Offset: INTEGER ;
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(b)
      THEN
         INC(Word, ORD(b[i]))
      END
   END ;
   Offset := GetOffset(DestSym) ;
   WriteInstruction(mov, GetUnitPrefix(Pack)) ;
   WriteString(' $0x') ; WriteHex(Word, 0) ;
   WriteString(', ') ;
   WriteScope(GetScope(DestSym), Offset, GeneralOffset, DestSym) ;
   WriteString(' #  ') ; DisplayString(b, Pack-1) ;
   WriteLn ;
   Assert(GetMode(DestSym)#LeftValue)    (* should be using MovePackedStringToIReg *)
END MovePackedStringTo ;


(*
   MovePackedStringToIReg - moves a packed string b into the memory location
                            indicated by (Reg)
                            GeneralOffset determines the number of bytes offset
                            from IReg the packed string is to be placed.
                            Pack determines the number of packed characters
                            must be stored at location (Reg+GeneralOffset).
                            If the string, b, contains less characters than
                            Pack then the remaining characters must be filled
                            in as nul.
*)

PROCEDURE MovePackedStringToIReg (Reg: Register; GeneralOffset: INTEGER;
                                  b: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   Offset: INTEGER ;
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(b)
      THEN
         INC(Word, ORD(b[i]))
      END
   END ;
   WriteInstruction(mov, GetUnitPrefix(Pack)) ;
   WriteString(' $0x') ; WriteHex(Word, 0) ;
   WriteString(', ') ;
   WritePositiveOffset(GeneralOffset) ;
   Write('(') ;
   WriteRegister(Reg, MaxDatum()) ;
   Write(')') ;
   WriteString(' #  ') ; DisplayString(b, Pack-1) ;
   WriteLn
END MovePackedStringToIReg ;


(*
   PushPackedString - moves a packed string b onto the stack.
                      Pack determines the number of packed characters
                      must be stored on the stack.
                      If the string, b, contains less characters than
                      Pack then the remaining characters must be filled
                      in as nul.
*)

PROCEDURE PushPackedString (b: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(b)
      THEN
         INC(Word, ORD(b[i]))
      END
   END ;
   WriteInstruction(push, GetUnitPrefix(Pack)) ;
   WriteString(' $0x') ;
   WriteHex(Word, 0) ;
   WriteString('  # ') ;
   DisplayString(b, Pack-1) ;
   WriteLn
END PushPackedString ;


(*
   PutStringMDReg - moves a packed string a into memory offset, Offset,
                    indexed by register, Reg.
                    Pack determines the number of packed characters
                    to be stored.
                    If the string, b, contains less characters than
                    Pack then the remaining characters must be filled
                    in as nul.
*)

PROCEDURE PutStringMDReg (Offset: INTEGER; Reg: Register;
                          a: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(a)
      THEN
         INC(Word, ORD(a[i]))
      END
   END ;
   WriteInstruction(mov, GetUnitPrefix(Pack)) ;
   WriteString(' $0x') ; WriteHex(Word, 0) ;
   WriteString(' ,') ;
   WritePositiveOffset(Offset) ;
   Write('(') ;
   WriteRegister(Reg, MaxDatum()) ;
   WriteString(')  #  ') ;
   DisplayString(a, Pack-1) ;
   WriteLn
END PutStringMDReg ;


(*
   DisplayString - displays a string, a, which contains, n, characters.
*)

PROCEDURE DisplayString (a: ARRAY OF CHAR; n: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   Write('"') ;
   FOR i := 0 TO n DO
      IF a[i]=nul
      THEN
         WriteString('<nul>')
      ELSE
         Write(a[i])
      END
   END ;
   WriteString('"   ( ') ;
   FOR i := n TO 0 BY -1 DO
      IF a[i]=nul
      THEN
         WriteString('<nul>')
      ELSE
         Write(a[i])
      END
   END ;
   WriteString(' )')
END DisplayString ;


(*
   CloseAssemblyFileAndTerminate - displays termination code
*)

PROCEDURE CloseAssemblyFileAndTerminate ;
BEGIN
   CloseAssemblyFile
END CloseAssemblyFileAndTerminate ;


(*
   FlushPipe - a dummy procedure which makes the interface to M2AS
               the same as the interface to M2Pipe586.
*)

PROCEDURE FlushPipe (cycles: CARDINAL) ;
BEGIN
END FlushPipe ;


BEGIN
   Init
END M2AS.
