IMPLEMENTATION MODULE M2AS ;


FROM SYSTEM IMPORT ADR ;

FROM M2AssemIO IMPORT OpenAssemblyFile, CloseAssemblyFile ;

FROM i486 IMPORT DWORDSIZE, WORDSIZE, BYTESIZE,
                 GetScope, GetOffset, GetSize, GetUnitPrefix,
                 MaxDatum ;

FROM M2System IMPORT IsPseudoSystemFunction ;

FROM M2GDB IMPORT FileInfo, TypeInfo,
                  VariableInfo, BeginInfo, EndInfo, MainInfo ;

FROM M2Configure IMPORT ReturnArgumentOnStack, SystemIsXenix ;
FROM M2AsmUtil IMPORT WriteAsmName, UnderScoreString, GetModuleInitName ;
FROM M2FileName IMPORT CalculateFileName ;
FROM M2ALU IMPORT PopInt ;
FROM NameKey IMPORT WriteKey, GetKey, MakeKey ;

FROM SymbolTable IMPORT NulSym,
                        ModeOfAddr,
                        GetMainModule,
                        IsConstString, GetString, GetStringLength,
                        IsProcedure, IsVar, IsModule, IsExported,
                        GetSymName, GetNth, GetMode, GetType, GetVarFather,
                        ForeachExportedDo,
                        ForeachImportedDo,
                        ForeachProcedureDo,
                        ForeachInnerModuleDo,
                        PushValue ;

FROM M2Debug IMPORT Assert ;
FROM M2Lexical IMPORT WriteError, NearToken, InternalError ;
FROM M2Error IMPORT BeginError, EndError ;
FROM StdIO IMPORT Write, PushOutput, PopOutput ;
FROM FIO IMPORT OpenToWrite, Close, WriteChar, IsNoError, File, WriteNBytes ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM NumberIO IMPORT WriteCard, CardToStr, WriteInt, WriteHex ;
FROM M2GDB IMPORT LineInfo ;

FROM M2Options IMPORT Optimizing, GenerateDebugging, Profiling,
                      WordLength ;
FROM StrLib IMPORT StrLen, StrEqual, StrConCat, StrCopy ;

FROM M2Quads IMPORT QuadOperator, GetQuad, GetNextQuad ;


CONST
   MaxString  =    255  ;
   nul        =      0C ;    (* p2c complains about ASCII_sub *)

VAR
   ProfileCount      : CARDINAL ; (* Gives a unique identifier for profile  *)
                                  (* count buckets.                         *)


(*
   TurnLineFeedOn - automatically generates a line feed when an
                    instruction is generated.
*)

PROCEDURE TurnLineFeedOn ;
BEGIN
END TurnLineFeedOn ;


(*
   TurnLineFeedOff - do not automatically generate a line feed when
                     an instruction is generated.
*)

PROCEDURE TurnLineFeedOff ;
BEGIN
END TurnLineFeedOff ;


(*
   CheckProfiling - checks to see whether profiling is turned on
                    and if so it emits a call to _mcount.
*)

PROCEDURE CheckProfiling ;
BEGIN
   IF Profiling
   THEN
      (* I guess we should call _mcount at some stage *)
   END
END CheckProfiling ;


(*
   InstRegConstant - instruction constant and memory.  (Reg <- inst <- Constant)
                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

                     Example:

                     mov   Reg, 12
                     add   Reg, 12
*)

PROCEDURE InstRegConstant (Inst: Instruction;
                           Unit: DatumPrefix;
                           Reg : Register;
                           Sym : CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WriteRegister(Reg, Unit) ;
   WriteString(', ') ; WriteConstant(Sym) ;
   WriteLn
END InstRegConstant ;


(*
   InstRegConstant - instruction register and constant.
      	       	     	   (cmp Constant, Reg)

                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

                     Example:

                     mov   Reg, 12
                     add   Reg, 12
*)

PROCEDURE InstConstantReg (Inst: Instruction;
                           Unit: DatumPrefix;
                           Sym : CARDINAL;
                           Reg : Register) ;
BEGIN
   WriteInstruction(Inst) ;
   WriteString('  ') ;
   WriteConstant(Sym) ;
   WriteString(', ') ;
   WriteRegister(Reg, Unit) ;
   WriteLn
END InstConstantReg ;


(*
   InstMDConstant - instruction constant and memory.  (Mem <- inst <- Constant)
                    Inst   - determines the instruction.
                    Unit   - determines the DatumPrefix of transfer.
                    o1, go - determine the offset into element1.
                    Scope  - determines the scope of element1.
                    Sym1   - element1 where the constant is to be placed.
                             Symbol requiring the memory reference.
                    Sym2   - constant element2.

                    Example:

                    mov   DWORD PTR [ebp+15], 12
*)

PROCEDURE InstMDConstant (Inst  : Instruction;
                          Unit  : DatumPrefix;
                          o1, go: INTEGER ;
                          Scope : KindOfScope;
                          Sym1,
                          Sym2  : CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WritePrefix(Unit) ;
   Write(' ') ;
   WriteScope(Scope, o1, go, Sym1) ;
   WriteString(', ') ; WriteConstant(Sym2) ;
   WriteLn
END InstMDConstant ;


(*
   InstMRIConstant - instruction constant and memory indexed by a register.
                     (Mem[Reg1] <- inst <- Constant)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     o1, go - determine the offset into element1.
                     Sym2   - constant element2.

                     Example:

                     mov   DWORD PTR [edi+15], 12
*)

PROCEDURE InstMRIConstant (Inst  : Instruction;
                           Unit  : DatumPrefix;
                           o1, go: INTEGER ;
                           Reg1  : Register;
                           Sym2  : CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WritePrefix(Unit) ;
   WriteString(' [') ;
   WriteRegister(Reg1, PDWord) ;
   WriteOffset(o1) ;
   WriteOffset(go) ;
   WriteString('], ') ; WriteConstant(Sym2) ;
   WriteLn
END InstMRIConstant ;


(*
   InstMRISymConstant - instruction constant and memory indexed by a register.
                        (Mem[Reg1] <- inst <- Constant)
                        Inst   - determines the instruction.
                        Unit   - determines the DatumPrefix of transfer.
                        Sym1   - global variable.
                        go     - determine the offset into element1.
                        Sym2   - constant element2.

                        Example:

                        mov   12, myglobal+3(edi)
*)

PROCEDURE InstMRISymConstant (Inst  : Instruction;
                              Unit  : DatumPrefix;
                              Sym1  : CARDINAL;
                              go    : INTEGER ;
                              Reg1  : Register;
                              Sym2  : CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WritePrefix(Unit) ;
   Write(' ') ;
   WriteScope(Global, 0, go, Sym1) ;
   Write('[') ;
   WriteRegister(Reg1, MaxDatum()) ;
   WriteString('], ') ;
   WriteConstant(Sym2) ;
   WriteLn
END InstMRISymConstant ;


(*
   InstMRICardinal - instruction constant and memory indexed by a register.
                     (Mem[Reg1] <- inst <- Constant)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     o1, go - determine the offset into element1.
                     Card   - cardinal constant.

                     Example:

                     mov   DWORD PTR [edi+15], 12H
*)

PROCEDURE InstMRICardinal (Inst  : Instruction;
                           Unit  : DatumPrefix;
                           o1, go: INTEGER ;
                           Reg1  : Register;
                           Card  : CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WritePrefix(Unit) ;
   WriteString(' [') ;
   WriteRegister(Reg1, PDWord) ;
   WriteOffset(o1) ;
   WriteOffset(go) ;
   WriteString('], ') ; WriteHex(Card, 0) ; Write('H') ;
   WriteLn
END InstMRICardinal ;


(*
   InstMRIReg - instruction constant and memory indexed by a register.
                (Mem[Reg1] <- inst <- Reg2)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                o1, go - determine the offset into element1.
                Reg1   - indexes memory.
                Reg2   - source of data.

                Example:

                mov   DWORD PTR [edi+15], eax
*)

PROCEDURE InstMRIReg (Inst  : Instruction;
                      Unit  : DatumPrefix;
                      o1, go: INTEGER ;
                      Reg1  ,
                      Reg2  : Register) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WritePrefix(Unit) ;
   WriteString(' [') ;
   WriteRegister(Reg1, PDWord) ;
   WriteOffset(o1) ;
   WriteOffset(go) ;
   WriteString('], ') ; WriteRegister(Reg2, Unit) ;
   WriteLn
END InstMRIReg ;


(*
   InstMDReg - inst reg into memory. (Mem1 <- inst <- Reg)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o1, go - determine the offset into element1.
               Reg    - determine the register.
               Scope1 - determines the scope of element1.
               Sym    - symbol requiring the memory reference.

               Example:

               mov   DWORD PTR [ebp+15], eax
*)

PROCEDURE InstMDReg (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     o1, go: INTEGER ;
                     Reg   : Register;
                     Scope1: KindOfScope;
                     Sym   : CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WritePrefix(Unit) ;
   Write(' ') ;
   WriteScope(Scope1, o1, go, Sym) ;
   WriteString(', ') ; WriteRegister(Reg, Unit) ;
   WriteLn
END InstMDReg ;


(*
   InstRegMD - inst memory into reg. (Reg <- inst <- Mem1)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o2, go - determine the offset into element2.
               Reg    - determine the register.
               Scope2 - determines the scope of element2.
               Sym    - symbol requiring the memory reference.

               Example:

               ; inst  reg, Mem{scope}+o1+go
               mov   eax, DWORD PTR [ebp+15]
*)

PROCEDURE InstRegMD (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg   : Register;
                     o1, go: INTEGER ;
                     Scope1: KindOfScope;
                     Sym   : CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WriteRegister(Reg, Unit) ;
   WriteString(', ') ;
   WritePrefix(Unit) ;
   Write(' ') ;
   WriteScope(Scope1, o1, go, Sym) ;
   WriteLn
END InstRegMD ;


(*
   InstRegMIO - inst memory into reg. (Reg1 <- inst <- Mem[Reg2])
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg1   - determine the destination register.
                Reg2   - determine the index register.
                Offset - offset to be added to the index register.

               Example:

               mov   eax, DWORD PTR [edi+4]
               ; mov   reg1, DWORD PTR [reg2+offset]
*)

PROCEDURE InstRegMIO (Inst  : Instruction;
                      Unit  : DatumPrefix;
                      Reg1,
                      Reg2  : Register;
                      Offset: INTEGER) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WriteRegister(Reg1, Unit) ;
   WriteString(', ') ;
   WritePrefix(Unit) ;
   WriteString(' [') ;
   WriteRegister(Reg2, PDWord) ;
   WriteOffset(Offset) ;
   Write(']') ;
   WriteLn
END InstRegMIO ;


(*
   InstRegMISym - inst memory into reg. (Reg1 <- inst <- global[Reg2])
                  Inst   - determines the instruction.
                  Unit   - determines the DatumPrefix of transfer.
                  Reg1   - determine the destination register.
                  Reg2   - determine the index register.
                  Offset - offset to be added to the index register.
                  Sym    - the global variable.

                  Example:

                  mov   myglobal+4(edi), eax
                ; mov   Sym+offset(reg2), reg1
*)

PROCEDURE InstRegMISym (Inst  : Instruction;
                        Unit  : DatumPrefix;
                        Reg1,
                        Reg2  : Register;
                        Offset: INTEGER;
                        Sym   : CARDINAL) ;
BEGIN
   Assert(NOT IsProcedure(GetVarFather(Sym))) ;
   WriteInstruction(Inst) ;
   Write(' ') ;
   WriteRegister(Reg1, Unit) ;
   WriteString(', ') ;
   WritePrefix(Unit) ;
   Write(' ') ;
   WriteScope(Global, 0, Offset, Sym) ;
   WriteString(' [') ;
   WriteRegister(Reg2, MaxDatum()) ;
   Write(']') ;
   WriteLn
END InstRegMISym ;


(*
   InstRegMI - inst memory into reg. (Reg1 <- inst <- Mem[Reg2])
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               Reg1   - determine the destination register.
               Reg2   - determine the index register.

               Example:

               mov   eax, DWORD PTR [edi]
               ; mov   reg1, DWORD PTR [reg2]
*)

PROCEDURE InstRegMI (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg1,
                     Reg2  : Register) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WriteRegister(Reg1, Unit) ;
   WriteString(', ') ;
   WritePrefix(Unit) ;
   WriteString(' [') ;
   WriteRegister(Reg2, PDWord) ;
   Write(']') ;
   WriteLn
END InstRegMI ;


(*
   InstMIOReg - inst memory into reg. (Mem[Reg1] <- inst <- Reg2)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg1   - determine the index register.
                Offset - determines the offset to be added to the index
                         register.
                Reg2   - determine the source register.

               Example:

               mov   DWORD PTR [edi], eax
               ; mov   DWORD PTR [reg1], reg2
*)

PROCEDURE InstMIOReg (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg1  : Register;
                     Offset: INTEGER;
                     Reg2  : Register) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WritePrefix(Unit) ;
   WriteString(' [') ;
   WriteRegister(Reg1, PDWord) ;
   WriteOffset(Offset) ;
   WriteString('], ') ;
   WriteRegister(Reg2, Unit) ;
   WriteLn
END InstMIOReg ;


(*
   InstMISymReg - inst reg into memory. (global[Reg2] <- inst <- Reg1)
                  Inst   - determines the instruction.
                  Unit   - determines the DatumPrefix of transfer.
                  Reg1   - determine the destination register.
                  Reg2   - determine the index register.
                  Offset - offset to be added to the index register.
                  Sym    - the global variable.

                  Example:

                  mov   %eax, myglobal+4(edi),
                ; mov   reg1, Sym+offset(reg2)
*)

PROCEDURE InstMISymReg (Inst  : Instruction;
                        Unit  : DatumPrefix;
                        Reg1,
                        Reg2  : Register;
                        Offset: INTEGER;
                        Sym   : CARDINAL) ;
BEGIN
   Assert(NOT IsProcedure(GetVarFather(Sym))) ;
   WriteInstruction(Inst) ;
   Write(' ') ;
   WritePrefix(Unit) ;
   Write(' ') ;
   WriteScope(Global, 0, Offset, Sym) ;
   WriteString(' [') ;
   WriteRegister(Reg2, MaxDatum()) ;
   WriteString('], ') ;
   WriteRegister(Reg1, Unit) ;
   WriteLn
END InstMISymReg ;


(*
   InstMIReg - inst memory into reg. (Mem[Reg1] <- inst <- Reg2)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               Reg1   - determine the index register.
               Reg2   - determine the source register.

               Example:

               mov   DWORD PTR [edi], eax
               ; mov   DWORD PTR [reg1], reg2
*)

PROCEDURE InstMIReg (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg1,
                     Reg2  : Register) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WritePrefix(Unit) ;
   WriteString(' [') ;
   WriteRegister(Reg1, PDWord) ;
   WriteString('], ') ;
   WriteRegister(Reg2, Unit) ;
   WriteLn
END InstMIReg ;


(*
   InstRegReg - inst reg into reg. (Reg1 <- inst <- Reg2)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg1   - determine the destination register.
                Reg2   - determine the index register.

                Example:

                mov   eax, ebp
               ; mov   reg1, reg2
*)

PROCEDURE InstRegReg (Inst: Instruction; Unit: DatumPrefix;
                      Reg1, Reg2: Register) ;
BEGIN
   WriteInstruction(Inst) ;
   WriteString('  ') ;
   WriteRegister(Reg1, Unit) ;
   WriteString(', ') ;
   WriteRegister(Reg2, Unit) ;
   WriteLn
END InstRegReg ;


(*
   InstRegInt - inst Reg with an Int. (Reg <- inst <- Int)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg    - determines the destination register.
                Int    - the integer constant as the second operand.

                Example:

                mov   eax, 15
               ; mov   Reg, Int
*)

PROCEDURE InstRegInt (Inst: Instruction; Unit: DatumPrefix;
                      Reg: Register; Int: INTEGER) ;
BEGIN
   WriteInstruction(Inst) ;
   WriteString('  ') ;
   WriteRegister(Reg, Unit) ;
   WriteString(', ') ;
   WriteInt(Int, 0) ;
   WriteLn
END InstRegInt ;


(*
   InstRegHexCard - inst Reg with an cardinal. (Reg <- inst <- Card)
                    Inst   - determines the instruction.
                    Unit   - determines the DatumPrefix of transfer.
                    Reg    - determines the destination register.
                    Card   - the cardinal constant as the second operand.

                    Example:

                    and   eax, 0FFFF0000H
                   ; mov   Reg, Int
*)

PROCEDURE InstRegHexCard (Inst: Instruction; Unit: DatumPrefix;
                          Reg: Register; Card: CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   WriteString('  ') ;
   WriteRegister(Reg, Unit) ;
   WriteString(', 0') ;
   WriteHex(Card, 0) ;
   Write('H') ;
   WriteLn
END InstRegHexCard ;


(*
   InstMDOffset - inst Mem with an offset Int. (Mem[Sym1] <- inst <- OFFSET Sym2)
                  Offset is the address of symbol Sym2.
                  Inst   - determines the instruction.
                  Unit   - determines the DatumPrefix of transfer.
                  Sym1   - indicates the memory location which will
                           contain the address of Sym2.
                  Sym2   - symbol requiring the memory reference.

                  Example:

                  mov   DWORD PTR [ebp-15], OFFSET _varname
                  ; mov   {unit prefix} {scope}, OFFSET _varname
*)

PROCEDURE InstMDOffset (Inst: Instruction; Unit: DatumPrefix;
                        Sym1, Sym2: CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   WriteString('  ') ;
   WritePrefix(Unit) ;
   WriteScope(GetScope(Sym1), GetOffset(Sym1), 0, Sym1) ;
   WriteString(', OFFSET ') ;
   WriteAsmName(Sym2) ;
   WriteLn
END InstMDOffset ;


(*
   InstRegOffset - inst reg with an offset Int. (Reg <- inst <- OFFSET Sym)
                   Offset is the address of symbol Sym.
                   Inst   - determines the instruction.
                   Unit   - determines the DatumPrefix of transfer.
                   Reg    - register which contain the address of Sym2.
                   Sym    - symbol requiring the memory reference.

                   Example:

                   mov   $_varname, $eax
                  ; mov  $_varname, {unit prefix} {reg}
*)

PROCEDURE InstRegOffset (Inst: Instruction; Unit: DatumPrefix;
                         Reg: Register; Sym: CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   WriteString('  ') ;
   WriteRegister(Reg, Unit) ;
   WriteString(', OFFSET ') ;
   WriteAsmName(Sym) ;
   WriteLn
END InstRegOffset ;


(*
   InstMDInt - inst Mem with an Int. (Mem[Sym1] <- inst <- Int)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o1, go - determine the offset into element1.
               Int    - the integer constant as the second operand.
               Scope  - determines the scope of element1.
               Sym    - symbol requiring the memory reference.

               Example:

               mov   DWORD PTR [ebp-15], 123
               ; mov   {unit prefix} {scope}, int
*)

PROCEDURE InstMDInt (Inst: Instruction; Unit: DatumPrefix;
                     o1, go: INTEGER; Int: INTEGER;
                     Scope: KindOfScope; Sym: CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   WriteString('  ') ;
   WritePrefix(Unit) ;
   WriteScope(Scope, o1, go, Sym) ;
   WriteString(', ') ;
   WriteInt(Int, 0) ;
   WriteLn
END InstMDInt ;


(*
   InstStr - inst followed by a string:
             inst String
             Inst   - determines the instruction.
             String - string that follows the Inst

             Example:

             Inst  String
             ; call  String
*)

PROCEDURE InstStr (Inst: Instruction; String: ARRAY OF CHAR) ;
BEGIN
   WriteInstruction(Inst) ;
   WriteString('  ') ;
   WriteString(String) ;
   WriteLn
END InstStr ;


(*
   InstKey - inst followed by a string:
             inst String
             Inst   - determines the instruction.
             String - string that follows the Inst

             Example:

             Inst  String
             ; call  String
*)

PROCEDURE InstKey (Inst: Instruction; string: CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ; WriteString('  ') ; WriteKey(string) ;
   WriteLn
END InstKey ;


(*
   BinInst - write a sole inst.
             Inst   - determines the instruction.

             Example:

             ret
*)

PROCEDURE BinInst (Inst: Instruction) ;
BEGIN
   WriteInstruction(Inst) ;
   WriteLn
END BinInst ;


(*
   BinInstUnit - write a sole inst.
                 Inst   - determines the instruction.

                 Example:

                 rep
                 movsl %esp:$0, %esp:$0
*)

PROCEDURE BinInstUnit (Inst: Instruction; Unit: DatumPrefix; Length: CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   WritePrefix(Unit) ; WriteString(' ES:0, ') ; WritePrefix(Unit) ; WriteString(' ES:0') ;
   WriteLn
END BinInstUnit ;


(*
   BinInstReg - inst operating on a reg. (inst <-> Reg)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg    - determine the operand register.

                Example:

                push   eax
*)

PROCEDURE BinInstReg (Inst: Instruction;
                      Unit: DatumPrefix;
                      Reg : Register) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WriteRegister(Reg, Unit) ;
   WriteLn
END BinInstReg ;


(*
   BinInstRegI - inst operating on a reg. ( inst <-> (Reg) )
                 Inst   - determines the instruction.
                 Unit   - determines the DatumPrefix of transfer.
                 Reg    - determine the operand register.

                 Example:

                 call   (eax)
*)

PROCEDURE BinInstRegI (Inst: Instruction;
                       Unit: DatumPrefix;
                       Reg : Register) ;
BEGIN
   WriteInstruction(Inst) ;
   WriteString(' [') ;
   WriteRegister(Reg, Unit) ;
   WriteString('] ') ;
   WriteLn
END BinInstRegI ;


(*
   BinInstConstant - inst operating on a reg. (inst <-> Reg)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     Sym    - determine the constant.

                     Example:

                     push   15
*)

PROCEDURE BinInstConstant (Inst: Instruction;
                           Unit: DatumPrefix;
                           Sym : CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WriteConstant(Sym) ;
   WriteLn
END BinInstConstant ;


(*
   BinInstMD - inst operating on a memory location. (inst <-> Mem)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o1, go - determine the operand memory location.
               Scope1 - determines the scope of the operand.
               Sym    - symbol requiring the memory reference.

               Example:

               push   DWORD PTR [ebp+15]
*)

PROCEDURE BinInstMD (Inst: Instruction; Unit: DatumPrefix;
                     o1, go  : INTEGER; Scope1: KindOfScope; Sym: CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   Write(' ') ;
   WritePrefix(Unit) ;
   Write(' ') ;
   WriteScope(Scope1, o1, go, Sym) ;
   WriteLn
END BinInstMD ;


(*
   BinInstHex - inst operating on a number. (inst <-> hex)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                c      - number.

                Example:

                push   0x1234
*)

PROCEDURE BinInstHex (Inst: Instruction;
                      Unit: DatumPrefix;
                      c   : CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   IF (Unit=PByte) AND (Inst=push)
   THEN
      WriteError('cannot push a byte on the 8088, 8086, 80386 or 80486')
   END ;
   WriteString(' 0') ;
   WriteHex(c, 0) ; Write('H') ; WriteLn ;
END BinInstHex ;


(*
   Comment - places string, a, into the assembly file.
*)

PROCEDURE Comment (a: ARRAY OF CHAR) ;
BEGIN
   WriteString('; ') ; WriteString(a) ; WriteLn
END Comment ;


(*
   WritePrefix - writes the prefix associated with transfering Size
                 bytes of data.
*)

PROCEDURE WritePrefix (Size: DatumPrefix) ;
BEGIN
   CASE Size OF

   PByte :  WriteString('BYTE PTR') |
   PWord :  WriteString('WORD PTR') |
   PDWord:  WriteString('DWORD PTR')

   ELSE
      WriteError('Expecting PByte, PWord, PDWord')
   END
END WritePrefix ;


(*
   WriteScope - writes the referencing information to a particular scope.
                Scope is either Global or Local.
                o1 is the offset of the element and go is the Delta
                offset. It is useful to contain both as separate
                entities as it makes human debugging easier!
                Sym - is the variable requiring memory access.
                      If global then the name Sym is written
                      otherwise the activation record offset in the form of
                      o1+go is written.
*)

PROCEDURE WriteScope (Scope: KindOfScope; o1, go: INTEGER; Sym: CARDINAL) ;
VAR
   Proc : CARDINAL ;
BEGIN
   IF Scope=Local
   THEN
      WriteString('[ebp') ; WriteOffset(o1) ; WriteOffset(go) ; Write(']')
   ELSE
      Write('[') ; WriteAsmName(Sym) ; WriteOffset(go) ; Write(']')
   END
END WriteScope ;


(*
   WriteOffset - writes an integer offset with a sign beforehand.
*)

PROCEDURE WriteOffset (o: INTEGER) ;
BEGIN
   IF o>=0
   THEN
      Write('+')
   END ;
   WriteInt(o, 0)
END WriteOffset ;


(*
   WriteConstant - writes a constant from a symbol.
                   Place in SymbolTable - must write out the following consts:

                   (1)  Integer, Cardinal value constants.
                   (2)  Character constants.
                   (3)  String constants.
*)

PROCEDURE WriteConstant (Sym: CARDINAL) ;
VAR
   Size: CARDINAL ;
   Str : ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IF IsConstString(Sym)
   THEN
      Size := GetStringLength(Sym) ;
      GetKey(GetString(Sym), Str) ;
      IF Size=1
      THEN
         IF Str[0]="'"
         THEN
            Write('"') ; Write(Str[0]) ; Write('"')
         ELSE
            Write("'") ; Write(Str[0]) ; Write("'")
         END
      ELSIF Size=0
      THEN
         WriteString(' 00H') ;  (* Empty string *)
      ELSE
         WriteString(Str)
      END
   ELSE
      PushValue(Sym) ;
      WriteInt(PopInt(), 0)
   END
END WriteConstant ;


(*
   WriteInstruction - writes an instruction.
*)

PROCEDURE WriteInstruction (Inst: Instruction) ;
BEGIN
   WriteString('    ') ;
   CASE Inst OF

   mov    : WriteString('mov  ') |
   add    : WriteString('add  ') |
   sub    : WriteString('sub  ') |
   mult   : WriteString('imul ') |
   idiv   : WriteString('idiv ') |
   div    : WriteString('div  ') |
   cdq    : WriteString('cdq  ') |
   mod    : WriteString('mod  ') |
   neg    : WriteString('neg  ') |
   push   : WriteString('push ') |
   pop    : WriteString('pop  ') |
   jmp    : WriteString('jmp  ') |
   je     : WriteString('je   ') |
   jne    : WriteString('jne  ') |
   jl     : WriteString('jl   ') |
   jg     : WriteString('jg   ') |
   jle    : WriteString('jle  ') |
   jge    : WriteString('jge  ') |
   jae    : WriteString('jae  ') |
   cmp    : WriteString('cmp  ') |
   repmovs: WriteString('rep movs ') |
   and    : WriteString('and  ') |
   or     : WriteString('or   ') |
   xor    : WriteString('xor  ') |
   sal    : WriteString('sal  ') |
   sar    : WriteString('sar  ') |
   shr    : WriteString('shr  ') |
   shl    : WriteString('shl  ') |
   ret    : WriteString('ret  ') |
   call   : WriteString('call ') |
   leave  : WriteString('leave') |
   inc    : WriteString('inc  ') |
   dec    : WriteString('dec  ') |
   nop    : WriteString('nop  ')

   ELSE
      WriteError('Instruction not known - expand case clause')
   END
END WriteInstruction ;


(*
   WriteRegister - writes a register.
*)

PROCEDURE WriteRegister (Reg: Register; Unit: DatumPrefix) ;
BEGIN
   CASE Unit OF

   PDWord: CASE Reg OF

           RegEAX: WriteString('eax') |
           RegEBX: WriteString('ebx') |
           RegECX: WriteString('ecx') |
           RegEDX: WriteString('edx') |
           RegEBP: WriteString('ebp') |
           RegESP: WriteString('esp') |
           RegESI: WriteString('esi') |
           RegEDI: WriteString('edi')

           ELSE
              WriteError('Register not known - expand case clause')
           END |

   PWord : CASE Reg OF

           RegEAX: WriteString('ax') |
           RegEBX: WriteString('bx') |
           RegECX: WriteString('cx') |
           RegEDX: WriteString('dx') |
           RegEBP: WriteString('bp') |
           RegESP: WriteString('sp') |
           RegESI: WriteString('si') |
           RegEDI: WriteString('di')

           ELSE
              WriteError('Register not known - expand case clause')
           END |

   PByte : CASE Reg OF

           RegEAX: WriteString('al') |
           RegEBX: WriteString('bl') |
           RegECX: WriteString('cl') |
           RegEDX: WriteString('dl') |
           RegEBP: WriteError('Cannot get byte access ebp') |
           RegESP: WriteError('Cannot get byte access esp') |
           RegESI: WriteError('Cannot get byte access esi') |
           RegEDI: WriteError('Cannot get byte access edi')

           ELSE
              WriteError('Register not known - expand case clause')
           END

   ELSE
      WriteError('DatumPrefix value not known')
   END
END WriteRegister ;


(*
   GenerateQuadLabel - writes a label refering to quad, QuadNo.
*)

PROCEDURE GenerateQuadLabel (QuadNo: CARDINAL) ;
BEGIN
   WriteString('L') ; WriteCard(QuadNo, 0) ; Write(':') ; WriteLn ;
END GenerateQuadLabel ;


(*
   OpenAssemblyFileAndInitialize - displays the initialization code.
*)

PROCEDURE OpenAssemblyFileAndInitialize ;
BEGIN
   OpenAssemblyFile(GetSymName(GetMainModule())) ;
   WriteLn ;
   WriteLn ;
   WriteString(";       Static Name Aliases") ;
   WriteString(";") ;
   WriteLn ;
   WriteString("        TITLE   $") ; WriteKey(GetSymName(GetMainModule())) ; WriteLn ;
   WriteLn ;
   WriteString("        .386") ; WriteLn ;
   WriteString("DGROUP  GROUP   CONST, _BSS, _DATA") ; WriteLn ;
   ForeachExportedDo(GetMainModule(), DisplayExportedProcedure) ;
   WriteString('PUBLIC  ') ; WriteKey(GetModuleInitName(GetMainModule())) ; WriteLn ;
   WriteString("_DATA   SEGMENT  DWORD USE32 PUBLIC 'DATA'") ; WriteLn ;
   DeclareGlobalVariables(GetMainModule()) ;
   ForeachImportedDo(GetMainModule(), DisplayImportedVar) ;
   ForeachProcedureDo(GetMainModule(), DisplayProcedureName) ;
   WriteString("_DATA      ENDS") ; WriteLn ;
   WriteString("_BSS    SEGMENT  DWORD USE32 PUBLIC 'BSS'") ; WriteLn ;
   WriteString("_BSS      ENDS") ; WriteLn ;
   WriteString("CONST   SEGMENT  DWORD USE32 PUBLIC 'CONST'") ; WriteLn ;
   WriteString("CONST      ENDS") ; WriteLn ;
   WriteString("_TEXT   SEGMENT  DWORD USE32 PUBLIC 'CODE'") ; WriteLn ;
   WriteString("        ASSUME   CS: _TEXT, DS: DGROUP, SS: DGROUP, ES: DGROUP") ; WriteLn ;
   ForeachImportedDo(GetMainModule(), DisplayImportedProcedure) ;
   IF GetSymName(GetMainModule())#MakeKey('M2RTS')
   THEN
      WriteString('EXTRN    ') ; UnderScoreString('M2RTS_HALT:NEAR') ; WriteLn ;
      WriteString('EXTRN    ') ; UnderScoreString('M2RTS_Terminate:NEAR') ; WriteLn ;
   END ;
   WriteLn ;
END OpenAssemblyFileAndInitialize ;


(*
   DisplayImportedProcedure - displays a symbol, Sym, (if a procedure)
                              in a MASM imported form.
*)

PROCEDURE DisplayImportedProcedure (Sym: CARDINAL) ;
BEGIN
   IF IsProcedure(Sym) AND (NOT IsPseudoSystemFunction(Sym))
   THEN
      WriteString('EXTRN   ') ; WriteAsmName(Sym) ; WriteString(':NEAR') ;
      WriteLn
   END
END DisplayImportedProcedure ;


(*
   DisplayExportedProcedure - displays a symbol, Sym, (if a procedure)
                              in a MASM exported form.
*)

PROCEDURE DisplayExportedProcedure (Sym: CARDINAL) ;
BEGIN
   IF IsProcedure(Sym) AND (NOT IsPseudoSystemFunction(Sym))
   THEN
      WriteString('PUBLIC   ') ; WriteAsmName(Sym) ;
      WriteLn
   END
END DisplayExportedProcedure ;


(*
   DisplayProcedureName - displays the procedure name in terms of byte strings.
*)

PROCEDURE DisplayProcedureName (Sym: CARDINAL) ;
BEGIN
   IF IsProcedure(Sym)
   THEN
      WriteString(';;;   ALIGN 4') ; WriteLn ;
      WriteString(';;;$S') ; WriteCard(Sym, 0) ;
      WriteString("   DB   '") ;
      WriteAsmName(Sym) ;
      WriteString("', 0aH, 00H") ; WriteLn
   END
END DisplayProcedureName ;


(*
   DisplayExportedVar - displays a symbol, Sym, (if a variable)
                        in a MASM exported form.
*)

PROCEDURE DisplayExportedVar (Sym: CARDINAL) ;
BEGIN
   IF IsVar(Sym)
   THEN
      WriteString('GLOBAL   ') ; WriteAsmName(Sym) ; WriteLn
   END
END DisplayExportedVar ;


(*
   DisplayImportedVar - displays a symbol, Sym, (if a variable)
                        in a MASM exported form.
*)

PROCEDURE DisplayImportedVar (Sym: CARDINAL) ;
BEGIN
   IF IsVar(Sym)
   THEN
      WriteString('EXTRN   ') ; WriteAsmName(Sym) ;
      WriteString(':NEAR') ; (* WriteSize(GetSize(Sym)) ; *)
      WriteLn
   END
END DisplayImportedVar ;


(*
   WritePrologue - writes the prologue code before the
                   initialization code of a module.
*)

PROCEDURE WritePrologue (ModSym: CARDINAL;
                         CurrentQuad: CARDINAL; Line: CARDINAL) ;
BEGIN
   WriteKey(GetModuleInitName(ModSym)) ; WriteString('   PROC NEAR') ; WriteLn ;
   BinInstReg(push, PDWord, RegEBP) ;
   InstRegReg(mov, PDWord, RegEBP, RegESP) ;
   BinInstReg(push, PDWord, RegEBX) ;
   BinInstReg(push, PDWord, RegEDI) ;
   BinInstReg(push, PDWord, RegESI)
END WritePrologue ;


(*
   WriteEpilogue - writes the termination code after the initialization
                   of a module.
*)

PROCEDURE WriteEpilogue (ModSym: CARDINAL) ;
BEGIN
   IF IsModule(GetMainModule()) AND (ModSym=GetMainModule())
   THEN
      WriteString('; clean up using Modula-2 M2RTS termination') ; WriteLn ;
      WriteInstruction(call) ;
      WriteString('  ') ;
      UnderScoreString('M2RTS_Terminate') ;
      WriteLn
   END ;
   BinInstReg(pop, PDWord, RegESI) ; 
   BinInstReg(pop, PDWord, RegEDI) ; 
   BinInstReg(pop, PDWord, RegEBX) ;
   BinInst(leave) ;
   BinInst(ret) ;
   WriteKey(GetModuleInitName(ModSym)) ; WriteString('   ENDP') ; WriteLn
END WriteEpilogue ;


(*
   CodeAlignment - emits a code alignment assembly statement.
*)

PROCEDURE CodeAlignment ;
BEGIN
   (* not yet implemented for masm *)
END CodeAlignment ;


(*
   DeclareGlobalVariables - lists the Global variables for Module ModSym
                            together with their offset.
*)

PROCEDURE DeclareGlobalVariables (ModSym: CARDINAL) ;
VAR
   s,
   n,
   Son: CARDINAL ;
BEGIN
   n := 1 ;
   Son := GetNth(ModSym, n) ;
   WHILE Son#NulSym DO
      IF IsExported(ModSym, Son)
      THEN
         (* Actually exported variables are defined as COMM ... *)
         (* was - WriteString('EXTRN  ') *)
         WriteString('COMM    ')
      ELSE
         WriteString('COMM    ')
      END ;
      WriteAsmName(Son) ; WriteString('  ') ;
      s := GetSize(Son) ;
      Write(':') ; WriteSize(s) ;
      WriteString(' ; offset') ;
      WriteInt(GetOffset(Son), 6) ;
      IF GetMode(Son)=LeftValue
      THEN
         WriteString(' (Pointer To ') ; WriteKey(GetSymName(GetType(Son))) ;
         Write(')') ;
      END ;
      WriteLn ;
      INC(n) ;
      Son := GetNth(ModSym, n)
   END ;
   ForeachInnerModuleDo(ModSym, DeclareGlobalVariables)
END DeclareGlobalVariables ;


(*
   DeclareLocalVariables - lists the Local variables for Procedure Sym
                           together with their offset.
*)

PROCEDURE DeclareLocalVariables (Sym: CARDINAL) ;
VAR
   s,
   n,
   Son: CARDINAL ;
BEGIN
   n := 1 ;   (* Later collect function return param *)
   Son := GetNth(Sym, n) ;
   WHILE Son#NulSym DO
      WriteString(';   ') ;
      WriteAsmName(Son) ; Write(':') ;
      s := GetSize(Son) ;
      WriteSize(s) ;
      WriteString(' ; offset') ;
      WriteInt(GetOffset(Son), 6) ;
      IF GetMode(Son)=LeftValue
      THEN
         WriteString(' (Pointer To ') ; WriteKey(GetSymName(GetType(Son))) ;
         Write(')') ;
      END ;
      WriteLn ;
      INC(n) ;
      Son := GetNth(Sym, n)
   END
END DeclareLocalVariables ;


(*
   WriteSize - writes the Size in bytes, but in a human form.
*)

PROCEDURE WriteSize (Size: CARDINAL) ;
BEGIN
   IF Size=DWORDSIZE
   THEN
      WriteString('DWORD')
   ELSIF Size=WORDSIZE
   THEN
      WriteString('WORD')
   ELSIF Size=BYTESIZE
   THEN
      WriteString('BYTE')
   ELSE
      WriteString('BYTE:') ; WriteCard(Size, 0)
   END
END WriteSize ;



(*
   CloseAssemblyFileAndTerminate - displays termination code
*)

PROCEDURE CloseAssemblyFileAndTerminate ;
BEGIN
   WriteString("_TEXT   ENDS") ; WriteLn ;
   WriteString("END") ; WriteLn ;
   CloseAssemblyFile
END CloseAssemblyFileAndTerminate ;


(*
   Init - initialize global variables for this module.
*)

PROCEDURE Init ;
BEGIN
   ProfileCount := 1
END Init ;


(*
   InstRegImmName -  instruction immediate name into register.  (Reg <- inst <- Constant)
                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

                     Example:

                     mov   $foo, Reg
*)

PROCEDURE InstRegImmName (Inst: Instruction;
                          Unit: DatumPrefix;
                          Reg : Register;
                          Sym : CARDINAL) ;
BEGIN
   WriteInstruction(mov) ;
   WriteString('  ') ; (*  means immediate in masm *)
   WriteRegister(Reg, Unit) ;
   WriteString(', ') ;
   WriteAsmName(Sym) ;
   WriteLn
END InstRegImmName ;


(*
   GenerateProcedureLabel - generates a procedure label for procedure, Sym.
                            The procedure is defined at CurrentQuad.
*)

PROCEDURE GenerateEntryProcedureLabel (CurrentQuad: CARDINAL; Sym: CARDINAL) ;
BEGIN
   WriteAsmName(Sym) ; WriteString('  PROC  NEAR') ; WriteLn
END GenerateEntryProcedureLabel ;


(*
   GenerateExitProcedureLabel - generates a procedure label for procedure, Sym.
*)

PROCEDURE GenerateExitProcedureLabel (Sym: CARDINAL) ;
BEGIN
   WriteAsmName(Sym) ; WriteString('  ENDP') ; WriteLn
END GenerateExitProcedureLabel ;


(*
   InstName - instruction to operate on a symbol name.
              Inst  - determines the instruction.
              Unit  - determines the DatumPrefix of manipulation
              Sym   - symbol whose name is used as the operand

              Example:

              call  foo
*)

PROCEDURE InstName (Inst: Instruction;
                    Unit: DatumPrefix;
                    Sym : CARDINAL) ;
BEGIN
   WriteInstruction(Inst) ;
   WriteString('  ') ;
   WriteAsmName(Sym) ; WriteLn ;
END InstName ;


(*
   MovePackedStringTo - moves a packed string b into the memory location
                        indicated by DestSym.
                        GeneralOffset determines the number of bytes offset
                        from DestSym the packed string is to be placed.
                        Pack determines the number of packed characters
                        must be stored at location DestSym.
                        If the string, b, contains less characters than
                        Pack then the remaining characters must be filled
                        in as nul.
*)

PROCEDURE MovePackedStringTo (DestSym: CARDINAL; GeneralOffset: INTEGER;
                              b: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   Offset: INTEGER ;
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(b)
      THEN
         INC(Word, ORD(b[i]))
      END
   END ;
   Offset := GetOffset(DestSym) ;
   WriteInstruction(mov) ;
   Write(' ') ;
   WritePrefix(GetUnitPrefix(Pack)) ;
   Write(' ') ;
   WriteScope(GetScope(DestSym), Offset, GeneralOffset, DestSym) ;
   WriteString(', 0') ; WriteHex(Word, 0) ;
   WriteString('H ; ') ; DisplayString(b, Pack-1) ;
   WriteLn ;
   Assert(GetMode(DestSym)#LeftValue)
END MovePackedStringTo ;


(*
   MovePackedStringToIReg - moves a packed string b into the memory location
                            indicated by (Reg)
                            GeneralOffset determines the number of bytes offset
                            from IReg the packed string is to be placed.
                            Pack determines the number of packed characters
                            must be stored at location (Reg+GeneralOffset).
                            If the string, b, contains less characters than
                            Pack then the remaining characters must be filled
                            in as nul.
*)

PROCEDURE MovePackedStringToIReg (Reg: Register; GeneralOffset: INTEGER;
                                  b: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   Offset: INTEGER ;
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(b)
      THEN
         INC(Word, ORD(b[i]))
      END
   END ;
   WriteInstruction(mov) ;
   Write(' ') ;
   WritePrefix(GetUnitPrefix(Pack)) ;
   WriteString(' [') ;
   WriteRegister(Reg, MaxDatum()) ;
   WriteOffset(GeneralOffset) ;
   WriteString('], 0') ; WriteHex(Word, 0) ;
   WriteString('H ; ') ; DisplayString(b, Pack-1) ;
   WriteLn
END MovePackedStringToIReg ;


(*
   PushPackedString - moves a packed string b onto the stack.
                      Pack determines the number of packed characters
                      must be stored on the stack.
                      If the string, b, contains less characters than
                      Pack then the remaining characters must be filled
                      in as nul.
*)

PROCEDURE PushPackedString (b: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(b)
      THEN
         INC(Word, ORD(b[i]))
      END
   END ;
   WriteInstruction(push) ;
   Write(' ') ;
   WritePrefix(GetUnitPrefix(Pack)) ;
   WriteString(' 0') ;
   WriteHex(Word, 0) ;
   WriteString('H ; ') ;
   DisplayString(b, Pack-1) ;
   WriteLn
END PushPackedString ;


(*
   PutStringMDReg - moves a packed string a into memory offset, Offset,
                    indexed by register, Reg.
                    Pack determines the number of packed characters
                    to be stored.
                    If the string, b, contains less characters than
                    Pack then the remaining characters must be filled
                    in as nul.
*)

PROCEDURE PutStringMDReg (Offset: INTEGER; Reg: Register;
                          a: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(a)
      THEN
         INC(Word, ORD(a[i]))
      END
   END ;
   WriteInstruction(mov) ;
   Write(' ') ;
   WritePrefix(GetUnitPrefix(Pack)) ;
   WriteString(' [') ; WriteRegister(Reg, PDWord) ;
   WriteOffset(Offset) ;
   WriteString('], 0') ; WriteHex(Word, 0) ; WriteString('H ; ') ;
   DisplayString(a, Pack-1) ;
   WriteLn
END PutStringMDReg ;


(*
   DisplayString - displays a string, a, which contains, n, characters.
*)

PROCEDURE DisplayString (a: ARRAY OF CHAR; n: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   Write('"') ;
   FOR i := 0 TO n DO
      IF a[i]=nul
      THEN
         WriteString('<nul>')
      ELSE
         Write(a[i])
      END
   END ;
   WriteString('"   ( ') ;
   FOR i := n TO 0 BY -1 DO
      IF a[i]=nul
      THEN
         WriteString('<nul>')
      ELSE
         Write(a[i])
      END
   END ;
   WriteString(' )')
END DisplayString ;


(*
   FlushPipe - a dummy procedure which makes the interface to M2AS
               the same as the interface to M2Pipe586.
*)

PROCEDURE FlushPipe (cycles: CARDINAL) ;
BEGIN
END FlushPipe ;


BEGIN
   Init
END M2AS.
