/* Header for module M2Alias, generated by p2c */
#ifndef M2AliasH
#define M2AliasH


/*
    Title      : M2Alias
    Author     : Gaius Mulley
    System     : UNIX (gm2)
    Date       : Thu May 18 10:31:38 1995
    Last edit  : Thu May 18 10:31:38 1995
    Description: provides an alias mechanism for the Modula-2 symbols.
*/


/*
   Author     : Gaius Mulley
   Title      : SymbolTable
   Date       : 7/3/87
   Description: SymbolTable provides the higher level routines to
                maintain a symbol table for the Modula-2 Compiler.
   Last update: 23/5/87
*/

#ifndef SymbolTableH
#include "SymbolTable.h"
#endif

#ifndef M2QuadsH
#include "M2Quads.h"
#endif


#ifdef M2AliasG
# define vextern
#else
# define vextern extern
#endif



#define M2Alias_MaxAlias  100


typedef struct M2Alias_AliasType {
  unsigned long Sym;
  SymbolTable_ModeOfAddr Mode;
} M2Alias_AliasType;


/*
   InitAlias - creates a new empty AliasList and returns it.
*/


extern void *M2Alias_InitAlias(void);


/*
   KillAlias - destroys an AliasList.
*/

extern void *M2Alias_KillAlias(void *a);


/*
   DuplicateAlias - duplicates an AliasList and returns the duplicate.
*/

extern void *M2Alias_DuplicateAlias(void *a);


/*
   AliasSymbol - alias a symbol [FromSymbol, FromMode] -> [ToSymbol, ToMode]
*/

extern void M2Alias_AliasSymbol(void *a, unsigned long FromSymbol,
  SymbolTable_ModeOfAddr FromMode, unsigned long ToSymbol,
  SymbolTable_ModeOfAddr ToMode);


/*
   RemoveAliasTo - removes all alias' to a symbol in the AliasList.
*/

extern void M2Alias_RemoveAliasTo(void *a, unsigned long Sym,
				  SymbolTable_ModeOfAddr Mode);


/*
   CollectAllAliases - collect all aliases in, a.
*/

extern unsigned long M2Alias_CollectAllAliases(void *a, const long l_LOW,
  const long l_HIGH, M2Alias_AliasType *l);


/*
   CollectAliases - collect all aliases to Sym and Mode into the list, l.
*/

extern unsigned long M2Alias_CollectAliases(void *a, const long l_LOW,
  const long l_HIGH, M2Alias_AliasType *l, unsigned long Sym,
  SymbolTable_ModeOfAddr Mode);


/*
   IsAliased - returns TRUE whether there exists an alias in, a, which matches
               [FromSymbol, FromMode] -> [ToSymbol, ToMode]
*/

extern P2C_BOOLEAN M2Alias_IsAliased(void *a, unsigned long FromSymbol,
  SymbolTable_ModeOfAddr FromMode, unsigned long ToSymbol,
  SymbolTable_ModeOfAddr ToMode);


/*
   NoTimesUsed - returns the number of times a symbol, [s:m] was used (read and write).
*/

extern unsigned long M2Alias_NoOfTimesUsed(void *a, unsigned long s,
					   SymbolTable_ModeOfAddr m);


/*
   MakeDirty - make the symbol, [s:m] dirty.
*/

extern void M2Alias_MakeDirty(void *a, unsigned long s,
			      SymbolTable_ModeOfAddr m);


/*
   MakeClean - make the symbol, [s:m] clean.
*/

extern void M2Alias_MakeClean(void *a, unsigned long s,
			      SymbolTable_ModeOfAddr m);


/*
   IsDirty - returns TRUE if the symbol, [s:m] is dirty.
*/

extern P2C_BOOLEAN M2Alias_IsDirty(void *a, unsigned long s,
				   SymbolTable_ModeOfAddr m);


/*
   IsClean - returns TRUE if the symbol, [s:m] is clean.
*/

extern P2C_BOOLEAN M2Alias_IsClean(void *a, unsigned long s,
				   SymbolTable_ModeOfAddr m);


/*
   IsAbleToBeCleaned - returns TRUE if a symbol [Sym:Mode] can be cleaned (if it is made dirty).
*/

extern P2C_BOOLEAN M2Alias_IsAbleToBeCleaned(void *a, unsigned long s,
					     SymbolTable_ModeOfAddr m);


#undef vextern

#endif /*M2AliasH*/

/* End. */
