IMPLEMENTATION MODULE M2AsmUtil ;


FROM StdIO IMPORT Write ;
FROM StrIO IMPORT WriteString ;
FROM StrLib IMPORT StrCopy, StrConCat ;
FROM NameKey IMPORT WriteKey, GetKey, MakeKey ;

FROM SymbolTable IMPORT NulSym,
                        GetSymName,
                        GetScopeAuthor,
                        GetBaseModule,
                        IsInnerModule,
                        IsVar,
                        IsProcedure,
                        IsModule,
                        IsDefImp,
                        IsExportQualified,
                        IsExported ;

FROM M2Lexical IMPORT WriteError ;
FROM M2Configure IMPORT UseUnderscoreForC, UseDotForJumpLabels ;


CONST
   MaxString = 4096 ;


(* %%%FORWARD%%%
PROCEDURE WriteModulePrefix (Sym, ModSym: CARDINAL) ; FORWARD ;
PROCEDURE GetModulePrefix (VAR Name: ARRAY OF CHAR;
                           Sym, ModSym: CARDINAL) ; FORWARD ;
   %%%FORWARD%%% *)


(*
   DotifyLabel - place a dot infront of the label if necessary.
*)

PROCEDURE DotifyLabel (VAR a: ARRAY OF CHAR) ;
BEGIN
   IF UseDotForJumpLabels
   THEN
      StrConCat('.', a, a)
   END
END DotifyLabel ;


(*
   GetAsmName - returns the NameKey for the assembler string of a symbol.
*)

PROCEDURE GetAsmName (Sym: CARDINAL) : CARDINAL ;
VAR
   Module ,
   Name   : ARRAY [0..MaxString] OF CHAR ;
   Father : CARDINAL ;
BEGIN
   IF IsVar(Sym) OR IsProcedure(Sym)
   THEN
      Father := GetScopeAuthor(Sym) ;
      IF UseUnderscoreForC
      THEN
         StrCopy('_', Module)
      ELSE
         StrCopy('', Module)
      END ;
      GetModulePrefix(Module, Sym, Father) ;
      GetKey(GetSymName(Sym), Name) ;
      StrConCat(Module, Name, Name) ;
      RETURN( MakeKey(Name) )
   ELSE
      WriteError('Expecting Var or Procedure symbol')
   END
END GetAsmName ;


(*
   GetFullSymName - returns the NameKey for the symbol name (which also
                    may contain the module name). This is the same as
                    GetAsmName except that it does not have the leading _
*)

PROCEDURE GetFullSymName (Sym: CARDINAL) : CARDINAL ;
VAR
   Module ,
   Name   : ARRAY [0..MaxString] OF CHAR ;
   Father : CARDINAL ;
BEGIN
   IF IsVar(Sym) OR IsProcedure(Sym)
   THEN
      Father := GetScopeAuthor(Sym) ;
      StrCopy('', Module) ;
      GetModulePrefix(Module, Sym, Father) ;
      GetKey(GetSymName(Sym), Name) ;
      StrConCat(Module, Name, Name) ;
      RETURN( MakeKey(Name) )
   ELSE
      WriteError('Expecting Var or Procedure symbol')
   END
END GetFullSymName ;


(*
   WriteAsmName - displays the symbol, Sym, name using module prefixes
                  if it is EXPORT QUALIFIED.
*)

PROCEDURE WriteAsmName (Sym: CARDINAL) ;
VAR
   Father: CARDINAL ;
BEGIN
   IF IsVar(Sym) OR IsProcedure(Sym)
   THEN
      IF UseUnderscoreForC
      THEN
         Write('_')
      END ;
      Father := GetScopeAuthor(Sym) ;
      WriteModulePrefix(Sym, Father) ;
      WriteKey(GetSymName(Sym))
   ELSE
      WriteError('Expecting Var or Procedure symbol')
   END
END WriteAsmName ;


(*
   WriteName - displays the symbol, Sym, name using module prefixes
               if it is EXPORT QUALIFIED.
               This procedure differs from the above procedure because
               it does not generate any _ prefix.
*)

PROCEDURE WriteName (Sym: CARDINAL) ;
VAR
   Father: CARDINAL ;
BEGIN
   IF IsVar(Sym) OR IsProcedure(Sym)
   THEN
      Father := GetScopeAuthor(Sym) ;
      WriteModulePrefix(Sym, Father) ;
      WriteKey(GetSymName(Sym))
   ELSE
      WriteError('Expecting Var or Procedure symbol')
   END
END WriteName ;


(*
   WriteModulePrefix - writes the module prefix for module, ModSym,
                       providing symbol, Sym, is exported.
*)

PROCEDURE WriteModulePrefix (Sym, ModSym: CARDINAL) ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   StrCopy('', a) ;
   GetModulePrefix(a, Sym, ModSym) ;
   WriteString(a)
END WriteModulePrefix ;


(*
   GetModulePrefix - gets the module prefix for module, ModSym,
                     providing symbol, Sym, is exported.
*)

PROCEDURE GetModulePrefix (VAR Name: ARRAY OF CHAR;
                           Sym, ModSym: CARDINAL) ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IF (ModSym#NulSym) AND (ModSym#GetBaseModule())
   THEN
      IF IsInnerModule(Sym) OR IsInnerModule(ModSym)
      THEN
         GetKey(GetSymName(ModSym), a) ;
         StrConCat('_', Name, Name) ;
         StrConCat(a, Name, Name) ;
         GetModulePrefix(Name, ModSym, GetScopeAuthor(ModSym))
      ELSIF IsDefImp(ModSym) AND IsExportQualified(Sym)
      THEN
         GetKey(GetSymName(ModSym), a) ;
         StrConCat(Name, a, Name) ;
         StrConCat(Name, '_', Name)
      END
   END
END GetModulePrefix ;


(*
   UnderScoreString - emits a string with a leading underscore if the C compiler
                      uses _ prefixes.
*)

PROCEDURE UnderScoreString (a: ARRAY OF CHAR) ;
BEGIN
   IF UseUnderscoreForC
   THEN
      Write('_')
   END ;
   WriteString(a)
END UnderScoreString ;


(*
   GetModuleInitName - returns the name of the initialization section of a module.
*)

PROCEDURE GetModuleInitName (Sym: CARDINAL) : CARDINAL ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   GetKey(GetSymName(Sym), a) ;
   GetModulePrefix(a, Sym, GetScopeAuthor(Sym)) ;
   StrConCat('_M2_', a, a) ;
   StrConCat(a, '_init', a) ;
   IF UseUnderscoreForC
   THEN
      StrConCat('_', a, a)
   END ;
   RETURN( MakeKey(a) )
END GetModuleInitName ;


END M2AsmUtil.
