IMPLEMENTATION MODULE M2AssemIO ;


FROM ASCII IMPORT cr, nul ;
FROM StdIO IMPORT PushOutput, PopOutput ;
FROM FIO IMPORT OpenToWrite, Close, WriteChar, IsNoError, File, WriteNBytes ;
FROM NameKey IMPORT GetKey ;
FROM M2Lexical IMPORT WriteError ;
FROM M2Options IMPORT AssemblerFileSpecified, GetAssemblerFileName ;
FROM StrLib IMPORT StrConCat ;
FROM SYSTEM IMPORT ADR ;


(* %%%FORWARD%%%
PROCEDURE FlushBuffer ; FORWARD ;
   %%%FORWARD%%% *)


CONST
   MaxBuffer = 65536 ;

VAR
   Buffer      : ARRAY [0..MaxBuffer] OF CHAR ;
   BufferLimit : CARDINAL ;
   AssemblyFile: File ;    (* Assembly file created by the compiler. *)


(*
   AssemblyWrite - writes a character to the assembly file.
*)

PROCEDURE AssemblyWrite (ch: CHAR) ;
BEGIN
   IF ch#cr
   THEN
      Buffer[BufferLimit] := ch ;
      INC(BufferLimit) ;
      IF BufferLimit=MaxBuffer
      THEN
         FlushBuffer
      END
   END
END AssemblyWrite ;


(*
   FlushBuffer - flushes any outstanding characters to file descriptor AssemblyFile.
*)

PROCEDURE FlushBuffer ;
BEGIN
   IF BufferLimit>0
   THEN
      WriteNBytes(AssemblyFile, BufferLimit, ADR(Buffer)) ;
      IF NOT IsNoError(AssemblyFile)
      THEN
         WriteError('error while writing to the output file')
      END ;
      BufferLimit := 0
   END
END FlushBuffer ;


(*
   OpenAssemblyFile - opens the assembly language file.
*)

PROCEDURE OpenAssemblyFile (ModuleName: CARDINAL) ;
VAR
   a: ARRAY [0..80] OF CHAR ;
BEGIN
   IF AssemblerFileSpecified
   THEN
      GetAssemblerFileName(a)
   ELSE
      GetKey(ModuleName, a) ;
      StrConCat(a, '.s', a)
   END ;
   AssemblyFile := OpenToWrite(a) ;
   IF IsNoError(AssemblyFile)
   THEN
      BufferLimit := 0 ;
      PushOutput(AssemblyWrite)
   ELSE
      WriteError('cannot open output file')
   END
END OpenAssemblyFile ;


(*
   CloseAssemblyFile - closes the assembly file.
*)

PROCEDURE CloseAssemblyFile ;
BEGIN
   FlushBuffer ;
   Close(AssemblyFile) ;
   PopOutput
END CloseAssemblyFile ;


END M2AssemIO.
