IMPLEMENTATION MODULE M2Atom ;


FROM NameKey IMPORT WriteKey, NulName ;

FROM SymbolTable IMPORT GetSym ;

FROM M2Lexical IMPORT GetToken, PutTokenLineOffset,
                      GetTokenOffset,
                      CurrentToken, TokenType, GetLineNo,
                      TypeOfToken,
                      TypeToCard, CardToType,
                      WriteError,
                      TokenIs, IsToken ;

FROM M2Reserved IMPORT NulTok, PeriodTok, SingleQuoteTok, DoubleQuotesTok,
                       SemiColonTok, EndTok, ElsifTok, ElseTok, ThenTok,
                       RepeatTok, WhileTok, ForTok, DoTok, LoopTok, CaseTok, AsmTok,
                       OfTok, BarTok, LCBraTok, EofTok ;

FROM M2Quads IMPORT PushTF, PopTF, PushT, PopT ;

FROM P2SymBuild IMPORT BuildString, BuildNumber ;

FROM M2Debug IMPORT WriteDebug, Assert ;


(* %%%FORWARD%%%
PROCEDURE SpecialTerminators (LookingForTok, CurrentTok: CARDINAL;
      	       	     	      TokenType: TypeOfToken) : BOOLEAN ; FORWARD ;
   %%%FORWARD%%% *)


CONST
   MaxAutoStack = 20 ;

VAR
   On      : BOOLEAN ;
   Stack   : ARRAY [0..MaxAutoStack] OF BOOLEAN ;
   StackPtr: CARDINAL ;


(*
   PushAutoOn - after this statement all calls to ident will stack
                the ident name.
                The previous Auto flag is pushed onto the stack.
*)

PROCEDURE PushAutoOn ;
BEGIN
   IF StackPtr<=MaxAutoStack
   THEN
      Stack[StackPtr] := On ;
      INC(StackPtr)
   ELSE
      WriteError('Auto stack overflow - in M2Atom')
   END ;
   On := TRUE
END PushAutoOn ;


(*
   PushAutoOff - after this statement all calls to ident will not stack
                 the ident name.
                 The previous Auto flag is pushed onto the stack.
*)

PROCEDURE PushAutoOff ;
BEGIN
   IF StackPtr<=MaxAutoStack
   THEN
      Stack[StackPtr] := On ;
      INC(StackPtr)
   ELSE
      WriteError('Auto stack overflow - in M2Atom')
   END ;
   On := FALSE
END PushAutoOff ;



(*
   IsAutoPushOn - returns true if the Auto ident push flag is On.
*)

PROCEDURE IsAutoPushOn () : BOOLEAN ;
BEGIN
   RETURN( On )
END IsAutoPushOn ;


(*
   PopAuto - pops the previous pushed Auto flag into the Auto flag.
*)

PROCEDURE PopAuto ;
BEGIN
   IF StackPtr>0
   THEN
      DEC(StackPtr) ;
      On := Stack[StackPtr] ;
   ELSE
      WriteError('Auto stack underflow - in M2Atom')
   END
END PopAuto ;


(*
   Found - Searches through the next symbols for token, Tok.
           Returns true if token Tok is found before a
      	   ; | - + * /
           END, ELSE, THEN ELSIF, OF token is encountered.
           The symbols input is restored to the state
           before entry into this procedure.
*)

PROCEDURE Found (Tok: CARDINAL) : BOOLEAN ;
VAR
   Start : CARDINAL ; (* Used ONLY for debugging *)
   Quotes: CARDINAL ;
   Token : CARDINAL ;
   Type  : CARDINAL ;
   Line  : CARDINAL ;
   found : BOOLEAN ;
BEGIN
   Start := CurrentToken ;
   PushT(NulName) ;
   PushTF(NulName, NulName) ;  (* End of the list of pushed tokens *)
   WHILE (CurrentToken#Tok) AND
         (CurrentToken#SemiColonTok) AND
         (CurrentToken#BarTok) AND
         (CurrentToken#EndTok) AND
         (CurrentToken#ElseTok) AND
         (CurrentToken#ThenTok) AND
         (CurrentToken#ElsifTok) AND
         (CurrentToken#OfTok) AND
         (CurrentToken#DoTok) AND
         (CurrentToken#LoopTok) AND
         (CurrentToken#RepeatTok) AND
         (CurrentToken#ForTok) AND
         (CurrentToken#WhileTok) AND
         (CurrentToken#CaseTok) AND
         (CurrentToken#AsmTok) AND
         (CurrentToken#EofTok) AND
      	 (NOT SpecialTerminators(Tok, CurrentToken, TokenType))
   DO
      IF (CurrentToken=SingleQuoteTok) OR (CurrentToken=DoubleQuotesTok)
      THEN
         Quotes := CurrentToken ;
         REPEAT
            PushT(GetLineNo()) ;
            PushTF(CurrentToken, TypeToCard(TokenType)) ; (* Save token *)
            GetToken
         UNTIL CurrentToken=Quotes
      END ;
      PushT(GetLineNo()) ;
      PushTF(CurrentToken, TypeToCard(TokenType)) ;  (* Save Token *)
      GetToken
   END ;
   found := (CurrentToken=Tok) ;
   PushT(GetLineNo()) ;
   PushTF(CurrentToken, TypeToCard(TokenType)) ;  (* Save Token *)
   PopTF(Token, Type) ;
   PopT(Line) ;
   REPEAT
      PutTokenLineOffset(Token, CardToType(Type), Line, GetTokenOffset()) ;  (* Restore tokens *)
      PopTF(Token, Type) ;
      PopT(Line) ;
   UNTIL Token=NulName ;
   GetToken ; (* Now back to the CurrentToken that we entered the procedure with *)
   Assert(Start=CurrentToken) ;
   RETURN( found )
END Found ;


(*
   SpecialTerminators - returns TRUE if we have found a special
                        terminator. The terminator depends upon
                        the LookingForTok token.
*)

PROCEDURE SpecialTerminators (LookingForTok, CurrentTok: CARDINAL;
      	       	     	      TokenType: TypeOfToken) : BOOLEAN ;
BEGIN
   IF LookingForTok=LCBraTok
   THEN
      (*
         We are looking for an { and so this is looking for a set
         We know that the set rule is [ Qualident ] { set }
         Hence we know that the only legal terminators before the
      	 { will be Identifiers and . which make up the Qualident.
      *)
      RETURN(
      	      (TokenType#IdentToken) AND (CurrentTok#PeriodTok)
      	    )
   ELSE
      RETURN( FALSE )
   END
END SpecialTerminators ;
   

(*
   Ident - returns true is a legal identifier was encountered.
           If true is returned and the Ident switch is On the
           the identifier found is pushed onto the stack.
*)

PROCEDURE Ident() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF TokenType=IdentToken
   THEN
      IF On
      THEN
         PushTF(CurrentToken, TypeToCard(TokenType))
      END ;
      GetToken ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Ident ;


(*
   Number - returns true if a legal number was encountered.
            If true is returned and the Ident switch is On the
            the number found is pushed onto the stack.
*)

PROCEDURE Number() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF (TokenType=IntegerToken) OR (TokenType=RealToken)
   THEN
      IF On
      THEN
         PushTF(CurrentToken, TypeToCard(TokenType)) ;
         BuildNumber
      END ;
      GetToken ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Number ;


(*
   String - returns true is a legal string was encountered.
            If true is returned and the Ident switch is On the
            the string found is pushed onto the stack.
*)

PROCEDURE String() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF TokenIs(SingleQuoteTok)
   THEN
      IF IsToken(SingleQuoteTok)
      THEN
         (* Found '' as a string *)
         IF On
         THEN
            PushTF(NulTok, TypeToCard(IdentToken)) ;
            BuildString
         END
      ELSE
         (* Found 'hello' as a string *)
         IF On
         THEN
            PushTF(CurrentToken, TypeToCard(IdentToken)) ;
            BuildString
         END ;
         GetToken
      END ;
      IF TokenIs(SingleQuoteTok)
      THEN
         Success := TRUE
      ELSE
         Success := FALSE ;
         WriteError('Syntax Error')
      END
   ELSIF TokenIs(DoubleQuotesTok)
   THEN
      IF IsToken(DoubleQuotesTok)
      THEN
         (* Found "" as a string *)
         IF On
         THEN
            PushTF(NulTok, TypeToCard(IdentToken)) ;
            BuildString
         END
      ELSE
         (* Found "hello" as a string *)
         IF On
         THEN
            PushTF(CurrentToken, TypeToCard(IdentToken)) ;
            BuildString
         END ;
         GetToken
      END ;
      IF TokenIs(DoubleQuotesTok)
      THEN
         Success := TRUE
      ELSE
         Success := FALSE ;
         WriteError('Syntax Error')
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END String ;


BEGIN
   On := TRUE ;
   StackPtr := 0
END M2Atom.
