IMPLEMENTATION MODULE M2Batch ;


FROM M2Debug IMPORT Assert ;

FROM SymbolTable IMPORT MakeModule, MakeDefImp, NulSym ;

FROM NameKey IMPORT GetKey, WriteKey ;

FROM StrIO IMPORT ReadString, WriteString, WriteLn ;

FROM NumberIO IMPORT WriteCard ;

FROM M2Lexical IMPORT WriteError ;


CONST
   MaxStack   = 100 ;
   MaxModules = 100 ;

TYPE
   Module = RECORD
               SymNo: CARDINAL ;
               Name : CARDINAL ;
            END ;

VAR
   Modules    : ARRAY [1..MaxModules] OF Module ;
   NoOfModules: CARDINAL ;
   Stack      : ARRAY [0..MaxStack] OF CARDINAL ;
   Ptr        : CARDINAL ;


(* %%%FORWARD%%%
PROCEDURE DisplayModules ; FORWARD ;
PROCEDURE Get (Name: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE Pop () : CARDINAL ; FORWARD ;
PROCEDURE Push (Sym: CARDINAL) ; FORWARD ;
PROCEDURE Put (Sym: CARDINAL; n: CARDINAL) ; FORWARD ;
   %%%FORWARD%%% *)

(*
   MakeProgramSource - is given a Name which is used to create a program module.
                       The program module will be placed onto the compilation
                       pending queue if it has not yet been compiled.
                       If the module has been compiled then no action is taken.
                       The Module Sym is returned.
*)

PROCEDURE MakeProgramSource (Name: CARDINAL) : CARDINAL ;
VAR
   Sym: CARDINAL ;
BEGIN
   Sym := Get(Name) ;
   IF Sym=NulSym
   THEN
      (* Neither been compiled or on the Pending Queue *)
      Sym := MakeModule(Name) ;
      Put(Sym, Name) ;
      Push(Sym)
   END ;
   RETURN( Sym )
END MakeProgramSource ;


(*
   MakeDefinitionSource - is given a Name which is used to create a Definition
                          module.
                          The Definition Module will be placed onto the
                          compilation pending queue if it has not yet been
                          compiled.
                          If the module has been compiled then no action is
                          taken. The Module Sym is returned.
*)

PROCEDURE MakeDefinitionSource (Name: CARDINAL) : CARDINAL ;
VAR
   Sym: CARDINAL ;
BEGIN
   Sym := Get(Name) ;
   IF Sym=NulSym
   THEN
      (* Neither been compiled or on the Pending Queue *)
      Sym := MakeDefImp(Name) ;
      Put(Sym, Name) ;
      Push(Sym)
   END ;
   RETURN( Sym )
END MakeDefinitionSource ;


(*
   MakeImplementationSource - is given a Name which is used to create an
                              implementation module.
                              The implementation Module will be placed onto
                              the compilation pending
                              queue if it has not yet been compiled.
                              If the module has been compiled then no
                              action is taken. The Module Sym is returned.
*)

PROCEDURE MakeImplementationSource (Name: CARDINAL) : CARDINAL ;
VAR
   Sym: CARDINAL ;
BEGIN
   Sym := Get(Name) ;
   IF Sym=NulSym
   THEN
      (* Neither been compiled or on the Pending Queue *)
      Sym := MakeDefImp(Name) ;
      Put(Sym, Name) ;
      Push(Sym)
   END ;
   RETURN( Sym )
END MakeImplementationSource ;


(*
   GetSource - returns with the symbol Sym of the next module to be compiled.
               If Sym returns with value 0 then no module should be compiled.
*)

PROCEDURE GetSource () : CARDINAL ;
BEGIN
   RETURN( Pop() )
END GetSource ;


(*
   GetModuleNo - returns with symbol number of the n th module which as
                 read in Pass 1.
*)

PROCEDURE GetModuleNo (n: CARDINAL) : CARDINAL ;
BEGIN
   Assert(n#0) ;
   IF n>NoOfModules
   THEN
      RETURN( NulSym )  (* No module *)
   ELSE
      RETURN( Modules[n].SymNo )
   END
END GetModuleNo ;


(*
   IsModuleKnown - returns TRUE if the, Name, matches a module.
*)

PROCEDURE IsModuleKnown (Name: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN( Get(Name)#NulSym )
END IsModuleKnown ;


(*
   Get - Returns the Module Sym of a module with name, Name.
*)

PROCEDURE Get (Name: CARDINAL) : CARDINAL ;
VAR
   i: CARDINAL ;
BEGIN
   i := 1 ;
   WHILE i<=NoOfModules DO
      IF Name=Modules[i].Name
      THEN
         RETURN( Modules[i].SymNo )
      ELSE
         INC(i)
      END
   END ;
   RETURN( NulSym )
END Get ;


PROCEDURE Put (Sym: CARDINAL; n: CARDINAL) ;
BEGIN
   INC(NoOfModules) ;
   IF NoOfModules>MaxModules
   THEN
      WriteError('Too many modules: increase MaxModules inside M2Batch') ;
      HALT
   ELSE
      WITH Modules[NoOfModules] DO
         SymNo := Sym ;
         Name := n
      END
   END
END Put ;


PROCEDURE Push (Sym: CARDINAL) ;
BEGIN
   IF Ptr>MaxStack
   THEN
      WriteError('Too many outstanding compilations: increase MaxStack') ;
      HALT
   ELSE
      Stack[Ptr] := Sym ;
      INC(Ptr)
   END
END Push ;


PROCEDURE Pop () : CARDINAL ;
BEGIN
   IF Ptr=0
   THEN
      RETURN( NulSym )
   ELSE
      DEC(Ptr) ;
      RETURN( Stack[Ptr] )
   END
END Pop ;


PROCEDURE DisplayModules ;
VAR
   i: CARDINAL ;
BEGIN
   i := 1 ;
   WHILE i<=NoOfModules DO
      WriteString('Module') ; WriteKey(Modules[i].Name) ; WriteCard(i, 4) ;
      WriteLn ;
      INC(i)
   END
END DisplayModules ;


BEGIN
   Ptr := 0 ;
   NoOfModules := 0
END M2Batch.
