IMPLEMENTATION MODULE M2EvalSym ;

FROM M2Configure IMPORT ReturnArgumentOnStack, PushParametersLeftToRight,
                        ActivationRecordOffset ;

FROM M2Lexical IMPORT InternalError, WriteErrorFormat2 ;

FROM M2Debug IMPORT Assert ;
FROM M2Options IMPORT CompilerDebugging, APUCodeGenerator, Alignment ;
FROM M2System IMPORT WordSize ;

FROM M2ALU IMPORT Pop, Push, Add, Sub, Mult,
                  PushCard, PopCard, PopInt, PushInt ;

FROM M2Base IMPORT Cardinal, Unbounded, MixTypes, TemplateProcedure, ActivationPointer ;
FROM M2System IMPORT Address ;
FROM NumberIO IMPORT WriteCard, WriteInt ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM NameKey IMPORT WriteKey ;
FROM M2Lexical IMPORT WriteError, NearToken, NearTokens ;
FROM M2Error IMPORT BeginError, EndError ;
FROM i486 IMPORT DWORDSIZE, PentiumDataAlignmentSize ;

FROM SymbolTable IMPORT NulSym,
                        ModeOfAddr,
                        FinalSymbol,
                        GetMode,

                        IsDefImp,
                        IsModule,
                        IsUnknown,
                        IsType,
                        IsProcedure,
                        IsProcedureNested,
                        IsVarParam,
                        IsPointer,
                        IsRecord,
                        IsVarient,
                        IsArray,
                        IsRecordField,
                        IsFieldVarient,
                        IsEnumeration,
                        IsProcType,
                        IsVar,
                        IsVarAParam,
                        IsConst,
                        IsDummy,
                        IsSubrange,
                        IsSubscript,
                        IsUnbounded,
      	       	     	IsSet,
                        IsConstString,
                        IsUnboundedParam,

                        GetDeclared,
                        GetNth,
                        GetNthParam,
                        GetType,
                        GetSymName,
                        GetSubrange,
                        PutSubrange,

                        NoOfParam,
                        NoOfElements,

                        IsSizeSolved,
                        IsOffsetSolved,
                        IsValueSolved,
                        IsSumOfParamSizeSolved,
                        PushSize,
                        PushVarSize,
                        PushOffset,
                        PushValue,
                        PopValue,
                        PopSize,
                        PopOffset,
                        PopSumOfParamSize ;

FROM M2Quads IMPORT DisplayQuadList, Head ;


(* %%%FORWARD%%%
PROCEDURE CheckAlignment (Sym: CARDINAL; IsAParam: BOOLEAN) ; FORWARD ;
PROCEDURE Eval (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveArray (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveBlock (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveEnumeration (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveSet (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolvePointer (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveProcType (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveProcTypeParam (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveProcedure (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveProcedureParametersRightToLeft (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveProcedureParametersRightToLeft (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveRecord (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveRecordField (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveSubrange (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveType (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveUnbounded (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveVar (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveVarient (Sym: CARDINAL; Offset: CARDINAL) ; FORWARD ;
PROCEDURE SolveVarientField (Sym: CARDINAL; Offset: CARDINAL) ; FORWARD ;
PROCEDURE SolveProcedureParametersLeftToRight (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveProcedureParametersLeftToRightIntel (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveProcedureParametersRightToLeft (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveProcedureParametersRightToLeftIntel (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveProcedureParametersLeftToRightAPU (Sym: CARDINAL) ; FORWARD ;
PROCEDURE SolveProcedureParametersRightToLeftAPU (Sym: CARDINAL) ; FORWARD ;
   %%%FORWARD%%% *)


(*
   Evaluate - evaluate all the symbols in the symbol table.
*)

PROCEDURE Evaluate ;
VAR
   i: CARDINAL ;
BEGIN
   FOR i := 1 TO FinalSymbol() DO
      Eval(i)
   END
END Evaluate ;


(*
   Eval - decides which routine will solve a Symbol Sym.
*)

PROCEDURE Eval (Sym: CARDINAL) ;
BEGIN
   IF CompilerDebugging
   THEN
      WriteString('Solving ') ; WriteKey(GetSymName(Sym)) ; WriteLn
   END ;
   IF IsDummy(Sym)
   THEN
   ELSIF IsConst(Sym)
   THEN
   ELSIF IsEnumeration(Sym)
   THEN
      SolveEnumeration(Sym)
   ELSIF IsProcedure(Sym)
   THEN
      SolveProcedure(Sym)
   ELSIF IsModule(Sym)
   THEN
      SolveBlock(Sym)
   ELSIF IsDefImp(Sym)
   THEN
      SolveBlock(Sym)
   ELSIF IsVar(Sym)
   THEN
      SolveVar(Sym)
   ELSIF IsType(Sym)
   THEN
      SolveType(Sym)
   ELSIF IsSet(Sym)
   THEN
      SolveSet(Sym)
   ELSIF IsRecord(Sym)
   THEN
      SolveRecord(Sym)
   ELSIF IsRecordField(Sym)
   THEN
      SolveRecordField(Sym)
   ELSIF IsPointer(Sym)
   THEN
      SolvePointer(Sym)
   ELSIF IsArray(Sym)
   THEN
      SolveArray(Sym)
   ELSIF IsSubrange(Sym)
   THEN
      SolveSubrange(Sym)
   ELSIF IsProcType(Sym)
   THEN
      SolveProcType(Sym)
   ELSIF IsUnbounded(Sym)
   THEN
      SolveUnbounded(Sym)
   ELSE
      (* Do not necessarily need to solve every symbol *)
      (* since we can leave parameters alone..         *)
   END
END Eval ;


(*
   SolveEnumeration - sets the size of an enumeration type to TSIZE(CARDINAL)
*)

PROCEDURE SolveEnumeration (Sym: CARDINAL) ;
BEGIN
   NearToken('', GetDeclared(Sym)) ;
   IF NOT IsSizeSolved(Sym)
   THEN
      PushSize(Cardinal) ;
      PopSize(Sym)
   END
END SolveEnumeration ;


(*
   SolveSet - works out the runtime space required for set type Sym.
*)

PROCEDURE SolveSet (Sym: CARDINAL) ;
VAR
   Type: CARDINAL ;
BEGIN
   IF NOT IsSizeSolved(Sym)
   THEN
      NearToken('', GetDeclared(Sym)) ;
      Type := GetType(Sym) ;
      IF NOT IsSizeSolved(Type)
      THEN
      	 Eval(Type)
      END ;
      IF IsSubrange(Type)
      THEN
      ELSIF IsEnumeration(Type)
      THEN
      ELSE
         NearTokens('Set should be declared with subrange or enumeration type',
                    GetDeclared(Sym), GetDeclared(Type))
      END ;
      PushSize(Cardinal) ;   (* for the moment all sets are WORD size *)
      PopSize(Sym)
   END
END SolveSet ;


(*
   MakeOffsetAlign - aligns, Offset, by, Align.
*)

PROCEDURE MakeOffsetAlign (Offset: INTEGER; Align: CARDINAL) : CARDINAL ;
BEGIN
   Assert(Offset>=0) ;
   IF Offset MOD Align#0
   THEN
      Offset := Offset + (Align - (Offset MOD Align))
   END ;
   RETURN( Offset )
END MakeOffsetAlign ;


(*
   MakeUpToAlignment - given a value on the ALU stack, return the number of
                       bytes required to make the value up to the correct alignment
                       entity. The value on the ALU stack is removed
*)

PROCEDURE MakeUpToAlignment (Sym: CARDINAL; IsAParam: BOOLEAN) : CARDINAL ;
VAR
   i, j: INTEGER ;
BEGIN
   IF Alignment
   THEN
      i := PopInt() ;
      PushInt(i) ;
      CheckAlignment(Sym, IsAParam) ;
      j := PopInt() ;
      IF i>j
      THEN
         RETURN( CARDINAL(i-j) )
      ELSE
         RETURN( CARDINAL(j-i) )
      END
   ELSE
      RETURN( 0 )
   END
END MakeUpToAlignment ;


(*
   CheckAlignment - checks to see whether the value on the stack should be aligned.
                    It assumes absolute value should INCREASE.
*)

PROCEDURE CheckAlignment (Sym: CARDINAL; IsAParam: BOOLEAN) ;
VAR
   Size         : CARDINAL ;
   CurrentOffset: INTEGER ;
BEGIN
   IF (Sym#NulSym) AND (Alignment OR IsAParam)
   THEN
      CurrentOffset := PopInt() ;
      IF IsVar(Sym)
      THEN
         PushVarSize(Sym)
      ELSE
         PushSize(Sym)
      END ;
      Size := PopCard() ;
      IF (Size>=PentiumDataAlignmentSize) AND (NOT IsAParam)
      THEN
         IF CurrentOffset>0
         THEN
            CurrentOffset := MakeOffsetAlign(CurrentOffset, PentiumDataAlignmentSize)
         ELSE
            CurrentOffset := -MakeOffsetAlign(-CurrentOffset, PentiumDataAlignmentSize)
         END
      ELSIF (Size>=WordSize) OR IsAParam
      THEN
         IF CurrentOffset>0
         THEN
            CurrentOffset := MakeOffsetAlign(CurrentOffset, WordSize)
         ELSE
            CurrentOffset := -MakeOffsetAlign(-CurrentOffset, WordSize)
         END
      END ;
      PushInt(CurrentOffset)
   END
END CheckAlignment ;


(*
   CheckNestedProcedure - if Sym is a nested procedure then make room for
                          an activation record pointer which will contain
                          the scope of the last parental static procedure.
*)

PROCEDURE CheckNestedProcedure (Sym: CARDINAL) ;
VAR
   ActivationOffset: INTEGER ;
BEGIN
   IF IsProcedureNested(Sym)
   THEN
      PushSize(Address) ;    (* room for activation record pointer *)
      Sub ;
      ActivationOffset := PopInt() ;
      PushInt(ActivationOffset) ;
      IF ActivationOffset#ActivationRecordOffset*WordSize
      THEN
         InternalError('configuration incorrect activation offset constant differs from that being used',
                       __FILE__, __LINE__)
      END
   ELSIF Sym=TemplateProcedure
   THEN
      IF NOT IsOffsetSolved(ActivationPointer)
      THEN
         ActivationOffset := PopInt() ;
         Eval(ActivationPointer) ;
         PushInt(ActivationOffset) ;
         PushSize(ActivationPointer) ;
         Sub ;
         PopOffset(ActivationPointer) ;
         PushInt(ActivationOffset) ;
         PushOffset(ActivationPointer) ;
         IF ActivationRecordOffset*WordSize#PopInt()
         THEN
            BeginError ;
            WriteString('ActivationRecordOffset is ') ; WriteCard(ActivationRecordOffset*WordSize, 0) ; WriteLn ;
            WriteString('ActivationPointer is ') ; PushOffset(ActivationPointer) ; WriteInt(PopInt(), 0) ; WriteLn ;
            InternalError('configuration incorrect activation offset constant differs from that being used',
                          __FILE__, __LINE__)
         END ;

      END
   END
END CheckNestedProcedure ;


(*
   SolveProcedure - solves the size and offset of a procedures Parameters
                    and local variables.
*)

PROCEDURE SolveProcedure (Sym: CARDINAL) ;
BEGIN
   IF PushParametersLeftToRight
   THEN
      SolveProcedureParametersLeftToRight(Sym)
   ELSE
      SolveProcedureParametersRightToLeft(Sym)
   END
END SolveProcedure ;


(*
   SolveProcedureParametersLeftToRight - solves the size and offset of a
                                         procedures Parameters and local
                                         variables.
                                         It assumes that parameters are
                                         pushed left to right.
*)

PROCEDURE SolveProcedureParametersLeftToRight (Sym: CARDINAL) ;
BEGIN
   IF APUCodeGenerator
   THEN
      SolveProcedureParametersLeftToRightAPU(Sym)
   ELSE
      SolveProcedureParametersLeftToRightIntel(Sym)
   END
END SolveProcedureParametersLeftToRight ;


(*
   SolveProcedureParametersLeftToRightIntel - Intel stacks grow down.
                                              Return value in %eax or %ax.
                                              %eip and %ebp in between params
                                              and local variables.
*)

PROCEDURE SolveProcedureParametersLeftToRightIntel (Sym: CARDINAL) ;
VAR
   Offset : INTEGER ;
   i      : CARDINAL ;
   Var    : CARDINAL ;
   Return : CARDINAL ;
   ParamNo: CARDINAL ;
   init   : CARDINAL ;
BEGIN
   (* WriteString('; Procedure :') ; WriteKey(GetSymName(Sym)) ; WriteLn ; *)
   IF NOT IsSizeSolved(Sym)
   THEN
      (* Size so far *)
      (* First Check to see if Sym is a function *)
      (* WriteString(' Procedure ') ; *)
      ParamNo := NoOfParam(Sym) ;
      Return := GetType(Sym) ;
      (* WriteString('type ') ; WriteKey(GetSymName(Return)) ; *)
      i := 1 ;
      (* Parameter N starts at offset ebp+8 *)
      PushCard(0) ;
      PushSize(Address) ;   (* Frame Ptr      *)
      Add ;
      PushSize(Address) ;   (* Return Address *)
      Add ;
      init := PopCard() ;
      PushCard(init) ;
      WHILE i<=ParamNo DO
         Var := GetNth(Sym, i) ;
         IF Var#NulSym
         THEN
            Eval(Var) ;
            Eval(GetType(Var)) ;
            IF NOT IsOffsetSolved(Var)
            THEN
               PopOffset(Var) ;
               PushOffset(Var)
(*
               ; WriteString('; Procedure :') ; WriteKey(GetSymName(Sym)) ;
               ; WriteString(' : Variable ') ; WriteKey(GetSymName(Var)) ;
               ; WriteString(' Offset :') ; WriteInt(PopInt(), 4) ; PushOffset(Var)
               ; WriteInt(PopInt(), 4) ; PushOffset(Var)
               ; WriteString(' Size:') ; PushSize(Var) ; WriteInt(PopInt(), 4)
*)
            END ;
(*
            ; ; ; WriteLn ;
*)
            PushVarSize(Var) ;
            Add ;  (* Add onto running total *)
            CheckAlignment(Var, TRUE)
         END ;
         INC(i)
      END ;
      IF Return#NulSym
      THEN
         (* WriteString('Procedure has return type') ; WriteLn ; *)
         (* Sym is a function *)
         SolveType(Return)
      END ;
      Offset := PopInt() ;
      PushCard(Offset-init) ;
      PopSumOfParamSize(Sym) ;
      (*
         Now work out where the parameters and local variables live
         Parameters and local variables are contained in the same
         list as regards GetNth.
         Local Variables follow the Procedure Parameters.

         +-------------+
         | Parameters  |
         |-------------|
         | PC'         |
         |-------------|
         | ebp'        |
         |-------------|  <- ebp
         | Local Vars  |
         |-------------|

      *)
      (*
         Now continue working out where local variables are stored
      *)
      PushCard(0) ;
      CheckNestedProcedure(Sym) ; (* parental static procedure scope *)
      i := ParamNo ;
      REPEAT
         INC(i) ;
         Var := GetNth(Sym, i) ;
         IF Var#NulSym
         THEN
            Eval(Var) ;
            Eval(GetType(Var)) ;
            PushVarSize(Var) ;
            Sub ;  (* Add onto running total *)
            CheckAlignment(Var, FALSE) ;
            IF NOT IsOffsetSolved(Var)
            THEN
               PopOffset(Var) ;
               PushOffset(Var)
(*
               ; WriteString('; Procedure :') ; WriteKey(GetSymName(Sym)) ;
               ; WriteString(' : Variable ') ; WriteKey(GetSymName(Var)) ;
               ; WriteString(' Offset :') ; WriteInt(PopInt(), 4) ; PushOffset(Var)
               ; WriteInt(PopInt(), 4) ; PushOffset(Var)
               ; WriteString(' Size:') ; PushSize(Var) ; WriteInt(PopInt(), 4)
*)
            END ;
(*
            ; ; ; WriteLn
*)
         END
      UNTIL Var=NulSym ;
      PopSize(Sym)
   END
(*
   ; WriteString('End of procedure') ; WriteLn
*)
END SolveProcedureParametersLeftToRightIntel ;


(*
   SolveProcedureParametersRightToLeft - solves the size and offset of a
                                         procedures Parameters and local
                                         variables.
                                         It assumes that parameters are
                                         pushed right to left.
*)

PROCEDURE SolveProcedureParametersRightToLeft (Sym: CARDINAL) ;
BEGIN
   IF APUCodeGenerator
   THEN
      SolveProcedureParametersRightToLeftAPU(Sym)
   ELSE
      SolveProcedureParametersRightToLeftIntel(Sym)
   END
END SolveProcedureParametersRightToLeft ;


(*
   SolveProcedureParametersRightToLeftIntel - Intel stacks grow down.
                                              Return value in %eax or %ax.
                                              %eip and %ebp in between params
                                              and local variables.
*)

PROCEDURE SolveProcedureParametersRightToLeftIntel (Sym: CARDINAL) ;
VAR
   Offset : INTEGER ;
   i      : CARDINAL ;
   Var    : CARDINAL ;
   Return : CARDINAL ;
   ParamNo: CARDINAL ;
   init   : CARDINAL ;
BEGIN
   (* WriteString('; Procedure :') ; WriteKey(GetSymName(Sym)) ; WriteLn ; *)
   IF NOT IsSizeSolved(Sym)
   THEN
      (* Size so far *)
      (* First Check to see if Sym is a function *)
      (* WriteString(' Procedure ') ; *)
      ParamNo := NoOfParam(Sym) ;
      Return := GetType(Sym) ;
      (* WriteString('type ') ; WriteKey(GetSymName(Return)) ; *)
      i := ParamNo ;
      (* Parameter N starts at offset ebp+8 *)
      PushCard(0) ;
      PushSize(Address) ;   (* Frame Ptr      *)
      Add ;
      PushSize(Address) ;   (* Return Address *)
      Add ;
      init := PopCard() ;
      PushCard(init) ;
      WHILE i>0 DO
         Var := GetNth(Sym, i) ;
         IF Var#NulSym
         THEN
            Eval(Var) ;
            Eval(GetType(Var)) ;
            IF NOT IsOffsetSolved(Var)
            THEN
               PopOffset(Var) ;
               PushOffset(Var)
(*
               ; WriteString('; Procedure :') ; WriteKey(GetSymName(Sym)) ;
               ; WriteString(' : Variable ') ; WriteKey(GetSymName(Var)) ;
               ; WriteString(' Offset :') ; WriteInt(PopInt(), 4) ; PushOffset(Var)
               ; WriteInt(PopInt(), 4) ; PushOffset(Var)
               ; WriteString(' Size:') ; PushSize(Var) ; WriteInt(PopInt(), 4)
*)
            END ;
(*
            ; ; ; WriteLn ;
*)
            PushVarSize(Var) ;
            Add ;  (* Add onto running total *)
            CheckAlignment(Var, TRUE)
         END ;
         DEC(i)
      END ;
      IF Return#NulSym
      THEN
         (* WriteString('Procedure has return type') ; WriteLn ; *)
         (* Sym is a function *)
         SolveType(Return)
      END ;
      Offset := PopInt() ;
      PushCard(Offset-init) ;
      PopSumOfParamSize(Sym) ;
      (*
         Now work out where the parameters and local variables live
         Parameters and local variables are contained in the same
         list as regards GetNth.
         Local Variables follow the Procedure Parameters.

         +-------------+
         | Parameters  |
         |-------------|
         | PC'         |
         |-------------|
         | ebp'        |
         |-------------|  <- ebp
         | Local Vars  |
         |-------------|

      *)
      (*
         Now continue working out where local variables are stored
      *)
      PushCard(0) ;
      CheckNestedProcedure(Sym) ; (* parental static procedure scope *)
      i := ParamNo ;
      REPEAT
         INC(i) ;
         Var := GetNth(Sym, i) ;
         IF Var#NulSym
         THEN
            Eval(Var) ;
            Eval(GetType(Var)) ;
            PushVarSize(Var) ;
            Sub ;  (* Add onto running total *)
            CheckAlignment(Var, FALSE) ;
            IF NOT IsOffsetSolved(Var)
            THEN
               PopOffset(Var) ;
               PushOffset(Var)
(*
               ; WriteString('; Procedure :') ; WriteKey(GetSymName(Sym)) ;
               ; WriteString(' : Variable ') ; WriteKey(GetSymName(Var)) ;
               ; WriteString(' Offset :') ; WriteInt(PopInt(), 4) ; PushOffset(Var)
               ; WriteInt(PopInt(), 4) ; PushOffset(Var)
               ; WriteString(' Size:') ; PushSize(Var) ; WriteInt(PopInt(), 4)
*)
            END ;
(*
            ; ; ; WriteLn
*)
         END
      UNTIL Var=NulSym ;
      PopSize(Sym)
   END
(*
   ; WriteString('End of procedure') ; WriteLn
*)
END SolveProcedureParametersRightToLeftIntel ;


(*
   SolveProcedureParametersLeftToRightAPU - APU stacks grow down.
                                            Return value on stack.
                                            %ip between params
                                            and local variables.
*)

PROCEDURE SolveProcedureParametersLeftToRightAPU (Sym: CARDINAL) ;
VAR
   Offset : INTEGER ;
   i      : CARDINAL ;
   Var    : CARDINAL ;
   Return : CARDINAL ;
   ParamNo: CARDINAL ;
   init   : CARDINAL ;
BEGIN
   IF NOT IsSizeSolved(Sym)
   THEN
      (* Parameter N starts at offset ra+8 *)
      PushCard(0) ;
      PushSize(Address) ;   (* Frame Ptr      *)
      Add ;
      PushSize(Address) ;   (* Return Address *)
      Add ;
      init := PopCard() ;
      PushCard(init) ;
      (* First Check to see if Sym is a function *)
      ParamNo := NoOfParam(Sym) ;
      Return := GetType(Sym) ;
      IF Return#NulSym
      THEN
         (* Sym is a function *)
         SolveType(Return) ;
         IF ReturnArgumentOnStack
         THEN
            PushSize(Return) ;
            Add
         END
      END ;
      i := 1 ;
      WHILE i<=ParamNo DO
         Var := GetNth(Sym, i) ;
         IF Var#NulSym
         THEN
            Eval(Var) ;
            Eval(GetType(Var)) ;
            IF NOT IsOffsetSolved(Var)
            THEN
               PopOffset(Var) ;
               PushOffset(Var)
            END ;
            PushVarSize(Var) ;
            Add ;  (* Add onto running total *)
            CheckAlignment(Var, TRUE)
         END ;
         INC(i)
      END ;

      (*
         Now work out where the parameters and local variables live
         Parameters and local variables are contained in the same
         list as regards GetNth.
         Local Variables follow the Procedure Parameters.

         +-------------+
         | Parameters  |
         |-------------|
         | pc'         |
         |-------------|
         | ra'         |   <- ra
         |-------------|
         | Local Vars  |
         |-------------|
      *)

      Offset := PopInt() ;  (* chuck value away now *)
      PushCard(Offset-init) ;
      PopSumOfParamSize(Sym) ;

      PushCard(0) ;         (* now work down the stack for local vars *)
      CheckNestedProcedure(Sym) ; (* parental static procedure scope *)

      (*
         Now continue working out where local variables are stored
      *)

      i := ParamNo ;
      REPEAT
         INC(i) ;
         Var := GetNth(Sym, i) ;
         IF Var#NulSym
         THEN
            Eval(Var) ;
            Eval(GetType(Var)) ;
            PushVarSize(Var) ;
            Sub ;  (* Add onto running total *)
            CheckAlignment(Var, FALSE) ;
            IF NOT IsOffsetSolved(Var)
            THEN
               PopOffset(Var) ;
               PushOffset(Var)
            END
         END
      UNTIL Var=NulSym ;
      PopSize(Sym)
   END
END SolveProcedureParametersLeftToRightAPU ;


(*
   SolveProcedureParametersRightToLeftAPU - APU stacks grow down.
                                            Return value on stack.
                                            %ip between params
                                            and local variables.
*)

PROCEDURE SolveProcedureParametersRightToLeftAPU (Sym: CARDINAL) ;
VAR
   Offset : INTEGER ;
   i      : CARDINAL ;
   Var    : CARDINAL ;
   Return : CARDINAL ;
   ParamNo: CARDINAL ;
   init   : CARDINAL ;
BEGIN
   (* WriteString('; Procedure :') ; WriteKey(GetSymName(Sym)) ; WriteLn ; *)
   IF NOT IsSizeSolved(Sym)
   THEN
      (* Parameter N starts at offset ra+8 *)
      PushCard(0) ;
      PushSize(Address) ;   (* Frame Ptr      *)
      Add ;
      PushSize(Address) ;   (* Return Address *)
      Add ;
      init := PopCard() ;
      PushCard(init) ;

      (* First Check to see if Sym is a function *)
      (* WriteString(' Procedure ') ; *)
      (* WriteString('type ') ; WriteKey(GetSymName(Return)) ; *)
      ParamNo := NoOfParam(Sym) ;
      Return := GetType(Sym) ;
      IF Return#NulSym
      THEN
         (* Sym is a function *)
         SolveType(Return) ;
         IF ReturnArgumentOnStack
         THEN
            PushSize(Return) ;
            Add
         END
      END ;
      i := ParamNo ;
      WHILE i>=1 DO
         Var := GetNth(Sym, i) ;
         IF Var#NulSym
         THEN
            Eval(Var) ;
            Eval(GetType(Var)) ;
            IF NOT IsOffsetSolved(Var)
            THEN
               PopOffset(Var) ;
               PushOffset(Var)
            END ;
            PushVarSize(Var) ;
            Add ;  (* Add onto running total *)
            CheckAlignment(Var, TRUE)
         END ;
         INC(i)
      END ;
      Offset := PopInt() ;  (* chuck value away now *)
      PushCard(Offset-init) ;
      PopSumOfParamSize(Sym) ;

      PushCard(0) ;  (* now work down the stack for local vars *)
      CheckNestedProcedure(Sym) ; (* parental static procedure scope *)
      (*
         Now work out where the parameters and local variables live
         Parameters and local variables are contained in the same
         list as regards GetNth.
         Local Variables follow the Procedure Parameters.

         +-------------+
         | ReturnValue |
         |-------------|
         | Parameter 1 |
         |-------------|
         | Parameter 2 |
         |-------------|
         | Parameter N |
         |-------------|
         | pc'         |
         |-------------|
         | ra'         |
         |-------------|   <- ra
         | Local Vars  |
         |-------------|

      *)

      (*
         Now continue working out where local variables are stored
      *)

      i := ParamNo ;
      REPEAT
         INC(i) ;
         Var := GetNth(Sym, i) ;
         IF Var#NulSym
         THEN
            Eval(Var) ;
            Eval(GetType(Var)) ;
            PushVarSize(Var) ;
            Sub ;  (* Add onto running total *)
            CheckAlignment(Var, FALSE) ;
            IF NOT IsOffsetSolved(Var)
            THEN
               PopOffset(Var) ;
               PushOffset(Var)
(*
               ; WriteString('; Procedure :') ; WriteKey(GetSymName(Sym)) ;
               ; WriteString(' : Variable ') ; WriteKey(GetSymName(Var)) ;
               ; WriteString(' Offset :') ; WriteInt(PopInt(), 4) ; PushOffset(Var)
               ; WriteInt(PopInt(), 4) ; PushOffset(Var)
               ; WriteString(' Size:') ; PushSize(Var) ; WriteInt(PopInt(), 4) ; WriteLn
*)
            END
         END
      UNTIL Var=NulSym ;
      PopSize(Sym)
   END
(*
   ; WriteString('End of procedure') ; WriteLn
*)
END SolveProcedureParametersRightToLeftAPU ;


(*
   SolveBlock - solves the size and offset of a Modules Global Variables.
                Determines the position of each Global Variable.
*)

PROCEDURE SolveBlock (Sym: CARDINAL) ;
VAR
   i  : CARDINAL ;
   Var: CARDINAL ;
BEGIN
(*
   WriteString('Module :') ; WriteKey(GetSymName(Sym)) ; WriteLn ;
*)
   IF NOT IsSizeSolved(Sym)
   THEN
      (* Sym Must be a Module therefore data starts at offset 0 *)
      PushCard(0) ;
      i := 0 ;
      REPEAT
         INC(i) ;
         Var := GetNth(Sym, i) ;
         IF Var#NulSym
         THEN
            Eval(Var) ;
            CheckAlignment(Var, FALSE) ;
            IF NOT IsOffsetSolved(Var)
            THEN
               PopOffset(Var) ;
               PushOffset(Var)
(*
               ; WriteString('Variable ') ; WriteKey(GetSymName(Var)) ;
               ; WriteString(' :') ; WriteCard(PopCard(), 6) ; PushOffset(Var)
               ; WriteLn
*)
            END ;
            PushVarSize(Var) ;
            Add   (* Add onto running total *)
         END
      UNTIL Var=NulSym ;
      PopSize(Sym)
   END ;
   PushCard(0) ;
   PopOffset(Sym)
(*
 ; WriteString('End of Module') ; WriteLn
*)
END SolveBlock ;


(*
   SolveType - Solves the size of a declared Modula-2 Type.
*)

PROCEDURE SolveType (Sym: CARDINAL) ;
VAR
   Type: CARDINAL ;
BEGIN
   IF NOT IsSizeSolved(Sym)
   THEN
      IF IsPointer(Sym)
      THEN
         SolvePointer(Sym)               (* added this clause 1/9/94 *)
      ELSIF IsEnumeration(Sym)
      THEN
         SolveEnumeration(Sym)
      ELSIF IsSet(Sym)
      THEN
      	 SolveSet(Sym)
      ELSE
         Type := GetType(Sym) ;
         IF Type=NulSym
         THEN
            BeginError ;
            WriteString('Should Never Get Here since all Base Types') ;
            WriteString(' SHOULD be solved') ; WriteLn ;
            WriteString('The type name with a problem is ') ;
            WriteKey(GetSymName(Sym)) ; WriteLn ;
            NearToken('base type should be solved', GetDeclared(Sym)) ;
            EndError ;
            (*
               Actually the problem is that a constant has no type
               and when used as a source, the intemediate destination
               also has no type. Sort this out soon.
            *)
            DisplayQuadList(Head) ;
            HALT ;
            (* Assume Cardinal *)
            PushSize(Cardinal)
         ELSE
            Eval(Type) ;
            PushSize(Type)
         END ;
         PopSize(Sym)
      END
   END
END SolveType ;


(*
   SolveProcType - solves the size of procedure types.
*)

PROCEDURE SolveProcType (Sym: CARDINAL) ;
BEGIN
   IF NOT IsSizeSolved(Sym)
   THEN
      (* Procedure Types have the same size as address *)
      PushSize(Address) ;
      PopSize(Sym)
   END ;
   IF NOT IsSumOfParamSizeSolved(Sym)
   THEN
      SolveProcTypeParam(Sym)
   END
END SolveProcType ;


(*
   SolveProcTypeParam - actually only solves the sum of paramaters.
*)

PROCEDURE SolveProcTypeParam (Sym: CARDINAL) ;
VAR
   i,
   ParamNo: CARDINAL ;
   Param,
   Type,
   Return : CARDINAL ;
BEGIN
   PushCard(0) ;
   (* First Check to see if Sym is a function *)
   ParamNo := NoOfParam(Sym) ;
   Return := GetType(Sym) ;
   IF Return#NulSym
   THEN
      (* Sym is a function *)
      SolveType(Return) ;
      IF ReturnArgumentOnStack
      THEN
         PushSize(Return) ;
         Add ;
         CheckAlignment(Return, TRUE)
      END
   END ;
   i := 1 ;
   WHILE i<=ParamNo DO
      Param := GetNthParam(Sym, i) ;
      IF Param#NulSym
      THEN
         Eval(Param) ;
         Eval(GetType(Param)) ;
         IF IsUnboundedParam(Sym, i)
         THEN
            Eval(Unbounded) ;  (* Make sure this has been calculated *)
            PushSize(Unbounded) ;
            Type := Unbounded
         ELSIF IsVarParam(Sym, i)
         THEN
            Type := Address ;
            PushSize(Type)
         ELSE
            Type := GetType(Param) ;
            PushSize(Type)
         END ;
         Add ;  (* Add onto running total *)
         CheckAlignment(Type, TRUE)
      END ;
      INC(i)
   END ;
   PopSumOfParamSize(Sym)
END SolveProcTypeParam ;


(*
   SolveVar - solves the size of a variable.
*)

PROCEDURE SolveVar (Sym: CARDINAL) ;
BEGIN
   IF NOT IsSizeSolved(Sym)
   THEN
      NearToken('', GetDeclared(Sym)) ;
      SolveType(Sym)   (* The Size of this variable depends on its Type *)
   END
END SolveVar ;


(*
   SolvePointer - all pointers have size ADDRESS
*)

PROCEDURE SolvePointer (Sym: CARDINAL) ;
VAR
   Type: CARDINAL ;
BEGIN
   IF NOT IsSizeSolved(Sym)
   THEN
      (* Pointers always have size ADDRESS *)
      (* This solves recursive data types  *)
      NearToken('', GetDeclared(Sym)) ;
      PushSize(Address) ;
      PopSize(Sym)
   END
END SolvePointer ;


(*
   SolveRecord - determines the size of a record and the relative offsets
                 of all the fields.
*)

PROCEDURE SolveRecord (Sym: CARDINAL) ;
VAR
   i            : CARDINAL ;
   Field        : CARDINAL ;
   j            : CARDINAL ;  (* Delete this when working *)
BEGIN
   IF NOT IsSizeSolved(Sym)
   THEN
      NearToken('', GetDeclared(Sym)) ;
(*
      WriteString('Record ') ; WriteKey(GetSymName(Sym)) ; WriteLn ;
*)
      PushCard(0) ;  (* Size total so far *)
      i := 1 ;
      REPEAT
         Field := GetNth(Sym, i) ;
         IF Field#NulSym
         THEN
(*
            WriteString('Field #') ; WriteCard(i, 4) ; WriteLn ;
*)
            IF IsVarient(Field)
            THEN
               j := PopCard() ;
               SolveVarient(Field, j) ;
               PushCard(j)
            ELSE
               Eval(Field) ;
               IF NOT IsOffsetSolved(Field)
               THEN
(*
                  ; WriteString('Field ') ;
                  ; WriteKey(GetSymName(Field)) ;
*)
                  j := PopCard() ;
                  PushCard(j) ;
(*
                  ; WriteString(' offset') ; WriteCard(j, 6) ; WriteLn ;
*)
                  CheckAlignment(Field, FALSE) ;
                  PopOffset(Field) ;
                  PushOffset(Field)
               END
            END ;
            PushSize(Field) ;
            Add ;   (* Add onto running total *)
            INC(i)
         END ;
      UNTIL Field=NulSym ;
      PopSize(Sym)
(*
      ; PushSize(Sym) ;
      ; WriteString('Record Size is') ; WriteCard(PopCard(), 6) ; WriteLn
*)
   END
END SolveRecord ;


(*
   SolveRecordField - solves the size of a record field, Sym.
*)

PROCEDURE SolveRecordField (Sym: CARDINAL) ;
BEGIN
   (* Offset solved by SolveRecord *)
   NearToken('', GetDeclared(Sym)) ;
   SolveType(Sym)
END SolveRecordField ;


(*
   SolveVarientField - solves the size of a varient record field, Sym.
*)

PROCEDURE SolveVarientField (Sym: CARDINAL; Offset: CARDINAL) ;
VAR
   Size : CARDINAL ;
   i    : CARDINAL ;
   Field: CARDINAL ;
   j    : CARDINAL ;  (* Delete this when working *)
BEGIN
   IF NOT IsSizeSolved(Sym)
   THEN
      NearToken('', GetDeclared(Sym)) ;
(*
      ; WriteString('VarientField') ; WriteCard(Sym, 4) ; WriteLn ;
*)
      Size := 0 ;     (* Total size of VarientField so far *)
      i := 1 ;
      REPEAT
         Field := GetNth(Sym, i) ;
         IF Field#NulSym
         THEN
            IF IsVarient(Field)
            THEN
               SolveVarient(Field, Offset+Size)
            ELSE
               IF NOT IsOffsetSolved(Field)
               THEN
(*
                  ; WriteString('Field ') ;
                  ; WriteKey(GetSymName(Field)) ;
                  ; WriteString(' offset') ; WriteCard(Offset+Size, 6) ; WriteLn ;
*)
                  PushCard(Offset+Size) ;
                  PopOffset(Field)
               END ;
               Assert(IsRecordField(Field)) ;
               Eval(Field)
            END ;
            PushSize(Field) ;
            INC(Size, PopCard()) ;   (* Add onto running total *)
            INC(i)
         END ;
      UNTIL Field=NulSym ;
      PushCard(Size) ;
      PopSize(Sym)
(*
      ; PushSize(Sym) ;
      ; WriteString('VarientField Size is') ; WriteCard(PopCard(), 6) ; WriteLn
*)
   END
END SolveVarientField ;


(*
   SolveVarient - solves the size of a varient list Sym.
                  The offset values for each varient
                  are assigned the value Offset.
*)

PROCEDURE SolveVarient (Sym: CARDINAL; Offset: CARDINAL) ;
VAR
   Max, Size: CARDINAL ;
   i        : CARDINAL ;
   Field    : CARDINAL ;
BEGIN
   IF NOT IsSizeSolved(Sym)
   THEN
      NearToken('', GetDeclared(Sym)) ;
(*
      ; WriteString('Solving varient record') ; WriteLn ;
*)
      Max := 0 ;
      i := 1 ;
      Field := GetNth(Sym, i) ;
      WHILE Field#NulSym DO
         SolveVarientField(Field, Offset) ;
         IF NOT IsOffsetSolved(Field)
         THEN
            PushCard(Offset) ;
            PopOffset(Field)     (* All Varient Field have the same offset *)
         END ;
         PushSize(Field) ;
         Size := PopCard() ;
         IF Size>Max
         THEN
            Max := Size
         END ;
         INC(i) ;
         Field := GetNth(Sym, i)
      END ;
      PushCard(Max) ;
      PopSize(Sym)
(*
    ; WriteString('Varient record size is') ; WriteCard(Max, 4) ; WriteLn
*)
   END
END SolveVarient ;


(*
   SolveSubrange - attributes the size of a subrange to TSIZE(CARDINAL)
*)

PROCEDURE SolveSubrange (Sym: CARDINAL) ;
VAR
   High, Low, Type: CARDINAL ;
BEGIN
   IF NOT IsSizeSolved(Sym)
   THEN
      NearToken('', GetDeclared(Sym)) ;
(*
    ; WriteString('Solving subrange: ') ; WriteKey(GetSymName(Sym)) ;
*)
      GetSubrange(Sym, High, Low) ;
      Eval(High) ;
      Eval(Low) ;
      Type := MixTypes(GetType(High), GetType(Low)) ;
(*
      ; WriteString(' base type of subrange is: ') ; WriteKey(GetSymName(Type)) ;
*)
      PutSubrange(Sym, Low, High, Type) ;
      IF IsConstString(High) AND IsConstString(Low)
      THEN
         PushSize(Cardinal)
      ELSE
         Eval(Type) ;
         PushSize(Type)
      END ;
      PopSize(Sym)
(*
     ; PushSize(Sym) ;
     ; WriteString(' .. has size') ; WriteCard(PopCard(), 4) ; WriteLn ;
*)
   END
END SolveSubrange ;


(*
   SolveArray - Solves the Size and Offset of an Array Symbol, Sym.
                The Offset is at memory location for Sym[0,...,0].
*)

PROCEDURE SolveArray (Sym: CARDINAL) ;
VAR
   i        : CARDINAL ;
   Subscript,
   Subrange : CARDINAL ;
   j        : INTEGER ;  (* Delete this when working *)
   High, Low: CARDINAL ;
   Type     : CARDINAL ;
BEGIN
   IF NOT IsSizeSolved(Sym)
   THEN
      NearToken('', GetDeclared(Sym)) ;
(*
      ; WriteString('Solving ') ; WriteKey(GetSymName(Sym)) ;
*)
      Type := GetType(Sym) ;
      Eval(Type) ;
      PushSize(Type) ;  (* Size for element 1 *)
      i := 1 ;
      REPEAT
         Subscript := GetNth(Sym, i) ;
         IF Subscript#NulSym
         THEN
            Assert(IsSubscript(Subscript)) ;
            IF NOT IsSizeSolved(Subscript)
            THEN
(*
               ; WriteKey(GetSymName(Subscript)) ;
               j := PopInt() ;
               PushInt(j) ;
               WriteString(' Subscript Size') ; WriteInt(j, 6) ; WriteLn ;
*)
               PopSize(Subscript) ;
               PushSize(Subscript) ;
            END ;
            Subrange := GetType(Subscript) ;
            IF NOT IsSubrange(Subrange)
            THEN
               BeginError ;
               WriteString('Error with array ') ; WriteKey(GetSymName(Sym)) ;
               WriteString(' subscript:') ; WriteCard(i, 4) ; WriteLn ;
               WriteString('No Subrange for this subscript') ; WriteLn ;
               WriteString('Instead the TYPE given is ') ;
               WriteKey(GetSymName(Subrange)) ; WriteLn ;
               NearToken('No subrange for this subscript within array', GetDeclared(Sym)) ;
               EndError
            END ;
            Assert(IsSubrange(Subrange)) ;
            Eval(Subrange) ;
            GetSubrange(Subrange, High, Low) ;
            IF IsVar(High)
            THEN
               BeginError ;
               WriteString('Error with subrange, ') ; WriteKey(GetSymName(Subrange)) ;
               WriteString(' = [') ;
               WriteKey(GetSymName(Low)) ; WriteString('..') ; WriteKey(GetSymName(High)) ;
               WriteString(']  (high element is not a constant)') ; WriteLn ;
               NearTokens('', GetDeclared(High), GetDeclared(Subrange)) ;
               WriteError('Subrange must be declared with constants and not variables') ; WriteLn ;
               EndError
            END ;
            PushValue(High) ;
(*            j := PopInt() ; PushInt(j) ; WriteString(' High:') ; WriteInt(j, 4) ; *)
            PushInt(1) ;   (* Add 1 since a[x..y] is y+1-x elements.. *)
            Add ;
(*            j := PopInt() ; PushInt(j) ; WriteString(' High+1:') ; WriteInt(j, 4) ; *)
            IF IsVar(Low)
            THEN
               BeginError ;
               WriteString('Error with subrange, ') ; WriteKey(GetSymName(Subrange)) ;
               WriteString(' = [') ;
               WriteKey(GetSymName(Low)) ; WriteString('..') ; WriteKey(GetSymName(High)) ;
               WriteString(']  (low element is not a constant)') ; WriteLn ;
               NearTokens('', GetDeclared(Low), GetDeclared(Subrange)) ;
               WriteError('Subrange must be declared with constants and not variables') ; WriteLn ;
               EndError
            END ;
            PushValue(Low) ;
(*            j := PopInt() ; PushInt(j) ; WriteString(' Low:') ; WriteInt(j, 4) ; *)
            Sub ;
(*            WriteString(' Range evaluation...') ;
            j := PopInt() ; PushInt(j) ; WriteString(' Range:') ; WriteInt(j, 4) ; *)
            Mult   (* Running total of size so far *)
(*            ; j := PopInt() ; PushInt(j) ; WriteString(' Mult:') ; WriteInt(j, 4) ; WriteLn *)
         END ;
         INC(i)
      UNTIL Subscript=NulSym ;
      PopSize(Sym)
(*      ; PushSize(Sym) ; j := PopInt() ; WriteString('Array size:') ;
        WriteInt(j, 6) ; WriteLn *)
   END
END SolveArray ;


(*
   SolveUnbounded - Solves the Size of an Unbounded Symbol, Sym.
*)

PROCEDURE SolveUnbounded (Sym: CARDINAL) ;
BEGIN
   IF NOT IsSizeSolved(Sym)
   THEN
      (*
         The Type of Unbounded symbols is the simple type ie
         ARRAY OF Type which determines the element types of the array.

         The unbounded size is determined by the Base symbol Unbounded
      *)
      Eval(Unbounded) ;  (* Make sure this has been calculated *)
      PushSize(Unbounded) ;
      PopSize(Sym)
(*
      ; WriteKey(GetSymName(Sym)) ;
      ; PushSize(Sym) ; WriteString('Array size:') ;
        WriteCard(PopCard(), 6) ; WriteLn
*)
   END
END SolveUnbounded ;


END M2EvalSym.
