IMPLEMENTATION MODULE M2GDB ;

(*
   Author     : Gaius Mulley
   Title      : M2GDB
   Date       : 3/4/91  [$Date: 1999/07/02 14:31:39 $]
   SYSTEM     : UNIX SUN and Logitech M2
   Description: Implements the Debugging facilities in the Modula-2 compiler.
   Version    : $Revision: 1.13 $
*)

(*
   Log        : $Log: M2GDB.mod,v $
   Log        : Revision 1.13  1999/07/02 14:31:39  gaius
   Log        :  - improved M2GenDyn486 to perform more optimization on indexed
   Log        :    addressing with strings.
   Log        :  - finally added nested procedures (the abomination of desolation!)
   Log        :    and retrofitted them to all code generators and gdb stabs generation.
   Log        :  - m2f is now a full Modula-2 compiler loosely complying to
   Log        :    Programming in Modula-2 Second Corrected Edition by N. Wirth
   Log        :  - fixed bugs in M2GenAPU.mod
   Log        :
   Log        : Revision 1.12  1998/05/05 09:04:10  gaius
   Log        : fixed bugs in M2GenAPU concerning inner module names.
   Log        :
   Log        : Revision 1.11  1997/10/22 12:14:22  gaius
   Log        : fixed - nested modules initialization code
   Log        :       - inner module assembler variable names
   Log        :       - improved two error messages
   Log        :
   Log        : Revision 1.10  1997/08/26 07:49:29  gaius
   Log        : fixed modulus rounding error in fpu code generator (it now rounds down).
   Log        : fixed gdb displaying of procedure variables.
   Log        : improved error messages.
   Log        : added optional removal of M2RTS from list of linked modules (realtime usage).
   Log        : implemented many of the WriteCardinal, WriteShort etc inside FIO.mod
   Log        :
   Log        : Revision 1.9  1997/03/12 15:49:18  gaius
   Log        : finally clearing up the last? bugs in M2GenDyn486. Compiler now compiles
   Log        : itself with the new code generator, minor inefficiencies remain.
   Log        : Hopefully will work fix this soon. Swaping variables caused a problem,
   Log        : which has been fixed.
   Log        :
   Log        : Revision 1.8  1996/11/28 15:47:05  gaius
   Log        : major modifications to M2Quads the indirection quadruples have been
   Log        : added and the mode of symbols do not now indicate indirection.
   Log        :
   Log        : many error messages have been tidied up and more errors are caught.
   Log        :
   Log        : the i586 code generator has been started but is in a very early stage
   Log        : (M2GenDyn486.mod)
   Log        :
   Log        : Revision 1.6  1996/09/16 07:43:57  gaius
   Log        : compiler now checks function parameters and return values.
   Log        : Fixed bug in -apu code generator brought about via CSE
   Log        :
   Log        : Revision 1.5  1996/07/26 11:47:50  gaius
   Log        : error reporting improved and many bugfixes made. The compiler now
   Log        : generally points to the symbol in error. Many modules now use
   Log        : the WriteErrorFormat1 or its varient (pseudo printf style).
   Log        :
   Log        : Found a bug in the case handling, which was somehow was introduced..
   Log        :
   Log        : Added M2Constants which allow for constant comparison, really
   Log        : helps M2SubExp reduce the number of quadruples. 12%..15% on average code.
   Log        : Speed improvements of up to 25% on the compiler..
   Log        :
   Log        : Common subexpression elimiation is now included when -O is present.
   Log        : -Ocse turns it off.
   Log        :
   Log        : Revision 1.4  1996/06/26 11:09:56  gaius
   Log        : improved GDB support, line number information is more accurate.
   Log        : return statements and repeat until sequences now handle line numbers
   Log        : correctly.
   Log        : updated the .stabs information for slackware GDB (ELF?)
   Log        : compiler now operates correctly with GDB on ELF and non ELF machines.
   Log        : should self configure during a rebuild.
   Log        :
   Log        : Revision 1.3  1996/06/20 08:17:28  gaius
   Log        : now handles VAR parameters correctly. GDB will now display
   Log        : their value and not their address.
   Log        :
   Log        : Revision 1.2  1996/06/03 07:00:29  gaius
   Log        : Fixed debugging symbols for the type BYTE
   Log        : and also fixed a parameter debugging message (fails on too many parameters)
   Log        :
   Log        : Revision 1.1.1.1  1996/05/28 10:13:01  gaius
   Log        : Modula-2 compiler sources imported
   Log        :
*)


FROM ASCII IMPORT nul ;
FROM M2Debug IMPORT Assert ;
FROM StdIO IMPORT Write ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM StrLib IMPORT StrConCat ;
FROM NumberIO IMPORT WriteCard, WriteInt, CardToStr ;
FROM M2Options IMPORT GenerateDebugging, GenerateLineDebug ;
FROM NameKey IMPORT WriteKey, MakeKey, GetKey, NulName ;
FROM M2AsmUtil IMPORT WriteAsmName, WriteName, GetAsmName, GetFullSymName, UnderScoreString, GetModuleInitName ;
FROM M2FileName IMPORT CalculateFileName ;
FROM M2Configure IMPORT UseDotForGDBLabels, UseShortStabLineNumbers ;

FROM Lists IMPORT List, InitList, IncludeItemIntoList,
                  GetIndexOfList, PutItemIntoList, GetItemFromList,
      	       	  IsItemInList, NoOfItemsInList ;

FROM M2Quads IMPORT QuadOperator, GetQuad, GetNextQuad ;

FROM SymbolTable IMPORT NulSym,
                        ModeOfAddr,
                        GetMode,
                        GetScopeAuthor,
                        GetNth, GetType,
                        MakeType, PutType,
      	       	     	GetSubrange,
      	       	     	NoOfParam,
                        IsTemporary, IsUnbounded, IsEnumeration,
      	       	     	IsSubrange, IsPointer, IsRecord, IsArray,
                        IsProcedure, IsProcedureNested, IsModule, IsDefImp,
      	       	     	IsSubscript, IsVarient, IsFieldVarient,
      	       	     	IsType, IsProcType, IsSet, IsConst,
                        IsFieldEnumeration,
                        IsExported, IsImported,
      	       	     	GetMainModule, GetBaseModule, GetModule,
                        IsAModula2Type,
                        GetSymName,
                        ForeachLocalSymDo, ForeachFieldEnumerationDo,
      	       	     	ForeachProcedureDo, ForeachModuleDo,
                        ForeachInnerModuleDo,
                        PopSize,
                        PushSize,
                        PushOffset,
                        PushValue ;

FROM M2Base IMPORT Cardinal, Char, Proc, Integer, Unbounded, LongInt, Real, LongReal ;
FROM M2System IMPORT Word, Bitset, Byte ;
FROM M2ALU IMPORT PopCard, PopInt, PushCard ;
FROM M2Lexical IMPORT InternalError ;


(* %%%FORWARD%%%
PROCEDURE GDBProcType (Sym: CARDINAL) ; FORWARD ;
PROCEDURE LastProcedureInfo (Sym: CARDINAL) ; FORWARD ;
PROCEDURE InitLabel ; FORWARD ;
PROCEDURE NextLabel ; FORWARD ;
PROCEDURE BeginInfo ; FORWARD ;
PROCEDURE EndInfo ; FORWARD ;
PROCEDURE WriteBeginLabel ; FORWARD ;
PROCEDURE WriteEndLabel ; FORWARD ;
PROCEDURE FileInfo (Sym: CARDINAL) ; FORWARD ;
PROCEDURE LineInfo (Line: CARDINAL) ; FORWARD ;
PROCEDURE ProcedureInfo (Sym: CARDINAL; Line: CARDINAL) ; FORWARD ;
PROCEDURE VariableInfo (ModSym: CARDINAL; Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteUnboundedChar (Son: CARDINAL) ; FORWARD ;
PROCEDURE IsTemporaryVariable (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE LocalVariableInfo (Sym: CARDINAL) ; FORWARD ;
PROCEDURE TypeInfo ; FORWARD ;
PROCEDURE GDBUnbounded (Sym: CARDINAL) ; FORWARD ;
PROCEDURE CheckNulName (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteDefaultTypes ; FORWARD ;
PROCEDURE TypesInModule (Sym: CARDINAL) ; FORWARD ;
PROCEDURE TypesInProcedure (Sym: CARDINAL) ; FORWARD ;
PROCEDURE MakeTypeInfo (Sym: CARDINAL) ; FORWARD ;
PROCEDURE DoType (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteOutstandingTypes ; FORWARD ;
PROCEDURE WriteKindOfType (Sym: CARDINAL) ; FORWARD ;
PROCEDURE IsBaseType (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE WriteFieldEnumeration (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteEnumeration (Sym: CARDINAL) ; FORWARD ;
PROCEDURE GDBEnumeration (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteSet (Sym: CARDINAL) ; FORWARD ;
PROCEDURE GDBSet (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteSubrange (Sym: CARDINAL) ; FORWARD ;
PROCEDURE GDBSubrange (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WritePointer (Sym: CARDINAL) ; FORWARD ;
PROCEDURE GDBPointer (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteRecord (Sym: CARDINAL) ; FORWARD ;
PROCEDURE GDBRecord (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteVarient (Sym: CARDINAL) ; FORWARD ;
PROCEDURE GDBVarient (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteProcType (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteArray (Sym: CARDINAL) ; FORWARD ;
PROCEDURE GDBArray (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteType (Sym: CARDINAL) ; FORWARD ;
PROCEDURE GDBType (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteTypeNo (Sym: CARDINAL) ; FORWARD ;
PROCEDURE ResolveType (Sym: CARDINAL) ; FORWARD ;
PROCEDURE GetAnotherType () : CARDINAL ; FORWARD ;
PROCEDURE AllDependantsWritten (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE PrintType (a: ARRAY OF CHAR; Sym: CARDINAL; b: BOOLEAN) ; FORWARD ;
PROCEDURE IsEnumerationDependantsWritten (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE IsSubrangeDependantsWritten (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE IsPointerDependantsWritten (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE IsRecordDependantsWritten (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE IsVarientDependantsWritten (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE IsArrayDependantsWritten (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE IsProcTypeDependantsWritten (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE IsTypeDependantsWritten (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
   %%%FORWARD%%% *)


VAR
   GDBTypeList  : List ;     (* Contains a list of all types.  *)
                             (* The position in this list is   *)
      	                     (* used to generate the GDB info  *)
      	       	     	     (* number.                        *)
   DefinedList  : List ;     (* This list contains all types   *)
                             (* that have had their info       *)
                             (* generated.                     *)
   ToDoList     : List ;     (* Contains a list of all         *)
                             (* outstanding types that need to *)
                             (* be written to the assembly     *)
                             (* file when its dependants have  *)
                             (* been written.                  *)
   LabelCount   : CARDINAL ; (* The label number for BEGIN END *)
                             (* sequences.                     *)
   AnotherType  : CARDINAL ; (* The number of AnotherTypes     *)
                             (* that have been produced.       *)
   VoidType     : CARDINAL ;
   LastLine     : CARDINAL ; (* The last line emitted as       *)
                             (* .stabs. Used to prevent        *)
                             (* duplicate lines.               *)
   LastProcedure: CARDINAL ; (* The last procedure emitted     *)
                             (* used for line numbers.         *)
   LineTimes    : CARDINAL ; (* Number of times this line has  *)
                             (* appeared.                      *)



(*
   InitLabel - initializes the label count to one.
*)

PROCEDURE InitLabel ;
BEGIN
   LastProcedure := NulSym ;
   LastLine := 0 ;
   LineTimes := 0 ;
   LabelCount := 1 ;
   AnotherType := 0
END InitLabel ;


(*
   NextLabel - increments the label count by one.
*)

PROCEDURE NextLabel ;
BEGIN
   INC(LabelCount)
END NextLabel ;


(*
   Dot - emit a dot providing the configuration BOOLEAN was TRUE.
*)

PROCEDURE Dot ;
BEGIN
   IF UseDotForGDBLabels
   THEN
      Write('.')
   END
END Dot ;


(*
   BeginInfo - writes out the GDB information about a BEGIN
               statement.
*)

PROCEDURE BeginInfo ;
BEGIN
   IF GenerateDebugging OR GenerateLineDebug
   THEN
      WriteBeginLabel ; Write(':') ; WriteLn
   END
END BeginInfo ;


(*
   EndInfo - writes out the GDB information about an END statement.
*)

PROCEDURE EndInfo ;
BEGIN
   IF GenerateDebugging OR GenerateLineDebug
   THEN
      WriteEndLabel ; Write(':') ; WriteLn ;
      NextLabel
   END
END EndInfo ;


(*
   MainInfo - writes out the GDB information about the module
              _init function.
*)

PROCEDURE MainInfo (Sym: CARDINAL; Line: CARDINAL) ;
BEGIN
   IF GenerateDebugging OR GenerateLineDebug
   THEN
      WriteString('.stabs "') ;
      WriteKey(GetModuleInitName(Sym)) ;
      IF GetMainModule()=Sym
      THEN
         WriteString(':F')    (* exported *)
      ELSE
         WriteString(':f')    (* keep inner module initialization local *)
      END ;
      WriteCard(VoidType, 0) ;
      WriteString('",36,0,') ; WriteCard(Line, 0) ; Write(',') ; WriteKey(GetModuleInitName(Sym)) ;
      WriteLn ;
      WriteString('.stabs "') ; UnderScoreString('argc:p') ;
      WriteTypeNo(Integer) ; WriteString('",160,0,0,8') ; WriteLn ;
      WriteString('.stabs "') ; UnderScoreString('argv:p') ; WriteTypeNo(GetAnotherType()) ;
      WriteString('=*') ; WriteTypeNo(GetAnotherType()) ;
      WriteString('=*') ; WriteTypeNo(Char) ;
      WriteString('",160,0,0,12') ; WriteLn ;
      WriteString('.stabn 192,0,0,') ; WriteBeginLabel ; WriteLn ;
      WriteString('.stabn 224,0,0,') ; WriteEndLabel ; WriteLn ;
      (* we dont call NextLabel here as we expect the main code to follow directly *)
      LastProcedureInfo(Sym)
   END
END MainInfo ;


(*
   WriteBeginLabel - writes out the begin label.
*)

PROCEDURE WriteBeginLabel ;
BEGIN
   Dot ;
   WriteString('LBB') ; WriteCard(LabelCount, 0)
END WriteBeginLabel ;


(*
   WriteEndLabel - writes out the end label.
*)

PROCEDURE WriteEndLabel ;
BEGIN
   Dot ;
   WriteString('LBE') ; WriteCard(LabelCount, 0)
END WriteEndLabel ;


(*
   FileInfo - generates the .stabs information for the source file for
              module Sym.
*)

PROCEDURE FileInfo (Sym: CARDINAL) ;
VAR
   ModuleName,
   FileName  : ARRAY [0..80] OF CHAR ;
BEGIN
   IF GenerateDebugging OR GenerateLineDebug
   THEN
      GetKey(GetSymName(Sym), ModuleName) ;
      CalculateFileName(ModuleName, 'mod', FileName) ;
      WriteString('        .stabs "') ; WriteString(FileName) ;
      WriteString('",100,0,0,Ltext') ; WriteLn ;
      WriteString('.text') ; WriteLn ;
      WriteString('Ltext:') ; WriteLn
   END
END FileInfo ;


(*
   WriteLastProcedure - writes the last procedure name.
*)

PROCEDURE WriteLastProcedure ;
BEGIN
   IF LastProcedure#NulSym
   THEN
      IF IsProcedure(LastProcedure)
      THEN
         Write('-') ; WriteAsmName(LastProcedure)
      ELSIF IsModule(LastProcedure) OR IsDefImp(LastProcedure)
      THEN
         Write('-') ; WriteKey(GetModuleInitName(LastProcedure))
      END
   END
END WriteLastProcedure ;


(*
   LineInfo - generates the .stabs information for a new source line, Line.
*)

PROCEDURE LineInfo (Line: CARDINAL) ;
BEGIN
   IF GenerateDebugging OR GenerateLineDebug
   THEN
      IF UseShortStabLineNumbers
      THEN
         WriteString('    .stabd 68,0,') ; WriteCard(Line, 0) ; WriteLn
      ELSE
         IF Line>=LastLine
         THEN
            IF Line=LastLine
            THEN
               INC(LineTimes)
            ELSE
               LineTimes := 0
            END ;
            WriteString('.stabn 68,0,') ; WriteCard(Line, 0) ;
            Write(',') ; Dot ; WriteString('LM') ; WriteCard(Line, 0) ;
            Write('T') ; WriteCard(LineTimes, 0) ;
            WriteLastProcedure ;
            WriteLn ;
            Dot ; WriteString('LM') ; WriteCard(Line, 0) ;
            Write('T') ; WriteCard(LineTimes, 0) ;
            Write(':') ; WriteLn ;
            LastLine := Line
         END
      END
   END
END LineInfo ;


(*
   LastProcedureInfo - sets the last procedure variable to Sym.
                       This allows line number stabs to reference
                       the current procedure encountered.
*)

PROCEDURE LastProcedureInfo (Sym: CARDINAL) ;
BEGIN
   LastProcedure := Sym ;
   IF GenerateDebugging
   THEN
      WriteString('  .type  ') ;
      IF IsProcedure(Sym)
      THEN
         WriteAsmName(Sym)
      ELSE
         WriteKey(GetModuleInitName(Sym))
      END ;
      WriteString(',@function') ; WriteLn
   END
END LastProcedureInfo ;


(*
   ProcedureInfo - generates the .stabs information for a procedure, Sym.
*)

PROCEDURE ProcedureInfo (Sym: CARDINAL; Line: CARDINAL) ;
BEGIN
   IF GenerateDebugging OR GenerateLineDebug
   THEN
      WriteString('.stabs "') ; WriteName(Sym) ;
      Write(':') ;
      IF IsExported(GetMainModule(), Sym)
      THEN
         Write('F')
      ELSE
         Write('f')
      END ;
      IF (GetType(Sym)=NulSym) OR GenerateLineDebug
      THEN
      	 (* Procedure - no return type *)
      	 WriteCard(VoidType, 0)
      ELSE
      	 (* Function - write return type *)
      	 WriteTypeNo(GetType(Sym))
      END ;
      IF IsProcedureNested(Sym)
      THEN
         IF IsProcedure(GetScopeAuthor(Sym))
         THEN
            Write(',') ;
            WriteName(Sym) ;
            Write(',') ;
            WriteName(GetScopeAuthor(Sym))
         END
      END ;
      WriteString('",36,0,') ; WriteCard(Line, 0) ; Write(',') ; WriteAsmName(Sym) ;
      WriteLn ;
      LocalVariableInfo(Sym) ;
      WriteString('.stabn 192,0,0,') ; WriteBeginLabel ; WriteLn ;
      WriteString('.stabn 224,0,0,') ; WriteEndLabel ; WriteLn ;
      LastProcedureInfo(Sym)
   END
END ProcedureInfo ;


(*
   VariableInfo - generates the .stabs information for a global
                  variable, Sym, declared in module, ModSym.
*)

PROCEDURE VariableInfo (ModSym: CARDINAL; Sym: CARDINAL) ;
BEGIN
   IF GenerateDebugging AND (NOT IsTemporary(Sym))
   THEN
      WriteString('.stabs "') ;
      WriteKey(GetFullSymName(Sym)) ;   (* was GetSymName(Sym) *)
      Write(':') ;
      IF IsExported(ModSym, Sym)
      THEN
         Write('G')
      ELSE
	 Write('S')
      END ;
      IF GetMode(Sym)=LeftValue
      THEN
         Write('*')
      END ;
      WriteTypeNo(GetType(Sym)) ;
      IF IsExported(ModSym, Sym)
      THEN
	 WriteString('",32,0,3,0') ; WriteLn
      ELSE
	 WriteString('",40,0,3,') ;
      	 WriteAsmName(Sym) ; WriteLn
      END
   END
END VariableInfo ;


(*
   WriteUnboundedChar - writes an unbounded char.
      	       	     	This is a special case of unbounded array.
*)

PROCEDURE WriteUnboundedChar (Son: CARDINAL) ;
BEGIN
   Write('s') ;
   PushSize(Unbounded) ;
   WriteCard(PopCard(), 0) ;
   UnderScoreString('ArrayAddress:') ;
   IF GetType(Son)=Char
   THEN
      Write('*')
   END ;
   WriteTypeNo(GetType(Son)) ;
   WriteString(',0,32;') ; UnderScoreString('ArrayHigh:3,32,32;')
END WriteUnboundedChar ;


(*
   IsTemporaryVariable - returns TRUE if Sym is a temporary
                         variable.
*)

PROCEDURE IsTemporaryVariable (Sym: CARDINAL) : BOOLEAN ;
VAR
   a: ARRAY [0..255] OF CHAR ;
BEGIN
   GetKey(GetSymName(Sym), a) ;
   RETURN( (a[0]='_') AND (a[1]='T') )
END IsTemporaryVariable ;


(*
   LocalVariableInfo - generates GDB debugging information for all
                       local variables in procedure Sym.
*)

PROCEDURE LocalVariableInfo (Sym: CARDINAL) ;
VAR
   s, n,
   p, Son: CARDINAL ;
BEGIN
   n := 1 ;   (* Later collect function return param *)
   Son := GetNth(Sym, n) ;
   p := NoOfParam(Sym) ;
   WHILE Son#NulSym DO
      IF NOT IsTemporaryVariable(Son)
      THEN
         WriteString('.stabs "') ;
         WriteKey(GetSymName(Son)) ;
         Write(':') ;
         IF n<=p
         THEN
            (* It is really a parameter *)
            IF GetMode(Son)=LeftValue
            THEN
               Write('v')
            ELSE
               Write('p')
            END ;
(* was
            Write('p') ;
            IF GetMode(Son)=LeftValue
            THEN
               WriteTypeNo(GetAnotherType()) ;
               WriteString('=*')
            END ;
*)
      	    WriteTypeNo(GetType(Son)) ;
            WriteString('",160,0,0,')
      	 ELSE
            (* Its really a local variable *)
            IF GetMode(Son)=LeftValue
            THEN
               WriteTypeNo(GetAnotherType()) ;
               WriteString('=*')
            END ;
            WriteTypeNo(GetType(Son)) ;
            WriteString('",128,0,0,')
      	 END ;
      	 PushOffset(Son) ;
      	 WriteInt(PopInt(), 0) ;
      	 WriteLn
      END ;
      INC(n) ;
      Son := GetNth(Sym, n)
   END
END LocalVariableInfo ;


(*
   TypeInfo - generates the .stabs information for all types used.
*)

PROCEDURE TypeInfo ;
BEGIN
   IF GenerateLineDebug
   THEN
      WriteDefaultTypes
   END ;
   IF GenerateDebugging
   THEN
      WriteDefaultTypes ;
      ForeachModuleDo(TypesInModule) ;
      WriteOutstandingTypes
   END
END TypeInfo ;


(*
   GDBUnbounded - generates the unbounded GDB format info.
*)

PROCEDURE GDBUnbounded (Sym: CARDINAL) ;
BEGIN
   IF GetType(Sym)=Char
   THEN
      WriteUnboundedChar(Sym)
   ELSE
      GDBRecord(Unbounded)
   END ;
   Write(';')
END GDBUnbounded ;


(*
   CheckNulName - if Sym has a NulName then we print out its type directly.
*)
      
PROCEDURE CheckNulName (Sym: CARDINAL) ;
BEGIN
   IF (Sym#NulSym) AND (GetSymName(Sym)=NulName)
   THEN
      IF IsRecord(Sym)
      THEN
      	 Write('=') ;
      	 GDBRecord(Sym)
      ELSIF IsVarient(Sym)
      THEN
         (* do not emit inline varients as this confuses GDB *)
      ELSIF IsSet(Sym)
      THEN
      	 Write('=') ;
      	 GDBSet(Sym)
      ELSIF IsPointer(Sym)
      THEN
      	 Write('=') ;
      	 GDBPointer(Sym)
      ELSIF IsArray(Sym)
      THEN
      	 Write('=') ;
      	 GDBArray(Sym)
      ELSIF IsSubrange(Sym)
      THEN
      	 Write('=') ;
      	 GDBSubrange(Sym)
      ELSIF IsEnumeration(Sym)
      THEN
      	 Write('=') ;
      	 GDBEnumeration(Sym)
      ELSIF IsUnbounded(Sym)
      THEN
      	 Write('=') ;
      	 GDBUnbounded(Sym)
      ELSIF IsType(Sym)
      THEN
      	 Write('=') ;
      	 GDBType(Sym)
      ELSIF IsProcType(Sym)
      THEN
      	 Write('=') ;
         GDBProcType(Sym)
      ELSE
      	 InternalError('need to interpret another type', __FILE__, __LINE__)
      END
   END
END CheckNulName ;


(*
   WriteDefaultTypes - writes the default GDB types.
*)

PROCEDURE WriteDefaultTypes ;
BEGIN
(*
   WriteString('// default types') ; WriteLn ;

   WriteString('.stabs "int:t1=r1;-2147483648;2147483647;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "char:t2=r2;0;127;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "long int:t3=r1;-2147483648;2147483647;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "unsigned int:t4=r1;0;-1;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "long unsigned int:t5=r1;0;-1;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "short int:t6=r1;-32768;32767;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "long long int:t7=r1;0;-1;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "short unsigned int:t8=r1;0;65535;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "long long unsigned int:t9=r1;0;-1;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "signed char:t10=r1;-128;127;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "unsigned char:t11=r1;0;255;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "float:t12=r1;4;0;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "double:t13=r1;8;0;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "long double:t14=r1;8;0;",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;

   WriteString('.stabs "void:t15=15",128,0,0,0') ;
   WriteLn ;
   Sym := GetAnotherType() ;
   VoidType := 15 ;


   (* Integer *)
   IncludeItemIntoList(GDBTypeList, Integer) ;
   IncludeItemIntoList(DefinedList, Integer) ;
   WriteString('.stabs "INTEGER:t16=1;",128,0,0,0') ;
   WriteLn ;
   (* Char *)
   IncludeItemIntoList(GDBTypeList, Char) ;
   IncludeItemIntoList(DefinedList, Char) ;
   WriteString('.stabs "CHAR:t17=2;",128,0,0,0') ;
   WriteLn ;
   (* Cardinal *)
   IncludeItemIntoList(GDBTypeList, Cardinal) ;
   IncludeItemIntoList(DefinedList, Cardinal) ;
   WriteString('.stabs "CARDINAL:t18=4;",128,0,0,0') ;
   WriteLn ;
   (* Word *)
   IncludeItemIntoList(GDBTypeList, Word) ;
   IncludeItemIntoList(DefinedList, Word) ;
   WriteString('.stabs "WORD:t19=4;",128,0,0,0') ;
   WriteLn ;
   (* Proc *)
   WriteString('.stabs "PROC:') ;
   WriteTypeNo(GetAnotherType()) ;
   WriteString('=*') ;
   WriteTypeNo(GetAnotherType()) ;
   WriteString('=f15",128,0,0,0') ;
   WriteLn ;
   (* Bitset *)
   IncludeItemIntoList(GDBTypeList, Bitset) ;
   IncludeItemIntoList(DefinedList, Bitset) ;
   WriteString('.stabs "BITSET:t22=1;",128,0,0,0') ;
   WriteLn ;
   (* Byte *)
   IncludeItemIntoList(GDBTypeList, Byte) ;
   IncludeItemIntoList(DefinedList, Byte) ;
   WriteString('.stabs "BYTE:t23=2;",128,0,0,0') ;
   WriteLn
*)

   WriteString('// default types') ;
   WriteLn ;
   (* Integer *)
   IncludeItemIntoList(GDBTypeList, Integer) ;
   IncludeItemIntoList(DefinedList, Integer) ;
   WriteString('.stabs "INTEGER:t1=r1;-2147483648;2147483647;",128,0,0,0') ;
   WriteLn ;
   (* Char *)
   IncludeItemIntoList(GDBTypeList, Char) ;
   IncludeItemIntoList(DefinedList, Char) ;
   WriteString('.stabs "CHAR:t2=r2;0;127;",128,0,0,0') ;
   WriteLn ;
   (* Cardinal *)
   IncludeItemIntoList(GDBTypeList, Cardinal) ;
   IncludeItemIntoList(DefinedList, Cardinal) ;
   WriteString('.stabs "CARDINAL:t3=r1;0;-1;",128,0,0,0') ;
   WriteLn ;
   (* Word *)
   IncludeItemIntoList(GDBTypeList, Word) ;
   IncludeItemIntoList(DefinedList, Word) ;
   WriteString('.stabs "WORD:t4=r1;0;-1;",128,0,0,0') ;
   WriteLn ;
   (* Proc *)
   IncludeItemIntoList(GDBTypeList, Proc) ;
   IncludeItemIntoList(DefinedList, Proc) ;
   WriteString('.stabs "void:t5=5",128,0,0,0') ;
   WriteLn ;
   VoidType := 5 ;
   WriteString('.stabs "PROC:') ;
   WriteTypeNo(GetAnotherType()) ;
   WriteString('=*') ;
   WriteTypeNo(GetAnotherType()) ;
   WriteString('=f5",128,0,0,0') ;
   WriteLn ;
   (* Bitset *)
   IncludeItemIntoList(GDBTypeList, Bitset) ;
   IncludeItemIntoList(DefinedList, Bitset) ;
   WriteString('.stabs "BITSET:t8=r1;0;-1;",128,0,0,0') ;
   WriteLn ;
   (* Byte *)
   IncludeItemIntoList(GDBTypeList, Byte) ;
   IncludeItemIntoList(DefinedList, Byte) ;
   WriteString('.stabs "BYTE:t9=r1;0;255;",128,0,0,0') ;
   WriteLn ;
   (* LongInt *)
   IncludeItemIntoList(GDBTypeList, LongInt) ;
   IncludeItemIntoList(DefinedList, LongInt) ;
   WriteString('.stabs "LONGINT:t10=r1;01000000000000000000000;0777777777777777777777;",128,0,0,0') ;
   WriteLn ;
   (* Real *)
   IncludeItemIntoList(GDBTypeList, Real) ;
   IncludeItemIntoList(DefinedList, Real) ;
   WriteString('.stabs "REAL:t11=r1;4;0;",128,0,0,0') ;
   WriteLn ;
   (* LongInt *)
   IncludeItemIntoList(GDBTypeList, LongReal) ;
   IncludeItemIntoList(DefinedList, LongReal) ;
   WriteString('.stabs "LONGREAL:t12=r1;8;0;",128,0,0,0') ;
   WriteLn

END WriteDefaultTypes ;


(*
   TypesInModule - generate the .stabs information for all types within
                   module, Sym.
*)

PROCEDURE TypesInModule (Sym: CARDINAL) ;
BEGIN
(*
   WriteString('// Types defined in ') ;
   WriteKey(GetSymName(Sym)) ;
   WriteLn ;
*)
   ForeachLocalSymDo(Sym, MakeTypeInfo) ;
   ForeachProcedureDo(Sym, TypesInProcedure) ;
   ForeachInnerModuleDo(Sym, TypesInModule)
END TypesInModule ;


(*
   TypesInProcedure - generate the .stabs information for all types within
                      procedure, Sym.
*)

PROCEDURE TypesInProcedure (Sym: CARDINAL) ;
BEGIN
(*
   WriteString('// Types defined in ') ;
   WriteKey(GetSymName(Sym)) ;
   WriteLn ;
*)
   ForeachLocalSymDo(Sym, MakeTypeInfo)
END TypesInProcedure ;


(*
   MakeTypeInfo - generates type information for a type symbol, Sym.
      	       	  The type symbol, Sym, is placed into the DefinedList
      	       	  if it has not already been added.
*)

PROCEDURE MakeTypeInfo (Sym: CARDINAL) ;
BEGIN
   (* WriteString('// checking symbol ') ; WriteKey(GetSymName(Sym)) ;
   WriteLn ; *)
   IF IsAModula2Type(Sym)
   THEN
      DoType(Sym)
   END
END MakeTypeInfo ;


(*
   DoType - any type symbol, will be transformed into GDB style
            debugging information.
            It transforms the symbol, Sym, into debugging information.
            It also transforms all Syms dependants into debugging information.
*)

PROCEDURE DoType (Sym: CARDINAL) ;
BEGIN
   IF ((GetSymName(Sym)#NulName) AND (NOT IsItemInList(DefinedList, Sym)) OR
       IsVarient(Sym)) AND IsAModula2Type(Sym)
   THEN
      (* only write out information about types declared in this module *)
      IF AllDependantsWritten(Sym)
      THEN
         IncludeItemIntoList(DefinedList, Sym) ;
         WriteKindOfType(Sym)
      ELSE
      	 (* WriteString('// Putting ') ; WriteKey(GetSymName(Sym)) ;
      	 WriteString(' into list ToDoList') ; WriteLn ; *)
         IncludeItemIntoList(ToDoList, Sym)
      END
   END
END DoType ;


(*
   WriteOutandingTypes - writes out any types that have their dependants
      	       	     	 solved.
*)

PROCEDURE WriteOutstandingTypes ;
VAR
   i, n         : CARDINAL ;
   NoMoreWritten: BOOLEAN ;
   Sym          : CARDINAL ;
BEGIN
   REPEAT
      NoMoreWritten := TRUE ;
      n := NoOfItemsInList(ToDoList) ;
      i := 1 ;
      WHILE i<=n DO
      	 Sym := GetItemFromList(ToDoList, i) ;
      	 (* WriteString('// WriteOutstandingTypes: ') ;
      	 WriteKey(GetSymName(Sym)) ; *)
      	 IF NOT IsItemInList(DefinedList, Sym)
      	 THEN
      	    (* WriteString(' has not yet been defined') ; *)
      	    IF AllDependantsWritten(Sym)
      	    THEN
      	       (* WriteString(' and has its dependants written') ; *)
      	       IncludeItemIntoList(DefinedList, Sym) ;
      	       WriteKindOfType(Sym) ;
      	       NoMoreWritten := FALSE
      	    END
      	 END ;
      	 (* WriteLn ; *)
      	 INC(i)
      END
   UNTIL NoMoreWritten ;
   (*
      self checking code
   *)
   NoMoreWritten := TRUE ;
   n := NoOfItemsInList(ToDoList) ;
   i := 1 ;
   WHILE i<=n DO
      Sym := GetItemFromList(ToDoList, i) ;
      IF NOT IsItemInList(DefinedList, Sym)
      THEN
      	 WriteString('// HELP should have written type') ;
      	 WriteCard(Sym, 4) ; Write(' ') ; WriteKey(GetSymName(Sym)) ; WriteLn
      END ;
      INC(i)
   END
END WriteOutstandingTypes ;


(*
   WriteKindOfType - writes the type for a symbol, Sym.
*)

PROCEDURE WriteKindOfType (Sym: CARDINAL) ;
BEGIN
   IF IsEnumeration(Sym)
   THEN
      WriteEnumeration(Sym)
   ELSIF IsSubrange(Sym)
   THEN
      WriteSubrange(Sym)
   ELSIF IsPointer(Sym)
   THEN
      WritePointer(Sym)
   ELSIF IsRecord(Sym)
   THEN
      WriteRecord(Sym)
   ELSIF IsVarient(Sym)
   THEN
      WriteVarient(Sym)
   ELSIF IsFieldVarient(Sym)
   THEN
      WriteVarient(Sym)
   ELSIF IsArray(Sym)
   THEN
      WriteArray(Sym)
   ELSIF IsProcType(Sym)
   THEN
      WriteProcType(Sym)
   ELSIF IsSet(Sym)
   THEN
      WriteSet(Sym)
   ELSE
      WriteType(Sym)
   END
END WriteKindOfType ;


(*
   IsBaseType - returns true if a type, Sym, is a base type and
                we use predefined GDB information to represent this
                type.
*)

PROCEDURE IsBaseType (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN(
           (Sym=Cardinal) OR (Sym=Integer) OR
           (Sym=Char)     OR (Sym=Proc)
         )
END IsBaseType ;


(*
   WriteFieldEnumeration - displays the enumeration field together
                           with its value in GDB format.
*)

PROCEDURE WriteFieldEnumeration (Sym: CARDINAL) ;
BEGIN
   WriteKey(GetSymName(Sym)) ; Write(':') ;
   PushValue(Sym) ;
   WriteCard(PopCard(), 0) ;
   Write(',')
END WriteFieldEnumeration ;


(*
   WriteEnumeration - displays GDB information about an enumeration type.
*)

PROCEDURE WriteEnumeration (Sym: CARDINAL) ;
BEGIN
   WriteString('.stabs "') ; WriteKey(GetSymName(Sym)) ;
   WriteString(':t') ; WriteTypeNo(Sym) ;
   Write('=') ;
   GDBEnumeration(Sym) ;
   WriteString('",128,0,0,0') ; WriteLn
END WriteEnumeration ;


(*
   GDBEnumeration - writes out the GDB format for an enumeration type.
*)

PROCEDURE GDBEnumeration (Sym: CARDINAL) ;
BEGIN
   Write('e') ;
   ForeachFieldEnumerationDo(Sym, WriteFieldEnumeration) ;
   Write(';')
END GDBEnumeration ;


(*
   WriteSet - writes the GDB information for set type Sym.
*)

PROCEDURE WriteSet (Sym: CARDINAL) ;
BEGIN
   WriteString('.stabs "') ; WriteKey(GetSymName(Sym)) ;
   WriteString(':t') ; WriteTypeNo(Sym) ;
   Write('=') ;
   GDBSet(Sym) ;
   WriteString('",128,0,0,0') ; WriteLn
END WriteSet ;


(*
   GDBSet - writes out the Set GDB format information.
*)

PROCEDURE GDBSet (Sym: CARDINAL) ;
BEGIN
   (* GDB doesn't understand about sets so we use INTEGER for now *)
   WriteTypeNo(Integer)
END GDBSet ;


(*
   WriteSubrange - displays the subrange type in GDB format.
*)

PROCEDURE WriteSubrange (Sym: CARDINAL) ;
BEGIN
   WriteString('.stabs "') ; WriteKey(GetSymName(Sym)) ;
   WriteString(':t') ; WriteTypeNo(Sym) ;
   Write('=') ;
   GDBSubrange(Sym) ;
   WriteString('",128,0,0,0') ;
   WriteLn
END WriteSubrange ;


(*
   GDBSubrange - writes out the subrange info in GDB format.
*)

PROCEDURE GDBSubrange (Sym: CARDINAL) ;
VAR
   High, Low: CARDINAL ;
BEGIN
   GetSubrange(Sym, High, Low) ;
   WriteTypeNo(GetType(Sym)) ;
   Write(':') ;
   PushValue(Low) ;
   WriteInt(PopInt(), 0) ;
   Write(';') ;
   PushValue(High) ;
   WriteInt(PopInt(), 0) ;
   Write(';')
END GDBSubrange ;


(*
   WritePointer - displays gdb type information about pointer type, Sym.
*)

PROCEDURE WritePointer (Sym: CARDINAL) ;
BEGIN
   WriteString('.stabs "') ; WriteKey(GetSymName(Sym)) ;
   WriteString(':t') ; WriteTypeNo(Sym) ;
   Write('=') ;
   GDBPointer(Sym) ;
   WriteString('",128,0,0,0') ; WriteLn
END WritePointer ;


(*
   GDBPointer - writes the GDB format of a pointer type.
*)

PROCEDURE GDBPointer (Sym: CARDINAL) ;
BEGIN
   Write('*') ;
   WriteTypeNo(GetType(Sym)) ;
   Write(';')
END GDBPointer ;


(*
   WriteRecord - displays gdb type information about record type, Sym.
*)

PROCEDURE WriteRecord (Sym: CARDINAL) ;
BEGIN
   WriteString('.stabs "') ; WriteKey(GetSymName(Sym)) ;
   WriteString(':T') ; WriteTypeNo(Sym) ;
   Write('=') ;
   GDBRecord(Sym) ;
   WriteString(';",128,0,0,0') ; WriteLn ;
   WriteString('.stabs "') ; WriteKey(GetSymName(Sym)) ;
   WriteString(':t') ; WriteTypeNo(Sym) ;
   WriteString('",128,0,0,0') ; WriteLn
END WriteRecord ;


(*
   GDBRecord - writes only the record information this must be used
               to display a record within a record.
*)

PROCEDURE GDBRecord (Sym: CARDINAL) ;
VAR
   i     : CARDINAL ;
   Size,
   Field,
   Offset: CARDINAL ;
BEGIN
   Write('s') ;
   PushSize(Sym) ;
   WriteCard(PopCard(), 0) ;
   Offset := 0 ;  (* Offset in Bytes *)
   i := 1 ;
   REPEAT
      Field := GetNth(Sym, i) ;
      IF Field#NulSym
      THEN
      	 IF IsVarient(Field)
      	 THEN
            WriteKey(GetSymName(Field)) ;
      	    Write(':') ;
      	    WriteTypeNo(Field) ;

      	    Write('=') ;
      	    GDBVarient(Field) ;
      	    Write(';') ;

            (* we do not emit inline varient unions as this confuses GDB *)
            WriteString(',0')
      	 ELSIF IsFieldVarient(Field)
      	 THEN
            WriteKey(GetSymName(Field)) ;
      	    WriteString(' Should not get here') ; WriteLn ;
      	    WriteTypeNo(GetType(Field)) ;
      	    WriteString(',0')
      	 ELSE
            WriteKey(GetSymName(Field)) ;
      	    Write(':') ;
            WriteTypeNo(GetType(Field)) ;
            PushOffset(Field) ;
            Offset := PopCard() ;
            Write(',') ;
            WriteCard(Offset * 8, 0) ;
      	 END ;
         PushSize(Field) ;
         Size := PopCard() ;
         Write(',') ;
         WriteCard(Size * 8, 0) ;
         Write(';')
      END ;
      INC(i)
   UNTIL Field=NulSym
END GDBRecord ;


(*
   WriteVarient - displays gdb type information about varient
                  record type, Sym.
*)

PROCEDURE WriteVarient (Sym: CARDINAL) ;
VAR
   i     : CARDINAL ;
   Size,
   Field,
   Offset: CARDINAL ;
BEGIN
   WriteString('.stabs "') ; WriteKey(GetSymName(Sym)) ;
   WriteString(':T') ; WriteTypeNo(Sym) ;
   Write('=') ;
   GDBVarient(Sym) ;
   WriteString(';",128,0,0,0') ; WriteLn ;
   WriteString('.stabs "') ; WriteKey(GetSymName(Sym)) ;
   WriteString(':t') ; WriteTypeNo(Sym) ;
   WriteString('",128,0,0,0') ; WriteLn
END WriteVarient ;


(*
   GDBVarient - writes only the record information this must be used
                to display a record within a record.

                Sym will always be inside a RECORD so we need to calculate all offsets
                relative to this value.
*)

PROCEDURE GDBVarient (Sym: CARDINAL) ;
VAR
   ParentOffsetFound: BOOLEAN ;
   i, j             : CARDINAL ;
   Size,
   Field1,
   Field2,
   ParentOffset,
   Offset           : CARDINAL ;
BEGIN
   Write('u') ;
   PushSize(Sym) ;
   WriteCard(PopCard(), 0) ;
   ParentOffsetFound := FALSE ;
   ParentOffset := 0 ;
   Offset := 0 ;  (* Offset in Bytes *)
   i := 1 ;
   REPEAT
      Field1 := GetNth(Sym, i) ;
      IF Field1#NulSym
      THEN
      	 Assert(IsFieldVarient(Field1)) ;
      	 j := 1 ;
      	 REPEAT
      	    Field2 := GetNth(Field1, j) ;
      	    IF Field2#NulSym
      	    THEN
               WriteKey(GetSymName(Field2)) ;
               Write(':') ;
               WriteTypeNo(GetType(Field2)) ;
               Write(',') ;
               PushSize(Field2) ;
               Size := PopCard() ;
               PushOffset(Field2) ;
               Offset := PopCard() ;
               (*
                  we need to make sure that the offsets for GDB are from zero, whereas in effect they continue from
                  the record structure where they were declared

               IF NOT ParentOffsetFound
               THEN
                  ParentOffsetFound := TRUE ;
                  ParentOffset      := Offset
               END ;
               *)
               DEC(Offset, ParentOffset) ;
               WriteCard(Offset * 8, 0) ;
               Write(',') ;
               WriteCard(Size * 8, 0) ;
               Write(';') ;
      	       INC(j)
      	    END
      	 UNTIL Field2=NulSym
      END ;
      INC(i)
   UNTIL Field1=NulSym
END GDBVarient ;


(*
   GDBProcType - writes only the proc type gdb information.
*)

PROCEDURE GDBProcType (Sym: CARDINAL) ;
VAR
   ReturnType: CARDINAL ;
BEGIN
   ReturnType := GetType(Sym) ;
   Write('*') ;
   WriteTypeNo(GetAnotherType()) ;
   WriteString('=f') ;
   IF ReturnType=NulSym
   THEN
      WriteCard(VoidType, 0)
   ELSE
      WriteTypeNo(ReturnType)
   END
END GDBProcType ;


(*
   WriteProcType - writes out the GDB information for a proc type, Sym.
*)

PROCEDURE WriteProcType (Sym: CARDINAL) ;
BEGIN
   Assert(IsProcType(Sym)) ;
   WriteString('.stabs "') ; WriteKey(GetSymName(Sym)) ;
   WriteString(':t') ; WriteTypeNo(Sym) ;
   Write('=') ;
   GDBProcType(Sym) ;
   WriteString('",128,0,0,0') ;
   WriteLn
END WriteProcType ;


(*
   WriteArray - writes out the GDB information for an array, Sym.
*)

PROCEDURE WriteArray (Sym: CARDINAL) ;
VAR
   i        : CARDINAL ;
   Subscript,
   Subrange : CARDINAL ;
   High, Low: CARDINAL ;
   Type     : CARDINAL ;
BEGIN
   Assert(IsArray(Sym)) ;
   WriteString('.stabs "') ; WriteKey(GetSymName(Sym)) ;
   WriteString(':t') ; WriteTypeNo(Sym) ;
   Write('=') ;
   GDBArray(Sym) ;
   WriteString('",128,0,0,0') ; WriteLn
END WriteArray ;


(*
   GDBArray - writes out the GDB information for an array, Sym.
      	      This procedure writes the low level information about
      	      an array such as the number of elements and their types
      	      and their ranges.
	      For example: this procedure is used to write array information
      	      for an array in a record.
*)

PROCEDURE GDBArray (Sym: CARDINAL) ;
VAR
   i        : CARDINAL ;
   Subscript,
   Subrange : CARDINAL ;
   High, Low: CARDINAL ;
   Type     : CARDINAL ;
BEGIN
   Assert(IsArray(Sym)) ;

   Type := GetType(Sym) ;
   PushSize(Type) ;  (* Size for element 1 *)
   i := 1 ;
   REPEAT
      Subscript := GetNth(Sym, i) ;
      IF Subscript#NulSym
      THEN
      	 IF i>1
      	 THEN
      	    WriteTypeNo(GetAnotherType()) ;
      	    Write('=')
      	 END ;
         Assert(IsSubscript(Subscript)) ;
         Subrange := GetType(Subscript) ;
         IF NOT IsSubrange(Subrange)
         THEN
            WriteString('Error with array ') ; WriteKey(GetSymName(Sym)) ;
            WriteString(' subscript:') ; WriteCard(i, 4) ; WriteLn ;
            WriteString('No Subrange for this subscript') ; WriteLn ;
            WriteString('Instead the TYPE given is ') ;
            WriteKey(GetSymName(Subrange)) ; WriteLn
         END ;
         Assert(IsSubrange(Subrange)) ;
	 WriteString('ar1;') ;   (* ar1; low; high; Type of element *)
         GetSubrange(Subrange, High, Low) ;
         PushValue(Low) ;
      	 WriteInt(PopInt(), 0) ;
      	 Write(';') ;
         PushValue(High) ;
         WriteInt(PopInt(), 0) ;
      	 Write(';')
      ELSE
      	 WriteTypeNo(Type)  (* Type of array *)
      END ;
      INC(i)
   UNTIL Subscript=NulSym
END GDBArray ;


(*
   WriteType - displays gdb type information about type, Sym.
*)

PROCEDURE WriteType (Sym: CARDINAL) ;
BEGIN
   IF GetType(Sym)=NulSym
   THEN
      WriteString('// Base type ') ;
      WriteKey(GetSymName(Sym)) ;
      WriteString(' not understood') ;
      WriteLn
   ELSE
      WriteString('.stabs "') ; WriteKey(GetSymName(Sym)) ;
      WriteString(':t') ; WriteTypeNo(Sym) ;
      Write('=') ;
      WriteTypeNo(GetType(Sym)) ;
      WriteString('",128,0,0,0') ; WriteLn
   END
END WriteType ;


(*
   GDBType - writes the GDB format info for type, Sym.
*)

PROCEDURE GDBType (Sym: CARDINAL) ;
BEGIN
   IF GetType(Sym)=NulSym
   THEN
      WriteString('// base type ') ;
      WriteKey(GetSymName(Sym)) ;
      WriteString(' not understood') ;
      WriteLn
   ELSE
      WriteTypeNo(GetType(Sym)) ;
      Write(';')
   END
END GDBType ;


(*
   WriteTypeNo - writes the type number of type, Sym.
*)

PROCEDURE WriteTypeNo (Sym: CARDINAL) ;
BEGIN
   IF GetSymName(Sym)=NulName
   THEN
      ResolveType(Sym) ;
      WriteCard(GetIndexOfList(GDBTypeList, Sym)+1, 0) ;
      CheckNulName(Sym)
   ELSE
      IF NOT IsItemInList(GDBTypeList, Sym)
      THEN
      	 IncludeItemIntoList(GDBTypeList, Sym)
      END ;
      WriteCard(GetIndexOfList(GDBTypeList, Sym)+1, 0) ;
   END
END WriteTypeNo ;


(*
   ResolveType - marks a type, Sym, as resolved.
*)

PROCEDURE ResolveType (Sym: CARDINAL) ;
BEGIN
   IF NOT IsItemInList(GDBTypeList, Sym)
   THEN
      IncludeItemIntoList(GDBTypeList, Sym)
   END ;
   IF NOT IsItemInList(DefinedList, Sym)
   THEN
      IncludeItemIntoList(DefinedList, Sym)
   END
END ResolveType ;


(*
   GetAnotherType - returns a CARDINAL number representing a new
                    intemediate type used for defining PROC types.
*)

PROCEDURE GetAnotherType () : CARDINAL ;
VAR
   Name: CARDINAL ;
   Type: CARDINAL ;
   Str : ARRAY [0..20] OF CHAR ;
BEGIN
   CardToStr(AnotherType, 0, Str) ;
   INC(AnotherType) ;
   StrConCat('_AnotherType', Str, Str) ;
   Type := MakeType(MakeKey(Str)) ;
   PutType(Type, NulSym) ;
   PushCard(0) ;
   PopSize(Type) ;
   IncludeItemIntoList(GDBTypeList, Type) ;
   RETURN( Type )
END GetAnotherType ;


(*
   AllDependantsWritten - returns true if the symbol, Sym, and
      	       	     	  all its dependants have been written
      	       	     	  out to the assembly file.
*)

PROCEDURE AllDependantsWritten (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsEnumeration(Sym)
   THEN
      RETURN( IsEnumerationDependantsWritten(Sym) )
   ELSIF IsSubrange(Sym)
   THEN
      RETURN( IsSubrangeDependantsWritten(Sym) )
   ELSIF IsPointer(Sym)
   THEN
      RETURN( IsPointerDependantsWritten(Sym) )
   ELSIF IsRecord(Sym)
   THEN
      RETURN( IsRecordDependantsWritten(Sym) )
   ELSIF IsVarient(Sym)
   THEN
      RETURN( IsVarientDependantsWritten(Sym) )
   ELSIF IsArray(Sym)
   THEN
      RETURN( IsArrayDependantsWritten(Sym) )
   ELSIF IsProcType(Sym)
   THEN
      RETURN( IsProcTypeDependantsWritten(Sym) )
   ELSIF IsType(Sym)
   THEN
      RETURN( IsTypeDependantsWritten(GetType(Sym)) )
   ELSE
      RETURN( TRUE )
   END
END AllDependantsWritten ;


(*
   PrintType - prints out the type, Sym, with a leading string
      	       and its name followed by a boolean.
*)

PROCEDURE PrintType (a: ARRAY OF CHAR; Sym: CARDINAL; b: BOOLEAN) ;
BEGIN
(*
   WriteString('// ') ; WriteString(a) ;
   WriteString(' : Sym #') ; WriteCard(Sym, 4) ;
   WriteString(' Name: ') ; WriteKey(GetSymName(Sym)) ;
   IF b
   THEN
      WriteString(' TRUE')
   ELSE
      WriteString('FALSE ')
   END ;
   WriteLn
*)
END PrintType ;


(*
   IsEnumerationDependantsWritten - returns true if the enumeration
                                    dependants have been written to
                                    the assembly file.
*)

PROCEDURE IsEnumerationDependantsWritten (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   PrintType('Enumeration', Sym, TRUE) ;
   RETURN( TRUE )
END IsEnumerationDependantsWritten ;


(*
   IsSubrangeDependantsWritten - returns true if the subrange
                                 dependants have been written to the
                                 assembly file.
*)

PROCEDURE IsSubrangeDependantsWritten (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   PrintType('Subrange', Sym, IsItemInList(DefinedList, GetType(Sym))) ;
   RETURN( IsItemInList(DefinedList, GetType(Sym)) )
END IsSubrangeDependantsWritten ;


(*
   IsPointerDependantsWritten - returns TRUE if the pointer symbol, Sym,
      	       	                dependants have been solved.
*)

PROCEDURE IsPointerDependantsWritten (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   PrintType('Pointer', Sym, IsItemInList(DefinedList, GetType(Sym))) ;
   PrintType('Pointer type', GetType(Sym),
              IsItemInList(DefinedList, GetType(Sym))) ;
   RETURN( TRUE ) ;
   RETURN( IsItemInList(DefinedList, GetType(Sym)) )
END IsPointerDependantsWritten ;


(*
   IsRecordDependantsWritten - returns TRUE if the symbol, Sym,
      	       	               dependants have been written
      	       	               to the assembly file.
*)

PROCEDURE IsRecordDependantsWritten (Sym: CARDINAL) : BOOLEAN ;
VAR
   i     : CARDINAL ;
   Size,
   Field,
   Offset: CARDINAL ;
BEGIN
(*
   WriteString('// IsRecordDependantsWritten Sym=') ; WriteCard(Sym, 0) ;
   WriteString(' name ') ; WriteKey(GetSymName(Sym)) ; WriteLn ;
*)
   i := 1 ;
   REPEAT
      Field := GetNth(Sym, i) ;
      IF Field#NulSym
      THEN
         IF IsVarient(Field)
         THEN
(*
            PrintType('Varient ', Field, IsItemInList(DefinedList, Field)) ;
            IF NOT IsItemInList(DefinedList, Field)
            THEN
               IncludeItemIntoList(ToDoList, Field) ;
               RETURN( FALSE )
            END ;
*)
      	    PrintType('VarientField', Field,
      	       	       IsVarientDependantsWritten(Field)) ;
            IF NOT IsVarientDependantsWritten(Field)
            THEN
               RETURN( FALSE )
      	    END
      	 ELSIF IsFieldVarient(Field)
      	 THEN
      	    WriteString('Should not get here') ; WriteLn ;
      	    WriteTypeNo(GetType(Field)) ;
      	    WriteString(',0')
      	 ELSE
(*
      	    WriteString('// Field ') ; WriteKey(GetSymName(Field)) ;
      	    PrintType(' Field Type', GetType(Field),
      	       	     	IsTypeDependantsWritten(GetType(Field))) ;
*)
      	    IF NOT IsTypeDependantsWritten(GetType(Field))
      	    THEN
               RETURN( FALSE )
      	    END
      	 END
      END ;
      INC(i)
   UNTIL Field=NulSym ;
   RETURN( TRUE )
END IsRecordDependantsWritten ;


(*
   IsVarientWritten - returns TRUE if all symbol, Sym, dependants
      	       	      have been written to the assembly file.
*)

PROCEDURE IsVarientDependantsWritten (Sym: CARDINAL) : BOOLEAN ;
VAR
   i, j  : CARDINAL ;
   Field1,
   Field2: CARDINAL ;
BEGIN
   i := 1 ;
   REPEAT
      Field1 := GetNth(Sym, i) ;
      IF Field1#NulSym
      THEN
      	 Assert(IsFieldVarient(Field1)) ;
      	 j := 1 ;
      	 REPEAT
      	    Field2 := GetNth(Field1, j) ;
      	    IF Field2#NulSym
      	    THEN
     	       PrintType('VarientField', Field2,
      	       	     	  IsTypeDependantsWritten(GetType(Field2))) ;
      	       IF NOT IsTypeDependantsWritten(GetType(Field2))
      	       THEN
      	       	  RETURN( FALSE )
      	       END ;
      	       INC(j)
      	    END
      	 UNTIL Field2=NulSym
      END ;
      INC(i)
   UNTIL Field1=NulSym ;
   RETURN( TRUE )
END IsVarientDependantsWritten ;


(*
   IsArrayDependantsWritten - returns TRUE if the symbol, Sym,
      	       	              dependants have
      	       	              been written into the assembly file.
*)

PROCEDURE IsArrayDependantsWritten (Sym: CARDINAL) : BOOLEAN ;
VAR
   i        : CARDINAL ;
   Subscript,
   Subrange : CARDINAL ;
   High, Low: CARDINAL ;
   Type     : CARDINAL ;
BEGIN
   Assert(IsArray(Sym)) ;

   Type := GetType(Sym) ;
   PrintType('[Array] Type', Type, IsItemInList(DefinedList, Type)) ;
 
   IF IsItemInList(DefinedList, Type)
   THEN
      i := 1 ;
      REPEAT
      	 Subscript := GetNth(Sym, i) ;
      	 IF Subscript#NulSym
      	 THEN
            Assert(IsSubscript(Subscript)) ;
            Subrange := GetType(Subscript) ;
            IF NOT IsSubrange(Subrange)
            THEN
               WriteString('Error with array ') ; WriteKey(GetSymName(Sym)) ;
               WriteString(' subscript:') ; WriteCard(i, 4) ; WriteLn ;
               WriteString('No Subrange for this subscript') ; WriteLn ;
               WriteString('Instead the TYPE given is ') ;
               WriteKey(GetSymName(Subrange)) ; WriteLn
            END ;
            Assert(IsSubrange(Subrange)) ;
            GetSubrange(Subrange, High, Low) ;
            IF IsFieldEnumeration(High)
            THEN
               PrintType('[Array] High', High, IsItemInList(DefinedList, GetType(High))) ;
               IF NOT IsItemInList(DefinedList, GetType(High))
               THEN
                  RETURN( FALSE )
               END
            ELSIF NOT IsConst(High)
            THEN
               PrintType('[Array] High', High, IsItemInList(DefinedList, High)) ;
               IF NOT IsItemInList(DefinedList, High)
               THEN
                  RETURN( FALSE )
               END
            END ;
            IF IsFieldEnumeration(Low)
            THEN
               PrintType('[Array] Low', Low, IsItemInList(DefinedList, GetType(Low))) ;
               IF NOT IsItemInList(DefinedList, GetType(Low))
               THEN
                  RETURN( FALSE )
               END
            ELSIF NOT IsConst(Low)
            THEN
               PrintType('[Array] Low',  Low, IsItemInList(DefinedList, Low)) ;
               IF (NOT IsItemInList(DefinedList, Low))
               THEN
                  RETURN( FALSE )
               END
            END
      	 ELSE
      	    PrintType('[Array] Subrange', Type, IsItemInList(DefinedList, Type)) ;
      	    IF NOT IsItemInList(DefinedList, Type)
      	    THEN
      	       RETURN( FALSE )    (* Type of array *)
      	    END
      	 END ;
      	 INC(i)
      UNTIL Subscript=NulSym
   ELSE
      RETURN( FALSE )
   END ;
   RETURN( TRUE )
END IsArrayDependantsWritten ;


(*
   IsProcTypeDependantsWritten - 
*)

PROCEDURE IsProcTypeDependantsWritten (Sym: CARDINAL) : BOOLEAN ;
VAR
   ReturnType: CARDINAL ;
BEGIN
   Assert(IsProcType(Sym)) ;
   ReturnType := GetType(Sym) ;
   PrintType('[ProcType] ReturnType', ReturnType,
      	     IsItemInList(DefinedList, ReturnType)) ;
   RETURN( IsItemInList(DefinedList, ReturnType) )
END IsProcTypeDependantsWritten ;


(*
   IsTypeDependantsWritten - returns TRUE if a type symbol, Sym, dependants
      	       	             have been written to the assembly file.
      	       	     	     If a symbol has no name then it will NOT
                             be written out by itself but as part
                             of another symbol who does have a name.
                             Hence we can return TRUE for such symbols.
*)

PROCEDURE IsTypeDependantsWritten (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   IF (Sym#NulSym) AND (GetSymName(Sym)#NulName)
   THEN
      IF NOT IsItemInList(DefinedList, Sym)
      THEN
      	 RETURN( FALSE )
      END
   END ;
   RETURN( TRUE )
END IsTypeDependantsWritten ;


BEGIN
   InitList(GDBTypeList) ;
   InitList(DefinedList) ;
   InitList(ToDoList) ;
   InitLabel
END M2GDB.
