IMPLEMENTATION MODULE M2GenDyn486 ;


FROM SYSTEM IMPORT WORD, BITSET ;
FROM StdIO IMPORT Write, PushOutput, PopOutput, GetCurrentOutput, ProcWrite ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM StrLib IMPORT StrConCat ;
FROM NumberIO IMPORT WriteCard, WriteInt, IntToStr, CardToStr ;

FROM NameKey IMPORT MakeKey, WriteKey, GetKey ;
FROM M2System IMPORT AddressSize ;
FROM M2Batch IMPORT MakeDefinitionSource ;
FROM M2Debug IMPORT Assert ;
FROM M2System IMPORT WordSize, Word ;
FROM M2AsmUtil IMPORT WriteAsmName ;
FROM M2Quads IMPORT QuadOperator, GetQuad, GetNextQuad, DisplayQuad ;
FROM M2Error IMPORT BeginError, EndError ;
FROM M2Lexical IMPORT InternalError, WriteError ;
FROM M2GDB IMPORT LineInfo ;
FROM M2Options IMPORT GenerateDebugging, OptimizeLength ;

FROM M2AS IMPORT InstRegConstant, InstRegReg, InstRegMD, InstMDReg, InstMDConstant,
                 WriteConstant, BinInstReg, BinInstMD, InstStr, InstKey,
                 InstMDOffset, InstRegImmName ;

IMPORT M2Time ;
IMPORT IO ;

FROM M2Alias IMPORT MaxAlias, AliasSymbol, AliasList, AliasType,
                    InitAlias, KillAlias, DuplicateAlias,
                    CollectAllAliases, CollectAliases, RemoveAliasTo,
                    MakeClean,
                    IsDirty, IsClean, IsAliased, IsAbleToBeCleaned,
                    MakeDirty, NoOfTimesUsed, DumpAliases ;

FROM i486 IMPORT Register, Instruction, KindOfScope, DatumPrefix,
                 GetVarTypeSize, EqualSizes, FitsInRegister,
                 GetOffset, GetSize, GetScope ;

FROM SymbolTable IMPORT NulSym, StartScope, EndScope,
                        MakeConstLit,
                        MakeTemporary, PutVar,
                        GetVarQuads, GetMode, GetStringLength,
                        GetSymName, GetType, GetScopeAuthor,
                        GetVarFather, GetSubrange, GetNth,
                        IsModule, IsTemporary, IsExported,
                        IsConst, IsConstString, IsProcedure, IsConstLit,
                        PushValue, PopValue, PushSize ;

FROM M2ALU IMPORT PopCard, Add, Sub, Mult, PopInt, PushInt ;
FROM M2Gen487 IMPORT IsCoprocessorType, IsCoprocessorUsed ;


CONST
   MaxCost      = 10000 ;
   MaxStack     =  5000 ;
   MaxParam     =     7 ;
   DebugOn      = FALSE ;
   DebugInst    =  TRUE ;
   MaxNumChars  =    20 ;
   MaxString    =  1024 ;
   MaxWordSize  =    32 ;  (* max bits in a word *)

TYPE
   Procedures = (noneproc, symtoreg, regtosym, symtosym,
                 codecopy, codeaddr, codebecm, codeline, codeadd,
                 codesub,
                 codegoto,
                 codecond, coderetval, codenew, codekill, coderet,
                 codebase,
                 peepadd, codeinc, codepeepinc,
                 peepsub, codedec, codepeepdec,
                 codeneg,
                 codemult, peepmult,
                 movefrom, moveinto,
                 makeflush, moveviareg, forregsym, makewillbeclean,
                 instmdreg, instregmd, instregreg, instregconst,
                 instmdconst, instbinmd, instbinreg, instbinstr,
                 instmdoffset, instregimmname,
                 makechanged,
                 makeclean, regtoreg, regtoext, exttoreg,
                 makeassign, makeleafclean, makeleafsafe) ;

   Parameters = RECORD
                   CurrentQuad: CARDINAL ;
                   w          : ARRAY [0..MaxParam] OF WORD ;
                END ;

   ProcedureDescription = RECORD
                             WhichProcedure : Procedures ;
                             WhichParameters: Parameters ;
                          END ;

   ProcedureList = RECORD
                      ListPtr         : CARDINAL ;
                      ListOfProcedures: ARRAY [0..MaxStack] OF ProcedureDescription ;
                   END ;

   Activation    = RECORD
                      CurrentAlias: AliasList ;
                      ArgPtr,
                      ArgStart,
                      ArgEnd,
                      AssemblyPtr,
                      CurrentCost,
                      BestCost    : CARDINAL ;
                   END ;

   EvaluationStack = ARRAY [0..MaxStack] OF Activation ;

   CombinationProc2 = PROCEDURE (CARDINAL, ModeOfAddr, CARDINAL, ModeOfAddr) ;
   CombinationProc3 = PROCEDURE (CARDINAL, ModeOfAddr, CARDINAL, ModeOfAddr, CARDINAL, ModeOfAddr) ;


VAR
   TwoToThePower      : ARRAY [0..MaxWordSize] OF CARDINAL ;   (* power of two constants *)
   Zero, One, Two     : CARDINAL ;    (* constants *)
   CurrentAssemblyCode,
   BestAssemblyCode,
   ArgStack           : ProcedureList ;
   EvalStack          : EvaluationStack ;
   NoOfCalls,
   StartQuad, EndQuad,
   EAX, EBX, ECX, EDX,
   ESP, EBP, ESI, EDI,
   CurrentLevel       : CARDINAL ;
   EvalPtr            : CARDINAL ;
   BestAlias          : AliasList ;


(*
   Strategy is to produce near optimal code by employing dynamic programming to search the tree
   of alternative code generation options.

   Essentially we maintain a number of lists:


   BestAssemblyCode    - the best assembly code found so far.
   CurrentAssemblyCode - the current assembly code we have found.
   ArgStack            - holds the functions and arguments described by the EvalStack
   EvalStack           - contains activation record for the ArgStack.
*)


(* %%%FORWARD%%%
   PROCEDURE IsAbleToCode (Start, End: CARDINAL) : BOOLEAN ; FORWARD ;
   PROCEDURE IsAbleToCodeQuad (q: CARDINAL) : BOOLEAN ; FORWARD ;
   PROCEDURE GenCode (Head, Start, End: CARDINAL) ; FORWARD ;
   PROCEDURE InitSearchLists ; FORWARD ;
   PROCEDURE GenerateOptimalCode (Head, Start, End: CARDINAL) ; FORWARD ;
   PROCEDURE NumberOfQuads (Start, End: CARDINAL) : CARDINAL ; FORWARD ;
   PROCEDURE CreateInitialActivation (a: AliasList) ; FORWARD ;
   PROCEDURE PutAllQuadsIntoMethods (Start, End, Count: CARDINAL) : CARDINAL ; FORWARD ;
   PROCEDURE FitsIntoRegister (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
   PROCEDURE TimeOf (h: CARDINAL) : CARDINAL ; FORWARD ;
   PROCEDURE EmitBestInstructions (Cost, Calls: CARDINAL) ; FORWARD ;
   PROCEDURE MakeConstLitFromInt (i: INTEGER) : CARDINAL ; FORWARD ;
   PROCEDURE ToCard (w: WORD) : CARDINAL ; FORWARD ;
   PROCEDURE ToInt (w: WORD) : INTEGER ; FORWARD ;
   PROCEDURE ToScope (w: WORD) : KindOfScope ; FORWARD ;
   PROCEDURE ToInst (w: WORD) : Instruction ; FORWARD ;
   PROCEDURE ToDatum (w: WORD) : DatumPrefix ; FORWARD ;
   PROCEDURE ToReg (w: WORD) : Register ; FORWARD ;
   PROCEDURE ToMode (w: WORD) : ModeOfAddr ; FORWARD ;
   PROCEDURE CopyTree (VAR From, To: ProcedureList) ; FORWARD ;
   PROCEDURE DoProcedure (ArgPtr: CARDINAL) ; FORWARD ;
   PROCEDURE DebugMethods ; FORWARD ;
   PROCEDURE Min (a, b: CARDINAL) : CARDINAL ; FORWARD ;
   PROCEDURE GetCurrentAlias () : AliasList ; FORWARD ;
   PROCEDURE DoEvaluate ; FORWARD ;
   PROCEDURE CodeCheckLeafIsCleanAndSave (h: CARDINAL) ; FORWARD ;
   PROCEDURE IsAllClean (h: CARDINAL) : BOOLEAN ; FORWARD ;
   PROCEDURE AddToCurrentCode (h: CARDINAL) ; FORWARD ;
   PROCEDURE MakeInstBinMethod (inst: Instruction; size: CARDINAL;
                             sym: CARDINAL; mode: ModeOfAddr) ; FORWARD ;
   PROCEDURE MakeInstBinStrMethod (inst: Instruction; str: CARDINAL) ; FORWARD ;
   PROCEDURE MakeInstRegConstantMethod (inst: Instruction; size: CARDINAL;
                                        reg: CARDINAL; sym: CARDINAL) ; FORWARD ;
   PROCEDURE MakeInstRegRegMethod (inst: Instruction; size: CARDINAL;
                                   toreg, fromreg: CARDINAL) ; FORWARD ;
   PROCEDURE MakeInstRegMDMethod (inst: Instruction; size: CARDINAL;
                                  reg: CARDINAL; offset1, offset2: CARDINAL;
                                  scope: KindOfScope; sym: CARDINAL) ; FORWARD ;
   PROCEDURE MakeInstMDRegMethod (inst: Instruction; size: CARDINAL;
                                  offset1, offset2: CARDINAL;
                                  scope: KindOfScope; sym: CARDINAL; reg: CARDINAL) ; FORWARD ;
   PROCEDURE MakeInstMDConstantMethod (inst: Instruction; size: CARDINAL;
                                       offset1, offset2: CARDINAL;
                                       scope: KindOfScope; sym: CARDINAL; const: CARDINAL) ; FORWARD ;
   PROCEDURE MakeInstMDOffsetMethod (inst: Instruction; size: CARDINAL;
                                     sym1, sym2: CARDINAL) ; FORWARD ;
   PROCEDURE MakeInstRegImmNameMethod (inst: Instruction; size: CARDINAL;
                                       reg: CARDINAL; sym: CARDINAL) ; FORWARD ;
   PROCEDURE MakeLeafSafeMethod ; FORWARD ;
   PROCEDURE CodeMakeLeafSafe (h: CARDINAL) ; FORWARD ;
   PROCEDURE IsSafe (a: AliasList) : BOOLEAN ; FORWARD ;
   PROCEDURE CleanRegisters (a: AliasList) : BOOLEAN ; FORWARD ;
   PROCEDURE MakeCheckLeafCleanMethod ; FORWARD ;
   PROCEDURE MakeFlushAllDirtySymbolsMethod ; FORWARD ;
   PROCEDURE MakeCleanSymbolMethod (sym: CARDINAL; mode: ModeOfAddr) ; FORWARD ;
   PROCEDURE MakeWillBeCleanMethod (sym: CARDINAL; mode: ModeOfAddr) ; FORWARD ;
   PROCEDURE MakeMoveFromToMethod (from: CARDINAL; frommode: ModeOfAddr;
                                   to  : CARDINAL; tomode  : ModeOfAddr) ; FORWARD ;
   PROCEDURE MakeSymToSymMethod (from: CARDINAL; frommode: ModeOfAddr;
                                 to  : CARDINAL; tomode  : ModeOfAddr) ; FORWARD ;
   PROCEDURE MakeRegToSymMethod (reg: CARDINAL;
                              to : CARDINAL; tomode: ModeOfAddr) ; FORWARD ;
   PROCEDURE MakeSymToRegMethod (from: CARDINAL; frommode: ModeOfAddr;
                              reg : CARDINAL) ; FORWARD ;
   PROCEDURE MakeRegToRegMethod (fromreg, toreg: CARDINAL; size: CARDINAL) ; FORWARD ;
   PROCEDURE MakeRegToExtMethod (reg: CARDINAL;
                              to : CARDINAL; tomode : ModeOfAddr) ; FORWARD ;
   PROCEDURE MakeExtToRegMethod (from: CARDINAL; frommode: ModeOfAddr;
                              reg : CARDINAL) ; FORWARD ;
   PROCEDURE MakeAssign (from: CARDINAL; frommode: ModeOfAddr;
                      to: CARDINAL; tomode: ModeOfAddr) ; FORWARD ;
   PROCEDURE MakeCodeBecmMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
   PROCEDURE MakeCodeLineMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
   PROCEDURE MakeCodeAddrMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
   PROCEDURE MakeCodeCopyMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
   PROCEDURE MakeCodeGotoMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
   PROCEDURE PushMethod (p: Procedures;
                       w0, w1, w2, w3, w4, w5, w6, w7: WORD) ; FORWARD ;
   PROCEDURE DisplayMethods ; FORWARD ;
   PROCEDURE WriteNSpaces (n: CARDINAL) ; FORWARD ;
   PROCEDURE WriteMethod (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteNoneProc (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteInstBinMD (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteInstBinReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteInstBinStr (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteForRegSym (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteSymToSym (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteRegToSym (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteSymToReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteRegToExt (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteExtToReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteRegToReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCodeCopy (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCodeAddr (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCodeBecm (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCodeLine (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCodeGoto (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteMoveFrom (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteMakeClean (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteMakeChanged (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteMakeWillBeClean (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteMakeFlush (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteMakeLeafSafe (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCheckLeafIsCleanAndSave (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteMakeAssign (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteMoveViaReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteInstMDReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteInstRegMD (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteInstRegReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteInstRegConst (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteInstMDConst (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteInstMDOffset (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteInstRegImmName (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteSymbol (Sym: CARDINAL; Mode: ModeOfAddr) ; FORWARD ;
   PROCEDURE WriteRegister (Sym: CARDINAL; Mode: ModeOfAddr) ; FORWARD ;
   PROCEDURE WriteRegisterName (Sym: CARDINAL) ; FORWARD ;
   PROCEDURE MakeQuadMethod (quad: CARDINAL) ; FORWARD ;
   PROCEDURE SaveCurrentIntoBest ; FORWARD ;
   PROCEDURE DumpCurrentInstructions ; FORWARD ;
   PROCEDURE CodeFlushAllDirtySymbols (h: CARDINAL) ; FORWARD ;
   PROCEDURE IsUsedOutSide (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
   PROCEDURE CodeCleanSymbol (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeAssign (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeWillBeClean (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeNone (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeAsm (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeLine (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeGoto (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeCopy (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeMoveFromTo (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeMoveFromTo (h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
   PROCEDURE MakeCleanMoveIntoMethod (from: CARDINAL; frommode: ModeOfAddr;
                                   to  : CARDINAL; tomode  : ModeOfAddr) ; FORWARD ;
   PROCEDURE CodeCleanMoveInto (h: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCleanMoveInto (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE CodeBecm (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeAddr (h: CARDINAL) ; FORWARD ;
   PROCEDURE MakeCodeAddMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCodeAdd (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE IsAliasedToZero (Sym: CARDINAL; Mode: ModeOfAddr) : BOOLEAN ; FORWARD ;
   PROCEDURE IsAliasedToOne (Sym: CARDINAL; Mode: ModeOfAddr) : BOOLEAN ; FORWARD ;
   PROCEDURE IsAliasedToTwo (Sym: CARDINAL; Mode: ModeOfAddr) : BOOLEAN ; FORWARD ;
   PROCEDURE OffsetUnlessConst (sym: CARDINAL) : ModeOfAddr ; FORWARD ;
   PROCEDURE DoCombinations2 (a: AliasList;
                           op1, op2: CARDINAL;
                           p: CombinationProc2) ; FORWARD ;
   PROCEDURE DoCombinations3 (a: AliasList;
                           op1, op2, op3: CARDINAL;
                           p: CombinationProc3) ; FORWARD ;
   PROCEDURE CodeAdd (h: CARDINAL) ; FORWARD ;
   PROCEDURE SwapSymbols (VAR s1: CARDINAL; VAR m1: ModeOfAddr; VAR s2: CARDINAL; VAR m2: ModeOfAddr) ; FORWARD ;
   PROCEDURE MakePeepAddMethod (op1: CARDINAL; m1: ModeOfAddr;
                              op2: CARDINAL; m2: ModeOfAddr;
                              op3: CARDINAL; m3: ModeOfAddr) ; FORWARD ;
   PROCEDURE CodePeepAdd (h: CARDINAL) ; FORWARD ;
   PROCEDURE WritePeepAdd (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCodeInc (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE MakeIncMethod (op1: CARDINAL; m1: ModeOfAddr; op2: CARDINAL; m2: ModeOfAddr) ; FORWARD ;
   PROCEDURE CodeInc (h: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCodePeepInc (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE MakePeepIncMethod (op1: CARDINAL; m1: ModeOfAddr; op2: CARDINAL; m2: ModeOfAddr) ; FORWARD ;
   PROCEDURE CodePeepInc (h: CARDINAL) ; FORWARD ;
   PROCEDURE MakeCodeSubMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCodeSub (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE CodeDec (h: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCodePeepDec (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE MakePeepDecMethod (op1: CARDINAL; m1: ModeOfAddr; op2: CARDINAL; m2: ModeOfAddr) ; FORWARD ;
   PROCEDURE CodePeepDec (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeSub (h: CARDINAL) ; FORWARD ;
   PROCEDURE MakePeepSubMethod (op1: CARDINAL; m1: ModeOfAddr;
                             op2: CARDINAL; m2: ModeOfAddr;
                             op3: CARDINAL; m3: ModeOfAddr) ; FORWARD ;
   PROCEDURE CodePeepSub (h: CARDINAL) ; FORWARD ;
   PROCEDURE WritePeepSub (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCodeDec (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE MakeDecMethod (op1: CARDINAL; m1: ModeOfAddr; op2: CARDINAL; m2: ModeOfAddr) ; FORWARD ;
   PROCEDURE MakeCodeNegMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
   PROCEDURE MakeNegMethod (op1: CARDINAL; m1: ModeOfAddr; op2: CARDINAL; m2: ModeOfAddr) ; FORWARD ;
   PROCEDURE WriteCodeNeg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE CodeNeg (h: CARDINAL) ; FORWARD ;
   PROCEDURE MakeCodeMultMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
   PROCEDURE CodeMult (h: CARDINAL) ; FORWARD ;
   PROCEDURE WriteCodeMult (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE MakePeepMultMethod (op1: CARDINAL; m1: ModeOfAddr;
                              op2: CARDINAL; m2: ModeOfAddr;
                              op3: CARDINAL; m3: ModeOfAddr) ; FORWARD ;
   PROCEDURE CodePeepMult (h: CARDINAL) ; FORWARD ;
   PROCEDURE WritePeepMult (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE MakeReturnMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
   PROCEDURE CodeRet (h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
   PROCEDURE WriteCodeRet (p: ProcedureList; m: CARDINAL) ; FORWARD ;
   PROCEDURE MakeCodeBaseMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
   PROCEDURE MakeChanged (sym: CARDINAL; mode: ModeOfAddr) ; FORWARD ;
   PROCEDURE CodeMakeClean (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeMakeChanged (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeSymToSym (h: CARDINAL) ; FORWARD ;
   PROCEDURE SymToSymMachineWord (h: CARDINAL) ; FORWARD ;
   PROCEDURE MakeForeachRegisterDoSymToSym (from: CARDINAL; frommode: ModeOfAddr;
                                         to: CARDINAL; tomode: ModeOfAddr) ; FORWARD ;
   PROCEDURE CodeForeachRegisterDoSymToSym (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeMoveThroughRegister (h: CARDINAL) ; FORWARD ;
   PROCEDURE MakeMoveThroughRegisterMethod (from: CARDINAL; frommode: ModeOfAddr;
                                         to  : CARDINAL; tomode  : ModeOfAddr;
                                         reg : CARDINAL) ; FORWARD ;
   PROCEDURE CodeSymToReg (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeExtToReg (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeRegToSym (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeRegToExt (h: CARDINAL) ; FORWARD ;
   PROCEDURE CodeRegToReg (h: CARDINAL) ; FORWARD ;
   PROCEDURE NoOfRegisters () : CARDINAL ; FORWARD ;
   PROCEDURE GetRegister (i: CARDINAL) : CARDINAL ; FORWARD ;
   PROCEDURE IsRegister (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
   PROCEDURE DefineConstants ; FORWARD ;
   PROCEDURE Init ; FORWARD ;
   %%%FORWARD%%% *)



(*
   IsAbleToCode - is the dynamic code generator able to encode these quadruples?
*)

PROCEDURE IsAbleToCode (Start, End: CARDINAL) : BOOLEAN ;
BEGIN
   WHILE Start#End DO
      IF IsAbleToCodeQuad(Start)
      THEN
         Start := GetNextQuad(Start)
      ELSE
         RETURN( FALSE )
      END
   END ;
   RETURN( IsAbleToCodeQuad(Start) )
END IsAbleToCode ;


(*
   IsAbleToCodeQuad - returns TRUE if we can encode this quadruple.
*)

PROCEDURE IsAbleToCodeQuad (q: CARDINAL) : BOOLEAN ;
VAR
   Op : QuadOperator ;
   Op1,
   Op2,
   Op3: CARDINAL ;
BEGIN
   GetQuad(q, Op, Op1, Op2, Op3) ;
   CASE Op OF

   AddrOp        :  RETURN( TRUE ) |
   BecomesOp     :  RETURN( TRUE ) |
   AddOp         :  RETURN( NOT IsCoprocessorUsed(Op1, Op2, Op3) ) |
   MultOp        :  RETURN( NOT IsCoprocessorUsed(Op1, Op2, Op3) ) |
   SubOp         :  RETURN( NOT IsCoprocessorUsed(Op1, Op2, Op3) ) |
   NegateOp      :  RETURN( NOT IsCoprocessorUsed(Op1, Op1, Op3) ) |
   BaseOp        :  RETURN( TRUE ) |
   GotoOp        :  RETURN( TRUE ) |
   OffsetOp      :  RETURN( TRUE ) |
   NewLocalVarOp :  RETURN( FALSE ) |
   KillLocalVarOp:  RETURN( FALSE ) |
   ReturnOp      :  RETURN( FALSE ) |
   LineNumberOp  :  RETURN( TRUE )

   ELSE
      WriteLn ;
      WriteString('cannot encode: ') ;
      DisplayQuad(q) ;
      RETURN( FALSE )
   END
END IsAbleToCodeQuad ;


(*
   GenCode - takes the list of quadruples from Head and generates
             assembly langauge between quadruples Start and End.
*)

PROCEDURE GenCode (Head, Start, End: CARDINAL) ;
BEGIN
   PushOutput(IO.Write) ; (* debugging purposes *)
   GenerateOptimalCode(Head, Start, End) ;
   PopOutput
END GenCode ;


(*
   InitSearchLists - initializes all the search data structures.
*)

PROCEDURE InitSearchLists ;
BEGIN
   CurrentLevel := 0 ;
   NoOfCalls := 0 ;
   EvalPtr := 0 ;
   CurrentAssemblyCode.ListPtr := 0 ;
   BestAssemblyCode.ListPtr := 0
END InitSearchLists ;


(*
   GenerateOptimalCode - 
*)

PROCEDURE GenerateOptimalCode (Head, Start, End: CARDINAL) ;
VAR
   Cost: CARDINAL ;
   a   : AliasList ;
BEGIN
   (* initialize alias list *)
   a := InitAlias() ;
   StartQuad := Start ;
   EndQuad := End ;
   REPEAT
      BestAlias := NIL ;
      (* zap all the procedure lists *)
      InitSearchLists ;
      CreateInitialActivation(a) ;
      WriteString('coding the following quads:') ; WriteLn ;
      Start := PutAllQuadsIntoMethods(Start, End, OptimizeLength) ;
      IF (Start=0) OR (Start=End)
      THEN
         MakeFlushAllDirtySymbolsMethod ; (* must clean up at end of BB     *)
         MakeCheckLeafCleanMethod         (* if all clean and better - save *)
      ELSE
         MakeLeafSafeMethod               (* do minimum to ensure we dont   *)
                                          (* need to backtrack from here    *)
      END ;
      DoEvaluate ;
      EmitBestInstructions(Cost, NoOfCalls) ;
      a := EvalStack[1].CurrentAlias ;
      IF a=NIL
      THEN
         InternalError('alias list should not be NIL', __FILE__, __LINE__)
      END ;
      a := KillAlias(a) ;
      a := BestAlias ;
      IF a=NIL
      THEN
         InternalError('best alias list should not be NIL', __FILE__, __LINE__)
      END
   UNTIL Start=0 ;
   a := KillAlias(a)
END GenerateOptimalCode ;


(*
   NumberOfQuads - returns the number of quads between, Start, and, End.
*)

PROCEDURE NumberOfQuads (Start, End: CARDINAL) : CARDINAL ;
VAR
   i: CARDINAL ;
BEGIN
   i := 1 ;
   WHILE (Start#End) AND (Start#0) DO
      Start := GetNextQuad(Start) ;
      INC(i)
   END ;
   RETURN( i )
END NumberOfQuads ;


(*
   CreateInitialActivation - 
*)

PROCEDURE CreateInitialActivation (a: AliasList) ;
BEGIN
   (* element 0 is only used to contain the final return result *)
   WITH EvalStack[0] DO
      ArgStart    := 0 ;
      ArgEnd      := 0 ;
      ArgPtr      := 0 ;
      AssemblyPtr := 0 ;
      CurrentCost := 0 ;
      BestCost    := MaxCost ;
      CurrentAlias:= a
   END ;
   (* this is the first element to contain any arguments *)
   WITH EvalStack[1] DO
      ArgStart    := 0 ;
      ArgEnd      := 0 ;
      ArgPtr      := 0 ;
      AssemblyPtr := 0 ;
      CurrentCost := 0 ;
      BestCost    := MaxCost ;
      CurrentAlias:= NIL
   END ;
   EvalPtr := 1
END CreateInitialActivation ;


(*
   PutAllQuadsIntoMethods - makes methods for all quadruples: Start..Start+Count
*)

PROCEDURE PutAllQuadsIntoMethods (Start, End, Count: CARDINAL) : CARDINAL ;
VAR
   op : QuadOperator ;
   op1,
   op2,
   op3: CARDINAL ;
BEGIN
   WHILE (Start#0) AND (Start<=End) AND (Count>0) DO
      DEC(Count) ;
      DisplayQuad(Start) ;
      GetQuad(Start, op, op1, op2, op3) ;
      IF op#LineNumberOp
      THEN
         MakeQuadMethod(Start)
      END ;
      IF Start=End
      THEN
         Start := 0  (* signifies end *)
      ELSE
         Start := GetNextQuad(Start)
      END
   END ;
   RETURN( Start )
END PutAllQuadsIntoMethods ;


(*
   FitsIntoRegister - returns TRUE if a symbol, Sym, fill fit into a register.
*)

PROCEDURE FitsIntoRegister (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsConst(Sym)
   THEN
      IF IsConstString(Sym)
      THEN
         RETURN( WordSize>GetStringLength(Sym) )
      ELSE
         (* we assume all constants can fit into a register, this will NOT work for LONGREAL *)
         RETURN( TRUE )
      END
   ELSE
      RETURN( WordSize>=GetSize(Sym) )
   END
END FitsIntoRegister ;


(*
   TimeOf - returns the number of cycles that the following instructions consume.
*)

PROCEDURE TimeOf (h: CARDINAL) : CARDINAL ;
VAR
   i: CARDINAL ;
BEGIN
   WITH ArgStack DO
      WITH ListOfProcedures[h].WhichParameters DO
         CASE ListOfProcedures[h].WhichProcedure OF

         instregconst  : RETURN( M2Time.InstRegConstant(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3])) ) |
         instregreg    : RETURN( M2Time.InstRegReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToReg(w[3])) ) |
         instregmd     : RETURN( M2Time.InstRegMD(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), w[3], w[4], ToScope(w[5]), w[6]) ) |
         instmdreg     : RETURN( M2Time.InstMDReg(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToReg(w[4]), ToScope(w[5]), w[6]) ) |
         instmdconst   : RETURN( M2Time.InstMDConstant(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5], w[6]) ) |
         instbinmd     : RETURN( M2Time.BinInstMD(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5]) ) |
         instbinreg    : RETURN( M2Time.BinInstReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2])) ) |
         instbinstr    : RETURN( M2Time.InstStr(ToInst(w[0]), ToCard(w[1])) ) |
         instmdoffset  : RETURN( M2Time.InstMDOffset(ToInst(w[0]), ToDatum(w[1]), w[2], w[3]) ) |
         instregimmname: RETURN( M2Time.InstRegImmName(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), w[3]) )

         ELSE
            InternalError('case statement should only be dealing with assembly language instructions', __FILE__, __LINE__)
         END
      END
   END
END TimeOf ;


(*
   EmitBestInstructions - 
*)

PROCEDURE EmitBestInstructions (Cost, Calls: CARDINAL) ;
BEGIN
   IF DebugInst
   THEN
      PopOutput ;
      WriteString('# dynamic found best instructions (cost = ') ; WriteCard(Cost, 0) ;
      WriteString(', calls = ') ; WriteCard(Calls, 0) ; Write(')') ; WriteLn ;
      DumpCurrentInstructions ;
      PushOutput(IO.Write)
   END ;
   WriteString('# dynamic found best instructions (cost = ') ; WriteCard(Cost, 0) ;
   WriteString(', calls = ') ; WriteCard(Calls, 0) ; Write(')') ; WriteLn ;
   DumpCurrentInstructions
END EmitBestInstructions ;


(*
   MakeConstLitFromInt - creates a constant symbol from an integer, i.
*)

PROCEDURE MakeConstLitFromInt (i: INTEGER) : CARDINAL ;
VAR
   num: ARRAY [0..MaxNumChars] OF CHAR ;
BEGIN
   IntToStr(i, 0, num) ;
   RETURN( MakeConstLit(MakeKey(num)) )
END MakeConstLitFromInt ;


(*
   ToCard - 
*)

PROCEDURE ToCard (w: WORD) : CARDINAL ;
BEGIN
   RETURN( VAL(CARDINAL, w) )
END ToCard ;


(*
   ToInt - 
*)

PROCEDURE ToInt (w: WORD) : INTEGER ;
BEGIN
   RETURN( VAL(INTEGER, w) )
END ToInt ;


(*
   ToScope - returns the KindOfScope represented by, w.
*)

PROCEDURE ToScope (w: WORD) : KindOfScope ;
BEGIN
   RETURN( VAL(KindOfScope, w) )
END ToScope ;


(*
   ToInst - convert a word into an instruction
*)

PROCEDURE ToInst (w: WORD) : Instruction ;
BEGIN
   RETURN( VAL(Instruction, w) )
END ToInst ;


(*
   ToDatum - convert a word into a DatumPrefix
*)

PROCEDURE ToDatum (w: WORD) : DatumPrefix ;
BEGIN
   CASE CARDINAL(w) OF

   1: RETURN( PByte ) |
   2: RETURN( PWord ) |
   4: RETURN( PDWord )

   ELSE
      InternalError('size unexpected in ToDatum', __FILE__, __LINE__)
   END
END ToDatum ;


(*
   ToReg - converts a WORD into a Register.
*)

PROCEDURE ToReg (w: WORD) : Register ;
BEGIN
   IF CARDINAL(w)=EAX
   THEN
      RETURN( RegEAX )
   ELSIF CARDINAL(w)=EBX
   THEN
      RETURN( RegEBX )
   ELSIF CARDINAL(w)=ECX
   THEN
      RETURN( RegECX )
   ELSIF CARDINAL(w)=EDX
   THEN
      RETURN( RegEDX )
   ELSIF CARDINAL(w)=EBP
   THEN
      RETURN( RegEBP )
   ELSIF CARDINAL(w)=ESP
   THEN
      RETURN( RegESP )
   ELSIF CARDINAL(w)=ESI
   THEN
      RETURN( RegESI )
   ELSIF CARDINAL(w)=EDI
   THEN
      RETURN( RegEDI )
   ELSE
      InternalError('unknown register', __FILE__, __LINE__)
   END
END ToReg ;


(*
   ToMode - 
*)

PROCEDURE ToMode (w: WORD) : ModeOfAddr ;
BEGIN
   RETURN( VAL(ModeOfAddr, w) )
END ToMode ;


(*
   CopyTree - copies, From, to, To.
              Both VAR parameters for the sake of speed :-(
*)

PROCEDURE CopyTree (VAR From, To: ProcedureList) ;
VAR
   i: CARDINAL ;
BEGIN
   WITH From DO
      i := ListPtr ;
      To.ListPtr := i ;
      WHILE i>0 DO
         To.ListOfProcedures[i] := ListOfProcedures[i] ;
         DEC(i)
      END
   END
END CopyTree ;


(*
   DoProcedure - 
*)

PROCEDURE DoProcedure (ArgPtr: CARDINAL) ;
BEGIN
   INC(NoOfCalls) ;
   CASE ArgStack.ListOfProcedures[ArgPtr].WhichProcedure OF

   makeleafclean  : CodeCheckLeafIsCleanAndSave  (ArgPtr) |
   makeleafsafe   : CodeMakeLeafSafe             (ArgPtr) |
   moveinto       : CodeCleanMoveInto            (ArgPtr) |
   makeassign     : CodeAssign                   (ArgPtr) |
   makewillbeclean: CodeWillBeClean              (ArgPtr) |
   forregsym      : CodeForeachRegisterDoSymToSym(ArgPtr) |
   instbinreg     : CodeAsm                      (ArgPtr) |
   instbinmd      : CodeAsm                      (ArgPtr) |
   instbinstr     : CodeAsm                      (ArgPtr) |
   instmdconst    : CodeAsm                      (ArgPtr) |
   instregconst   : CodeAsm                      (ArgPtr) |
   instmdreg      : CodeAsm                      (ArgPtr) |
   instregmd      : CodeAsm                      (ArgPtr) |
   instregreg     : CodeAsm                      (ArgPtr) |
   instmdoffset   : CodeAsm                      (ArgPtr) |
   instregimmname : CodeAsm                      (ArgPtr) |
   moveviareg     : CodeMoveThroughRegister      (ArgPtr) |
   makeclean      : CodeMakeClean                (ArgPtr) |
   makechanged    : CodeMakeChanged              (ArgPtr) |
   makeflush      : CodeFlushAllDirtySymbols     (ArgPtr) |
   movefrom       : CodeMoveFromTo               (ArgPtr) |
   symtosym       : CodeSymToSym                 (ArgPtr) |
   symtoreg       : CodeSymToReg                 (ArgPtr) |
   regtosym       : CodeRegToSym                 (ArgPtr) |
   exttoreg       : CodeExtToReg                 (ArgPtr) |
   regtoext       : CodeRegToExt                 (ArgPtr) |
   regtoreg       : CodeRegToReg                 (ArgPtr) |
   codecopy       : CodeCopy                     (ArgPtr) |
   codeaddr       : CodeAddr                     (ArgPtr) |
   codeline       : CodeLine                     (ArgPtr) |
   codegoto       : CodeGoto                     (ArgPtr) |
   codebecm       : CodeBecm                     (ArgPtr) |
   codeadd        : CodeAdd                      (ArgPtr) |
   codesub        : CodeSub                      (ArgPtr) |
   codemult       : CodeMult                     (ArgPtr) |
   peepadd        : CodePeepAdd                  (ArgPtr) |
   peepsub        : CodePeepSub                  (ArgPtr) |
   peepmult       : CodePeepMult                 (ArgPtr) |
   codeinc        : CodeInc                      (ArgPtr) |
   codepeepinc    : CodePeepInc                  (ArgPtr) |
   codedec        : CodeDec                      (ArgPtr) |
   codepeepdec    : CodePeepDec                  (ArgPtr) |
   codeneg        : CodeNeg                      (ArgPtr) |
(*
   codecond       : CodeCond                     (ArgPtr) |
   coderetval     : CodeRetVal                   (ArgPtr) |
   codenew        : CodeNew                      (ArgPtr) |
   codekill       : CodeKill                     (ArgPtr) |
   coderet        : CodeRet                      (ArgPtr) |
*)
   noneproc       : CodeNone                     (ArgPtr)

   ELSE
      InternalError('case statement does not know about this procedure', __FILE__, __LINE__)
   END
END DoProcedure ;


(*
   DebugMethods - 
*)

PROCEDURE DebugMethods ;
BEGIN
   WITH EvalStack[EvalPtr-1] DO
      WriteLn ;
      WriteString('=================================================================') ; WriteLn ;
      WriteString('                            (Eval stack  :') ; WriteCard(EvalPtr-1, 4) ; Write(')') ; WriteLn ;
      WriteString('                            (Current cost:') ; WriteCard(CurrentCost, 4) ; Write(')') ; WriteLn ;
      WriteString('                            (Best    cost:') ; WriteCard(BestCost, 4) ; Write(')') ; WriteLn ;
      WriteString('methods:') ; WriteLn ;
      DisplayMethods ;
      WriteLn
   END
END DebugMethods ;


(*
   Min - 
*)

PROCEDURE Min (a, b: CARDINAL) : CARDINAL ;
BEGIN
   IF a<b
   THEN
      RETURN( a )
   ELSE
      RETURN( b )
   END
END Min ;


(*
   GetCurrentAlias - returns the CurrentAlias
*)

PROCEDURE GetCurrentAlias () : AliasList ;
BEGIN
   RETURN( EvalStack[EvalPtr-1].CurrentAlias )
END GetCurrentAlias ;


(*
   DoEvaluate - evaluates the next procedure on the ArgStack as described by
                the EvalStack. This procedure traverses the search space in
                depth first order.

                NOTE that   EvalPtr - 2 = Previous
                            EvalPtr - 1 = Current
                            EvalPtr - 0 = Next
*)

PROCEDURE DoEvaluate ;
VAR
   PreviousAlias: AliasList ;
BEGIN
   (* increment evaluation stack and reset the start of our sons arguments *)
   INC(EvalPtr) ;
   IF EvalPtr=MaxStack
   THEN
      InternalError('increase MaxStack', __FILE__, __LINE__)
   END ;
   WITH EvalStack[EvalPtr] DO
      ArgStart := ArgStack.ListPtr ;
      ArgPtr   := ArgStack.ListPtr ;
      ArgEnd   := ArgStack.ListPtr    (* to a safe value for DisplayMethods *)
   END ;

   IF TRUE
   THEN
      DebugMethods
   END ;

   (* now process Current level procedures *)
   WITH EvalStack[EvalPtr-1] DO
      (* inherit BestCost and CurrentCost, set up alias *)
      BestCost        := EvalStack[EvalPtr-2].BestCost ;
      CurrentCost     := EvalStack[EvalPtr-2].CurrentCost ;
      CurrentAlias    := DuplicateAlias(EvalStack[EvalPtr-2].CurrentAlias) ;
      AssemblyPtr     := EvalStack[EvalPtr-2].AssemblyPtr ;
      ArgEnd          := ArgStack.ListPtr ;  (* the end of the arguments for this function *)

      WHILE (CurrentCost<BestCost) AND (ArgPtr<ArgEnd) DO
         DoProcedure(ArgPtr) ;
         INC(ArgPtr)
      END ;

      IF ArgPtr=ArgEnd
      THEN
         IF CurrentCost<=BestCost
         THEN
            (* return BestCost *)
            EvalStack[EvalPtr-2].BestCost  := BestCost ;
            WriteString('found new: BestCost = ') ; WriteCard(BestCost, 6) ; WriteLn
         ELSE
            WriteString('nothing interesting found') ; WriteLn
         END ;
         (* return CurrentAlias *)
         EvalStack[EvalPtr-2].CurrentAlias := KillAlias(EvalStack[EvalPtr-2].CurrentAlias) ;
         EvalStack[EvalPtr-2].CurrentAlias := CurrentAlias ;
         EvalStack[EvalPtr-2].CurrentCost  := CurrentCost ;
         EvalStack[EvalPtr-2].AssemblyPtr  := AssemblyPtr
      ELSE
         (* tidy up *)
         CurrentAlias := KillAlias(CurrentAlias) ;
         WriteString('cut off searching any furthur: BestCost = ') ; WriteCard(BestCost, 6) ; WriteLn
      END
   END ;
   DEC(EvalPtr) ;
END DoEvaluate ;


(*
   CodeCheckLeafCleanAndSave - checks to see whether all symbols are clean.
                               If so it tests the current cost against the
                               best cost and saves the better.
*)

PROCEDURE CodeCheckLeafIsCleanAndSave (h: CARDINAL) ;
BEGIN
   IF DebugOn
   THEN
      WriteString(' (leaf)')
   END ;
   IF IsAllClean(h)
   THEN
      WITH EvalStack[EvalPtr-1] DO
         IF CurrentCost<BestCost
         THEN
            (*
               at this point all symbols are clean and we are at the end of
               the basic block. We also know that the current cost is better
               than our previous best and so we can update our best score.
            *)
            IF DebugOn
            THEN
               WriteString(' best so far...') ; WriteLn
            END ;
            SaveCurrentIntoBest
         ELSE
            IF DebugOn
            THEN
               WriteString(' [not as good as the best]') ; WriteLn
            END
         END
      END
   ELSE
      IF DebugOn
      THEN
         WriteString(' [symbols are not clean]') ; WriteLn
      END
   END
END CodeCheckLeafIsCleanAndSave ;


(*
   IsAllClean - 
*)

PROCEDURE IsAllClean (h: CARDINAL) : BOOLEAN ;
VAR
   Start, End: CARDINAL ;
   n, i      : CARDINAL ;
   al        : ARRAY [0..MaxAlias] OF AliasType ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      n := CollectAllAliases(GetCurrentAlias(), al) ;
      i := 0 ;
      WHILE i<n DO
         IF NOT IsConst(al[i].Sym)
         THEN
            IF IsDirty(GetCurrentAlias(), al[i].Sym, al[i].Mode) AND IsUsedOutSide(al[i].Sym)
            THEN
               RETURN( FALSE )
            END
         END ;
         INC(i)
      END
   END ;
   RETURN( TRUE )
END IsAllClean ;


(*
   AddToCurrentCode - adds procedure, h, to the current code pile of assembly instructions
                      and increments the current cost held on the EvalStack.
*)

PROCEDURE AddToCurrentCode (h: CARDINAL) ;
BEGIN
   WITH CurrentAssemblyCode DO
      IF EvalStack[EvalPtr-1].AssemblyPtr>MaxStack
      THEN
         InternalError('increase MaxStack', __FILE__, __LINE__)
      ELSE
         ListOfProcedures[EvalStack[EvalPtr-1].AssemblyPtr] := ArgStack.ListOfProcedures[h] ;
         INC(EvalStack[EvalPtr-1].CurrentCost, TimeOf(h))
      END ;
      INC(EvalStack[EvalPtr-1].AssemblyPtr)
   END
END AddToCurrentCode ;


(*
   MakeInstBinMethod - 
*)

PROCEDURE MakeInstBinMethod (inst: Instruction; size: CARDINAL;
                             sym: CARDINAL; mode: ModeOfAddr) ;
BEGIN
   IF IsRegister(sym)
   THEN
      PushMethod(instbinreg,
                 inst, size, sym, mode, 0, 0, 0, 0)
   ELSE
      PushMethod(instbinmd,
                 inst, size, GetOffset(sym), 0, GetScope(sym), sym, mode, 0)
   END
END MakeInstBinMethod ;


(*
   MakeInstBinStrMethod - 
*)

PROCEDURE MakeInstBinStrMethod (inst: Instruction; str: CARDINAL) ;
BEGIN
   PushMethod(instbinstr,
              inst, str, 0, 0, 0, 0, 0, 0)
END MakeInstBinStrMethod ;


(*
   MakeInstRegConstantMethod - 
*)

PROCEDURE MakeInstRegConstantMethod (inst: Instruction; size: CARDINAL;
                                     reg: CARDINAL; sym: CARDINAL) ;
BEGIN
   PushMethod(instregconst,
              inst, size, reg, sym, 0, 0, 0, 0)
END MakeInstRegConstantMethod ;


(*
   MakeInstRegRegMethod - 
*)

PROCEDURE MakeInstRegRegMethod (inst: Instruction; size: CARDINAL;
                                toreg, fromreg: CARDINAL) ;
BEGIN
   PushMethod(instregreg,
              inst, size, toreg, fromreg, 0, 0, 0, 0)
END MakeInstRegRegMethod ;


(*
   MakeInstRegMDMethod - 
*)

PROCEDURE MakeInstRegMDMethod (inst: Instruction; size: CARDINAL;
                               reg: CARDINAL; offset1, offset2: CARDINAL;
                               scope: KindOfScope; sym: CARDINAL) ;
BEGIN
   PushMethod(instregmd,
              inst, size, reg, offset1, offset2, scope, sym, 0)
END MakeInstRegMDMethod ;


(*
   MakeInstMDRegMethod - 
*)

PROCEDURE MakeInstMDRegMethod (inst: Instruction; size: CARDINAL;
                               offset1, offset2: CARDINAL;
                               scope: KindOfScope; sym: CARDINAL; reg: CARDINAL) ;
BEGIN
   PushMethod(instmdreg,
              inst, size, offset1, offset2, reg, scope, sym, 0)
END MakeInstMDRegMethod ;


(*
   MakeInstMDConstantMethod - 
*)

PROCEDURE MakeInstMDConstantMethod (inst: Instruction; size: CARDINAL;
                                    offset1, offset2: CARDINAL;
                                    scope: KindOfScope; sym: CARDINAL; const: CARDINAL) ;
BEGIN
   PushMethod(instmdconst,
              inst, size, offset1, offset2, scope, sym, const, 0)
END MakeInstMDConstantMethod ;


(*
   MakeInstMDOffsetMethod - inst Mem with an offset Int. (Mem[Sym1] <- inst <- OFFSET Sym2)
*)

PROCEDURE MakeInstMDOffsetMethod (inst: Instruction; size: CARDINAL;
                                  sym1, sym2: CARDINAL) ;
BEGIN
   PushMethod(instmdoffset,
              inst, size, sym1, sym2, 0, 0, 0, 0)
END MakeInstMDOffsetMethod ;


(*
   MakeInstRegImmNameMethod -  instruction immediate name into register.  (Reg <- inst <- Constant)
*)

PROCEDURE MakeInstRegImmNameMethod (inst: Instruction; size: CARDINAL;
                                    reg: CARDINAL; sym: CARDINAL) ;
BEGIN
   PushMethod(instregimmname,
              inst, size, reg, sym, 0, 0, 0, 0)
END MakeInstRegImmNameMethod ;


(*
   MakeLeafSafeMethod - makes a leaf safe so that we will never need to
                        backtrack beyond this point.
*)

PROCEDURE MakeLeafSafeMethod ;
BEGIN
   PushMethod(makeleafsafe,
              0, 0, 0, 0, 0, 0, 0, 0)
END MakeLeafSafeMethod ;


(*
   CodeMakeLeafSafe - makes a leaf safe so that we will never need to
                      backtrack beyond this point.
*)

PROCEDURE CodeMakeLeafSafe (h: CARDINAL) ;
VAR
   i, n: CARDINAL ;
   al  : ARRAY [0..MaxAlias] OF AliasType ;
BEGIN
   IF IsSafe(GetCurrentAlias())
   THEN
      WITH EvalStack[EvalPtr-1] DO
         IF CurrentCost<BestCost
         THEN
            IF DebugOn
            THEN
               WriteString(' best so far...') ; WriteLn
            END ;
            SaveCurrentIntoBest
         END
      END
   ELSE
      IF CleanRegisters(GetCurrentAlias())
      THEN
         (* we are able to clean the registers *)
         n := CollectAllAliases(GetCurrentAlias(), al) ;
         i := 0 ;
         WHILE i<n DO
            IF IsRegister(al[i].Sym) AND IsDirty(GetCurrentAlias(), al[i].Sym, al[i].Mode)
            THEN
               MakeCleanSymbolMethod(al[i].Sym, al[i].Mode)
            END ;
            INC(i)
         END ;
         MakeLeafSafeMethod ;
         DoEvaluate
      ELSE
         (* not able to clean the registers *)
      END
   END
END CodeMakeLeafSafe ;


(*
   IsSafe - 
*)

PROCEDURE IsSafe (a: AliasList) : BOOLEAN ;
VAR
   i, n: CARDINAL ;
   al  : ARRAY [0..MaxAlias] OF AliasType ;
BEGIN
   n := CollectAllAliases(a, al) ;
   i := 0 ;
   WHILE i<n DO
      IF IsRegister(al[i].Sym) AND IsDirty(a, al[i].Sym, al[i].Mode)
      THEN
         RETURN( FALSE )
      END ;
      INC(i)
   END ;
   RETURN( TRUE )
END IsSafe ;


(*
   CleanRegisters - 
*)

PROCEDURE CleanRegisters (a: AliasList) : BOOLEAN ;
VAR
   i, n: CARDINAL ;
   al  : ARRAY [0..MaxAlias] OF AliasType ;
BEGIN
   n := CollectAllAliases(a, al) ;
   i := 0 ;
   WHILE i<n DO
      IF IsRegister(al[i].Sym) AND IsDirty(a, al[i].Sym, al[i].Mode)
      THEN
         IF NOT IsAbleToBeCleaned(a, al[i].Sym, al[i].Mode)
         THEN
            RETURN( FALSE )
         END
      END ;
      INC(i)
   END ;
   (* all registers can be cleaned *)
   RETURN( TRUE )
END CleanRegisters ;


(*
   MakeCheckLeafCleanMethod - checks to see whether all symbols are clean.
                              If so it tests the current cost against the
                              best cost and saves the better.
*)

PROCEDURE MakeCheckLeafCleanMethod ;
BEGIN
   PushMethod(makeleafclean,
              0, 0, 0, 0, 0, 0, 0, 0)
END MakeCheckLeafCleanMethod ;


(*
   MakeFlushAllDirtySymbolsMethod - makes a FlushAllDirtySymbols method.
*)

PROCEDURE MakeFlushAllDirtySymbolsMethod ;
BEGIN
   PushMethod(makeflush,
              0, 0, 0, 0, 0, 0, 0, 0)
END MakeFlushAllDirtySymbolsMethod ;


(*
   MakeCleanSymbolMethod - makes a clean symbol method.
*)

PROCEDURE MakeCleanSymbolMethod (sym: CARDINAL; mode: ModeOfAddr) ;
BEGIN
   PushMethod(makeclean,
              sym, mode, 0, 0, 0, 0, 0, 0)
END MakeCleanSymbolMethod ;


(*
   MakeWillBeCleanMethod - 
*)

PROCEDURE MakeWillBeCleanMethod (sym: CARDINAL; mode: ModeOfAddr) ;
BEGIN
   PushMethod(makewillbeclean,
              sym, mode, 0, 0, 0, 0, 0, 0)
END MakeWillBeCleanMethod ;



(*
   MakeMoveFromToMethod -
*)

PROCEDURE MakeMoveFromToMethod (from: CARDINAL; frommode: ModeOfAddr;
                                to  : CARDINAL; tomode  : ModeOfAddr) ;
BEGIN
   IF (from#to) OR (frommode#tomode)
   THEN
      PushMethod(movefrom,
                 from, frommode, to, tomode, 0, 0, 0, 0)
   END
END MakeMoveFromToMethod ;


(*
   MakeSymToSymMethod - 
*)

PROCEDURE MakeSymToSymMethod (from: CARDINAL; frommode: ModeOfAddr;
                              to  : CARDINAL; tomode  : ModeOfAddr) ;
BEGIN
   IF (from#to) OR (frommode#tomode)
   THEN
      PushMethod(symtosym,
                 from, frommode, to, tomode, 0, 0, 0, 0)
   END
END MakeSymToSymMethod ;


(*
   MakeRegToSymMethod - 
*)

PROCEDURE MakeRegToSymMethod (reg: CARDINAL;
                              to : CARDINAL; tomode: ModeOfAddr) ;
BEGIN
   PushMethod(regtosym,
              reg, to, tomode, 0, 0, 0, 0, 0)
END MakeRegToSymMethod ;


(*
   MakeSymToRegMethod - 
*)

PROCEDURE MakeSymToRegMethod (from: CARDINAL; frommode: ModeOfAddr;
                              reg : CARDINAL) ;
BEGIN
   PushMethod(symtoreg,
              from, frommode, reg, 0, 0, 0, 0, 0)
END MakeSymToRegMethod ;


(*
   MakeRegToRegMethod - 
*)

PROCEDURE MakeRegToRegMethod (fromreg, toreg: CARDINAL; size: CARDINAL) ;
BEGIN
   IF fromreg#toreg
   THEN
      PushMethod(regtoreg,
                 fromreg, toreg, size, 0, 0, 0, 0, 0)
   END
END MakeRegToRegMethod ;


(*
   MakeRegToExtMethod - 
*)

PROCEDURE MakeRegToExtMethod (reg: CARDINAL;
                              to : CARDINAL; tomode : ModeOfAddr) ;
BEGIN
   PushMethod(regtoext,
              reg, to, tomode, 0, 0, 0, 0, 0)
END MakeRegToExtMethod ;


(*
   MakeExtToRegMethod - 
*)

PROCEDURE MakeExtToRegMethod (from: CARDINAL; frommode: ModeOfAddr;
                              reg : CARDINAL) ;
BEGIN
   PushMethod(exttoreg,
              from, frommode, reg, 0, 0, 0, 0, 0)
END MakeExtToRegMethod ;


(*
   MakeAssign - 
*)

PROCEDURE MakeAssign (from: CARDINAL; frommode: ModeOfAddr;
                      to: CARDINAL; tomode: ModeOfAddr) ;
BEGIN
   IF (from#to) OR (frommode#tomode)
   THEN
      PushMethod(makeassign,
                 from, frommode, to, tomode, 0, 0, 0, 0)
   END
END MakeAssign ;


(*
   MakeCodeBecmMethod - 
*)

PROCEDURE MakeCodeBecmMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
VAR
   Size1, Size3: CARDINAL ;
BEGIN
   (* check for assignment errors *)
   IF IsConst(op1) AND IsConst(op3)
   THEN
   ELSIF IsConstString(op3)
   THEN
      InternalError('not yet implemented', __FILE__, __LINE__) ;
   ELSIF IsProcedure(op3)
   THEN
   ELSE
      Size1 := GetVarTypeSize(op1) ;
      IF IsConst(op3)
      THEN
         Size3 := Size1
      ELSE
         Size3 := GetVarTypeSize(op3)
      END ;
      IF (Size1#Size3) AND (NOT (IsCoprocessorType(op1) AND (IsCoprocessorType(op3))))
      THEN
         BeginError ;
         WriteLn ;
         WriteKey(GetSymName(op1)) ; WriteString(' := ') ; WriteKey(GetSymName(op3)) ; WriteLn ;
         WriteString('Type of ') ; WriteKey(GetSymName(op1)) ;
         WriteString(' is ') ; WriteKey(GetSymName(GetType(op1))) ; WriteLn ;
         WriteString('Type of ') ; WriteKey(GetSymName(op3)) ;
         WriteString(' is ') ; WriteKey(GetSymName(GetType(op3))) ; WriteLn ;
         WriteError('Different sizes in :=')
      END
   END ;
   (* end of checking *)
   PushMethod(codebecm,
              op1, GetMode(op1), op3, GetMode(op3), 0, 0, 0, 0)
END MakeCodeBecmMethod ;


(*
   MakeCodeLineMethod - 
*)

PROCEDURE MakeCodeLineMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   PushMethod(codeline,
              op3, 0, 0, 0, 0, 0, 0, 0)
END MakeCodeLineMethod ;


(*
   MakeCodeAddrMethod - 
*)

PROCEDURE MakeCodeAddrMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   PushMethod(codeaddr,
              op1, GetMode(op1), op3, GetMode(op3), 0, 0, 0, 0)
END MakeCodeAddrMethod ;


(*
   MakeCodeCopyMethod - 
*)

PROCEDURE MakeCodeCopyMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   PushMethod(codecopy,
              op1, GetMode(op1), op3, GetMode(op3), 0, 0, 0, 0)
END MakeCodeCopyMethod ;


(*
   MakeCodeGotoMethod - 
*)

PROCEDURE MakeCodeGotoMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   MakeFlushAllDirtySymbolsMethod ;   (* must clean up before goto *)
   PushMethod(codegoto,
              op3, 0, 0, 0, 0, 0, 0, 0)
END MakeCodeGotoMethod ;


(*
   PushMethod - 
*)

PROCEDURE PushMethod (p: Procedures;
                      w0, w1, w2, w3, w4, w5, w6, w7: WORD) ;
BEGIN
   WITH ArgStack DO
      IF ListPtr=MaxStack
      THEN
         InternalError('internal error - increase MaxStack in M2GenDyn486', __FILE__, __LINE__)
      ELSE
         WITH ListOfProcedures[ListPtr] DO
            WhichProcedure := p ;
            WITH WhichParameters DO
               w[0] := w0 ;
               w[1] := w1 ;
               w[2] := w2 ;
               w[3] := w3 ;
               w[4] := w4 ;
               w[5] := w5 ;
               w[6] := w6 ;
               w[7] := w7
            END
         END ;
         INC(ListPtr)
      END
   END
END PushMethod ;


(*
   DisplayMethods - 
*)

PROCEDURE DisplayMethods ;
VAR
   f, a: CARDINAL ;
BEGIN
   f := 0 ;
   WHILE f<EvalPtr DO
      WITH EvalStack[f] DO
         a := ArgStart ;
         WriteNSpaces(f*2) ;
         WHILE a<ArgEnd DO
            WriteString(' (') ;
            WriteMethod(ArgStack, a) ;
            Write(')') ;
            INC(a)
         END ;
         WriteLn ;
      END ;
      INC(f)
   END
END DisplayMethods ;


(*
   WriteNSpaces - writes, n, spaces.
*)

PROCEDURE WriteNSpaces (n: CARDINAL) ;
BEGIN
   WHILE n>0 DO
      Write(' ') ;
      DEC(n)
   END
END WriteNSpaces ;


(*
   WriteMethod - writes a method to the default output stream.
*)

PROCEDURE WriteMethod (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m] DO
      CASE WhichProcedure OF

      makeleafclean   :  WriteCheckLeafIsCleanAndSave(p ,m) |
      makeleafsafe    :  WriteMakeLeafSafe(p, m) |
      moveinto        :  WriteCleanMoveInto(p, m) |
      makeassign      :  WriteMakeAssign(p, m) |
      makewillbeclean :  WriteMakeWillBeClean(p, m) |
      forregsym       :  WriteForRegSym(p, m) |
      noneproc        :  WriteNoneProc(p, m) |
      symtosym        :  WriteSymToSym(p, m) |
      symtoreg        :  WriteSymToReg(p, m) |
      regtosym        :  WriteRegToSym(p, m) |
      regtoreg        :  WriteRegToReg(p, m) |
      exttoreg        :  WriteExtToReg(p, m) |
      regtoext        :  WriteRegToExt(p, m) |
      codecopy        :  WriteCodeCopy(p, m) |
      codeaddr        :  WriteCodeAddr(p, m) |
      codebecm        :  WriteCodeBecm(p, m) |
      codeline        :  WriteCodeLine(p, m) |
      codegoto        :  WriteCodeGoto(p, m) |
(*
      codecond        :  WriteCodeCond(p, m) |
      coderetval      :  WriteCodeRetVal(p, m) |
      codenew         :  WriteCodeNew(p, m) |
      codekill        :  WriteCodeKill(p, m) |
      coderet         :  WriteCodeRet(p, m) |
*)
      codeadd         :  WriteCodeAdd(p, m) |
      peepadd         :  WritePeepAdd(p, m) |
      codeinc         :  WriteCodeInc(p, m) |
      codepeepinc     :  WriteCodePeepInc(p, m) |
      codepeepdec     :  WriteCodePeepDec(p, m) |
      codesub         :  WriteCodeSub(p, m) |
      peepsub         :  WritePeepSub(p, m) |
      codedec         :  WriteCodeDec(p, m) |
      codeneg         :  WriteCodeNeg(p, m) |
      codemult        :  WriteCodeMult(p, m) |
      peepmult        :  WritePeepMult(p, m) |
      movefrom        :  WriteMoveFrom(p, m) |
      makeclean       :  WriteMakeClean(p, m) |
      makechanged     :  WriteMakeChanged(p, m) |
      makeflush       :  WriteMakeFlush(p, m) |
      moveviareg      :  WriteMoveViaReg(p, m) |
      instmdreg       :  WriteInstMDReg(p, m) |
      instregmd       :  WriteInstRegMD(p, m) |
      instregreg      :  WriteInstRegReg(p, m) |
      instregconst    :  WriteInstRegConst(p, m) |
      instmdconst     :  WriteInstMDConst(p, m) |
      instbinmd       :  WriteInstBinMD(p, m) |
      instbinreg      :  WriteInstBinReg(p, m) |
      instbinstr      :  WriteInstBinStr(p, m) |
      instmdoffset    :  WriteInstMDOffset(p, m) |
      instregimmname  :  WriteInstRegImmName(p, m)


      ELSE
         InternalError('unknown procedure in case', __FILE__, __LINE__)
      END
   END
END WriteMethod ;


(*
   WriteNoneProc - 
*)

PROCEDURE WriteNoneProc (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WriteString('(noneproc)')
END WriteNoneProc ;


(*
   WriteInstBinMD - 
*)

PROCEDURE WriteInstBinMD (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.BinInstMD(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5]), 0) ;
      WriteString('>" ') ;
      BinInstMD(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5]) ;
      WriteString('" ')
   END
END WriteInstBinMD ;


(*
   WriteInstBinReg - 
*)

PROCEDURE WriteInstBinReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.BinInstReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2])), 0) ;
      WriteString('>" ') ;
      BinInstReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2])) ;
      WriteString('" ')
   END
END WriteInstBinReg ;


(*
   WriteInstBinStr - 
*)

PROCEDURE WriteInstBinStr (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstStr(ToInst(w[0]), ToCard(w[1])), 0) ;
      WriteString('>" ') ;
      InstKey(ToInst(w[0]), ToCard(w[1])) ;
      WriteString('" ')
   END
END WriteInstBinStr ;


(*
   WriteForRegSym - 
*)

PROCEDURE WriteForRegSym (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('ForRegSym( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ANYREG -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteForRegSym ;


(*
   WriteSymToSym - 
*)

PROCEDURE WriteSymToSym (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('SymToSym( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteSymToSym ;


(*
   WriteRegToSym - 
*)

PROCEDURE WriteRegToSym (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('RegToSym( ') ;
      WriteSymbol(ToCard(w[0]), RightValue) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[1]), ToMode(w[2])) ;
      WriteString(' )')
   END
END WriteRegToSym ;


(*
   WriteSymToReg - 
*)

PROCEDURE WriteSymToReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('SymToReg( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), RightValue) ;
      WriteString(' )')
   END
END WriteSymToReg ;


(*
   WriteRegToExt - 
*)

PROCEDURE WriteRegToExt (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('RegToExt( ') ;
      WriteSymbol(ToCard(w[0]), RightValue) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[1]), ToMode(w[2])) ;
      WriteString(' )')
   END
END WriteRegToExt ;


(*
   WriteExtToReg - 
*)

PROCEDURE WriteExtToReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('ExtToReg( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), RightValue) ;
      WriteString(' )')
   END
END WriteExtToReg ;


(*
   WriteRegToReg - 
*)

PROCEDURE WriteRegToReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('RegToReg( ') ;
      WriteSymbol(ToCard(w[0]), RightValue) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[1]), RightValue) ;
      WriteString(' )')
   END
END WriteRegToReg ;


(*
   WriteCodeCopy - 
*)

PROCEDURE WriteCodeCopy (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('CodeCopy( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteCodeCopy ;


(*
   WriteCodeAddr - 
*)

PROCEDURE WriteCodeAddr (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('CodeAddr( ADR(') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(') -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteCodeAddr ;


(*
   WriteCodeBecm - 
*)

PROCEDURE WriteCodeBecm (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('CodeBecm( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' := ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteCodeBecm ;


(*
   WriteCodeLine - 
*)

PROCEDURE WriteCodeLine (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('(line ') ;
      WriteCard(ToCard(w[0]), 0) ;
      Write(')')
   END
END WriteCodeLine ;


(*
   WriteCodeGoto - 
*)

PROCEDURE WriteCodeGoto (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('(goto ') ;
      WriteCard(ToCard(w[0]), 0) ;
      Write(')')
   END
END WriteCodeGoto ;


(*
   WriteMoveFrom - 
*)

PROCEDURE WriteMoveFrom (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('MoveFrom( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteMoveFrom ;


(*
   WriteMakeClean - 
*)

PROCEDURE WriteMakeClean (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('MakeClean( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WriteMakeClean ;


(*
   WriteMakeChanged - 
*)

PROCEDURE WriteMakeChanged (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('MakeChanged( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WriteMakeChanged ;


(*
   WriteMakeWillBeClean - 
*)

PROCEDURE WriteMakeWillBeClean (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('isclean( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WriteMakeWillBeClean ;


(*
   WriteMakeFlush - 
*)

PROCEDURE WriteMakeFlush (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WriteString('MakeFlush() ')
END WriteMakeFlush ;

(*
   WriteMakeLeafSafe - 
*)

PROCEDURE WriteMakeLeafSafe (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WriteString('MakeLeafSafe() ')
END WriteMakeLeafSafe ;

(*
   WriteCheckLeafIsCleanAndSave - 
*)

PROCEDURE WriteCheckLeafIsCleanAndSave (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WriteString('CheckLeafIsCleanAndSave() ')
END WriteCheckLeafIsCleanAndSave ;

(*
   WriteMakeAssign - 
*)

PROCEDURE WriteMakeAssign (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('assign( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteMakeAssign ;


(*
   WriteMoveViaReg - 
*)

PROCEDURE WriteMoveViaReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('MoveViaReg( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteRegister(ToCard(w[4]), RightValue) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteMoveViaReg ;


(*
   WriteInstMDReg - 
*)

PROCEDURE WriteInstMDReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstMDReg(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToReg(w[4]), ToScope(w[5]), w[6]), 0) ;
      WriteString('>" ') ;
      InstMDReg(ToInst(w[0]), ToDatum(w[1]), ToCard(w[2]), ToCard(w[3]),
                ToReg(w[4]), ToScope(w[5]), ToCard(w[6])) ;
      WriteString('" ')
   END
END WriteInstMDReg ;


(*
   WriteInstRegMD - 
*)

PROCEDURE WriteInstRegMD (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstRegMD(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), w[3], w[4], ToScope(w[5]), w[6]), 0) ;
      WriteString('>" ') ;
      InstRegMD(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3]),
                ToCard(w[4]), ToScope(w[5]), ToCard(w[6])) ;
      WriteString('" ')
   END
END WriteInstRegMD ;


(*
   WriteInstRegReg -
*)

PROCEDURE WriteInstRegReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstRegReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToReg(w[3])), 0) ;
      WriteString('>" ') ;
      InstRegReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToReg(w[3])) ;
      WriteString('" ')
   END
END WriteInstRegReg ;


(*
   WriteInstRegConst -
*)

PROCEDURE WriteInstRegConst (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstRegConstant(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3])), 0) ;
      WriteString('>" ') ;
      InstRegConstant(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3])) ;
      WriteString('" ')
   END
END WriteInstRegConst ;


(*
   WriteInstMDConst -
*)

PROCEDURE WriteInstMDConst (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstMDConstant(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5], w[6]), 0) ;
      WriteString('>" ') ;
      InstMDConstant(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5], w[6]) ;
      WriteString('" ')
   END
END WriteInstMDConst ;


(*
   WriteInstMDOffset - 
*)

PROCEDURE WriteInstMDOffset (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstMDOffset(ToInst(w[0]), ToDatum(w[1]), ToCard(w[2]), ToCard(w[3])), 0) ;
      WriteString('>" ') ;
      InstMDOffset(ToInst(w[0]), ToDatum(w[1]), ToCard(w[2]), ToCard(w[3])) ;
      WriteString('" ')
   END
END WriteInstMDOffset ;


(*
   WriteInstRegImmName - 
*)

PROCEDURE WriteInstRegImmName (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstRegImmName(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3])), 0) ;
      WriteString('>" ') ;
      InstRegImmName(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3])) ;
      WriteString('" ')
   END
END WriteInstRegImmName ;


(*
   WriteSymbol - 
*)

PROCEDURE WriteSymbol (Sym: CARDINAL; Mode: ModeOfAddr) ;
BEGIN
   IF IsRegister(Sym)
   THEN
      WriteRegister(Sym, Mode)
   ELSE
      IF (Mode=ImmediateValue) OR IsConst(Sym)
      THEN
         Write('#') ;
         IF IsConst(Sym)
         THEN
            WriteConstant(Sym)
         ELSE
            WriteAsmName(Sym)
         END
      ELSIF Mode=RightValue
      THEN
         WriteAsmName(Sym)
      ELSIF Mode=LeftValue
      THEN
         Write('[') ; WriteAsmName(Sym) ; Write(']')
      ELSE
         InternalError('incorrect mode', __FILE__, __LINE__)
      END
   END
END WriteSymbol ;


(*
   WriteRegister - 
*)

PROCEDURE WriteRegister (Sym: CARDINAL; Mode: ModeOfAddr) ;
BEGIN
   IF Mode=LeftValue
   THEN
      Write('[') ; WriteRegisterName(Sym) ; Write(']')
   ELSIF Mode=RightValue
   THEN
      WriteRegisterName(Sym)
   ELSE
      InternalError('incorrect mode for register', __FILE__, __LINE__)
   END
END WriteRegister ;


(*
   WriteRegisterName - 
*)

PROCEDURE WriteRegisterName (Sym: CARDINAL) ;
BEGIN
   IF Sym=EAX
   THEN
      WriteString('%eax')
   ELSIF Sym=EBX
   THEN
      WriteString('%ebx')
   ELSIF Sym=ECX
   THEN
      WriteString('%ecx')
   ELSIF Sym=EDX
   THEN
      WriteString('%edx')
   ELSIF Sym=ESP
   THEN
      WriteString('%esp')
   ELSIF Sym=EBP
   THEN
      WriteString('%ebp')
   ELSIF Sym=EDI
   THEN
      WriteString('%edi')
   ELSIF Sym=ESI
   THEN
      WriteString('%esi')
   ELSE
      InternalError('no such register', __FILE__, __LINE__)
   END
END WriteRegisterName ;


(*
   MakeQuadMethod
*)

PROCEDURE MakeQuadMethod (quad: CARDINAL) ;
VAR
   Op: QuadOperator;
   Op1, Op2, Op3: CARDINAL ;
BEGIN
   GetQuad(quad, Op, Op1, Op2, Op3) ;
   CASE Op OF

   GotoOp        : MakeCodeGotoMethod    (Op, Op1, Op2, Op3) |
   AddrOp        : MakeCodeAddrMethod    (Op, Op1, Op2, Op3) |
   LineNumberOp  : MakeCodeLineMethod    (Op, Op1, Op2, Op3) |
   BecomesOp     : MakeCodeBecmMethod    (Op, Op1, Op2, Op3) |
   AddOp         : MakeCodeAddMethod     (Op, Op1, Op2, Op3) |
   SubOp         : MakeCodeSubMethod     (Op, Op1, Op2, Op3) |
   NegateOp      : MakeCodeNegMethod     (Op, Op1, Op2, Op3) |
   MultOp        : MakeCodeMultMethod    (Op, Op1, Op2, Op3) |
   ReturnOp      : MakeReturnMethod      (Op, Op1, Op2, Op3) |
   BaseOp        : MakeCodeBaseMethod    (Op, Op1, Op2, Op3)
(*
   NewLocalVarOp : MakeNewLocalVarMethod (Op, Op1, Op2, Op3) |
   KillLocalVarOp: MakeKillLocalVarMethod(Op, Op1, Op2, Op3) |
   ReturnValueOp : MakeReturnValueMethod (Op, Op1, Op2, Op3) |
   IfEquOp       : MakeConditionMethod   (Op, Op1, Op2, Op3)
*)

   ELSE
      InternalError('unknown quadruple in case', __FILE__, __LINE__)
   END
END MakeQuadMethod ;


(*
   SaveCurrentIntoBest - saves the current assembly instructions into the best
                         list of assembly instructions.
*)

PROCEDURE SaveCurrentIntoBest ;
VAR
   i: CARDINAL ;
BEGIN
   WITH EvalStack[EvalPtr-1] DO
      WriteString('------------------------------------------------------') ; WriteLn ;
      WriteString('current list of assembly instructions:  (Cost=') ; WriteCard(CurrentCost, 0) ; Write(')') ; WriteLn ;
      WriteString('                                     :  (Best=') ; WriteCard(BestCost, 0) ; Write(')') ; WriteLn ;
      DisplayMethods ; WriteLn ;
      DumpAliases(GetCurrentAlias()) ;
      WriteString('yield the following instructions:') ; WriteLn ;
      DumpCurrentInstructions ;
      (* save alias *)
      IF BestAlias#NIL
      THEN
         BestAlias := KillAlias(BestAlias)
      END ;
      BestAlias := DuplicateAlias(GetCurrentAlias()) ;
      (* save instructions *)
      i := 0 ;
      WITH CurrentAssemblyCode DO
         i := ListPtr ;
         BestAssemblyCode.ListPtr := i ;
         WHILE i>0 DO
            BestAssemblyCode.ListOfProcedures[i] := ListOfProcedures[i] ;
            DEC(i)
         END
      END ;
      BestCost := CurrentCost
   END
END SaveCurrentIntoBest ;


(*
   DumpCurrentInstructions - dumps the current assembly instructions to the standard output
*)

PROCEDURE DumpCurrentInstructions ;
VAR
   i: CARDINAL ;
BEGIN
   WITH CurrentAssemblyCode DO
      i := 0 ;
      WHILE i<EvalStack[EvalPtr-1].AssemblyPtr DO
         WITH ListOfProcedures[i].WhichParameters DO
            CASE ListOfProcedures[i].WhichProcedure OF

            instbinmd     : BinInstMD(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5]) |
            instbinreg    : BinInstReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2])) |
            instbinstr    : InstKey(ToInst(w[0]), ToCard(w[1])) |
            instmdconst   : InstMDConstant(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5], w[6]) |
            instregconst  : InstRegConstant(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3])) |
            instregreg    : InstRegReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToReg(w[3])) |
            instregmd     : InstRegMD(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3]),
                                      ToCard(w[4]), ToScope(w[5]), ToCard(w[6])) |
            instmdreg     : InstMDReg(ToInst(w[0]), ToDatum(w[1]), ToCard(w[2]), ToCard(w[3]),
                                      ToReg(w[4]), ToScope(w[5]), ToCard(w[6])) |
            codeline      : LineInfo(ToCard(w[0])) |
            instmdoffset  : InstMDOffset(ToInst(w[0]), ToDatum(w[1]), w[2], w[3]) |
            instregimmname: InstRegImmName(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), w[3])

            ELSE
               InternalError('case statement should only be dealing with assembly language instructions', __FILE__, __LINE__)
            END
         END ;
         INC(i)
      END
   END
END DumpCurrentInstructions ;


(*
   CodeFlushAllDirtySymbols - 
*)

PROCEDURE CodeFlushAllDirtySymbols (h: CARDINAL) ;
VAR
   n, i: CARDINAL ;
   al  : ARRAY [0..MaxAlias] OF AliasType ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      n := CollectAllAliases(GetCurrentAlias(), al) ;
      i := 0 ;
      WHILE i<n DO
         IF NOT IsConst(al[i].Sym)
         THEN
            IF IsDirty(GetCurrentAlias(), al[i].Sym, al[i].Mode) AND IsUsedOutSide(al[i].Sym)
            THEN
               IF DebugOn
               THEN
                  Write('{') ; WriteSymbol(al[i].Sym, al[i].Mode) ; WriteString(' is dirty}')
               END ;
               (*
                  symbol is dirty and it is used outside this basic block
                  - hence we must flush it to memory
               *)
               MakeCleanSymbolMethod(al[i].Sym, al[i].Mode) ;
               MakeFlushAllDirtySymbolsMethod ;
               DoEvaluate ;
               RETURN  (* return as we have recurively explored all possabilities *)
            END
         END ;
         INC(i)
      END
   END ;
   DoEvaluate
END CodeFlushAllDirtySymbols ;


(*
   IsUsedOutSide - returns TRUE if symbol is used outside the current basic block.
*)

PROCEDURE IsUsedOutSide (Sym: CARDINAL) : BOOLEAN ;
VAR
   Father,
   Start, End: CARDINAL ;
BEGIN
   Father := GetScopeAuthor(Sym) ;
   IF IsModule(Father) AND IsExported(Father, Sym)
   THEN
      RETURN( TRUE )
   ELSE
      GetVarQuads(Sym, Start, End) ;
      IF (Start<StartQuad) OR (EndQuad<End)
      THEN
         (* is used outside *)
         RETURN( TRUE )
      ELSE
         (* not used outside *)
         IF IsTemporary(Sym)
         THEN
            (* user couldn't care less about temporaries *)
            RETURN( FALSE )
         ELSE
            (* if we are debugging then user may want to see variable anyhow (via gdb) *)
            RETURN( GenerateDebugging )
         END
      END
   END
END IsUsedOutSide ;


(*
   CodeCleanSymbol - 
*)
(* ******************** NEVER USED *******************8
PROCEDURE CodeCleanSymbol (h: CARDINAL) ;
VAR
   c, n, i: CARDINAL ;
   al     : ARRAY [0..MaxAlias] OF AliasType ;
   sym    : CARDINAL ;
   mode   : ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      sym := ToCard(w[0]) ;
      mode := ToMode(w[1]) ;
      IF IsDirty(GetCurrentAlias(), sym, mode)
      THEN
         c := 0 ;
         n := CollectAliases(GetCurrentAlias(), al, sym, mode) ;
         i := 0 ;
         WHILE i<n DO
            IF IsClean(GetCurrentAlias(), al[i].Sym, al[i].Mode) AND ((al[i].Sym#sym) OR (al[i].Mode#mode))
            THEN
               (* ok not moving (a -> a) .. *)
               INC(c) ;
               MakeSymToSymMethod(al[i].Sym, al[i].Mode, sym, mode) ;
               DoEvaluate
            END ;
            INC(i)
         END ;
         IF c=0
         THEN
            DebugMethods ;
            InternalError('failed to find a clean symbol in alias list', __FILE__, __LINE__)
         END
      ELSE
         DoEvaluate
      END
   END
END CodeCleanSymbol ;
*********************************************************** *)

(*
   CodeAssign - updates the Alias list to the effect that an assignment has taken place.
*)

PROCEDURE CodeAssign (h: CARDINAL) ;
VAR
   i, n,
   from, to        : CARDINAL ;
   frommode, tomode: ModeOfAddr ;
   al              : ARRAY [0..MaxAlias] OF AliasType ;
   Found           : BOOLEAN ;
BEGIN
   WITH ArgStack DO
      WITH ListOfProcedures[h].WhichParameters DO
         from     := ToCard(w[0]) ;
         frommode := ToMode(w[1]) ;
         to       := ToCard(w[2]) ;
         tomode   := ToMode(w[3]) ;
         IF (from=to) AND (frommode=tomode)
         THEN
            (* nothing to do *)
            DoEvaluate
         ELSE
            n := CollectAliases(GetCurrentAlias(), al, from, frommode) ;
            RemoveAliasTo(GetCurrentAlias(), to, tomode) ;
            AliasSymbol(GetCurrentAlias(), from, frommode, to, tomode) ;
            MakeClean(GetCurrentAlias(), to, tomode) ;
            (* inherit all from's aliases *)
            i := 0 ;
            WHILE i<n DO
               IF (al[i].Sym#to) OR (al[i].Mode#tomode)
               THEN
                  AliasSymbol(GetCurrentAlias(), al[i].Sym, al[i].Mode, to, tomode)
               END ;
               INC(i)
            END ;
            DoEvaluate
         END
      END
   END
END CodeAssign ;


(*
   CodeWillBeClean - 
*)

PROCEDURE CodeWillBeClean (h: CARDINAL) ;
VAR
   sym : CARDINAL ;
   mode: ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      sym := ToCard(w[0]) ;
      mode := ToMode(w[1]) ;
      MakeClean(GetCurrentAlias(), sym, mode) ;
      DoEvaluate
   END
END CodeWillBeClean ;


(*
   CodeNone -
*)

PROCEDURE CodeNone (h: CARDINAL) ;
BEGIN
   DoEvaluate
END CodeNone ;


(*
   CodeAsm - generic assembly code procedure.
*)

PROCEDURE CodeAsm (h: CARDINAL) ;
BEGIN
   AddToCurrentCode(h) ;  (* assembly code onto current pile *)
   DoEvaluate
END CodeAsm ;


(*
   CodeLine - 
*)

PROCEDURE CodeLine (h: CARDINAL) ;
BEGIN
   AddToCurrentCode(h) ;   (* assembly code onto current pile *)
   DoEvaluate
END CodeLine ;

(*
   CodeGoto - 
*)

PROCEDURE CodeGoto (h: CARDINAL) ;
VAR
   str: ARRAY [0..20] OF CHAR ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      CardToStr(ToCard(w[0]), 0, str) ;
      StrConCat('L', str, str) ;
      MakeInstBinStrMethod(jmp, MakeKey(str)) ;
      DoEvaluate
   END
END CodeGoto ;


(*
------------------------------------------------------------------------------
   Copy Operator
------------------------------------------------------------------------------
   Sym1<O>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
   Sym1<V>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
   Sym1<O>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
*)

PROCEDURE CodeCopy (h: CARDINAL) ;
VAR
   op1, op3: CARDINAL ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      op3 := ToCard(w[2]) ;
      IF IsConst(op3)
      THEN
         (*
            Sym1<O>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
            Sym1<V>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
         *)
         MakeMoveFromToMethod(op3, ImmediateValue, op1, RightValue) ;
         DoEvaluate
      ELSIF GetMode(op3)=RightValue
      THEN
         (*
            Sym1<O>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
            Sym1<V>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
         *)
         MakeMoveFromToMethod(op3, RightValue, op1, RightValue) ;
         DoEvaluate
      ELSE
         (*
            Sym1<O>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
            Sym1<V>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
         *)
         MakeMoveFromToMethod(op3, LeftValue, op1, RightValue) ;
         DoEvaluate
      END
   END
END CodeCopy ;


(*
   CodeMoveFromTo - should only be called via DoNextProcedure
*)

PROCEDURE CodeMoveFromTo (h: CARDINAL) ;
VAR
   i, n            : CARDINAL ;
   al              : ARRAY [0..MaxAlias] OF AliasType ;
   from, to        : CARDINAL ;
   frommode, tomode: ModeOfAddr ;
BEGIN
   WITH ArgStack DO
      WITH ListOfProcedures[h].WhichParameters DO
         from     := ToCard(w[0]) ;
         frommode := ToMode(w[1]) ;
         to       := ToCard(w[2]) ;
         tomode   := ToMode(w[3]) ;
         IF IsAliased(GetCurrentAlias(), from, frommode, to, tomode)
         THEN
            IF IsDirty(GetCurrentAlias(), from, frommode)
            THEN
               (*
                  choices:

                  (i)    do nothing
                  (ii)   clean up dirty symbol
               *)
               IF IsAbleToBeCleaned(GetCurrentAlias(), from, frommode)
               THEN
                  (* now explore doing nothing (i) *)
                  DoEvaluate ;

                  (* try (ii) clean up from symbol *)

                  MakeCleanSymbolMethod(from, frommode) ;
                  DoEvaluate
               ELSE
                  WriteString('unable to clean symbol ') ; WriteSymbol(from, frommode) ; WriteLn ;
                  RETURN ;   (* as we are unable to move any further *)
               END
            ELSE
               (* it is clean *)
               DoEvaluate
            END
         ELSE
            (*
               not aliased therefore choices are:

               (i)   alias it
               (ii)  move it across
               (iii) spill, to, into a previous dirty alias

               (in fact to achieve (i) and (ii) all we need do is to alias it, make it dirty
                and then call MoveFromTo again).

               But before we do that we should check to see whether we should
               copy, to, into a previous alias.
            *)

            IF DebugOn
            THEN
               DumpAliases(GetCurrentAlias())
            END ;
            IF IsClean(GetCurrentAlias(), to, tomode)
            THEN
               n := CollectAliases(GetCurrentAlias(), al, to, tomode) ;
               i := 0 ;
               WHILE i<n DO
                  IF IsDirty(GetCurrentAlias(), al[i].Sym, al[i].Mode) AND
                     (NOT IsAbleToBeCleaned(GetCurrentAlias(), to, tomode))
                  THEN
                     MakeCleanSymbolMethod(al[i].Sym, al[i].Mode)
                  END ;
                  INC(i)
               END
(* we surely do not care if to:tomode is dirty? as we are going to overwrite it?
               ELSE
                  MakeCleanSymbolMethod(to, tomode)
*)
            END ;
            MakeCleanMoveIntoMethod(from, frommode, to, tomode) ;
            DoEvaluate
         END
      END
   END
END CodeMoveFromTo ;


(*
               (*
                  not aliased therefore choices are:

                  (i)   alias it
                  (ii)  move it across
                  (iii) spill, to, into a previous dirty alias

                  (in fact to achieve (i) and (ii) all we need do is to alias it, make it dirty
                   and then call MoveFromTo again).

                  But before we do that we should check to see whether we should
                  copy, to, into a previous alias.
               *)
               IF IsClean(GetCurrentAlias(), to, tomode)
               THEN
                  n := CollectAliases(GetCurrentAlias(), al, to, tomode) ;
                  i := 0 ;
                  WHILE i<n DO
                     IF IsDirty(GetCurrentAlias(), al[i].Sym, al[i].Mode) AND
                        (NOT IsAbleToBeCleaned(GetCurrentAlias(), al[i].Sym, al[i].Mode))
                     THEN
                        MakeCleanSymbolMethod(al[i].Sym, al[i].Mode)
                     END ;
                     INC(i)
                  END
               ELSE
                  MakeCleanSymbolMethod(to, tomode)
               END ;
               MakeCleanMoveIntoMethod(from, frommode, to, tomode) ;
               BestCost := DoNextProcedure(GetCurrentAlias(), BestCost, CurrentCost)
            END
*)



(*
   CodeMoveFromTo - should only be called via DoNextProcedure
*)

(* B R O K E N
PROCEDURE CodeMoveFromTo (h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   i, n            : CARDINAL ;
   al              : ARRAY [0..MaxAlias] OF AliasType ;
   from, to        : CARDINAL ;
   frommode, tomode: ModeOfAddr ;
BEGIN
   WITH ArgStack DO
      WITH ListOfProcedures[h].WhichParameters DO
         from     := ToCard(w[0]) ;
         frommode := ToMode(w[1]) ;
         to       := ToCard(w[2]) ;
         tomode   := ToMode(w[3]) ;
         IF (from=to) AND (frommode=tomode)
         THEN
            (* nothing to do *)
            BestCost := DoNextProcedure(a, BestCost, CurrentCost)
         ELSE
            IF IsAliased(a, from, frommode, to, tomode)
            THEN
               (* nothing to do *)
               BestCost := DoNextProcedure(a, BestCost, CurrentCost)
            ELSE
               (*
                  not aliased therefore:

                  - check to see whether "to" IsClean
                  - if so then check to see whether any dirty
                    symbol is dependent upon this symbol for cleaning
                    then
                       clean symbol
                    fi
                  - alias from -> to
               *)

               (*
                  actually we require more intelligence as we only
                  need to clean the symbol if someone uses the result
                  outside this BB or later on inside the BB - but
                  we cannot tell what will happen later on inside this
                  BB yet. So we will just try and if we get a non infinite
                  cost then it will work anyway...
               *)
               MakeCleanMoveIntoMethod(from, frommode, to, tomode) ;
               BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;
               IF BestCost<MaxCost
               THEN
                  (* how about that - it worked anyhow *)
               ELSE
                  WriteString('no go..') ; WriteLn ;
                  DebugMethods ;
                  WriteLn ;

(*
                  IF IsClean(a, to, tomode)
                  THEN
                     MakeDirty(a, to, tomode) ; (* see whether this is important *)
                     n := CollectAliases(a, al, to, tomode) ;
                     i := 0 ;
                     WHILE i<n DO
                        IF IsDirty(a, al[i].Sym, al[i].Mode) AND
                           (NOT IsAbleToBeCleaned(a, al[i].Sym, al[i].Mode))
                        THEN
                           MakeCleanSymbolMethod(al[i].Sym, al[i].Mode)
                        END ;
                        INC(i)
                     END ;
                     MakeClean(a, to, tomode)    (* restore orig state *)
                  ELSE
                     InternalError('CodeMoveFromTo should never get here', __FILE__, __LINE__)
                  END ;
                  BestCost := DoNextProcedure(a, BestCost, CurrentCost)
*)
               END
            END
         END
      END
   END
END CodeMoveFromTo ;
**********************)

(*
   MakeCleanMoveIntoMethod - 
*)

PROCEDURE MakeCleanMoveIntoMethod (from: CARDINAL; frommode: ModeOfAddr;
                                   to  : CARDINAL; tomode  : ModeOfAddr) ;
BEGIN
   PushMethod(moveinto,
              from, frommode, to, tomode, 0, 0, 0, 0)
END MakeCleanMoveIntoMethod ;


(*
   CodeCleanMoveInto - the symbol [to:tomode] is now clean, we simply remove
                       all aliases to [to:tomode] and then alias [from:frommode]
                       to this symbol after which we make [to:tomode] dirty.
                       We then call MakeMoveFrom again.
*)

PROCEDURE CodeCleanMoveInto (h: CARDINAL) ;
VAR
   from, to        : CARDINAL ;
   frommode, tomode: ModeOfAddr ;
   b               : AliasList ;
BEGIN
   WITH ArgStack DO
      WITH ListOfProcedures[h].WhichParameters DO
         from     := ToCard(w[0]) ;
         frommode := ToMode(w[1]) ;
         to       := ToCard(w[2]) ;
         tomode   := ToMode(w[3]) ;
         IF IsDirty(GetCurrentAlias(), from, frommode)
         THEN
            IF IsAbleToBeCleaned(GetCurrentAlias(), from, frommode)
            THEN
               b := DuplicateAlias(GetCurrentAlias()) ;
               RemoveAliasTo(GetCurrentAlias(), to, tomode) ;   (* remove all previous aliases to [to:tomode] *)
               AliasSymbol(GetCurrentAlias(), from, frommode, to, tomode) ;
               MakeDirty(GetCurrentAlias(), to, tomode) ;

               (* now explore doing nothing (i) *)
               DoEvaluate ;

               EvalStack[EvalPtr-1].CurrentAlias := KillAlias(EvalStack[EvalPtr-1].CurrentAlias) ;
               EvalStack[EvalPtr-1].CurrentAlias := b ;

               (* try (ii) clean up from symbol *)
               MakeCleanSymbolMethod(from, frommode) ;
               MakeCleanMoveIntoMethod(from, frommode, to, tomode) ;    (* recursion but it will be clean next time *)
               DoEvaluate
            ELSE
               WriteString('cannot clean symbol ') ; WriteSymbol(from, frommode) ; WriteLn ;
               RETURN
            END
         ELSE
            (* must be clean *)
            RemoveAliasTo(GetCurrentAlias(), to, tomode) ;   (* remove all previous aliases to [to:tomode] *)
            AliasSymbol(GetCurrentAlias(), from, frommode, to, tomode) ;
            MakeDirty(GetCurrentAlias(), to, tomode) ;
            DoEvaluate
         END
      END
   END
END CodeCleanMoveInto ;


(*
   WriteCleanMoveInto -
*)

PROCEDURE WriteCleanMoveInto (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('CleanMoveInto( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteCleanMoveInto ;


(*
------------------------------------------------------------------------------
   := Operator
------------------------------------------------------------------------------
   Sym1<I> := Sym2<I>           := produces a constant
   Sym1<O> := Sym2<O>           := has the effect Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O> := Sym2<V>           := has the effect
                                             Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V> := Sym2<O>           := has the effect
                                             Mem[Mem[Sym1<I>]] := Mem[Sym2<I>]
   Sym1<V> := Sym2<V>           := has the effect
                                        Mem[Mem[Sym1<I>]] := Mem[Mem[Sym2<I>]]
*)

PROCEDURE CodeBecm (h: CARDINAL) ;
VAR
   op1, op3: CARDINAL ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      op3 := ToCard(w[2]) ;

      IF IsConst(op1) AND IsConst(op3)
      THEN
         PushValue(op3) ;
         PopValue(op1) ;
         DoEvaluate
      ELSIF IsConstString(op3)
      THEN
         InternalError('not yet implemented MoveStringFromTo', __FILE__, __LINE__) ;
         (* MoveStringFromTo(Operand3, Operand1) ; *)
         DoEvaluate
      ELSIF IsProcedure(op3)
      THEN
         MakeMoveFromToMethod(op3, GetMode(op3),
                              op1, GetMode(op1)) ;
         DoEvaluate
      ELSE
         IF IsCoprocessorType(op1) OR IsCoprocessorType(op3)
         THEN
            InternalError('not yet implemented floating point for this optimized code generator', __FILE__, __LINE__) ;
            (* CoprocessorBecomes(Operand1, Operand3) *)
         ELSE
            MakeMoveFromToMethod(op3, GetMode(op3),
                                 op1, GetMode(op1)) ;
            DoEvaluate
         END
      END
   END
END CodeBecm ;


(*
   Addr Operator  - contains the address of a variable.

   Yields the address of a variable - need to add the frame pointer if
   a variable is local to a procedure.

   Sym1<O>   Addr   Sym2<O>     meaning     Mem[Sym1<I>] := Sym2<I>
   Sym1<V>   Addr   Sym2<O>     meaning     Mem[Sym1<I>] := Sym2<I>
   Sym1<O>   Addr   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Addr   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
*)

PROCEDURE CodeAddr (h: CARDINAL) ;
VAR
   op1, op3: CARDINAL ;
   i, n    : CARDINAL ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      op3 := ToCard(w[2]) ;

      (*
         Follow the Quadruple rules:
      *)
      IF IsConst(op3)
      THEN
         WriteError('Not expecting a constant')
      ELSE
         IF GetMode(op3)=RightValue
         THEN
            (*
               Mem[Op1] := Op3
            *)
            IF IsProcedure(GetVarFather(op3))
            THEN
               (* (i)   ebp -> r; r += offset(op3); r -> op1 *)
               (* (ii)  offset(op3) -> r; r += ebp; r -> op1 *)

               i := 0 ;
               n := NoOfRegisters() ;
               WHILE i<n DO
                  IF (GetRegister(i)#EBP) OR (GetOffset(op3)=0)
                  THEN
                     (* must not mess with our frame pointer *)

                     (* (i) *)
                     MakeMoveFromToMethod(EBP, RightValue, GetRegister(i), RightValue) ;
                     IF GetOffset(op3)>0
                     THEN
                        MakeIncMethod(GetRegister(i), RightValue, MakeConstLitFromInt(GetOffset(op3)), ImmediateValue)
                     ELSIF GetOffset(op3)<0
                     THEN
                        MakeDecMethod(GetRegister(i), RightValue, MakeConstLitFromInt(GetOffset(op3)), ImmediateValue)
                     END ;
                     MakeMoveFromToMethod(GetRegister(i), RightValue, op1, GetMode(op1)) ;
                     DoEvaluate ;

                     (* (ii) *)

                     MakeMoveFromToMethod(MakeConstLitFromInt(GetOffset(op3)), ImmediateValue, GetRegister(i), RightValue) ;
                     MakeDecMethod(GetRegister(i), RightValue, EBP, RightValue) ;
                     MakeMoveFromToMethod(GetRegister(i), RightValue, op1, GetMode(op1)) ;
                     DoEvaluate
                  END ;
                  INC(i)
               END
            ELSE
               (* (i)   ADR(op3) -> op1 *)
               MakeCleanSymbolMethod(op1, GetMode(op1)) ;
               MakeInstMDOffsetMethod(mov, WordSize, op1, op3) ;
               MakeChanged(op1, GetMode(op1)) ;
               DoEvaluate ;

               (* (ii)  ADR(op3) -> r -> op1 *)
               i := 0 ;
               n := NoOfRegisters() ;
               WHILE i<n DO
                  MakeCleanSymbolMethod(GetRegister(i), RightValue) ;
                  MakeInstRegImmNameMethod(mov, WordSize, GetRegister(i), op3) ;
                  MakeChanged(GetRegister(i), RightValue) ;
                  MakeMoveFromToMethod(GetRegister(i), RightValue, op1, GetMode(op1)) ;
                  DoEvaluate ;
                  INC(i)
               END
            END
         ELSE
            (*
               Mem[Op1] := Mem[Op3]
            *)

            Assert(GetMode(op3)=LeftValue) ;
            (*
               Consistancy check to see whether the SIZE(Operand1)=SIZE(ADDRESS)
               If this should ever fail then we know the compiler would produce
               incorrect code.
            *)
            Assert(GetSize(op1)=AddressSize) ;

            MakeMoveFromToMethod(op3, OffsetUnlessConst(op3), op1, OffsetUnlessConst(op1)) ;
            DoEvaluate
         END
      END
   END
END CodeAddr ;


(*
   MakeCodeAddMethod - Briefly:
                                Operand1 := Operand2 + Operand3
                                Checks for Constant folding - otherwise a method is generated.
*)

PROCEDURE MakeCodeAddMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsConst(op1)
   THEN
      PushValue(op3) ;
      PushValue(op2) ;
      Add ;
      PopValue(op1)
   ELSIF IsCoprocessorUsed(op1, op2, op3)
   THEN
      InternalError('not yet implemented', __FILE__, __LINE__) ;
      (* CoprocessorAdd(Operand1, Operand2, Operand3) *)
   ELSE
      PushMethod(codeadd,
                 op1, op2, op3, 0, 0, 0, 0, 0)
   END
END MakeCodeAddMethod ;


(*
   WriteCodeAdd - 
*)

PROCEDURE WriteCodeAdd (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[1]), RightValue) ;
      WriteString(' + ') ; WriteSymbol(ToCard(w[2]), RightValue) ;
      WriteString(' -> ') ; WriteSymbol(ToCard(w[0]), RightValue) ;
      WriteString(' )')
   END
END WriteCodeAdd ;


(*
   IsAliasedToZero - 
*)

PROCEDURE IsAliasedToZero (Sym: CARDINAL; Mode: ModeOfAddr) : BOOLEAN ;
VAR
   al: ARRAY [0..MaxAlias] OF AliasType ;
   n,
   i   : CARDINAL ;
BEGIN   
   n := CollectAliases(GetCurrentAlias(), al, Sym, Mode) ;
   i := 0 ;
   WHILE i<n DO
      IF al[i].Sym=Zero
      THEN
         IF al[i].Mode#ImmediateValue
         THEN
            InternalError('constant type is not immediate ??', __FILE__, __LINE__)
         END ;
         RETURN( TRUE )
      END ;
      INC(i)
   END ;
   RETURN( FALSE )
END IsAliasedToZero ;


(*
   IsAliasedToOne - 
*)

PROCEDURE IsAliasedToOne (Sym: CARDINAL; Mode: ModeOfAddr) : BOOLEAN ;
VAR
   al: ARRAY [0..MaxAlias] OF AliasType ;
   n,
   i   : CARDINAL ;
BEGIN   
   n := CollectAliases(GetCurrentAlias(), al, Sym, Mode) ;
   i := 0 ;
   WHILE i<n DO
      IF al[i].Sym=One
      THEN
         IF al[i].Mode#ImmediateValue
         THEN
            InternalError('constant type is not immediate ??', __FILE__, __LINE__)
         END ;
         RETURN( TRUE )
      END ;
      INC(i)
   END ;
   RETURN( FALSE )
END IsAliasedToOne ;


(*
   IsAliasedToTwo - 
*)

PROCEDURE IsAliasedToTwo (Sym: CARDINAL; Mode: ModeOfAddr) : BOOLEAN ;
VAR
   al: ARRAY [0..MaxAlias] OF AliasType ;
   n,
   i   : CARDINAL ;
BEGIN   
   n := CollectAliases(GetCurrentAlias(), al, Sym, Mode) ;
   i := 0 ;
   WHILE i<n DO
      IF al[i].Sym=Two
      THEN
         IF al[i].Mode#ImmediateValue
         THEN
            InternalError('constant type is not immediate ??', __FILE__, __LINE__)
         END ;
         RETURN( TRUE )
      END ;
      INC(i)
   END ;
   RETURN( FALSE )
END IsAliasedToTwo ;


(*
   OffsetUnlessConst - returns RightValue unless, sym, is a constant in
                       which case ImmediateValue is returned.
*)

PROCEDURE OffsetUnlessConst (sym: CARDINAL) : ModeOfAddr ;
BEGIN
   IF IsConst(sym)
   THEN
      RETURN( ImmediateValue )
   ELSE
      RETURN( RightValue )
   END
END OffsetUnlessConst ;


(*
   DoCombinations2 - runs through all combinations of op1, op2, aliases returning the best cost.
                     It only tests each combination once.
*)

PROCEDURE DoCombinations2 (a: AliasList;
                           op1, op2: CARDINAL;
                           p: CombinationProc2) ;
VAR
   l1, l2,
   b1, b2: ARRAY [0..MaxAlias] OF AliasType ;
   IsNew : BOOLEAN ;
   c, j,
   i1, i2,
   n1, n2: CARDINAL ;
BEGIN
   c := 0 ;  (* how many combinations have we tried? *)
   n1 := CollectAliases(a, b1, op1, OffsetUnlessConst(op1)) ;
   n2 := CollectAliases(a, b2, op2, OffsetUnlessConst(op2)) ;
   i1 := 0 ;
   WHILE i1<n1 DO
      IF NOT IsConstLit(b1[i1].Sym)
      THEN
         i2 := 0 ;
         WHILE i2<n2 DO
            j := 0 ;
            IsNew := TRUE ;
            WHILE (j<c) AND IsNew DO
               IF (b1[i1].Sym=l1[j].Sym) AND (b2[i2].Sym=l2[j].Sym) AND
                  (b1[i1].Mode=l1[j].Mode) AND (b2[i2].Mode=l2[j].Mode)
               THEN
                  IsNew := FALSE
               END ;
               INC(j)
            END ;
            IF IsNew
            THEN
               l1[c].Sym  := b1[i1].Sym ;
               l1[c].Mode := b1[i1].Mode ;
               l2[c].Sym  := b2[i2].Sym ;
               l2[c].Mode := b2[i2].Mode ;
               INC(c) ;
               p(b1[i1].Sym, b1[i1].Mode, b2[i2].Sym, b2[i2].Mode) ;
               DoEvaluate
            END ;
            INC(i2)
         END
      END ;
      INC(i1)
   END ;
   p(op1, OffsetUnlessConst(op1), op2, OffsetUnlessConst(op2)) ;
   DoEvaluate
END DoCombinations2 ;


(*
   DoCombinations3 - runs through all combinations of op1, op2, op3 aliases returning the best cost.
                     It only tests each combination once. (This routine was initially part of CodeAdd
                     and CodeSub but now has been factored into a separate routine).
*)

PROCEDURE DoCombinations3 (a: AliasList;
                           op1, op2, op3: CARDINAL;
                           p: CombinationProc3) ;
VAR
   l1, l2, l3,
   b1, b2, b3   : ARRAY [0..MaxAlias] OF AliasType ;
   IsNew        : BOOLEAN ;
   c, j,
   i1, i2, i3,
   n1, n2, n3   : CARDINAL ;
BEGIN
   c := 0 ;  (* how many combinations have we tried? *)
   n1 := CollectAliases(a, b1, op1, OffsetUnlessConst(op1)) ;
   n2 := CollectAliases(a, b2, op2, OffsetUnlessConst(op2)) ;
   n3 := CollectAliases(a, b3, op3, OffsetUnlessConst(op3)) ;
   i1 := 0 ;
   WHILE i1<n1 DO
      IF NOT IsConstLit(b1[i1].Sym)
      THEN
         i2 := 0 ;
         WHILE i2<n2 DO
            i3 := 0 ;
            WHILE i3<n3 DO
               j := 0 ;
               IsNew := TRUE ;
               WHILE (j<c) AND IsNew DO
                  IF (b1[i1].Sym=l1[j].Sym) AND (b2[i2].Sym=l2[j].Sym) AND (b3[i3].Sym=l3[j].Sym) AND
                     (b1[i1].Mode=l1[j].Mode) AND (b2[i2].Mode=l2[j].Mode) AND (b3[i3].Mode=l3[j].Mode)
                  THEN
                     IsNew := FALSE
                  END ;
                  INC(j)
               END ;
               IF IsNew
               THEN
                  l1[c].Sym  := b1[i1].Sym ;
                  l1[c].Mode := b1[i1].Mode ;
                  l2[c].Sym  := b2[i2].Sym ;
                  l2[c].Mode := b2[i2].Mode ;
                  l3[c].Sym  := b3[i3].Sym ;
                  l3[c].Mode := b3[i3].Mode ;
                  INC(c) ;
                  p(b1[i1].Sym, b1[i1].Mode, b2[i2].Sym, b2[i2].Mode, b3[i3].Sym, b3[i3].Mode) ;
                  DoEvaluate
               END ;
               INC(i3)
            END ;
            INC(i2)
         END
      END ;
      INC(i1)
   END ;
   p(op1, OffsetUnlessConst(op1), op2, OffsetUnlessConst(op2), op3, OffsetUnlessConst(op3)) ;
   DoEvaluate
END DoCombinations3 ;


(*
   CodeAdd - symbols are referenced by Mem[Sym]
*)

PROCEDURE CodeAdd (h: CARDINAL) ;
VAR
   op1, op2, op3: CARDINAL ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      op2 := ToCard(w[1]) ;
      op3 := ToCard(w[2]) ;
      IF IsAliasedToZero(op2, OffsetUnlessConst(op2))
      THEN
         MakeMoveFromToMethod(op3, OffsetUnlessConst(op3), op1, OffsetUnlessConst(op1)) ;
         DoEvaluate
      ELSIF IsAliasedToZero(op3, OffsetUnlessConst(op3))
      THEN
         MakeMoveFromToMethod(op2, OffsetUnlessConst(op2), op1, OffsetUnlessConst(op1)) ;
         DoEvaluate
      ELSE
         DoCombinations3(GetCurrentAlias(), op1, op2, op3, MakePeepAddMethod)
      END
   END
END CodeAdd ;


(*
   SwapSymbols - 
*)

PROCEDURE SwapSymbols (VAR s1: CARDINAL; VAR m1: ModeOfAddr; VAR s2: CARDINAL; VAR m2: ModeOfAddr) ;
VAR
   t: CARDINAL ;
   n: ModeOfAddr ;
BEGIN
   t := s1 ; s1 := s2 ; s2 := t ;
   n := m1 ; m1 := m2 ; m2 := n
END SwapSymbols ;


(*
   MakePeepAddMethod - 
*)

PROCEDURE MakePeepAddMethod (op1: CARDINAL; m1: ModeOfAddr;
                             op2: CARDINAL; m2: ModeOfAddr;
                             op3: CARDINAL; m3: ModeOfAddr) ;
BEGIN
   PushMethod(peepadd,
              op1, m1, op2, m2, op3, m3, 0, 0)
END MakePeepAddMethod ;


(*
   CodePeepAdd - 
*)

PROCEDURE CodePeepAdd (h: CARDINAL) ;
VAR
   i, j, n,
   size,
   op1, op2, op3: CARDINAL ;
   m1, m2, m3   : ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      m1 := ToMode(w[1]) ;
      op2 := ToCard(w[2]) ;
      m2 := ToMode(w[3]) ;
      op3 := ToCard(w[4]) ;
      m3 := ToMode(w[5]) ;

(*
      WriteLn ;
      WriteString('( ') ;
      WriteSymbol(op1, m1) ; WriteString(' := ') ;
      WriteSymbol(op2, m2) ; WriteString(' + ') ;
      WriteSymbol(op3, m3) ; WriteString(')') ;
      WriteLn ;
*)
      IF IsConst(op2) AND IsConst(op3)
      THEN
         IF IsConst(op1)
         THEN
            PushValue(op2) ;
            PushValue(op3) ;
            Add ;
            PopValue(op1) ;
         ELSE
            PushValue(op2) ;
            PushValue(op3) ;
            Add ;
            MakeMoveFromToMethod(MakeConstLitFromInt(PopInt()), ImmediateValue, op1, m1) ;
            DoEvaluate
         END
      ELSE
         IF IsAliased(GetCurrentAlias(), op2, OffsetUnlessConst(op2), op1, OffsetUnlessConst(op1))
         THEN
            MakeIncMethod(op1, OffsetUnlessConst(op1), op3, OffsetUnlessConst(op3)) ;
            DoEvaluate
         END ;
         IF IsAliased(GetCurrentAlias(), op3, OffsetUnlessConst(op3), op1, OffsetUnlessConst(op1))
         THEN
            MakeIncMethod(op1, OffsetUnlessConst(op1), op2, OffsetUnlessConst(op2)) ;
            DoEvaluate
         END ;

         IF IsConst(op2)
         THEN
            SwapSymbols(op2, m2, op3, m3)
         END ;
         size := GetSize(op2) ;   (* op2 is not a Const *)
         (*
            A number of choices:

            (i)    op2 -> op1 ; inc op1, op3   when op1 is not op3
            (ii)   op3 -> op1 ; inc op1, op2   when op1 is not op2
         *)

         IF op1#op3
         THEN
            (* (i)   op2 -> op1 ; op1 += op3 *)
            MakeMoveFromToMethod(op2, m2, op1, m1) ;
            MakeIncMethod(op1, m1, op3, m3) ;
            DoEvaluate
         END ;

         IF op1#op2
         THEN
            (* (ii)  op3 -> op1 ; op1 += op2 *)
            MakeMoveFromToMethod(op3, m3, op1, m1) ;
            MakeIncMethod(op1, m1, op2, m2) ;
            DoEvaluate
         END ;

         (* (iii) op3 -> r ; r += op2 ; r -> op1 *)
         i := 0 ;
         n := NoOfRegisters() ;
         WHILE i<n DO
            IF (op3=op2) OR (GetRegister(i)#op2)
            THEN
               MakeMoveFromToMethod(op3, m3, GetRegister(i), RightValue) ;
               MakeIncMethod(GetRegister(i), RightValue, op2, m2) ;
               MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
               DoEvaluate
            END ;
            INC(i)
         END ;
         (* (iv)  op2 -> r ; r += op3 ; r -> op1 *)
         i := 0 ;
         n := NoOfRegisters() ;
         WHILE i<n DO
            IF (op2#op3) OR (GetRegister(i)#op3)
            THEN
               MakeMoveFromToMethod(op2, m2, GetRegister(i), RightValue) ;
               MakeIncMethod(GetRegister(i), RightValue, op3, m3) ;
               MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
               DoEvaluate
            END ;
            INC(i)
         END ;
         (* (v)  op3 -> r1 ; op2 -> r2 ; r1 += r2 ; r1 -> op1 *)
         IF IsAliased(GetCurrentAlias(), op2, OffsetUnlessConst(op2), op3, OffsetUnlessConst(op3))
         THEN
            (* (v case i)  op3 -> r ; r += r ; r -> op1 *)
            i := 0 ;
            n := NoOfRegisters() ;
            WHILE i<n DO
               MakeMoveFromToMethod(op3, m3, GetRegister(i), RightValue) ;
               MakeIncMethod(GetRegister(i), RightValue, GetRegister(i), RightValue) ;
               MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
               DoEvaluate ;
               INC(i)
            END
         ELSE
            (* (v)  op3 -> r1 ; op2 -> r2 ; r1 += r2 ; r1 -> op1 *)
            i := 0 ;
            n := NoOfRegisters() ;
            WHILE i<n DO
               j := 0 ;
               WHILE j<n DO
                  IF (j#i) AND (GetRegister(i)#op2) AND (GetRegister(j)#op3)
                  THEN
                     MakeMoveFromToMethod(op3, m3, GetRegister(i), RightValue) ;
                     MakeMoveFromToMethod(op2, m2, GetRegister(j), RightValue) ;
                     MakeIncMethod(GetRegister(i), RightValue, GetRegister(j), RightValue) ;
                     MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
                     DoEvaluate
                  END ;
                  INC(j)
               END ;
               INC(i)
            END
         END
      END
   END
END CodePeepAdd ;


(*
   WritePeepAdd - 
*)

PROCEDURE WritePeepAdd (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' peep+ ') ; WriteSymbol(ToCard(w[4]), ToMode(w[5])) ;
      WriteString(' -> ') ; WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WritePeepAdd ;


(*
   WriteCodeInc - 
*)

PROCEDURE WriteCodeInc (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' += -> ') ; WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WriteCodeInc ;


(*
   MakeIncMethod - 
*)

PROCEDURE MakeIncMethod (op1: CARDINAL; m1: ModeOfAddr; op2: CARDINAL; m2: ModeOfAddr) ;
BEGIN
   PushMethod(codeinc,
              op1, m1, op2, m2, 0, 0, 0, 0)
END MakeIncMethod ;


(*
   CodeInc -
*)

PROCEDURE CodeInc (h: CARDINAL) ;
VAR
   al      : ARRAY [0..MaxAlias] OF AliasType ;
   size,
   i, j, n,
   op1, op2: CARDINAL ;
   m1, m2  : ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      m1 := ToMode(w[1]) ;
      op2 := ToCard(w[2]) ;
      m2 := ToMode(w[3]) ;

      IF IsConstLit(op1)
      THEN
         InternalError('should not be incrementing a constant', __FILE__, __LINE__)
      END ;
      size := GetSize(op1) ;
      IF IsAliasedToOne(op2, m2)
      THEN
         n := CollectAliases(GetCurrentAlias(), al, op1, m1) ;
         i := 0 ;
         WHILE i<n DO
            WITH al[i] DO
               MakeCleanSymbolMethod(Sym, Mode) ;
               MakeInstBinMethod(inc, size, Sym, Mode) ;
               MakeChanged(Sym, Mode) ;
               MakeMoveFromToMethod(Sym, Mode, op1, m1) ;
               DoEvaluate
            END ;
            INC(i)
         END
      END ;
      IF IsRegister(op1) AND IsRegister(op2)
      THEN
         MakeCleanSymbolMethod(op2, m2) ;
         MakeCleanSymbolMethod(op1, m1) ;
         MakeInstRegRegMethod(add, size, op1, op2) ;
         MakeChanged(op1, m1) ;
         DoEvaluate
      END ;
      i := 0 ;
      n := NoOfRegisters() ;
      WHILE i<n DO
         IF op2#GetRegister(i)
         THEN
            (* (i)  op1 -> r; r += op2 ; r -> op1 *)
            MakeMoveFromToMethod(op1, m1, GetRegister(i), RightValue) ;
            MakePeepIncMethod(GetRegister(i), RightValue, op2, m2) ;
            MakeChanged(GetRegister(i), RightValue) ;
            MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
            DoEvaluate
         END ;

         IF op1#GetRegister(i)
         THEN
            (* (ii)  op2 -> r; r += op1 ; r -> op1 *)
            MakeMoveFromToMethod(op2, m2, GetRegister(i), RightValue) ;
            MakePeepIncMethod(GetRegister(i), RightValue, op1, m1) ;
            MakeChanged(GetRegister(i), RightValue) ;
            MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
            DoEvaluate
         END ;

         j := 0 ;
         WHILE j<n DO
            IF (i#j) OR ((op1=op2) AND (m1=m2))
            THEN
               (* (iii)  op1 -> r1 ; op2 -> r2 ; r1 += r2 ; r1 -> op1 *)
               MakeMoveFromToMethod(op1, m1, GetRegister(i), RightValue) ;
               MakeMoveFromToMethod(op2, m2, GetRegister(j), RightValue) ;
               MakePeepIncMethod(GetRegister(i), RightValue, GetRegister(j), RightValue) ;
               MakeChanged(GetRegister(i), RightValue) ;
               MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
               DoEvaluate
            END ;
            INC(j)
         END ;
         INC(i) ;
      END
   END
END CodeInc ;


(*
   WriteCodePeepInc - 
*)

PROCEDURE WriteCodePeepInc (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' peep += -> ') ; WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WriteCodePeepInc ;


(*
   MakePeepIncMethod - 
*)

PROCEDURE MakePeepIncMethod (op1: CARDINAL; m1: ModeOfAddr; op2: CARDINAL; m2: ModeOfAddr) ;
BEGIN
   IF NOT (IsRegister(op1) OR IsRegister(op2))
   THEN
      InternalError('expecting one of the operands to be a register', __FILE__, __LINE__)
   END ;
   PushMethod(codepeepinc,
              op1, m1, op2, m2, 0, 0, 0, 0)
END MakePeepIncMethod ;


(*
   CodePeepInc - it assumes that op1 or op2 is a register.
*)

PROCEDURE CodePeepInc (h: CARDINAL) ;
VAR
   al      : ARRAY [0..MaxAlias] OF AliasType ;
   size,
   i, j, n,
   op1, op2: CARDINAL ;
   m1, m2  : ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      m1 := ToMode(w[1]) ;
      op2 := ToCard(w[2]) ;
      m2 := ToMode(w[3]) ;
      size := GetSize(op1) ;
      IF IsConstLit(op1)
      THEN
         InternalError('should not be incrementing a constant', __FILE__, __LINE__)
      END ;
      size := GetSize(op1) ;
      IF IsAliasedToOne(op2, m2)
      THEN
         n := CollectAliases(GetCurrentAlias(), al, op1, m1) ;
         i := 0 ;
         WHILE i<n DO
            MakeCleanSymbolMethod(op1, m1) ;
            MakeInstBinMethod(inc, size, op1, m1) ;
            MakeChanged(op1, m1) ;
            DoEvaluate ;
            INC(i)
         END
      END ;
      (* add op1, op2 *)
      MakeCleanSymbolMethod(op2, m2) ;
      MakeCleanSymbolMethod(op1, m1) ;
      IF IsRegister(op1) AND IsRegister(op2)
      THEN
         MakeInstRegRegMethod(add, size, op1, op2)
      ELSIF IsRegister(op1)
      THEN
         IF IsConst(op2)
         THEN
            (* add r, #op2 *)
            MakeInstRegConstantMethod(add, size, op1, op2)
         ELSE
            MakeInstRegMDMethod(add, size, op1, GetOffset(op2), 0, GetScope(op2), op2)
         END
      ELSIF IsRegister(op2)
      THEN
         MakeInstMDRegMethod(add, size, 0, GetOffset(op1), GetScope(op1), op1, op2)
      END ;
      MakeChanged(op1, m1) ;
      DoEvaluate
   END
END CodePeepInc ;


(*
   MakeCodeSubMethod - Briefly:
                                Operand1 := Operand2 - Operand3
                                Checks for Constant folding - otherwise a method is generated.
*)

PROCEDURE MakeCodeSubMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsConst(op1)
   THEN
      PushValue(op3) ;
      PushValue(op2) ;
      Sub ;
      PopValue(op1)
   ELSIF IsCoprocessorUsed(op1, op2, op3)
   THEN
      InternalError('not yet implemented', __FILE__, __LINE__) ;
      (* CoprocessorSub(Operand1, Operand2, Operand3) *)
   ELSE
      PushMethod(codesub,
                 op1, op2, op3, 0, 0, 0, 0, 0)
   END
END MakeCodeSubMethod ;


(*
   WriteCodeSub - 
*)

PROCEDURE WriteCodeSub (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[1]), RightValue) ;
      WriteString(' - ') ; WriteSymbol(ToCard(w[2]), RightValue) ;
      WriteString(' -> ') ; WriteSymbol(ToCard(w[0]), RightValue) ;
      WriteString(' )')
   END
END WriteCodeSub ;


(*
   CodeDec -
*)

PROCEDURE CodeDec (h: CARDINAL) ;
VAR
   al      : ARRAY [0..MaxAlias] OF AliasType ;
   size,
   i, j, n,
   op1, op2: CARDINAL ;
   m1, m2  : ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      m1 := ToMode(w[1]) ;
      op2 := ToCard(w[2]) ;
      m2 := ToMode(w[3]) ;

      IF IsConstLit(op1)
      THEN
         InternalError('should not be decrementing a constant', __FILE__, __LINE__)
      END ;
      size := GetSize(op1) ;
      IF IsAliasedToOne(op2, m2)
      THEN
         n := CollectAliases(GetCurrentAlias(), al, op1, m1) ;
         i := 0 ;
         WHILE i<n DO
            WITH al[i] DO
               MakeCleanSymbolMethod(Sym, Mode) ;
               MakeInstBinMethod(dec, size, Sym, Mode) ;
               MakeChanged(Sym, Mode) ;
               MakeMoveFromToMethod(Sym, Mode, op1, m1) ;
               DoEvaluate
            END ;
            INC(i)
         END
      END ;
      IF IsRegister(op1) AND IsRegister(op2)
      THEN
         MakeCleanSymbolMethod(op1, m1) ;
         MakeCleanSymbolMethod(op2, m2) ;
         MakeInstRegRegMethod(sub, size, op1, op2) ;
         MakeChanged(op1, m1) ;
         DoEvaluate
      END ;
      i := 0 ;
      n := NoOfRegisters() ;
      WHILE i<n DO
         (* (i)  op1 -> r; r -= op2 ; r -> op1 *)
         MakeMoveFromToMethod(op1, m1, GetRegister(i), RightValue) ;
         MakePeepDecMethod(GetRegister(i), RightValue, op2, m2) ;
         MakeChanged(GetRegister(i), RightValue) ;
         MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
         DoEvaluate ;

         j := 0 ;
         WHILE j<n DO
            IF (i#j) OR ((op1=op2) AND (m1=m2))
            THEN
               (* (ii)  op1 -> r1 ; op2 -> r2 ; r1 -= r2 ; r1 -> op1 *)
               MakeMoveFromToMethod(op1, m1, GetRegister(i), RightValue) ;
               MakeMoveFromToMethod(op2, m2, GetRegister(j), RightValue) ;
               MakePeepDecMethod(GetRegister(i), RightValue, GetRegister(j), RightValue) ;
               MakeChanged(GetRegister(i), RightValue) ;
               MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
               DoEvaluate
            END ;
            INC(j)
         END ;
         INC(i) ;
      END
   END
END CodeDec ;


(*
   WriteCodePeepDec - 
*)

PROCEDURE WriteCodePeepDec (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' peep -= -> ') ; WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WriteCodePeepDec ;


(*
   MakePeepDecMethod - 
*)

PROCEDURE MakePeepDecMethod (op1: CARDINAL; m1: ModeOfAddr; op2: CARDINAL; m2: ModeOfAddr) ;
BEGIN
   IF NOT (IsRegister(op1) OR IsRegister(op2))
   THEN
      InternalError('expecting one of the operands to be a register', __FILE__, __LINE__)
   END ;
   PushMethod(codepeepdec,
              op1, m1, op2, m2, 0, 0, 0, 0)
END MakePeepDecMethod ;


(*
   CodePeepDec - it assumes that op1 or op2 is a register.
*)

PROCEDURE CodePeepDec (h: CARDINAL) ;
VAR
   al      : ARRAY [0..MaxAlias] OF AliasType ;
   size,
   i, j, n,
   op1, op2: CARDINAL ;
   m1, m2  : ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      m1 := ToMode(w[1]) ;
      op2 := ToCard(w[2]) ;
      m2 := ToMode(w[3]) ;
      size := GetSize(op1) ;
      IF IsConstLit(op1)
      THEN
         InternalError('should not be incrementing a constant', __FILE__, __LINE__)
      END ;
      size := GetSize(op1) ;
      IF IsAliasedToOne(op2, m2)
      THEN
         n := CollectAliases(GetCurrentAlias(), al, op1, m1) ;
         i := 0 ;
         WHILE i<n DO
            MakeCleanSymbolMethod(op1, m1) ;
            MakeInstBinMethod(dec, size, op1, m1) ;
            MakeChanged(op1, m1) ;
            DoEvaluate ;
            INC(i)
         END
      END ;
      (* sub op1, op2 *)
      MakeCleanSymbolMethod(op1, m1) ;
      MakeCleanSymbolMethod(op2, m2) ;
      IF IsRegister(op1) AND IsRegister(op2)
      THEN
         MakeInstRegRegMethod(sub, size, op1, op2)
      ELSIF IsRegister(op1)
      THEN
         IF IsConst(op2)
         THEN
            (* sub op1, #op2 *)
            MakeInstRegConstantMethod(sub, size, op1, op2)
         ELSE
            MakeInstRegMDMethod(sub, size, op1, GetOffset(op2), 0, GetScope(op2), op2)
         END
      ELSIF IsRegister(op2)
      THEN
         MakeInstMDRegMethod(sub, size, 0, GetOffset(op1), GetScope(op1), op1, op2)
      END ;
      MakeChanged(op1, m1) ;
      DoEvaluate
   END
END CodePeepDec ;


(*
   CodeSub - symbols are referenced by Mem[Sym]
*)

PROCEDURE CodeSub (h: CARDINAL) ;
VAR
   op1, op2, op3: CARDINAL ;
   b1, b2, b3   : ARRAY [0..MaxAlias] OF AliasType ;
   i1, i2, i3,
   n1, n2, n3   : CARDINAL ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      op2 := ToCard(w[1]) ;
      op3 := ToCard(w[2]) ;
      IF IsAliasedToZero(op2, OffsetUnlessConst(op2))
      THEN
         MakeNegMethod(op1, OffsetUnlessConst(op1), op3, OffsetUnlessConst(op3)) ;
         DoEvaluate
      END ;
      IF IsAliasedToZero(op3, OffsetUnlessConst(op3))
      THEN
         MakeMoveFromToMethod(op2, OffsetUnlessConst(op2), op1, OffsetUnlessConst(op1)) ;
         DoEvaluate
      END ;
      IF IsAliased(GetCurrentAlias(), op2, OffsetUnlessConst(op2), op1, OffsetUnlessConst(op1))
      THEN
         MakeDecMethod(op1, OffsetUnlessConst(op1), op3, OffsetUnlessConst(op3)) ;
         DoEvaluate
      END ;
      IF IsAliased(GetCurrentAlias(), op3, OffsetUnlessConst(op3), op1, OffsetUnlessConst(op1))
      THEN
         MakeDecMethod(op1, OffsetUnlessConst(op1), op2, OffsetUnlessConst(op2)) ;
         DoEvaluate
      END ;
      IF IsAliased(GetCurrentAlias(), op2, OffsetUnlessConst(op2), op3, OffsetUnlessConst(op3))
      THEN
         MakeMoveFromToMethod(Zero, ImmediateValue, op1, OffsetUnlessConst(op1)) ;
         DoEvaluate
      ELSE
         DoCombinations3(GetCurrentAlias(), op1, op2, op3, MakePeepSubMethod)
      END
   END
END CodeSub ;


(*
   MakePeepSubMethod - 
*)

PROCEDURE MakePeepSubMethod (op1: CARDINAL; m1: ModeOfAddr;
                             op2: CARDINAL; m2: ModeOfAddr;
                             op3: CARDINAL; m3: ModeOfAddr) ;
BEGIN
   PushMethod(peepsub,
              op1, m1, op2, m2, op3, m3, 0, 0)
END MakePeepSubMethod ;


(*
   CodePeepSub - 
*)

PROCEDURE CodePeepSub (h: CARDINAL) ;
VAR
   i, j, n,
   size,
   op1, op2, op3: CARDINAL ;
   m1, m2, m3   : ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      m1 := ToMode(w[1]) ;
      op2 := ToCard(w[2]) ;
      m2 := ToMode(w[3]) ;
      op3 := ToCard(w[4]) ;
      m3 := ToMode(w[5]) ;

      WriteLn ;
      WriteString('( ') ;
      WriteSymbol(op1, m1) ; WriteString(' := ') ;
      WriteSymbol(op2, m2) ; WriteString(' - ') ;
      WriteSymbol(op3, m3) ; WriteString(')') ;
      WriteLn ;

      IF IsConst(op2) AND IsConst(op3)
      THEN
         IF IsConst(op1)
         THEN
            PushValue(op2) ;
            PushValue(op3) ;
            Sub ;
            PopValue(op1) ;
            DoEvaluate
         ELSE
            PushValue(op2) ;
            PushValue(op3) ;
            Sub ;
            MakeMoveFromToMethod(MakeConstLitFromInt(PopInt()), ImmediateValue, op1, m1) ;
            DoEvaluate
         END
      ELSE
         (*
            Not as many choices as add - in fact only two:

            (i)    op2 -> op1 ; dec op1, op3         when op1 is not op3
            (ii)   op2 -> r ; r -= op3 ; r -> op1
         *)

         IF op1#op3
         THEN
            (* (i)   op2 -> op1 ; op1 -= op3 *)
            MakeMoveFromToMethod(op2, m2, op1, m1) ;
            MakeDecMethod(op1, m1, op3, m3) ;
            DoEvaluate
         END ;

         (* (ii)  op2 -> r ; r -= op3 ; r -> op1 *)
         i := 0 ;
         n := NoOfRegisters() ;
         WHILE i<n DO
            IF (op2#op3) OR (GetRegister(i)#op3)
            THEN
               MakeMoveFromToMethod(op2, m2, GetRegister(i), RightValue) ;
               MakeDecMethod(GetRegister(i), RightValue, op3, m3) ;
               MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
               DoEvaluate
            END ;
            INC(i)
         END ;

         (* (iii)  op3 -> r1 ; op2 -> r2 ; r1 -= r2 ; r1 -> op1 *)
         IF IsAliased(GetCurrentAlias(), op2, OffsetUnlessConst(op2), op3, OffsetUnlessConst(op3))
         THEN
            (* (v case i)  op3 -> r ; r += r ; r -> op1 *)
            i := 0 ;
            n := NoOfRegisters() ;
            WHILE i<n DO
               MakeMoveFromToMethod(op3, m3, GetRegister(i), RightValue) ;
               MakeIncMethod(GetRegister(i), RightValue, GetRegister(i), RightValue) ;
               MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
               DoEvaluate ;
               INC(i)
            END
         ELSE
            (* (v)  op3 -> r1 ; op2 -> r2 ; r1 += r2 ; r1 -> op1 *)
            i := 0 ;
            n := NoOfRegisters() ;
            WHILE i<n DO
               j := 0 ;
               WHILE j<n DO
                  IF (j#i) AND (GetRegister(i)#op2) AND (GetRegister(j)#op3)
                  THEN
                     MakeMoveFromToMethod(op3, m3, GetRegister(i), RightValue) ;
                     MakeMoveFromToMethod(op2, m2, GetRegister(j), RightValue) ;
                     MakeIncMethod(GetRegister(i), RightValue, GetRegister(j), RightValue) ;
                     MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
                     DoEvaluate
                  END ;
                  INC(j)
               END ;
               INC(i)
            END
         END
      END
   END
END CodePeepSub ;


(*
   WritePeepSub - 
*)

PROCEDURE WritePeepSub (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' peep- ') ; WriteSymbol(ToCard(w[4]), ToMode(w[5])) ;
      WriteString(' -> ') ; WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WritePeepSub ;


(*
   WriteCodeDec - 
*)

PROCEDURE WriteCodeDec (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' -= -> ') ; WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WriteCodeDec ;


(*
   MakeDecMethod - 
*)

PROCEDURE MakeDecMethod (op1: CARDINAL; m1: ModeOfAddr; op2: CARDINAL; m2: ModeOfAddr) ;
BEGIN
   PushMethod(codedec,
              op1, m1, op2, m2, 0, 0, 0, 0)
END MakeDecMethod ;


(*
   MakeCodeNegMethod - Briefly:
                                Operand1 := - Operand3
                                Checks for Constant folding - otherwise a method is generated.
*)

PROCEDURE MakeCodeNegMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsConst(op1)
   THEN
      PushInt(0) ;
      PushValue(op3) ;
      Sub ;
      PopValue(op1)
   ELSIF IsCoprocessorUsed(op1, op2, op3)
   THEN
      InternalError('not yet implemented', __FILE__, __LINE__) ;
      (* CoprocessorNeg(Operand1, Operand3) *)
   ELSE
      MakeNegMethod(op1, GetMode(op1), op3, GetMode(op3))
   END
END MakeCodeNegMethod ;


(*
   MakeNegMethod - 
*)

PROCEDURE MakeNegMethod (op1: CARDINAL; m1: ModeOfAddr; op2: CARDINAL; m2: ModeOfAddr) ;
BEGIN
   PushMethod(codeneg,
              op1, m1, op2, m2, 0, 0, 0, 0)
END MakeNegMethod ;


(*
   WriteCodeNeg - 
*)

PROCEDURE WriteCodeNeg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ;
      WriteString(' - ') ; WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WriteCodeNeg ;


(*
   CodeNeg -
*)

PROCEDURE CodeNeg (h: CARDINAL) ;
VAR
   al      : ARRAY [0..MaxAlias] OF AliasType ;
   size,
   i, j, n,
   op1, op2: CARDINAL ;
   m1, m2  : ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      m1 := ToMode(w[1]) ;
      op2 := ToCard(w[2]) ;
      m2 := ToMode(w[3]) ;

      IF IsConstLit(op1)
      THEN
         InternalError('should not be negating a constant', __FILE__, __LINE__)
      END ;
      size := GetSize(op1) ;
      n := CollectAliases(GetCurrentAlias(), al, op2, m2) ;
      i := 0 ;
      WHILE i<n DO
         WITH al[i] DO
            IF IsRegister(Sym) AND (Mode=RightValue)
            THEN
               MakeCleanSymbolMethod(Sym, Mode) ;
               MakeInstBinMethod(neg, size, Sym, Mode) ;
               MakeChanged(Sym, Mode) ;
               MakeMoveFromToMethod(Sym, Mode, op1, m1)
            ELSIF IsConst(Sym)
            THEN
               PushInt(0) ;
               PushValue(Sym) ;
               Sub ;
               MakeMoveFromToMethod(MakeConstLitFromInt(PopInt()), ImmediateValue, op1, m1)
            ELSE
               MakeCleanSymbolMethod(Sym, Mode) ;
               MakeInstBinMethod(neg, size, Sym, Mode) ;
               MakeChanged(Sym, Mode) ;
               MakeMoveFromToMethod(Sym, Mode, op1, m1)
            END ;
            DoEvaluate
         END ;
         INC(i)
      END
   END
END CodeNeg ;


(*
   MakeCodeMultMethod - Briefly:
                                Operand1 := Operand2 * Operand3
                                Checks for Constant folding - otherwise a method is generated.
*)

PROCEDURE MakeCodeMultMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsConst(op1)
   THEN
      PushValue(op3) ;
      PushValue(op2) ;
      Mult ;
      PopValue(op1)
   ELSIF IsCoprocessorUsed(op1, op2, op3)
   THEN
      InternalError('not yet implemented', __FILE__, __LINE__) ;
      (* CoprocessorAdd(Operand1, Operand2, Operand3) *)
   ELSE
      PushMethod(codemult,
                 op1, op2, op3, 0, 0, 0, 0, 0)
   END
END MakeCodeMultMethod ;


(*
   CodeMult - symbols are referenced by Mem[Sym]
*)

PROCEDURE CodeMult (h: CARDINAL) ;
VAR
   op1, op2, op3: CARDINAL ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      op2 := ToCard(w[1]) ;
      op3 := ToCard(w[2]) ;
      IF IsAliasedToZero(op2, OffsetUnlessConst(op2))
      THEN
         MakeMoveFromToMethod(Zero, ImmediateValue, op1, OffsetUnlessConst(op1)) ;
         DoEvaluate
      ELSIF IsAliasedToZero(op3, OffsetUnlessConst(op3))
      THEN
         MakeMoveFromToMethod(Zero, ImmediateValue, op1, OffsetUnlessConst(op1)) ;
         DoEvaluate
      ELSIF IsAliasedToOne(op3, OffsetUnlessConst(op3))
      THEN
         MakeMoveFromToMethod(op2, OffsetUnlessConst(op2), op1, OffsetUnlessConst(op1)) ;
         DoEvaluate
      ELSIF IsAliasedToOne(op3, OffsetUnlessConst(op3))
      THEN
         MakeMoveFromToMethod(op2, OffsetUnlessConst(op2), op1, OffsetUnlessConst(op1)) ;
         DoEvaluate
      ELSE
         (* these choices may not be as fast? *)
         IF IsAliasedToTwo(op2, OffsetUnlessConst(op2))
         THEN
            (* try adding rather than multiplying *)
            MakeCodeAddMethod(AddOp, op1, op3, op3) ;
            DoEvaluate
         END ;
         IF IsAliasedToTwo(op3, OffsetUnlessConst(op3))
         THEN
            (* try adding rather than multiplying *)
            MakeCodeAddMethod(AddOp, op1, op2, op2) ;
            DoEvaluate
         END ;
         (* finally try brute force *)
         DoCombinations3(GetCurrentAlias(), op1, op2, op3, MakePeepMultMethod)
      END
   END
END CodeMult ;


(*
   WriteCodeMult - 
*)

PROCEDURE WriteCodeMult (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[1]), RightValue) ;
      WriteString(' * ') ; WriteSymbol(ToCard(w[2]), RightValue) ;
      WriteString(' -> ') ; WriteSymbol(ToCard(w[0]), RightValue) ;
      WriteString(' )')
   END
END WriteCodeMult ;


(*
   MakePeepMultMethod - 
*)

PROCEDURE MakePeepMultMethod (op1: CARDINAL; m1: ModeOfAddr;
                              op2: CARDINAL; m2: ModeOfAddr;
                              op3: CARDINAL; m3: ModeOfAddr) ;
BEGIN
   PushMethod(peepmult,
              op1, m1, op2, m2, op3, m3, 0, 0)
END MakePeepMultMethod ;


(*
   CodePeepMult - 
*)

PROCEDURE CodePeepMult (h: CARDINAL) ;
VAR
   i, j, n,
   size,
   op1, op2, op3: CARDINAL ;
   m1, m2, m3   : ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      m1 := ToMode(w[1]) ;
      op2 := ToCard(w[2]) ;
      m2 := ToMode(w[3]) ;
      op3 := ToCard(w[4]) ;
      m3 := ToMode(w[5]) ;

(*
      WriteLn ;
      WriteString('( ') ;
      WriteSymbol(op1, m1) ; WriteString(' := ') ;
      WriteSymbol(op2, m2) ; WriteString(' * ') ;
      WriteSymbol(op3, m3) ; WriteString(')') ;
      WriteLn ;
*)
      IF IsConst(op2) AND IsConst(op3)
      THEN
         IF IsConst(op1)
         THEN
            PushValue(op2) ;
            PushValue(op3) ;
            Mult ;
            PopValue(op1) ;
            DoEvaluate
         ELSE
            PushValue(op2) ;
            PushValue(op3) ;
            Mult ;
            MakeMoveFromToMethod(MakeConstLitFromInt(PopInt()), ImmediateValue, op1, m1) ;
            DoEvaluate
         END
      ELSE
         IF IsConst(op2)
         THEN
            SwapSymbols(op2, m2, op3, m3)
         END ;
         size := GetSize(op2) ;   (* op2 is not a Const *)

         (* (i) op3 -> r ; r *= op2 ; r -> op1 *)
         i := 0 ;
         n := NoOfRegisters() ;
         WHILE i<n DO
            IF (op3=op2) OR (GetRegister(i)#op2)
            THEN
               MakeMoveFromToMethod(op3, m3, GetRegister(i), RightValue) ;
               MakeCleanSymbolMethod(GetRegister(i), RightValue) ;
               MakeCleanSymbolMethod(op2, m2) ;
               IF IsConst(op2)
               THEN
                  (* mult r, #op2 *)
                  MakeInstRegConstantMethod(mult, size, GetRegister(i), op2)
               ELSE
                  MakeInstRegMDMethod(mult, size, GetRegister(i), GetOffset(op2), 0, GetScope(op2), op2)
               END ;
               MakeChanged(GetRegister(i), RightValue) ;
               MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
               DoEvaluate
            END ;
            INC(i)
         END ;
         (* (ii)  op2 -> r ; r *= op3 ; r -> op1 *)
         i := 0 ;
         n := NoOfRegisters() ;
         WHILE i<n DO
            IF (op3=op2) OR (GetRegister(i)#op2)
            THEN
               MakeMoveFromToMethod(op2, m2, GetRegister(i), RightValue) ;
               MakeCleanSymbolMethod(GetRegister(i), RightValue) ;
               MakeCleanSymbolMethod(op3, m3) ;
               IF IsConst(op3)
               THEN
                  (* mult r, #op3 *)
                  MakeInstRegConstantMethod(mult, size, GetRegister(i), op3)
               ELSE
                  MakeInstRegMDMethod(mult, size, GetRegister(i), GetOffset(op3), 0, GetScope(op3), op3)
               END ;
               MakeChanged(GetRegister(i), RightValue) ;
               MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
               DoEvaluate
            END ;
            INC(i)
         END ;
         (* (iii)  op3 -> r1 ; op2 -> r2 ; r1 *= r2 ; r1 -> op1 *)
         i := 0 ;
         n := NoOfRegisters() ;
         WHILE i<n DO
            j := 0 ;
            WHILE j<n DO
               IF ((op1=op2) AND (m2=m3)) OR           (* op1 and op2 are the same OR *)
                  ((j#i) AND (GetRegister(i)#op2) AND (GetRegister(j)#op3))  (* different regs *)
               THEN
                  MakeMoveFromToMethod(op3, m3, GetRegister(i), RightValue) ;
                  MakeMoveFromToMethod(op2, m2, GetRegister(j), RightValue) ;
                  MakeCleanSymbolMethod(GetRegister(i), RightValue) ;
                  MakeCleanSymbolMethod(GetRegister(j), RightValue) ;
                  MakeInstRegRegMethod(mult, size, GetRegister(i), GetRegister(j)) ;
                  MakeChanged(GetRegister(i), RightValue) ;
                  MakeMoveFromToMethod(GetRegister(i), RightValue, op1, m1) ;
                  DoEvaluate
               END ;
               INC(j)
            END ;
            INC(i)
         END
      END
   END
END CodePeepMult ;


(*
   WritePeepMult - 
*)

PROCEDURE WritePeepMult (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' peep* ') ; WriteSymbol(ToCard(w[4]), ToMode(w[5])) ;
      WriteString(' -> ') ; WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WritePeepMult ;


(*
   MakeReturnMethod - fills in a return method. (Return does nothing on the i486)
*)

PROCEDURE MakeReturnMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   (* deliberately blank *)
END MakeReturnMethod ;


(*
   CodeRet - 
*)

PROCEDURE CodeRet (h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
BEGIN
   InternalError('return does nothing on the *86 processor', __FILE__, __LINE__) ;
   RETURN( 0 )
END CodeRet ;


(*
   WriteCodeRet - 
*)

PROCEDURE WriteCodeRet (p: ProcedureList; m: CARDINAL) ;
BEGIN
   InternalError('return does nothing on the *86 processor', __FILE__, __LINE__)
END WriteCodeRet ;


(*
   MakeCodeBaseMethod - 
*)

PROCEDURE MakeCodeBaseMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   InternalError('base operator should have been optimized away in M2Optimize', __FILE__, __LINE__)
END MakeCodeBaseMethod ;


(*
   MakeChanged - indicates that the symbol, [sym:mode] has changed and
                 therefore we must remove all previous aliases to this symbol.
                 This procedure should be called when we attempto to evauluate
                 binary operations (eg. inc, dec).
*)

PROCEDURE MakeChanged (sym: CARDINAL; mode: ModeOfAddr) ;
BEGIN
   PushMethod(makechanged,
              sym, mode, 0, 0, 0, 0, 0, 0)
END MakeChanged ;


(*
   CodeMakeClean - makes a dirty symbol clean by moving one alias into it.
*)

PROCEDURE CodeMakeClean (h: CARDINAL) ;
VAR
   al   : ARRAY [0..MaxAlias] OF AliasType ;
   i, n : CARDINAL ;
   sym  : CARDINAL ;
   mode : ModeOfAddr ;
   found: BOOLEAN ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      sym  := ToCard(w[0]) ;
      mode := ToMode(w[1]) ;
      IF IsDirty(GetCurrentAlias(), sym, mode)
      THEN
         n := CollectAliases(GetCurrentAlias(), al, sym, mode) ;
         IF n<=1
         THEN
            WriteLn ;
            WriteString('hmm no aliases for symbol ') ; WriteSymbol(sym, mode) ;
            WriteString('alias dump') ; WriteLn ;
            DumpAliases(GetCurrentAlias()) ;
            RETURN
         END ;
         i := 0 ;
         found := FALSE ;
         WHILE i<n DO
            (* try all aliases and take the cheapest *)
            IF (al[i].Sym#sym) OR (al[i].Mode#mode)
            THEN
               (* ok not moving (a -> a) .. *)
               IF IsClean(GetCurrentAlias(), al[i].Sym, al[i].Mode)
               THEN
                  MakeSymToSymMethod(al[i].Sym, al[i].Mode, sym, mode) ;
                  MakeWillBeCleanMethod(sym, mode) ;
                  DoEvaluate ;
                  found := TRUE
               END
            END ;
            INC(i)
         END ;
         IF NOT found
         THEN
            WriteLn ;
            WriteString('aborting this line - as cannot clean symbol: ') ; WriteSymbol(sym, mode) ;
            WriteLn ;
            DebugMethods ;
            DumpAliases(GetCurrentAlias()) ;
            RETURN ;
            IF DebugOn
            THEN
               WriteLn ;
               WriteString('aborting this line - as cannot clean symbol: ') ; WriteSymbol(sym, mode)
            END ;
            (* DebugMethods ; *)
            (* InternalError('cannot find make a symbol clean', __FILE__, __LINE__) ; *)
            (* at a pinch you could just return MaxCost at this point but I believe this should never occur - no *)
         END
      ELSE
         DoEvaluate
      END
   END
END CodeMakeClean ;


(*
   CodeMakeChanged - notes that a symbol has changed and therefore removes any aliases to it.
*)

PROCEDURE CodeMakeChanged (h: CARDINAL) ;
VAR
   sym : CARDINAL ;
   mode: ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      sym  := ToCard(w[0]) ;
      mode := ToMode(w[1]) ;
      RemoveAliasTo(GetCurrentAlias(), sym, mode) ;
      MakeWillBeCleanMethod(sym, mode) ;
      DoEvaluate
   END
END CodeMakeChanged ;


(*
   CodeSymToSym - should only be called via DoNextProcedure
*)

PROCEDURE CodeSymToSym (h: CARDINAL) ;
VAR
   from, to        : CARDINAL ;
   frommode, tomode: ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      frommode := ToMode(w[1]) ;
      to       := ToCard(w[2]) ;
      tomode   := ToMode(w[3]) ;
      MakeAssign(from, frommode, to, tomode) ;
      IF FitsIntoRegister(from) AND FitsIntoRegister(to)
      THEN
         SymToSymMachineWord(h)
      ELSE
         InternalError('SymToSymBlock not written yet', __FILE__, __LINE__) ;
         (* SymToSymBlock(h) *)
      END
   END
END CodeSymToSym ;


(*
   SymToSymMachineWord - 
*)

PROCEDURE SymToSymMachineWord (h: CARDINAL) ;
VAR
   to, from        : CARDINAL ;
   tomode, frommode: ModeOfAddr ;
BEGIN
   (* really does do the move, providing not a clean alias *)
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      frommode := ToMode(w[1]) ;
      to       := ToCard(w[2]) ;
      tomode   := ToMode(w[3]) ;
      IF (from=to) AND (frommode=tomode) OR
         (IsAliased(GetCurrentAlias(), from, frommode, to, tomode) AND IsClean(GetCurrentAlias(), to, tomode))
      THEN
         (* ignore as it is clean aliased *)
         DoEvaluate
      ELSE
         IF IsConst(from)
         THEN
            (* just see whether it is cheaper to move constant directly into memory without using a register *)
            MakeInstMDConstantMethod(mov, GetSize(to), GetOffset(to), 0, GetScope(to), to, from) ;
            DoEvaluate
         END ;
         (* now do the move *)
         MakeForeachRegisterDoSymToSym(from, frommode, to, tomode) ;
         IF DebugOn
         THEN
            WriteString('AFTER FOREACH REGISTER') ; WriteLn ;
            DebugMethods
         END ;
         DoEvaluate ;
         IF DebugOn
         THEN
            WriteString('FINISHED FOREACH REGISTER') ; WriteLn ;
            DebugMethods
         END
      END
   END
END SymToSymMachineWord ;


(*
   MakeForeachRegisterDoSymToSym - 
*)

PROCEDURE MakeForeachRegisterDoSymToSym (from: CARDINAL; frommode: ModeOfAddr;
                                         to: CARDINAL; tomode: ModeOfAddr) ;
BEGIN
   PushMethod(forregsym,
              from, frommode, to, tomode, 0, 0, 0, 0)
END MakeForeachRegisterDoSymToSym ;


(*
   CodeForeachRegisterDoSymToSym - 
*)

PROCEDURE CodeForeachRegisterDoSymToSym (h: CARDINAL) ;
VAR
   n, i            : CARDINAL ;
   to, from        : CARDINAL ;
   tomode, frommode: ModeOfAddr ;
   found           : BOOLEAN ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      frommode := ToMode(w[1]) ;
      to       := ToCard(w[2]) ;
      tomode   := ToMode(w[3]) ;
      IF (from=to) AND (frommode=tomode)
      THEN
         (* nothing to do *)
         DoEvaluate
      ELSE
         found := FALSE ;
         n := NoOfRegisters() ;
         i := 0 ;
         WHILE i<n DO
            IF (to=GetRegister(i)) AND (tomode=RightValue)
            THEN
               (* going into a register - let us try this directly *)
               found := TRUE ;
               MakeSymToRegMethod(from, frommode, GetRegister(i)) ;
               DoEvaluate
            END ;
            IF (from=GetRegister(i)) AND (frommode=RightValue)
            THEN
               (* coming from a register - let us try this directly *)
               found := TRUE ;
               MakeRegToSymMethod(GetRegister(i), to, tomode) ;
               DoEvaluate
            END ;
            (* finally try a slower method, which might be useful later on *)
            IF IsClean(GetCurrentAlias(), GetRegister(i), RightValue) OR (NoOfTimesUsed(GetCurrentAlias(), GetRegister(i), RightValue)=0) OR
               ((from=GetRegister(i)) AND (frommode=RightValue)) OR
               ((to=GetRegister(i)) AND (tomode=RightValue))
            THEN
               found := TRUE ;
               MakeMoveThroughRegisterMethod(from, frommode, to, tomode, GetRegister(i)) ;
               DoEvaluate
            END ;
            INC(i)
         END ;
         IF NOT found
         THEN
            (* cannot perform SymToSym *)
            (* BestCost := MaxCost *)
         END
      END
   END
END CodeForeachRegisterDoSymToSym ;


(*
   CodeMoveThroughRegister - 
*)

PROCEDURE CodeMoveThroughRegister (h: CARDINAL) ;
VAR
   to,  from: CARDINAL;
   tomode,
   frommode : ModeOfAddr;
   reg      : CARDINAL ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      frommode := ToMode(w[1]) ;
      to       := ToCard(w[2]) ;
      tomode   := ToMode(w[3]) ;
      reg      := ToCard(w[4]) ;
      IF (from=to) AND (frommode=tomode)
      THEN
         (* nothing to do *)
         DoEvaluate
      ELSE
         IF (reg#from) OR (frommode#RightValue)
         THEN
            MakeSymToRegMethod(from, frommode, reg)   (* (i)  *)
         END ;
         IF (reg#to) OR (tomode#RightValue)
         THEN
            MakeRegToSymMethod(reg, to, tomode)       (* (ii) *)
         END ;
         DoEvaluate
      END
   END
END CodeMoveThroughRegister ;


(*
   MakeMoveThroughRegisterMethod - 
*)

PROCEDURE MakeMoveThroughRegisterMethod (from: CARDINAL; frommode: ModeOfAddr;
                                         to  : CARDINAL; tomode  : ModeOfAddr;
                                         reg : CARDINAL) ;
BEGIN
   PushMethod(moveviareg,
              from, frommode, to, tomode, reg, 0, 0, 0)
END MakeMoveThroughRegisterMethod ;


(*
   CodeSymToReg - 
*)

PROCEDURE CodeSymToReg (h: CARDINAL) ;
VAR
   to,  from: CARDINAL ;
   frommode : ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      frommode := ToMode(w[1]) ;
      to       := ToCard(w[2]) ;
      IF (to=from) AND (frommode=RightValue) OR
         (IsAliased(GetCurrentAlias(), from, frommode, to, RightValue) AND IsClean(GetCurrentAlias(), to, RightValue))
      THEN
         (* ignore as it is clean aliased *)
         DoEvaluate
      ELSIF IsRegister(from)
      THEN
         (* maybe this should be size of symbol? *)
         MakeRegToRegMethod(from, to, WordSize) ;
         MakeAssign(from, frommode, to, RightValue) ;
         DoEvaluate
      ELSE
         MakeExtToRegMethod(from, frommode, to) ;
         MakeAssign(from, frommode, to, RightValue) ;
         DoEvaluate
      END
   END
END CodeSymToReg ;


(*
   CodeExtToReg - places a symbol Sym into a spare register.
                  If the Sym is Value or Offset then reg contains
                  Mem[Sym]
                  CASE Mode(Sym) OF

                  Const:  Reg := Constant(Sym) |
                  Value,
                  Offset: Reg := Mem[Sym]

                  END
*)

PROCEDURE CodeExtToReg (h: CARDINAL) ;
VAR
   Size       : CARDINAL ;
   toreg, from: CARDINAL;
   frommode   : ModeOfAddr;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      frommode := ToMode(w[1]) ;
      toreg    := ToCard(w[2]) ;
      IF IsConstString(from)
      THEN
         Size := GetStringLength(from) ;
         IF Size=1
         THEN
            MakeInstRegConstantMethod(mov, WordSize, toreg, from) ;
            MakeAssign(from, frommode, toreg, RightValue) ;
            DoEvaluate
         ELSE
            InternalError('cannot place a string with more than 1 character in it into a register', __FILE__, __LINE__)
         END
      ELSIF IsConst(from)
      THEN
         MakeInstRegConstantMethod(mov, WordSize, toreg, from) ;
         MakeAssign(from, frommode, toreg, RightValue) ;
         DoEvaluate
      ELSE
         Size := GetSize(from) ;
         Assert(Size<=WordSize) ;
         MakeInstRegMDMethod(mov, Size, toreg, GetOffset(from), 0, GetScope(from), from) ;
         MakeAssign(from, frommode, toreg, RightValue) ;
         DoEvaluate
      END
   END
END CodeExtToReg ;


(*
   CodeRegToSym - 
*)

PROCEDURE CodeRegToSym (h: CARDINAL) ;
VAR
   from, to: CARDINAL ;
   tomode  : ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      to       := ToCard(w[1]) ;
      tomode   := ToMode(w[2]) ;
      IF ((from=to) AND (tomode=RightValue)) OR
         (IsAliased(GetCurrentAlias(), from, RightValue, to, tomode) AND IsClean(GetCurrentAlias(), to, tomode))
      THEN
         (* ignore as it is clean aliased *)
         DoEvaluate
      ELSIF IsRegister(to)
      THEN
         MakeRegToRegMethod(from, to, WordSize) ;
         MakeAssign(from, RightValue, to, tomode) ;
         DoEvaluate
      ELSE
         MakeRegToExtMethod(from, to, tomode) ;
         MakeAssign(from, RightValue, to, tomode) ;
         DoEvaluate
      END
   END
END CodeRegToSym ;


(*
   CodeRegToExt - places register contents into a symbol Sym.
                  If the Sym is Value or Offset then reg is placed into Mem[Sym]
                  CASE Mode(Sym) OF

                  Const:  Error |
                  Value,
                  Offset: Mem[Sym] := Reg

                  END
*)

PROCEDURE CodeRegToExt (h: CARDINAL) ;
VAR
   Size     : CARDINAL ;
   to,  from: CARDINAL;
   tomode,
   frommode : ModeOfAddr;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      to       := ToCard(w[1]) ;
      tomode   := ToMode(w[2]) ;
      IF IsConst(to)
      THEN
         InternalError('Cannot place a register into a constant', __FILE__, __LINE__)
      ELSE
         Size := GetSize(to) ;
         Assert(Size<=WordSize) ;
         (* Correct since we ignore LeftValue mode and treat as RightValue *)
         MakeInstMDRegMethod(mov, Size, GetOffset(to), 0, GetScope(to), to, from) ;
         MakeAssign(from, RightValue, to, tomode) ;
         DoEvaluate
      END
   END
END CodeRegToExt ;


(*
   CodeRegToReg - 
*)

PROCEDURE CodeRegToReg (h: CARDINAL) ;
VAR
   size            : CARDINAL ;
   from, to        : CARDINAL ;
   tomode, frommode: ModeOfAddr ;
BEGIN
   WITH ArgStack.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      to       := ToCard(w[1]) ;
      size     := ToCard(w[2]) ;
      IF (from=to) OR
         (IsAliased(GetCurrentAlias(), from, RightValue, to, RightValue) AND IsClean(GetCurrentAlias(), to, RightValue))
      THEN
         (* ignore as it is clean aliased *)
         DoEvaluate
      ELSE
         MakeInstRegRegMethod(mov, size, to, from) ;
         MakeAssign(from, RightValue, to, RightValue) ;
         DoEvaluate
      END
   END
END CodeRegToReg ;


(*
   NoOfRegisters - returns the number of registers for this processor.
*)

PROCEDURE NoOfRegisters () : CARDINAL ;
BEGIN
   RETURN( 2 )  (* should be 5 - (2 for testing) *)
END NoOfRegisters ;


(*
   GetRegister - returns the symbol representing register, i.
*)

PROCEDURE GetRegister (i: CARDINAL) : CARDINAL ;
BEGIN
   CASE i OF

   0:  RETURN( EAX ) |
   1:  RETURN( EBX ) |
   2:  RETURN( ECX ) |
   3:  RETURN( EDX ) |
   4:  RETURN( ESI ) |
   5:  RETURN( EDI )

   ELSE
      InternalError('this register does not exist', __FILE__, __LINE__)
   END
END GetRegister ;


(*
   IsRegister - 
*)

PROCEDURE IsRegister (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN(
          (Sym=EAX) OR (Sym=EBX) OR (Sym=ECX) OR (Sym=EDX) OR
          (Sym=ESP) OR (Sym=EBP) OR (Sym=ESI) OR (Sym=EDI)
         )
END IsRegister ;


(*
   DefineConstants - 
*)

PROCEDURE DefineConstants ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
   i: CARDINAL ;
   p: INTEGER ;
   b: BITSET ;
BEGIN
   FOR i := 0 TO WordSize*8 - 1 DO
      b := {} ;
      INCL(b, i) ;
      IntToStr(INTEGER(b), 0, a) ;
      TwoToThePower[i] := MakeConstLit(MakeKey(a)) ;
   END ;
   Zero := MakeConstLit(MakeKey('0')) ;  (* MakeConstLit(MakeKey('0.0')) *)
   One := MakeConstLit(MakeKey('1')) ;  (* MakeConstLit(MakeKey('1.0')) *)
   Two := MakeConstLit(MakeKey('2'))
END DefineConstants ;


(*
   Init - 
*)

PROCEDURE Init ;
VAR
   System: CARDINAL ;
BEGIN
   (* define some useful constants *)
   DefineConstants ;

   (* use SYSTEM module to create our registers *)
   System := MakeDefinitionSource(MakeKey('SYSTEM')) ;
   StartScope(System) ;

   EAX := MakeTemporary(RightValue) ; PutVar(EAX, Word) ;
   EBX := MakeTemporary(RightValue) ; PutVar(EBX, Word) ;
   ECX := MakeTemporary(RightValue) ; PutVar(ECX, Word) ;
   EDX := MakeTemporary(RightValue) ; PutVar(EDX, Word) ;
   ESP := MakeTemporary(RightValue) ; PutVar(ESP, Word) ;
   EBP := MakeTemporary(RightValue) ; PutVar(EBP, Word) ;
   ESI := MakeTemporary(RightValue) ; PutVar(ESI, Word) ;
   EDI := MakeTemporary(RightValue) ; PutVar(EDI, Word) ;

   EndScope

END Init ;


BEGIN
   Init
END M2GenDyn486.
