IMPLEMENTATION MODULE M2GenStatic486 ;


FROM SYSTEM IMPORT ADR ;
FROM ASCII IMPORT nul, cr ;
FROM NameKey IMPORT WriteKey, GetKey, MakeKey ;
FROM M2Configure IMPORT ReturnArgumentOnStack, SystemIsXenix, UseDotForGDBLabels ;
FROM M2AsmUtil IMPORT WriteAsmName, DotifyLabel, GetModuleInitName ;
FROM M2FileName IMPORT CalculateFileName ;
FROM M2System IMPORT IsPseudoSystemFunction, WordSize ;

FROM M2Quads IMPORT QuadOperator, GetQuad, IsReferenced, GetNextQuad,
                    IsProfileOn, IsOptimizeOn, IsCodeOn,
                    QuadToLineNo ;

FROM SymbolTable IMPORT PushOffset, PushSize, PushValue, PopValue,
                        PushVarSize,
                        PushSumOfLocalVarSize,
                        PushSumOfParamSize,
                        MakeConstLit,
                        GetMainModule, GetScopeAuthor,
                        GetVarFather, GetSymName, ModeOfAddr, GetMode,
                        IsModule,
                        IsConstString, GetString, GetStringLength,
                        IsConst, IsProcedure, IsProcType, IsProcedureNested,
                        IsVar, IsVarParam,
                        IsUnbounded, IsArray, IsSet,
                        IsProcedureVariable,
                        IsExportQualified,
                        IsExported,
                        IsSubrange,
                        ForeachExportedDo,
                        ForeachImportedDo,
                        ForeachProcedureDo,
                        ForeachInnerModuleDo,
                        GetType, GetNth, GetNthParam,
                        GetSubrange, NoOfElements,
                        GetFirstUsed,
                        NulSym ;

FROM M2Debug IMPORT Assert ;
FROM M2Lexical IMPORT WriteError, NearToken, NearTokens, InternalError ;
FROM M2Error IMPORT BeginError, EndError ;
FROM M2Base IMPORT MixTypes, ActivationPointer, IsMathType ;

FROM M2ALU IMPORT PopCard, PushCard,
                  PopInt, PushInt,
                  Add, Sub, Mult, Div, Mod,
      	       	  SetOr, SetAnd, SetIn, Bit,
                  SetDifference, SetSymmetricDifference,
                  Equ, NotEqu, Less, Gre, LessEqu, GreEqu ;

FROM StdIO IMPORT Write, PushOutput, PopOutput ;
FROM FIO IMPORT OpenToWrite, Close, WriteChar, IsNoError, File, WriteNBytes ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM NumberIO IMPORT WriteCard, CardToStr, WriteInt, WriteHex ;
FROM M2Base IMPORT Cardinal ;
FROM M2System IMPORT Address, AddressSize ;

FROM M2Options IMPORT Optimizing, GenerateDebugging, Profiling,
                      WordLength ;

FROM M2Inline IMPORT WriteInlineBuffer ;

FROM M2GDB IMPORT LineInfo, ProcedureInfo,
                  VariableInfo, BeginInfo, EndInfo ;

FROM StrLib IMPORT StrLen, StrEqual, StrConCat, StrCopy ;

FROM i486 IMPORT DWORDSIZE, WORDSIZE, BYTESIZE,
                 ModeOfMove, KindOfScope, Register, Instruction,
                 DatumPrefix,
                 GetScope, GetOffset, GetSize, GetUnitPrefix,
                 GetVarTypeSize, EqualSizes, FitsInRegister,
                 MaxDatum ;

FROM M2AS IMPORT CheckProfiling,
                 InstRegConstant, InstConstantReg, InstMDConstant,
                 InstMRIConstant, InstMRICardinal, InstMRIReg,
                 InstMDReg, InstRegMD, InstRegMIO, InstRegMI,
                 InstMIOReg, InstMIReg, InstRegReg, InstRegInt,
                 InstRegHexCard, InstMDOffset, InstMDInt,
                 InstRegImmName, InstName,
                 InstStr,
                 BinInst, BinInstUnit, BinInstReg, BinInstRegI, BinInstConstant,
                 BinInstMD, BinInstHex, Comment,
                 MovePackedStringToIReg, MovePackedStringTo,
                 DeclareLocalVariables,
                 GenerateQuadLabel,
                 GenerateEntryProcedureLabel,
                 GenerateExitProcedureLabel,
                 OpenAssemblyFileAndInitialize,
                 CloseAssemblyFileAndTerminate,
                 WritePrologue, WriteEpilogue ;

FROM M2Gen487 IMPORT IsCoprocessorUsed,
                     CoprocessorMath,
                     CoprocessorIndrX,
                     CoprocessorXIndr,
                     CoprocessorBecomes,
                     CoprocessorConvert,
                     CoprocessorAdd,
                     CoprocessorSub,
                     CoprocessorMult,
                     CoprocessorDiv,
                     CoprocessorMod,
                     CoprocessorNeg,
                     CoprocessorIfEqu,
                     CoprocessorIfNotEqu,
                     CoprocessorIfLess,
                     CoprocessorIfLessEqu,
                     CoprocessorIfGre,
                     CoprocessorIfGreEqu,
                     CoprocessorFunctValue,
                     CoprocessorReturnValue,
                     CoprocessorParam,
                     CoprocessorLineInfo,
                     FlushCoprocessor,
                     CoprocessorInUse,
                     CoprocessorTerminate ;


CONST
   MaxString       = 255 ;
   NearQuadGuess   =  10 ;        (* Guess that a quadruple will be <=127   *)
                                  (* bytes from +-NearQuadGuess quadruples. *)
                                  (* Reduce the value if it is wrong.       *)
   MemcopyThreshold=  30 ;        (* use reps if we move >=30 bytes         *)


VAR
   CurrentProcedure,
   PreviousScope      : CARDINAL ;(* the current procedure being compiled   *)
   CurrentQuad        : CARDINAL ;
   AbsoluteHead       : CARDINAL ;
   LastLine           : CARDINAL ;(* The Last Line number emmitted with the *)
                                  (* generated code.                        *)
   UnboundedOnStack   : BOOLEAN ; (* Determines whether Unbounded Array     *)
                                  (* Data has been placed on the stack,     *)
                                  (* if true then the SP must be reduced    *)
                                  (* after the procedure call - to destroy  *)
                                  (* this data. Old SP is kept in RegC.     *)
   CompilingMainModule: BOOLEAN ; (* Determines whether the main module     *)
                                  (* quadrules are being processed.         *)
   LastOperator       : QuadOperator ; (* The last operator processed.      *)


(*
   Rules for Quadruples
   ====================

   Rules
   =====

   All program declared variables are given the mode, Offset.
   All constants have mode, Immediate.

   Operators
   =========

------------------------------------------------------------------------------
   Array Operators
------------------------------------------------------------------------------
   Sym<I>   Base   a            Delivers a constant result if a is a
                                Global variable. If a is a local variable
                                then the Frame pointer needs to be added.
                                Base yields the effective location in memory
                                of, a, array [0,0, .. ,0] address.
   Sym<I>   ElementSize 1       Always delivers a constant. The number
                                indicates which specified element is chosen.
                                ElementSize is the TypeSize for that element.
   Unbounded  Op1 Op3           Initializes the Op1 StartAddress of the array
                                Op3. Op3 can be a normal array or unbounded array.
                                This operator pushes the value of the array
                                onto the stack and sets Op1 to the
                                start of this array copy.
                                Op1 (is the Unbounded.ArrayAddress) :=
                                ADR(CopyOfOp3)
------------------------------------------------------------------------------
   := Operator
------------------------------------------------------------------------------
   Sym1<I> := Sym2<I>           := produces a constant
   Sym1<O> := Sym2<O>           := has the effect Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O> := Sym2<V>           := has the effect
                                             Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V> := Sym2<O>           := has the effect
                                             Mem[Mem[Sym1<I>]] := Mem[Sym2<I>]
   Sym1<V> := Sym2<V>           := has the effect
                                        Mem[Mem[Sym1<I>]] := Mem[Mem[Sym2<I>]]
------------------------------------------------------------------------------
   Addr Operator  - contains the address of a variable - may need to add
------------------------------------------------------------------------------
   Yields the address of a variable - need to add the frame pointer if
   a variable is local to a procedure.

   Sym1<O>   Addr   Sym2<O>     meaning     Mem[Sym1<I>] := Sym2<I>
   Sym1<V>   Addr   Sym2<O>     meaning     Mem[Sym1<I>] := Sym2<I>
   Sym1<O>   Addr   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Addr   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
------------------------------------------------------------------------------
   Xindr Operator  ( *a = b)
------------------------------------------------------------------------------
   Sym1<O>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
   Sym1<V>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
   Sym1<O>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
------------------------------------------------------------------------------
   IndrX Operator  (a = *b)   where <X> means any value
------------------------------------------------------------------------------
   Sym1<X>   IndrX  Sym2<I>     meaning     Mem[Sym1<I>] := Mem[constant]
   Sym1<X>   IndrX  Sym2<I>     meaning     Mem[Sym1<I>] := Mem[constant]

   Sym1<X>   IndrX  Sym2<X>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<X>   IndrX  Sym2<X>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
------------------------------------------------------------------------------
   + - / * Operators
------------------------------------------------------------------------------
   Sym1<I>   +      Sym2<I>  Sym3<I>  meaning Sym1<I> := Sym2<I> + Sym3<I>
   Sym1<O>   +      Sym2<O>  Sym3<I>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Sym3<I>
   Sym1<O>   +      Sym2<O>  Sym3<O>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
   Sym1<O>   +      Sym2<O>  Sym3<V>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
   Sym1<V>   +      Sym2<O>  Sym3<V>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
   Sym1<V>   +      Sym2<V>  Sym3<V>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
------------------------------------------------------------------------------
   Base Operator
------------------------------------------------------------------------------
   Sym1<O>   Base  Sym2   Sym3<O>     meaning     Mem[Sym1<I>] := Sym3<I>
   Sym1<V>   Base  Sym2   Sym3<O>     meaning     Should Never Occur But If it did..
                                                  Mem[Mem[Sym1<I>]] := Sym3<I>
   Sym1<O>   Base  Sym2   Sym3<V>     meaning     Mem[Sym1<I>] := Mem[Sym3<I>]
   Sym1<V>   Base  Sym2   Sym3<V>     meaning     Should Never Occur But If it did..
                                                  Mem[Mem[Sym1<I>]] := Mem[Sym3<I>]
                   Sym2 is the array type
------------------------------------------------------------------------------

Notes

1)   All references to memory Mem[Sym1<I>] may need to add the base
     pointer of the activation record if the symbol is a procedure variable.

2)   Addr provides the address of variables and therefore can be used to
     perform the pseudo system call function ADR( variable ), provided
     one obeys rule (1).
*)


(*
   Problems/improvements:

   Access registers from SYSTEM.regname
*)

(*
   Register Usage:

   RegRA points to the base of the local Variables on the stack.
   RegRB is used to dereference pointers, arrays and indirect procedures.

   Parameter passing:

   The Call Frame is built,
   All parameters are pushed onto the stack,
   And the procedure is called.

   Build Frame
   Put P0 into Param0 ; return function result
   Put P1 into Param1 ;
   Put P2 into Param2 ;
   Put P3 into Param3 ;
   Put Pn into Param n ;
   Save Volitiles
   Call Procedure
   Restore Volitiles
   Destroy Frame ; Frame -SIZE(P0)  if Procedure is a function
   ; result is now on the stack
*)


VAR
   Coding: BOOLEAN ;

(* %%%FORWARD%%%
        PROCEDURE CodeMath ; FORWARD ;
        PROCEDURE DivMod3AddressUnSigned (Inst: Instruction;
                                          Sym1, Sym2, Sym3: CARDINAL) ; FORWARD ;
        PROCEDURE DivMod3AddressSigned (Inst: Instruction;
                                        Sym1, Sym2, Sym3: CARDINAL) ; FORWARD ;
        PROCEDURE PushConstant (Unit: DatumPrefix; Sym: CARDINAL) ; FORWARD ;
        PROCEDURE GenerateCommentString (a: ARRAY OF CHAR; n: CARDINAL;
                                         VAR b: ARRAY OF CHAR) ; FORWARD ;
        PROCEDURE PushUnbounded (Sym: CARDINAL) ; FORWARD ;
        PROCEDURE PackString (a: ARRAY OF CHAR; DestSize, i: CARDINAL;
                              VAR b: ARRAY OF CHAR) : CARDINAL ; FORWARD ;
        PROCEDURE MoveStringFromTo (StringSym: CARDINAL; DestSym: CARDINAL; DestSize: CARDINAL; m: ModeOfMove) ; FORWARD ;
        PROCEDURE MoveStringFromToDirect (StringSym: CARDINAL; DestSym: CARDINAL; DestSize: CARDINAL) ; FORWARD ;
        PROCEDURE MoveStringFromToIndirect (StringSym: CARDINAL; DestSym: CARDINAL; DestSize: CARDINAL) ; FORWARD ;
        PROCEDURE PushArrayToStack (ArraySym: CARDINAL) ; FORWARD ;
        PROCEDURE PushPackedString (b: ARRAY OF CHAR; Pack: CARDINAL) ; FORWARD ;
        PROCEDURE PushStringToStack (StringSym: CARDINAL) ; FORWARD ;
        PROCEDURE PrepareToPushUnboundedData ; FORWARD ;
        PROCEDURE RegToSym (Reg: Register; Sym: CARDINAL) ; FORWARD ;
        PROCEDURE RemoveAnyUnboundedData ; FORWARD ;
        PROCEDURE InstRegSym (Inst: Instruction; Unit: DatumPrefix;
                              Reg: Register; Sym: CARDINAL) ; FORWARD ;
        PROCEDURE SymToReg (Sym: CARDINAL) : Register ; FORWARD ;
        PROCEDURE InstSymReg (Inst: Instruction; Unit: DatumPrefix;
                              Sym: CARDINAL; Reg: Register) ; FORWARD ;
        PROCEDURE Inst2Address (Inst: Instruction;
                                Sym1, Sym2: CARDINAL) : Register ; FORWARD ;
        PROCEDURE Inst2ModeAddress (Inst: Instruction;
                                    Sym1, Sym2: CARDINAL;
                                    VAR Reg: Register; VAR Unit: DatumPrefix) : Register ; FORWARD ;
        PROCEDURE Inst3Address (Inst: Instruction;
                                Sym1, Sym2, Sym3: CARDINAL) ; FORWARD ;
        PROCEDURE FetchAddress (Sym: CARDINAL; Reg: Register) ; FORWARD ;
        PROCEDURE FetchDatum (Sym         : CARDINAL;
                              Mode        : ModeOfMove;
                              Unit        : DatumPrefix;
                              Scope       : KindOfScope;
                              Offset,
                              GlobalOffset: INTEGER;
                              Reg         : Register) ; FORWARD ;
        PROCEDURE GetFreeRegister ( (* Sym: CARDINAL *) ) : Register ; FORWARD ;
        PROCEDURE MoveFromToDelta (Sym1, Sym2  : CARDINAL;
                                   m1, m2      : ModeOfMove;
                                   o1, o2      : INTEGER;
                                   Unit        : DatumPrefix;
                                   s1, s2      : KindOfScope;
                                   GlobalOffset: INTEGER) ; FORWARD ;
        PROCEDURE GetDatum (Sym         : CARDINAL;
                            Mode        : ModeOfMove;
                            Unit        : DatumPrefix;
                            Scope       : KindOfScope;
                            Offset,
                            GlobalOffset: INTEGER) : Register ; FORWARD ;
        PROCEDURE PutDatum (Sym         : CARDINAL;
                            Mode        : ModeOfMove;
                            Unit        : DatumPrefix;
                            Scope       : KindOfScope;
                            Offset,
                            GlobalOffset: INTEGER;
                            Reg         : Register) ; FORWARD ;
        PROCEDURE PushReg (Unit: DatumPrefix; Reg: Register;
                           Offset, GlobalOffset: INTEGER) ; FORWARD ;
        PROCEDURE GetUnitOfTransfer (Size: CARDINAL) : CARDINAL ; FORWARD ;
        PROCEDURE IncreaseStack (c: CARDINAL) ; FORWARD ;
        PROCEDURE DecreaseStack (c: CARDINAL) ; FORWARD ;
	PROCEDURE CheckLine ; FORWARD ;
        PROCEDURE CodeLogicalOr ; FORWARD ;
        PROCEDURE CodeLogicalAnd ; FORWARD ;
        PROCEDURE CodeSymmetricDifference ; FORWARD ;
	PROCEDURE CodeAdd ; FORWARD ;
	PROCEDURE CodeAddr ; FORWARD ;
	PROCEDURE CodeBase ; FORWARD ;
	PROCEDURE CodeBecomes ; FORWARD ;
	PROCEDURE CodeCall ; FORWARD ;
	PROCEDURE CodeCodeOff ; FORWARD ;
	PROCEDURE CodeCodeOn ; FORWARD ;
	PROCEDURE CodeCoerce ; FORWARD ;
	PROCEDURE CodeConvert ; FORWARD ;
	PROCEDURE CodeCopy ; FORWARD ;
	PROCEDURE CodeDirectCall ; FORWARD ;
	PROCEDURE CodeDiv ; FORWARD ;
	PROCEDURE CodeElementSize ; FORWARD ;
	PROCEDURE CodeEnd ; FORWARD ;
	PROCEDURE CodeEndFile ; FORWARD ;
	PROCEDURE CodeFunctValue ; FORWARD ;
	PROCEDURE CodeGoto ; FORWARD ;
	PROCEDURE CodeHigh ; FORWARD ;
	PROCEDURE CodeIfIn ; FORWARD ;
        PROCEDURE CodeIfNotIn ; FORWARD ;
	PROCEDURE CodeIfEqu ; FORWARD ;
	PROCEDURE CodeIfGre ; FORWARD ;
	PROCEDURE CodeIfGreEqu ; FORWARD ;
	PROCEDURE CodeIfLess ; FORWARD ;
	PROCEDURE CodeIfLessEqu ; FORWARD ;
	PROCEDURE CodeIfNotEqu ; FORWARD ;
	PROCEDURE CodeIndirectCall ; FORWARD ;
	PROCEDURE CodeInline ; FORWARD ;
	PROCEDURE CodeKillLocalVar ; FORWARD ;
	PROCEDURE CodeLineNumber ; FORWARD ;
	PROCEDURE CodeMod ; FORWARD ;
	PROCEDURE CodeMult ; FORWARD ;
	PROCEDURE CodeNegate ; FORWARD ;
	PROCEDURE CodeBit ; FORWARD ;
	PROCEDURE CodeNewLocalVar ; FORWARD ;
	PROCEDURE CodeOffset ; FORWARD ;
	PROCEDURE CodeOptimizeOff ; FORWARD ;
	PROCEDURE CodeOptimizeOn ; FORWARD ;
	PROCEDURE CodeParam ; FORWARD ;
	PROCEDURE CodeProfileOff ; FORWARD ;
	PROCEDURE CodeProfileOn ; FORWARD ;
	PROCEDURE CodeReturn ; FORWARD ;
	PROCEDURE CodeReturnValue ; FORWARD ;
	PROCEDURE CodeStart ; FORWARD ;
	PROCEDURE CodeStartDefFile ; FORWARD ;
	PROCEDURE CodeStartModFile ; FORWARD ;
	PROCEDURE CodeStatement ; FORWARD ;
	PROCEDURE CodeSub ; FORWARD ;
	PROCEDURE CodeUnbounded ; FORWARD ;
        PROCEDURE CodeIndrX ; FORWARD ;
        PROCEDURE CodeXIndr ; FORWARD ;
	PROCEDURE Compare (Inst: Instruction; Sym1, Sym2: CARDINAL) ; FORWARD ;
   %%%FORWARD%%% *)


(*
   GenCode - Codes the list of quadruples described by Head.
             Codes quadruples Start .. End in 80386 assembly language.
*)

PROCEDURE GenCode (Head, Start, End: CARDINAL) ;
VAR
   LastQuad: CARDINAL ;
BEGIN
   AbsoluteHead := Head ;
   CurrentQuad := Start ;
   REPEAT
      LastQuad := CurrentQuad ;
      CodeStatement
   UNTIL LastQuad=End ;
   FlushCoprocessor
END GenCode ;


(*
   Init ;
   InitializationCode ;
   ...
   End
*)

(*
   CodeMain - Codes all the modules quadruples which are not in a procedure.
*)

PROCEDURE CodeMain (Head: CARDINAL) ;
VAR
   Level   : CARDINAL ;
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   AbsoluteHead := Head ;
   CurrentQuad := Head ;
   Level := 0 ;
   WHILE CurrentQuad#0 DO
      GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
      IF Operator=NewLocalVarOp
      THEN
         INC(Level)
      END ;
      IF Level=0
      THEN
         CodeStatement
      ELSE
         CurrentQuad := GetNextQuad(CurrentQuad)
      END ;
      IF Operator=KillLocalVarOp
      THEN
         DEC(Level)
      END
   END ;
   FlushCoprocessor
END CodeMain ;


(*
   CodeProcedures - Codes quadruples which are in a procedure.
*)

PROCEDURE CodeProcedures (Head: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Level   : CARDINAL ;
BEGIN
   AbsoluteHead := Head ;
   CurrentQuad := Head ;
   Level := 0 ;
   WHILE CurrentQuad#0 DO
      GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
      IF Operator=NewLocalVarOp
      THEN
         INC(Level)
      END ;
      IF Level#0
      THEN
         CodeStatement
      ELSE
         IF Operator=StartModFileOp
         THEN
            CodeStartModFile
         ELSIF Operator=StartDefFileOp
         THEN
            CodeStartDefFile
         ELSIF Operator=EndFileOp
         THEN
            CodeEndFile
         END ;
         CurrentQuad := GetNextQuad(CurrentQuad)
      END ;
      IF Operator=KillLocalVarOp
      THEN
         DEC(Level) ;
         FlushCoprocessor
      END
   END
END CodeProcedures ;


(*
   CheckLine - checks the current quadruples line number,
               if the line number has altered from the LastLine
               then a new line comment is emitted.
*)

PROCEDURE CheckLine ;
VAR
   Line: CARDINAL ;
BEGIN
   Line := QuadToLineNo(CurrentQuad) ;
   IF (LastLine#Line) AND CompilingMainModule
   THEN
      LastLine := Line
   END
END CheckLine ;


(*
   GoingToUseCoprocessor - returns TRUE if we are about to use the coprocessor.
*)

PROCEDURE GoingToUseCoprocessor (Operator: QuadOperator; Op1, Op2, Op3: CARDINAL) : BOOLEAN ;
BEGIN
   CASE Operator OF

   AddOp        :  RETURN( IsCoprocessorUsed(Op1, Op2, Op3) ) |
   SubOp        :  RETURN( IsCoprocessorUsed(Op1, Op2, Op3) ) |
   MultOp       :  RETURN( IsCoprocessorUsed(Op1, Op2, Op3) ) |
   DivOp        :  RETURN( IsCoprocessorUsed(Op1, Op2, Op3) ) |
   ModOp        :  RETURN( IsCoprocessorUsed(Op1, Op2, Op3) ) |
   NegateOp     :  RETURN( IsCoprocessorUsed(Op1, Op1, Op3) ) |
   BecomesOp    :  RETURN( IsMathType(GetType(Op1)) OR IsMathType(GetType(Op3)) ) |
   ConvertOp    :  RETURN( IsMathType(Op2) OR IsMathType(GetType(Op3)) ) |
   IfEquOp      :  RETURN( IsMathType(GetType(Op1)) OR IsMathType(GetType(Op2)) ) |
   IfNotEquOp   :  RETURN( IsMathType(GetType(Op1)) OR IsMathType(GetType(Op2)) ) |
   IfLessOp     :  RETURN( IsMathType(GetType(Op1)) OR IsMathType(GetType(Op2)) ) |
   IfLessEquOp  :  RETURN( IsMathType(GetType(Op1)) OR IsMathType(GetType(Op2)) ) |
   IfGreOp      :  RETURN( IsMathType(GetType(Op1)) OR IsMathType(GetType(Op2)) ) |
   IfGreEquOp   :  RETURN( IsMathType(GetType(Op1)) OR IsMathType(GetType(Op2)) ) |
   FunctValueOp :  RETURN( IsMathType(GetType(Op1)) ) |
   ReturnValueOp:  RETURN( IsMathType(GetType(Op3)) ) |
   XIndrOp      :  RETURN( IsMathType(Op2) ) |    (* do not use GetType here as op2 is a type *)
   IndrXOp      :  RETURN( IsMathType(Op2) ) |    (* do not use GetType here as op2 is a type *)
   MathOp       :  RETURN( TRUE ) |
   LineNumberOp :  RETURN( TRUE )

   ELSE
      RETURN( FALSE )
   END
END GoingToUseCoprocessor ;


(*
   CodeStatement - A Multi-Way decision call depending on the current
                   quadruple statement.
*)

PROCEDURE CodeStatement ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   (* WriteString('// Quad') ; WriteCard(CurrentQuad, 6) ; WriteLn ; *)
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF NOT GoingToUseCoprocessor(Operator, Operand1, Operand2, Operand3)
   THEN
      FlushCoprocessor
   END ;
   IF IsReferenced(AbsoluteHead, CurrentQuad) AND (Operator#NewLocalVarOp)
   THEN
      FlushCoprocessor ;
      CheckLine ;
      LastOperator := AddOp ;   (* we must disable the line no opt *)
      GenerateQuadLabel(CurrentQuad)
   ELSE
      CheckLine
   END ;

   CASE Operator OF

   DummyOp            : |
   LogicalOrOp        : CodeLogicalOr |
   LogicalAndOp       : CodeLogicalAnd |
   LogicalXorOp       : CodeSymmetricDifference |
   AddOp              : CodeAdd |
   SubOp              : CodeSub |
   MultOp             : CodeMult |
   DivOp              : CodeDiv |
   ModOp              : CodeMod |
   NegateOp           : CodeNegate |
   BitOp              : CodeBit |
   ConvertOp          : CodeConvert |
   CoerceOp           : CodeCoerce |
   NewLocalVarOp      : CodeNewLocalVar |
   KillLocalVarOp     : CodeKillLocalVar |
   CallOp             : CodeCall |
   ParamOp            : CodeParam |
   AddrOp             : CodeAddr |
   IndrXOp            : CodeIndrX |
   XIndrOp            : CodeXIndr |
   BecomesOp          : CodeBecomes |
   GotoOp             : CodeGoto |
   StartDefFileOp     : CodeStartDefFile |
   StartModFileOp     : CodeStartModFile |
   EndFileOp          : CodeEndFile |
   StartOp            : CodeStart |
   EndOp              : CodeEnd |
   IfInOp             : CodeIfIn |
   IfNotInOp          : CodeIfNotIn |
   IfEquOp            : CodeIfEqu |
   IfNotEquOp         : CodeIfNotEqu |
   IfGreEquOp         : CodeIfGreEqu |
   IfLessEquOp        : CodeIfLessEqu |
   IfLessOp           : CodeIfLess |
   IfGreOp            : CodeIfGre |
   BaseOp             : CodeBase |
   ElementSizeOp      : CodeElementSize |
   OffsetOp           : CodeOffset |
   FunctValueOp       : CodeFunctValue |
   ReturnOp           : CodeReturn |
   ReturnValueOp      : CodeReturnValue |
   UnboundedOp        : CodeUnbounded |
   MathOp             : CodeMath |

   (* And the Compiler Flag Pseudo Quadruples *)
   InlineOp      : CodeInline |
   LineNumberOp  : CodeLineNumber |
   CodeOnOp      : CodeCodeOn |
   CodeOffOp     : CodeCodeOff |
   ProfileOnOp   : CodeProfileOn |
   ProfileOffOp  : CodeProfileOff |
   OptimizeOnOp  : CodeOptimizeOn |
   OptimizeOffOp : CodeOptimizeOff

   ELSE
      BeginError ;
      WriteString('Quad') ; WriteCard(CurrentQuad, 6) ; WriteLn ;
      InternalError('quadruple not implemented yet', __FILE__, __LINE__)
   END ;
   LastOperator := Operator ;
   CurrentQuad := GetNextQuad(CurrentQuad)
END CodeStatement ;


(*
   CodeStartModFile - StartModFileOp is a quadruple which has the following
                      format:

                      StartModFileOp  _  _  ModuleSym

                      Its function is to reset the source file to another
                      file, hence all line numbers emmitted with the
                      generated code will be relative to this source file.
*)

PROCEDURE CodeStartModFile ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   LastLine := 0 ;
   CompilingMainModule := GetMainModule()=Operand3
END CodeStartModFile ;


(*
   CodeStartDefFile - StartDefFileOp is a quadruple which has the following
                      format:

                      StartDefFileOp  _  _  ModuleSym

                      Its function is to reset the source file to another
                      file, hence all line numbers emmitted with the
                      generated code will be relative to this source file.
*)

PROCEDURE CodeStartDefFile ;
BEGIN
   LastLine := 0 ;
   CompilingMainModule := FALSE
END CodeStartDefFile ;


(*
   CodeEndFile - FileOp is a quadruple which has the following format:

                 EndFileOp  _  _  ModuleSym

                 Its function is to reset the source file to another
                 file, hence all line numbers emmitted with the
                 generated code will be relative to this source file.
*)

PROCEDURE CodeEndFile ;
BEGIN
   (* CompilingMainModule := FALSE *)
END CodeEndFile ;


(*
   CodeInline - InlineOp is a quadruple which has the following format:

                InlineOp   NulSym  NulSym  Buf

                The Inline Buffer, Buf, is written to standard output.
*)

PROCEDURE CodeInline ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   (*
      Must Flush the out standing instructions in the code buffer
      otherwise the Inline code will be emmitted in the incorrect
      order.
   *)

   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   WriteInlineBuffer(Operand3)
END CodeInline ;


(*
   CodeLineNumber - LineNumberOp is a quadruple which has the following format:

                    LineNumberOp   NulSym  NulSym  LineNo

                    The line number, LineNo, is ommited if we are producing
                    runtime debugging information.
*)

PROCEDURE CodeLineNumber ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF LastOperator#LineNumberOp
   THEN
      (*
         Must Flush the out standing instructions in the code buffer
         otherwise the LineNumber code will be emitted in the incorrect
         order.
      *)
      IF CoprocessorInUse()
      THEN
         CoprocessorLineInfo(Operand3) (* place line into in middle of *)
                                       (* the coprocessor instructions *)
      ELSE
         LineInfo(Operand3)            (* Generate GDB information *)
      END
   END
END CodeLineNumber ;


(*
   CodeCodeOn - CodeOnOp is a quadruple which switches the compiler
                Code flag to On.
*)

PROCEDURE CodeCodeOn ;
BEGIN
END CodeCodeOn ;


(*
   CodeCodeOff - CodeOffOp is a quadruple which switches the compiler
                 Code flag to Off.
*)

PROCEDURE CodeCodeOff ;
BEGIN
END CodeCodeOff ;


(*
   CodeProfileOn - ProfileOnOp is a quadruple which switches the compiler
                   Profile flag to On.
*)

PROCEDURE CodeProfileOn ;
BEGIN
END CodeProfileOn ;


(*
   CodeProfileOff - CodeProfileOp is a quadruple which switches the compiler
                    Profile flag to Off.
*)

PROCEDURE CodeProfileOff ;
BEGIN
END CodeProfileOff ;


(*
   CodeOptimizeOn - OptimizeOnOp is a quadruple which switches the compiler
                    Optimize flag to On.
*)

PROCEDURE CodeOptimizeOn ;
BEGIN
END CodeOptimizeOn ;


(*
   CodeOptimizeOff - CodeOptimizeOp is a quadruple which switches the compiler
                     Optimize flag to Off.
*)

PROCEDURE CodeOptimizeOff ;
BEGIN
END CodeOptimizeOff ;


(*
   IncReg - increments a register, Reg, by, o.
*)

PROCEDURE IncReg (Reg: Register; o: INTEGER) ;
BEGIN
   IF o>0
   THEN
      InstRegInt(add, MaxDatum(), Reg, o)
   ELSIF o<0
   THEN
      InstRegInt(sub, MaxDatum(), Reg, -o)
   END
END IncReg ;


(*
   BlockMove - utilize the reps trying to make use of the long word moves.
*)

PROCEDURE BlockMove (Size: CARDINAL) ;
BEGIN
   IF Size>WordSize
   THEN
      InstRegInt(mov, MaxDatum(), RegECX, INTEGER(Size DIV WordSize)) ;
      BinInstUnit(repmovs, MaxDatum(), Size DIV WordSize) ;
      Size := Size MOD WordSize
   END ;
   IF Size>0
   THEN
      InstRegInt(mov, MaxDatum(), RegECX, INTEGER(Size)) ;
      BinInstUnit(repmovs, PByte, Size)
   END
END BlockMove ;


(*
   Memcopy - moves a datum from storage element1 to storage element2.
             Sym1 and Sym2 are the symbols of the elements to be moved.
             The mode of the move, whether register, stack, memory,
             is determined by m1 and m2.
             o1 and o2 determine the fixed offset.
             Size determines the total size of datum transfer.
             s1 and s2 determine whether the elements are Local or Global
             variables. This field should be ignored if moving a constant,
             or to/from the stack.
*)

PROCEDURE Memcopy (Sym1, Sym2: CARDINAL;
                   m1, m2    : ModeOfMove;
                   o1, o2    : INTEGER;
                   s1, s2    : KindOfScope;
                   Size      : CARDINAL) ;
BEGIN
   IF (m1=StackDirect) AND (m2=StackDirect)
   THEN
      InternalError('cannot get and put items onto the stack - this wont work', __FILE__, __LINE__)
   END ;
   IF m1=StackDirect
   THEN
      InstRegReg(mov, MaxDatum(), RegESI, RegESP)
   ELSIF m1=MemoryIndirect
   THEN
      FetchAddress(Sym1, RegEAX) ;
      InstRegMIO(mov, MaxDatum(), RegESI, RegEAX, 0)
   ELSE
      FetchAddress(Sym1, RegESI)
   END ;
   IF m2=StackDirect
   THEN
      InstRegInt(sub, MaxDatum(), RegESP, INTEGER(Size)) ;
      InstRegReg(mov, MaxDatum(), RegEDI, RegESP)
   ELSIF m2=MemoryIndirect
   THEN
      FetchAddress(Sym2, RegEAX) ;
      InstRegMIO(mov, MaxDatum(), RegEDI, RegEAX, 0)
   ELSE
      FetchAddress(Sym2, RegEDI)
   END ;
   (*
      rep movs          ;  while ecx>0 do
                        ;     mem[edi] := mem[esi]
                        ;     inc(edi)
                        ;     inc(esi)
                        ;     dec(ecx)
                        ;  end
   *)
   BlockMove(Size)
END Memcopy ;


(*
   MoveFromTo - moves a datum from storage element1 to storage element2.
                Sym1 and Sym2 are the symbols of the elements to be moved.
                The mode of the move, whether register, stack, memory,
                is determined by m1 and m2.
                o1 and o2 determine the fixed offset.
                Size determines the total size of datum transfer.
                s1 and s2 determine whether the elements are Local or Global
                variables. This field should be ignored if moving a constant,
                or to/from the stack.
*)

PROCEDURE MoveFromTo (Sym1, Sym2: CARDINAL;
                      m1, m2    : ModeOfMove;
                      o1, o2    : INTEGER;
                      Size      : CARDINAL;
                      s1, s2    : KindOfScope) ;
VAR
   i   : INTEGER ;
   Unit: CARDINAL ;
BEGIN
   IF (m1=StackDirect) AND (m2=StackDirect)
   THEN
      InternalError('cannot get and put items onto the stack - this wont work', __FILE__, __LINE__)
   END ;
   IF Size>=MemcopyThreshold
   THEN
      Memcopy(Sym1, Sym2, m1, m2, o1, o2, s1, s2, Size)
   ELSE
      i := Size ;
      WHILE i>0 DO
         Unit := GetUnitOfTransfer(CARDINAL(i)) ;
         DEC(i, Unit) ;
         MoveFromToDelta(Sym1, Sym2, m1, m2, o1, o2, GetUnitPrefix(Unit),
                         s1, s2, i)
      END
   END
END MoveFromTo ;


(*
   GetMoveMode - returns the move mode of a symbol, Sym.
*)

PROCEDURE GetMoveMode (Sym: CARDINAL) : ModeOfMove ;
VAR
   m: ModeOfAddr ;
BEGIN
   m := GetMode(Sym) ;
   CASE m OF

   NoValue       : NearToken('Should never reach here - why has Sym no ModeOfAddr?', GetFirstUsed(Sym)) |
   LeftValue     : RETURN( MemoryIndirect ) |
   RightValue    : RETURN( MemoryDirect ) |
   ImmediateValue: RETURN( Constant )

   ELSE
      InternalError('mode enumeration has increased - expand case clause', __FILE__, __LINE__)
   END
END GetMoveMode ;


(*
   GetUnitOfTransfer - returns the number of bytes that can be transferred
                       at a time, if Size bytes are to be transferred.
*)

PROCEDURE GetUnitOfTransfer (Size: CARDINAL) : CARDINAL ;
BEGIN
   IF Size=1
   THEN
      RETURN( BYTESIZE )
   ELSIF (Size=2) OR (Size=3)
   THEN
      RETURN( WORDSIZE )
   ELSIF Size>3
   THEN
      RETURN( WordSize )
   ELSE
      InternalError('Cannot transfer 0 bytes of memory!', __FILE__, __LINE__)
   END
END GetUnitOfTransfer ;


(*
   MoveFromToDelta - moves a datum from storage element1 to storage element2.
                     Sym1 and Sym2 are the symbols of the elements to be moved.
                     The mode of the move, whether register, stack, memory,
                     is determined by m1 and m2.
                     o1 and o2 determine the fixed offset.
                     Size determines the Delta size of datum transfer.
                     s1 and s2 determine whether the elements are Local or
                     Global variables. This field should be ignored if
                     moving a constant, or to/from the stack.
                     GlobalOffset contains an offset into element1 and element2
                     which determines the whereabouts of Delta piece of datum!
*)

PROCEDURE MoveFromToDelta (Sym1, Sym2  : CARDINAL;
                           m1, m2      : ModeOfMove;
                           o1, o2      : INTEGER;
                           Unit        : DatumPrefix;
                           s1, s2      : KindOfScope;
                           GlobalOffset: INTEGER) ;
VAR
   RegTemp: Register ;
BEGIN
   IF (m1=Constant) AND (NOT IsProcedure(Sym1))
   THEN
      IF m2=Constant
      THEN
         PushValue(Sym1) ;
         PopValue(Sym2)
      ELSIF m2=MemoryDirect
      THEN
         InstMDConstant(mov, Unit, o2, GlobalOffset, s2, Sym2, Sym1)
      ELSIF m2=MemoryIndirect
      THEN
         InstRegMD(mov, MaxDatum(), RegEDI, o2, GlobalOffset, s2, Sym2) ;
         InstMRIConstant(mov, Unit, 0, 0, RegEDI, Sym1)
      ELSIF m2=StackDirect
      THEN
         PushConstant(Unit, Sym1)
      ELSE
         InternalError('not expecting to reach here - check value of m2', __FILE__, __LINE__)
      END
   ELSE
      (* Data transfer must travel through a register *)
      RegTemp := GetDatum(Sym1, m1, Unit, s1, o1, GlobalOffset) ;
      PutDatum(Sym2, m2, Unit, s2, o2, GlobalOffset, RegTemp)
   END
END MoveFromToDelta ;


(*
   PutDatum - places a Register into a memory location.
              (Mem[Sym] <- Reg)

              Sym          - determines the symbol
                             having its runtime value updated.
              Mode         - the mode of the symbol memory reference.
              Unit         - the size of the datum transfer.
              Scope        - indicates whether Sym is a local or global
                             variable.
              Offset       - offset into the local or global variable space.
              GlobalOffset - offset within the local variable itself (optional).
              Reg          - the source register.
*)

PROCEDURE PutDatum (Sym         : CARDINAL;
                    Mode        : ModeOfMove;
                    Unit        : DatumPrefix;
                    Scope       : KindOfScope;
                    Offset,
                    GlobalOffset: INTEGER;
                    Reg         : Register) ;
BEGIN
   CASE Mode OF

   MemoryDirect  :  InstMDReg(mov, Unit, Offset, GlobalOffset, Reg, Scope, Sym) |
   MemoryIndirect:  InstRegMD(mov, MaxDatum(), RegEDI, Offset, 0, Scope, Sym) ;
                    InstMIOReg(mov, Unit, RegEDI, GlobalOffset, Reg) ;
                    IF Reg=RegEDI
                    THEN
                       InternalError('attempting to index through the value register', __FILE__, __LINE__)
                    END |
   StackDirect   :  PushReg(Unit, Reg, Offset, GlobalOffset) |
   Constant      :  WriteError('cannot put into a constant')

   ELSE
      InternalError('not expected to reach this point', __FILE__, __LINE__)
   END
END PutDatum ;


(*
   PushConstant - pushes a constant to the stack. The 80386 cannot
                  push an 8 bit entity as 8 bits only so we manufacture
                  this instruction.
*)

PROCEDURE PushConstant (Unit: DatumPrefix; Sym: CARDINAL) ;
BEGIN
   BinInstConstant(push, MaxDatum(), Sym)
(* ************************************************
   IF Unit=PDWord
   THEN
      (* no problem - pushl works *)
      BinInstConstant(push, Unit, Sym)
   ELSIF Unit=PByte
   THEN
      (* must manufacture this *)
      InstRegInt(sub, MaxDatum(), RegESP, 1) ;
      IF WordSize=4
      THEN
         (* 80386 and 80486 *)
         InstMRIConstant(mov, Unit, 0, 0, RegESP, Sym)
      ELSE
         InstRegReg(mov, MaxDatum(), RegEDI, RegESP) ;
         (* WriteString('  seg ss') ; WriteLn ; *)
         InstMRIConstant(mov, Unit, 0, 0, RegEDI, Sym)
      END
   ELSIF Unit=PWord
   THEN
      IF WordSize=4
      THEN
         (* 80386 and 80486 *)
         InstRegInt(sub, MaxDatum(), RegESP, 2) ;
         InstMRIConstant(mov, Unit, 0, 0, RegESP, Sym)
      ELSE
         (* no problem as 8088 and 8086 can push 16 bits *)
         BinInstConstant(push, Unit, Sym)
      END
   ELSE
      InternalError('unrecognized Unit', __FILE__, __LINE__)
   END
  ***************************************************** *)
END PushConstant ;


(*
   PushReg - pushes a register to the stack.
             The entire register is pushed, parameter passing has been
             made more efficient (speed) by pushing word quantities.

             **********************************************************
             no longer push 8 bit parameters via method below
             **********************************************************
             It is incredible that you are not allowed to "push %al"
             on the 80*86. So we have to make up for this deficiency
             by:
                 sub  %esp, 1
                 mov  al, (%esp)

             more badness is that the 8088 and 8086 do not allow
                 mov al, (%sp)
             therefore we have to use another register and
                 mov  %sp, %di
                 mov  al, (%di)

             this procedure is an attempt to isolate these brain dead features.
*)

PROCEDURE PushReg (Unit: DatumPrefix; Reg: Register;
                   Offset, GlobalOffset: INTEGER) ;
BEGIN
   BinInstReg(push, MaxDatum(), Reg)

(* *************************************************************
   was 
   *************************************************************

   IF Unit=PByte
   THEN
      (*
         example of a retarded instruction set:
         not allowed to: push al
      *)
      InstRegInt(sub, MaxDatum(), RegESP, 1) ;
      IF WordSize=4
      THEN
         (* 80386 or 80486 *)
         InstMRIReg(mov, Unit, Offset, GlobalOffset, RegESP, Reg)
      ELSE
         IF Reg=RegEDI
         THEN
            InternalError('trying to index through the value register', __FILE__, __LINE__)
         END ;
         InstRegReg(mov, MaxDatum(), RegEDI, RegESP) ;
         (* WriteString('  seg ss') ; WriteLn ; *)
         InstMRIReg(mov, Unit, Offset, GlobalOffset, RegEDI, Reg)
      END
   ELSIF Unit=PWord
   THEN
      IF WordSize=4
      THEN
         (* 80386 and 80486 *)
         InstRegInt(sub, MaxDatum(), RegESP, 2) ;
         InstMRIReg(mov, Unit, 0, 0, RegESP, Reg)
      ELSE
         (* no problem as 8088 and 8086 can push 16 bits *)
         BinInstReg(push, Unit, Reg)
      END
   ELSIF Unit=PDWord
   THEN
      BinInstReg(push, Unit, Reg)
   ELSE
      InternalError('unrecognized Unit', __FILE__, __LINE__)
   END
*)
END PushReg ;


(*
   PushCardinal - pushes a CARDINAL to the stack.
*)

PROCEDURE PushCardinal (c: CARDINAL) ;
BEGIN
   BinInstHex(push, MaxDatum(), c)
END PushCardinal ;


(*
   PushString - pushes a 1..4 byte string to the stack. The 80386 cannot
                push an 8 bit entity as 8 bits only so we manufacture
                this instruction.
*)

PROCEDURE PushString (Pack: CARDINAL; a: ARRAY OF CHAR) ;
VAR
   i,
   Word: CARDINAL ;
   b   : ARRAY [0..MaxString] OF CHAR ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(a)
      THEN
         INC(Word, ORD(a[i]))
      END
   END ;
   GenerateCommentString(a, Pack-1, b) ;
   Comment(b) ;
   PushCardinal(Word)
END PushString ;


(*
   GenerateCommentString - generates a comment string, b, from a string, a,
                           which contains, n, characters.
*)

PROCEDURE GenerateCommentString (a: ARRAY OF CHAR; n: CARDINAL;
                                 VAR b: ARRAY OF CHAR) ;
VAR
   i: CARDINAL ;
   c: ARRAY [0..1] OF CHAR ;
BEGIN
   StrCopy('"', b) ;
   FOR i := 0 TO n DO
      IF a[i]=nul
      THEN
         StrConCat(b, '<nul>', b)
      ELSE
         c[0] := a[i] ;
         c[1] := nul ;
         StrConCat(b, c, b)
      END
   END ;
   StrConCat(b, '"   ( ', b) ;
   FOR i := n TO 0 BY -1 DO
      IF a[i]=nul
      THEN
         StrConCat(b, '<nul>', b)
      ELSE
         c[0] := a[i] ;
         c[1] := nul ;
         StrConCat(b, c, b)
      END
   END ;
   StrConCat(b, ' )', b)
END GenerateCommentString ;


(*
   GetDatum - places a Symbol runtime value into a register.
              (Reg <- Mem[Sym])

              Sym          - determines the symbol
                             having its runtime value updated.
              Mode         - the mode of the symbol memory reference.
              Unit         - the size of the datum transfer.
              Scope        - indicates whether Sym is a local or global
                             variable.
              Offset       - offset into the local or global variable space.
              GlobalOffset - offset within the local variable itself (optional).

              The register used to contain the symbol value is returned.
*)

PROCEDURE GetDatum (Sym         : CARDINAL;
                    Mode        : ModeOfMove;
                    Unit        : DatumPrefix;
                    Scope       : KindOfScope;
                    Offset,
                    GlobalOffset: INTEGER) : Register ;
VAR
   Reg: Register ;
BEGIN
   Reg := GetFreeRegister() ;
   FetchDatum(Sym, Mode, Unit, Scope, Offset, GlobalOffset, Reg) ;
   RETURN( Reg )
END GetDatum ;


(*
   FetchDatum - places a Symbol runtime value into a register.
                (Reg <- Mem[Sym])

                Sym          - determines the symbol
                               having its runtime value updated.
                Mode         - the mode of the symbol memory reference.
                Unit         - the size of the datum transfer.
                Scope        - indicates whether Sym is a local or global
                               variable.
                Offset       - offset into the local or global variable space.
                GlobalOffset - offset within the local variable itself
                               (optional).
                Reg          - the destination register.
*)

PROCEDURE FetchDatum (Sym         : CARDINAL;
                      Mode        : ModeOfMove;
                      Unit        : DatumPrefix;
                      Scope       : KindOfScope;
                      Offset,
                      GlobalOffset: INTEGER;
                      Reg         : Register) ;
BEGIN
   IF IsProcedure(Sym)
   THEN
      FetchAddress(Sym, Reg)
   ELSE
      CASE Mode OF
   
      MemoryDirect  :  InstRegMD(mov, Unit, Reg, Offset, GlobalOffset, Scope, Sym) |
      MemoryIndirect:  InstRegMD(mov, MaxDatum(), RegEDI, Offset, 0, Scope, Sym) ;
                       InstRegMIO(mov, Unit, Reg, RegEDI, GlobalOffset) ;
                       IF Reg=RegEDI
                       THEN
                          InternalError('Attempting to index through the value register', __FILE__, __LINE__)
                       END |
      StackDirect   :  BinInstReg(pop, Unit, Reg) |
      Constant      :  InstRegConstant(mov, Unit, Reg, Sym)
   
      ELSE
         InternalError('not expected to reach this point', __FILE__, __LINE__)
      END
   END
END FetchDatum ;


(*
   FetchAddress - set Reg with the address of Sym. (Reg <- OFFSET Sym)
                  Sym    - symbol requiring the memory reference.
                  Reg    - indicates the register which will
                           contain the address of Sym.
*)

PROCEDURE FetchAddress (Sym: CARDINAL; Reg: Register) ;
VAR
   Proc  : CARDINAL ;
   Offset: INTEGER ;
BEGIN
   IF GetScope(Sym)=Local
   THEN
      IF WordSize=4
      THEN
         IF GetOffset(Sym)=0
         THEN
            (* so unlikely - it may even be impossible? *)
            InstRegReg(mov, MaxDatum(), Reg, RegEBP)
         ELSE
            InstRegMD(lea, MaxDatum(), Reg, GetOffset(Sym), 0, GetScope(Sym), Sym)
         END
      ELSE
         InstRegReg(mov, MaxDatum(), Reg, RegEBP) ;
         Offset := GetOffset(Sym) ;
         IF Offset>0
         THEN
            InstRegInt(add, MaxDatum(), Reg, Offset)
         ELSIF Offset<0
         THEN
            InstRegInt(sub, MaxDatum(), Reg, -Offset)
         END
      END
   ELSE
      InstRegImmName(mov, MaxDatum(), Reg, Sym)
   END
END FetchAddress ;


(*
   GetFreeRegister - returns a free register to be used with symbol Sym.
                     Need to work on the register allocation schema.
                     For now RegEAX is the only genaral purpose
                     temporary register exercised.
                     Ideally we should have a list of possible candidates
                     together with which symbols they reference.
                     The list must also contain how the registers relate
                     to symbols, whether address or value.
*)

PROCEDURE GetFreeRegister ( (* Sym: CARDINAL *) ) : Register ;
BEGIN
   RETURN( RegEAX )
END GetFreeRegister ;


(*
   Inst3Address - takes a 3 address instruction and produces equivalent
                  386 code.
                  Typical cases are add, sub, mult.
                  In all cases the memory is referenced by Mem[Sym]
*)

PROCEDURE Inst3Address (Inst: Instruction;
                        Sym1, Sym2, Sym3: CARDINAL) ;
VAR
   Reg: Register ;
BEGIN
   (*
      format: (by calling Inst2Address)

              mov   reg, mem[sym2]
              inst  reg, mem[sym3]
   *)
   Reg := Inst2Address(Inst, Sym2, Sym3) ;
   (*
      Now produce the following code:

              mov   mem[sym1], reg
   *)
   InstSymReg(mov, GetUnitPrefix(GetSize(Sym1)), Sym1, Reg)
END Inst3Address ;


(*
   Inst2Address - takes a 2 address instruction and produces equivalent
                  386 code.
                  Typical cases are add, sub, mult.
                  The result of the operation is held in a register,
                  this register is returned from this routine.
                  In all cases the memory is referenced by Mem[Sym]
*)

PROCEDURE Inst2Address (Inst: Instruction;
                        Sym1, Sym2: CARDINAL) : Register ;
VAR
   Reg : Register ;
   Unit: DatumPrefix ;
BEGIN
   IF IsConst(Sym1) AND IsConst(Sym2)
   THEN
      Comment('improve Should have folded constants') ;
      Unit := MaxDatum()
   ELSE
      IF IsConst(Sym1)
      THEN
         Unit := GetUnitPrefix(GetSize(Sym2))
      ELSE
         Unit := GetUnitPrefix(GetSize(Sym1))
      END
   END ;
(*
   ELSE
*)
      IF IsConst(Sym1)
      THEN
         (* Reg := Sym1 *)
         Reg := SymToReg(Sym1) ;
         InstRegSym(Inst, Unit, Reg, Sym2) ;
         (* Reg now contains the result value *)
      ELSIF IsConst(Sym2)
      THEN
         (* Reg := Mem[Sym1] *)
         Reg := SymToReg(Sym1) ;
         (* Inst  Reg, Const *)
         InstRegConstant(Inst, Unit, Reg, Sym2) ;
         (* Reg now contains the result value *)
      ELSE
         (* Neither Sym1 or Sym2 are constants *)
         (*
            format:

                    mov   reg, mem[sym1]
                    inst  reg, mem[sym2]
         *)
         (* Reg := Mem[Sym1] *)
         Reg := SymToReg(Sym1) ;
         (* Inst  Reg, Const *)
         InstRegSym(Inst, Unit, Reg, Sym2) ;
         (* Reg now contains the result value *)
      END ;
      RETURN( Reg )
(*
   END ;
*)
   ; InternalError('Fatel error during code generation - terminate', __FILE__, __LINE__)
END Inst2Address ;


(*
   Inst2ModeAddress - takes a 2 address instruction and produces equivalent
                      386 code.
                      It sets the Reg and Unit depending upon the size of
                      Sym1 and Sym2.
                      The result register, if any, is returned.

                      Typical case is cmp.

                      In all cases the memory is referenced by:
                      CASE GetMode(Sym) OF

                      MemoryDirect: Mem[Sym] |
                      MemoryIndirect: Mem[Mem[Sym]] |
                      Constant      : Sym

                      END

      	       	      Note that we do NOT switch the operands Sym1 & Sym2
      	       	      around because their ordering may be important.
      	       	      This is definitely the case for cmp !

      	       	      Example:
      	       	      cmp   Sym1, Sym2
*)

PROCEDURE Inst2ModeAddress (Inst: Instruction;
                            Sym1, Sym2: CARDINAL;
                            VAR Reg: Register; VAR Unit: DatumPrefix) : Register ;
BEGIN
   IF IsConst(Sym1) AND IsConst(Sym2)
   THEN
      Comment('improve Should have folded constants') ;
      Unit := MaxDatum()
   ELSE
      IF IsConst(Sym1)
      THEN
         Unit := GetUnitPrefix(GetVarTypeSize(Sym2))
      ELSE
         Unit := GetUnitPrefix(GetVarTypeSize(Sym1))
      END
   END ;

   IF IsConst(Sym1)
   THEN
      (* Reg := Sym2 *)
      Reg := GetDatum(Sym2, GetMoveMode(Sym2), Unit, GetScope(Sym2),
                      GetOffset(Sym2), 0) ;
      (* RegEDX := Sym1 *)
      InstRegConstant(mov, Unit, RegEDX, Sym1) ;
      (* Inst  Reg, RegEDX *)
      InstRegReg(Inst, Unit, RegEDX, Reg) ;
      RETURN( RegEDX )
   ELSIF IsConst(Sym2)
   THEN
      (* Reg := Sym1 *)
      Reg := GetDatum(Sym1, GetMoveMode(Sym1), Unit, GetScope(Sym1),
                      GetOffset(Sym1), 0) ;
      (* Inst  Reg, Const *)
      InstRegConstant(Inst, Unit, Reg, Sym2) ;
      (* Reg now contains the result value *)
      RETURN( Reg )
   ELSE
      (* Neither Sym1 or Sym2 are constants *)
      (*
          format:

                    reg1 := Sym1
                    edx := Sym2
                    inst  reg1, edx
      *)
      (* Reg := Sym1 *)
      Reg := GetDatum(Sym1, GetMoveMode(Sym1), Unit, GetScope(Sym1),
                      GetOffset(Sym1), 0) ;
      (* RegEDX := Sym2 *)
      FetchDatum(Sym2, GetMoveMode(Sym2), Unit,
                 GetScope(Sym2), GetOffset(Sym2), 0, RegEDX) ;
      InstRegReg(Inst, Unit, Reg, RegEDX) ;
      RETURN( Reg )
   END
END Inst2ModeAddress ;


(*
   DivMod3Address - takes a 3 address DIV, MOD instruction and produces
                    equivalent 386 code.
                    The 386 MOD and div instructions are
                    significantly different from other instructions (ie ADD SUB)
                    and therefore require a separate procedure.
                    In all cases the memory is referenced by Mem[Sym]

                    This procedure decides whether to use signed or unsigned
                    division.
*)

PROCEDURE DivMod3Address (Inst: Instruction;
                          Sym1, Sym2, Sym3: CARDINAL) ;
VAR
   High, Low,
   Mix      : CARDINAL ;
BEGIN
   IF IsConst(Sym2) AND IsConst(Sym3)
   THEN
      InternalError('should have folded constants', __FILE__, __LINE__)
   ELSE
      IF GetType(Sym2)=NulSym
      THEN
         Mix := GetType(Sym3)
      ELSIF GetType(Sym3)=NulSym
      THEN
         Mix := GetType(Sym2)
      ELSE
         Mix := MixTypes(GetType(Sym2), GetType(Sym3))
      END ;
      IF (Mix#NulSym) AND IsSubrange(Mix)
      THEN
         GetSubrange(Mix, High, Low) ;
         PushValue(Low) ;
         PushCard(0) ;
         IF Less()
         THEN
            (* signed division required *)
            DivMod3AddressSigned(Inst, Sym1, Sym2, Sym3)
         ELSE
            DivMod3AddressUnSigned(Inst, Sym1, Sym2, Sym3)
         END
      ELSE
         DivMod3AddressUnSigned(Inst, Sym1, Sym2, Sym3)
      END ;
      IF Inst=mod
      THEN
         InstSymReg(mov, GetUnitPrefix(GetSize(Sym1)), Sym1,RegEDX)   (* Modulo result held in edx *)
      ELSE
         InstSymReg(mov, GetUnitPrefix(GetSize(Sym1)), Sym1,RegEAX)   (* Division result held in eax *)
      END
   END
END DivMod3Address ;


(*
   DivMod3AddressSigned - takes a 3 address DIV, MOD instruction and produces
                          equivalent 386 code.
                          The 386 MOD and div instructions are
                          significantly different from other instructions (ie ADD SUB)
                          and therefore require a separate procedure.
                          In all cases the memory is referenced by Mem[Sym]
*)

PROCEDURE DivMod3AddressSigned (Inst: Instruction;
                                Sym1, Sym2, Sym3: CARDINAL) ;
BEGIN
   IF IsConst(Sym2) AND IsConst(Sym3)
   THEN
      InternalError('Should have folded constants', __FILE__, __LINE__)
   ELSE
      IF IsConst(Sym2)
      THEN
         (* mov  eax, Constant *)
         InstRegConstant(mov, MaxDatum(), RegEAX, Sym2) ;
         BinInst(cdq) ;
         BinInstMD(idiv, MaxDatum(), GetOffset(Sym3), 0, GetScope(Sym3), Sym3) ;
         (* eax or edx now contains the value for Sym1 *)
      ELSIF IsConst(Sym3)
      THEN
         (* mov  eax, mem[Sym2] *)
         InstRegConstant(mov, MaxDatum(), RegECX, Sym3) ;
         InstRegSym(mov, GetUnitPrefix(GetSize(Sym2)), RegEAX, Sym2) ;
         BinInst(cdq) ;
         BinInstReg(idiv, MaxDatum(), RegECX) ;
         (* eax or edx now contains the value for Sym1 *)
      ELSE
         (* Neither Sym2 or Sym3 are constants *)
         (*
            format:

                    mov   eax, mem[sym2]
                    cdq
                    inst  mem[sym3]
                    mov   mem[sym1], reg
         *)
         (* mov  eax, mem[Sym2] *)
         InstRegSym(mov, GetUnitPrefix(GetSize(Sym2)), RegEAX, Sym2) ;
         BinInst(cdq) ;
         BinInstMD(idiv, MaxDatum(), GetOffset(Sym3), 0, GetScope(Sym3), Sym3) ;
         (* eax or edx now contains the value for Sym1 *)
      END
   END
END DivMod3AddressSigned ;


(*
   DivMod3AddressUnSigned - takes a 3 address DIV, MOD instruction and produces
                            equivalent 386 code.
                            The 386 MOD and div instructions are
                            significantly different from other instructions (ie ADD SUB)
                            and therefore require a separate procedure.
                            In all cases the memory is referenced by Mem[Sym]
*)

PROCEDURE DivMod3AddressUnSigned (Inst: Instruction;
                                  Sym1, Sym2, Sym3: CARDINAL) ;
BEGIN
   IF IsConst(Sym2) AND IsConst(Sym3)
   THEN
      InternalError('should have folded constants', __FILE__, __LINE__)
   ELSE
      IF IsConst(Sym2)
      THEN
         (* mov  eax, Constant *)
         InstRegConstant(mov, MaxDatum(), RegEAX, Sym2) ;
         InstRegReg(xor, MaxDatum(), RegEDX, RegEDX) ;
         BinInstMD(div, MaxDatum(), GetOffset(Sym3), 0, GetScope(Sym3), Sym3) ;
         (* eax or edx now contains the value for Sym1 *)
      ELSIF IsConst(Sym3)
      THEN
         (* mov  eax, mem[Sym2] *)
         InstRegConstant(mov, MaxDatum(), RegECX, Sym3) ;
         InstRegSym(mov, GetUnitPrefix(GetSize(Sym2)), RegEAX, Sym2) ;
         InstRegReg(xor, MaxDatum(), RegEDX, RegEDX) ;
         BinInstReg(div, MaxDatum(), RegECX) ;
         (* eax or edx now contains the value for Sym1 *)
      ELSE
         (* Neither Sym2 or Sym3 are constants *)
         (*
            format:

                    mov   eax, mem[sym2]
                    xor   edx, edx      // edx = 0
                    inst  mem[sym3]
                    mov   mem[sym1], reg
         *)
         (* mov  eax, mem[Sym2] *)
         InstRegSym(mov, GetUnitPrefix(GetSize(Sym2)), RegEAX, Sym2) ;
         InstRegReg(xor, MaxDatum(), RegEDX, RegEDX) ;
         BinInstMD(div, MaxDatum(), GetOffset(Sym3), 0, GetScope(Sym3), Sym3) ;
         (* eax or edx now contains the value for Sym1 *)
      END
   END
END DivMod3AddressUnSigned ;


(*
   Mult3Address - takes a 3 address MULT instruction and produces
                  equivalent 386 code.
                  The 386 imull instruction is
                  significantly different from other instructions (ie ADD SUB)
                  and therefore require a separate procedure.
                  In all cases the memory is referenced by Mem[Sym]
*)

PROCEDURE Mult3Address (Inst: Instruction;
                        Sym1, Sym2, Sym3: CARDINAL) ;
VAR
   Reg: Register ;
BEGIN
   IF IsConst(Sym2) AND IsConst(Sym3)
   THEN
      Comment('should have folded constants')
   END ;
      IF IsConst(Sym2)
      THEN
         (* mov    eax, Sym3 *)
         (* imul   eax, Sym2 *)
         (* mov    Sym3, eax *)
         Reg := SymToReg(Sym3) ;
         InstRegConstant(Inst, MaxDatum(), Reg, Sym2) ;
         RegToSym(Reg, Sym1)
         (*
            when time prevails, alter to perform the following:

            imull  eax, Sym3, Constant ;  eax := Sym3 * Constant
            MultRegMDConstant(RegEAX, Sym3, Sym2) ;
            RegToSym(RegEAX, Sym1)
         *)
      ELSIF IsConst(Sym3)
      THEN
         (* mov    eax, Sym3 *)
         (* imul   eax, Sym2 *)
         (* mov    Sym3, eax *)
         Reg := SymToReg(Sym2) ;
         InstRegConstant(Inst, MaxDatum(), Reg, Sym3) ;
         RegToSym(Reg, Sym1)
         (*
            when time prevails, alter to perform the following:

            imull  eax, Sym2, Constant ;  eax := Sym2 * Constant
            MultRegMDConstant(RegEAX, Sym2, Sym3) ;
            RegToSym(RegEAX, Sym1)
         *)
      ELSE
         (* Neither Sym2 or Sym3 are constants *)
         (*
            format:

                    mov   eax, mem[sym2]
                    imul  eax, mem[sym3]
                    mov   mem[sym1], eax
         *)
         (* mov    eax, Sym3 *)
         (* imul   eax, Sym2 *)
         (* mov    Sym3, eax *)
         Reg := SymToReg(Sym2) ;
         InstRegSym(Inst, GetUnitPrefix(GetSize(Sym3)), Reg, Sym3) ;
         RegToSym(Reg, Sym1)
      END
(*
   END
*)
END Mult3Address ;


(*
   SymToReg - places a symbol Sym into a spare register.
              The register is returned.
              Note that if the Sym is a constant then
              the register contains the constant value.
              If the Sym is Value or Offset then reg contains
              Mem[Sym]
              CASE Mode(Sym) OF

              Const:  Reg := Constant(Sym) |
              Value,
              Offset: Reg := Mem[Sym]

              END
*)

PROCEDURE SymToReg (Sym: CARDINAL) : Register ;
VAR
   Father: CARDINAL ;
   i,
   Size,
   Offset: CARDINAL ;
   Reg   : Register ;
BEGIN
   IF IsConstString(Sym)
   THEN
      Size := GetStringLength(Sym) ;
      IF Size=1
      THEN
         (* GetKey(GetString(Sym), Str) ; *)
         Reg := GetFreeRegister() ;
         InstRegConstant(mov, MaxDatum(), Reg, Sym)
      ELSE
         InternalError('cannot place a string with more than 1 character in it into a register', __FILE__, __LINE__)
      END
   ELSIF IsConst(Sym)
   THEN
      Reg := GetFreeRegister() ;
      InstRegConstant(mov, MaxDatum(), Reg, Sym)
   ELSE
      Size := GetSize(Sym) ;
      Assert(Size<=WordSize) ;
      Reg := GetFreeRegister() ;
      InstRegMD(mov, GetUnitPrefix(GetUnitOfTransfer(Size)), Reg, GetOffset(Sym), 0, GetScope(Sym), Sym)
   END ;
   RETURN( Reg )
END SymToReg ;


(*
   RegToSym - places register contents into a symbol Sym.
              If the Sym is Value or Offset then reg is placed into Mem[Sym]
              CASE Mode(Sym) OF

              Const:  Error |
              Value,
              Offset: Mem[Sym] := Reg

              END
*)

PROCEDURE RegToSym (Reg: Register; Sym: CARDINAL) ;
VAR
   Size: CARDINAL ;
BEGIN
   IF IsConst(Sym)
   THEN
      WriteError('cannot place a register into a constant')
   ELSE
      Size := GetSize(Sym) ;
      Assert(Size<=WordSize) ;
      (* Correct since we ignore LeftValue mode and treat as OffsetMode *)
      InstMDReg(mov, GetUnitPrefix(GetUnitOfTransfer(Size)), GetOffset(Sym), 0, Reg, GetScope(Sym), Sym)
   END
END RegToSym ;


(*
   IntToSym - places an integer into a symbol Sym.
              If the Sym is Value or Offset then reg is placed into Mem[Sym]
              CASE Mode(Sym) OF

              Const:  Error |
              Value,
              Offset: Mem[Sym] := Int

              END
*)

PROCEDURE IntToSym (Int: INTEGER; Sym: CARDINAL) ;
VAR
   Size: CARDINAL ;
BEGIN
   IF IsConst(Sym)
   THEN
      NearToken('should never need to place integer into a constant', GetFirstUsed(Sym))
   ELSE
      Size := GetSize(Sym) ;
      Assert(Size<=WordSize) ;
      (* Correct since we ignore LeftValue mode and treat as OffsetMode *)
      InstMDInt(mov, GetUnitPrefix(GetUnitOfTransfer(Size)), GetOffset(Sym), 0, Int, GetScope(Sym), Sym)
   END
END IntToSym ;


(*
   GetAdr - returns the address of symbol, Sym, in a free register.
*)

PROCEDURE GetAdr (Sym: CARDINAL) : Register ;
VAR
   Reg: Register ;
BEGIN
   IF IsProcedureVariable(Sym)
   THEN
      IF GetMode(Sym)=RightValue
      THEN
         Reg := GetFreeRegister() ;
         InstRegReg(mov, MaxDatum(), Reg, RegEBP) ;
         IF GetOffset(Sym)<0
         THEN
            InstRegInt(sub, MaxDatum(), Reg, -GetOffset(Sym))
         ELSE
            InstRegInt(add, MaxDatum(), Reg, GetOffset(Sym))
         END
      ELSE
         Assert(GetMode(Sym)=LeftValue) ;
         Reg := SymToReg(Sym)
      END
   ELSE
      IF GetMode(Sym)=RightValue
      THEN
         Reg := GetFreeRegister() ;
         InstRegImmName(mov, MaxDatum(), Reg, Sym)
      ELSE
         Assert(GetMode(Sym)=LeftValue) ;
         Reg := SymToReg(Sym)
      END
   END ;
   RETURN( Reg )
END GetAdr ;


(*
   InstSymReg - places a register into a symbol Sym.
                The code for Offset and Value cases is the same ie:
                Mem[Sym] <- inst <- Reg
*)

PROCEDURE InstSymReg (Inst: Instruction; Unit: DatumPrefix;
                      Sym: CARDINAL; Reg: Register) ;
BEGIN
   IF IsConst(Sym)
   THEN
      WriteError('cannot move a register to a constant')
   ELSE
      InstMDReg(Inst, Unit, GetOffset(Sym), 0, Reg, GetScope(Sym), Sym)
   END
END InstSymReg ;


(*
   InstRegSym - places a symbol Sym into register Reg.
                The code for Offset and Value cases is the same ie:
                Reg <- inst <- Mem[Sym]
*)

PROCEDURE InstRegSym (Inst: Instruction; Unit: DatumPrefix;
                      Reg: Register; Sym: CARDINAL) ;
BEGIN
   IF IsConst(Sym)
   THEN
      InstRegConstant(Inst, Unit, Reg, Sym)
   ELSE
      InstRegMD(Inst, Unit, Reg, GetOffset(Sym), 0, GetScope(Sym), Sym)
   END
END InstRegSym ;


(*
   CodeAdd - Briefly:
                      Operand1 := Operand2 + Operand3
             Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeAdd ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Reg     : Register ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      Add ;
      PopValue(Operand1)
   ELSIF IsCoprocessorUsed(Operand1, Operand2, Operand3)
   THEN
      CoprocessorAdd(Operand1, Operand2, Operand3) ;
   ELSE
      Inst3Address(add, Operand1, Operand2, Operand3)
   END
END CodeAdd ;


(*
   CodeLogicalOr - Briefly:
                            Operand1 := Operand2 LOR Operand3
                   Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeLogicalOr ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Reg     : Register ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      SetOr ;
      PopValue(Operand1)
   ELSE
      Inst3Address(or, Operand1, Operand2, Operand3)
   END
END CodeLogicalOr ;


(*
   CodeSub - Briefly:
                      Operand1 := Operand2 - Operand3
             Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeSub ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   i       : INTEGER ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand2) ;
      PushValue(Operand3) ;
      Sub ;
      PopValue(Operand1)
   ELSIF IsCoprocessorUsed(Operand1, Operand2, Operand3)
   THEN
      CoprocessorSub(Operand1, Operand2, Operand3)
   ELSE
      Inst3Address(sub, Operand1, Operand2, Operand3)
   END
END CodeSub ;


(*
   CodeMult - Briefly:
                       Operand1 := Operand2 * Operand3
              Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeMult ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Type    : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   Type := GetType(Operand1) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      IF (Type#NulSym) AND IsSet(Type)
      THEN
      	 SetAnd
      ELSE
      	 Mult
      END ;
      PopValue(Operand1)
   ELSE
      IF (Type#NulSym) AND IsSet(Type)
      THEN
      	 Mult3Address(and, Operand1, Operand2, Operand3)
      ELSIF IsCoprocessorUsed(Operand1, Operand2, Operand3)
      THEN
         CoprocessorMult(Operand1, Operand2, Operand3)
      ELSE
      	 Mult3Address(mult, Operand1, Operand2, Operand3)
      END
   END
END CodeMult ;


(*
   CodeLogicalAnd - Briefly:
                             Operand1 := Operand2 LAND Operand3
                    Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeLogicalAnd ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      SetAnd ;
      PopValue(Operand1)
   ELSE
      Mult3Address(and, Operand1, Operand2, Operand3)
   END
END CodeLogicalAnd ;


(*
   CodeDiv - Briefly:
                      Operand1 := Operand2 / Operand3
             Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeDiv ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      Div ;
      PopValue(Operand1)
   ELSIF IsCoprocessorUsed(Operand1, Operand2, Operand3)
   THEN
      CoprocessorDiv(Operand1, Operand2, Operand3)
   ELSE
      DivMod3Address(idiv, Operand1, Operand2, Operand3)
   END
END CodeDiv ;


(*
   CodeSymmetricDifference - Briefly:
                                      Operand1 := Operand2 / Operand3
                             Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeSymmetricDifference ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      SetSymmetricDifference ;
      PopValue(Operand1)
   ELSE
      Mult3Address(xor, Operand1, Operand2, Operand3)
   END
END CodeSymmetricDifference ;


(*
   CodeMod - Briefly:
                      Operand1 := Operand2 MOD Operand3
             Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeMod ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand2) ;
      PushValue(Operand3) ;
      Mod ;
      PopValue(Operand1)
   ELSIF IsCoprocessorUsed(Operand1, Operand2, Operand3)
   THEN
      CoprocessorMod(Operand1, Operand2, Operand3)
   ELSE
      DivMod3Address(mod, Operand1, Operand2, Operand3)
   END
END CodeMod ;


(*
   CodeNegate - Briefly:
                      Operand1 := - Operand3
                Checks for Constant folding - otherwise 386 code is generated.
*)

PROCEDURE CodeNegate ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   i       : INTEGER ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushCard(0) ;
      PushValue(Operand3) ;
      Sub ;
      PopValue(Operand1)
   ELSIF IsCoprocessorUsed(Operand1, Operand1, Operand3)
   THEN
      CoprocessorNeg(Operand1, Operand3)
   ELSIF IsConst(Operand3)
   THEN
      IF GetSize(GetType(Operand1)) > WordSize
      THEN
         WriteError('cannot negate a variable larger than register size (WordSize bytes)')
      END ;
      PushCard(0) ;
      PushValue(Operand3) ;
      Sub ;
      i := PopInt() ;
      InstMDInt(neg, GetUnitPrefix(GetVarTypeSize(Operand1)),
                GetOffset(Operand1), 0, i, GetScope(Operand1), Operand1)
   ELSE
      IF Operand1=Operand3
      THEN
         (* Peephole Operand1 = Operand3 *)
         BinInstMD(neg, GetUnitPrefix(GetVarTypeSize(Operand1)),
                   GetOffset(Operand1), 0, GetScope(Operand1), Operand1)
      ELSE
         InstRegMD(mov, GetUnitPrefix(GetVarTypeSize(Operand3)),
                   RegEAX, GetOffset(Operand3), 0, GetScope(Operand3),
                   Operand3) ;
         BinInstReg(neg, GetUnitPrefix(GetVarTypeSize(Operand3)), RegEAX) ;
         InstMDReg(mov, GetUnitPrefix(GetVarTypeSize(Operand1)),
                   GetOffset(Operand1), 0, RegEAX, GetScope(Operand1),
                   Operand1)
      END
   END
END CodeNegate ;


(*
   CodeBit - Briefly:
             Operand1 :=  (1 << Operand3)
             Checks for Constant folding - otherwise 386 code is generated.
*)

PROCEDURE CodeBit ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   i       : INTEGER ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      Bit ;
      PopValue(Operand1)
   ELSIF IsConst(Operand3)
   THEN
      IF GetSize(GetType(Operand1)) > WordSize
      THEN
         WriteError('cannot find the bit shift of a variable larger than register size (4 bytes)')
      END ;
      PushValue(Operand3) ;
      Bit ;
      i := PopInt() ;
      InstMDInt(mov, GetUnitPrefix(GetVarTypeSize(Operand1)),
                GetOffset(Operand1), 0, i, GetScope(Operand1), Operand1)
   ELSE
      InstRegMD(mov, GetUnitPrefix(GetVarTypeSize(Operand3)),
                RegECX, GetOffset(Operand3), 0, GetScope(Operand3),
                Operand3) ;
      InstRegConstant(mov, MaxDatum(),
                      RegEAX, MakeConstLit(MakeKey('1'))) ;
      InstRegReg(sal, GetUnitPrefix(GetVarTypeSize(Operand3)),
                 RegEAX, RegECX) ;
      InstMDReg(mov, GetUnitPrefix(GetVarTypeSize(Operand1)),
                GetOffset(Operand1), 0, RegEAX, GetScope(Operand1),
                Operand1)
   END
END CodeBit ;


(*
   CodeMath - codes the inbuilt mathematical operators.
*)

PROCEDURE CodeMath ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   CoprocessorMath(Operand1, Operand2, Operand3)
END CodeMath ;


(*
   CodeConvert - Converts Operand3 to type Operand2 placing the result into
                 Operand1.
                 Convert will, if need be, alter the machine representation
                 of Operand3 to comply with TYPE Operand2.
                 However with the simple APU all base types are the same size,
                 this simple code generator does not implement changing
                 more complex types.
                 This must be added later.
*)

PROCEDURE CodeConvert ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Reg     : Register ;
   s1, s3  : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   s1 := GetVarTypeSize(Operand1) ;
   IF IsMathType(GetType(Operand1)) OR IsMathType(GetType(Operand2)) OR
      IsMathType(GetType(Operand3))
   THEN
      CoprocessorConvert(Operand1, Operand2, Operand3)
   ELSIF IsConst(Operand3)
   THEN
      IF s1>WordSize
      THEN
         WriteError('implementation restriction - can only CONVERT constant types to a type whose size <= 4 bytes')
      END ;
      (* Constants assume any size up to WordSize *)
      MoveFromTo(Operand3, Operand1,
                 GetMoveMode(Operand3), GetMoveMode(Operand1),
                 GetOffset(Operand3), GetOffset(Operand1),
                 GetVarTypeSize(Operand1),
                 GetScope(Operand3), GetScope(Operand1))
   ELSE
      s3 := GetVarTypeSize(Operand3) ;
      IF (s1 <= WordSize) AND (s3 <= WordSize)
      THEN
         (*
         PushAPU(Operand3) ;
         PopAPU(Operand1)
         *)
         (*
            The APU code seems not to take into account the mode, I think
            this is actually a bug and it should take mode into account.
            Should behave like :=
         *)
         Reg := GetDatum(Operand3, GetMoveMode(Operand3),
                         GetUnitPrefix(GetVarTypeSize(Operand3)), GetScope(Operand3),
                         GetOffset(Operand3), 0) ;
         IF s1>s3
         THEN
            IF s3=BYTESIZE
            THEN
               InstRegHexCard(and, MaxDatum(), Reg, 0FFH)
            ELSIF s3=WORDSIZE
            THEN
               InstRegHexCard(and, MaxDatum(), Reg, 0FFFFH)
            ELSE
               WriteError('peculiar type size - check this out!')
            END
         END ;
         PutDatum(Operand1, GetMoveMode(Operand1),
                  GetUnitPrefix(GetVarTypeSize(Operand1)), GetScope(Operand1),
                  GetOffset(Operand1), 0, Reg)
      ELSE
         WriteError('implementation restriction - can only CONVERT types which have sizes <= 4 bytes')
         (* Alter this when time permits *)
      END
   END
END CodeConvert ;


(*
   CodeCoerce - Coerce Operand3 to type Operand2 placing the result into
                Operand1.
                Coerce will NOT alter the machine representation
                of Operand3 to comply with TYPE Operand2.
                Therefore it INSISTS that under all circumstances that the
                type sizes of Operand1 and Operand3 are the same.
                CONVERT will perform machine manipulation to change variable
                types, coerce does no such thing.
*)

PROCEDURE CodeCoerce ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Must not call GetVarTypeSize if sym is a constant.
      Constants can assume any size from BYTESIZE .. WordSize
   *)
   IF IsProcedure(Operand3)
   THEN
      IF GetVarTypeSize(Operand1)=WordSize
      THEN
         MoveFromTo(Operand3, Operand1,
                    GetMoveMode(Operand3), GetMoveMode(Operand1),
                    GetOffset(Operand3), GetOffset(Operand1),
                    GetVarTypeSize(Operand1),
                    GetScope(Operand3), GetScope(Operand1))
      ELSE
         WriteError('procedure address can only be stored in a word size operand')
      END
   ELSIF IsConst(Operand3) OR (GetVarTypeSize(Operand1)=GetVarTypeSize(Operand3))
   THEN
      (*
         it should take mode into account, just like :=
      *)
      MoveFromTo(Operand3, Operand1,
                 GetMoveMode(Operand3), GetMoveMode(Operand1),
                 GetOffset(Operand3), GetOffset(Operand1),
                 GetVarTypeSize(Operand1),
                 GetScope(Operand3), GetScope(Operand1))
   ELSE
      WriteError('TYPE Coersion can only be achieved with types of the same size')
      (* Alter this when time permits *)
   END
END CodeCoerce ;


(*
   Notes on procedure call using Unbounded arrays.

   (i)   The Unbounded data is pushed onto the stack. Unbounded Op1
         is then initialised to contain the address of this data.
         Op1 is effectively (Unbounded.ArrayAddress).
   (ii)  The old sp ie sp' is copied into RegC.
   (iii) RegC and all other registers are preserved when calling the
         required procedure.
   (iv)  After the procedure returns the Unbounded data must be destroyed
         ie SP := RegC
   (v)   This should be performed in different places depending whether
         the procedure has a return argument (function).
         (a)  Functions perform this after FunctValue quadruple.
         (b)  Procedures perform this after the Call quadruple.
*)

(*
   The Call Mechanism:


   Stack of the APU - grows upwards!


   Before Call
   ===========
   .             .
   .             .
   .             .
   |             |   <- sp'
   |-------------|
   | rd''        |
   |-------------|
   | rc''        |
   |-------------|
   | rb''        |
   |-------------|
   | ra''        |
   |-------------|
   |             |
   | Local       |
   |             |
   | Variables   |
   |             |
   |-------------|
   | pc''        |
   |-------------|
   | Parameters  |
   |-------------|
   | Return Arg  |   <- ra'
   |-------------|
   |             |
   .             .
   .             .
   .             .


   Stacking Parameters                    res := proc(a, b, c)
   ===================

   .             .
   .             .
   .             .
   |             |   <- sp
   |-------------|
   | c           |
   |-------------|
   | b           |
   |-------------|
   | a           |
   |-------------|
   | res         |
   |-------------|
   | Unbounded   |
   :             :
   | Array space |   <- sp'     <- rc
   |-------------|
   | rd''        |
   |-------------|
   | rc''        |
   |-------------|
   | rb''        |
   |-------------|
   | ra''        |
   |-------------|
   |             |
   | Local       |
   |             |
   | Variables   |
   |             |
   |-------------|
   | pc''        |
   |-------------|
   | Parameters  |
   |-------------|
   | Return Arg  |   <- ra'
   |-------------|
   |             |
   .             .
   .             .
   .             .


   NewLocalVar
   ===========

   .             .
   .             .
   .             .
   |             |   <- sp
   |-------------|
   |             |
   | Local       |
   |             |
   | Variables   |
   |             |
   |-------------|
   | pc'         |
   |-------------|
   | c           |
   |-------------|
   | b           |
   |-------------|
   | a           |
   |-------------|
   | res         |   <- ra
   |-------------|
   | Unbounded   |
   :             :
   | Array space |   <- sp'     <- rc'
   |-------------|
   | rd''        |
   |-------------|
   | rc''        |
   |-------------|
   | rb''        |
   |-------------|
   | ra''        |
   |-------------|
   |             |
   | Local       |
   |             |
   | Variables   |
   |             |
   |-------------|
   | pc''        |
   |-------------|
   | Parameters  |
   |-------------|
   | Return Arg  |   <- ra'
   |-------------|
   |             |
   .             .
   .             .
   .             .


   KillLocalVar
   ============

   .             .
   .             .
   .             .
   |             |   <- sp
   |-------------|
   | res         |
   |-------------|
   | Unbounded   |
   :             :
   | Array space |   <- sp'     <- rc
   |-------------|
   | rd''        |
   |-------------|
   | rc''        |
   |-------------|
   | rb''        |
   |-------------|
   | ra''        |
   |-------------|
   |             |
   | Local       |
   |             |
   | Variables   |
   |             |
   |-------------|
   | pc''        |
   |-------------|
   | Parameters  |
   |-------------|
   | Return Arg  |   <- ra'
   |-------------|
   |             |
   .             .
   .             .
   .             .


*)



(*
   CodeNewLocalVar - Builds a new frame on the stack to contain the procedure
                     local variables.
*)

PROCEDURE CodeNewLocalVar ;
VAR
   Operator: QuadOperator ;
   Operand1: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, PreviousScope, CurrentProcedure) ;
   ProcedureInfo(CurrentProcedure, Operand1) ;   (* Generate GDB information *)
   GenerateEntryProcedureLabel(CurrentQuad, CurrentProcedure) ;
   LineInfo(Operand1) ;                  (* Generate GDB information *)

   (* Create new frame *)
   BinInstReg(push, MaxDatum(), RegEBP) ;
   InstRegReg(mov, MaxDatum(), RegEBP, RegESP) ;

   IF IsProcedureNested(CurrentProcedure)
   THEN
      BinInstReg(push, MaxDatum(), RegECX)
   END ;

   DeclareLocalVariables(CurrentProcedure) ;

   (* Make local variable space *)

   PushSumOfLocalVarSize(CurrentProcedure) ;
   DecreaseStack(PopCard()) ;

   CheckProfiling ;

   (* Push Registers *)

   BinInstReg(push, MaxDatum(), RegEBX) ;
   BinInstReg(push, MaxDatum(), RegEDI) ;
   BinInstReg(push, MaxDatum(), RegESI) ;

   (* Generate debugging trace infomation *)

   BeginInfo

END CodeNewLocalVar ;


(*
   CodeKillLocalVar - Resets RA to the old value before the CodeNewLocalVar
                      and kills off local variables.
*)

PROCEDURE CodeKillLocalVar ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, CurrentProcedure) ;

   (* Generate debugging trace infomation *)

   EndInfo ;

   (* Pop Registers *)

   BinInstReg(pop, MaxDatum(), RegESI) ;
   BinInstReg(pop, MaxDatum(), RegEDI) ;
   BinInstReg(pop, MaxDatum(), RegEBX) ;

   (* Kill local variable space *)

   PushSumOfLocalVarSize(CurrentProcedure) ;
   IncreaseStack(PopCard()) ;

   (* Kill activation record - (call frame) *)
   IF IsProcedureNested(CurrentProcedure)
   THEN
      BinInstReg(pop, MaxDatum(), RegECX)
   END ;

   BinInst(leave) ;
   BinInst(ret) ;

   GenerateExitProcedureLabel(CurrentProcedure) ;
   CurrentProcedure := NulSym ;
   PreviousScope    := NulSym
END CodeKillLocalVar ;


(*
   DecreaseStack - decreases the runtime stack by c.
*)

PROCEDURE DecreaseStack (c: CARDINAL) ;
BEGIN
   IF c>0
   THEN
      InstRegInt(sub, MaxDatum(), RegESP, VAL(INTEGER, c))
   END
END DecreaseStack ;


(*
   IncreaseStack - increases the runtime stack by c.
*)

PROCEDURE IncreaseStack (c: CARDINAL) ;
BEGIN
   IF c>0
   THEN
      InstRegInt(add, MaxDatum(), RegESP, VAL(INTEGER, c))
   END
END IncreaseStack ;


(*
   CodeCall - determines whether the procedure call is a direct call
              or an indirect procedure call.
*)

PROCEDURE CodeCall ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Operator : CallOp
      Operand3 : Procedure
   *)
   IF IsProcedure(Operand3)
   THEN
      CodeDirectCall
   ELSIF IsProcType(GetType(Operand3))
   THEN
      CodeIndirectCall
   ELSE
      InternalError('Expecting Procedure or ProcType', __FILE__, __LINE__)
   END
END CodeCall ;


(*
   CheckNestedProcedure - checks whether we are about to call a nested procedure.
                          If the procedure is nested we must set up the activation
                          record pointer:

                            o    if we are the statically defined parent scope then we
                                 pass our frame pointer in ecx
                            o    otherwise we find out the previous statically defined parent
                                 and pass this in ecx.
*)

PROCEDURE CheckNestedProcedure (Sym: CARDINAL) ;
VAR
   BackScope: CARDINAL ;
BEGIN
   IF IsProcedureNested(Sym)
   THEN
      (*
         following Advanced Compiler Design and Implementation by Steven S.Muchnick p114

         (i)   if the procedure being called is nested directly within its caller, its
               static link points to its callers frame.
         (ii)  if the procedure is at the same level of nesting as its caller, then its
               static link is the same as the callers static link
         (iii) if the procedure being called is n levels higher than the caller in the
               nesting structure, then its static link can be determined by following, n,
               static links back from the callers static link and copying the static link
               found there.
      *)
      IF GetScopeAuthor(Sym)=CurrentProcedure
      THEN
         (* found (i) *)
         InstRegReg(mov, MaxDatum(), RegECX, RegEBP)
      ELSIF GetScopeAuthor(Sym)=PreviousScope
      THEN
         (* found (ii) *)
         InstRegSym(mov, MaxDatum(), RegECX, ActivationPointer)
      ELSE
         (* must be (iii) *)
         InstRegSym(mov, MaxDatum(), RegECX, ActivationPointer) ;
         BackScope := PreviousScope ;
         WHILE BackScope#GetScopeAuthor(Sym) DO
            PushOffset(ActivationPointer) ;
            InstRegMIO(mov, MaxDatum(), RegECX, RegECX, PopInt()) ;
            BackScope := GetScopeAuthor(BackScope)
         END
      END
   END
END CheckNestedProcedure ;


(*
   CodeDirectCall - saves all volitiles and jumps to a subroutine.
*)

PROCEDURE CodeDirectCall ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;

   (*
      At a Later Stage:

      If we are calling a procedure which is in
      a different interrupt priority MODULE to the current MODULE
      then
         push current interrupt mask
         set current interrupt mask to that of the called procedure MODULE
      end
   *)

   CheckNestedProcedure(Operand3) ;
   InstName(call, MaxDatum(), Operand3) ;

   (*
      At a later stage we could add MODULE priority here:

      If we have called a procedure which is in
      a different interrupt priority MODULE to the current MODULE
      then
         pop current interrupt mask
      end
   *)

   (*
      Now restore sp to above param 0.
      The Stack Pointer will be just above the procedure Parameters
   *)

   IF GetType(Operand3)=NulSym   (* Is it a function *)
   THEN
      (* Procedure *)
      IF UnboundedOnStack
      THEN
         RemoveAnyUnboundedData
      ELSE
         PushSumOfParamSize(Operand3) ;
         IncreaseStack(PopCard())
      END
   ELSE
      (* Function *)
      PushSumOfParamSize(Operand3) ;
      IF ReturnArgumentOnStack
      THEN
         PushSize(GetType(Operand3)) ;
         Sub
         (* Stack Should be above Param 0 *)
      END ;
      IncreaseStack(PopCard())
   END

END CodeDirectCall ;


(*
   CodeIndirectCall - saves all volitiles and jumps to a subroutine.
*)

PROCEDURE CodeIndirectCall ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Proc    : CARDINAL ;
   s       : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      At a Later Stage:

      If we are calling a procedure which is in
      a different interrupt priority MODULE to the current MODULE
      then
         push current interrupt mask
         set current interrupt mask to that of the called procedure MODULE
      end
   *)
   FetchDatum(Operand3, GetMoveMode(Operand3), GetUnitPrefix(GetSize(Operand3)),
              GetScope(Operand3), GetOffset(Operand3), 0, RegEAX) ;
   BinInstRegI(call, MaxDatum(), RegEAX) ;

   (*
      At a later stage we could add MODULE priority here:

      If we have called a procedure which is in
      a different interrupt priority MODULE to the current MODULE
      then
         pop current interrupt mask
      end
   *)

   (*
      Now restore sp to above param 0.
      The Stack Pointer will be just above the procedure Parameters
   *)

   Proc := GetType(Operand3) ;
   Assert(IsProcType(Proc)) ;
   IF GetType(Proc)=NulSym   (* Is it a function *)
   THEN
      (* Procedure *)
      IF UnboundedOnStack
      THEN
         RemoveAnyUnboundedData
      ELSE
         PushSumOfParamSize(Proc) ;
         IncreaseStack(PopCard())
      END
   ELSE
      (* Function *)
      PushSumOfParamSize(Proc) ;
      IF ReturnArgumentOnStack
      THEN
         PushSize(GetType(Operand3)) ;
         Sub
         (* Stack Should be above Param 0 *)
      END ;
      IncreaseStack(PopCard())
   END

END CodeIndirectCall ;


(*
   CodeParam - pushes the parameter onto the stack, if the parameter,
               number is greater than 0, otherwise the parameter
               is effectively space for a return value.

               NOTE that we almost can treat VAR and NON VAR parameters the same, expect for
                    some types:

                    procedure parameters
                    unbounded parameters

                    these require special attention and thus it is easier to test individually
                    for VAR and NON VAR parameters.

               NOTE that we CAN ignore ModeOfAddr though
*)

PROCEDURE CodeParam ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size    : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF Operand1=0
   THEN
      IF ReturnArgumentOnStack
      THEN
         (*
            Parameter 0 is the return value for a function. (Operand3)
            - must leave space for the value on the stack.
         *)
         PushSize(GetType(Operand3)) ;
         DecreaseStack(PopCard())
      END
   ELSE
      IF IsVarParam(Operand2, Operand1)
      THEN
         IF IsConst(Operand3)
         THEN
            WriteError('cannot pass a constant as a VAR parameter')
         ELSE
            (* ignore LeftAddr and RightAddr, but must be careful about size of Operand3 *)
            (* SIZE(operand3) will normally be TSIZE(ADDRESS) but NOT when it is unbounded *)
            MoveFromTo(Operand3, NulSym, MemoryDirect, StackDirect,
                       GetOffset(Operand3), 0, GetSize(Operand3),
                       GetScope(Operand3), NoScope)
         END
      ELSE
         IF IsConst(Operand3)
         THEN
            IF IsMathType(GetType(GetNthParam(Operand2, Operand1)))
            THEN
               CoprocessorParam(Operand3, GetType(GetNthParam(Operand2, Operand1)))
            ELSE
               MoveFromTo(Operand3, NulSym, Constant, StackDirect,
                          0, 0,
                          GetSize(GetType(GetNthParam(Operand2, Operand1))),
                          GetScope(Operand3), NoScope)
            END
         ELSE
            (* ignore LeftAddr and RightAddr for ParamOp *)
            (*
               we must use GetSize(GetType(GetNth ... as we might want to pass a procedure as a parameter.
               (a procedure does not know about its size)
            *)
            MoveFromTo(Operand3, NulSym, MemoryDirect, StackDirect,
                       GetOffset(Operand3), 0, GetSize(GetType(GetNthParam(Operand2, Operand1))),
                       GetScope(Operand3), NoScope)
         END
      END
   END
END CodeParam ;


(*
   CodeFunctValue - retrieves the function return value from the stack.
                    The result of the function should be on top of the
                    APU stack.
*)

PROCEDURE CodeFunctValue ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size    : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Operator : FunctValueOp
      Operand1 : The Returned Variable
      Operand2 : The Function Returning this Variable
   *)
   IF ReturnArgumentOnStack
   THEN
      MoveFromTo(NulSym, Operand1, StackDirect, GetMoveMode(Operand1),
                 0, GetOffset(Operand1), GetVarTypeSize(Operand1),
                 NoScope, GetScope(Operand1))
   ELSIF IsMathType(GetType(Operand1)) AND (GetMoveMode(Operand1)#MemoryIndirect)
   THEN
      CoprocessorFunctValue(Operand1)
   ELSE
      (*
         move RegEAX into Symbol Operand1 ignoring its mode (always assumed to be offset, see M2Quads.mod).
         note that the variable Operand1 might actually have a different mode (LeftValue) but we will
         treat it as RightValue since this is what M2Quads.mod expects. M2SubExp might reorder temporary
         usage, but it will still expect NO indirection to occur.
      *)
      IF GetMoveMode(Operand1)=MemoryDirect
      THEN
         (* straightforward *)
         PutDatum(Operand1, GetMoveMode(Operand1),
                  GetUnitPrefix(GetVarTypeSize(Operand1)), GetScope(Operand1),
                  GetOffset(Operand1), 0, RegEAX)
      ELSE
         (* M2SubExp has rearranged temporaries *)
         PutDatum(Operand1, MemoryDirect,
                  GetUnitPrefix(AddressSize), GetScope(Operand1),
                  GetOffset(Operand1), 0, RegEAX)
      END
   END ;
   RemoveAnyUnboundedData
END CodeFunctValue ;


(*
   CodeReturn - does nothing in the 80x86 code generator (return is done by KillLocalVar).
*)

PROCEDURE CodeReturn ;
BEGIN
END CodeReturn ;


(*
   CodeReturnValue - places the operand into the return value space
                     allocated by the function call.
*)

PROCEDURE CodeReturnValue ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Type    : CARDINAL ;
   Size, i : CARDINAL ;
   Num     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF ReturnArgumentOnStack
   THEN
(* APU
      Size := PushOperandMode(Operand3) ;
      FOR i := Size-1 TO 0 BY -1 DO
         CardToStr(i, 0, Num) ;
         Emit('', 'Store', 'idx', 'ra', Num)
      END
*)
   ELSE
      Type := GetType(Operand3) ; (* Fetch Function Return Type *)
      (* Move Operand1 into RegEAX taking into account its mode *)
      IF IsMathType(GetType(Operand3))
      THEN
         CoprocessorReturnValue(Operand1, Operand3)
      ELSIF FitsInRegister(Type)
      THEN
         FetchDatum(Operand1, GetMoveMode(Operand1),
                    GetUnitPrefix(GetSize(Type)), GetScope(Operand1),
                    GetOffset(Operand1), 0, RegEAX)
      ELSE
         WriteError('can only return values <= register size')
      END
   END
END CodeReturnValue ;


(*
   AlignStackPointer - aligns the stack pointer on a sensible boundary.
*)

PROCEDURE AlignStackPointer ;
BEGIN
   IF WordSize=DWORDSIZE
   THEN
      InstRegHexCard(and, MaxDatum(), RegESP, 0FFFFFFF8H)
   END
END AlignStackPointer ;


(*
   CodeUnbounded - codes the creation of an unbounded parameter
                   variable.
                   Pushes Op3 onto the stack and sets Op1 to the start
                   of this data.
*)

PROCEDURE CodeUnbounded ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size    : CARDINAL ;
BEGIN
   PrepareToPushUnboundedData ;
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Set Op1 to the start of the copy of Op3

      Push Op3 onto the stack - therefore Op1 := esp
   *)
   IF IsConstString(Operand3)
   THEN
      PushStringToStack(Operand3) ;
      (*
         Set Op1 to the start of the copy of Op3
         Remember 80386 stack grows from HIGH -> LOW memory.
         Therefore Op1 := esp
         Last entity on the stack is the first element of the unbounded array.
      *)
      PutDatum(Operand1, GetMoveMode(Operand1), MaxDatum(),
               GetScope(Operand1), GetOffset(Operand1), 0, RegESP)
   ELSE
      (*
         the following unbounded cases use reps so we should align the stack pointer
         after saving it to the string ArrayAddress unbounded field.
      *)
      IF IsArray(GetType(Operand3))
      THEN
         PushArrayToStack(Operand3)
      ELSIF IsUnbounded(GetType(Operand3))
      THEN
         PushUnbounded(Operand3)
      ELSE
         (*
            we should only execute this if we are passing an object to an ARRAY OF BYTE or ARRAY OF WORD
            parameter. Parameter checking will have been done in Pass 3 and we have no way of double
            checking - so here we go...
         *)
         PushArrayToStack(Operand3)
      END ;
      (*
         Set Op1 to the start of the copy of Op3
         Remember 80386 stack grows from HIGH -> LOW memory.
         Therefore Op1 := esp
         Last entity on the stack is the first element of the unbounded array.
      *)
      PutDatum(Operand1, GetMoveMode(Operand1), MaxDatum(),
               GetScope(Operand1), GetOffset(Operand1), 0, RegESP) ;
      AlignStackPointer
   END
END CodeUnbounded ;


(*
   PushUnbounded - pushes an Unbounded Array Sym data onto the stack.
*)

PROCEDURE PushUnbounded (Sym: CARDINAL) ;
VAR
   Size: CARDINAL ;
   Num1,
   Num2: ARRAY [0..20] OF CHAR ;
BEGIN
   (*
      esi := ArrayAddress
   *)
   FetchDatum(Sym, GetMoveMode(Sym), MaxDatum(), GetScope(Sym), GetOffset(Sym),
              0, RegESI) ;   (* Record field offset = 0 *)
   (*
      ecx := ArrayHigh
   *)
   FetchDatum(Sym, GetMoveMode(Sym), MaxDatum(), GetScope(Sym), GetOffset(Sym),
              INTEGER(WordSize), RegECX) ;
                             (* Record field offset = WordSize *)
   (*
      INC(ecx, 1+1)  because:  Size of Array -1 = HIGH
                               Possible null on array of char also 1
   *)
   InstRegInt(add, MaxDatum(), RegECX, 1+1) ;
   (*
      sub  esp, ecx,
      mov  edi, esp
      rep movs          ;  while ecx>0 do
                        ;     mem[edi] := mem[esi]
                        ;     inc(edi)
                        ;     inc(esi)
                        ;     dec(ecx)
                        ;  end
   *)
   InstRegReg(sub, MaxDatum(), RegESP, RegECX) ;
   InstRegReg(mov, MaxDatum(), RegEDI, RegESP) ;
   BinInstUnit(repmovs, PByte, 0) ;
END PushUnbounded ;


(*
   RemoveAnyUnboundedData - attempts to remove any undounded data.
                            Unbounded data will be on the stack if the
                            boolean, UnboundedOnStack, is true.
                            In which case the SP := RC.
*)

PROCEDURE RemoveAnyUnboundedData ;
BEGIN
   IF UnboundedOnStack
   THEN
      UnboundedOnStack := FALSE ;
      InstRegReg(mov, MaxDatum(), RegESP, RegEBX)  (* RegESP := RegEBX *)
   END
END RemoveAnyUnboundedData ;


(*
   PrepareToPushUnboundedData - prepares to push some unbounded data.
                                If this has not been called before
                                then Register RC := SP.
                                The value of SP is before any unbounded data
                                has been pushed - therefore take a copy of SP
                                into a reserve register.
*)

PROCEDURE PrepareToPushUnboundedData ;
BEGIN
   IF NOT UnboundedOnStack
   THEN
      (*
         Not done yet -
         so do this now and never again before the procedure call.
      *)
      UnboundedOnStack := TRUE ;
      InstRegReg(mov, MaxDatum(), RegEBX, RegESP)  (* RegEBX := old RegESP *)
   END
END PrepareToPushUnboundedData ;


(*
   CodeHigh - codes the procedure HIGH for a static array.
              Op1 is a cardinal or variable which can contain
                  a numerial ie INTEGER, etc.
              Op2 is NulSym
              Op3 contains the Array Symbol which should have its
                  highest bounds calculated.
*)

PROCEDURE CodeHigh ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   InternalError('compile time assignment - add to M2Optimize', __FILE__, __LINE__)
END CodeHigh ;


(*
   Addr Operator  - contains the address of a variable.

   Yields the address of a variable - need to add the frame pointer if
   a variable is local to a procedure.

   Sym1<X>   Addr   Sym2<X>     meaning     Mem[Sym1<I>] := Sym2<I>
*)

PROCEDURE CodeAddr ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand3)
   THEN
      WriteError('not expecting to be asked to find the address of a constant')
   ELSE
      (*
         Mem[Op1] := Op3
      *)
      IF IsProcedure(GetVarFather(Operand3))
      THEN
         IF WordSize=4
         THEN
            IF GetOffset(Operand3)=0
            THEN
               (* so unlikely - it may even be impossible? *)
               InstRegReg(mov, MaxDatum(), RegEAX, RegEBP)
            ELSE
               InstRegMD(lea, MaxDatum(), RegEAX, GetOffset(Operand3), 0, GetScope(Operand3), Operand3)
            END
         ELSE
            InstRegReg(mov, MaxDatum(), RegEAX, RegEBP) ;
            IF GetOffset(Operand3)<0
            THEN
               InstRegInt(sub, MaxDatum(), RegEAX, -GetOffset(Operand3))
            ELSIF GetOffset(Operand3)>0
            THEN
               InstRegInt(add, MaxDatum(), RegEAX, GetOffset(Operand3))
            END
         END ;
         InstMDReg(mov, MaxDatum(), GetOffset(Operand1), 0,
                   RegEAX, GetScope(Operand1), Operand1)
      ELSE
         InstMDOffset(mov, MaxDatum(), Operand1, Operand3)
      END
   END
END CodeAddr ;


(*
------------------------------------------------------------------------------
   Copy Operator
------------------------------------------------------------------------------
   Sym1<O>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
   Sym1<V>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
   Sym1<O>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]


PROCEDURE CodeCopy ;
VAR
   Operator        : QuadOperator ;
   Operand1,
   Operand2,
   Operand3        : CARDINAL ;
   Size1, Size3    : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Follow the Quadruple rules:
   *)

   IF IsConstString(Operand3)
   THEN
      MoveStringFromTo(Operand3, Operand1)
   ELSIF ((GetMode(Operand1)=RightValue) AND IsMathType(GetType(Operand1))) OR IsMathType(GetType(Operand3))
   THEN
      CoprocessorCopy(Operand1, Operand3)
   ELSIF IsConst(Operand3)
   THEN
      PushSize(Operand1) ;
      Size1 := PopCard() ;
      (*
         Mem[Op1] := constant
      *)
      MoveFromTo(Operand3, Operand1,
                 Constant, MemoryDirect,
                 GetOffset(Operand3), GetOffset(Operand1),
                 Size1,
                 GetScope(Operand3), GetScope(Operand1))
   ELSIF IsProcedure(Operand3)
   THEN
      PushSize(Operand1) ;
      Size1 := PopCard() ;
      MoveFromTo(Operand3, Operand1,
                 GetMoveMode(Operand3), MemoryDirect,
                 GetOffset(Operand3), GetOffset(Operand1),
                 Size1,
                 GetScope(Operand3), GetScope(Operand1))
   ELSE
      PushSize(Operand1) ;
      Size1 := PopCard() ;
      PushSize(Operand3) ;
      Size3 := PopCard() ;
      IF Size1#Size3
      THEN
         WriteError('cannot move different size entities into each other')
      END ;
      IF GetMode(Operand3)=RightValue
      THEN
         (*
            Mem[Op1] := Mem[Op3]
         *)
         MoveFromTo(Operand3, Operand1,
                    MemoryDirect, MemoryDirect,
                    GetOffset(Operand3), GetOffset(Operand1),
                    Size1,
                    GetScope(Operand3), GetScope(Operand1))
      ELSE
         (*
            Mem[Op1] := Mem[Mem[Op3]]
         *)
         Assert(GetMode(Operand3)=LeftValue) ;
         MoveFromTo(Operand3, Operand1,
                    MemoryIndirect, MemoryDirect,
                    GetOffset(Operand3), GetOffset(Operand1),
                    Size1,
                    GetScope(Operand3), GetScope(Operand1))
      END
   END
END CodeCopy ;
*)


(*
------------------------------------------------------------------------------
   IndrX Operator           a = *b
------------------------------------------------------------------------------
   Sym1<X>   IndrX   Sym2<I>     Meaning     Mem[Sym1<I>] := Mem[constant]
   Sym1<X>   IndrX   Sym2<X>     Meaning     Mem[Sym1<I>] := Mem[Mem[Sym3<I>]]

   (Operand2 is the type of the data being indirectly copied)
*)

PROCEDURE CodeIndrX ;
VAR
   Operator        : QuadOperator ;
   Operand1,
   Operand2,
   Operand3        : CARDINAL ;
   Size1, Size3    : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Follow the Quadruple rules:
   *)

   IF IsConstString(Operand3)
   THEN
      InternalError('not expecting to index through a constant string', __FILE__, __LINE__)
   ELSIF IsConst(Operand3)
   THEN
      InternalError('this needs more thought gaius FIXME indirection via a constant', __FILE__, __LINE__) ;
   ELSIF IsMathType(Operand2)
   THEN
      CoprocessorIndrX(Operand1, Operand2, Operand3)
   ELSE
      PushSize(Operand2) ;    (* Operand2 is the TYPE of the variable being copied *)
      Size1 := PopCard() ;
      (*
         Mem[Op1] := Mem[Mem[Op3]]
      *)
      MoveFromTo(Operand3, Operand1,
                 MemoryIndirect, MemoryDirect,
                 GetOffset(Operand3), GetOffset(Operand1),
                 Size1,
                 GetScope(Operand3), GetScope(Operand1))
   END
END CodeIndrX ;


(*
------------------------------------------------------------------------------
   XIndr Operator           *a = b
------------------------------------------------------------------------------
   Sym1<I>   XIndr   Sym2<X>     Meaning     Mem[constant]     := Mem[Sym3<I>]
   Sym1<X>   XIndr   Sym2<X>     Meaning     Mem[Mem[Sym1<I>]] := Mem[Sym3<I>]

   (Operand2 is the type of the data being indirectly copied)
*)

PROCEDURE CodeXIndr ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size1   : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Follow the Quadruple rules:
   *)

   IF IsConstString(Operand3)
   THEN
      PushSize(Operand2) ;
      MoveStringFromTo(Operand3, Operand1, PopCard(), MemoryIndirect)
   ELSIF IsConst(Operand1)
   THEN
      (* FIXME - this is legal *)
      InternalError('this needs more thought gaius FIXME indirection via a constant', __FILE__, __LINE__) ;
   ELSIF IsMathType(Operand2)
   THEN
      CoprocessorXIndr(Operand1, Operand2, Operand3)
   ELSE
      PushSize(Operand2) ;    (* Operand2 is the TYPE of the variable being copied *)
      Size1 := PopCard() ;
      (*
         Mem[Mem[Op1]] := Mem[Op3]
      *)
      MoveFromTo(Operand3, Operand1,
                 GetMoveMode(Operand3), MemoryIndirect,
                 GetOffset(Operand3), GetOffset(Operand1),
                 Size1,
                 GetScope(Operand3), GetScope(Operand1))
   END
END CodeXIndr ;


(*
------------------------------------------------------------------------------
   := Operator
------------------------------------------------------------------------------
   Sym1<I> := Sym3<I>           := produces a constant
   Sym1<O> := Sym3<O>           := has the effect Mem[Sym1<I>] := Mem[Sym3<I>]
*)

PROCEDURE CodeBecomes ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   m3      : ModeOfMove ;
   Offset3,
   Offset1 : INTEGER ;
   Size1,
   Size3   : CARDINAL ;
   Reg     : Register ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand3)
   THEN
      PushValue(Operand3) ;
      PopValue(Operand1)
   ELSIF IsConstString(Operand3)
   THEN
      MoveStringFromTo(Operand3, Operand1, GetVarTypeSize(Operand1), MemoryDirect)
   ELSIF IsProcedure(Operand3)
   THEN
      MoveFromTo(Operand3, Operand1,
                 GetMoveMode(Operand3), MemoryDirect,
                 GetOffset(Operand3), GetOffset(Operand1),
                 GetVarTypeSize(Operand1),
                 GetScope(Operand3), GetScope(Operand1))
   ELSE
      IF GetMode(Operand1)=LeftValue
      THEN
         PushSize(Address) ;
         Size1 := PopCard()
      ELSE
         Size1 := GetVarTypeSize(Operand1)
      END ;
      Offset1 := GetOffset(Operand1) ;
      IF IsConst(Operand3)
      THEN
         Size3 := Size1 ;
         Offset3 := 0 ;
         m3 := Constant
      ELSE
         IF GetMode(Operand3)=LeftValue
         THEN
            PushSize(Address) ;
            Size3 := PopCard()
         ELSE
            Size3 := GetVarTypeSize(Operand3)
         END ;
         Offset3 := GetOffset(Operand3) ;
         m3 := MemoryDirect
      END ;
      IF (IsMathType(GetType(Operand1)) OR IsMathType(GetType(Operand3))) AND
         ((GetMode(Operand1)#LeftValue) AND (GetMode(Operand3)#LeftValue))
      THEN
         (* coprocessor used and we are not calculating addresses *)
         CoprocessorBecomes(Operand1, Operand3)
      ELSIF (Size1#Size3) AND (NOT (IsMathType(GetType(Operand1)) AND (IsMathType(GetType(Operand3)))))
      THEN
         BeginError ;
         WriteLn ;
         WriteKey(GetSymName(Operand1)) ; WriteString(' := ') ; WriteKey(GetSymName(Operand3)) ; WriteLn ;
         WriteString('Type of ') ; WriteKey(GetSymName(Operand1)) ;
         WriteString(' is ') ; WriteKey(GetSymName(GetType(Operand1))) ; WriteLn ;
         WriteString('Type of ') ; WriteKey(GetSymName(Operand3)) ;
         WriteString(' is ') ; WriteKey(GetSymName(GetType(Operand3))) ; WriteLn ;
         WriteError('different sizes in :=')
      ELSE
         MoveFromTo(Operand3, Operand1,
                    m3, MemoryDirect,
                    Offset3, Offset1,
                    Size1,
                    GetScope(Operand3), GetScope(Operand1))
      END
   END
END CodeBecomes ;


(*
   MoveStringFromTo - moves a string, StringSym, into DestSym memory location.
                      The addressing mode, m, is obeyed and the destination size of
                      the variable is DestSize.
*)

PROCEDURE MoveStringFromTo (StringSym: CARDINAL; DestSym: CARDINAL; DestSize: CARDINAL; m: ModeOfMove) ;
BEGIN
   IF m=MemoryIndirect
   THEN
      MoveStringFromToIndirect(StringSym, DestSym, DestSize)
   ELSE
      MoveStringFromToDirect(StringSym, DestSym, DestSize)
   END
END MoveStringFromTo ;


(*
   MoveStringFromToIndirect - moves a string, StringSym, to an indirect destination.
*)

PROCEDURE MoveStringFromToIndirect (StringSym: CARDINAL; DestSym: CARDINAL; DestSize: CARDINAL) ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
   b: ARRAY [0..3] OF CHAR ;
   RegTemp   : Register ;
   Pack,
   i,
   String,
   SourceSize: CARDINAL ;
BEGIN
   (* Make sure the string is the same size as DestSym *)
   String := GetString(StringSym) ;
   SourceSize := GetStringLength(StringSym) ;
   RegTemp := SymToReg(DestSym) ;   (* RegTemp := DestSym *)
   GetKey(String, a) ;
   i := 0 ;
   WHILE i<SourceSize DO
      Pack := PackString(a, DestSize, i, b) ;
      MovePackedStringToIReg(RegTemp, INTEGER(i), b, Pack) ;
      INC(i, Pack)
   END ;
   IF i>DestSize
   THEN
      WriteError('string is too large for the variable')
   END ;
   (* Now fill in the remaining space with nul characters *)
   WHILE i<DestSize DO
      Pack := PackString('', DestSize, i, b) ;
      MovePackedStringToIReg(RegTemp, INTEGER(i), b, Pack) ;
      INC(i, Pack)
   END
END MoveStringFromToIndirect ;


(*
   MoveStringFromToDirect - moves a string, StringSym, to a direct destination.
*)

PROCEDURE MoveStringFromToDirect (StringSym: CARDINAL; DestSym: CARDINAL; DestSize: CARDINAL) ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
   b: ARRAY [0..3] OF CHAR ;
   Pack,
   i,
   String,
   SourceSize: CARDINAL ;
BEGIN
   (* Make sure the string is the same size as DestSym *)
   String := GetString(StringSym) ;
   SourceSize := GetStringLength(StringSym) ;
   IF SourceSize>DestSize
   THEN
      NearTokens('attempting to copy a string into a destination which is too small (possibly increase destination size)',
                 GetFirstUsed(StringSym), GetFirstUsed(DestSym))
   END ;
   GetKey(String, a) ;
   i := 0 ;
   WHILE i<SourceSize DO
      Pack := PackString(a, DestSize, i, b) ;
      MovePackedStringTo(DestSym, INTEGER(i), b, Pack) ;
      INC(i, Pack)
   END ;
   (* Now fill in the remaining space with nul characters *)
   WHILE i<DestSize DO
      Pack := PackString('', DestSize, i, b) ;
      MovePackedStringTo(DestSym, INTEGER(i), b, Pack) ;
      INC(i, Pack)
   END
END MoveStringFromToDirect ;


(*
   PushArrayToStack - moves an array, ArraySym, onto the stack.
                      Pushed in reverse order and therefore the start
                      of the array must be the top of the stack.
                      80386 stack grows from HIGH -> LOW
*)

PROCEDURE PushArrayToStack (ArraySym: CARDINAL) ;
VAR
   Size: CARDINAL ;
   Proc: CARDINAL ;
BEGIN
   CASE GetMoveMode(ArraySym) OF

   MemoryDirect:   Size := GetSize(ArraySym) ;
                   (* make room for array on the stack. *)
                   InstRegInt(sub, MaxDatum(), RegESP, INTEGER(Size)) ;
                   InstRegReg(mov, MaxDatum(), RegEDI, RegESP) ;
                   (* find start address of ArraySym *)
                   FetchAddress(ArraySym, RegESI) ;
                   BlockMove(Size) |
   MemoryIndirect: Size := GetVarTypeSize(ArraySym) ;
                   (* make room for array on the stack. *)
                   InstRegInt(sub, MaxDatum(), RegESP, INTEGER(Size)) ;
                   InstRegReg(mov, MaxDatum(), RegEDI, RegESP) ;
                   (* find start address of ArraySym *)
                   InstRegMD(mov, MaxDatum(), RegESI, GetOffset(ArraySym), 0, GetScope(ArraySym), ArraySym) ;
                   BlockMove(Size)

   ELSE
      InternalError('not expecting to reach here', __FILE__, __LINE__)
   END
END PushArrayToStack ;


(*
   PaddString - padd a string 'a' into another string 'b',
                starting at position i in string 'a'.
                The number of characters packed into b are returned.
                String b is assummed to contain 4 elements and we
                will use all of them.
*)

PROCEDURE PaddString (a: ARRAY OF CHAR; DestSize: CARDINAL; i: CARDINAL;
                      VAR b: ARRAY OF CHAR) ;
VAR
   SourceSize, j: CARDINAL ;
BEGIN
   SourceSize := StrLen(a) ;
   FOR j := 0 TO WordSize-1 DO
      IF i+j>=SourceSize
      THEN
         b[j] := nul
      ELSE
         b[j] := a[i+j]
      END
   END ;
   IF WordSize<HIGH(b)
   THEN
      b[WordSize] := nul     (* May need to terminate the string, b *)
   END
END PaddString ;


(*
   PushStringToStack - moves a string, StringSym, onto the stack.
                       The string is terminated with a nul character.
*)

PROCEDURE PushStringToStack (StringSym: CARDINAL) ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
   b: ARRAY [0..3] OF CHAR ;
   i,
   String,
   Length: CARDINAL ;
BEGIN
   String := GetString(StringSym) ;
   Length := GetStringLength(StringSym)+1 ;  (* +1 to include nul *)
   GetKey(String, a) ;
   i := Length ;
   IF i MOD WordSize#0
   THEN
      i := i+WordSize-(i MOD WordSize)
   END ;
   WHILE i>=WordSize DO
      DEC(i, WordSize) ;
      PaddString(a, Length, i, b) ;
      PushString(WordSize, b)
   END
END PushStringToStack ;


(*
   PackString - packs a string 'a' into another string 'b',
                starting at position i in string 'a'.
                The number of characters packed into b are returned.
                String b is assummed to contain 4 elements.
*)

PROCEDURE PackString (a: ARRAY OF CHAR; DestSize, i: CARDINAL;
                      VAR b: ARRAY OF CHAR) : CARDINAL ;
VAR
   SourceSize,
   Pack, j   : CARDINAL ;
BEGIN
   SourceSize := StrLen(a) ;
   IF (DestSize>SourceSize) AND
      (((SourceSize-i=3) AND (WordSize=DWORDSIZE)) OR (SourceSize-i=1))
   THEN
      (* Padd out remaining boundary byte with the nul character *)
      Pack := SourceSize-i+1
   ELSE
      Pack := GetUnitOfTransfer(DestSize-i)
   END ;
   FOR j := 0 TO Pack-1 DO
      IF i+j>=SourceSize
      THEN
         b[j] := nul
      ELSE
         b[j] := a[i+j]
      END
   END ;
   IF Pack<HIGH(b)
   THEN
      b[Pack] := nul     (* May need to terminate the string, b *)
   END ;
   RETURN( Pack )
END PackString ;


(* ****************************************************************
(*
   PushStringToStack - moves a string, StringSym, onto the stack.
                       The string is terminated with a nul character.
*)

PROCEDURE PushStringToStack (StringSym: CARDINAL) ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
   b: ARRAY [0..3] OF CHAR ;
   i     : INTEGER ;
   Pack,
   String,
   Length: CARDINAL ;
BEGIN
   String := GetString(StringSym) ;
   Length := GetStringLength(StringSym)+1 ;  (* +1 to include nul *)
   GetKey(String, a) ;
   (* Fill in 0..Length-1 *)
   i := Length ;
   CASE (i MOD WordSize) OF

   1:  DEC(i) ;
       Pack := PackString(a, Length, CARDINAL(i), b) ;
       PushString(Pack, b) ;
       Assert(Pack=1) |
   2:  DEC(i, 2) ;
       Pack := PackString(a, Length, CARDINAL(i), b) ;
       PushString(Pack, b) ;
       Assert(Pack=2) |
   3:  DEC(i) ;
       Pack := PackString(a, Length, CARDINAL(i), b) ;
       PushString(Pack, b) ;
       Assert(Pack=1) ;
       DEC(i, 2) ;
       Pack := PackString(a, Length, CARDINAL(i), b) ;
       PushString(Pack, b) ;
       Assert(Pack=2)

   ELSE
   END ;
   IF Length>=WordSize
   THEN
      DEC(i, WordSize) ;
      WHILE i>=0 DO
         Pack := PackString(a, Length, CARDINAL(i), b) ;
         Assert(Pack=WordSize) ;
         PushString(Pack, b) ;
         DEC(i, WordSize)
      END
   ELSE
      Assert(i=0)
   END
END PushStringToStack ;


 ************************************************************* *)

PROCEDURE CodeGoto ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   CardToStr(Operand3, 0, Str) ;
   StrConCat('L', Str, Str) ;
   DotifyLabel(Str) ;
   InstStr(jmp, Str)
END CodeGoto ;


(*
   CallInnerInit - produce a call to inner module initialization routine.
*)

PROCEDURE CallInnerInit (Sym: CARDINAL) ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   GetKey(GetModuleInitName(Sym), a) ;
   InstStr(call, a)
END CallInnerInit ;


(*
   CodeStart - emits starting code before the main BEGIN END of the
               current module.
*)

PROCEDURE CodeStart ;
VAR
   op           : QuadOperator ;
   op1, op2, op3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, op, op1, op2, op3) ;
   IF CompilingMainModule
   THEN
      WritePrologue(op3, CurrentQuad, op1) ;
      ForeachInnerModuleDo(op3, CallInnerInit)
   END
END CodeStart ;


(*
   CodeEnd - emits terminating code after the main BEGIN END of the
             current module.
*)

PROCEDURE CodeEnd ;
VAR
   op           : QuadOperator ;
   op1, op2, op3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, op, op1, op2, op3) ;
   WriteEpilogue(op3) ;
   IF GetMainModule()=op3
   THEN
      FlushCoprocessor
   END
END CodeEnd ;


PROCEDURE CodeOffset ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      InternalError('offset should be calculated when folding constants', __FILE__, __LINE__)
(*
      PushOffset(Operand3) ;
      PopValue(Operand1)
*)
   ELSE
      PushOffset(Operand3) ;
      IntToSym(PopInt(), Operand1)
   END
END CodeOffset ;


(*
   CodeBase - calculates the offset from the address of the array. This operation
              is now calculated with M2Optimize as the value will be constant.
*)

PROCEDURE CodeBase ;
BEGIN
   InternalError('BaseOp should have been optimized away in M2Optimize', __FILE__, __LINE__)
END CodeBase ;


PROCEDURE CodeElementSize ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Subscript: CARDINAL ;
   Type     : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   Type := Operand2 ;
   IF NoOfElements(Type)>=Operand3
   THEN
      Subscript := GetNth(Type, Operand3) ;
      PushSize(Subscript) ;
      IF IsConst(Operand1)
      THEN
         PopValue(Operand1)
      ELSE
         IntToSym(PopInt(), Operand1)
      END
   ELSE
      WriteError('too many elements') ;
   END
END CodeElementSize ;


(*
   IsNear - returns TRUE if two quadruples are near each other.
            This is a guess and it guesses that two quadruples will
            be <=127 bytes appart.
*)

PROCEDURE IsNear (a, b: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN(
          ((a>=b) AND (a-b<=NearQuadGuess)) OR
          ((a<b) AND (b-a<=NearQuadGuess))
         )
END IsNear ;


(*
   Reverse - returns the opposite instruction.
*)

PROCEDURE Reverse (CondJump: Instruction) : Instruction ;
BEGIN
   CASE CondJump OF

   je : RETURN( jne ) |
   jne: RETURN( je ) |
   jl : RETURN( jge ) |
   jge: RETURN( jl ) |
   jg : RETURN( jle ) |
   jle: RETURN( jg )

   ELSE
      WriteError('dont understand this kind of conditional jump')
   END
END Reverse ;


(*
   CreateJump - creates a jump to the quadruple, To, using
                the instruction, Jump. If we are using the 8086/8088
                then we have to be careful not to overstep the byte
                offset.
*)

PROCEDURE CreateJump (To, From: CARDINAL; Jump: Instruction) ;
VAR
   Str: ARRAY [0..20] OF CHAR ;
   New: Instruction ;
BEGIN
   IF (WordSize=4) OR IsNear(To, From)
   THEN
      (*
         80386 and 80486 don't have any silly restrictions.
         Or we guess that the ten quadruples can be reached by 128 bytes.
      *)
      CardToStr(To, 0, Str) ;
      StrConCat('L', Str, Str) ;
      DotifyLabel(Str) ;
      InstStr(Jump, Str)
   ELSE
      (* reverse condition of jump and branch around an unconditional jump *)
      New := Reverse(Jump) ;
      CardToStr(From, 0, Str) ;
      StrConCat('Lt', Str, Str) ;
      DotifyLabel(Str) ;
      InstStr(New, Str) ;
      (* unconditional jump *)
      CardToStr(To, 0, Str) ;
      StrConCat('L', Str, Str) ;
      DotifyLabel(Str) ;
      InstStr(jmp, Str) ;
      (* create new label *)
      CardToStr(From, 0, Str) ;
      StrConCat('Lt', Str, Str) ;
      DotifyLabel(Str) ;
      WriteString(Str) ; Write(':') ; WriteLn
   END
END CreateJump ;


(*
   CodeIfIn - codes a bit test determining whether operand1 is in operand2.
*)

PROCEDURE CodeIfIn ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF SetIn()
      THEN
         CodeGoto
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSE
      Compare(and, Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, jne)
   END
END CodeIfIn ;


(*
   CodeIfNotIn - codes a bit test determining whether operand1 is not
                 in operand2.
*)

PROCEDURE CodeIfNotIn ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF SetIn()
      THEN
         (* Do nothing since never arrises! *)
      ELSE
         CodeGoto
      END
   ELSE
      Compare(and, Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, je)
   END
END CodeIfNotIn ;


PROCEDURE CodeIfEqu ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF Equ()
      THEN
         CodeGoto
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSIF IsMathType(GetType(Operand1)) OR IsMathType(GetType(Operand2))
   THEN
      CoprocessorIfEqu(Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, je)
   ELSE
      Compare(cmp, Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, je)
   END
END CodeIfEqu ;


PROCEDURE CodeIfNotEqu ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF NotEqu()
      THEN
         CodeGoto
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSIF IsMathType(GetType(Operand1)) OR IsMathType(GetType(Operand2))
   THEN
      CoprocessorIfNotEqu(Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, jne)
   ELSE
      Compare(cmp, Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, jne)
   END
END CodeIfNotEqu ;


PROCEDURE CodeIfLess ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF Less()
      THEN
         CodeGoto
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSIF IsMathType(GetType(Operand1)) OR IsMathType(GetType(Operand2))
   THEN
      CoprocessorIfLess(Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, jae)
   ELSE
      Compare(cmp, Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, jl)
   END
END CodeIfLess ;


PROCEDURE CodeIfGre ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF Gre()
      THEN
         CodeGoto
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSIF IsMathType(GetType(Operand1)) OR IsMathType(GetType(Operand2))
   THEN
      CoprocessorIfGre(Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, jne)
   ELSE
      Compare(cmp, Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, jg)
   END
END CodeIfGre ;


PROCEDURE CodeIfLessEqu ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF LessEqu()
      THEN
         CodeGoto
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSIF IsMathType(GetType(Operand1)) OR IsMathType(GetType(Operand2))
   THEN
      CoprocessorIfLessEqu(Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, jne)
   ELSE
      Compare(cmp, Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, jle)
   END
END CodeIfLessEqu ;


PROCEDURE CodeIfGreEqu ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF GreEqu()
      THEN
         CodeGoto
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSIF IsMathType(GetType(Operand1)) OR IsMathType(GetType(Operand2))
   THEN
      CoprocessorIfGreEqu(Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, jne)
   ELSE
      Compare(cmp, Operand1, Operand2) ;
      CreateJump(Operand3, CurrentQuad, jge)
   END
END CodeIfGreEqu ;


(*
   Compare - produces code that compares symbols Sym1 and Sym2.
             The mode of the variable is taken into consideration.
*)

PROCEDURE Compare (Inst: Instruction; Sym1, Sym2: CARDINAL) ;
VAR
   ResultReg,
   Reg      : Register ;
   Unit     : DatumPrefix ;
BEGIN
   (* Only allow register sized variables to be compared at the moment *)
   IF NOT EqualSizes(Sym1, Sym2)
   THEN
      WriteError('size of data must be the same for comparison')
   END ;
   (* Testing for size - word size *)
   IF (IsConst(Sym1) OR FitsInRegister(GetType(Sym1))) AND
      (IsConst(Sym2) OR FitsInRegister(GetType(Sym2)))
   THEN
      ResultReg := Inst2ModeAddress(Inst, Sym1, Sym2, Reg, Unit) ;
      IF (Inst=and) OR (Inst=or)
      THEN
         (* test reg,reg *)
         InstRegReg(test, Unit, ResultReg, ResultReg)
      END
   ELSE
      WriteError('implementation restriction - only allowed to compare <= register size')
   END
END Compare ;


(*
   TerminationCode - displays termination code
*)

PROCEDURE TerminationCode ;
BEGIN
   CoprocessorTerminate ;
   CloseAssemblyFileAndTerminate
END TerminationCode ;


(*
   InitializationCode - displays the initialization code.
*)

PROCEDURE InitializationCode ;
BEGIN
   OpenAssemblyFileAndInitialize   
END InitializationCode ;


(*
   Init - initialize global variables for this module.
*)

PROCEDURE Init ;
BEGIN
   UnboundedOnStack := FALSE ;
   PreviousScope    := NulSym ;
   CurrentProcedure := NulSym ;

   LastLine := 0 ;  (* Delete this line when the Quadruple FileOp is implemented *)
   LastOperator := AddOp
END Init ;


BEGIN
   Init
END M2GenStatic486.
