IMPLEMENTATION MODULE M2Inline ;


FROM ASCII IMPORT nul, cr ;
FROM StdIO IMPORT Write ;
FROM StrIO IMPORT WriteLn ;


CONST
   MaxInline = 20000 ;      (* Maximum amount of inline text allowed. *)
   MaxBuffer =   100 ;      (* Maximum number of inline buffers.      *)

TYPE
   Frame  = RECORD
               Start,                (* Start of this buffers text.   *)
               Size : CARDINAL ;     (* Size of this buffers text.    *)
            END ;


VAR
   Frames       : ARRAY [1..MaxBuffer] OF Frame ;
                                          (* Inline text storage.     *)
   Buffer       : ARRAY [0..MaxInline] OF CHAR ;
   CurrentPos   : CARDINAL ;              (* Index into buffer.       *)
   CurrentBuffer: CARDINAL ;              (* Buffer Currently in use. *)


(*
   NewInlineBuffer - creates a new inline buffer and returns the
                     buffer index.
*)

PROCEDURE NewInlineBuffer () : CARDINAL ;
BEGIN
   INC(CurrentBuffer) ;
   WITH Frames[CurrentBuffer] DO
      Start := CurrentPos ;
      Size := 0
   END ;
   RETURN( CurrentBuffer )
END NewInlineBuffer ;


(*
   PutInlineBuffer - places a character, ch, into the current buffer.
*)

PROCEDURE PutInlineBuffer (ch: CHAR) ;
BEGIN
   WITH Frames[CurrentBuffer] DO
      Buffer[CurrentPos] := ch ;
      INC(Size) ;
      CurrentPos := Start+Size
   END
END PutInlineBuffer ;


(*
   WriteInlineBuffer - writes a buffer indexed by, Buf.
*)

PROCEDURE WriteInlineBuffer (Buf: CARDINAL) ;
VAR 
   i: CARDINAL ;
BEGIN
   WITH Frames[Buf] DO
      i := Start ;
      WHILE Buffer[i]#nul DO
         IF Buffer[i]=cr
         THEN
            (* indicates a WriteLn necessary *)
            WriteLn
         ELSE
            Write(Buffer[i])
         END ;
         INC(i)
      END
   END ;
   WriteLn
END WriteInlineBuffer ;


(*
   Init - initializes the data structures for this module.
*)

PROCEDURE Init ;
BEGIN
   CurrentBuffer := 0 ;
   CurrentPos := 0
END Init ;


BEGIN
   Init
END M2Inline.
