IMPLEMENTATION MODULE M2Math ;


FROM SymbolTable IMPORT NulSym,
                        SetCurrentModule,
                        StartScope,
                        EndScope,
      	       	     	MakeConstVar,
                        MakeProcedure,
                        PutFunction,
                        PutConst,
                        PutConstString,
                        PopValue ;

FROM M2Base IMPORT Real, LongReal ;
FROM M2Batch IMPORT MakeDefinitionSource ;
FROM NameKey IMPORT MakeKey ;
FROM M2ALU IMPORT PushLongReal ;


(*
   IsPseudoMathFunction - returns TRUE if Sym is an in built math function.
*)

PROCEDURE IsPseudoMathFunction (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN(
          (Sym=sin) OR (Sym=cos) OR (Sym=tan) OR (Sym=atan) OR (Sym=sqrt)
         )
END IsPseudoMathFunction ;


(*
   InitMath - initializes the module MATH
*)

PROCEDURE InitMath ;
VAR
   Math: CARDINAL ;
BEGIN
   (* create MATH module *)
   Math := MakeDefinitionSource(MakeKey('MATH')) ;
   StartScope(Math) ;

   (* predefined constants *)
   pi := MakeConstVar(MakeKey('pi')) ;
   PutConst(pi, LongReal) ;
   PushLongReal(3.1415926535897931158542633056640625) ;
      (*
         we should really implement this by using a hex constant
            - a trade off between accuracy and portability.
            - we shouldn't really do the above as we introduce
              more errors each time we convert string -> longreal.
      *)

   PopValue(pi) ;

   (* And now the predefined pseudo functions *)

   sin := MakeProcedure(MakeKey('sin')) ;     (* Function        *)
(*
   PutFunction(sin, Real) ;                   (* Return Type     *)
                                              (* REAL/LONGREAL   *)
*)

   cos := MakeProcedure(MakeKey('cos')) ;     (* Function        *)
(*
   PutFunction(cos, Real) ;                   (* Return Type     *)
                                              (* REAL/LONGREAL   *)
*)

   tan := MakeProcedure(MakeKey('tan')) ;     (* Function        *)
(*
   PutFunction(tan, Real) ;                   (* Return Type     *)
                                              (* REAL/LONGREAL   *)
*)
   atan := MakeProcedure(MakeKey('atan')) ;   (* Function        *)
(*
   PutFunction(atan, Real) ;                  (* Return Type     *)
                                              (* REAL/LONGREAL   *)
*)
   sqrt := MakeProcedure(MakeKey('sqrt')) ;   (* Function        *)
(*
   PutFunction(sqrt, Real) ;                  (* Return Type     *)
                                              (* REAL/LONGREAL   *)
*)
   EndScope
END InitMath ;


END M2Math.
