IMPLEMENTATION MODULE M2Optimize ;


FROM M2Debug IMPORT Assert ;

FROM NameKey IMPORT WriteKey, MakeKey, GetKey ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM NumberIO IMPORT WriteCard ;

FROM M2Lexical IMPORT NearToken, NearTokens, InternalError, WriteErrorFormat2 ;
FROM M2Batch IMPORT GetModuleNo ;
FROM M2Options IMPORT Quiet, CompilerDebugging ;
FROM M2EvalSym IMPORT Evaluate ;

FROM SymbolTable IMPORT IsConst, IsValueSolved, IsOffsetSolved,
                        PushValue, PopValue, PushOffset, PushSize,
                        IsConstString, GetString, PutConstString,
                        IsSet, IsRecordField,
                        GetSymName,
                        GetProcedureQuads, GetModuleQuads,
                        GetModule, GetNthProcedure,
                        GetSubrange,
                        PutProcedureReachable, IsProcedureReachable,
                        PutProcedureStartQuad, PutProcedureEndQuad,
                        IsProcedure, IsUnbounded, IsArray,
                        GetDeclared, GetFirstUsed,
                        GetNth, NoOfElements, GetType,
                        IsExportQualified, IsExportUnQualified,
                        IsSizeSolved,
                        PutConst,
                        NulSym ;

FROM M2Quads IMPORT QuadOperator, GetQuad, GetNextQuad, PutQuad, SubQuad,
                    Opposite, IsReferenced, GetRealQuad, QuadToTokenNo ;

FROM M2ALU IMPORT Add, Sub, Mult, Div, Mod,
      	       	  SetOr, SetAnd, Bit, BitRange,
                  SetDifference, SetSymmetricDifference,
                  Equ, NotEqu, Less, LessEqu, Gre, GreEqu,
                  PushCard, PopInt, PushInt ;

FROM M2Constants IMPORT MakeNewConstFromValue ;
FROM M2Base IMPORT IsCompatible, MixTypes, Cardinal ;
FROM M2System IMPORT Address ;
FROM StrLib IMPORT StrConCat ;


CONST
   MaxString = 255 ;


(* %%%FORWARD%%%
PROCEDURE FoldOr (VAR Head: CARDINAL; QuadNo: CARDINAL;
                  Operand1, Operand2, Operand3: CARDINAL;
                  VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldAnd (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL;
                   VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldXor (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL;
                   VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldSetDifference (VAR Head: CARDINAL; QuadNo: CARDINAL;
                             Operand1, Operand2, Operand3: CARDINAL;
                             VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE MakeExportedProceduresReachable ; FORWARD ;
PROCEDURE Delete (VAR Head: CARDINAL; Start, End: CARDINAL) ; FORWARD ;
PROCEDURE DeleteUnReachableProcedures (VAR Head: CARDINAL) ; FORWARD ;
PROCEDURE DisplayReachable ; FORWARD ;
PROCEDURE FoldMod (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL;
                   VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldBit (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldBitRange (VAR Head: CARDINAL; QuadNo: CARDINAL;
                        Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldAdd (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldBase (VAR Head: CARDINAL; QuadNo: CARDINAL;
                    Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldBecomes (VAR Head: CARDINAL; QuadNo: CARDINAL;
                       Operand1, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldDiv (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldElementSize (VAR Head: CARDINAL; QuadNo: CARDINAL;
                           Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldElementSizeForArray (VAR Head: CARDINAL;
                                   QuadNo: CARDINAL;
                                   Operand1, Operand2,
                                   Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldElementSizeForUnbounded (VAR Head: CARDINAL;
                                       QuadNo: CARDINAL;
                                       Operand1, Operand2,
                                       Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldEqu (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldGre (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldGreEqu (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldHigh (VAR Head: CARDINAL; QuadNo: CARDINAL;
                    Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldLess (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldLessEqu (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldMult (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldMultipleGoto (QuadNo: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE FoldNegate (VAR Head: CARDINAL; QuadNo: CARDINAL;
                      Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldNotEqu (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldOffset (VAR Head: CARDINAL; QuadNo: CARDINAL;
                      Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldSize (VAR Head: CARDINAL; QuadNo: CARDINAL;
                    Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE FoldSub (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL; VAR Unresolved: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE KnownReach (Sym: CARDINAL) ; FORWARD ;
PROCEDURE KnownReachable (Start, End: CARDINAL) ; FORWARD ;
PROCEDURE KnownReachableInitCode(Start, End: CARDINAL) ; FORWARD ;
PROCEDURE ReduceBranch (VAR Head: CARDINAL;
                        Operator: QuadOperator;
                        CurrentQuad,
                        CurrentOperand1, CurrentOperand2,
                        CurrentOperand3: CARDINAL;
                        VAR NextQuad: CARDINAL;
                        Folded: BOOLEAN) : BOOLEAN ; FORWARD ;
PROCEDURE IsBasicBlock (Head, From, To: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE ReduceGoto (VAR Head: CARDINAL;
                      CurrentQuad, CurrentOperand3, NextQuad: CARDINAL;
                      Folded: BOOLEAN) : BOOLEAN ; FORWARD ;

   %%%FORWARD%%% *)


(*
   FoldConstants - folds as many constants as possible in the list of
                   quadruples provided by Head.
                   - It does not instantiate variables as constants
                     this should be attempted when BasicBlock works
                     as at the moment the looping structure prevents
                     this optimization.


                   NOTE we must fold constants before we call Evaluate
                        this ensures that Named Constants (ConstVar's)
                        have been assigned values.
*)

PROCEDURE FoldConstants (VAR Head: CARDINAL) ;
VAR
   NeverCalledEvaluate,
   Previous,
   Unresolved,
   Folded             : BOOLEAN ;
   Operator           : QuadOperator ;
   Operand1,
   Operand2,
   Operand3           : CARDINAL ;
   i                  : CARDINAL ;
BEGIN
   NeverCalledEvaluate := TRUE ;
   REPEAT
      i := Head ;
      Folded := FALSE ;
      Unresolved := FALSE ;
      WHILE i#0 DO
         Previous := Unresolved ;
         GetQuad(i, Operator, Operand1, Operand2, Operand3) ;
         CASE Operator OF

         BitOp              : Folded := FoldBit(Head, i, Operand1, Operand3, Unresolved) |
         BitRangeOp         : Folded := FoldBitRange(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         BecomesOp          : Folded := FoldBecomes(Head, i, Operand1, Operand3, Unresolved) |

         NegateOp           : Folded := FoldNegate(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         AddOp              : Folded := FoldAdd(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         SubOp              : Folded := FoldSub(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         MultOp             : Folded := FoldMult(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         DivOp              : Folded := FoldDiv(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         ModOp              : Folded := FoldMod(Head, i, Operand1, Operand2, Operand3, Unresolved) |

         LogicalOrOp        : Folded := FoldOr(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         LogicalAndOp       : Folded := FoldAnd(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         LogicalXorOp       : Folded := FoldXor(Head, i, Operand1, Operand2, Operand3, Unresolved) |

         IfEquOp            : Folded := FoldEqu(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         IfNotEquOp         : Folded := FoldNotEqu(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         IfLessOp           : Folded := FoldLess(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         IfGreOp            : Folded := FoldGre(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         IfLessEquOp        : Folded := FoldLessEqu(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         IfGreEquOp         : Folded := FoldGreEqu(Head, i, Operand1, Operand2, Operand3, Unresolved) |

         ElementSizeOp      : Folded := FoldElementSize(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         BaseOp             : Folded := FoldBase(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         OffsetOp           : Folded := FoldOffset(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         SizeOp             : Folded := FoldSize(Head, i, Operand1, Operand2, Operand3, Unresolved) |
         HighOp             : Folded := FoldHigh(Head, i, Operand1, Operand2, Operand3, Unresolved)

         ELSE
         END ;
         IF CompilerDebugging AND (NOT Previous) AND Unresolved
         THEN
            WriteString('unresolved symbol found at quad: ') ; WriteCard(i, 0) ; WriteLn
         END ;
         i := GetNextQuad(i)
      END ;
      IF (NOT Folded) AND Unresolved
      THEN
         NeverCalledEvaluate := FALSE ;
         Evaluate
      END
   UNTIL (NOT Folded) AND (NOT Unresolved) ;
   IF NeverCalledEvaluate
   THEN
      Evaluate
   END
END FoldConstants ;


(*
   FoldElementSize - folds the element size for an ArraySym or UnboundedSym.
                     ElementSize returns a constant which defines the
                     multiplier to be multiplied by this element index.
*)

PROCEDURE FoldElementSize (VAR Head: CARDINAL; QuadNo: CARDINAL;
                           Operand1, Operand2, Operand3: CARDINAL;
                           VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsUnbounded(Operand2)
   THEN
      RETURN(
              FoldElementSizeForUnbounded(Head, QuadNo,
                                          Operand1, Operand2, Operand3,
                                          Unresolved)
            )
   ELSIF IsArray(Operand2)
   THEN
      RETURN(
              FoldElementSizeForArray(Head, QuadNo,
                                      Operand1, Operand2, Operand3,
                                      Unresolved)
            )
   ELSE
      InternalError('Expecting UnboundedSym or ArraySym', __FILE__, __LINE__)
   END
END FoldElementSize ;


(*
   FoldElementSizeForArray - attempts to calculate the Subscript
                             multiplier for the index Operand3.
*)

PROCEDURE FoldElementSizeForArray (VAR Head: CARDINAL;
                                   QuadNo: CARDINAL;
                                   Operand1, Operand2, Operand3: CARDINAL;
                                   VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   Subscript: CARDINAL ;
   Type     : CARDINAL ;
BEGIN
   IF IsConst(Operand1) AND (NOT IsValueSolved(Operand1))
   THEN
      Type := Operand2 ;
      IF NoOfElements(Type)>=Operand3
      THEN
         Subscript := GetNth(Type, Operand3) ;
         IF IsSizeSolved(Subscript)
         THEN
            PushSize(Subscript) ;
            PopValue(Operand1) ;
            PutConst(Operand1, Cardinal) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      ELSE
         InternalError('too many elements', __FILE__, __LINE__) ;
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldElementSizeForArray ;


(*
   FoldElementSizeForUnbounded - Unbounded arrays only have one index,
                                 therefore element size will be the
                                 TSIZE(Type) where Type is defined as:
                                 ARRAY OF Type.
*)

PROCEDURE FoldElementSizeForUnbounded (VAR Head: CARDINAL;
                                       QuadNo: CARDINAL;
                                       Operand1, Operand2, Operand3: CARDINAL;
                                       VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   ArrayType: CARDINAL ;
   Type     : CARDINAL ;
BEGIN
   IF IsConst(Operand1)
   THEN
      IF IsValueSolved(Operand1)
      THEN
         InternalError('cannot assign a value twice to a constant', __FILE__, __LINE__)
      ELSE
         ArrayType := Operand2 ;
         Assert(IsUnbounded(ArrayType)) ;
         Assert(Operand3=1) ;
         Assert(NoOfElements(ArrayType)=Operand3) ;
         Type := GetType(ArrayType) ;
         IF IsSizeSolved(Type)
         THEN
            PushSize(Type) ;
            PopValue(Operand1) ;
            PutConst(Operand1, Cardinal) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      END
   ELSE
      RETURN( FALSE )
   END
END FoldElementSizeForUnbounded ;


(*
   FoldSize - removes a SizeOp quadruple if the size of Operand3
              is known. In practice Operand3 should be known
              as at this stage the evaluation of all symbols should
              have taken place.
*)

PROCEDURE FoldSize (VAR Head: CARDINAL; QuadNo: CARDINAL;
                    Operand1, Operand2, Operand3: CARDINAL;
                    VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsConst(Operand1) AND (NOT IsValueSolved(Operand1))
   THEN
      IF IsSizeSolved(Operand3)
      THEN
         PushSize(Operand3) ;
         PopValue(Operand1) ;
         SubQuad(Head, QuadNo) ;
         RETURN( TRUE )
      ELSE
         Unresolved := TRUE ;
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldSize ;


(*
   FoldHigh - if the array is not dynamic then we should be able to
              remove the HighOp quadruple and assign Operand1 with
              the known compile time HIGH(Operand3).
*)

PROCEDURE FoldHigh (VAR Head: CARDINAL; QuadNo: CARDINAL;
                    Operand1, Operand2, Operand3: CARDINAL;
                    VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   Type    : CARDINAL ;
   i       : CARDINAL ;
   Offset  : CARDINAL ;
   High,
   Low     : CARDINAL ;
   Subscript,
   Subrange: CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   Type := GetType(Operand3) ;
   IF NoOfElements(Type)#1
   THEN
      NearToken('HIGH operator only allowed on one dimensional arrays',
               QuadToTokenNo(QuadNo))
   END ;
   Subscript := GetNth(Type, 1) ;
   Subrange := GetType(Subscript) ;
   GetSubrange(Subrange, High, Low) ;
   IF IsConst(Operand1) AND (NOT IsValueSolved(Operand1))
   THEN
      IF IsSizeSolved(Operand3) AND IsValueSolved(High)
      THEN
         NearToken('', GetDeclared(High)) ;
         PushValue(High) ;
         PopValue(Operand1) ;
         PutConst(Operand1, Cardinal) ;
         SubQuad(Head, QuadNo) ;
         RETURN( TRUE )
      ELSE
         Unresolved := TRUE ;
         RETURN( FALSE )
      END
   ELSIF IsSizeSolved(Operand3) AND IsValueSolved(High)
   THEN
      PushValue(High) ;
      PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
      RETURN( TRUE )
   ELSE
      RETURN( FALSE )
   END
END FoldHigh ;


(*
   CodeBase - Operand1 is a constant and BaseOp will calculate the offset
              of the virtual start of the array  ie a[0,0,0,0..,0]
              from the address of the array &a.

              Operand2 is the type of the array
              Operand3 is the array
*)

PROCEDURE FoldBase (VAR Head: CARDINAL; QuadNo: CARDINAL;
                    Operand1, Operand2, Operand3: CARDINAL;
                    VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   i       : CARDINAL ;
   Offset  : INTEGER ;
   High,
   Low     : CARDINAL ;
   Subscript,
   Subrange: CARDINAL ;
BEGIN
   IF IsConst(Operand1)
   THEN
      IF NOT IsValueSolved(Operand1)
      THEN
         i := 1 ;
         Offset := 0 ;
         Subscript := GetNth(Operand2, i) ;
         WHILE Subscript#NulSym DO
            IF IsSizeSolved(Subscript)
            THEN
               PushSize(Subscript) ;
               Subrange := GetType(Subscript) ;
               GetSubrange(Subrange, High, Low) ;
               IF IsValueSolved(Low)
               THEN
                  PushValue(Low) ;
                  Mult ;
                  DEC(Offset, PopInt()) ;
                  INC(i) ;
                  Subscript := GetNth(Operand2, i)
               ELSE
                  i := PopInt() ;  (* throw away working value *)
                  Unresolved := TRUE ;
                  RETURN( FALSE )
               END
            ELSE
               Unresolved := TRUE ;
               RETURN( FALSE )
            END
         END ;
         PushInt(Offset) ;
         PopValue(Operand1) ;
         SubQuad(Head, QuadNo) ;
         RETURN( TRUE )
      END
   ELSE
      InternalError('expecting constant operand1 for BaseOp', __FILE__, __LINE__)
   END ;
   RETURN( FALSE )
END FoldBase ;


PROCEDURE FoldOffset (VAR Head: CARDINAL; QuadNo: CARDINAL;
                      Operand1, Operand2, Operand3: CARDINAL;
                      VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsConst(Operand1)
   THEN
      IF IsValueSolved(Operand1)
      THEN
         InternalError('cannot assign a value twice to a constant', __FILE__, __LINE__)
      ELSE
         IF IsOffsetSolved(Operand3)
         THEN
            PushOffset(Operand3) ;
            PopValue(Operand1) ;
            PutConst(Operand1, Address) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      END
   ELSE
      RETURN( FALSE )
   END
END FoldOffset ;


(*
   FoldOr - attempts to fold the logical OR of Operand2 and Operand3.
            TRUE is returned if the folding was achieved.
*)

PROCEDURE FoldOr (VAR Head: CARDINAL; QuadNo: CARDINAL;
                  Operand1, Operand2, Operand3: CARDINAL;
                  VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   Type   : CARDINAL ;
   Op1Name,
   Op2Name,
   Op3Name: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2) AND
      IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand1)
      THEN
         InternalError('cannot assign a value twice to a constant', __FILE__, __LINE__)
      ELSE
         IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
         THEN
            NearToken('', QuadToTokenNo(QuadNo)) ;
            PushValue(Operand2) ;
            PushValue(Operand3) ;
            Type := MixTypes(GetType(Operand3), GetType(Operand2)) ;
            SetOr ;
            PopValue(Operand1) ;
            PutConst(Operand1, Type) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      END
   ELSIF IsConst(Operand2) AND IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
      THEN
         PushValue(Operand2) ;
         PushValue(Operand3) ;
         SetOr ;
         PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldOr ;


(*
   FoldAnd - attempts to fold the logical OR of Operand2 and Operand3.
             TRUE is returned if the folding was achieved.
*)

PROCEDURE FoldAnd (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL;
                   VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   Type   : CARDINAL ;
   Op1Name,
   Op2Name,
   Op3Name: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2) AND
      IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand1)
      THEN
         InternalError('cannot assign a value twice to a constant', __FILE__, __LINE__)
      ELSE
         IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
         THEN
            NearToken('', QuadToTokenNo(QuadNo)) ;
            PushValue(Operand2) ;
            PushValue(Operand3) ;
            Type := MixTypes(GetType(Operand3), GetType(Operand2)) ;
            SetAnd ;
            PopValue(Operand1) ;
            PutConst(Operand1, Type) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      END
   ELSIF IsConst(Operand2) AND IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
      THEN
         PushValue(Operand2) ;
         PushValue(Operand3) ;
         SetAnd ;
         PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldAnd ;


(*
   FoldXor - attempts to fold the logical XOR of Operand2 and Operand3.
             TRUE is returned if the folding was achieved.
*)

PROCEDURE FoldXor (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL;
                   VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   Type   : CARDINAL ;
   Op1Name,
   Op2Name,
   Op3Name: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2) AND
      IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand1)
      THEN
         InternalError('cannot assign a value twice to a constant', __FILE__, __LINE__)
      ELSE
         IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
         THEN
            NearToken('', QuadToTokenNo(QuadNo)) ;
            PushValue(Operand2) ;
            PushValue(Operand3) ;
            Type := MixTypes(GetType(Operand3), GetType(Operand2)) ;
            SetSymmetricDifference ;
            PopValue(Operand1) ;
            PutConst(Operand1, Type) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      END
   ELSIF IsConst(Operand2) AND IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
      THEN
         PushValue(Operand2) ;
         PushValue(Operand3) ;
         SetSymmetricDifference ;
         PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldXor ;


(*
   FoldSetDifference - attempts to fold the logical set difference of Operand2 and Operand3.
                       TRUE is returned if the folding was achieved.
*)

PROCEDURE FoldSetDifference (VAR Head: CARDINAL; QuadNo: CARDINAL;
                             Operand1, Operand2, Operand3: CARDINAL;
                             VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   Type   : CARDINAL ;
   Op1Name,
   Op2Name,
   Op3Name: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2) AND
      IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand1)
      THEN
         InternalError('cannot assign a value twice to a constant', __FILE__, __LINE__)
      ELSE
         IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
         THEN
            NearToken('', QuadToTokenNo(QuadNo)) ;
            PushValue(Operand2) ;
            PushValue(Operand3) ;
            Type := MixTypes(GetType(Operand3), GetType(Operand2)) ;
            SetDifference ;
            PopValue(Operand1) ;
            PutConst(Operand1, Type) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      END
   ELSIF IsConst(Operand2) AND IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
      THEN
         PushValue(Operand2) ;
         PushValue(Operand3) ;
         SetDifference ;
         PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldSetDifference ;


(*
   FoldBit - attempts to fold the operation:

             BitOp  Operand1 _ Operand3.

             TRUE is returned if the folding was achieved.
             (Note that folding should always be successful with BitOp).
*)

PROCEDURE FoldBit (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand3: CARDINAL;
                   VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand3)
   THEN
      IF NOT IsValueSolved(Operand1)
      THEN
         IF IsValueSolved(Operand3)
         THEN
            PushValue(Operand3) ;
            Bit ;
            PopValue(Operand1) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      ELSE
         InternalError('this quadruple should have been deleted', __FILE__, __LINE__)
      END
   ELSIF IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand3)
      THEN
         PushValue(Operand3) ;
         Bit ;
         PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   END ;
   RETURN( FALSE )
END FoldBit ;


(*
   FoldBitRange - attempts to fold the operation:

                  BitRangeOp  Operand1 Operand2 Operand3.

                  TRUE is returned if the folding was achieved.
                  (Note that folding should always be successful with
                   BitRangeOp).
*)

PROCEDURE FoldBitRange (VAR Head: CARDINAL; QuadNo: CARDINAL;
                        Operand1, Operand2, Operand3: CARDINAL;
                        VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand3)
   THEN
      IF NOT IsValueSolved(Operand1)
      THEN
         IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
         THEN
            PushValue(Operand2) ;
            PushValue(Operand3) ;
            BitRange ;
            PopValue(Operand1) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      ELSE
         InternalError('this quadruple should have been deleted', __FILE__, __LINE__)
      END
   ELSIF IsConst(Operand2) AND IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
      THEN
         PushValue(Operand2) ;
         PushValue(Operand3) ;
         BitRange ;
         PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   END ;
   RETURN( FALSE )
END FoldBitRange ;


(*
   FoldNegate - attempts to fold the negation of Operand3.
                TRUE is returned if the folding was achieved.
*)

PROCEDURE FoldNegate (VAR Head: CARDINAL; QuadNo: CARDINAL;
                      Operand1, Operand2, Operand3: CARDINAL;
                      VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand1)
      THEN
         InternalError('cannot assign a value twice to a constant', __FILE__, __LINE__)
      ELSE
         IF IsValueSolved(Operand3)
         THEN
            PushCard(0) ;
            PushValue(Operand3) ;
            Sub ;
            PopValue(Operand1) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      END
   ELSIF IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand3)
      THEN
         PushCard(0) ;
         PushValue(Operand3) ;
         Sub ;
         PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldNegate ;


(*
   FoldAdd - attempts to fold the addition of Operand2 and Operand3.
             TRUE is returned if the folding was achieved.
*)

PROCEDURE FoldAdd (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL;
                   VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   Type   : CARDINAL ;
   Op1Name,
   Op2Name,
   Op3Name: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2) AND
      IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand1)
      THEN
         InternalError('cannot assign a value twice to a constant', __FILE__, __LINE__)
      ELSE
         IF IsConstString(Operand2) AND IsConstString(Operand3)
         THEN
            GetKey(GetString(Operand2), Op2Name) ;
            GetKey(GetString(Operand3), Op3Name) ;
            StrConCat(Op2Name, Op3Name, Op1Name) ;
            PutConstString(Operand1, MakeKey(Op1Name)) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSIF IsConstString(Operand2) OR IsConstString(Operand3)
         THEN
            NearToken('cannot mix numerical and string addition',
                      QuadToTokenNo(QuadNo))
         ELSIF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
         THEN
            NearToken('', QuadToTokenNo(QuadNo)) ;
            PushValue(Operand2) ;
            PushValue(Operand3) ;
            Type := MixTypes(GetType(Operand3), GetType(Operand2)) ;
            IF (Type#NulSym) AND IsSet(Type)
            THEN
               InternalError('why does Add perform Or in the quadruples?', __FILE__, __LINE__) ;
               SetOr
            ELSE
               Add
            END ;
            PopValue(Operand1) ;
            PutConst(Operand1, Type) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      END
   ELSIF IsConst(Operand2) AND IsConst(Operand3) AND
         (NOT IsConstString(Operand2)) AND (NOT IsConstString(Operand3))
   THEN
      IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
      THEN
         PushValue(Operand2) ;
         PushValue(Operand3) ;
         Add ;
         PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldAdd ;


(*
   FoldSub - attempts to fold the subtraction of Operand2 and Operand3.
             TRUE is returned if the folding was achieved.
*)

PROCEDURE FoldSub (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL;
                   VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   Type: CARDINAL ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2) AND
      IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand1)
      THEN
         InternalError('cannot assign a value twice to a constant', __FILE__, __LINE__)
      ELSE
         IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
         THEN
            NearToken('', QuadToTokenNo(QuadNo)) ;
            PushValue(Operand2) ;
            PushValue(Operand3) ;
            Type := MixTypes(GetType(Operand3), GetType(Operand2)) ;
            IF (Type#NulSym) AND IsSet(Type)
            THEN
               InternalError('why does Sub perform XOr in the quadruples?', __FILE__, __LINE__) ;
               SetDifference
            ELSE
               Sub
            END ;
            PopValue(Operand1) ;
            PutConst(Operand1, MixTypes(GetType(Operand3), GetType(Operand2))) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      END
   ELSIF IsConst(Operand2) AND IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
      THEN
         PushValue(Operand2) ;
         PushValue(Operand3) ;
         Sub ;
         PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldSub ;


(*
   FoldMult - attempts to fold the multiplication of Operand2 and Operand3.
              TRUE is returned if the folding was achieved.
*)

PROCEDURE FoldMult (VAR Head: CARDINAL; QuadNo: CARDINAL;
                    Operand1, Operand2, Operand3: CARDINAL;
                    VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   Type: CARDINAL ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2) AND
      IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand1)
      THEN
         InternalError('cannot assign a value twice to a constant', __FILE__, __LINE__)
      ELSE
         IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
         THEN
            NearToken('', QuadToTokenNo(QuadNo)) ;
            PushValue(Operand2) ;
            PushValue(Operand3) ;
            Type := MixTypes(GetType(Operand3), GetType(Operand2)) ;
            IF (Type#NulSym) AND IsSet(Type)
            THEN
               InternalError('why does mult perform and in the quadruples?', __FILE__, __LINE__) ;
               SetAnd
            ELSE
               Mult
            END ;
            PopValue(Operand1) ;
            PutConst(Operand1, Type) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      END
   ELSIF IsConst(Operand2) AND IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
      THEN
         PushValue(Operand2) ;
         PushValue(Operand3) ;
         Mult ;
         PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldMult ;


(*
   FoldDiv - attempts to fold the addition of Operand2 and Operand3.
             TRUE is returned if the folding was achieved.
*)

PROCEDURE FoldDiv (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL;
                   VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   Type: CARDINAL ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2) AND
      IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand1)
      THEN
         InternalError('cannot assign a value twice to a constant', __FILE__, __LINE__)
      ELSE
         IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
         THEN
            NearToken('', QuadToTokenNo(QuadNo)) ;
            PushValue(Operand2) ;
            PushValue(Operand3) ;
            Type := MixTypes(GetType(Operand3), GetType(Operand2)) ;
            IF (Type#NulSym) AND IsSet(Type)
            THEN
               InternalError('why does div perform xor in the quadruples?', __FILE__, __LINE__) ;
               SetSymmetricDifference
            ELSE
               Div
            END ;
            PopValue(Operand1) ;
            PutConst(Operand1, MixTypes(GetType(Operand3), GetType(Operand2))) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      END
   ELSIF IsConst(Operand2) AND IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
      THEN
         PushValue(Operand2) ;
         PushValue(Operand3) ;
         Div ;
         PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldDiv ;


(*
   FoldMod - attempts to fold the modulus of Operand2 and Operand3.
             TRUE is returned if the folding was achieved.
*)

PROCEDURE FoldMod (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL;
                   VAR Unresolved: BOOLEAN) : BOOLEAN ;
VAR
   Type: CARDINAL ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2) AND
      IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand1)
      THEN
         InternalError('cannot assign a value twice to a constant', __FILE__, __LINE__)
      ELSE
         IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
         THEN
            NearToken('', QuadToTokenNo(QuadNo)) ;
            PushValue(Operand2) ;
            PushValue(Operand3) ;
            Type := MixTypes(GetType(Operand3), GetType(Operand2)) ;
            IF (Type#NulSym) AND IsSet(Type)
            THEN
               InternalError('modulus on sets makes no sense', __FILE__, __LINE__)
            ELSE
               Mod
            END ;
            PopValue(Operand1) ;
            PutConst(Operand1, MixTypes(GetType(Operand3), GetType(Operand2))) ;
            SubQuad(Head, QuadNo) ;
            RETURN( TRUE )
         ELSE
            Unresolved := TRUE ;
            RETURN( FALSE )
         END
      END
   ELSIF IsConst(Operand2) AND IsConst(Operand3)
   THEN
      IF IsValueSolved(Operand2) AND IsValueSolved(Operand3)
      THEN
         PushValue(Operand2) ;
         PushValue(Operand3) ;
         Mod ;
         PutQuad(QuadNo, BecomesOp, Operand1, NulSym, MakeNewConstFromValue()) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldMod ;


(*
   FoldBecomes - attempts to fold Operand1 := Operand3.
                 It will fold string assignment and numerical
                 assignment, but not a mixture of both!
*)

PROCEDURE FoldBecomes (VAR Head: CARDINAL; QuadNo: CARDINAL;
                       Operand1, Operand3: CARDINAL;
                       VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand3)
   THEN
      IF IsConstString(Operand3)
      THEN
         PutConstString(Operand1, GetString(Operand3)) ;
         SubQuad(Head, QuadNo) ;
         RETURN( TRUE )
      ELSIF IsValueSolved(Operand1)
      THEN
         NearTokens('constant is declared with a value and later reassigned a new value',
                   QuadToTokenNo(QuadNo), GetDeclared(Operand1)) ;
      ELSIF IsValueSolved(Operand3)
      THEN
         PushValue(Operand3) ;
         PopValue(Operand1) ;
         IF GetType(Operand1)=NulSym
         THEN
            PutConst(Operand1, GetType(Operand3))
         END ;
         SubQuad(Head, QuadNo) ;
         RETURN( TRUE )
      ELSE
         Unresolved := TRUE ;
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldBecomes ;


PROCEDURE FoldEqu (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL;
                   VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      IF IsValueSolved(Operand1) AND IsValueSolved(Operand2)
      THEN
         PushValue(Operand1) ;
         PushValue(Operand2) ;
         IF Equ()
         THEN
            PutQuad(QuadNo, GotoOp, NulSym, NulSym, Operand3)
         ELSE
            SubQuad(Head, QuadNo)
         END ;
         RETURN( TRUE )
      ELSE
         Unresolved := TRUE ;
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldEqu ;


PROCEDURE FoldNotEqu (VAR Head: CARDINAL; QuadNo: CARDINAL;
                      Operand1, Operand2, Operand3: CARDINAL;
                      VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      IF IsValueSolved(Operand1) AND IsValueSolved(Operand2)
      THEN
         PushValue(Operand1) ;
         PushValue(Operand2) ;
         IF NotEqu()
         THEN
            PutQuad(QuadNo, GotoOp, NulSym, NulSym, Operand3)
         ELSE
            SubQuad(Head, QuadNo)
         END ;
         RETURN( TRUE )
      ELSE
         Unresolved := TRUE ;
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldNotEqu ;


PROCEDURE FoldLess (VAR Head: CARDINAL; QuadNo: CARDINAL;
                    Operand1, Operand2, Operand3: CARDINAL;
                    VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      IF IsValueSolved(Operand1) AND IsValueSolved(Operand2)
      THEN
         PushValue(Operand1) ;
         PushValue(Operand2) ;
         IF Less()
         THEN
            PutQuad(QuadNo, GotoOp, NulSym, NulSym, Operand3)
         ELSE
            SubQuad(Head, QuadNo)
         END ;
         RETURN( TRUE )
      ELSE
         Unresolved := TRUE ;
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldLess ;


PROCEDURE FoldGre (VAR Head: CARDINAL; QuadNo: CARDINAL;
                   Operand1, Operand2, Operand3: CARDINAL;
                   VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      IF IsValueSolved(Operand1) AND IsValueSolved(Operand2)
      THEN
         PushValue(Operand1) ;
         PushValue(Operand2) ;
         IF Gre()
         THEN
            PutQuad(QuadNo, GotoOp, NulSym, NulSym, Operand3)
         ELSE
            SubQuad(Head, QuadNo)
         END ;
         RETURN( TRUE )
      ELSE
         Unresolved := TRUE ;
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldGre ;


PROCEDURE FoldLessEqu (VAR Head: CARDINAL; QuadNo: CARDINAL;
                       Operand1, Operand2, Operand3: CARDINAL;
                       VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      IF IsValueSolved(Operand1) AND IsValueSolved(Operand2)
      THEN
         PushValue(Operand1) ;
         PushValue(Operand2) ;
         IF LessEqu()
         THEN
            PutQuad(QuadNo, GotoOp, NulSym, NulSym, Operand3)
         ELSE
            SubQuad(Head, QuadNo)
         END ;
         RETURN( TRUE )
      ELSE
         Unresolved := TRUE ;
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldLessEqu ;


PROCEDURE FoldGreEqu (VAR Head: CARDINAL; QuadNo: CARDINAL;
                      Operand1, Operand2, Operand3: CARDINAL;
                      VAR Unresolved: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      IF IsValueSolved(Operand1) AND IsValueSolved(Operand2)
      THEN
         PushValue(Operand1) ;
         PushValue(Operand2) ;
         IF GreEqu()
         THEN
            PutQuad(QuadNo, GotoOp, NulSym, NulSym, Operand3)
         ELSE
            SubQuad(Head, QuadNo)
         END ;
         RETURN( TRUE )
      ELSE
         Unresolved := TRUE ;
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END FoldGreEqu ;


(*
   FoldBranches - folds unneccessary branches in the list of quadruples.
                  It searches for the following patterns:

                  [x]  GotoOp  _   _   y                GotoOp   _   _   z
                  ...                                   ...
                  [y]  GotoOp  _   _   z      "deleted"

                  WHERE ... may contain 0..n Pseudo Quads


                  OR


                  [x]  IfREL   _   _   z      If NOT REL  _   _   a
                  ...                         ...
                  [y]  Goto    _   _   a      "deleted"
                  ...                         ...
                  [z]


                  WHERE ... may contain 0..n Pseudo Quads
                  but in this case they must not be a
                      target of any other quad.
*)

PROCEDURE FoldBranches (VAR Head: CARDINAL) ;
VAR
   Folded     : BOOLEAN ;
   i,
   Right      : CARDINAL ;
   Operator   : QuadOperator ;
   Operand1,
   Operand2,
   Operand3   : CARDINAL ;
BEGIN
   REPEAT
      i := Head ;
      Folded := FALSE ;
      WHILE i#0 DO
         Right := GetRealQuad(GetNextQuad(i)) ;
         GetQuad(i, Operator, Operand1, Operand2, Operand3) ;
         CASE Operator OF

         GotoOp             : Folded := ReduceGoto(Head, i, Operand3,
                                                   Right, Folded) |
         IfInOp, IfNotInOp,
         IfNotEquOp, IfEquOp,
         IfLessEquOp, IfGreEquOp,
         IfGreOp, IfLessOp  : Folded := ReduceBranch(Head, Operator, i,
                                                     Operand1, Operand2, Operand3,
                                                     Right, Folded)

         ELSE
         END ;
         i := Right
      END
   UNTIL NOT Folded
END FoldBranches ;



(*
   ReduceBranch - searches for the following pattern:

                  [x]  IfREL   _   _   z      If NOT REL  _   _   a
                  ...                         ...
                  [y]  Goto    _   _   a      "deleted"
                  ...                         ...
                  [z]


                  WHERE ... may contain 0..n Pseudo Quads
                  but in this case they must not be a
                      target of any other quad.

*)

PROCEDURE ReduceBranch (VAR Head: CARDINAL;
                        Operator: QuadOperator;
                        CurrentQuad,
                        CurrentOperand1, CurrentOperand2,
                        CurrentOperand3: CARDINAL;
                        VAR NextQuad: CARDINAL;
                        Folded: BOOLEAN) : BOOLEAN ;
VAR
   OpNext : QuadOperator ;
   NextPlusOne,
   Op1Next,
   Op2Next,
   Op3Next,
   From,
   To     : CARDINAL ;
BEGIN
   (* If op         NextQuad+1 *)
   (* Goto          x          *)

   IF NextQuad#0
   THEN
      IF (GetNextQuad(CurrentQuad)=CurrentOperand3) OR
         (GetRealQuad(GetNextQuad(CurrentQuad))=CurrentOperand3)
      THEN
         SubQuad(Head, CurrentQuad) ;
         Folded := TRUE
      ELSE
         From := GetNextQuad(CurrentQuad) ;  (* start after CurrentQuad *)
         To := NextQuad ;
         CurrentOperand3 := GetRealQuad(CurrentOperand3) ;

         NextPlusOne := GetRealQuad(GetNextQuad(NextQuad)) ;
         GetQuad(NextQuad, OpNext, Op1Next, Op2Next, Op3Next) ;
         IF (OpNext=GotoOp) AND (NextPlusOne=CurrentOperand3) AND
            IsBasicBlock(Head, From, To)
         THEN
            (*       Op3Next := GetRealQuad(Op3Next) ; *)
            SubQuad(Head, NextQuad) ;
            PutQuad(CurrentQuad, Opposite(Operator),
                    CurrentOperand1, CurrentOperand2, Op3Next) ;
            NextQuad := NextPlusOne ;
            Folded := TRUE
         END
      END ;
      IF FoldMultipleGoto(CurrentQuad)
      THEN
         Folded := TRUE
      END
   END ;
   RETURN( Folded )
END ReduceBranch ;


(*
   IsBasicBlock - returns TRUE if no other quadruple jumps inbetween
                  the range From..To.
                  It assumes that there are no jumps in the quadruples
                  From..To.
*)

PROCEDURE IsBasicBlock (Head, From, To: CARDINAL) : BOOLEAN ;
BEGIN
   WHILE From#To DO
      IF IsReferenced(Head, From)
      THEN
         RETURN( FALSE )
      ELSE
         IF From>To
         THEN
            InternalError('Optimization: assert failed From should NEVER be larger than To', __FILE__, __LINE__)
         END ;
         From := GetNextQuad(From)
      END
   END ;
   RETURN( TRUE )
END IsBasicBlock ;


(*
   ReduceGoto - searches for the following patterns:

                [x]  GotoOp  _   _   y                GotoOp   _   _   z
                  ...                                   ...
                [y]  GotoOp  _   _   z      "deleted"


*)

PROCEDURE ReduceGoto (VAR Head: CARDINAL;
                      CurrentQuad, CurrentOperand3, NextQuad: CARDINAL;
                      Folded: BOOLEAN) : BOOLEAN ;
BEGIN
   CurrentOperand3 := GetRealQuad(CurrentOperand3) ;
   (* IF next quad is a GotoOp *)
   IF CurrentOperand3=NextQuad
   THEN
      SubQuad(Head, CurrentQuad) ;
      Folded := TRUE
   ELSE
      (* Does Goto point to another Goto ? *)
      IF FoldMultipleGoto(CurrentQuad)
      THEN
         Folded := TRUE
      END
   END ;
   RETURN( Folded )
END ReduceGoto ;


(*
   FoldMultipleGoto - takes a QuadNo and if it jumps to another GotoOp
                      then it takes the later target as a replacement
                      for its own.

                      NOTE  it does not remove any quadruples.
*)

PROCEDURE FoldMultipleGoto (QuadNo: CARDINAL) : BOOLEAN ;
VAR
   Operator,
   Op      : QuadOperator ;
   Op1, Op2,
   Op3,
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(QuadNo, Operator, Operand1, Operand2, Operand3) ;
   Operand3 := GetRealQuad(Operand3) ;  (* skip pseudo quadruples *)
   GetQuad(Operand3, Op, Op1, Op2, Op3) ;
   IF Op=GotoOp
   THEN
      PutQuad(QuadNo, Operator, Operand1, Operand2, Op3) ;
      (* Dont want success to be returned if in fact the Goto *)
      (* line number has not changed... otherwise we loop     *)
      (* forever.                                             *)
      RETURN( Op3#Operand3 )
   ELSE
      RETURN( FALSE )
   END
END FoldMultipleGoto ;


(*
   RemoveProcedures - removes any procedures that are never referenced
                      by the quadruples. If any reduction of quadruples
                      has occurred then true is returned.
*)

PROCEDURE RemoveProcedures (VAR Head: CARDINAL) ;
VAR
   n,
   Start,
   End,
   Module: CARDINAL ;
BEGIN
   (* DisplayReachable ; *)
   n := 1 ;
   REPEAT
      Module := GetModuleNo(n) ;
      (* Should be at least one module *)
      IF Module#NulSym
      THEN
         (* WriteString('Module ') ; WriteKey(GetSymName(Module)) ; *)
         GetModuleQuads(Module, Start, End) ;
         IF Start#End
         THEN
            (* Module does have some initialization code *)
            (*
            WriteString(' has initialization code  ') ;
            WriteCard(Start, 0) ; WriteString('..') ; WriteCard(End, 0) ;
            WriteLn ;
            *)
            KnownReachableInitCode(Start, End)
         ELSE
            (* WriteString(' has no initialization code') ; WriteLn *)
         END ;
         INC(n)
      END
   UNTIL Module=NulSym ;
   MakeExportedProceduresReachable ;
   DeleteUnReachableProcedures(Head)
END RemoveProcedures ;


(*
   MakeExportedProceduresReachable - make all exported procedures
                                     reachable
*)

PROCEDURE MakeExportedProceduresReachable ;
VAR
   Start,
   End,
   n, m,
   Module,
   Proc  : CARDINAL ;
BEGIN
   m := 1 ;
   REPEAT
      Module := GetModuleNo(m) ;
      IF Module#NulSym
      THEN
         n := 1 ;
         Proc := GetNthProcedure(Module, n) ;
         WHILE Proc#NulSym DO
            IF (IsExportQualified(Proc) OR IsExportUnQualified(Proc)) AND
               (NOT IsProcedureReachable(Proc))
            THEN
               PutProcedureReachable(Proc) ;
               GetProcedureQuads(Proc, Start, End) ;
               KnownReachable(Start, End)
            END ;
            INC(n) ;
            Proc := GetNthProcedure(Module, n)
         END ;
         INC(m)
      END
   UNTIL Module=NulSym
END MakeExportedProceduresReachable ;


(*
   KnownReachableInitCode - All quadruples between "Start..End-1" and
                            non procedural code are reachable.
*)

PROCEDURE KnownReachableInitCode(Start, End: CARDINAL) ;
VAR
   Op           : QuadOperator ;
   Op1, Op2, Op3: CARDINAL ;
BEGIN
   IF Start#0
   THEN
      REPEAT
         GetQuad(Start, Op, Op1, Op2, Op3) ;
         IF Op=NewLocalVarOp
         THEN
            (* Skip all procedural code of a Modules Initialization Code *)
            REPEAT
               Start := GetNextQuad(Start) ;
               GetQuad(Start, Op, Op1, Op2, Op3)
            UNTIL (Start=End) OR (Op=ReturnOp)
         ELSE
            (* Can be sure that quads are at a base level Zero, outside procedures *)
            CASE Op OF

            AddrOp,
            CallOp,
            ParamOp,
            XIndrOp,
            BecomesOp: KnownReach(Op3)

            ELSE
            END ;
            IF Op=DummyOp
            THEN
               (* Start been deleted therefore just increment *)
               INC(Start)
            ELSE
               (* Follow the quad list *)
               Start := GetNextQuad(Start)
            END
         END
      UNTIL Start=End
   END
END KnownReachableInitCode ;


PROCEDURE KnownReachable (Start, End: CARDINAL) ;
VAR
   Op           : QuadOperator ;
   Op1, Op2, Op3: CARDINAL ;
BEGIN
   IF Start#0
   THEN
      REPEAT
         GetQuad(Start, Op, Op1, Op2, Op3) ;
         CASE Op OF

         AddrOp,
         CallOp,
         ParamOp,
         XIndrOp,
         BecomesOp: KnownReach(Op3)

         ELSE
         END ;
         IF Op=DummyOp
         THEN
            (* Start been deleted therefore just increment *)
            INC(Start)
         ELSE
            (* Follow the quad list *)
            Start := GetNextQuad(Start)
         END
      UNTIL Start=End
   END
END KnownReachable ;


PROCEDURE KnownReach (Sym: CARDINAL) ;
VAR
   Start,
   End  : CARDINAL ;
BEGIN
   IF IsProcedure(Sym)
   THEN
      IF NOT IsProcedureReachable(Sym)
      THEN
         (* Not reachable at present - therefore make it reachable *)
         PutProcedureReachable(Sym) ;
         GetProcedureQuads(Sym, Start, End) ;
         KnownReachable(Start, End)
      END
   END
END KnownReach ;


(*
   DeleteUnReachableProcedures - Deletes all procedures that are unreachable.
*)

PROCEDURE DeleteUnReachableProcedures (VAR Head: CARDINAL) ;
VAR
   n, m,
   Start,
   End,
   Module,
   Proc  : CARDINAL ;
BEGIN
   m := 1 ;
   REPEAT
      Module := GetModuleNo(m) ;
      IF Module#NulSym
      THEN
         n := 1 ;
         Proc := GetNthProcedure(Module, n) ;
         WHILE Proc#NulSym DO
            IF IsProcedureReachable(Proc) OR
               IsExportQualified(Proc) OR IsExportUnQualified(Proc)
            THEN
               (* is reachable - do not delete it *)
            ELSE
               IF NOT Quiet
               THEN
                  WriteString('[') ; WriteKey(GetSymName(Proc)) ;
                  WriteString(']') ; WriteLn
               END ;

               GetProcedureQuads(Proc, Start, End) ;
               IF Start#0
               THEN
                  Delete(Head, Start, End) ;
                  (* No Longer any Quads for this Procedure *)
                  PutProcedureStartQuad(Proc, 0) ;
                  PutProcedureEndQuad(Proc, 0)
               END
            END ;
            INC(n) ;
            Proc := GetNthProcedure(Module, n)
         END ;
         INC(m)
      END
   UNTIL Module=NulSym
END DeleteUnReachableProcedures ;


(*
   Delete - deletes all quadruples from Start..End
            or the end of the procedure.
*)

PROCEDURE Delete (VAR Head: CARDINAL; Start, End: CARDINAL) ;
VAR
   Last,
   i   : CARDINAL ;
   Op  : QuadOperator ;
   Op1,
   Op2,
   Op3 : CARDINAL ;
BEGIN
   Last := GetNextQuad(End) ;
   WHILE (Head#0) AND (Start#0) AND (Last#Start) DO
      GetQuad(Start, Op, Op1, Op2, Op3) ;
      IF Op=DummyOp
      THEN
         (* Start has already been deleted - try next quad *)
         INC(Start)
      ELSIF Op=ReturnOp
      THEN
         (* Found end of procedure therefore just delete and exit *)
         (* WriteString('Deleting') ; WriteCard(Start, 6) ; WriteLn ; *)
         SubQuad(Head, Start) ;
         Start := Last
      ELSE
         (* Following the list of quadruples to the End *)
         i := GetNextQuad(Start) ;
         (* WriteString('Deleting') ; WriteCard(Start, 6) ; WriteLn ; *)
         SubQuad(Head, Start) ;
         Start := i
      END
   END
END Delete ;


(*
   DisplayReachable - Displays the data structures surrounding Reachablity.
*)

PROCEDURE DisplayReachable ;
VAR
   n, m,
   Start,
   End,
   Module,
   Proc  : CARDINAL ;
BEGIN
   m := 1 ;
   REPEAT
      Module := GetModuleNo(m) ;
      IF Module#NulSym
      THEN
         WriteString('Module') ; WriteCard(m, 3) ; WriteKey(GetSymName(Module)) ;
         GetModuleQuads(Module, Start, End) ; WriteString(' Reachable ') ;
         WriteCard(Start, 4) ; WriteCard(End, 4) ; WriteLn ;
         n := 1 ;
         Proc := GetNthProcedure(Module, n) ;
         WHILE Proc#NulSym DO
            WriteString('Procedure ') ; WriteKey(GetSymName(Proc)) ;
            GetProcedureQuads(Proc, Start, End) ;
            WriteString(' Quads: ') ; WriteCard(Start, 6) ; WriteCard(End, 6) ;
            IF NOT IsProcedureReachable(Proc)
            THEN
               WriteString(' UN reachable')
            ELSE
               WriteString(' IS reachable')
            END ;
            WriteLn ;
            INC(n) ;
            Proc := GetNthProcedure(Module, n)
         END ;
         INC(m)
      END
   UNTIL Module=NulSym
END DisplayReachable ;


END M2Optimize.
