IMPLEMENTATION MODULE M2Search ;


FROM FIO IMPORT Exists ;
FROM StrLib IMPORT StrEqual, StrCopy, StrConCat, StrLen ;
FROM StdIO IMPORT Write ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM CmdArgs IMPORT Narg, GetArg ;
FROM M2Defaults IMPORT GetSearchPath ;
FROM M2Configure IMPORT DefaultLibraryPath ;

 
CONST
   MaxStringLen =   79 ;
   Directory    =  '/' ;
 
 
VAR
   Main      : ARRAY [0..MaxStringLen] OF CHAR ;
   SearchPath: ARRAY [0..1000] OF CHAR ;

(* %%%FORWARD%%%
PROCEDURE FindSourceFile (FileName: ARRAY OF CHAR;
                          VAR FullPath: ARRAY OF CHAR) : BOOLEAN ; FORWARD ;
PROCEDURE WriteFileName (FileName: ARRAY OF CHAR) ; FORWARD ;
PROCEDURE SetSearchPath (Path: ARRAY OF CHAR) ; FORWARD ;
PROCEDURE Init ; FORWARD ;
   %%%FORWARD%%% *)

(*
   FindSourceFile - attempts to locate the source file FileName.
                    If a file is found then TRUE is returned otherwise
                    FALSE is returned.
                    The parameter FullPath is set which indicates the
                    absolute location of source FileName.
                    
*)

PROCEDURE FindSourceFile (FileName: ARRAY OF CHAR;
                          VAR FullPath: ARRAY OF CHAR) : BOOLEAN ;
VAR
   i, n: CARDINAL ;
BEGIN
   n := Narg(SearchPath) ;
   i := 0 ;
   WHILE i<n DO
      IF GetArg(SearchPath, i, FullPath)
      THEN
      	 IF StrEqual('.', FullPath)
      	 THEN
      	    StrCopy(FileName, FullPath)
      	 ELSE
            StrConCat(FullPath, Directory, FullPath) ;
            StrConCat(FullPath, FileName, FullPath)
      	 END ;
         IF Exists(FullPath)
         THEN
            RETURN( TRUE )
         END
      END ;
      INC(i)
   END ;
   RETURN( FALSE )
END FindSourceFile ;


(*
   WriteFileName - displays a file name, FileName, with formatted spaces
                   after the string.
*)

PROCEDURE WriteFileName (FileName: ARRAY OF CHAR) ;
CONST
   MaxSpaces = 20 ;
VAR
   i, High: CARDINAL ;
BEGIN
   WriteString(FileName) ;
   High := StrLen(FileName) ;
   IF High<MaxSpaces
   THEN
      High := MaxSpaces - High ;
      i := 0 ;
      WHILE i<High DO
         Write(' ') ;
         INC(i)
      END
   END
END WriteFileName ;


(*
   SetSearchPath - assigns the search path to Path.
                   The string Path may take the form:

                   Path           ::= IndividualPath { " " IndividualPath }
                   IndividualPath ::= "." | DirectoryPath
                   DirectoryPath  ::= Name { "/" Name }
                   Name           ::= Letter { (Letter | Number) }
                   Letter         ::= A..Z | a..z
                   Number         ::= 0..9
*)

PROCEDURE SetSearchPath (Path: ARRAY OF CHAR) ;
BEGIN
   StrCopy(Path, SearchPath)
END SetSearchPath ;


(*
   Init - initializes the search path to '.'.
*)

PROCEDURE Init ;
BEGIN
   GetSearchPath(SearchPath) ;
   IF StrLen(SearchPath)=0
   THEN
      StrCopy(DefaultLibraryPath, SearchPath)   (* Default to current directory *)
   END
END Init ;


BEGIN
   Init
END M2Search.
