IMPLEMENTATION MODULE M2System ;


FROM StrIO IMPORT WriteString, WriteLn ;
FROM NameKey IMPORT MakeKey, NulName ;

FROM SymbolTable IMPORT NulSym,
                        SetCurrentModule,
                        StartScope,
                        EndScope,
      	       	     	MakeConstLit,
                        MakePointer,
                        MakeType,
                        MakeProcedure,
      	       	     	MakeSet,
      	       	     	MakeSubrange,
                        PutFunction,
                        PutType, PutPointer,
      	       	     	PutSet,
      	       	     	PutSubrange,
                        GetSym,
                        PopSize ;

FROM M2Batch IMPORT MakeDefinitionSource ;
FROM M2Options IMPORT WordLength, APUCodeGenerator ;
FROM M2Base IMPORT Cardinal ;
FROM M2ALU IMPORT PushCard ;


(*
   InitSizes - initializes the sizes of word, byte, address
               and bitset depending upon the WordLength.
*)

PROCEDURE InitSizes ;
BEGIN
   IF APUCodeGenerator
   THEN
      WordSize    := 4 ;
      ByteSize    := 1 ;
      AddressSize := 4 ;
      BitsetSize  := 4
   ELSIF WordLength=32
   THEN
      WordSize    := 4 ;
      ByteSize    := 1 ;
      AddressSize := 4 ;
      BitsetSize  := 4
   ELSE
      WordSize    := 2 ;
      ByteSize    := 1 ;
      AddressSize := 2 ;
      BitsetSize  := 2
   END
END InitSizes ;
   
   
(*
   InitSystem - creates the system dependant types and procedures.
                Note that they are not exported here, but they are
                exported in the textual module: SYSTEM.def.
                Therefore the compiler must have access to this module
                to ensure correct operation.
*)

PROCEDURE InitSystem ;
VAR
   System     : CARDINAL ;
   BitsetRange: CARDINAL ;
BEGIN
   (* create SYSTEM module *)
   System := MakeDefinitionSource(MakeKey('SYSTEM')) ;
   StartScope(System) ;

   Word := MakeType(MakeKey('WORD')) ;
   PutType(Word, NulSym) ;                    (* Base Type       *)
   PushCard(WordSize) ;
   PopSize(Word) ;

   (* ADDRESS = POINTER TO WORD *)

   Address := MakePointer(MakeKey('ADDRESS')) ;
   PutPointer(Address, Word) ;                (* Base Type       *)
   PushCard(AddressSize) ;
   PopSize(Address) ;

   Byte := MakeType(MakeKey('BYTE')) ;
   PutType(Byte, NulSym) ;                    (* Base Type       *)
   PushCard(ByteSize) ;
   PopSize(Byte) ;

   Bitset := MakeSet(MakeKey('BITSET')) ;     (* Base Type       *)

   BitsetRange := MakeSubrange(NulName) ;
   PutSubrange(
      	        BitsetRange,
      	        MakeConstLit(MakeKey('0')), MakeConstLit(MakeKey('31')),
      	        Word
      	      ) ;
   PutSet(Bitset, BitsetRange) ;

(*   PutSet(Bitset, Cardinal) ; surely this is a mistake? *)
   PushCard(BitsetSize) ;
   PopSize(Bitset) ;

   (* And now the predefined pseudo functions *)

   Size := MakeProcedure(MakeKey('SIZE')) ;   (* Function        *)
   PutFunction(Size, Cardinal) ;              (* Return Type     *)
                                              (* Cardinal        *)

   Adr := MakeProcedure(MakeKey('ADR')) ;     (* Function        *)
   PutFunction(Adr, Address) ;                (* Return Type     *)
                                              (* Address         *)

   TSize := MakeProcedure(MakeKey('TSIZE')) ; (* Function        *)
   PutFunction(TSize, Cardinal) ;             (* Return Type     *)
                                              (* Cardinal        *)

   EndScope
END InitSystem ;


(*
   GetSystemTypeMinMax - returns the minimum and maximum values for a given system type.
*)

PROCEDURE GetSystemTypeMinMax (type: CARDINAL; VAR min, max: CARDINAL) ;
BEGIN
   min := MakeConstLit(MakeKey('0')) ;
   IF (type=Word) OR (type=Address)
   THEN
      IF WordSize=4
      THEN
         max := MakeConstLit(MakeKey('0FFFFFFFFH'))
      ELSE
         max := MakeConstLit(MakeKey('0FFFFH'))
      END
   ELSIF type=Byte
   THEN
      max := MakeConstLit(MakeKey('0FFH'))
   END
END GetSystemTypeMinMax ;


(*
   IsPseudoSystemFunction - returns true if Sym is a SYSTEM pseudo function.
*)

PROCEDURE IsPseudoSystemFunction (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN( (Sym=Adr) OR (Sym=TSize) OR (Sym=Size) )
END IsPseudoSystemFunction ;


(*
   IsSystemType - returns TRUE if Sym is a SYSTEM (inbuilt) type.
                  It does not search your SYSTEM implementation module.
*)

PROCEDURE IsSystemType (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN(
          (Sym=Word)    OR (Sym=Byte) OR
          (Sym=Address) OR (Sym=Bitset)
         )
END IsSystemType ;


END M2System.
