IMPLEMENTATION MODULE NameKey ;


FROM Storage IMPORT ALLOCATE ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM StdIO IMPORT Write ;
FROM NumberIO IMPORT WriteCard ;
FROM StrLib IMPORT StrLen ;
FROM ASCII IMPORT nul ;


CONST
   TableLength = 1000000 ;
   LineWidth   =    4096 ;

TYPE
   Table   = POINTER TO Word ;
   Word    = RECORD
                Key  : CARDINAL ;   (* Key Table Index *)
                Left,
                Right: Table ;
             END ;


   Relation= (less, equal, greater) ;

VAR
   asc    : ARRAY [0..TableLength+LineWidth] OF CHAR ;
   ascinx : CARDINAL ;
   Names  : Table ;


(* %%%FORWARD%%%
PROCEDURE PrintItem (t: Table) ; FORWARD ;
PROCEDURE Rel (k: CARDINAL) : Relation ; FORWARD ;
PROCEDURE Search (VAR id, father: Table; VAR r: Relation) ; FORWARD ;
PROCEDURE TraverseTree (t: Table) ; FORWARD ;
   %%%FORWARD%%% *)

(*
   GetKey - returns the name, a, of the key, Key.
*)

PROCEDURE GetKey (Key: CARDINAL; VAR a: ARRAY OF CHAR) ;
VAR
   i, higha: CARDINAL ;
BEGIN
   (* Pseudo MACRO - place asc[] into a[] *)
   i := 0 ;
   higha := HIGH(a) ;
   WHILE (i<higha) AND (asc[Key+i]#nul) DO
      a[i] := asc[Key+i] ;
      INC(i)
   END ;
   IF i<higha
   THEN
      a[i] := nul
   END
   (* END of MACRO *)
END GetKey ;


(*
   IsKey - returns TRUE if string, a, is currently a key.
*)

PROCEDURE IsKey (a: ARRAY OF CHAR) : BOOLEAN ;
VAR
   r     : Relation ;
   f, c  : Table ;
   i,
   higha : CARDINAL ;
BEGIN
   (* Pseudo MACRO - place a[] into asc[] *)
   i := 0 ;
   higha := StrLen(a) ;
   WHILE i<higha DO
      asc[ascinx+i] := a[i] ;
      INC(i)
   END ;
   asc[ascinx+i] := nul ;
   (* END of MACRO *)

   Search( c, f, r ) ;
   RETURN( (c#NIL) AND (r=equal) )
END IsKey ;


PROCEDURE MakeKey (a: ARRAY OF CHAR) : CARDINAL ;
VAR
   r     : Relation ;
   f, c  : Table ;
   i, n,
   higha : CARDINAL ;
BEGIN
   IF ascinx>TableLength
   THEN
      (* Halt('fatal error - symbol table overflow') *)
      HALT
   ELSE
      (* Pseudo MACRO - place a[] into asc[] *)
      i := 0 ;
      higha := StrLen(a) ;
      WHILE i<higha DO
         asc[ascinx+i] := a[i] ;
         INC( i )
      END ;
      asc[ascinx+i] := nul ;
      (* END of MACRO *)

      Search( c, f, r ) ;
      IF c=NIL
      THEN
         IF r=less
         THEN
            NEW( f^.Left ) ;
            c := f^.Left
         ELSIF r=greater
         THEN
            NEW( f^.Right ) ;
            c := f^.Right
         END ;
         WITH c^ DO
            Right := NIL ;
            Left := NIL ;
            Key := ascinx
         END ;
         n := ascinx ;
         INC( ascinx, higha+1 ) 
      ELSE
         n := c^.Key
      END
   END ;
   RETURN( n )
END MakeKey ;


(*
   LengthKey - returns the StrLen of Key.
*)

PROCEDURE LengthKey (Key: CARDINAL) : CARDINAL ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE asc[Key]#nul DO
      INC(i) ;
      INC(Key)
   END ;
   RETURN( i )
END LengthKey ;


(*
   Search table, t,with name equal to a. id is set to the node if
   an entry is found, father is set to the node above id.
*)

PROCEDURE Search (VAR id, father: Table; VAR r: Relation) ;
BEGIN
   father := Names ;
   id := Names^.Left ;
   IF id=NIL
   THEN
      r := less
   ELSE
      REPEAT
         r := Rel( id^.Key ) ;
         IF r=less
         THEN
            father := id ;
            id := id^.Left
         ELSIF r=greater
         THEN
            father := id ;
            id := id^.Right
         END
      UNTIL (id=NIL) OR (r=equal)
   END
END Search ;


(*
   Compare asc[ascinx] with asc[k]
*)

PROCEDURE Rel (k: CARDINAL) : Relation ;
VAR
   i: CARDINAL ;
   r: Relation ;
   x, y: CHAR ;
BEGIN
   i := 0 ;
   x := asc[ascinx] ;
   y := asc[k] ;
   r := equal ;
   REPEAT
      IF x<y
      THEN
         r := less
      ELSIF x>y
      THEN
         r := greater
      ELSE
         INC(i) ;
         x := asc[ascinx+i] ;
         INC(k) ;
         y := asc[k] ;
         IF (x=nul) AND (y#nul)
         THEN
            r := less
         ELSIF (x#nul) AND (y=nul)
         THEN
            r := greater
         END
      END
   UNTIL (r#equal) OR (x=nul) OR (y=nul) ;
   RETURN(r)
END Rel ;


(*
   IsSameExcludingCase - returns TRUE if key1 and key2 are
                         the same. It is case insensitive.
                         This function is delibertley manually
                         inlined for speed.
*)

PROCEDURE IsSameExcludingCase (key1, key2: CARDINAL) : BOOLEAN ;
VAR
   c1, c2: CHAR ;
BEGIN
   IF key1=key2
   THEN
      RETURN( TRUE )
   ELSE
      c1 := asc[key1] ;
      c2 := asc[key2] ;
      WHILE (c1#nul) AND (c2#nul) DO
         IF (c1=c2) OR
            (((c1>='A') AND (c1<='Z')) AND (c2=CHR(ORD(c1)-ORD('A')+ORD('a')))) OR
            (((c2>='A') AND (c2<='Z')) AND (c1=CHR(ORD(c2)-ORD('A')+ORD('a'))))
         THEN
            INC(key1) ;
            INC(key2) ;
            c1 := asc[key1] ;
            c2 := asc[key2]
         ELSE
            (* difference found *)
            RETURN( FALSE )
         END
      END ;
      RETURN( c1=c2 )
   END
END IsSameExcludingCase ;


PROCEDURE Tabulate ;
BEGIN
   WriteLn ;
   TraverseTree(Names^.Left)
END Tabulate ;


PROCEDURE TraverseTree (t: Table) ;
BEGIN
   IF t#NIL
   THEN
      TraverseTree(t^.Left) ;
      PrintItem(t) ;
      TraverseTree(t^.Right)
   END
END TraverseTree ;


PROCEDURE PrintItem (t: Table) ;
BEGIN
   WriteCard(t^.Key, 6) ;
   WriteString('   ') ;
   WriteKey(t^.Key) ;
   WriteLn
END PrintItem ;


PROCEDURE WriteKey (Key: CARDINAL) ;
VAR
   ch : CHAR ;
BEGIN
   ch := asc[Key] ;
   WHILE ch#nul DO
      Write(ch) ;
      INC(Key) ;
      ch := asc[Key]
   END
END WriteKey ;


BEGIN
   ascinx := 1 ;
   asc[0] := nul ; (* Now NulName can be written the same as all other keys *)
   NEW( Names ) ;
   Names^.Left := NIL
END NameKey.
