IMPLEMENTATION MODULE PrettyEBNF ;


FROM NameKey IMPORT LengthKey, WriteKey, MakeKey ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM StdIO IMPORT Write ;
FROM StrLib IMPORT StrLen ;
FROM YacTok IMPORT GetTok ;


(* %%%FORWARD%%%
PROCEDURE ParseLiteral (VAR Index: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE ParseAtom (VAR Index: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE Init ; FORWARD ;
PROCEDURE StartComment ; FORWARD ;
PROCEDURE EndComment ; FORWARD ;
PROCEDURE NewLineComment ; FORWARD ;
PROCEDURE CreateRule ; FORWARD ;
PROCEDURE Production () : BOOLEAN ; FORWARD ;
PROCEDURE DisplayExpression (VAR Index: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE ParseExpression (VAR Index: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE SplitExpression (VAR Index: CARDINAL) ; FORWARD ;
PROCEDURE DisplayTerm (VAR Index: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE ParseTerm (VAR Index: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE SplitTerm (VAR Index: CARDINAL) ; FORWARD ;
PROCEDURE DisplayFactor (VAR Index: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE ParseFactor (VAR Index: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE SplitFactor (VAR Index: CARDINAL) ; FORWARD ;
PROCEDURE ParseSymbol (VAR Index: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE DisplaySymbol (VAR Index: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE DisplaySymIs (a: ARRAY OF CHAR; VAR Index: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE ParseSymIs (a: ARRAY OF CHAR; VAR Index: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE DisplayAtom (VAR Index: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE DisplayLiteral (VAR Index: CARDINAL) : BOOLEAN ; FORWARD ;
   %%%FORWARD%%% *)


CONST
   BaseRightLimit  = 75 ;
   BaseRightMargin = 60 ;
   BaseNewLine     =  3 ;

VAR
   BeginningOfLine: BOOLEAN ;
   CurrentSymbol,
   Indent,
   NewLineIndent,
   RightLimit,
   RightMargin,
   LeftMargin     : CARDINAL ;


(*
   Init - initializes the global variables used in this module.
*)

PROCEDURE Init ;
BEGIN
   BeginningOfLine := TRUE ;
   RightLimit := BaseRightLimit ;
   RightMargin := BaseRightMargin ;
   LeftMargin := 0 ;
   Indent := 0 ;
   NewLineIndent := BaseNewLine ;
   CurrentSymbol := GetTok(0)
END Init ;


(*
   StartComment - starts writing the bnf rule as a comment.
*)

PROCEDURE StartComment ;
BEGIN
   Init ;
   WriteLn ; WriteLn ;
   WriteString('(*') ;
   NewLineComment
END StartComment ;


(*
   EndComment - ends writing the bnf rule in a comment.
*)

PROCEDURE EndComment ;
BEGIN
   WriteLn ;
   WriteString('*)') ; WriteLn
END EndComment ;


(*
   NewLineComment - generate a newline within a comment.
*)

PROCEDURE NewLineComment ;
BEGIN
   WriteLn ;
   BeginningOfLine := TRUE ;
   Indent := 0 ;
   WHILE Indent<NewLineIndent DO
      Write(' ') ;
      INC(Indent)
   END
END NewLineComment ;


(*
   CreateRule - generates the start of a bnf rule.
*)

PROCEDURE CreateRule ;
BEGIN
   NewLineIndent := Indent + LengthKey(GetTok(0)) + StrLen(':=') + 2 ;
   Indent := NewLineIndent ;
   WriteKey(GetTok(0)) ; WriteString(' := ')
END CreateRule ;


(*
   CheckForNewLineInComment - checks to see whether we need to insert
                              a newline within the bnf rule.
*)

PROCEDURE CheckForNewLineInComment ;
BEGIN
   IF Indent>BaseRightMargin
   THEN
      NewLineComment
   END
END CheckForNewLineInComment ;


(*
   PrettyPrintRule - displays an EBNF rule.
*)

PROCEDURE PrettyPrintRule ;
BEGIN
   StartComment ;
   IF Production()
   THEN
   END ;
   EndComment
END PrettyPrintRule ;


(*
   Production := Symbol " := " Expression =:
*)

PROCEDURE Production () : BOOLEAN ;
VAR
   Success         : BOOLEAN ;
   ExpLength       : CARDINAL ;
   BeforeExpression: CARDINAL ;
   Index           : CARDINAL ;
BEGIN
   Index := 0 ;
   IF ParseSymbol(Index)>0
   THEN
      IF ParseSymIs(':=', Index)>0
      THEN
         CreateRule ;
         BeforeExpression := Index ;
         ExpLength := ParseExpression(Index) ;
         IF ExpLength>0
         THEN
            IF Indent+ExpLength<BaseRightMargin
            THEN
               Success := DisplayExpression(BeforeExpression)
            ELSE
               SplitExpression(BeforeExpression)
            END ;
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Production ;


(*
   DisplayExpression := Term { " | " Term } =:
*)

PROCEDURE DisplayExpression (VAR Index: CARDINAL) : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF DisplayTerm(Index)
   THEN
      WHILE DisplaySymIs('|', Index) DO
         IF DisplayTerm(Index)
         THEN
         END
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END DisplayExpression ;


(*
   ParseExpression := Term { " | " Term } =:
*)

PROCEDURE ParseExpression (VAR Index: CARDINAL) : CARDINAL ;
VAR
   LengthSym,
   Length   : CARDINAL ;
BEGIN
   Length := ParseTerm(Index) ;
   IF Length>0
   THEN
      LengthSym := ParseSymIs('|', Index) ;
      INC(Length, LengthSym) ;
      WHILE LengthSym>0 DO
         LengthSym := ParseTerm(Index) ;
         IF LengthSym>0
         THEN
            INC(Length, LengthSym)
         END ;
         LengthSym := ParseSymIs('|', Index) ;
         INC(Length, LengthSym)
      END
   END ;
   RETURN( Length )
END ParseExpression ;


(*
   Expression := Term { " | " Term } =:
*)

PROCEDURE SplitExpression (VAR Index: CARDINAL) ;
VAR
   Success   : BOOLEAN ;
   BeforeTerm,
   l,
   Length    : CARDINAL ;
BEGIN
   BeforeTerm := Index ;
   Length := ParseTerm(Index) ;
   IF Length>0
   THEN
      IF Indent+Length<BaseRightMargin
      THEN
         Success := DisplayTerm(BeforeTerm)
      ELSE
         SplitTerm(BeforeTerm)
      END ;
      BeforeTerm := Index ;
      l := ParseSymIs('|', Index) ;
      WHILE l>0 DO
         Success := DisplaySymIs('|', BeforeTerm) ;
         Index := BeforeTerm ;
         l := ParseTerm(Index) ;
         IF l>0
         THEN
            IF Indent+l<BaseRightMargin
            THEN
               Success := DisplayTerm(BeforeTerm)
            ELSIF NOT BeginningOfLine
            THEN
               NewLineComment ;   (* natural to break line after | *)
               IF Indent+l<BaseRightMargin
               THEN
                  Success := DisplayTerm(BeforeTerm)
               ELSE
                  SplitTerm(BeforeTerm)
               END
            ELSE
               SplitTerm(BeforeTerm)
            END
         END
      END
   END
END SplitExpression ;


(*
   Term := Factor { Factor } =:
*)

PROCEDURE DisplayTerm (VAR Index: CARDINAL) : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF DisplayFactor(Index)
   THEN
      WHILE DisplayFactor(Index) DO
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END DisplayTerm ;


(*
   Term := Factor { Factor } =:
*)

PROCEDURE ParseTerm (VAR Index: CARDINAL) : CARDINAL ;
VAR
   l,
   Length: CARDINAL ;
BEGIN
   Length := ParseFactor(Index) ;
   IF Length>0
   THEN
      l := ParseFactor(Index) ;
      WHILE l>0 DO
         INC(Length, l) ;
         l := ParseFactor(Index)
      END
   END ;
   RETURN( Length )
END ParseTerm ;


(*
   Term := Factor { Factor } =:
*)

PROCEDURE SplitTerm (VAR Index: CARDINAL) ;
VAR
   Success     : BOOLEAN ;
   BeforeFactor,
   l,
   Length      : CARDINAL ;
BEGIN
   BeforeFactor := Index ;
   Length := ParseFactor(Index) ;
   IF Length>0
   THEN
      IF Indent+Length<BaseRightMargin
      THEN
         Success := DisplayFactor(BeforeFactor)
      ELSE
         SplitFactor(BeforeFactor)
      END ;
      BeforeFactor := Index ;
      l := ParseFactor(Index) ;
      WHILE l>0 DO
         IF Indent+l<BaseRightMargin
         THEN
            Success := DisplayFactor(BeforeFactor)
         ELSE
            SplitFactor(BeforeFactor)
         END ;
         BeforeFactor := Index ;
         l := ParseFactor(Index)
      END
   END
END SplitTerm ;


(*
   Factor := Symbol | " { " ( Expression " } " ) | " [ " ( Expression " ] " ) | " ( " ( Expression " ) " ) =:
*)

PROCEDURE DisplayFactor (VAR Index: CARDINAL) : BOOLEAN ;
VAR
   OldIndex: CARDINAL ;
   Success : BOOLEAN ;
BEGIN
   OldIndex := Index ;
   IF DisplayAtom(Index)
   THEN
      Success := TRUE
   ELSIF DisplaySymIs('{', Index)
   THEN
      IF DisplayExpression(Index)
      THEN
         IF DisplaySymIs('}', Index)
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSIF DisplaySymIs('[', Index)
   THEN
      IF DisplayExpression(Index)
      THEN
         IF DisplaySymIs(']', Index)
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSIF DisplaySymIs('(', Index)
   THEN
      IF DisplayExpression(Index)
      THEN
         IF DisplaySymIs(')', Index)
         THEN
            Success := TRUE
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   IF Success
   THEN
      RETURN( TRUE )
   ELSE
      Index := OldIndex ;
      RETURN( FALSE )
   END
END DisplayFactor ;


(*
   Factor := Symbol | " { " ( Expression " } " ) | " [ " ( Expression " ] " ) | " ( " ( Expression " ) " ) =:
*)

PROCEDURE ParseFactor (VAR Index: CARDINAL) : CARDINAL ;
VAR
   OldIndex,
   Length, l: CARDINAL ;
BEGIN
   OldIndex := Index ;
   Length := ParseAtom(Index) ;
   IF Length>0
   THEN
      RETURN( Length )
   ELSE
      Length := ParseSymIs('{', Index) ;
      IF Length>0
      THEN
         l := ParseExpression(Index) ;
         IF l>0
         THEN
            INC(Length, l) ;
            l := ParseSymIs('}', Index) ;
            IF l>0
            THEN
               RETURN( l+Length )
            END
         END
      ELSE
         Length := ParseSymIs('[', Index) ;
         IF Length>0
         THEN
            l := ParseExpression(Index) ;
            IF l>0
            THEN
               INC(Length, l) ;
               l := ParseSymIs(']', Index) ;
               IF l>0
               THEN
                  RETURN( l+Length )
               END
            END
         ELSE
            Length := ParseSymIs('(', Index) ;
            IF Length>0
            THEN
               l := ParseExpression(Index) ;
               IF l>0
               THEN
                  INC(Length, l) ;
                  l := ParseSymIs(')', Index) ;
                  IF l>0
                  THEN
                     RETURN( l+Length )
                  END
               END
            END
         END
      END
   END ;
   Index := OldIndex ;
   RETURN( 0 )
END ParseFactor ;


(*
   Factor := Symbol | " { " ( Expression " } " ) | " [ " ( Expression " ] " ) | " ( " ( Expression " ) " ) =:
*)

PROCEDURE SplitFactor (VAR Index: CARDINAL) ;
VAR
   OldIndex,
   l,
   Length: CARDINAL ;
   Success: BOOLEAN ;
BEGIN
   OldIndex := Index ;
   Length := ParseAtom(Index) ;
   IF Length>0
   THEN
      (* cannot split a single symbol so just display it and finish *)
      IF Indent+Length<BaseRightMargin
      THEN
         IF NOT BeginningOfLine
         THEN
            NewLineComment
         END
      END ;
      Success := DisplaySymbol(OldIndex)
   ELSE
      Length := ParseSymIs('{', Index) ;
      IF Length>0
      THEN
         l := ParseExpression(Index) ;
         IF l>0
         THEN
            INC(Length, l) ;
            l := ParseSymIs('}', Index) ;
            IF l>0
            THEN
               IF Indent+l+Length<BaseRightMargin
               THEN
                  Success := DisplaySymIs('{', OldIndex) ;
                  Success := DisplayExpression(OldIndex) ;
                  Success := DisplaySymIs('}', OldIndex)
               ELSE
                  IF NOT BeginningOfLine
                  THEN
                     NewLineComment ;
                  END ;
                  Success := DisplaySymIs('{', OldIndex) ;
                  INC(NewLineIndent, 2) ;
                  Index := OldIndex ;
                  Length := ParseExpression(Index) ;
                  IF Indent+Length<BaseRightMargin
                  THEN
                     Success := DisplayExpression(OldIndex)
                  ELSE
                     SplitExpression(OldIndex)
                  END ;
                  Success := DisplaySymIs('}', Index) ;
                  DEC(NewLineIndent, 2)
               END
            END
         END
      ELSE
         Length := ParseSymIs('[', Index) ;
         IF Length>0
         THEN
            l := ParseExpression(Index) ;
            IF l>0
            THEN
               INC(Length, l) ;
               l := ParseSymIs(']', Index) ;
               IF l>0
               THEN
                  IF Indent+l+Length<BaseRightMargin
                  THEN
                     Success := DisplaySymIs('[', OldIndex) ;
                     Success := DisplayExpression(OldIndex) ;
                     Success := DisplaySymIs(']', OldIndex)
                  ELSE
                     IF NOT BeginningOfLine
                     THEN
                        NewLineComment ;
                     END ;
                     Success := DisplaySymIs('[', OldIndex) ;
                     Index := OldIndex ;
                     Length := ParseExpression(Index) ;
                     INC(NewLineIndent, 2) ;
                     IF Indent+Length<BaseRightMargin
                     THEN
                        Success := DisplayExpression(OldIndex)
                     ELSE
                        SplitExpression(OldIndex)
                     END ;
                     Success := DisplaySymIs(']', Index) ;
                     DEC(NewLineIndent, 2)
                  END
               END
            END
         ELSE
            Length := ParseSymIs('(', Index) ;
            IF Length>0
            THEN
               l := ParseExpression(Index) ;
               IF l>0
               THEN
                  INC(Length, l) ;
                  l := ParseSymIs(')', Index) ;
                  IF l>0
                  THEN
                     IF Indent+l+Length<BaseRightMargin
                     THEN
                        Success := DisplaySymIs('(', OldIndex) ;
                        Success := DisplayExpression(OldIndex) ;
                        Success := DisplaySymIs(')', OldIndex)
                     ELSE
                        IF NOT BeginningOfLine
                        THEN
                           NewLineComment ;
                        END ;
                        Success := DisplaySymIs('(', OldIndex) ;
                        Index := OldIndex ;
                        Length := ParseExpression(Index) ;
                        INC(NewLineIndent, 2) ;
                        IF Indent+Length<BaseRightMargin
                        THEN
                           Success := DisplayExpression(OldIndex)
                        ELSE
                           SplitExpression(OldIndex)
                        END ;
                        Success := DisplaySymIs(')', Index) ;
                        DEC(NewLineIndent, 2)
                     END
                  END
               END
            END
         END
      END
   END
END SplitFactor ;


(*
   ParseAtom - parse an atom and return its length.
*)

PROCEDURE ParseAtom (VAR Index: CARDINAL) : CARDINAL ;
VAR
   Length: CARDINAL ;
BEGIN
   Length := ParseLiteral(Index) ;
   IF Length>0
   THEN
      RETURN( Length )
   ELSE
      RETURN( ParseSymbol(Index) )
   END
END ParseAtom ;


(*
   DisplayAtom - displays an atom and returns whether
                 the atom was found or not.
                 Index is only incremented if an atom was
                 found.
*)

PROCEDURE DisplayAtom (VAR Index: CARDINAL) : BOOLEAN ;
BEGIN
   IF DisplayLiteral(Index)
   THEN
      RETURN( TRUE )
   ELSE
      RETURN( DisplaySymbol(Index) )
   END
END DisplayAtom ;


(*
   ParseLiteral - parses a literal and returns its length.
*)

PROCEDURE ParseLiteral (VAR Index: CARDINAL) : CARDINAL ;
VAR
   l,
   Length: CARDINAL ;
BEGIN
   Length := ParseSymIs("'", Index) ;
   IF Length>0
   THEN
      INC(Length, LengthKey(GetTok(Index))+1) ;
      INC(Index) ;
      l := ParseSymIs("'", Index) ;
      INC(Length, l+1) ;
      RETURN( Length )
   ELSE
      Length := ParseSymIs('"', Index) ;
      IF Length>0
      THEN
         INC(Length, LengthKey(GetTok(Index))+1) ;
         INC(Index) ;
         l := ParseSymIs('"', Index) ;
         INC(Length, l+1) ;
         RETURN( Length )
      ELSE
         RETURN( 0 )
      END
   END
END ParseLiteral ;


(*
   DisplayLiteral - displays a literal.
                    It returns true if a literal was found and
                    increments, Index. If false is returned
                    then, Index is not incremented.
*)

PROCEDURE DisplayLiteral (VAR Index: CARDINAL) : BOOLEAN ;
VAR
   l,
   Length: CARDINAL ;
BEGIN
   IF GetTok(Index)=MakeKey("'")
   THEN
      WriteString("' ") ; WriteKey(GetTok(Index+1)) ; WriteString(" ' ") ;
      INC(Indent, LengthKey(Index+1)+5) ;
      INC(Index, 3) ;
      RETURN( TRUE )
   ELSIF GetTok(Index)=MakeKey('"')
   THEN
      WriteString('" ') ; WriteKey(GetTok(Index+1)) ; WriteString(' " ') ;
      INC(Indent, LengthKey(Index+1)+5) ;
      INC(Index, 3) ;
      RETURN( TRUE )
   ELSE
      RETURN( FALSE )
   END
END DisplayLiteral ;


(*
   ParseSymIs - if, a, equals the symbol indicated by, Index, then
                its length is returned and, Index, is incremented.
                Otherwise zero is returned and, Index, is left alone.
*)

PROCEDURE ParseSymIs (a: ARRAY OF CHAR; VAR Index: CARDINAL) : CARDINAL ;
BEGIN
   IF GetTok(Index)=MakeKey(a)
   THEN
      INC(Index) ;                (* onto the next token *)
      RETURN( LengthKey(GetTok(Index-1))+1 )
   ELSE
      RETURN( 0 )
   END
END ParseSymIs ;


(*
   DisplaySymIs - if a equals the symbol indicated by Index then
                  true is returned, the symbol is written
                  and Index is incremented.
                  Otherwise false is returned and Index is left alone.
*)

PROCEDURE DisplaySymIs (a: ARRAY OF CHAR; VAR Index: CARDINAL) : BOOLEAN ;
BEGIN
   IF GetTok(Index)=MakeKey(a)
   THEN
      WriteString(a) ; Write(' ') ;
      INC(Indent, StrLen(a)+1) ;  (* update the Indent   *)
      INC(Index) ;                (* onto the next token *)
      RETURN( TRUE )
   ELSE
      RETURN( FALSE )
   END
END DisplaySymIs ;
                                                

(*
   DisplaySymbol - displays symbol, Index, and returns true.
*)

PROCEDURE DisplaySymbol (VAR Index: CARDINAL) : BOOLEAN ;
VAR
   Sym: CARDINAL ;
BEGIN
   (*
      needs to specifically check for a symbol - ie not a { } [ ] ( ) etc.
   *)
   Sym := GetTok(Index) ;
   IF (Sym=MakeKey('{')) OR (Sym=MakeKey('}')) OR
      (Sym=MakeKey('[')) OR (Sym=MakeKey(']')) OR
      (Sym=MakeKey('(')) OR (Sym=MakeKey(')')) OR
      (Sym=MakeKey('|')) OR (Sym=MakeKey(':=')) OR (Sym=MakeKey('=:')) OR
      (Sym=MakeKey('BNF')) OR (Sym=MakeKey('FNB'))
   THEN
      RETURN( FALSE )
   ELSE
      BeginningOfLine := FALSE ;
      WriteKey(GetTok(Index)) ; Write(' ') ;
      INC(Indent, LengthKey(Sym)+1) ;
      INC(Index) ;
      RETURN( TRUE )
   END
END DisplaySymbol ;


(*
   ParseSymbol - return the length of the current symbol and increment Index.
*)

PROCEDURE ParseSymbol (VAR Index: CARDINAL) : CARDINAL ;
VAR
   Length,
   Sym   : CARDINAL ;
BEGIN
   (*
      needs to specifically check for a symbol - ie not a { } [ ] ( ) etc.
   *)
   Sym := GetTok(Index) ;
   IF (Sym=MakeKey('{')) OR (Sym=MakeKey('}')) OR
      (Sym=MakeKey('[')) OR (Sym=MakeKey(']')) OR
      (Sym=MakeKey('(')) OR (Sym=MakeKey(')')) OR
      (Sym=MakeKey('|')) OR (Sym=MakeKey(':=')) OR (Sym=MakeKey('=:')) OR
      (Sym=MakeKey('BNF')) OR (Sym=MakeKey('FNB'))
   THEN
      RETURN( 0 )
   ELSE
      Length := LengthKey(GetTok(Index)) ;
      INC(Index) ;
      RETURN( Length )
   END
END ParseSymbol ;


END PrettyEBNF.
