MODULE gm2 ;

(*
   Author     : Gaius Mulley
   Title      : gm2
   Date       : 1987  [$Date: 1997/05/21 08:17:29 $]
   SYSTEM     : UNIX SUN and Logitech M2
   Description: Main module of the compiler, collects arguments and
                starts the compilation.
   Version    : $Revision: 1.2 $
*)

(*
   Log        : $Log: gm2.mod,v $
   Log        : Revision 1.2  1997/05/21 08:17:29  gaius
   Log        : fixed install scripts inside makefiles. Removed confusion between
   Log        : release and install. Both are now treated the same. Made compiler
   Log        : handle -h option correctly.
   Log        :
   Log        : Revision 1.1.1.1  1996/05/28 10:13:03  gaius
   Log        : Modula-2 compiler sources imported
   Log        :
   Revision 1.1  91/05/04  12:42:41  comp
   Initial revision
*)


IMPORT DebugPMD ;
IMPORT CmdArgs ;
FROM M2Search IMPORT FindSourceFile ;
FROM M2Options IMPORT IsAnOption, ParseOptions, QuietString, QuietLn ;
FROM M2Comp IMPORT Compile ;
FROM M2Batch IMPORT MakeProgramSource, MakeImplementationSource ;
FROM SymbolTable IMPORT SetMainModule ;
FROM NameKey IMPORT MakeKey ;
FROM M2FileName IMPORT CalculateFileName ;
FROM Args IMPORT GetArg, Narg ;
FROM StrLib IMPORT StrEqual, StrLen ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM libc IMPORT exit ;


CONST
   MaxStringLen = 1000 ;


(* %%%FORWARD%%%
PROCEDURE StartParsing ; FORWARD ;
PROCEDURE CompileModule (Name: ARRAY OF CHAR) ; FORWARD ;
   %%%FORWARD%%% *)


(*
   StartParsing - scans the command line and processes the arguments.
                  It attempts to compile any modules supplied.
*)

PROCEDURE StartParsing ;
VAR
   a: ARRAY [0..MaxStringLen] OF CHAR ;
   n: CARDINAL ;
BEGIN
   ParseOptions ;
   n := Narg() ;
   IF (n>1) AND GetArg(a, n-1)
   THEN
      (* WriteString('Argument found: ') ; WriteString(a) ; WriteLn ; *)
      IF StrEqual(a, '-M')
      THEN
      ELSIF StrEqual(a, '-version')
      THEN
      ELSIF NOT IsAnOption(a)
      THEN
         CompileModule(a)
      END
   END
END StartParsing ;


(*
   CompileModule - attempts to locate a module Name and then compile
                   this module.
*)

PROCEDURE CompileModule (Name: ARRAY OF CHAR) ;
VAR
   Sym     : CARDINAL ;
   FullPath,
   FileName: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   CalculateFileName(Name, 'def', FileName) ;
   IF FindSourceFile(FileName, FullPath)
   THEN
      Sym := MakeImplementationSource(MakeKey(Name)) ;
      CalculateFileName(Name, 'mod', FileName) ;
      IF FindSourceFile(FileName, FullPath)
      THEN 
         SetMainModule(Sym) ;
      	 QuietString('Compiling ') ; QuietString(FullPath) ; QuietLn ;
         Compile(Sym)
      ELSE
         WriteString('File: ') ; WriteString(FileName) ;
         WriteString(' not found') ; WriteLn ;
         exit(1)
      END
   ELSE
      CalculateFileName(Name, 'mod', FileName) ;
      IF FindSourceFile(FileName, FullPath)
      THEN
      	 QuietString('Compiling ') ; QuietString(FullPath) ; QuietLn ;
         Sym := MakeProgramSource(MakeKey(Name)) ;
         SetMainModule(Sym) ;
         Compile(Sym)
      ELSE
         WriteString('File: ') ; WriteString(FileName) ;
         WriteString(' not found') ; WriteLn ;
         exit(1)
      END
   END
END CompileModule ;


BEGIN
   StartParsing
END gm2.
