MODULE gm2l ;
 
(*
   Author     : Gaius Mulley
   Title      : gm2l
   Date       : Date: Sat 16-09-1989 Time: 17:49:34.18
              : [$Date: 1997/07/11 08:03:52 $]
   SYSTEM     : UNIX SUN and Logitech M2
   Description: generates the list of initialization order for the modules.
   Version    : $Revision: 1.2 $
*)

(*
   Log        : $Log: gm2l.mod,v $
   Log        : Revision 1.2  1997/07/11 08:03:52  gaius
   Log        : altered gm2l.mod to contain an 8k char path variable and not 79 chars.
   Log        :
   Log        : Revision 1.1.1.1  1996/05/28 10:13:02  gaius
   Log        : Modula-2 compiler sources imported
   Log        :
*)

IMPORT Break ;
IMPORT DebugPMD ;
IMPORT M2Search ;
FROM M2Lex IMPORT OpenSource, GetSymbol, IsSym, SymIs, CloseSource,
                  CurrentSymbol ;
FROM M2FileName IMPORT CalculateFileName ;
FROM StrLib IMPORT StrEqual, StrCopy, StrConCat, StrLen ;
FROM NumberIO IMPORT WriteCard ;
FROM StdIO IMPORT Write ;
FROM StrIO IMPORT WriteString, WriteLn ;
IMPORT Args ;
FROM M2Depth IMPORT MakeDependant, GetDepth ;
FROM CmdArgs IMPORT Narg, GetArg ;
FROM NameKey IMPORT GetKey, MakeKey, WriteKey ;


(* %%%FORWARD%%%
PROCEDURE ScanSources ; FORWARD ;
PROCEDURE ScanImport (s: CARDINAL) ; FORWARD ;
PROCEDURE MakeModule (ModuleName: CARDINAL) ; FORWARD ;
PROCEDURE FindSourceFile (s: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE WriteFileName (FileName: ARRAY OF CHAR) ; FORWARD ;
PROCEDURE CalculateDepth ; FORWARD ;
PROCEDURE SortSources ; FORWARD ;
PROCEDURE DisplaySources ; FORWARD ;
PROCEDURE ScanArgs() : BOOLEAN ; FORWARD ;
   %%%FORWARD%%% *)

 
CONST
   MaxStringLen = 8192 ;
   MaxNoOfFiles =  100 ;
   Comment      = '# ' ;  (* Comment identifier *)
 
TYPE
   Source = RECORD
               Name : CARDINAL ;
               Depth: CARDINAL ;
            END ;
 
 
VAR
   Main        : ARRAY [0..MaxStringLen] OF CHAR ;
   SourceNo    : CARDINAL ;
   Sources     : ARRAY [1..MaxNoOfFiles] OF Source ;
   IncludeM2RTS: BOOLEAN ;   (* do we automatically include M2RTS into the top module? *)


(*
   ScanSources - scans all the source files for IMPORTs and places all
                 imports into the Sources array.
*)

PROCEDURE ScanSources ;
VAR
   i: CARDINAL ;
BEGIN
   SourceNo := 1 ;
   WITH Sources[SourceNo] DO
      Name := MakeKey(Main) ;
      Depth := 0
   END ;
   IF IncludeM2RTS
   THEN
      (*
         we should include M2RTS as a dependant module otherwise it simply wont link
      *)
      MakeModule(MakeKey('M2RTS')) ;
      MakeDependant(MakeKey(Main), MakeKey('M2RTS'))
   END ;
   i := 0 ;
   WHILE i<SourceNo DO
      INC(i) ;
      ScanImport(i)
   END
END ScanSources ;


(*
   ScanImport - scans Source[s] for imports. Any imported file is added
                to the Sources array.
*)
 
PROCEDURE ScanImport (s: CARDINAL) ;
VAR
   done : BOOLEAN ;
   mod  : CARDINAL ;
BEGIN
   IF FindSourceFile(s)
   THEN
      done := FALSE ;
      REPEAT
         GetSymbol ;
         WHILE (NOT IsSym('FROM')) AND (NOT IsSym('END')) AND
               (NOT IsSym('IMPORT')) AND (NOT IsSym(';'))
         DO
            GetSymbol
         END ;
         IF IsSym('FROM')
         THEN
            GetSymbol ;
            MakeModule(MakeKey(CurrentSymbol)) ;
            MakeDependant(Sources[s].Name, MakeKey(CurrentSymbol)) ;
            (* Eat up to end of IMPORT line *)
            WHILE NOT IsSym(';') DO
               GetSymbol
            END
         ELSIF IsSym('IMPORT')
         THEN
            REPEAT
               GetSymbol ;
               MakeModule(MakeKey(CurrentSymbol)) ;
               MakeDependant(Sources[s].Name, MakeKey(CurrentSymbol)) ;
               GetSymbol
            UNTIL IsSym(';')
         ELSIF IsSym('BEGIN') OR IsSym('END') OR IsSym('VAR') OR IsSym('TYPE')
         THEN
            done := TRUE
         END
      UNTIL done ;
      CloseSource
   END
END ScanImport ;


(*
   MakeModule - makes a module for ModuleName.
*)

PROCEDURE MakeModule (ModuleName: CARDINAL) ;
VAR
   i    : CARDINAL ;
   Found: BOOLEAN ;
BEGIN
   Found := FALSE ;
   i := 1 ;
   WHILE (i<=SourceNo) AND (NOT Found) DO
      WITH Sources[i] DO
         IF Name=ModuleName
         THEN
            Found := TRUE
         ELSE
            INC(i)
         END
      END
   END ;
   IF NOT Found
   THEN
      (* Not found so create *)
      INC(SourceNo) ;
      WITH Sources[SourceNo] DO
         Name := ModuleName ;
         Depth := 0
      END
   END
END MakeModule ;


(*
   FindSourceFile - attempts to find the source file for source, s.
*)

PROCEDURE FindSourceFile (s: CARDINAL) : BOOLEAN ;
VAR
   FileName,
   Path    : ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   GetKey(Sources[s].Name, FileName) ;
   WriteString(Comment) ;
   CalculateFileName(FileName, 'mod', FileName) ;
   IF M2Search.FindSourceFile(FileName, Path) AND OpenSource(Path)
   THEN
      WriteString('Module ') ; WriteFileName(FileName) ;
      WriteString(' In file: ') ; WriteString(Path) ; WriteLn ;
      RETURN( TRUE )
   ELSE
      (* WriteString(Path) ; WriteString(' : ') ; *)
   END ;
   RETURN( FALSE )
END FindSourceFile ;


(*
   WriteFileName - displays a file name, FileName, with formatted spaces
                   after the string.
*)

PROCEDURE WriteFileName (FileName: ARRAY OF CHAR) ;
CONST
   MaxSpaces = 20 ;
VAR
   i, High: CARDINAL ;
BEGIN
   WriteString(FileName) ;
   High := StrLen(FileName) ;
   IF High<MaxSpaces
   THEN
      High := MaxSpaces - High ;
      i := 0 ;
      WHILE i<High DO
         Write(' ') ;
         INC(i)
      END
   END
END WriteFileName ;


(*
   CalculateDepth - Calculates the modules depth.
*)

PROCEDURE CalculateDepth ;
VAR
   i: CARDINAL ;
BEGIN
   i := 1 ;
   WHILE i<=SourceNo DO
      WITH Sources[i] DO
         Depth := GetDepth(Name)
      END ;
      INC(i)
   END
END CalculateDepth ;


(*
   SortSources - sorts the module sources into the order of runtime
                 initialization.
*)

PROCEDURE SortSources ;
VAR
   i, j: CARDINAL ;
   t   : Source ;
BEGIN
   i := 1 ;
   WHILE i<=SourceNo DO
      j := 1 ;
      WHILE j<=SourceNo DO
         IF (i#j) AND (Sources[i].Depth>Sources[j].Depth)
         THEN
            t := Sources[i] ;
            Sources[i] := Sources[j] ;
            Sources[j] := t
         END ;
         INC(j)
      END ;
      INC(i)
   END
END SortSources ;


(*
   DisplaySources - displays the source names 1..SourceNo.
*)

PROCEDURE DisplaySources ;
VAR
   i: CARDINAL ;
   a: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   i := 1 ;
   WHILE i<=SourceNo DO
      WITH Sources[i] DO
         GetKey(Name, a) ;
         WriteString(Comment) ;
         WriteFileName(a) ; WriteCard(Depth, 4) ; WriteLn
      END ;
      INC(i)
   END ;
   WriteString(Comment) ; WriteLn ;
   WriteString(Comment) ; WriteString('Initialization order') ; WriteLn ;
   WriteString(Comment) ; WriteLn ;
   i := 1 ;
   WHILE i<=SourceNo DO
      WITH Sources[i] DO
         WriteKey(Name) ; WriteLn
      END ;
      INC(i)
   END
END DisplaySources ;


(*
   ScanArgs - scans the argument list and returns TRUE if the main source
              module is found.
*)

PROCEDURE ScanArgs () : BOOLEAN ;
VAR
   i, n: CARDINAL ;
   a   : ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   IncludeM2RTS := TRUE ;
   StrCopy('', Main) ;
   n := Args.Narg() ;
   IF n=1
   THEN
      WriteString('Usage: gm2l [ -M searchpath ] modulename') ; WriteLn
   ELSE
      i := 1 ;
      REPEAT
         IF Args.GetArg(a, i)
         THEN
            (* WriteString('Argument found: ') ; WriteString(a) ; WriteLn ; *)
            IF StrEqual(a, '-M')
            THEN
               INC(i) ;
               IF Args.GetArg(a, i)
               THEN
                  M2Search.SetSearchPath(a)
               ELSE
                  WriteString('Expecting path after -M flag') ; WriteLn
               END
               (* ; WriteString('Paths are: ') ; WriteString(a) ; WriteLn *)
            ELSIF StrEqual(a, '-M2RTS')
            THEN
               (* no M2RTS to be automatically linked imported from the main module *)
               IncludeM2RTS := FALSE
            ELSE
               StrCopy(a, Main)
            END
         END ;
         INC(i)
      UNTIL i>n
   END ;
   RETURN( NOT StrEqual('', Main) )
END ScanArgs ;


BEGIN
   IF ScanArgs()
   THEN
      ScanSources ;
      CalculateDepth ;
      SortSources ;
      DisplaySources
   END
END gm2l.
