MODULE gm2lcc ;

(*
   Author     : Gaius Mulley
   Title      : gm2lcc
   Date       : Fri Jul 24 11:45:08 BST 1992
   Description: Generates the cc command for linking the suit of modules.
   Last update: Fri Sep 15 14:42:38 BST 1989
*)

FROM libc IMPORT system, exit ;
FROM NameKey IMPORT MakeKey, WriteKey ;
FROM M2Search IMPORT SetSearchPath, FindSourceFile ;
FROM M2FileName IMPORT CalculateFileName ;
FROM Args IMPORT GetArg ;
FROM StrLib IMPORT StrEqual, StrLen, StrCopy ;
FROM StrIO IMPORT ReadString, WriteString, WriteLn ;
FROM StdIO IMPORT Write ;
FROM ASCII IMPORT eof ;  (* UNIX *)


(* %%%FORWARD%%%
PROCEDURE ScanSources ; FORWARD ;
PROCEDURE ScanImport (s: CARDINAL) ; FORWARD ;
PROCEDURE MakeModule (ModuleName: CARDINAL) ; FORWARD ;
PROCEDURE WriteFileName (FileName: ARRAY OF CHAR) ; FORWARD ;
PROCEDURE CalculateDepth ; FORWARD ;
PROCEDURE SortSources ; FORWARD ;
PROCEDURE DisplaySources ; FORWARD ;
PROCEDURE ScanArgs() : BOOLEAN ; FORWARD ;
   %%%FORWARD%%% *)

CONST
   MaxName     = 255  ; (* Length of an identifier *)
(*
   eof         =  32C ; (* Dos                     *)
*)
   Comment     = '#'  ; (* Comment identifier.     *)
   MaxArchives = 1000 ;

VAR
   ApuFound    : BOOLEAN ;
   DebugFound  : BOOLEAN ;
   CheckFound  : BOOLEAN ;
   ProfileFound: BOOLEAN ;
   Target      : ARRAY [0..MaxName] OF CHAR ;
   TargetFound : BOOLEAN ;
   PathFound   : BOOLEAN ;
   Path        : ARRAY [0..MaxName] OF CHAR ;
   StartupFound: BOOLEAN ;
   StartupFile : ARRAY [0..MaxName] OF CHAR ;
   Archives    : ARRAY [0..MaxArchives] OF CARDINAL ;
   NoOfArchives: CARDINAL ;



(*
   AddArchive - adds an archive into the archive list.
*)

PROCEDURE AddArchive (Name: CARDINAL) ;
BEGIN
   IF NoOfArchives=MaxArchives
   THEN
      WriteString('inrease MaxArchives in gm2lcc.mod') ; WriteLn ;
      exit(1)
   ELSE
      Archives[NoOfArchives] := Name ;
      INC(NoOfArchives)
   END
END AddArchive ;


(*
   GenerateCommand - generate the appropriate linkage command
                     with the correct options.
*)

PROCEDURE GenerateCommand ;
BEGIN
   IF StartupFound
   THEN
      IF ApuFound
      THEN
         WriteString('ld-is32 -Ttext 0x0 ')
      ELSE
         WriteString('ld -A 0000 -i ')
      END ;
      WriteString(StartupFile)
   ELSIF ApuFound
   THEN
      WriteString('ld-is32 -Ttext 0x0 ')
   ELSE
      WriteString('gcc')
   END ;
   IF DebugFound
   THEN
      WriteString(' -g')
   END ;
   IF ProfileFound
   THEN
      WriteString(' -p')
   END ;
   IF TargetFound
   THEN
      WriteString(' -o ') ;
      WriteString(Target)
   END ;
   IF ProfileFound
   THEN
      WriteString(' -lgmon')
   END
END GenerateCommand ;


(*
   GenCC - writes out the linkage command for the C compiler.
*)

PROCEDURE GenCC ;
VAR
   a, b: ARRAY [0..MaxName] OF CHAR ;
   i   : CARDINAL ;
BEGIN
   GenerateCommand ;
   IF ApuFound
   THEN
      WriteString(' crt0.obj mod_init.obj')
   ELSE
      WriteString(' mod_init.o')
   END ;
   ReadString(a) ;
   WHILE a[0]#eof DO
      IF (a[0]#Comment) AND (NOT StrEqual(a, ''))
      THEN
         StrCopy(a, b) ;
         IF ApuFound
         THEN
            CalculateFileName(a, 'obj', a)
         ELSE
            CalculateFileName(a, 'o', a)
         END ;
         IF FindSourceFile(a, a)
         THEN
            WriteString('  ') ;
            WriteString(a)
         ELSE
            CalculateFileName(b, 'a', b) ;
            IF FindSourceFile(b, b)
            THEN
               AddArchive(MakeKey(b))
            ELSE
               WriteString('# have not found .o or .a ') ; WriteString(b) ; WriteLn ;
               exit(1)
            END
         END
      END ;
      ReadString(a)
   END ;
   i := 0 ;
   WHILE i<NoOfArchives DO
      WriteString('  ') ;
      WriteKey(Archives[i]) ;
      INC(i)
   END ;
   WriteLn
END GenCC ;


(*
   WriteModuleName - displays a module name, ModuleName, with formatted spaces
                     after the string.
*)

PROCEDURE WriteModuleName (ModuleName: ARRAY OF CHAR) ;
CONST
   MaxSpaces = 20 ;
VAR
   i, High: CARDINAL ;
BEGIN
   WriteString(ModuleName) ;
   High := StrLen(ModuleName) ;
   IF High<MaxSpaces
   THEN
      High := MaxSpaces - High ;
      i := 0 ;
      WHILE i<High DO
         Write(' ') ;
         INC(i)
      END
   END
END WriteModuleName ;


(*
   CheckCC - checks to see whether all the object files can be found
             for each module.
*)

PROCEDURE CheckCC ;
VAR
   a, b : ARRAY [0..MaxName] OF CHAR ;
   Error: INTEGER ;
BEGIN
   Error := 0 ;
   ReadString(a) ;
   WHILE a[0]#eof DO
      IF (a[0]#Comment) AND (NOT StrEqual(a, ''))
      THEN
         WriteString('Module ') ;
         WriteModuleName(a) ;
         StrCopy(a, b) ;
         IF ApuFound
         THEN
            CalculateFileName(a, 'obj', a)
         ELSE
            CalculateFileName(a, 'o', a)
         END ;
         IF FindSourceFile(a, a)
         THEN
            WriteString(' : ') ;
            WriteString(a)
         ELSE
            CalculateFileName(b, 'a', b) ;
            IF FindSourceFile(b, b)
            THEN
               WriteString(' : ') ;
               WriteString(b)
            ELSE
               WriteString(' : Not found') ;
               Error := 1
            END
         END ;
         WriteLn
      END ;
      ReadString(a)
   END ;
   exit(Error)
END CheckCC ;


(*
   ProcessTarget - copies the specified target file into Target
                   and sets the boolean TargetFound.
*)

PROCEDURE ProcessTarget (i: CARDINAL) ;
BEGIN
   IF NOT GetArg(Target, i+1)
   THEN
      HALT  (* Should always be able to get argument *)
   END ;
   TargetFound := TRUE
END ProcessTarget ;


(*
   ProcessStartupFile - copies the specified startup file name into StartupFile
                        and sets the boolean StartupFound.
*)

PROCEDURE ProcessStartupFile (i: CARDINAL) ;
BEGIN
   IF NOT GetArg(StartupFile, i+1)
   THEN
      HALT  (* Should always be able to get argument *)
   END ;
   StartupFound := TRUE
END ProcessStartupFile ;


(*
   ScanArguments - scans arguments for flags: -M -g
*)

PROCEDURE ScanArguments ;
VAR
   a: ARRAY [0..MaxName] OF CHAR ;
   i: CARDINAL ;
BEGIN
   i := 1 ;
   WHILE GetArg(a, i) DO
      IF StrEqual(a, '-apu')
      THEN
         ApuFound := TRUE
      ELSIF StrEqual(a, '-g')
      THEN
         DebugFound := TRUE
      ELSIF StrEqual(a, '-c')
      THEN
         CheckFound := TRUE
      ELSIF StrEqual(a, '-p')
      THEN
         ProfileFound := TRUE
      ELSIF StrEqual(a, '-M')
      THEN
         INC(i) ;
         IF GetArg(a, i)
         THEN
            SetSearchPath(a)
         END
      ELSIF StrEqual(a, '-o')
      THEN
         ProcessTarget(i) ;
         INC(i)                   (* Target found *)
      ELSIF StrEqual(a, '-startup')
      THEN
         ProcessStartupFile(i) ;
         INC(i)                   (* Target found *)
      END ;
      INC(i)
   END
END ScanArguments ;


(*
   Init - initializes the global variables.
*)

PROCEDURE Init ;
BEGIN
   NoOfArchives := 0 ;
   DebugFound := FALSE ;
   CheckFound := FALSE ;
   TargetFound := FALSE ;
   ProfileFound := FALSE ;
   ApuFound := FALSE ;
   ScanArguments ;
   IF CheckFound
   THEN
      CheckCC
   ELSE
      GenCC
   END
END Init ;


BEGIN
   Init
END gm2lcc.
