MODULE gm2lgen ;

(*
   Author     : Gaius Mulley
   Title      : gm2lgen
   Date       : Fri Sep 15 14:42:17 BST 1989
   Description: Generates the C main.c code, from a list of module names.
   Last update: Fri Sep 15 14:42:38 BST 1989
*)

FROM StrLib IMPORT StrEqual, StrCopy ;
FROM StrIO IMPORT ReadString, WriteString, WriteLn ;
FROM StdIO IMPORT Write ;
FROM ASCII IMPORT eof ;
FROM Args IMPORT GetArg ;


(* %%%FORWARD%%%
PROCEDURE GenInitializationCalls (ApuFound: BOOLEAN) ; FORWARD ;
   %%%FORWARD%%% *)


CONST
   MaxName = 255  ; (* Length of an identifier *)
   Comment = '#'  ; (* Comment identifier      *)


(*
   GenMain - writes out the main() function together with module initialization
             calls.
*)

PROCEDURE GenMain ;
VAR
   ApuFound: BOOLEAN ;
   i       : CARDINAL ;
   a, b    : ARRAY [0..MaxName] OF CHAR ;
BEGIN
   i := 1 ;
   ApuFound := FALSE ;
   StrCopy('main', b) ;
   WHILE GetArg(a, i) DO
      IF StrEqual(a, '-apu')
      THEN
         ApuFound := TRUE
      ELSE
         StrCopy(a, b)
      END ;
      INC(i)
   END ;
   IF ApuFound
   THEN
      WriteString('    .global ') ; Write('_') ; WriteString(b) ; WriteLn ;
      Write('_') ; WriteString(b) ; Write(':') ; WriteLn ;
      WriteString('    load.d.d  %ra') ; WriteLn ;
      WriteString('    load.d.d  %sp') ; WriteLn ;
      WriteString('    store.d.d %ra') ; WriteLn ;
      WriteString('    load.d.d  %rb') ; WriteLn ;
      WriteString('    load.d.d  %rc') ; WriteLn ;  (* so gdb can get hold of this function *)
      GenInitializationCalls(ApuFound) ;
      WriteString('    store.d.d %rc') ; WriteLn ;
      WriteString('    store.d.d %rb') ; WriteLn ;
      WriteString('    store.d.d %ra') ; WriteLn ;
      WriteString('    store.d.d %pc') ; WriteLn ;
      WriteString('    nop') ; WriteLn ;            (* just so our PC never goes our of range *)
      WriteString('    .end') ; WriteLn
   ELSE
      WriteString(b) ; WriteString('(argc, argv)') ; WriteLn ;
      WriteString('int   argc ;') ; WriteLn ;
      WriteString('char  *argv[];') ; WriteLn ;
      WriteString('{') ; WriteLn ;
      GenInitializationCalls(ApuFound) ;
      WriteString('   ') ; WriteString('libc_exit(0);') ; WriteLn ;
      WriteString('}') ; WriteLn
   END
END GenMain ;


(*
   GenInitializationCalls - writes out the initialization calls for the modules
                            in the application suit.
*)

PROCEDURE GenInitializationCalls (ApuFound: BOOLEAN) ;
VAR
   a: ARRAY [0..MaxName] OF CHAR ;
BEGIN
   ReadString(a) ;
   WHILE a[0]#eof DO
      IF (a[0]#Comment) AND (NOT StrEqual(a, ''))
      THEN
         IF ApuFound
         THEN
            WriteString('  .extern __M2_') ; WriteString(a) ; WriteString('_init') ; WriteLn ;
            WriteString('  load.d.d $__M2_') ; WriteString(a) ; WriteString('_init') ; WriteLn ;
            WriteString('  call') ; WriteLn
         ELSE
            WriteString('   _M2_') ;
            WriteString(a) ; WriteString('_init(argc, argv) ;') ; WriteLn
         END
      END ;
      ReadString(a)
   END
END GenInitializationCalls ;


BEGIN
   GenMain
END gm2lgen.
