IMPLEMENTATION MODULE i486 ;


FROM SymbolTable IMPORT IsConst, IsProcedure, IsVar, GetType,
                        PushSize, PushVarSize, PushOffset,
                        MakeTemporary, PutVar,
                        StartScope, EndScope,
                        GetFirstUsed, ModeOfAddr,
                        GetVarFather ;

FROM NameKey IMPORT MakeKey ;
FROM M2Batch IMPORT MakeDefinitionSource ;
FROM M2Options IMPORT WordLength ;
FROM M2ALU IMPORT PopInt, PopCard ;
FROM M2System IMPORT WordSize, Word ;
FROM M2Lexical IMPORT NearToken, InternalError ;


VAR
   MaxDatumValue : DatumPrefix ;


(*
   MaxDatum - returns the maximum size of data that can be contained in
              a register.
*)

PROCEDURE MaxDatum () : DatumPrefix ;
BEGIN
   RETURN( MaxDatumValue )
END MaxDatum ;


(*
   GetUnitOfTransfer - returns the number of bytes that can be transferred
                       at a time, if Size bytes are to be transferred.
*)

PROCEDURE GetUnitOfTransfer (Size: CARDINAL) : CARDINAL ;
BEGIN
   IF Size=1
   THEN
      RETURN( BYTESIZE )
   ELSIF (Size=2) OR (Size=3)
   THEN
      RETURN( WORDSIZE )
   ELSIF Size>3
   THEN
      RETURN( WordSize )
   ELSE
      InternalError('Cannot transfer 0 bytes of memory!', __FILE__, __LINE__)
   END
END GetUnitOfTransfer ;


(*
   GetSizeOfPrefix - returns the number of bytes associated with
                     the prefix, p.
*)

PROCEDURE GetSizeOfPrefix (p: DatumPrefix) : CARDINAL ;
BEGIN
   CASE p OF

   PByte :  RETURN( BYTESIZE ) |
   PWord :  RETURN( WORDSIZE ) |
   PDWord:  RETURN( DWORDSIZE )

   ELSE
      InternalError('unexpected prefix size', __FILE__, __LINE__)
   END
END GetSizeOfPrefix ;


(*
   GetScope - returns the scope of Sym.
              The scope is either Local or Global.
*)

PROCEDURE GetScope (Sym: CARDINAL) : KindOfScope ;
BEGIN
   IF IsConst(Sym) OR IsProcedure(Sym)
   THEN
      RETURN( NoScope )
   ELSIF IsProcedure(GetVarFather(Sym))
   THEN
      RETURN( Local )
   ELSE
      RETURN( Global )
   END
END GetScope ;


(*
   GetOffset - returns the offset for a symbol, Sym.
*)

PROCEDURE GetOffset (Sym: CARDINAL) : INTEGER ;
BEGIN
   IF IsConst(Sym) OR IsProcedure(Sym)
   THEN
      RETURN( 0 )
   ELSE
      PushOffset(Sym) ;
      RETURN( PopInt() )
   END
END GetOffset ;


(*
   GetSize - returns the size of a symbol, Sym.
             returns the amount of space variable, Sym, requires.
             NOTE this is not the same as GetVarTypeSize since
                  if Sym is LeftValue then the size will be
                  TSIZE(ADDRESS) since Sym is in fact a "pointer".
                  ie VAR parameter...
*)

PROCEDURE GetSize (Sym: CARDINAL) : CARDINAL ;
BEGIN
   IF IsVar(Sym)
   THEN
      PushVarSize(Sym)
   ELSE
      PushSize(Sym)
   END ;
   RETURN( PopCard() )
END GetSize ;


(*
   GetUnitPrefix - returns the prefix mode which corresponds to size.
*)

PROCEDURE GetUnitPrefix (Size: CARDINAL) : DatumPrefix ;
BEGIN
   IF WordLength=32
   THEN
      CASE Size OF

      BYTESIZE : RETURN( PByte ) |
      WORDSIZE : RETURN( PWord ) |
      DWORDSIZE: RETURN( PDWord )

      ELSE
         InternalError('no prefix for this size', __FILE__, __LINE__)
      END
   ELSE
      CASE Size OF

      BYTESIZE : RETURN( PByte ) |
      WORDSIZE : RETURN( PWord )

      ELSE
         InternalError('no prefix for this size using 16 bit code generation', __FILE__, __LINE__)
      END
   END
END GetUnitPrefix ;


(*
   GetVarTypeSize - returns the size of a variable symbol, Sym, type.
*)

PROCEDURE GetVarTypeSize (Sym: CARDINAL) : CARDINAL ;
BEGIN
   IF IsVar(Sym)
   THEN
      RETURN( GetSize(GetType(Sym)) )
   ELSE
      NearToken('Expecting Sym to be a variable', GetFirstUsed(Sym))
   END
END GetVarTypeSize ;


(*
   EqualSizes - returns true if the two symbols have the same size.
*)

PROCEDURE EqualSizes (Sym1, Sym2: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsConst(Sym1) OR IsConst(Sym2)
   THEN
      RETURN( TRUE )   (* Constants can be compared *)
   ELSE
      RETURN( GetVarTypeSize(Sym1)=GetVarTypeSize(Sym2) )
   END
END EqualSizes ;


(*
   FitsInRegister - returns true if Sym will fit within a register.
*)

PROCEDURE FitsInRegister (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsConst(Sym)
   THEN
      RETURN( TRUE )   (* Constants will match any register size *)
   ELSE
      RETURN( GetSize(Sym) <= WordSize )
   END
END FitsInRegister ;


(*
   LowReg - translates the symbol table representation of a register into
            the enumerated type Register.
*)

PROCEDURE LowReg (Sym: CARDINAL) : Register ;
BEGIN
   IF Sym=EAX
   THEN
      RETURN( RegEAX )
   ELSIF Sym=EBX
   THEN
      RETURN( RegEBX )
   ELSIF Sym=ECX
   THEN
      RETURN( RegECX )
   ELSIF Sym=EDX
   THEN
      RETURN( RegEDX )
   ELSIF Sym=ESI
   THEN
      RETURN( RegESI )
   ELSIF Sym=EDI
   THEN
      RETURN( RegEDI )
   ELSIF Sym=ESP
   THEN
      RETURN( RegESP )
   ELSIF Sym=EBP
   THEN
      RETURN( RegEBP )
   ELSE
      InternalError('this register does not exist', __FILE__, __LINE__)
   END
END LowReg ;


(*
   IsRegister - returns TRUE if a symbol, sym, is a register.
*)

PROCEDURE IsRegister (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN(
          (Sym=EAX) OR (Sym=EBX) OR (Sym=ECX) OR (Sym=EDX) OR
          (Sym=ESP) OR (Sym=EBP) OR (Sym=ESI) OR (Sym=EDI)
         )
END IsRegister ;


(*
   HighReg - translates the i486 register representation
             into symbol table register representation
*)

PROCEDURE HighReg (r: Register) : CARDINAL ;
BEGIN
   CASE r OF

   RegEAX:  RETURN( EAX ) |
   RegEBX:  RETURN( EBX ) |
   RegECX:  RETURN( ECX ) |
   RegEDX:  RETURN( EDX ) |
   RegESI:  RETURN( ESI ) |
   RegEDI:  RETURN( EDI ) |
   RegESP:  RETURN( ESP ) |
   RegEBP:  RETURN( EBP )

   ELSE
      InternalError('this register does not exist', __FILE__, __LINE__)
   END
END HighReg ;


(*
   Init - initializes global module data.
*)

PROCEDURE Init ;
VAR
   System: CARDINAL ;
BEGIN
   IF WordLength=32
   THEN
      MaxDatumValue := PDWord
   ELSE
      MaxDatumValue := PWord
   END ;

   (* use SYSTEM module to create our registers *)
   System := MakeDefinitionSource(MakeKey('SYSTEM')) ;
   StartScope(System) ;

   EAX := MakeTemporary(RightValue) ; PutVar(EAX, Word) ;
   EBX := MakeTemporary(RightValue) ; PutVar(EBX, Word) ;
   ECX := MakeTemporary(RightValue) ; PutVar(ECX, Word) ;
   EDX := MakeTemporary(RightValue) ; PutVar(EDX, Word) ;
   ESP := MakeTemporary(RightValue) ; PutVar(ESP, Word) ;
   EBP := MakeTemporary(RightValue) ; PutVar(EBP, Word) ;
   ESI := MakeTemporary(RightValue) ; PutVar(ESI, Word) ;
   EDI := MakeTemporary(RightValue) ; PutVar(EDI, Word) ;

   EndScope

END Init ;


BEGIN
   Init
END i486.
