MODULE m2m ;
 
(*
   Author     : Gaius Mulley
   Title      : M2M
   Date       : 3/4/86  [$Date: 1997/07/11 08:03:53 $  $Time$]
   SYSTEM     : UNIX SUN and Logitech M2
   Description: Implements the input to Make
   Version    : $Revision: 1.4 $
   Log        : $Log: m2m.mod,v $
   Log        : Revision 1.4  1997/07/11 08:03:53  gaius
   Log        : altered gm2l.mod to contain an 8k char path variable and not 79 chars.
   Log        :
   Log        : Revision 1.3  1996/09/16 07:43:59  gaius
   Log        : compiler now checks function parameters and return values.
   Log        : Fixed bug in -apu code generator brought about via CSE
   Log        :
   Log        : Revision 1.2  1996/07/26 11:47:56  gaius
   Log        : error reporting improved and many bugfixes made. The compiler now
   Log        : generally points to the symbol in error. Many modules now use
   Log        : the WriteErrorFormat1 or its varient (pseudo printf style).
   Log        :
   Log        : Found a bug in the case handling, which was somehow was introduced..
   Log        :
   Log        : Added M2Constants which allow for constant comparison, really
   Log        : helps M2SubExp reduce the number of quadruples. 12%..15% on average code.
   Log        : Speed improvements of up to 25% on the compiler..
   Log        :
   Log        : Common subexpression elimiation is now included when -O is present.
   Log        : -Ocse turns it off.
   Log        :
   Log        : Revision 1.1.1.1  1996/05/28 10:13:02  gaius
   Log        : Modula-2 compiler sources imported
   Log        :
*)
 
FROM ASCII IMPORT nul, tab ;
FROM Storage IMPORT ALLOCATE ;
FROM M2Lex IMPORT GetSymbol, OpenSource, CloseSource, CurrentSymbol,
                  LastSymbol ;
FROM StrLib IMPORT StrEqual, StrCopy, StrConCat, StrLen ;
FROM NumberIO IMPORT WriteCard ;
FROM StdIO IMPORT Write ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM M2FileName IMPORT CalculateFileName, CalculateStemName ;
FROM M2Search IMPORT FindSourceFile, SetSearchPath ;
FROM Args IMPORT Narg, GetArg ;
FROM M2Defaults IMPORT GetOptions, GetSearchPath ;
FROM NameKey IMPORT WriteKey, MakeKey, GetKey, NulName ;
FROM SymbolKey IMPORT SymbolTree, InitTree, GetSymKey, PutSymKey, NulKey ;
FROM libc IMPORT exit ;


CONST
   MaxStringLen = 8192 ;
   MaxNoOfSons  =  100 ;
   MaxNoOfFiles =  500 ;
   MaxLineBreak =   70 ;     (* Any line beyond this is broken into *)
                             (* multiple lines.                     *)
   eof          = 032C ;     (* end of file character               *)

   IsSymFile    = FALSE ;    (* Compile time option *)
 
TYPE
   PtrToNode = POINTER TO Node ;
   Node      = RECORD
                  Index    : CARDINAL ;  (* Index into Source *)
                  NoOfSons : CARDINAL ;
                  Sons     : ARRAY [0..MaxNoOfSons] OF PtrToNode ;
                  Father   : PtrToNode ; (* Father of this node *)
               END ;

   (* A Son field element may be NIL, in which case the son has been *)
   (* replaced in another part of the dependancy tree.               *)

   Source    = RECORD
                  Name       : CARDINAL ;
                  Ext        : CARDINAL ; 
                  FullPath   : CARDINAL ;
                  Date       : CARDINAL ;
                  Modify     : BOOLEAN ;
                  Depth      : CARDINAL ;
                  SourceNode : PtrToNode ;  (* Node which corresponds to *)
                                            (* source.                   *)
               END ;
 
 
VAR
   main         : PtrToNode ;
   FileTable    : SymbolTree ; (* Table of each file in files. Name -> index *)
   DefExistsTree: SymbolTree ; (* Table of definition modules found or not   *)
   ModExistsTree: SymbolTree ; (* Table of imp/program modules found or not  *)
   FileNo       : CARDINAL ;
   Files        : ARRAY [1..MaxNoOfFiles] OF Source ;
   ExeName,
   ObjName,
   SymName,
   DefName,
   ModName      : CARDINAL ;
   KeyFound,
   KeyNotFound  : CARDINAL ;
   CurrentArg   : CARDINAL ;
   GM2Path,
   M2MPath,
   DestDirectory: ARRAY [0..MaxStringLen] OF CHAR ;
   GM2PathFound,
   M2MPathFound : BOOLEAN ;
   DependOnly   : BOOLEAN ;


(*
   NewNode - creates a new node, n, with zero sons.
*)
 
PROCEDURE NewNode (VAR n: PtrToNode) ;
BEGIN
   NEW( n ) ;
   WITH n^ DO
      NoOfSons := 0 ;
   END
END NewNode ;
 

(*
   NewSource - creates a new source file, s.
*)
 
PROCEDURE NewSource (VAR s: CARDINAL) ;
BEGIN
   IF FileNo=MaxNoOfFiles
   THEN
      WriteString('Too many source files [increase MaxNoOfFiles]') ;
      WriteLn ;
      HALT
   ELSE
      INC( FileNo ) ;
      s := FileNo
   END
END NewSource ;
 

(*
   MakeExe - makes a top level node and attempts to import any
             dependancies from this source file and it then increases
             all dependancies by this top level.
*)
 
PROCEDURE MakeExe (ModuleName: CARDINAL; n: PtrToNode; Level: CARDINAL) ;
VAR
   i, j : CARDINAL ;
   t    : PtrToNode ;
BEGIN
   MakeObj(ModuleName, n, Level+1 ) ;
   i := n^.NoOfSons ;
   ImportViaMod(ModuleName, ObjName, n, Level+1) ;
   j := i+1 ;
   WHILE j<=n^.NoOfSons DO
      t := n^.Sons[j] ;
      IF (t#NIL) AND (t^.Father=n)
      THEN
         Obj( Files[t^.Index].Name, t, Level+1 )
      END ;
      INC( j )
   END
END MakeExe ;


(*
   MakeObj - makes an object level node.
*)

PROCEDURE MakeObj (ModuleName: CARDINAL; n: PtrToNode; Level: CARDINAL) ;
VAR
   son: PtrToNode ;
BEGIN
   IF CreateSon(son, ModuleName, ObjName, n, Level+1)
   THEN
      IF ModExists(ModuleName)
      THEN
         MakeMod(ModuleName, son, Level+1)
      END
   END
END MakeObj ;
 

(*
   MakeMod - make a module level node.
*)

PROCEDURE MakeMod (ModuleName: CARDINAL; n: PtrToNode; Level: CARDINAL) ;
VAR
   son : PtrToNode ;
   i, j: CARDINAL ;
   t   : PtrToNode ;
BEGIN
   IF CreateSon(son, ModuleName, ModName, n, Level+1 )
   THEN
      IF ModExists(ModuleName)
      THEN
         i := n^.NoOfSons ;
         MakeSym(ModuleName, n, Level) ;  (* Bypasses .sym is NOT IsSymFile *)
         IF IsSymFile
         THEN
            ImportViaMod(ModuleName, SymName, n, Level+1)
         ELSE
            ImportViaMod(ModuleName, DefName, n, Level+1)
         END ;
         j := i+1 ;
         WHILE j<=n^.NoOfSons DO
            t := n^.Sons[j] ;
            IF (t#NIL) AND (t^.Father=n)
            THEN
               IF IsSymFile
      	       THEN
               	  Sym( Files[t^.Index].Name, t, Level+1 )
      	       ELSE
               	  Sym( Files[t^.Index].Name, t, Level+1 )  (* Really use Sym *)
      	       END ;
               (*
      	       	  As an aside - we check for .lnk into the symbol table and
      	       	  graph of dependency. We also now need to check that .lnk is
                  also in the graph of dependency, since we can only LINK the
                  final main module when ALL implementation modules have been
                  compiled. Defs can be in the dependency without their Imps
                  using a tree strategy. We must ensure that the LNK's are
                  also included.
      	       *)
               MakeObj( Files[t^.Index].Name, main, 1 )
               (* Attach to main.lod for lnk *)
            END ;
            INC(j)
         END
      ELSE
         WriteString('Warning no matching ') ;
         WriteKey(ModuleName) ; WriteString('.mod') ;
         WriteLn
      END
   END
END MakeMod ;
 

(*
   MakeSym - make a sym level node.
*)
 
PROCEDURE MakeSym (ModuleName: CARDINAL; n: PtrToNode; Level: CARDINAL) ;
VAR
   son: PtrToNode ;
BEGIN
   (* Only create a .sym file if there is a corresponding .def file *)
   IF DefExists(ModuleName)
   THEN
      IF IsSymFile
      THEN
      	 IF CreateSon(son, ModuleName, SymName, n, Level+1)
      	 THEN
            MakeDef(ModuleName, son, Level+1)
      	 END
      ELSE
      	 IF CreateSon(son, ModuleName, DefName, n, Level+1)
      	 THEN
            MakeDef(ModuleName, son, Level+1)
      	 END
      END
   ELSE
      WriteString('# Warning no matching ') ;
      WriteKey(ModuleName) ; WriteString('.def') ; WriteLn
   END
END MakeSym ;


(*
   MakeDef - make a definition module level node.
*)

PROCEDURE MakeDef (ModuleName: CARDINAL; n: PtrToNode; Level: CARDINAL) ;
VAR
   son, t : PtrToNode ;
   i, j   : CARDINAL ;
BEGIN
   IF CreateSon( son, ModuleName, DefName, n, Level+1 )
   THEN
      i := n^.NoOfSons ;
      IF IsSymFile
      THEN
         ImportViaDef(ModuleName, SymName, n, Level)
      ELSE
      	 ImportViaDef(ModuleName, DefName, n, Level)
      END ;
      j := i+1 ;
      WHILE j<=n^.NoOfSons DO
         t := n^.Sons[j] ;
         IF (t#NIL) AND (t^.Father=n)
         THEN
      	    IF IsSymFile
      	    THEN
               Sym( Files[t^.Index].Name, t, Level+1 ) ; (* was +2 ?? *)
      	    ELSE
               Sym( Files[t^.Index].Name, t, Level+1 ) ; (* was +2 ?? *)
      	       (* Really use Sym *)
      	    END ;
            (* As an aside - we check for .lnk into the symbol table and    *)
            (* graph of dependency. We also now need to check that .lnk is  *)
            (* also in the graph of dependency, since we can only LINK the  *)
            (* final main module when ALL implementation modules have been  *)
            (* compiled. Defs can be in the dependency without their Imps   *)
            (* using a tree strategy. We must ensure that the LNK's are     *)
            (* also included.                                               *)
 
            MakeObj(Files[t^.Index].Name, main, 1)
            (* Attach to main.lod for lnk                                   *)
            (* End of the aside!                                            *)
         END ;
         INC( j )
      END
   END
END MakeDef ;
 

(*
   ImportViaDef - find all imports from a definition module.
                  Build up this subtree.
*)
 
PROCEDURE ImportViaDef (ModuleName, ExtName: CARDINAL;
                        n: PtrToNode; Level: CARDINAL) ;
VAR
   son        : PtrToNode ;
   a, b       : ARRAY [0..MaxStringLen] OF CHAR ;
   done, found: BOOLEAN ;
   from, i    : CARDINAL ;
   Symbol     : CARDINAL ;
BEGIN
   done := FALSE ;
   from := n^.NoOfSons ;
   IF NOT DefExists(ModuleName)
   THEN
      HALT
   END ;
   GetKey(ModuleName, a) ;
   CalculateFileName(a, 'def', b) ;
   IF NOT (FindSourceFile(b, b) AND OpenSource(b))
   THEN
      HALT
   END ;
   GetSymbol ;
   REPEAT
      WHILE (CurrentSymbol[0]#eof) AND
            (NOT StrEqual(CurrentSymbol, 'FROM')) AND
            (NOT StrEqual(CurrentSymbol, 'END')) AND
            (NOT (StrEqual(CurrentSymbol, 'IMPORT') AND
                  ChrEqual(LastSymbol, ';')))
      DO
         GetSymbol
      END ;
      IF StrEqual(CurrentSymbol, 'FROM') OR StrEqual(CurrentSymbol, 'IMPORT')
      THEN
         GetSymbol ;
         Symbol := MakeKey(CurrentSymbol) ;
         IF DefExists(Symbol)
         THEN
      	    IF IsSymFile
      	    THEN
               found := CreateSon(son, Symbol, SymName, n, Level+1 )
      	    ELSE
      	       found := CreateSon(son, Symbol, DefName, n, Level+1 )
      	    END
         END
      ELSE
         done := TRUE
      END
   UNTIL done ;
   CloseSource
END ImportViaDef ;
 

(*
   ImportViaMod - creates a dependancy tree from the module, ModuleName and
                  ExtName.
*)

PROCEDURE ImportViaMod (ModuleName, ExtName: CARDINAL ;
                        n: PtrToNode; Level: CARDINAL) ;
VAR
   son        : PtrToNode ;
   a, b       : ARRAY [0..MaxStringLen] OF CHAR ;
   done, found: BOOLEAN ;
   from, i    : CARDINAL ;
   Symbol     : CARDINAL ;
BEGIN
   done := FALSE ;
   from := n^.NoOfSons ;
   IF NOT ModExists(ModuleName)
   THEN
      WriteString('# Fatel Error: ') ;
      WriteString(a) ; WriteString('.mod') ; WriteLn ;
      WriteString('does not exist') ;
      HALT
   END ;
   GetKey(ModuleName, a) ;
   CalculateFileName( a, 'mod', b ) ;
   IF NOT (FindSourceFile(b, b) AND OpenSource(b))
   THEN
      HALT
   END ;
   GetSymbol ;
   REPEAT
      WHILE (CurrentSymbol[0]#eof) AND
            (NOT StrEqual(CurrentSymbol, 'FROM')) AND
            (NOT StrEqual(CurrentSymbol, 'END')) AND
            (NOT (StrEqual(CurrentSymbol, 'IMPORT') AND
                  ChrEqual(LastSymbol, ';')))
      DO
         GetSymbol
      END ;
      IF StrEqual(CurrentSymbol, 'FROM') OR StrEqual(CurrentSymbol, 'IMPORT')
      THEN
         GetSymbol ;
         Symbol := MakeKey(CurrentSymbol) ;
         IF ModExists(Symbol)
         THEN
            found := CreateSon(son, Symbol, ExtName, n, Level+1)
         END
      ELSE
         done := TRUE
      END
   UNTIL done ;
   CloseSource
END ImportViaMod ;
 

(*
   Sym - similar to MakeSym except Sym does not create a son
         as it assumes this has already been done.
         Hence it sends the son as a parameter not the father.
*)
 
PROCEDURE Sym (ModuleName: CARDINAL; son: PtrToNode; Level: CARDINAL) ;
BEGIN
   IF DefExists(ModuleName)
   THEN
      MakeDef(ModuleName, son, Level+1)
   ELSE
      WriteString('# Warning no matching ') ;
      WriteKey(ModuleName) ; WriteString('.def') ; WriteLn
   END
END Sym ;
 

(*
   Obj - creates an object level node if the module, ModuleName.mod,
         exists.
*)

PROCEDURE Obj (ModuleName: CARDINAL; son: PtrToNode; Level: CARDINAL) ;
BEGIN
   IF ModExists(ModuleName)
   THEN
      MakeMod(ModuleName, son, Level+1)
   ELSE
      WriteString('# Warning no matching ') ;
      WriteKey(ModuleName) ; WriteString('.mod') ; WriteLn
   END
END Obj ;


(*
   InFileTable - returns TRUE if ModuleName and ExtName are
                 to be found in the FileTable tree.
*)

PROCEDURE InFileTable (ModuleName, ExtName: CARDINAL;
                       VAR index: CARDINAL) : BOOLEAN ;
VAR
   a, b, c: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   GetKey(ModuleName, a) ;
   GetKey(ExtName, b) ;
   CalculateFileName(a, b, c) ;
   index := GetSymKey(FileTable, MakeKey(c)) ;
   RETURN( index#NulKey )
END InFileTable ;


(*
   CreateSon - returns TRUE if we create a new son.
*)

PROCEDURE CreateSon (VAR son: PtrToNode;
                     ModuleName, ExtName: CARDINAL;
                     n: PtrToNode; Level: CARDINAL) : BOOLEAN ;
VAR
   a, b, c: ARRAY [0..MaxStringLen] OF CHAR ;
   i      : CARDINAL ;
BEGIN
   GetKey(ModuleName, a) ;
   GetKey(ExtName, b) ;
   CalculateFileName(a, b, c) ;
   IF InFileTable(ModuleName, ExtName, i)
   THEN
      IF Files[i].Depth<Level
      THEN
         (* Move SubTree Files[i].SourceNode furthur down the Tree to *)
         (* be a son of n. But since Files[i].SourceNode is higher up *)
         (* the Tree we must make certain that n is not a SubNode of  *)
         (* Files[i].SourceNode. Otherwise we would be in serious     *)
         (* trouble if we attempted to move a SubTree under those     *)
         (* conditions.                                               *)
         IF NOT SubNode( Files[i].SourceNode, n )
         THEN
            WriteString('# Moving subtree') ;
            WriteKey(Files[i].Name) ;
      	    Write('.') ; WriteKey(Files[i].Ext) ;
            WriteString(' to be a SubTree of ') ;
            IF n#NIL
            THEN
               WriteKey(Files[n^.Index].Name) ; Write('.') ;
      	       WriteKey(Files[n^.Index].Ext) ; WriteLn
            ELSE
               WriteLn
            END ;

            (* Move Files[i].SourceNode SubTree to be a son of n.     *)
            (* Need to delete Files[i].SourceNode from its father.    *)
            DeleteFromFather( Files[i].SourceNode ) ;

            (* Delete Psedo sons - only need proper son if available  *)
            DeleteSons( n, Files[i].SourceNode ) ;

            (* Now add Files[i].SourceNode to be a son of n.          *)
            AddSon( n, Files[i].SourceNode ) ;
            Files[i].SourceNode^.Father := n ;

            (* Now need to increase the Depth of Files[i].SourceNode  *)
            (* and children.                                          *)
            (* All need to increased by Level-Files[i].Depth.         *)
            IncreaseLevel( Files[i].SourceNode, Level-Files[i].Depth )
         ELSE
            WriteString('# Found Circular Import: ') ;
            WriteString(c) ; WriteLn
         END
      ELSE
         (* Pseudo son - son does not acknowledge his father in this case *)
         (* need this because father is still dependant upon son. Only   *)
         (* that the son has another father deeper down the tree.        *)
         IF NOT IsASon( n, Files[i].SourceNode )
         THEN
            son := Files[i].SourceNode ;
            AddSon( n, Files[i].SourceNode )
         END
         (* Do not set the father field here for reason above.           *)
      END ;
      RETURN( FALSE )
   ELSE
      NewNode(son) ;
      AddSon(n, son) ;
      son^.Father := n ;
      NewSource(i) ;
      PutSymKey(FileTable, MakeKey(c), i) ;
      WITH Files[i] DO
         Name       := ModuleName ;
         Ext        := ExtName ;
         Depth      := Level ;
         Date       := 0 ;  (* Needs to be looked up really... *)
         Modify     := FALSE ;
         SourceNode := son
      END ;
      son^.Index := i ;
      RETURN( TRUE )
   END
END CreateSon ;


(*
   IsASon - returns TRUE if, son, is a son of, father.
*)

PROCEDURE IsASon (father, son: PtrToNode) : BOOLEAN ;
VAR
   i: CARDINAL ;
   t: PtrToNode ;
   found: BOOLEAN ;
BEGIN
   i := 1 ;
   found := FALSE ;
   WHILE (i<=father^.NoOfSons) AND (NOT found) DO
      t := father^.Sons[i] ;
      found := (t#NIL) AND (t=son) ;
      INC( i )
   END ;
   RETURN( found )
END IsASon ;


(*
   DeleteSons - deletes all sons, son, from, father.
*)

PROCEDURE DeleteSons (father, son: PtrToNode) ;
VAR
   i: CARDINAL ;
BEGIN
   i := 1 ;
   WHILE (i<=father^.NoOfSons) DO
      IF father^.Sons[i]=son
      THEN
         father^.Sons[i] := NIL
      END ;
      INC( i )
   END
END DeleteSons ;


(*
   SubNode - returns TRUE if son is a desendant of father, otherwise
             FALSE is returned.
*)

PROCEDURE SubNode (father, son: PtrToNode) : BOOLEAN ;
VAR
   i     : CARDINAL ;
   found : BOOLEAN ;
   t     : PtrToNode ;
BEGIN
   IF father=son
   THEN
      found := TRUE
   ELSE
      i := 1 ;
      found := FALSE ;
      WHILE (i<=father^.NoOfSons) AND (NOT found) DO
         t := father^.Sons[i] ;
         IF (t#NIL) AND (t^.Father=father)
         THEN
            found := SubNode( t, son )
         END ;
         INC( i )
      END
   END ;
   RETURN( found )
END SubNode ;


(*
   DeleteFromFather - Deletes all entries for son from its direct Father.
*)

PROCEDURE DeleteFromFather (son: PtrToNode) ;
VAR
   father: PtrToNode ;
   i     : CARDINAL ;
BEGIN
   father := son^.Father ;
   IF father#NIL
   THEN
      i := 1 ;
      WHILE i<=father^.NoOfSons DO
         IF father^.Sons[i]=son
         THEN
            father^.Sons[i] := NIL
         END ;
         INC(i)
      END
   END
END DeleteFromFather ;


(*
   IncreaseLevel - adds inc to the Depth field for all n and its SubTree
*)

PROCEDURE IncreaseLevel (n: PtrToNode ; inc: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   INC( Files[n^.Index].Depth, inc ) ;
   i := 1 ;
   WHILE i<=n^.NoOfSons DO
      IF n^.Sons[i]#NIL
      THEN
         IncreaseLevel( n^.Sons[i], inc )
      END ;
      INC( i )
   END
END IncreaseLevel ;


(*
   AddSon - adds son to be a son of father, it only alters the Son array
            and the NoOfSons but no other fields. This procedure should be used
            since it reuses maybe NIL fields within the Son ARRAY.
            Caused by moving subtrees of the dependancy graph.
*)

PROCEDURE AddSon (father, son: PtrToNode) ; 
VAR
  i : CARDINAL ;
BEGIN
   i := 1 ;
   WHILE (i<=father^.NoOfSons) AND (father^.Sons[i]#NIL) DO
      INC(i)
   END ;
   IF i<=father^.NoOfSons
   THEN
      father^.Sons[i] := son   (* Resuse a NIL field *)
   ELSIF father^.NoOfSons<MaxNoOfSons
   THEN
      (* Add son at the end of the ARRAY *)
      INC( father^.NoOfSons ) ;
      father^.Sons[father^.NoOfSons] := son
   ELSE
      (* Run out of space -- Error *)
      WriteString('# m2m: Too many sons --- Too many imports in one file') ; WriteLn ;
      WriteString('# [Increase MaxNoOfSons]') ; WriteLn ;
      HALT
   END
END AddSon ;


(*
   ModExists - returns TRUE if ModuleName.mod exists in a file.
*)

PROCEDURE ModExists (ModuleName: CARDINAL) : BOOLEAN ;
VAR
   a, b : ARRAY [0..MaxStringLen] OF CHAR ;
   found: BOOLEAN ;
   Key  : CARDINAL ;
BEGIN
   GetKey(ModuleName, a) ;
   Key := GetSymKey(ModExistsTree, ModuleName) ;
   IF Key=KeyFound
   THEN
      found := TRUE
   ELSIF Key=KeyNotFound
   THEN
      found := FALSE
   ELSIF Key=NulKey
   THEN
      CalculateFileName(a, 'mod', b) ;
      IF FindSourceFile(b, b)
      THEN
         found := TRUE ;
         PutSymKey(ModExistsTree, MakeKey(a), KeyFound)
      ELSE
         found := FALSE ;
         PutSymKey(ModExistsTree, MakeKey(a), KeyNotFound)
      END
   ELSE
      HALT (* incorrect value of key *)
   END ;
   RETURN( found )
END ModExists ;


(*
   DefExists - returns TRUE if ModuleName.def exists in a file.
*)

PROCEDURE DefExists (ModuleName: CARDINAL) : BOOLEAN ;
VAR
   a, b : ARRAY [0..MaxStringLen] OF CHAR ;
   found: BOOLEAN ;
   Key  : CARDINAL ;
BEGIN
   GetKey(ModuleName, a) ;
   Key := GetSymKey(DefExistsTree, ModuleName) ;
   IF Key=KeyFound
   THEN
      found := TRUE
   ELSIF Key=KeyNotFound
   THEN
      found := FALSE
   ELSIF Key=NulKey
   THEN
      CalculateFileName(a, 'def', b) ;
      IF FindSourceFile(b, b)
      THEN
         found := TRUE ;
         PutSymKey(DefExistsTree, MakeKey(a), KeyFound)
      ELSE
         found := FALSE ;
         PutSymKey(DefExistsTree, MakeKey(a), KeyNotFound)
      END
   ELSE
      HALT (* incorrect value of key *)
   END ;
   RETURN( found )
END DefExists ;


(*
   DisplayTree - dumps the tree in english.
*)

PROCEDURE DisplayTree (n: PtrToNode) ;
VAR
   a, b, c: ARRAY [0..MaxStringLen] OF CHAR ;
   i      : CARDINAL ;
   t      : PtrToNode ;
BEGIN
   WriteKey(Files[n^.Index].Name) ; Write('.') ; WriteKey(Files[n^.Index].Ext) ;
   WriteString(' has sons') ;
   WriteLn ;
   i := 1 ;
   WHILE i<=n^.NoOfSons DO
      t := n^.Sons[i] ;
      IF t#NIL
      THEN
         DisplayTree(t)
      END ;
      INC(i)
   END ;
   GetKey(Files[n^.Index].Name, a) ;
   GetKey(Files[n^.Index].Ext , b) ;
   CalculateFileName(a, b, c) ;
   WriteString(c) ;
   WriteString(' at level: ') ;
   i := GetSymKey(FileTable, MakeKey(c)) ;
   IF i=NulKey
   THEN
      WriteString('error: no entry in the FileTable') ; WriteLn
   END ;
   WriteCard(Files[i].Depth, 6) ;
   WriteLn
END DisplayTree ;
 
 
(*
   DisplaySource - displays the source files.
*)
 
PROCEDURE DisplaySource ;
VAR
   i       : CARDINAL ;
   Def, Mod: CARDINAL ;
BEGIN
   Def := MakeKey('def') ;
   Mod := MakeKey('mod') ;
   i := 1 ;
   WHILE i<=FileNo DO
      IF (Files[i].Ext=DefName) OR (Files[i].Ext=ModName)
      THEN
         WriteCard(Files[i].Depth, 6) ; WriteString('  ') ;
         WriteKey(Files[i].Name) ; Write('.') ; WriteKey(Files[i].Ext) ;
         WriteLn
      END ;
      INC(i)
   END
END DisplaySource ;
 

(*
   DisplayDependancy - Version for GPCM make...
*)
(*
PROCEDURE DisplayDependency (n: PtrToNode) ;
VAR
   i, j: CARDINAL ;
   t   : PtrToNode ;
   base: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   IF n^.NoOfSons>0
   THEN
      StrCopy( Files[n^.Index].Name, base ) ;
      WriteString(Files[n^.Index].Name) ; Write('.') ; WriteString(Files[n^.Index].Ext) ;
      WriteString(' --> ') ;
      i := 1 ;
      WHILE i<=n^.NoOfSons DO
         t := n^.Sons[i] ;
         IF t#NIL
         THEN
            j := t^.Index ;
            WriteString('  ') ;
            WriteString( Files[j].Name ) ;
            WriteString( Files[j].Ext )
         END ;
         INC( i )
      END ;
      WriteString(' ; ') ;
      IF StrEqual( Files[n^.Index].Ext, 'lod') AND
         Exists( base, 'mod' )
      THEN
         WriteString(' m2l ') ; WriteString(Files[n^.Index].Name)
      ELSIF StrEqual( Files[n^.Index].Ext, 'lnk') AND
            Exists( base, 'mod' )
      THEN
         WriteString(' m2c ') ; WriteString(Files[n^.Index].Name)
      ELSIF StrEqual( Files[n^.Index].Ext, 'sym') AND
            Exists( base, 'def' )
      THEN
         WriteString(' m2c ') ; WriteString(Files[n^.Index].Name) ; WriteString('.def')
      END ;
      WriteString(' ; ') ;
      WriteLn ;
      i := 1 ;
      WHILE i<=n^.NoOfSons DO
         t := n^.Sons[i] ;
         IF (t#NIL) AND (t^.Father=n)
         THEN
            DisplayDependency( t )
         END ;
         INC( i )
      END
   END
END DisplayDependency ;


(*
   DisplayDependancy - Version for UNIX make or GNU make.
*)

PROCEDURE DisplayDependency (n: PtrToNode) ;
VAR
   i, j: CARDINAL ;
   t   : PtrToNode ;
   Base: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   IF n^.NoOfSons>0
   THEN
      GetKey(Files[n^.Index].Name, Base) ;
      WriteKey(Files[n^.Index].Name) ;
      Write('.') ; WriteKey(Files[n^.Index].Ext) ;
      WriteString(': ') ;
      i := 1 ;
      WHILE i<=n^.NoOfSons DO
         t := n^.Sons[i] ;
         IF t#NIL
         THEN
            j := t^.Index ;
            WriteString('  ') ;
            WriteKey(Files[j].Name) ;
            Write('.') ; WriteKey(Files[j].Ext)
         END ;
         INC(i)
      END ;
      WriteLn ;
      IF (Files[n^.Index].Ext=ExeName) AND ModExists(Files[n^.Index].Name)
      THEN
         Write(tab) ;
         WriteString('$(M2L) ') ;
         WriteKey(Files[n^.Index].Name)
      ELSIF (Files[n^.Index].Ext=ObjName) AND ModExists(Files[n^.Index].Name)
      THEN
         Write(tab) ;
         WriteString('$(M2C) $(OPTIONS) ') ;
         WriteKey(Files[n^.Index].Name) ;
         WriteString('.mod')
      ELSIF (Files[n^.Index].Ext=Sym) AND DefExists(Files[n^.Index].Name)
      THEN
         Write(tab) ;
         WriteString('$(M2C) $(OPTIONS) ') ;
         WriteKey(Files[n^.Index].Name) ; WriteString('.def')
      END ;
      WriteLn ;
      i := 1 ;
      WHILE i<=n^.NoOfSons DO
         t := n^.Sons[i] ;
         IF (t#NIL) AND (t^.Father=n)
         THEN
            DisplayDependency( t )
         END ;
         INC( i )
      END
   END
END DisplayDependency ;


(*
   UnixMake - generates the makefile for UNIX SUN M2
*)

PROCEDURE UnixMake (main: PtrToNode) ;
VAR
   a: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   WriteString('OPTIONS= ') ;
   IF GetArg(a, 3)
   THEN
      WriteString(a)
   END ;
   WriteLn ;
   WriteString('M2C= ') ;
   IF GetArg(a, 2)
   THEN
      WriteString(a)
   END ;
   WriteLn ;
   WriteString('M2L= $(M2C) $(OPTIONS) -o ') ;
   IF GetArg(a, 1)
   THEN
      (* WriteString(a) *)
      WriteString('a.out')
   END ;
   WriteString(' -e') ;
   WriteLn ;
   DisplayDependency( main )
END UnixMake ;
*)


(*
   DisplayDependants - displays the dependants of node, n.
*)

PROCEDURE DisplayDependants (n: PtrToNode) ;
VAR
   i, j: CARDINAL ;
   t   : PtrToNode ;
   Ext,
   a, b: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   GetKey(Files[n^.Index].Name, a) ;
   GetKey(Files[n^.Index].Ext , b) ;
   IF Files[n^.Index].Ext=ObjName
   THEN
      GetKey(ObjName, Ext) ;
      StrConCat(Ext, '  ', Ext) ;
      CalculateFileName(a, Ext, b) ;
      StrConCat(DestDirectory, b, b)
   ELSIF Files[n^.Index].Ext=ExeName
   THEN
      StrCopy(a, b)
   ELSE
      CalculateFileName(a, b, b)
   END ;
   FormatString(b) ;
   FormatString(': ') ;
   i := 1 ;
   WHILE i<=n^.NoOfSons DO
      t := n^.Sons[i] ;
      IF t#NIL
      THEN
         j := t^.Index ;
         GetKey(Files[j].Name, a) ;
         GetKey(Files[j].Ext , b) ;
         IF Files[j].Ext=ObjName
         THEN
            FormatString('  ') ;
            GetKey(ObjName, Ext) ;
            StrConCat(Ext, '  ', Ext) ;
      	    CalculateFileName(a, Ext, b) ;
            StrConCat(DestDirectory, b, b) ;
            FormatString(b)
      	 ELSE
            FormatString('  ') ;
      	    CalculateFileName(a, b, b) ;
            IF FindSourceFile(b, b)
            THEN
            END ;
            FormatString(b)
         END
      END ;
      INC(i)
   END ;
   FormatLn
END DisplayDependants;


(*
   DisplayBuild - displays the command to build node, n.
*)

PROCEDURE DisplayBuild (n: PtrToNode) ;
VAR
   Ext,
   a, b: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   IF NOT DependOnly
   THEN
      GetKey(Files[n^.Index].Name, a) ;
      GetKey(Files[n^.Index].Ext , b) ;
      IF (Files[n^.Index].Ext=ExeName) AND
         ModExists(Files[n^.Index].Name)
      THEN
         FormatChar(tab) ;
         FormatString('$(M2L) ') ;
         FormatString(a) ;
         FormatLn
      ELSIF (Files[n^.Index].Ext=ObjName) AND
         ModExists(Files[n^.Index].Name)
      THEN
         FormatChar(tab) ;
         FormatString('$(M2C) $(OPTIONS) ') ;
         IF NOT StrEqual(DestDirectory, '')
         THEN
            FormatString(' -o ') ;
            GetKey(ObjName, Ext) ;
            StrConCat(Ext, '  ', Ext) ;
            CalculateFileName(a, Ext, a) ;
            StrConCat(DestDirectory, a, a) ;
            FormatString(a)
         END ;
         GetKey(Files[n^.Index].Name, a) ;
         FormatString(a) ;
         FormatLn
      END
   END
END DisplayBuild ;


(*
   Version for UNIX make - gm2
*)

PROCEDURE DisplayDependency (n: PtrToNode) ;
VAR
   i: CARDINAL ;
   t: PtrToNode ;
BEGIN
   IF n^.NoOfSons>0
   THEN
      DisplayDependants(n) ;
      DisplayBuild(n) ;
      i := 1 ;
      WHILE i<=n^.NoOfSons DO
         t := n^.Sons[i] ;
         IF (t#NIL) AND (t^.Father=n)
         THEN
            DisplayDependency(t)
         END ;
         INC( i )
      END
   END
END DisplayDependency ;


(*
   DisplayObjects - displays a target which depends on each object.
*)

PROCEDURE DisplayObjects (n: PtrToNode) ;
VAR
   i, j: CARDINAL ;
   t   : PtrToNode ;
   Ext,
   a, b: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   IF n^.NoOfSons>0
   THEN
      FormatString('OBJS:') ;
      i := 1 ;
      WHILE i<=n^.NoOfSons DO
         t := n^.Sons[i] ;
         IF t#NIL
         THEN
            j := t^.Index ;
            IF Files[j].Ext=ObjName
            THEN
               GetKey(Files[j].Name, a) ;
               FormatString('  ') ;
               GetKey(ObjName, Ext) ;
               StrConCat(Ext, '  ', Ext) ;
               CalculateFileName(a, Ext, a) ;
               StrConCat(DestDirectory, a, a) ;
               FormatString(a)
            ELSE
               GetKey(Files[j].Name, a) ;
               GetKey(Files[j].Ext , b) ;
               FormatString('  ') ;
               CalculateFileName(a, b, a) ;
               IF FindSourceFile(a, a)
               THEN
               END ;
               FormatString(a)
            END
         END ;
         INC( i )
      END ;
      FormatLn
   END
END DisplayObjects ;


(*
   UnixMake - generates the makefile for UNIX gm2
*)

PROCEDURE UnixMake (main: PtrToNode) ;
VAR
   a: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   IF NOT DependOnly
   THEN
      FormatString('OPTIONS= ') ;
      INC(CurrentArg) ;
      IF GetArg(a, CurrentArg)
      THEN
         WriteString(a) ;
         INC(CurrentArg)
      ELSE
         IF NOT GM2PathFound
         THEN
            GetSearchPath(GM2Path)
         END ;
         IF NOT StrEqual(GM2Path, '')
         THEN
            FormatString('-M "') ; FormatString(GM2Path) ; FormatString('" ')
         END ;
         GetOptions(a) ;
         IF NOT StrEqual(a, '')
         THEN
            FormatChar(' ') ; FormatString(a)
         END
      END ;
      FormatLn ;
      FormatString('M2C= ') ;
      IF GetArg(a, CurrentArg)
      THEN
         FormatString(a) ;
         INC(CurrentArg)
      ELSE
         FormatString('m2f')
      END ;
      FormatLn ;
      FormatString('M2L= $(M2C) $(OPTIONS) -l ') ;
      IF GetArg(a, 1)
      THEN
         (* WriteString(a) *)
      END ;
      FormatLn
   END ;
   DisplayDependency( main ) ;
   DisplayObjects( main )
END UnixMake ;


(*
   JJL - dependency input
*)
(*
PROCEDURE DisplayDependency (n: PtrToNode) ;
VAR
   i, j: CARDINAL ;
   t   : PtrToNode ;
   base: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   IF n^.NoOfSons>0
   THEN
      i := 1 ;
      WHILE i<=n^.NoOfSons DO
         StrCopy( Files[n^.Index].Name, base ) ;
         WriteString(Files[n^.Index].Name) ; Write('.') ; WriteString(Files[n^.Index].Ext) ;
         WriteString('# #') ;
         t := n^.Sons[i] ;
         IF t#NIL
         THEN
            j := t^.Index ;
            WriteString( Files[j].Name ) ;
            Write('.') ; WriteString( Files[j].Ext )
         END ;
         INC( i ) ;
         WriteLn
      END ;
      i := 1 ;
      WHILE i<=n^.NoOfSons DO
         t := n^.Sons[i] ;
         IF t#NIL
         THEN
            DisplayDependency(t)
         END ;
         INC( i )
      END
   END
END DisplayDependency ;
*)

PROCEDURE ChrEqual (a: ARRAY OF CHAR; ch: CHAR) : BOOLEAN ;
BEGIN
   RETURN( (a[0]=ch) AND (a[1]=nul) )
END ChrEqual ;


VAR
   CurrentLineBreak: CARDINAL ;

(*
   FormatChar - writes a character and wraps around MaxLineBreak.
*)

PROCEDURE FormatChar (ch: CHAR) ;
BEGIN
   IF ch=tab
   THEN
      INC(CurrentLineBreak, 8)
   ELSE
      INC(CurrentLineBreak)
   END ;
   Write(ch)
END FormatChar ;


(*
   FormatString - writes a string which wraps around MaxLineBreak.
*)

PROCEDURE FormatString (a: ARRAY OF CHAR) ;
BEGIN
   IF StrLen(a)+CurrentLineBreak>MaxLineBreak
   THEN
      WriteString(' \') ;
      FormatLn ;
      FormatChar(tab)
   END ;
   WriteString(a) ;
   INC(CurrentLineBreak, StrLen(a))
END FormatString ;


(*
   FormatLn - writes a newline and sets CurrentLineBreak to zero.
*)

PROCEDURE FormatLn ;
BEGIN
   WriteLn ;
   CurrentLineBreak := 0
END FormatLn ;


(*
   ParseArgs - parses the arguments and then calls the main program.
*)

PROCEDURE ParseArgs ;
VAR
   Ext,
   a, b: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   CurrentArg := 1 ;
   IF Narg()>1
   THEN
      WHILE GetArg(a, CurrentArg) DO
         IF StrEqual(a, '-M')
         THEN
            INC(CurrentArg) ;
            GM2PathFound := GetArg(GM2Path, CurrentArg)
         ELSIF StrEqual(a, '-P')
         THEN
            INC(CurrentArg) ;
            M2MPathFound := GetArg(M2MPath, CurrentArg)
         ELSIF StrEqual(a, '-o')
         THEN
            INC(CurrentArg) ;
            IF NOT GetArg(DestDirectory, CurrentArg)
            THEN
               WriteString('-o expecting a destination directory') ;
               WriteLn ; exit(1)
            END
         ELSIF StrEqual(a, '-e')
         THEN
            INC(CurrentArg) ;
            IF NOT GetArg(Ext, CurrentArg)
            THEN
               WriteString('-e expecting an extension argument') ;
               WriteLn ;
               exit(1)
            END ;
            ObjName := MakeKey(Ext)
         ELSIF StrEqual(a, '-depend')
         THEN
            DependOnly := TRUE
         ELSE
            IF M2MPathFound
            THEN
               SetSearchPath(M2MPath)
            ELSIF GM2PathFound
            THEN
               SetSearchPath(GM2Path)
            END ;
            IF ModExists(MakeKey(a))
            THEN
               NewNode( main ) ;
               main^.Father := NIL ;
               NewSource( main^.Index ) ;
               WITH Files[main^.Index] DO
                  SourceNode := main ;
                  Name := MakeKey(a) ;
                  Files[main^.Index].Ext := ExeName ;
                  CalculateFileName( a, 'lod', b ) ;
                  PutSymKey(FileTable, MakeKey(b), main^.Index ) ;
                  MakeExe(Name, main, 0) ;
                  UnixMake( main )
               END
            ELSE
               WriteString('m2m: cannot find ') ; WriteString(a) ;
               WriteString('.mod') ; WriteLn
            END
         END ;
         INC(CurrentArg)
      END
   ELSE
      WriteString('Usage: m2m [-e extension][-o destination directory][-depend][-P m2m searchpath][-M compiler searchpath] modulename [compiler] [options]') ; WriteLn
   END
END ParseArgs ;


(*
   Init - initializes the data structures and calls ParseArgs
*)

PROCEDURE Init ;
BEGIN
   ExeName     := MakeKey('exe') ;
   ObjName     := MakeKey('o') ;
   SymName     := MakeKey('sym') ;
   DefName     := MakeKey('def') ;
   ModName     := MakeKey('mod') ;
   KeyFound    := MakeKey('KeyFound') ;
   KeyNotFound := MakeKey('KeyNotFound') ;

   CurrentLineBreak := 0 ;
   FileNo := 0 ;
   InitTree(FileTable) ;
   InitTree(DefExistsTree) ;
   InitTree(ModExistsTree) ;
   StrCopy('', DestDirectory) ;
   StrCopy('', M2MPath) ;
   StrCopy('', GM2Path) ;
   M2MPathFound := FALSE ;
   GM2PathFound := FALSE ;
   DependOnly := FALSE ;
   ParseArgs
END Init ;


BEGIN
   Init
END m2m.
