IMPLEMENTATION MODULE M2Lexical ;


FROM StdIO IMPORT Read ;

FROM M2Pass IMPORT IsPass1, IsPass2, IsPassCodeGeneration ;

FROM M2Inline IMPORT NewInlineBuffer, PutInlineBuffer, WriteInlineBuffer ;

FROM M2Reserved IMPORT IsReserved ;

FROM M2Options IMPORT CompilerDebugging,
                      DefaultProfiling, DefaultCoding, DefaultOptimizing,
                      SetOptionProfiling, SetOptionCoding,
                      SetOptionOptimizing ;

FROM M2Quads IMPORT BuildInline, PushT, GetLastQuadNo, QuadToLineNo ;

FROM M2Debug IMPORT WriteDebug ;

FROM NameKey IMPORT MakeKey, GetKey, WriteKey, LengthKey ;



(* Logitech Library *)
FROM FileSystem IMPORT Lookup, ReadChar, Close, Reset, File, Response ;

FROM StdIO IMPORT Write ;

FROM StrIO IMPORT ReadString, WriteString, WriteLn ;

FROM NumberIO IMPORT WriteCard, WriteOct ;


(* UNIX M-2 *)
FROM ASCII IMPORT nul, nl ;

(*
(* Logitech Library *)
FROM ASCII IMPORT nul, EOL ;
*)

FROM StrLib IMPORT StrCopy, StrEqual, StrLen ;


CONST

   (* UNIX M-2 *)
   EOL           = nl ;

   MaxSourceSize = 10000  ;
   MaxTokenSize  =   255  ;
   MaxBuffer     =   150  ;
   MaxFileName   =    80  ;
   eof           =   032C ;
   MaxStack      =    50  ;

TYPE
   TokenInfo        = RECORD
                         TokName : CARDINAL ;
                         TokType : TypeOfToken ;
                      END ;

   TokenAndPosition = RECORD
                         Token   : TokenInfo ;
                         TokIndex: CARDINAL ; (* Index into source file *)
                      END ;

VAR
   f: File ;
   FileName      : ARRAY [0..MaxFileName] OF CHAR ;   (* Current file name *)
   Opened        : BOOLEAN ;
   CurrentChar   : CHAR ;
   NextChar      : CHAR ;
   Eof           : BOOLEAN ;   (* End of source file.                      *)
   InQuotes      : BOOLEAN ;   (* If we are in quotes.                     *)
   QuoteChar     : CHAR ;      (* Quote character expected.                *)
   Stack         : ARRAY [0..MaxStack] OF TokenAndPosition ;
   StackPtr      : CARDINAL ;
   TokenLength   : CARDINAL ;
   TokenSource   : ARRAY [0..MaxSourceSize] OF TokenAndPosition ;
   IndexToSource : CARDINAL ;  (* How far through the source file in Pass2 *)
   SourceSize    : CARDINAL ;  (* SourceSize in tokens.                    *)
   FoundEofToken : BOOLEAN ;   (* End of file token found.                 *)
   EofToken      : CARDINAL ;  (* End of file token.                       *)



(*
   AddTokenToSource - adds a Token to the source buffer, ready for Pass 2.
*)

PROCEDURE AddTokenToSource (Name: CARDINAL; Type: TypeOfToken;
                            Index: CARDINAL) ;
BEGIN
   WITH TokenSource[SourceSize] DO
      Token.TokName := Name ;
      Token.TokType := Type ;
      TokIndex := Index
   END ;
   INC(SourceSize)
END AddTokenToSource ;


(*
   ReadTokenFromSource - reads a Token and Type from the source buffer.
*)

PROCEDURE ReadTokenFromSource (VAR Name: CARDINAL; VAR Type: TypeOfToken;
                               VAR Index: CARDINAL) ;
BEGIN
   IF IndexToSource<SourceSize
   THEN
      WITH TokenSource[IndexToSource] DO
         Name := Token.TokName ;
         Type := Token.TokType ;
         Index := TokIndex
      END ;
      INC(IndexToSource)
   ELSE
      WriteString('Exceeded source buffer') ; HALT
   END
END ReadTokenFromSource ;


(*
   GetLineNo - returns the current line number reached
               in the source file by the lexical analaysis.
*)

PROCEDURE GetLineNo () : CARDINAL ;
BEGIN
   RETURN( CurSymLine )
END GetLineNo ;


(*
   WriteError - displays the source line and points to the symbol in error.
                The message, a, is displayed.
*)

PROCEDURE WriteError (a: ARRAY OF CHAR) ;
VAR
   i: CARDINAL ;
BEGIN
   IF Opened
   THEN
      IF IsPassCodeGeneration()
      THEN
         (* Cannot use source line as it is no longer representative
            of the line causing the problem.
         *)
         WriteString('Code Generation') ; WriteLn ;
         WriteString('Error occurred at line') ;
         WriteCard(QuadToLineNo(GetLastQuadNo()), 6) ; WriteLn ;
      ELSIF IsPass1()
      THEN
         WriteString('Pass1 :- Error in: ') ; WriteString(FileName) ;
         WriteString(' at line') ;
         WriteCard(CurSymLine, 6) ; WriteLn ;
         WriteLn ;
         WriteBuffer ;
         WriteLn ;
         WriteLn ;
         (* ConsumeAndDisplayTokens *)
      ELSIF IsPass2()
      THEN
         WriteString('Pass2 :- Error in: ') ; WriteString(FileName) ;
         WriteString(' at line') ;
         WriteCard(CurSymLine, 6) ; WriteLn ;
         ConsumeAndDisplayTokens
      END
   ELSE
      WriteString('Source not opened yet: ') ;
   END ;
   WriteString(a) ; WriteLn ;
   HALT
END WriteError ;


(*
   ConsumeAndDisplayTokens - consumes the rest of the available tokens
                             in the current pass.
*)

PROCEDURE ConsumeAndDisplayTokens ;
BEGIN
   WriteString('up to:') ; WriteLn ;
   WHILE NOT FoundEofToken DO
      WriteKey(CurrentToken) ;
      GetToken
   END
END ConsumeAndDisplayTokens ;


PROCEDURE SetTokenLength ;
BEGIN
   TokenLength := LengthKey(CurrentToken)
END SetTokenLength ;


(*
   WriteToken - displays the token and its type. A debugging measure.
*)

PROCEDURE WriteToken (Token: CARDINAL; Type: TypeOfToken) ;
VAR
   a: ARRAY [0..MaxTokenSize] OF CHAR ;
BEGIN
   WriteString('**** ') ;
   GetKey(Token, a) ;
   WriteString(a) ; WriteString('  ') ;
   WriteString('**** ') ;
   IF Type=IdentToken
   THEN
      WriteString('Ident Token')
   ELSIF Type=ReservedToken
   THEN
      WriteString('Reserved Token')
   ELSIF Type=RealToken
   THEN
      WriteString('Real Token')
   ELSE
      WriteString('Integer Token')
   END ;
   WriteLn
END WriteToken ;


(*
   TypeToCard - Converts the Type to a CARDINAL.
                Care must be taken when using this procedure.
*)

PROCEDURE TypeToCard (Type: TypeOfToken) : CARDINAL ;
BEGIN
   RETURN (ORD(Type))
END TypeToCard ;


(*
   CardToType - Converts the CARDINAL to a TypeOfToken type.
                Care must be taken when using this procedure.
*)

PROCEDURE CardToType (c: CARDINAL) : TypeOfToken ;
BEGIN
   RETURN (VAL(TypeOfToken, c))
END CardToType ;


(*
   TokenIs - if Name is equal to the CurrentToken the next Token is read
             and true is returned, otherwise false is returned.
*)

PROCEDURE TokenIs (Name: CARDINAL) : BOOLEAN ;
BEGIN
   IF Name=CurrentToken
   THEN
      GetToken ;
      RETURN( TRUE )
   ELSE
      RETURN( FALSE )
   END
END TokenIs ;


(*
   IsToken - returns the result of the comparison between CurrentToken
             and Name.
*)

PROCEDURE IsToken (Name: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN( Name=CurrentToken )
END IsToken ;


(*
   OpenSource - Attempts to open the source file, a.
                The success of the operation is returned.
*)

PROCEDURE OpenSource (a: ARRAY OF CHAR) : BOOLEAN ;
BEGIN
   IF IsPass1()
   THEN
      OpenPass1(a)
   ELSIF IsPass2()
   THEN
      OpenPass2(a)
   ELSE
      WriteError('Open called outside Pass 1 and Pass 2')
   END ;
   WriteString('OPENED ') ; WriteString(a) ; WriteString(' Symbol = ') ;
   WriteToken(CurrentToken, TokenType) ;
   RETURN( Opened )
END OpenSource ;


(*
   OpenPass1 - opens the source file, a, for reading.
*)

PROCEDURE OpenPass1 (a: ARRAY OF CHAR) ;
BEGIN
   Lookup( f, a, FALSE ) ;
   IF f.res=done
   THEN
      Reset( f ) ;
      Opened := TRUE ;
      StrCopy(a, FileName) ;
      InitEofState ;
      InitToken
   ELSE
      Opened := FALSE ;
      Eof := TRUE
   END
END OpenPass1 ;


(*
   OpenPass2 - opens file, a, for reading during pass 2.
               In reality the source file is not opened, the
               tokens stored in the TokenSource array during Pass 1 are
               read from this array in Pass 2.
*)

PROCEDURE OpenPass2 (a: ARRAY OF CHAR) ;
BEGIN
   InitEofState ;
   Opened := TRUE ;
   StrCopy(a, FileName) ;
   IF StackPtr>0
   THEN
      WriteError('Check this... StackPtr should be 0 at beginning of a Pass')
   END ;
   GetToken
END OpenPass2 ;


PROCEDURE InitEofState ;
BEGIN
   Eof := FALSE ;
   FoundEofToken := FALSE ;
   NextChar := ' ' ;
   CurSymLine := 0
END InitEofState ;

   
PROCEDURE Init ;
VAR
   a: ARRAY [0..1] OF CHAR ;
BEGIN
   WriteString('Initialization of M2Lexical') ; WriteLn ;
   NextChar := ' ' ;  (* Thrown away - needs to be initialised. *)
   CurSymLine := 0 ;
   StackPtr := 0 ;
   InQuotes := FALSE ;
   TokenLength := 0 ;
   BufferSize := 0 ;
   BufferIndex := 0 ;
   BufferOffset := 0 ;
   SourceIndex := 0 ;
   OutBufferIndex := 0 ;
   TokenIndex := 0 ;
   IndexToSource := 0 ;
   SourceSize := 0 ;
   a[0] := eof ;
   a[1] := nul ;
   EofToken := MakeKey(a) ;
END Init ;


PROCEDURE InitToken ;
BEGIN
   PutBuffer(EOL) ; (* EOL determines end of the last line for WriteBuffer *)
   AdvanceChar ;
   AdvanceChar ;
   GetToken
END InitToken ;


(*
   CloseSource - Closes the current open file.
*)

PROCEDURE CloseSource ;
BEGIN
   WriteString('CLOSE called ') ; WriteLn ;
   IF IsPass1()
   THEN
      IF Opened=TRUE
      THEN
         WHILE NOT FoundEofToken DO
            GetToken
         END ;
         Opened := FALSE ;
         Close( f )
      END
   ELSIF IsPass2()
   THEN
      IF Opened=TRUE
      THEN
         WHILE NOT FoundEofToken DO
            GetToken
         END ;
         Opened := FALSE
      END
   ELSE
      WriteError('CloseFile called outside Pass 1 and Pass 2')
   END
END CloseSource ;


(*
   GetToken - gets the next token into CurrentToken.
*)

PROCEDURE GetToken ;
BEGIN
   IF StackPtr>0
   THEN
      DEC(StackPtr) ;
      WITH Stack[StackPtr] DO
         CurrentToken := Token.TokName ;
         TokenType := Token.TokType ;
         TokenIndex := TokIndex
      END
   ELSE
      IF IsPass1()
      THEN
         GetTokenPass1
      ELSIF IsPass2()
      THEN
         GetTokenPass2
      ELSE
         WriteError('GetToken should only be called during Pass 1 and Pass 2')
      END
   END ;
   SetTokenLength
 ; IF CompilerDebugging
   THEN
      WriteToken(CurrentToken, TokenType)
   END ;
   (* ; HALT *)
END GetToken ;


(*
   GetTokenPass1 - reads the next token providing that EofToken was not found.
                   The read token is then added to the pass 2 token buffer.
*)

PROCEDURE GetTokenPass1 ;
BEGIN
   IF NOT FoundEofToken
   THEN
      ReadToken(CurrentToken, TokenType, TokenIndex) ;
    (* ; WriteString('TokenIndex => ') ; Write(BufferSource[TokenIndex MOD MaxBuffer]) ; WriteLn ; *)
      AddTokenToSource(CurrentToken, TokenType, CurSymLine) ;
      IF CurrentToken=EofToken
      THEN
         WriteLn ;
         WriteString('******* EOF FOUND PASS 1 *******') ; WriteLn ;
         FoundEofToken := TRUE
      END
   END
END GetTokenPass1 ;


(*
   GetTokenPass2 - reads the next token providing that FoundEofToken has not
                   been found.
*)

PROCEDURE GetTokenPass2 ;
BEGIN
   IF NOT FoundEofToken
   THEN
      ReadTokenFromSource(CurrentToken, TokenType, CurSymLine) ;
      IF CurrentToken=EofToken
      THEN
         WriteLn ;
         WriteString('******* EOF FOUND PASS 2 *******') ; WriteLn ;
         FoundEofToken := TRUE
      END
   END
END GetTokenPass2 ;


(*
   PutToken - pushes a symbol, Name, back onto the input.
              GetToken will set CurrentToken to, Name.
*)

PROCEDURE PutToken (Name: CARDINAL; Type: TypeOfToken) ;
BEGIN
   IF StackPtr=MaxStack
   THEN
      WriteError('Maximum push back symbol exceeded - Increase CONST MaxStack')
   ELSE
      WITH Stack[StackPtr] DO
         Token.TokName := Name ;
         Token.TokType := Type ;
         TokIndex := TokenIndex
      END ;
      INC(StackPtr)
   END
END PutToken ;


(*
   PutTokenAndIndex - pushes a symbol, Name, back onto the input.
                      GetToken will set CurrentToken to, Name.
                      The token index is set to Index.
*)

PROCEDURE PutTokenAndIndex (Name: CARDINAL; Type: TypeOfToken; Index: CARDINAL) ;
BEGIN
   IF StackPtr=MaxStack
   THEN
      WriteError('Maximum push back symbol exceeded - Increase CONST MaxStack')
   ELSE
      WITH Stack[StackPtr] DO
         Token.TokName := Name ;
         Token.TokType := Type ;
         TokIndex := Index
      END ;
      INC(StackPtr)
   END
END PutTokenAndIndex ;


(*
   CodeQuads - sets a switch in the quadruples depending on the
               flags = + -.
*)

PROCEDURE CodeQuads ;
BEGIN
   IF CurrentChar='='
   THEN
      SetOptionCoding(DefaultCoding) ;
      AdvanceChar
   ELSIF CurrentChar='+'
   THEN
      SetOptionCoding(TRUE) ;
      AdvanceChar
   ELSIF CurrentChar='-'
   THEN
      SetOptionCoding(FALSE) ;
      AdvanceChar
   END
END CodeQuads ;


(*
   ProfileCode - sets a switch in the quadruples depending on the
                 flags = + -.
*)

PROCEDURE ProfileCode ;
BEGIN
   IF CurrentChar='='
   THEN
      SetOptionProfiling(DefaultProfiling) ;
      AdvanceChar
   ELSIF CurrentChar='+'
   THEN
      SetOptionProfiling(TRUE) ;
      AdvanceChar
   ELSIF CurrentChar='-'
   THEN
      SetOptionProfiling(FALSE) ;
      AdvanceChar
   END
END ProfileCode ;


(*
   OptimizeCode - sets a switch in the quadruples depending on the
                  flags = + -.
*)

PROCEDURE OptimizeCode ;
BEGIN
   IF CurrentChar='='
   THEN
      SetOptionOptimizing(DefaultOptimizing) ;
      AdvanceChar
   ELSIF CurrentChar='+'
   THEN
      SetOptionOptimizing(TRUE) ;
      AdvanceChar
   ELSIF CurrentChar='-'
   THEN
      SetOptionOptimizing(FALSE) ;
      AdvanceChar
   END
END OptimizeCode ;


(*
   InsertInlineCode - inserts code after the $I up until the next $I
                      into the Inline buffer.
*)

PROCEDURE InsertInlineCode ;
VAR
   Buf : CARDINAL ;
BEGIN
   Buf := NewInlineBuffer() ;
   PushT(Buf) ;
   BuildInline ;
   REPEAT
      PutInlineBuffer(CurrentChar) ;
      AdvanceChar ;
   (* IF NextLine#CurLine
      THEN
         (* Insert an effective WriteLn *)
         PutInlineBuffer(EOL)
      END
   *)
   UNTIL (CurrentChar='$') AND (NextChar='I') ;
   (* end of inline code *)
   PutInlineBuffer(nul)
END InsertInlineCode ;


(*
   ParseOption - parses compiler options.
*)

PROCEDURE ParseOption ;
BEGIN
   AdvanceChar ;
   IF CurrentChar='I'
   THEN
      AdvanceChar ;
      InsertInlineCode
   ELSIF CurrentChar='P'
   THEN
      AdvanceChar ;
      ProfileCode
   ELSIF CurrentChar='O'
   THEN
      AdvanceChar ;
      OptimizeCode
   ELSIF CurrentChar='C'
   THEN
      AdvanceChar ;
      CodeQuads
   END
END ParseOption ;


(*
   ParseComment - Skips comments and nested comments.
                  Inside comments compiler directives may be set.
                  These options are:

                  $C        Code Switch
                  $P        Profile Switch
                  $O        Optimization Switch
                  $I        Insert Assembly Language Switch
*)

PROCEDURE ParseComment ;
VAR
   Level: CARDINAL ;
BEGIN
   Level := 0 ;
   REPEAT
      IF (CurrentChar='(') AND (NextChar='*')
      THEN
         INC(Level)
      ELSIF (CurrentChar='*') AND (NextChar=')')
      THEN
         DEC(Level)
      ELSIF (CurrentChar='$') AND (Level=1)
      THEN
         ParseOption
      END ;
      AdvanceChar ;
   UNTIL (Level=0) OR Eof ;
END ParseComment ;


(*
   ConsumeSpaces - gets rid of all excess spaces before the next token.
*)

PROCEDURE ConsumeSpaces ;
VAR
   ok: BOOLEAN ;
BEGIN
   REPEAT
      IF CurrentChar=' '
      THEN
         WHILE (CurrentChar=' ') AND (NOT Eof) DO
            AdvanceChar
         END ;
         ok := FALSE
      ELSIF CurrentChar=EOL
      THEN
         AdvanceChar ;
         ok := FALSE
      ELSIF (CurrentChar='(') AND (NextChar='*')
      THEN
         ParseComment ;
         AdvanceChar ;
         ok := FALSE
      ELSE
        ok := TRUE
      END
   UNTIL ok
END ConsumeSpaces ;


(*
   BuildString - collects characters up to but not including the
                 start quote character and places these characters in, a.
*)

PROCEDURE BuildString (VAR a: ARRAY OF CHAR; VAR i: CARDINAL) ;
VAR
   High: CARDINAL ;
BEGIN
   High := HIGH(a) ;
   (* Fill in string or character *)
   REPEAT
      a[i] := CurrentChar ;
      INC(i) ;
      AdvanceChar
   UNTIL (i=High) OR (CurrentChar=QuoteChar) OR (CurrentChar=EOL) OR Eof ;
   IF CurrentChar=EOL
   THEN
      WriteError('Missing quote expected: strings cannot span more than one line')
   END
END BuildString ;


(*
   HandleQuotes - builds a string, a, from the quote character or string
                  depending whether the string has been consumed yet.
                  Firstly generate string, secondly generate quote string.
                  This procedure is called if InQuotes is true.
                  If InQuotes is true and the CurrentChar=QuoteChar
                  then a string was the last token to be built
                  and therefore the next token will be a quote.
                  If, however, CurrentChar#QuoteChar then we need to build
                  a string since InQuotes is TRUE.
*)

PROCEDURE HandleQuotes (VAR a: ARRAY OF CHAR; VAR i: CARDINAL) ;
BEGIN
   (* Have we consumed the string yet? *)
   IF CurrentChar=QuoteChar
   THEN
      (* Yes, therefore build a token from the final quote *)
      InQuotes := FALSE ;
      a[i] := QuoteChar ;
      INC(i) ;
      AdvanceChar
   ELSE
      (*
         ok build string and then the next time this procedure is called
         build a quote.
      *)
      BuildString(a, i)
      (* If all is well the CurrentChar=QuoteChar *)
   END
END HandleQuotes ;


(*
   ReadToken - checks the CurrentChar, NextChar and generates the next token.
*)

PROCEDURE ReadToken (VAR Name: CARDINAL;
                     VAR Type: TypeOfToken;
                     VAR Index: CARDINAL) ;
VAR
   i    : CARDINAL ;
   ok   : BOOLEAN ;
   a    : ARRAY [0..MaxTokenSize] OF CHAR ;
BEGIN
   i := 0 ;
   IF NOT Eof
   THEN
      IF InQuotes
      THEN
         Index := OutBufferIndex-2 ;
         HandleQuotes(a, i)
      ELSE
         ConsumeSpaces ;
         Index := OutBufferIndex-2 ;
         i := 0 ;
         IF (CurrentChar='"') OR (CurrentChar="'")
         THEN
            InQuotes := TRUE ;
            QuoteChar := CurrentChar ;
            a[i] := CurrentChar ;
            AdvanceChar ;
            INC(i)
         ELSIF DoubleDelimiter()
         THEN
            a[i] := CurrentChar ;
            AdvanceChar ;
            INC(i) ;
            a[i] := CurrentChar ;
            AdvanceChar ;
            INC(i)
         ELSIF Delimiter()
         THEN
            a[i] := CurrentChar ;
            AdvanceChar ;
            INC(i)
         ELSIF Digit(CurrentChar)
         THEN
            ParseNumber(a, i, Type)
         ELSE
            ParseIdent(a, i) ;
            Type := IdentToken
         END
      END
   ELSE
      (* eof *)
      i := 0 ;
      a[i] := eof ;
      INC(i)
   END ;
   IF i<=HIGH(a)
   THEN
      a[i] := nul
   END ;
   Name := MakeKey(a) ;
   IF IsReserved(Name)
   THEN
      Type := ReservedToken
   END
END ReadToken ;


(*
   ParseIdent - reads an identifier into string, a, beginning at index, i.
*)

PROCEDURE ParseIdent (VAR a: ARRAY OF CHAR; VAR i: CARDINAL) ;
BEGIN
   REPEAT
      a[i] := CurrentChar ;
      AdvanceChar ;
      INC(i)
   UNTIL Delimiter() OR (i>HIGH(a)) OR (CurrentChar=' ') OR
         (CurrentChar=EOL) OR Eof
END ParseIdent ;


(*
   ParseNumber - returns with a legal number in string, a, index, i, is
                 set together with the Type of the number.
*)

PROCEDURE ParseNumber (VAR a: ARRAY OF CHAR; VAR i: CARDINAL;
                       VAR Type: TypeOfToken) ;
BEGIN
   REPEAT
      a[i] := CurrentChar ;
      AdvanceChar ;
      INC(i)
   UNTIL NOT Digit(CurrentChar) ;
   IF (CurrentChar='.') AND (NextChar#'.')
   THEN
      a[i] := CurrentChar ;
      AdvanceChar ;
      INC(i) ;
      WHILE Digit(CurrentChar) DO
         a[i] := CurrentChar ;
         AdvanceChar ;
         INC(i)
      END ;
      IF CurrentChar='E'
      THEN
         a[i] := CurrentChar ;
         AdvanceChar ;
         INC(i) ;
         IF (CurrentChar='+') OR (CurrentChar='-')
         THEN
            a[i] := CurrentChar ;
            AdvanceChar ;
            INC(i)
         END ;
         WHILE Digit(CurrentChar) DO
            a[i] := CurrentChar ;
            AdvanceChar ;
            INC(i)
         END
      END ;
      Type := RealToken
   ELSE
      IF (CurrentChar='H') OR
         (HexDigit(CurrentChar) AND (HexDigit(NextChar) OR (NextChar='H')))
         (*
            The problem is that B and C are both legal hex digits!
            B and C also denote binary and octal character constants.
         *)
      THEN
         WHILE HexDigit(CurrentChar) DO
            a[i] := CurrentChar ;
            AdvanceChar ;
            INC(i)
         END ;
         IF CurrentChar='H'
         THEN
            a[i] := CurrentChar ;
            AdvanceChar ;
            INC(i)
         ELSE
            WriteError('Hex number missing H')
         END
      ELSIF CurrentChar='C'
      THEN
         a[i] := CurrentChar ;
         AdvanceChar ;
         INC(i)
      ELSIF CurrentChar='B'
      THEN
         a[i] := CurrentChar ;
         AdvanceChar ;
         INC(i)
      END ;
      Type := IntegerToken
   END
END ParseNumber ;


(* Delimiter returns true if and only if CurrentChar is a delimiter *)

PROCEDURE Delimiter() : BOOLEAN ;
BEGIN
   IF (CurrentChar='-') OR
      (CurrentChar='+') OR (CurrentChar='*') OR (CurrentChar='\') OR
      (CurrentChar='|') OR (CurrentChar='(') OR (CurrentChar=')') OR
      (CurrentChar='"') OR (CurrentChar="'") OR (CurrentChar='{')
   THEN
      RETURN( TRUE )
   ELSIF
      (CurrentChar='}') OR (CurrentChar='[') OR (CurrentChar=']') OR
      (CurrentChar='#') OR (CurrentChar='=') OR (CurrentChar='<')
   THEN
      RETURN( TRUE )
   ELSIF
      (CurrentChar='>') OR (CurrentChar='.') OR (CurrentChar=';') OR
      (CurrentChar=':') OR (CurrentChar='^') OR (CurrentChar=',')
   THEN
      RETURN( TRUE )
   ELSE
      RETURN( FALSE )
   END
END Delimiter ;


PROCEDURE DoubleDelimiter () : BOOLEAN ;
BEGIN
   RETURN (
           ((CurrentChar='>') AND (NextChar='=')) OR
           ((CurrentChar='<') AND (NextChar='=')) OR
           ((CurrentChar='<') AND (NextChar='>')) OR
           ((CurrentChar=':') AND (NextChar='=')) OR
           ((CurrentChar='.') AND (NextChar='.'))
          )
END DoubleDelimiter ;


PROCEDURE Digit (ch: CHAR) : BOOLEAN ;
BEGIN
   RETURN( (ch>='0') AND (ch<='9') )
END Digit ;


PROCEDURE HexDigit (ch: CHAR) : BOOLEAN ;
BEGIN
   RETURN( Digit(ch) OR ((ch>='A') AND (ch<='F')) )
END HexDigit ;


(*
   AdvanceChar - advances the CurrentChar to the NextChar and reads a
                 character from the source file into NextChar.
                 AdvanceChar actually buffers the characters, to allow
                 for a reasonable snapshot of text if an error should
                 arise.
                 It reads a line at a time into the Buffer, thus allowing
                 the WriteError routine to display the complete line
                 with the erroroneous symbol underlined.
*)

PROCEDURE AdvanceChar ;
BEGIN
   IF NOT Eof
   THEN
      CurrentChar := NextChar ;
      IF NOT GetBuffer(NextChar)
      THEN
         REPEAT
            ReadChar(f, NextChar) ;
            (* Write(NextChar) ; WriteOct(ORD(NextChar), 4) ; WriteLn ; *)
            IF f.res#done
            THEN
               NextChar := eof
            END ;
            PutBuffer(NextChar)
         UNTIL (NextChar=EOL) OR (NextChar=eof) ;
         IF NOT GetBuffer(NextChar)
         THEN
            WriteError('AdvanceChar has failed')
         END
      END ;
      IF CurrentChar=eof
      THEN
         Eof := TRUE
      END
   END
END AdvanceChar ;


(*
   The following routines manipulate the BufferSource which contains
   a snapshot of the source file being passed.
*)

VAR
   BufferSource  : ARRAY [0..MaxBuffer-1] OF CHAR ; (* Source buffer       *)
   BufferIndex   : CARDINAL ;  (* End of the buffer.                       *)
   BufferOffset  : CARDINAL ;  (* Where the buffer starts within the file. *)
   BufferSize    : CARDINAL ;  (* Amount of the buffer currently used.     *)
   OutBufferIndex: CARDINAL ;  (* Specifies the character where the        *)
                               (* GetBuffer routine has read.              *)
   SourceIndex   : CARDINAL ;  (* Index into the source file, nth char     *)
   CurSymLine    : CARDINAL ;  (* Line number of the CurrentSymbol         *)



(*
   PutBuffer - places the character, ch, into the circular buffer.
               It also updates the CurSymLine if a <CR> is found.
*)

PROCEDURE PutBuffer (ch: CHAR) ;
BEGIN
   (* IF ch=EOL THEN WriteLn ELSE Write(ch) END ; *)
   BufferSource[BufferIndex] := ch ;
   BufferIndex := (BufferIndex+1) MOD MaxBuffer ;
   IF ch=EOL
   THEN
      INC(CurSymLine)
   END ;
   INC(SourceIndex) ;
   IF BufferSize=MaxBuffer
   THEN
      INC(BufferOffset)
   ELSE
      INC(BufferSize)
   END
END PutBuffer ;


(*
   GetBuffer - returns true if a character is found in the BufferSource,
               otherwise false is returned.
*)

PROCEDURE GetBuffer (VAR ch: CHAR) : BOOLEAN ;
BEGIN
   IF OutBufferIndex<SourceIndex
   THEN
      ch := BufferSource[OutBufferIndex MOD MaxBuffer] ;
      INC(OutBufferIndex) ;
      RETURN( TRUE )
   ELSE
      RETURN( FALSE )
   END
END GetBuffer ;


(*
   WriteBuffer - writes the buffer with the CurrentSymbol underlined.
*)

PROCEDURE WriteBuffer ;
VAR
   OffsetIntoSource,
   OffsetIntoBuffer,
   Indent          : CARDINAL ;
BEGIN
   OffsetIntoBuffer := StartOfIndexInBuffer() ;
   OffsetIntoSource := BufferOffset ;
   FindNextLineInBuffer(OffsetIntoBuffer, OffsetIntoSource) ;

   (* Found end of first line, now display BufferSource *)

   Indent := DisplayUpToToken(OffsetIntoBuffer, OffsetIntoSource) ;
   WriteNChars(' ', Indent) ;
   WriteNChars('^', TokenLength) ;
   WriteLn ;
   DisplayRestOfBuffer(OffsetIntoBuffer, OffsetIntoSource)
END WriteBuffer ;


(*
   FindNextLineInBuffer - sets the index, OffsetIntoBuffer,
                          into BufferSource which
                          indicates the start of the next available
                          textual line inside BufferSource.
                          OffsetIntoSource is set which indicates the
                          total number of characters into the source file
                          OffsetIntoBuffer represents.
*)

PROCEDURE FindNextLineInBuffer (VAR OffsetIntoBuffer: CARDINAL;
                                VAR OffsetIntoSource: CARDINAL) ;
BEGIN
   WHILE (OffsetIntoSource<=SourceIndex) AND
         (BufferSource[OffsetIntoBuffer]#EOL)
   DO
      OffsetIntoBuffer := (OffsetIntoBuffer+1) MOD MaxBuffer ;
      INC(OffsetIntoSource)
   END
END FindNextLineInBuffer ;


(*
   DisplayUpToToken - displays the BufferSource up to and including the line
                      of CurrentToken.
                      The Token indent offset is returned.
*)

PROCEDURE DisplayUpToToken (VAR OffsetIntoBuffer: CARDINAL ;
                            VAR OffsetIntoSource: CARDINAL) : CARDINAL ;
VAR
   Indent,
   EndOfToken: CARDINAL ;
BEGIN
   EndOfToken := TokenIndex+TokenLength-1 ;
   IF SourceIndex<EndOfToken THEN HALT END ;
   Indent := 0 ;
   REPEAT
      WHILE (BufferSource[OffsetIntoBuffer]#EOL) AND
            (OffsetIntoSource<SourceIndex)
      DO
         Write(BufferSource[OffsetIntoBuffer]) ;
         IF OffsetIntoSource<TokenIndex
         THEN
            INC(Indent)
         END ;
         OffsetIntoBuffer := (OffsetIntoBuffer+1) MOD MaxBuffer ;
         INC(OffsetIntoSource)
      END ;
      IF BufferSource[OffsetIntoBuffer]=EOL
      THEN
         WriteLn ;
         IF OffsetIntoSource<TokenIndex
         THEN
            Indent := 0
         END ;
         OffsetIntoBuffer := (OffsetIntoBuffer+1) MOD MaxBuffer ;
         INC(OffsetIntoSource)
      END
   UNTIL OffsetIntoSource>=EndOfToken ;
   RETURN( Indent )
END DisplayUpToToken ;


(*
   DisplayRestOfBuffer - display the rest of the buffer, from, OffsetIntoBuffer.*)

PROCEDURE DisplayRestOfBuffer (OffsetIntoBuffer: CARDINAL;
                               OffsetIntoSource: CARDINAL) ;
BEGIN
   WHILE OffsetIntoSource<SourceIndex DO
      IF BufferSource[OffsetIntoBuffer]=EOL
      THEN
         WriteLn
      ELSE
         Write(BufferSource[OffsetIntoBuffer])
      END ;
      OffsetIntoBuffer := (OffsetIntoBuffer+1) MOD MaxBuffer ;
      INC(OffsetIntoSource)
   END
END DisplayRestOfBuffer ;


(*
   DisplayBuffer - display the complete buffer of source text.
*)

PROCEDURE DisplayBuffer ;
VAR
   OffsetIntoBuffer,
   OffsetIntoSource: CARDINAL ;
BEGIN
   OffsetIntoBuffer := StartOfIndexInBuffer() ;
   OffsetIntoSource := BufferOffset ;
   DisplayRestOfBuffer(OffsetIntoBuffer, OffsetIntoSource)
END DisplayBuffer ;


(*
   StartOfIndexInBuffer - returns the start of the textual data in SourceBuffer.
*)

PROCEDURE StartOfIndexInBuffer () : CARDINAL ;
BEGIN
   RETURN( (BufferIndex+MaxBuffer-BufferSize) MOD MaxBuffer)
END StartOfIndexInBuffer ;


PROCEDURE DisplayBufferVars ;
VAR
   ch: CHAR ;
BEGIN
   WriteString('TokenIndex  :') ; WriteCard(TokenIndex, 4) ; WriteLn ;
   WriteString('SourceIndex :') ; WriteCard(SourceIndex, 4) ; WriteLn ;
   WriteString('BufferSize  :') ; WriteCard(BufferSize, 4) ; WriteLn ;
   WriteString('CurrentToken:"') ; WriteKey(CurrentToken) ; WriteString('"') ; WriteLn ;
   WriteString('TokenLength :') ; WriteCard(TokenLength, 4) ; WriteLn ;
   IF TokenLength#LengthKey(CurrentToken) THEN WriteString('ERROR') ; WriteLn ; HALT END ;
(*
   Read(ch) ;
   IF ch='d' THEN HALT END ;
*)
END DisplayBufferVars ;


(*
   WriteNChars - writes, n, characters, ch, to the default output.
*)

PROCEDURE WriteNChars (ch: CHAR; n: CARDINAL) ;
BEGIN
   WHILE n>0 DO
      Write(ch) ;
      DEC(n)
   END
END WriteNChars ;


(*
   IsInRange - returns true if i is in the range of r1..r2
*)

PROCEDURE IsInRange (i, r1, r2: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN( (i>=r1) AND (i<=r2) )
END IsInRange ;


BEGIN
   Init
END M2Lexical.
