MODULE m2where ;

(*
    Title      : m2where
    Author     : Gaius Mulley
    System     : UNIX (gm2)
    Date       : Tue Feb 25 20:31:41 1992
    Last edit  : Tue Feb 25 20:31:41 1992
    Description: gives the absolute path to the module given as the
                 first argument.
*)

FROM Args IMPORT GetArg ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM M2Search IMPORT FindSourceFile ;
FROM M2FileName IMPORT CalculateFileName ;
FROM StrLib IMPORT IsSubString, StrLen, StrCopy ;
FROM ASCII IMPORT nul ;


CONST
   MaxPath = 1000 ;


(*
   Init - looks for argument one, and then attempts to
          find the path for this module.
*)

PROCEDURE Init ;
VAR
   FileName,
   Stem,
   Extension,
   ModuleName: ARRAY [0..MaxPath] OF CHAR ;
BEGIN
   IF GetArg(ModuleName, 1)
   THEN
      (* found module name from argument one *)
      IF HasExtension(ModuleName)
      THEN
      	 (*
      	    Build ModuleName up from its stem and extension,
      	    so to get around the name length restriction of the
      	    filesystem.
      	 *)
      	 ExtractStemAndExtension(ModuleName, Stem, Extension) ;
      	 CalculateFileName(Stem, Extension, ModuleName)
      ELSE
      	 (* No extension given - therefore default to .mod *)
      	 CalculateFileName(ModuleName, 'mod', ModuleName)
      END ;
      IF FindSourceFile(ModuleName, FileName)
      THEN
      	 WriteString(FileName) ; WriteLn
      END
   END
END Init ;


(*
   HasExtension - returns true if module, a, has a definition extension
                  or an implementation extension.
*)

PROCEDURE HasExtension (a: ARRAY OF CHAR) : BOOLEAN ;
BEGIN
   RETURN( IsSubString(a, '.def') OR IsSubString(a, '.mod') )
END HasExtension ;


(*
   ExtractStemAndExtension - fills in Stem and Extension with the
                             appropriate components of FullName.
                             FullName is expected to contain:
                             { Letter } " . " { Letter }
                             Stem       " . " Extension
*)

PROCEDURE ExtractStemAndExtension (FullName: ARRAY OF CHAR;
                                   VAR Stem, Extension: ARRAY OF CHAR) ;
VAR
   s, e,
   n, i, j: CARDINAL ;
BEGIN
   n := StrLen(FullName) ;
   e := HIGH(Extension) ;
   s := HIGH(Stem) ;
   i := 0 ;
   WHILE (i<n) AND (FullName[i]#'.') AND (i<=s) DO
      Stem[i] := FullName[i] ;
      INC(i)
   END ;
   IF (i<=s)
   THEN
      Stem[i] := nul
   END ;
   IF (i<n) AND (FullName[i]='.')
   THEN
      INC(i) ;  (* We dont want the '.' *)
      j := 0 ;
      WHILE (i+j<n) AND (FullName[i+j]#nul) AND (j<=e) DO
      	 Extension[j] := FullName[i+j] ;
      	 INC(j)
      END ;
      IF j<=e
      THEN
      	 Extension[j] := nul
      END
   ELSE
      StrCopy('', Extension)
   END ;
END ExtractStemAndExtension ;

BEGIN
   Init
END m2where.
