MODULE varient5 ;


FROM Storage IMPORT ALLOCATE, DEALLOCATE ;

FROM StrLib IMPORT StrConCat ;

FROM StrIO IMPORT WriteString, WriteLn ;

FROM M2Debug IMPORT Assert ;

FROM NumberIO IMPORT CardToStr,
                     WriteCard ;  (* Delete this line when all works *)

FROM M2Configure IMPORT ReturnArgumentOnStack ;

FROM M2ALU IMPORT InitValue, PtrToValue, PushCard, PopCard,
                  Pop, PushString, Push,
                  IsSolved ;

FROM M2Lexical IMPORT WriteError, PedanticError ;

FROM Lists IMPORT List, InitList, GetItemFromList, PutItemIntoList,
                  IsItemInList, IncludeItemIntoList, NoOfItemsInList,
                  ForeachItemInListDo ;

FROM NameKey IMPORT MakeKey, NulName, WriteKey, LengthKey, GetKey ;

FROM SymbolKey IMPORT NulKey, SymbolTree,
                      PerformOperation, InitTree, Tabulate,
                      GetSymKey, PutSymKey, DelSymKey, IsEmptyTree,
                      DoesTreeContainAny, ForeachNodeDo ;

FROM M2Base IMPORT InitBase, Char, Integer ;

FROM M2System IMPORT Address ;

FROM M2Comp IMPORT CompilingDefinitionModule,
                   CompilingImplementationModule ;


CONST
   MaxScopes  =  20 ; (* Maximum number of scopes at any one time.           *)
   MaxSymbols = 100 ;

TYPE
   TypeOfAddr = (Relative, Absolute) ;
   ModeOfAddr = (Value, Addr, None) ;

   (* TypeOfSymbol denotes the type of symbol.                               *)

   TypeOfSymbol
              = (RecordSym, VarientSym,
                 VarSym, EnumerationSym, SubrangeSym, ArraySym,
                 ConstStringSym, ConstVarSym, ConstLitSym,
                 VarParamSym, ParamSym, PointerSym,
                 UndefinedSym, TypeSym,
                 RecordFieldSym, VarientFieldSym, EnumerationFieldSym,
                 DefImpSym, ModuleSym, SetSym, ProcedureSym, ProcTypeSym,
                 DummySym, SubscriptSym, UnboundedSym) ;

   SymUndefined = RECORD
                     Name      : CARDINAL ;   (* Index into name array, name *)
                                              (* of record.                  *)
                  END ;

   SymVarient = RECORD
                   Size        : PtrToValue ; (* Size at runtime of symbol.  *)
                   ListOfSons  : List ;       (* ListOfSons contains a list  *)
                                              (* of SymRecordField and       *)
                                              (* SymVarients                 *)
                                              (* declared by the source      *)
                                              (* file.                       *)
                   Parent      : CARDINAL ;   (* Points to the Father Symbol *)
               END ;

   SymRecord = RECORD
                  Name         : CARDINAL ;   (* Index into name array, name *)
                                              (* of record.                  *)
                  LocalSymbols : SymbolTree ; (* Contains all record fields. *)
                  Size         : PtrToValue ; (* Size at runtime of symbol.  *)
                  ListOfSons   : List ;       (* ListOfSons contains a list  *)
                                              (* of SymRecordField and       *)
                                              (* SymVarients                 *)
                                              (* declared by the source      *)
                                              (* file.                       *)
                  Parent       : CARDINAL ;   (* Points to the Father Symbol *)
               END ;

   SymSubrange = RECORD
                    Name       : CARDINAL ;   (* Index into name array, name *)
                                              (* of subrange.                *)
                    Low        : CARDINAL ;   (* Index to symbol for lower   *)
                    High       : CARDINAL ;   (* Index to symbol for higher  *)
                    Size       : PtrToValue ; (* Size of subrange type.      *)
                    Type       : CARDINAL ;   (* Index to type symbol for    *)
                                              (* the type of subrange.       *)
                 END ;

   SymEnumeration =
                RECORD
                   Name        : CARDINAL ;   (* Index into name array, name *)
                                              (* of enumeration.             *)
                   NoOfElements: CARDINAL ;   (* No elements in enumeration  *)
                   LocalSymbols: SymbolTree ; (* Contains all enumeration    *)
                                              (* fields.                     *)
                   Size        : PtrToValue ; (* Size at runtime of symbol.  *)
                   Parent      : CARDINAL ;   (* Parent is used as an index  *)
                                              (* to the parent block. The    *)
                                              (* enumeration type is a       *)
                                              (* pseudo block which has its  *)
                                              (* fields as LocalSymbols.     *)
                END ;

   SymArray = RECORD
                 Name        : CARDINAL ;     (* Index into name array, name *)
                                              (* of array.                   *)
                 ListOfSubs  : List ;         (* Contains a list of all      *)
                                              (* subscripts for this array.  *)
                 Size        : PtrToValue ;   (* Size at runtime of symbol.  *)
                 Offset      : PtrToValue ;   (* Offset at runtime of symbol *)
                 Type        : CARDINAL ;     (* Type of the Array.          *)
              END ;

   SymSubscript = RECORD
                     Type       : CARDINAL ;   (* Index to a subrange symbol. *)
                     Size       : PtrToValue ; (* Size of this indice in*Size *)
                     Offset     : PtrToValue ; (* Offset at runtime of symbol *)
                                               (* Pseudo ie: Offset+Size*i    *)
                                               (* 1..n. The array offset is   *)
                                               (* the real memory offset.     *)
                                               (* This offset allows the a[i] *)
                                               (* to be calculated without    *)
                                               (* the need to perform         *)
                                               (* subtractions when a[4..10]  *)
                                               (* needs to be indexed.        *)
                  END ;

   SymUnbounded = RECORD
                     Type       : CARDINAL ;   (* Index to Simple type symbol *)
                     Size       : PtrToValue ; (* Max No of words ever        *)
                                               (* passed to this type.        *)
                  END ;

   SymProcedure
            = RECORD
                 Name        : CARDINAL ;   (* Index into name array, name   *)
                                            (* of procedure.                 *)
                 ListOfParam : List ;       (* Contains a list of all the    *)
                                            (* parameters in this procedure. *)
                 ParamDefined: BOOLEAN ;    (* Have the parameters been      *)
                                            (* defined yet?                  *)
                 DefinedInDef: BOOLEAN ;    (* Were the parameters defined   *)
                                            (* in the Definition module?     *)
                                            (* Note that this depends on     *)
                                            (* whether the compiler has read *)
                                            (* the .def or .mod first.       *)
                                            (* The second occurence is       *)
                                            (* compared to the first.        *)
                 DefinedInImp: BOOLEAN ;    (* Were the parameters defined   *)
                                            (* in the Implementation module? *)
                                            (* Note that this depends on     *)
                                            (* whether the compiler has read *)
                                            (* the .def or .mod first.       *)
                                            (* The second occurence is       *)
                                            (* compared to the first.        *)
                 Father      : CARDINAL ;   (* Father scope of procedure.    *)
                 StartQuad   : CARDINAL ;   (* Index into quads for start    *)
                                            (* of procedure.                 *)
                 EndQuad     : CARDINAL ;   (* Index into quads for end of   *)
                                            (* procedure.                    *)
                 Reachable   : BOOLEAN ;    (* Defines if procedure will     *)
                                            (* ever be called by the main    *)
                                            (* Module.                       *)
                 ReturnType  : CARDINAL ;   (* Return type for function.     *)
                 Offset      : CARDINAL ;   (* Location of procedure used    *)
                                            (* in Pass 2 and if procedure    *)
                                            (* is a syscall.                 *)
                 LocalSymbols: SymbolTree ; (* Contains all symbols declared *)
                                            (* within this procedure.        *)
                 EnumerationScopeList: List ;
                                            (* Enumeration scope list which  *)
                                            (* contains a list of all        *)
                                            (* enumerations which are        *)
                                            (* visable within this scope.    *)
                 ListOfVars  : List ;       (* List of variables in this     *)
                                            (* scope.                        *)
                 Size        : PtrToValue ; (* Activation record size.       *)
              END ;

   SymProcType
            = RECORD
                 Name        : CARDINAL ;   (* Index into name array, name   *)
                                            (* of procedure.                 *)
                 ListOfParam : List ;       (* Contains a list of all the    *)
                                            (* parameters in this procedure. *)
                 ReturnType  : CARDINAL ;   (* Return type for function.     *)
                 Size     : PtrToValue ;    (* Runtime size of symbol.       *)
              END ;

   SymParam = RECORD
                 Name        : CARDINAL ;     (* Index into name array, name *)
                                              (* of param.                   *)
                 Type        : CARDINAL ;     (* Index to the type of param. *)
              END ;

   SymVarParam = RECORD
                    Name     : CARDINAL ;     (* Index into name array, name *)
                                              (* of param.                   *)
                    Type     : CARDINAL ;     (* Index to the type of param. *)
                 END ;

   SymConstString
               = RECORD
                    Name     : CARDINAL ;     (* Index into name array, name *)
                                              (* of const.                   *)
                    String   : CARDINAL ;     (* Value of string.            *)
                    Length   : CARDINAL ;     (* StrLen(String)              *)
                 END ;

   SymConstLit = RECORD
                    Name     : CARDINAL ;     (* Index into name array, name *)
                                              (* of const.                   *)
                    Value    : PtrToValue ;   (* Value of the constant.      *)
                    Type     : CARDINAL ;     (* TYPE of constant, char etc  *)
                 END ;

   SymConstVar = RECORD
                    Name     : CARDINAL ;     (* Index into name array, name *)
                                              (* of const.                   *)
                    Value    : PtrToValue ;   (* Value of the constant       *)
                    Type     : CARDINAL ;     (* TYPE of constant, char etc  *)
                 END ;

   SymVar = RECORD
               Name     : CARDINAL ;          (* Index into name array, name *)
                                              (* of const.                   *)
               Type     : CARDINAL ;          (* Index to a type symbol.     *)
               Size     : PtrToValue ;        (* Runtime size of symbol.     *)
               Offset   : PtrToValue ;        (* Offset at runtime of symbol *)
               AddrMode : ModeOfAddr ;        (* Type of Addressing mode.    *)
               Father   : CARDINAL ;          (* Father scope of variable.   *)
               IsTemp   : BOOLEAN ;           (* Is variable a temporary?    *)
            END ;

   SymType = RECORD
                Name     : CARDINAL ;         (* Index into name array, name *)
                                              (* of type.                    *)
                Type     : CARDINAL ;         (* Index to a type symbol.     *)
                Size     : PtrToValue ;       (* Runtime size of symbol.     *)
             END ;

   SymPointer
           = RECORD
                Name     : CARDINAL ;         (* Index into name array, name *)
                                              (* of pointer.                 *)
                Type     : CARDINAL ;         (* Index to a type symbol.     *)
                Size     : PtrToValue ;       (* Runtime size of symbol.     *)
             END ;

   SymRecordField =
             RECORD
                Name     : CARDINAL ;         (* Index into name array, name *)
                                              (* of record field.            *)
                Type     : CARDINAL ;         (* Index to a type symbol.     *)
                Size     : PtrToValue ;       (* Runtime size of symbol.     *)
                Offset   : PtrToValue ;       (* Offset at runtime of symbol *)
                Parent   : CARDINAL ;         (* Index into symbol table to  *)
                                              (* determine the parent symbol *)
                                              (* for this record field. Used *)
                                              (* for BackPatching.           *)
             END ;

   SymVarientField =
             RECORD
                Size     : PtrToValue ;       (* Runtime size of symbol.     *)
                Offset   : PtrToValue ;       (* Offset at runtime of symbol *)
                Parent   : CARDINAL ;         (* Index into symbol table to  *)
                                              (* determine the parent symbol *)
                                              (* for this record field. Used *)
                                              (* for BackPatching.           *)
                ListOfSons: List ;            (* Contains a list of the      *)
                                              (* RecordField symbols.        *)
             END ;

   SymEnumerationField =
             RECORD
                Name     : CARDINAL ;         (* Index into name array, name *)
                                              (* of enumeration field.       *)
                Value    : PtrToValue ;       (* Enumeration field value.    *)
                Type     : CARDINAL ;         (* Index to the father.        *)
             END ;

   SymSet  = RECORD
                Name     : CARDINAL ;         (* Index into name array, name *)
                                              (* of set.                     *)
                Type     : CARDINAL ;         (* Index to a type symbol.     *)
                Size     : PtrToValue ;       (* Runtime size of symbol.     *)
             END ;

   SymDefImp =
            RECORD
               Name          : CARDINAL ;   (* Index into name array, name   *)
                                            (* of record field.              *)
               Type          : CARDINAL ;   (* Index to a type symbol.       *)
               Size          : PtrToValue ; (* Runtime size of symbol.       *)
               Offset        : PtrToValue ; (* Offset at runtime of symbol   *)
               ExportQualifiedTree: SymbolTree ;
                                            (* Holds all the export          *)
                                            (* Qualified identifiers.        *)
                                            (* This tree may be              *)
                                            (* deleted at the end of Pass 1. *)
               ExportUnQualifiedTree: SymbolTree ;
                                            (* Holds all the export          *)
                                            (* UnQualified identifiers.      *)
                                            (* This tree may be              *)
                                            (* deleted at the end of Pass 1. *)
               ExportRequest : SymbolTree ; (* Contains all identifiers that *)
                                            (* have been requested by other  *)
                                            (* modules before this module    *)
                                            (* declared its export list.     *)
               IncludeList   : List ;       (* Contains all included symbols *)
                                            (* which are included by         *)
                                            (* IMPORT modulename ;           *)
                                            (* modulename.Symbol             *)
               ImportTree    : SymbolTree ; (* Contains all IMPORTed         *)
                                            (* identifiers.                  *)
               ExportUndeclared: SymbolTree ;
                                            (* ExportUndeclared contains all *)
                                            (* the identifiers which were    *)
                                            (* exported but have not yet     *)
                                            (* been declared.                *)
               NeedToBeImplemented: SymbolTree ;
                                            (* NeedToBeImplemented contains  *)
                                            (* the identifiers which have    *)
                                            (* been exported and declared    *)
                                            (* but have not yet been         *)
                                            (* implemented.                  *)
               LocalSymbols  : SymbolTree ; (* The LocalSymbols hold all the *)
                                            (* variables declared local to   *)
                                            (* the block. It contains the    *)
                                            (* IMPORT r ;                    *)
                                            (* FROM _ IMPORT x, y, x ;       *)
                                            (*    and also                   *)
                                            (* MODULE WeAreHere ;            *)
                                            (*    x y z visiable by localsym *)
                                            (*    MODULE Inner ;             *)
                                            (*       EXPORT x, y, z ;        *)
                                            (*    END Inner ;                *)
                                            (* END WeAreHere.                *)
               EnumerationScopeList: List ; (* Enumeration scope list which  *)
                                            (* contains a list of all        *)
                                            (* enumerations which are        *)
                                            (* visable within this scope.    *)
               Priority      : CARDINAL ;   (* Priority of the module. This  *)
                                            (* is an index to a constant.    *)
               Unresolved    : SymbolTree ; (* All symbols currently         *)
                                            (* unresolved in this module.    *)
               StartQuad     : CARDINAL ;   (* Signify the initialization    *)
                                            (* code.                         *)
               EndQuad       : CARDINAL ;   (* EndQuad should point to a     *)
                                            (* goto quad.                    *)
               System        : BOOLEAN ;    (* True if this module was       *)
                                            (* defined as a $SYSTEM$ module. *)
                                            (* If true the addresses are     *)
                                            (* absolute not relative.        *)
               ContainsHiddenType: BOOLEAN ;(* True if this module           *)
                                            (* implements a hidden type.     *)
               ListOfVars    : List ;       (* List of variables in this     *)
                                            (* scope.                        *)
               ListOfProcs   : List ;       (* List of all procedures        *)
                                            (* declared within this module.  *)
            END ;

   SymModule =
            RECORD
               Name          : CARDINAL ;   (* Index into name array, name   *)
                                            (* of record field.              *)
               Size          : PtrToValue ; (* Runtime size of symbol.       *)
               Offset        : PtrToValue ; (* Offset at runtime of symbol   *)
               LocalSymbols  : SymbolTree ; (* The LocalSymbols hold all the *)
                                            (* variables declared local to   *)
                                            (* the block. It contains the    *)
                                            (* IMPORT r ;                    *)
                                            (* FROM _ IMPORT x, y, x ;       *)
                                            (*    and also                   *)
                                            (* MODULE WeAreHere ;            *)
                                            (*    x y z visiable by localsym *)
                                            (*    MODULE Inner ;             *)
                                            (*       EXPORT x, y, z ;        *)
                                            (*    END Inner ;                *)
                                            (* END WeAreHere.                *)
               ExportTree    : SymbolTree ; (* Holds all the exported        *)
                                            (* identifiers.                  *)
                                            (* This tree may be              *)
                                            (* deleted at the end of Pass 1. *)
               IncludeList   : List ;       (* Contains all included symbols *)
                                            (* which are included by         *)
                                            (* IMPORT modulename ;           *)
                                            (* modulename.Symbol             *)
               ImportTree    : SymbolTree ; (* Contains all IMPORTed         *)
                                            (* identifiers.                  *)
               ExportUndeclared: SymbolTree ;
                                            (* ExportUndeclared contains all *)
                                            (* the identifiers which were    *)
                                            (* exported but have not yet     *)
                                            (* been declared.                *)
               EnumerationScopeList: List ; (* Enumeration scope list which  *)
                                            (* contains a list of all        *)
                                            (* enumerations which are        *)
                                            (* visable within this scope.    *)
               Father        : CARDINAL ;   (* Father scope of module.       *)
               Priority      : CARDINAL ;   (* Priority of the module. This  *)
                                            (* is an index to a constant.    *)
               Unresolved    : SymbolTree ; (* All symbols currently         *)
                                            (* unresolved in this module.    *)
               StartQuad     : CARDINAL ;   (* Signify the initialization    *)
                                            (* code.                         *)
               EndQuad       : CARDINAL ;   (* EndQuad should point to a     *)
                                            (* goto quad.                    *)
               System        : BOOLEAN ;    (* True if this module was       *)
                                            (* defined as a $SYSTEM$ module. *)
                                            (* If true the addresses are     *)
                                            (* absolute not relative.        *)
               ListOfVars    : List ;       (* List of variables in this     *)
                                            (* scope.                        *)
               ListOfProcs   : List ;       (* List of all procedures        *)
                                            (* declared within this module.  *)
            END ;

   SymDummy =
            RECORD
               NextFree     : CARDINAL      (* Link to the next free symbol. *)
            END ;


   Symbol = RECORD
               CASE SymbolType : TypeOfSymbol OF
                                            (* Determines the type of symbol *)

               RecordSym           : Record           : SymRecord |
               VarientSym          : Varient          : SymVarient |
               VarSym              : Var              : SymVar |
               EnumerationSym      : Enumeration      : SymEnumeration |
               SubrangeSym         : Subrange         : SymSubrange |
               SubscriptSym        : Subscript        : SymSubscript |
               ArraySym            : Array            : SymArray |
               UnboundedSym        : Unbounded        : SymUnbounded |
               ConstVarSym         : ConstVar         : SymConstVar |
               ConstLitSym         : ConstLit         : SymConstLit |
               ConstStringSym      : ConstString      : SymConstString |
               VarParamSym         : VarParam         : SymVarParam |
               ParamSym            : Param            : SymParam |
               UndefinedSym        : Undefined        : SymUndefined |
               TypeSym             : Type             : SymType |
               PointerSym          : Pointer          : SymPointer |
               RecordFieldSym      : RecordField      : SymRecordField |
               VarientFieldSym     : VarientField     : SymVarientField |
               EnumerationFieldSym : EnumerationField : SymEnumerationField |
               DefImpSym           : DefImp           : SymDefImp |
               ModuleSym           : Module           : SymModule |
               SetSym              : Set              : SymSet |
               ProcedureSym        : Procedure        : SymProcedure |
               ProcTypeSym         : ProcType         : SymProcType |
               DummySym            : Dummy            : SymDummy

               END
            END ;

   CallFrame = RECORD
                  Main  : CARDINAL ;  (* Main scope for insertions        *)
                  Search: CARDINAL ;  (* Search scope for symbol searches *)
                  Start : CARDINAL ;  (* ScopePtr value before StartScope *)
                                      (* was called.                      *)
               END ;


VAR
   Symbols       : ARRAY [1..MaxSymbols] OF Symbol ;
   ScopeCallFrame: ARRAY [1..MaxScopes] OF CallFrame ;
   FreeSymbol    : CARDINAL ;    (* The Head of the free symbol list   *)
   DefModuleTree : SymbolTree ;
   ModuleTree    : SymbolTree ;  (* Tree of all modules ever used.     *)
   ConstLitStringTree
                 : SymbolTree ;  (* String Literal Constants only need *)
                                 (* to be declared once.               *)
   ConstLitTree  : SymbolTree ;  (* Numerical Literal Constants only   *)
                                 (* need to be declared once.          *)
   CurrentModule : CARDINAL ;    (* Index into symbols determining the *)
                                 (* current module being compiled.     *)
   MainModule    : CARDINAL ;    (* Index into symbols determining the *)
                                 (* module the user requested to       *)
                                 (* compile.                           *)
   ScopePtr      : CARDINAL ;    (* An index to the ScopeCallFrame.    *)
                                 (* ScopePtr determines the top of the *)
                                 (* ScopeCallFrame.                    *)
   BaseScopePtr  : CARDINAL ;    (* An index to the ScopeCallFrame of  *)
                                 (* the top of BaseModule. BaseModule  *)
                                 (* is always left at the bottom of    *)
                                 (* stack since it is used so          *)
                                 (* frequently. When the BaseModule    *)
                                 (* needs to be searched the ScopePtr  *)
                                 (* is temporarily altered to          *)
                                 (* BaseScopePtr and GetScopeSym is    *)
                                 (* called.                            *)
   BaseModule    : CARDINAL ;    (* Index to the symbol table of the   *)
                                 (* Base pseudo modeule declaration.   *)
   TemporaryNo   : CARDINAL ;    (* The next temporary number.         *)

VAR
   c: CARDINAL ;
BEGIN
   FOR c := 1 TO MaxSymbols DO
      Symbols[c].SymbolType := UndefinedSym
   END
END varient5.
