IMPLEMENTATION MODULE RealIO ;


FROM StrIO IMPORT ReadString, WriteString ;
FROM StrLib IMPORT StrLen ;
FROM ASCII IMPORT nul ;


CONST
   MaxLineLength = 79 ;


PROCEDURE ReadReal (VAR x: REAL) ;
VAR
   a: ARRAY [0..MaxLineLength] OF CHAR ;
BEGIN
   ReadString( a ) ;
   StrToReal( a, x )
END ReadReal ;


PROCEDURE WriteReal (x: REAL ; n: CARDINAL) ;
VAR
   a: ARRAY [0..MaxLineLength] OF CHAR ;
BEGIN
   RealToStr( x, n, a ) ;
   WriteString( a )
END WriteReal ;


PROCEDURE StrToReal (a: ARRAY OF CHAR ; VAR x: REAL) ;
VAR
   i, high : CARDINAL ;
   negative: BOOLEAN ;
   fraction: REAL ;
   exp     : REAL ;
BEGIN
   high := StrLen(a) ;
   i := 0 ;
   IF (i<high) AND (a[i]='-')
   THEN
      negative := TRUE ;
      INC(i)
   ELSE
      negative := FALSE
   END ;
   x := 0.0 ;
   WHILE (i<high) AND (a[i]>='0') AND (a[i]<='9') DO
      x := x*10.0+FLOAT(ORD(a[i])-ORD('0')) ;
      INC(i)
   END ;
   IF i<high
   THEN
      IF a[i]='.'
      THEN
         exp := 10.0 ;
         fraction := 0.0 ;
         INC(i) ;
         WHILE (i<high) AND (a[i]>='0') AND (a[i]<='9') DO
            fraction := fraction+FLOAT(ORD(a[i])-ORD('0'))/exp ;
            exp := exp*10.0 ;
            INC(i)
         END ;
         x := x+fraction
      END
   END
END StrToReal ;


(*
   Power10 - returns a REAL the value of 10 ^ power
*)

PROCEDURE Power10 (r: REAL; power: CARDINAL) : REAL;
VAR 
   i: CARDINAL;
BEGIN
   FOR i  := 1 TO power DO
      r := r * 10.0
   END ;
   RETURN( r )
END Power10 ;


PROCEDURE RealToStr (x: REAL ; n: CARDINAL ; VAR a: ARRAY OF CHAR) ;
VAR
   b       : ARRAY [0..MaxLineLength] OF CARDINAL ;
   high,
   i, j    : CARDINAL ;
   negative: BOOLEAN ;
   exp     : REAL ;
BEGIN
   high := HIGH(a) ;
   IF x<0.0
   THEN
      negative := TRUE ;
      x := -x
   ELSE
      negative := FALSE
   END ;
   exp := 10.0 ;
   WHILE exp<=x DO
      exp := exp*10.0
   END ;
   exp := exp/10.0 ;
   j := 0 ;
   REPEAT
      b[j] := TRUNC(x/exp) ;
      x := x-FLOAT(b[j])*exp ;
      INC(j) ;
      exp := exp/10.0
   UNTIL exp<1.0 ;
   IF negative AND (n>0)
   THEN
      DEC(n)
   END ;
   i := 0 ;
   WHILE (n>j) AND (i<high) DO
      a[i] := ' ' ;
      INC(i) ;
      DEC(n)
   END ;
   IF negative AND (i<high)
   THEN
      a[i] := '-' ;
      INC(i)
   END ;
   n := 0 ;
   WHILE (n<j) AND (i<high) DO
      a[i] := CHR(b[n]+ORD('0')) ;
      INC(i) ;
      INC(n)
   END ;
   IF i<high
   THEN
      a[i] := nul
   END
END RealToStr ;


PROCEDURE ReadLongReal (VAR x: LONGREAL) ;
VAR
   a: ARRAY [0..MaxLineLength] OF CHAR ;
BEGIN
   ReadString( a ) ;
   StrToLongReal( a, x )
END ReadLongReal ;


PROCEDURE WriteLongReal (x: LONGREAL ; n: CARDINAL) ;
VAR
   a: ARRAY [0..MaxLineLength] OF CHAR ;
BEGIN
   LongRealToStr( x, n, a ) ;
   WriteString( a )
END WriteLongReal ;


PROCEDURE StrToLongReal (a: ARRAY OF CHAR ; VAR x: LONGREAL) ;
VAR
   i, high : CARDINAL ;
   negative: BOOLEAN ;
   fraction: LONGREAL ;
   exp     : LONGREAL ;
BEGIN
   high := StrLen(a) ;
   i := 0 ;
   IF (i<high) AND (a[i]='-')
   THEN
      negative := TRUE ;
      INC(i)
   ELSE
      negative := FALSE
   END ;
   x := 0.0 ;
   WHILE (i<high) AND (a[i]>='0') AND (a[i]<='9') DO
      x := x*10.0+FLOAT(ORD(a[i])-ORD('0')) ;
      INC(i)
   END ;
   IF i<high
   THEN
      IF a[i]='.'
      THEN
         exp := 10.0 ;
         fraction := 0.0 ;
         INC(i) ;
         WHILE (i<high) AND (a[i]>='0') AND (a[i]<='9') DO
            fraction := fraction+FLOAT(ORD(a[i])-ORD('0'))/exp ;
            exp := exp*10.0 ;
            INC(i)
         END ;
         x := x+fraction
      END
   END
END StrToLongReal ;


PROCEDURE LongRealToStr (x: LONGREAL ; n: CARDINAL ; VAR a: ARRAY OF CHAR) ;
VAR
   b       : ARRAY [0..MaxLineLength] OF CARDINAL ;
   high,
   i, j    : CARDINAL ;
   negative: BOOLEAN ;
   exp     : LONGREAL ;
BEGIN
   high := HIGH(a) ;
   IF x<0.0
   THEN
      negative := TRUE ;
      x := -x
   ELSE
      negative := FALSE
   END ;
   exp := 10.0 ;
   WHILE exp<=x DO
      exp := exp*10.0
   END ;
   exp := exp/10.0 ;
   j := 0 ;
   REPEAT
      b[j] := TRUNC(x/exp) ;
      x := x-FLOAT(b[j])*exp ;
      INC(j) ;
      exp := exp/10.0
   UNTIL exp<1.0 ;
   IF negative AND (n>0)
   THEN
      DEC(n)
   END ;
   i := 0 ;
   WHILE (n>j) AND (i<high) DO
      a[i] := ' ' ;
      INC(i) ;
      DEC(n)
   END ;
   IF negative AND (i<high)
   THEN
      a[i] := '-' ;
      INC(i)
   END ;
   n := 0 ;
   WHILE (n<j) AND (i<high) DO
      a[i] := CHR(b[n]+ORD('0')) ;
      INC(i) ;
      INC(n)
   END ;
   IF i<high
   THEN
      a[i] := nul
   END
END LongRealToStr ;


END RealIO.
