IMPLEMENTATION MODULE FpuIO ;


FROM StrIO IMPORT ReadString, WriteString, WriteLn ;
FROM StrLib IMPORT StrLen, StrRemoveWhitePrefix ;
FROM ASCII IMPORT nul ;


(* %%%FORWARD%%%


   %%%FORWARD%%% *)


CONST
   MaxLineLength = 79 ;
   MaxInt        = 2147483647 ;
   MaxDigits     = 20 ;


(*
   Power10 - returns a REAL the value of 10 ^ power
*)

PROCEDURE Power10 (r: LONGREAL; power: CARDINAL) : REAL;
VAR 
   i: CARDINAL;
BEGIN
   i := 0 ;
   WHILE i<power DO
      r := r * 10.0 ;
      INC(i)
   END ;
   RETURN( VAL(REAL, r) )
END Power10 ;



PROCEDURE ReadReal (VAR x: REAL) ;
VAR
   a: ARRAY [0..MaxLineLength] OF CHAR ;
BEGIN
   ReadString( a ) ;
   StrToReal( a, x )
END ReadReal ;


(*
   WriteReal - converts a REAL number, Real, which has a, TotalWidth, and
               FractionWidth into a string.
*)

PROCEDURE WriteReal (Real: REAL; TotalWidth, FractionWidth: CARDINAL) ;
VAR
   a: ARRAY [0..MaxLineLength] OF CHAR ;
BEGIN
   RealToStr(Real, TotalWidth, FractionWidth, a) ;
   WriteString(a)
END WriteReal ;


PROCEDURE StrToReal (a: ARRAY OF CHAR ; VAR x: REAL) ;
VAR
   lr: LONGREAL ;
BEGIN
   StrToLongReal(a, lr) ;
   x := lr
END StrToReal ;


PROCEDURE ReadLongReal (VAR x: LONGREAL) ;
VAR
   a: ARRAY [0..MaxLineLength] OF CHAR ;
BEGIN
   ReadString( a ) ;
   StrToLongReal( a, x )
END ReadLongReal ;


(*
   WriteLongReal - converts a LONGREAL number, Real, which has a, TotalWidth, and
                   FractionWidth into a string.
*)

PROCEDURE WriteLongReal (Real: LONGREAL; TotalWidth, FractionWidth: CARDINAL) ;
VAR
   a: ARRAY [0..MaxLineLength] OF CHAR ;
BEGIN
   LongRealToStr(Real, TotalWidth, FractionWidth, a) ;
   WriteString(a)
END WriteLongReal ;


PROCEDURE StrToLongReal (a: ARRAY OF CHAR ; VAR x: LONGREAL) ;
VAR
   i, high : CARDINAL ;
   negative: BOOLEAN ;
   fraction: LONGREAL ;
   exp     : LONGREAL ;
BEGIN
   StrRemoveWhitePrefix(a, a) ;
   high := StrLen(a) ;
   i := 0 ;
   IF (i<high) AND (a[i]='-')
   THEN
      negative := TRUE ;
      INC(i)
   ELSE
      negative := FALSE
   END ;
   x := 0.0 ;
   WHILE (i<high) AND (a[i]>='0') AND (a[i]<='9') DO
      x := x*10.0+FLOAT(ORD(a[i])-ORD('0')) ;
      INC(i)
   END ;
   IF i<high
   THEN
      IF a[i]='.'
      THEN
         exp := 10.0 ;
         fraction := 0.0 ;
         INC(i) ;
         WHILE (i<high) AND (a[i]>='0') AND (a[i]<='9') DO
            fraction := fraction+FLOAT(ORD(a[i])-ORD('0'))/exp ;
            exp := exp*10.0 ;
            INC(i)
         END ;
         x := x+fraction
      END
   END ;
   IF negative
   THEN
      x := -x
   END
END StrToLongReal ;


(*
   Add - places, ch, into, a, at index, i, providing, i<High
*)

PROCEDURE Add (VAR a: ARRAY OF CHAR; High: CARDINAL; i: CARDINAL; ch: CHAR) ;
BEGIN
   IF i<High
   THEN
      a[i] := ch
   END
END Add ;


(*
   RealToStr - converts a LONGREAL number, Real, which has, TotalWidth, and
               FractionWidth into a string.
*)

PROCEDURE RealToStr (Real: REAL; TotalWidth, FractionWidth: CARDINAL; VAR a: ARRAY OF CHAR) ;
VAR
   integer, 
   int         : INTEGER;
   firstwidth,
   fieldwidth,
   integerwidth,
   ccount, 
   count, 
   rcount      : CARDINAL;
   remainder   : LONGREAL;
   neg         : BOOLEAN;
   FinalIndex  : CARDINAL ;
   buf         : ARRAY [0..80] OF CHAR;
   LogPower    : CARDINAL ;
   MaxPower,
   MaxTrunc    : REAL ;
   High        : CARDINAL ;
BEGIN
   High := HIGH(a) ;
   MaxTrunc := FLOAT(MaxInt)/10.0 ;  (* 80387 doesn't like performing TRUNC on 1E+10 *)
   MaxPower := 1.0 ;
   LogPower := 0 ;
   WHILE MaxPower*10.0<MaxTrunc DO
      MaxPower := MaxPower * 10.0 ;
      INC(LogPower)
   END ;

   neg := Real<0.0 ;
   IF neg
   THEN
      Real := -Real
   END;
   FOR count := 0 TO HIGH(buf) DO
      buf[count] := '0'
   END ;
   count := 0 ;
   REPEAT
      (* Divide it down until it can safely be TRUNCed	*)
      rcount := 0 ;
      WHILE Real/Power10(1.0, rcount) >= MaxTrunc DO
         INC(rcount) ;
      END ;
      Real := Real / Power10(1.0, rcount) ;

      (*  Now store the non-fractional digits *)
      integer := TRUNC(Real) ;
      int := integer ;
      ccount := 0 ;
      REPEAT
         buf[ccount+rcount] := CHR((CARDINAL (int MOD 10)) + ORD('0')) ;
         int := int DIV 10 ;
         INC(ccount)
      UNTIL int=0 ;
      IF count=0
      THEN
         count := rcount + ccount
      END ;

      (*  recover fractional part and restore to true significance *)
      remainder := Real - FLOAT(integer) ;
      IF rcount=0
      THEN
         Real := remainder
      ELSE
         Real := Power10(remainder, rcount)
      END
   UNTIL rcount = 0 ;

   IF neg
   THEN
      buf[count] := '-' ;
      INC(count)
   END ;

   integerwidth := count ;
   firstwidth := integerwidth ;

   IF FractionWidth>0
   THEN
      (*  make room for decimal point *)
      INC(firstwidth)
   END ;
   IF firstwidth>=TotalWidth
   THEN
      integerwidth := TotalWidth
   ELSE
      IF firstwidth+FractionWidth<TotalWidth
      THEN
         integerwidth := TotalWidth ;
         IF FractionWidth>0
         THEN
            DEC(integerwidth, FractionWidth+1)
         END
      END
   END ;

   FinalIndex := 0 ;
   fieldwidth := integerwidth ;
   WHILE fieldwidth>TotalWidth DO
      Add(a, High, FinalIndex, ' ') ;
      INC(FinalIndex) ;
      DEC(fieldwidth)
   END ;

   WHILE FinalIndex+count<integerwidth DO
      Add(a, High, FinalIndex, ' ') ;
      INC(FinalIndex)
   END ;
   
   REPEAT
      DEC(count);
      Add(a, High, FinalIndex, buf[count]) ;
      INC(FinalIndex)
   UNTIL count=0 ;
   IF integerwidth<TotalWidth
   THEN
      Add(a, High, FinalIndex, '.') ;
      INC(FinalIndex) ;
      count := integerwidth+2 ;
      WHILE count<=TotalWidth DO
         remainder := remainder * MaxPower ;
         integer := TRUNC(remainder) ;
         remainder := remainder - FLOAT(integer) ;
         ccount := 0 ;
         REPEAT
            buf[ccount] := CHR((CARDINAL (integer MOD 10)) + ORD('0')) ;
            integer := integer DIV 10 ;
            INC(ccount)
         UNTIL integer=0 ;
         INC(count, ccount) ;
         (* add leading zero's *)
         int := ccount ;
         WHILE (int<LogPower) AND (FinalIndex<High) DO
            Add(a, High, FinalIndex, '0') ;
            INC(FinalIndex) ;
            INC(int)
         END ;
         REPEAT
            DEC(ccount) ;
            Add(a, High, FinalIndex, buf[ccount]) ;
            INC(FinalIndex)
         UNTIL ccount=0
      END ;
      IF FinalIndex>TotalWidth
      THEN
         FinalIndex := TotalWidth
      END
   END ;
   Add(a, High, FinalIndex, nul)
END RealToStr ;


(*
   LongRealToStr - converts a LONGREAL number, Real, which has, TotalWidth, and
                   FractionWidth into a string.
*)

PROCEDURE LongRealToStr (Real: LONGREAL; TotalWidth, FractionWidth: CARDINAL; VAR a: ARRAY OF CHAR) ;
VAR
   integer, 
   int         : INTEGER;
   firstwidth,
   fieldwidth,
   integerwidth,
   ccount, 
   count, 
   rcount      : CARDINAL;
   remainder   : LONGREAL;
   neg         : BOOLEAN;
   FinalIndex  : CARDINAL ;
   buf         : ARRAY [0..80] OF CHAR;
   LogPower    : CARDINAL ;
   MaxPower,
   MaxTrunc    : LONGREAL ;
   High        : CARDINAL ;
BEGIN
   High := HIGH(a) ;
   MaxTrunc := FLOAT(MaxInt)/10.0 ;  (* 80387 doesn't like performing TRUNC on 1E+10 *)
   MaxPower := 1.0 ;
   LogPower := 0 ;
   WHILE MaxPower*10.0<MaxTrunc DO
      MaxPower := MaxPower * 10.0 ;
      INC(LogPower)
   END ;

   neg := Real<0.0 ;
   IF neg
   THEN
      Real := -Real
   END;
   FOR count := 0 TO HIGH(buf) DO
      buf[count] := '0'
   END ;
   count := 0 ;
   REPEAT
      (* Divide it down until it can safely be TRUNCed	*)
      rcount := 0 ;
      WHILE Real/Power10(1.0, rcount) >= MaxTrunc DO
         INC(rcount) ;
      END ;
      Real := Real / Power10(1.0, rcount) ;

      (*  Now store the non-fractional digits *)
      integer := TRUNC(Real) ;
      int := integer ;
      ccount := 0 ;
      REPEAT
         buf[ccount+rcount] := CHR((CARDINAL (int MOD 10)) + ORD('0')) ;
         int := int DIV 10 ;
         INC(ccount)
      UNTIL int=0 ;
      IF count=0
      THEN
         count := rcount + ccount
      END ;

      (*  recover fractional part and restore to true significance *)
      remainder := Real - FLOAT(integer) ;
      IF rcount=0
      THEN
         Real := remainder
      ELSE
         Real := Power10(remainder, rcount)
      END
   UNTIL rcount = 0 ;

   IF neg
   THEN
      buf[count] := '-' ;
      INC(count)
   END ;

   integerwidth := count ;
   firstwidth := integerwidth ;

   IF FractionWidth>0
   THEN
      (*  make room for decimal point *)
      INC(firstwidth)
   END ;
   IF firstwidth>=TotalWidth
   THEN
      integerwidth := TotalWidth
   ELSE
      IF firstwidth+FractionWidth<TotalWidth
      THEN
         integerwidth := TotalWidth ;
         IF FractionWidth>0
         THEN
            DEC(integerwidth, FractionWidth+1)
         END
      END
   END ;

   FinalIndex := 0 ;
   fieldwidth := integerwidth ;
   WHILE fieldwidth>TotalWidth DO
      Add(a, High, FinalIndex, ' ') ;
      INC(FinalIndex) ;
      DEC(fieldwidth)
   END ;
   
   REPEAT
      DEC(count);
      Add(a, High, FinalIndex, buf[count]) ;
      INC(FinalIndex)
   UNTIL count=0 ;
   IF integerwidth<TotalWidth
   THEN
      Add(a, High, FinalIndex, '.') ;
      INC(FinalIndex) ;
      count := integerwidth+2 ;
      WHILE count<=TotalWidth DO
         remainder := remainder * MaxPower ;
         integer := TRUNC(remainder) ;
         remainder := remainder - VAL(LONGREAL, integer) ;
         ccount := 0 ;
         REPEAT
            buf[ccount] := CHR((CARDINAL (integer MOD 10)) + ORD('0')) ;
            integer := integer DIV 10 ;
            INC(ccount)
         UNTIL integer=0 ;
         INC(count, ccount) ;
         (* add leading zero's *)
         int := ccount ;
         WHILE (int<LogPower) AND (FinalIndex<High) DO
            Add(a, High, FinalIndex, '0') ;
            INC(FinalIndex) ;
            INC(int)
         END ;
         REPEAT
            DEC(ccount) ;
            Add(a, High, FinalIndex, buf[ccount]) ;
            INC(FinalIndex)
         UNTIL ccount=0
      END ;
      IF FinalIndex>TotalWidth
      THEN
         FinalIndex := TotalWidth
      END
   END ;
   Add(a, High, FinalIndex, nul)
END LongRealToStr ;


PROCEDURE ReadLongInt (VAR x: LONGINT) ;
VAR
   a : ARRAY [0..MaxLineLength] OF CHAR ;
BEGIN
   ReadString( a ) ;
   StrToLongInt( a, x )
END ReadLongInt ;


PROCEDURE WriteLongInt (x: LONGINT; n: CARDINAL) ;
VAR
   a : ARRAY [0..MaxLineLength] OF CHAR ;
BEGIN
   LongIntToStr( x, n, a ) ;
   WriteString( a )
END WriteLongInt ;


PROCEDURE LongIntToStr (x: LONGINT; n: CARDINAL ; VAR a: ARRAY OF CHAR) ;
VAR
   i, j,
   Higha    : CARDINAL ;
   buf      : ARRAY [1..MaxDigits] OF CARDINAL ;
   Negative : BOOLEAN ;
BEGIN
   IF x<0
   THEN
      Negative := TRUE ;
      IF n>0
      THEN
         DEC(n)
      END ;
      x := -x
   ELSE
      Negative := FALSE ;
   END ;
   i := 0 ;
   REPEAT
      INC(i) ;
      IF i>MaxDigits
      THEN
         WriteString('increase MaxDigits in FpuIO') ; WriteLn ;
         HALT
      END ;
      buf[i] := VAL(CARDINAL, x MOD 10) ;
      x := x DIV 10 ;
   UNTIL x=0 ;
   j := 0 ;
   Higha := HIGH(a) ;
   WHILE (n>i) AND (j<=Higha) DO
      a[j] := ' ' ;
      INC(j) ;
      DEC(n)
   END ;
   IF Negative
   THEN
      a[j] := '-' ;
      INC(j)
   END ;
   WHILE (i#0) AND (j<=Higha) DO
      a[j] := CHR( buf[i] + ORD('0') ) ;
      INC(j) ;
      DEC(i)
   END ;
   IF j<=Higha
   THEN
      a[j] := nul
   END
END LongIntToStr ;


PROCEDURE StrToLongInt (a: ARRAY OF CHAR ; VAR x: LONGINT) ;
VAR
   i        : CARDINAL ;
   ok,
   Negative : BOOLEAN ;
   higha    : CARDINAL ;
BEGIN
   StrRemoveWhitePrefix(a, a) ;
   higha := StrLen(a) ;
   i := 0 ;
   Negative := FALSE ;
   ok := TRUE ;
   WHILE ok DO
      IF i<higha
      THEN
         IF a[i]='-'
         THEN
            INC(i) ;
            Negative := NOT Negative
         ELSIF (a[i]<'0') OR (a[i]>'9')
         THEN
            INC(i)
         ELSE
            ok := FALSE
         END
      ELSE
         ok := FALSE
      END
   END ;
   x := 0 ;
   IF i<=higha
   THEN
      ok := TRUE ;
      REPEAT
         IF Negative
         THEN
            x := 10*x - INTEGER(ORD(a[i])-ORD('0'))
         ELSE
            x := 10*x + INTEGER(ORD(a[i])-ORD('0'))
         END ;
         IF i<higha
         THEN
            INC(i) ;
            IF (a[i]<'0') OR (a[i]>'9')
            THEN
               ok := FALSE
            END
         ELSE
            ok := FALSE
         END
      UNTIL NOT ok ;
   END
END StrToLongInt ;


END FpuIO.
