IMPLEMENTATION MODULE Jump ;


FROM SYSTEM IMPORT ADDRESS ;

TYPE
   RegEnv = RECORD
               eip,
               esp,
               ebp: ADDRESS ;
            END ;


(*
   SetJump - saves enough state to be used by LongJump.
             LongJump assumes that the procedure calling SetJump
             has not finished.
             SetJump will return zero when called direct
             and non zero when indirectly called by LongJump.
*)

PROCEDURE SetJump (VAR e: RegEnv) : INTEGER ;
BEGIN
   (*
    *  remember that the prologue to this function was:


      |---------|
      | &e      |
      |---------|
      | eip     |
      |---------|
      | ebp     |     <- ebp
      |---------|
      | ebx     |
      |---------|
      | edi     |
      |---------|
      | esi     |     <- esp
      |---------|
      |         |
      |---------|

   pushl   &e
   call    SetJump
   addl    %esp, 4
   if      %eax=0
   ...

SetJump:
   pushl   %ebp
   movl    %esp, %ebp
   pushl   %ebx
   pushl   %edi
   pushl   %esi

    * no need to decrease stack by local vars as they = 0
    *)

   ASM
      #
      # we save stack and frame pointer values which existed
      # before the call to SetJump
      #
      movl  8(%ebp), %ebx       # e
      movl  3*4(%esp), %esi     # old ebp into esi
      movl  %esi, 8(%ebx)       # and save it in e^.ebp

      movl  4*4(%esp), %esi     # old eip into esi
      movl  %esi, 0(%ebx)       # and save it in e^.eip

      movl  %esp, %esi          # calculate old esp
      addl  $5*4, %esi
      movl  %esi, 4(%ebx)       # and save it in e^.esp

      movl  $0, %eax            # and we have returned from SetJump
   END
END SetJump ;


(*
   LongJump - invokes the longjump returning the value, value.
              value must be non zero. If a value of zero is given
              one will be returned.
*)

PROCEDURE LongJump (VAR e: RegEnv; value: INTEGER) ;
BEGIN
   IF value=0
   THEN
      value := 1
   END ;
   ASM
      # we must "return" just after SetJump with value in %eax
      movl   12(%ebp), %eax    # value into %eax
      movl    8(%ebp), %ebx    # e into %ebx
      movl    8(%ebx), %ebp    # restore old ebp
      movl    4(%ebx), %esp    # finally restore old esp
      jmp     0(%ebx)          # jump to just after call to SetJump (magic!)
   END
END LongJump ;


END Jump.
