IMPLEMENTATION MODULE Queue ;

FROM SYSTEM IMPORT ADR, ADDRESS ;
FROM Storage IMPORT ALLOCATE, DEALLOCATE ;

TYPE
   QUEUE = POINTER TO RECORD
                         Left, Right: QUEUE ;
                         DataStart  : ADDRESS ;
                         DataSize   : CARDINAL ;
                      END ;


(*
   InitQueue - initializes a queue with one unfilled element.
*)

PROCEDURE InitQueue () : QUEUE ;
VAR
   q: QUEUE ;
BEGIN
   NEW(q) ;
   InitContents(q) ;
   RETURN( q )
END InitQueue ;


(*
   InitContents - set each field to a safe value.
*)

PROCEDURE InitContents (q: QUEUE) ;
BEGIN
   WITH q^ DO
      Left      := q ;
      Right     := q ;
      DataStart := NIL ;
      DataSize  := 0
   END
END InitContents ;


(*
   KillQueue - kills a queue (with all elements) and returns NIL
*)

PROCEDURE KillQueue (q: QUEUE) : QUEUE ;
VAR
   p, t: QUEUE ;
BEGIN
   IF p#NIL
   THEN
      p := q ;
      REPEAT
         IF p^.DataSize>0
         THEN
            DEALLOCATE(p^.DataStart, p^.DataSize-1)
         END ;
         t := p^.Right ;
         DISPOSE(t) ;
         p := t
      UNTIL p=q ;
   END ;
   RETURN( NIL )
END KillQueue ;


(*
   PutData - places the, data, into queue element, q, and returns
             q.
*)

PROCEDURE PutData (q: QUEUE; VAR data: ARRAY OF BYTE) : QUEUE ;
VAR
   p: POINTER TO BYTE ;
   i: CARDINAL ;
BEGIN
   WITH q^ DO
      DataSize  := HIGH(data)+1 ;
      ALLOCATE(DataStart, DataSize) ;
      i := 0 ;
      p := DataStart ;
      WHILE i<DataSize DO
         p^ := data[i] ;
         INC(p) ;
         INC(i)
      END
   END ;
   RETURN( q )
END PutData ;


(*
   GetData - gets data from the queue element, q, and returns q.
*)

PROCEDURE GetData (q: QUEUE; VAR data: ARRAY OF BYTE) : QUEUE ;
VAR
   i: CARDINAL ;
   p: POINTER TO BYTE ;
BEGIN
   WITH q^ DO
      i := 0 ;
      p := DataStart ;
      WHILE i<DataSize DO
         data[i] := p ;
         INC(i) ;
         INC(p)
      END ;
      DEALLOCATE(DataStart, DataSize)
   END ;
   RETURN( q )
END GetData ;


(*
   IsEmpty - returns TRUE if the queue, q, is empty
*)

PROCEDURE IsEmpty (q: QUEUE) : BOOLEAN ;
BEGIN
   RETURN( q=NIL )
END IsEmpty ;


(*
   MoveRight - returns the right neighbough of q.
*)

PROCEDURE MoveRight (q: QUEUE) : QUEUE ;
BEGIN
   RETURN( q^.Right )
END MoveRight ;


(*
   MoveLeft - returns the left neighbough of q.
*)

PROCEDURE MoveLeft (q: QUEUE) : QUEUE ;
BEGIN
   RETURN( q^.Left )
END MoveLeft ;


(*
   AddToQueue - adds an element the end of a queue.
                It returns the new element added to the end of the queue.
*)
 
PROCEDURE AddToQueue (VAR h: QUEUE; e: QUEUE) : QUEUE ;
BEGIN
   IF h=NIL
   THEN
      h := e ;
      e^.Left := e ;
      e^.Right := e
   ELSE
      e^.Right := h ;
      e^.Left := h^.Left ;
      h^.Left^.Right := e ;
      h^.Left := e
   END ;
   RETURN( e )
END AddToQueue ;
 
 
(*
   SubFromQueue - removes an element from Queue.
                  It returns the element, e, removed from
                  the end of the queue.
*)
 
PROCEDURE SubFromQueue (VAR h: QUEUE; e: QUEUE) : QUEUE ;
BEGIN
   IF (e^.Left=e) AND (e=h)
   THEN
      h := NIL
   ELSE
      IF h=e
      THEN
         h := h^.Right
      END ;
      e^.Left^.Right := e^.Right ;
      e^.Right^.Left := e^.Left
   END ;
   RETURN( e )
END SubFromQueue ;


END Queue.
