IMPLEMENTATION MODULE SysStorage ;

(*
   Standard dynamic storage management

Storage management for dynamic variables. Calls to the
Modula-2 standard procedures NEW and DISPOSE are
translated into calls to ALLOCATE and DEALLOCATE. The
standard way to provide these two procedures is to
import them from this module 'Storage'.

*)

FROM libc IMPORT malloc, cfree ;
FROM Debug IMPORT Halt ;


(*
- Allocate some dynamic storage (contiguous memory area).

in:     Size    number of bytes to allocate,

out:    a       ADDRESS of allocated storage.
*)

PROCEDURE ALLOCATE (VAR a: ADDRESS ; Size: CARDINAL) ;
BEGIN
   a := malloc(Size) ;
   IF a=NIL
   THEN
      Halt('out of memory error', __LINE__, __FILE__)
   END
END ALLOCATE ;


(*
- Release some dynamic storage (contiguous memory area).

in:     a       ADDRESS of the area to release,
        Size    number of bytes to be released,

out:    a       set to NIL.

The storage area released is made available for subsequent
calls to ALLOCATE.
*)

PROCEDURE DEALLOCATE (a: ADDRESS; Size: CARDINAL);
BEGIN
   cfree(a)
END DEALLOCATE ;


(*
- Test whether some number of bytes could be allocated.

in:     Size    number of bytes

out:    TRUE if ALLOCATE (p, Size) would succeed.
*)

PROCEDURE Available (Size: CARDINAL) : BOOLEAN;
VAR
   a: ADDRESS ;
BEGIN
   a := malloc(Size) ;
   IF a=NIL
   THEN
      RETURN( FALSE )
   ELSE
      cfree(a) ;
      RETURN( TRUE )
   END
END Available ;


(*
   Init - 
*)

PROCEDURE Init ;
BEGIN
END Init ;


END SysStorage.
