.TH M2F L
.SH NAME
m2f \- a complete modula-2 compiler conforming to PIM-2 Edition 2
.SH SYNOPSIS
.B "m2f [-l] [-g] [-ng] [-p] [-bounds] [-students] [-Odynamic] [-Oi586] [-Oi686] [-On] [-O] [-Ouncalled] [-Omemcopy] [-nomath] [-M2RTS] [-m] [-S] [-16|-32] [-i386|-apu] [-gas|-masm|-bas] [-dryrun] [-v] [-o \fIdestinition\fB] [-orderfile \fIyourfile\fB] [-orderonly] [-quiet] [-verbose] [-statistics] [-pedantic] [-M \fISearchPath\fB] \fIModuleName\fB"
.br
.SH DESCRIPTION
.I m2f
is the front end to gm2, a modula-2 compiler.
It takes a number of arguments and invokes many subcomponents to
build, assemble and link a modula-2 application.
Many of the arguments given to \fIm2f\fR are in fact passed through to
the subcomponents. For example in the synopsis given above the \fB-g\fR
and the \fB-quiet\fR onwards are passed through to the main compiler
component \fIgm2\fR.
However the purpose of \fIm2f\fR is to shield the user of knowing the
interaction of the subcomponents by presenting a simple interface.
.br
.SH ENVIRONMENT
There are two useful environment variables that may be used with the
modula-2 utilities, these are \fBM2PATH\fR and \fBM2OPTIONS\fR.
The \fBM2PATH\fR can be set to determine the directory order that
the compiler will search to find library modules.
The path is specified by a path, a space and more paths ie:
.sp
.in +.5i
\fBexport M2PATH="/user/gaius/m2/comp/libs /usr/m2/libs"\fR
.in -.5i
.sp
This indicates that the compiler will search the directory /user/gaius/m2/comp/libs
before it searches /usr/m2/libs to find library modules.
Note that you need to explicitly place a \fB.\fR in the \fBM2PATH\fR
variable to make the compiler search your current working directory.
Note that the default path for the compiler may vary depending
upon the target. To find out the default path type: \fIm2f -apu -h\fR
or \fIm2f -gas -h\fR or \fIm2f -bas -h\fR.
.br
The \fBM2OPTIONS\fR environment variable can be set to contain any of
the non path related options. Thus whenever the compiler is invoked
\fBwithout\fR any command line options these environment options
are used.
Note that if an option is specified on the command line then \fBM2OPTIONS\fR
is ignored. If the \fB-M\fR command line option is used then it too will
override the \fBM2PATH\fR environment variable.
.sp
This mechanism allows users to configure their environment and not
have to tediously specify for every compilation the options required.
.SH OPTIONS
.IP "-l"
link the module specified on the command line. The module is assumed to have
already been compiled.
.IP "-nomath"
tell the compiler that you do \fInot\fR expect to use the coprocessor.
The compiler will generate an error if any source code attempts to
generate coprocessor instructions. This option is useful for
realtime systems which are targetted for processors without
floating point instructions.
.IP "-M2RTS"
do not automatically link with the module M2RTS. You should not specify
this flag for normal UNIX applications. It \fImight\fR be appropriate
to use this flag for a realtime system.
(The linker will automatically include the module M2RTS even if no imports
are specified in your program module. The UNIX environment requires M2RTS
as HALT and some termination code is implemented inside this module).
.IP "-m"
at link stage display all the modules that are to be linked.
.IP "-g"
compile or link generating GNU GDB compatible debugging information.
Note that occasionally the line number debugging information is not
absolutely correct if the -O option is also supplied.
Most of the statements correspond with the line numbers but procedure
calls at the end of an IF THEN END statement sequence maybe misaligned
by a few source lines.
The -O option removes dead code and sometimes rearranges code sequences.
.IP "-ng"
do not generate GNU GDB compatible debugging information.
This option will override a \fB-g\fR found in the environment.
.IP "-bounds"
generate code to check the bounds of subranges and array indexes.
The default is that this flag is FALSE.
.IP "-return"
generate code to check that functions always exit with a RETURN and do not
fall out at the end.
The default is that this flag is FALSE.
.IP "-p"
generate profiling information. When the final binary is executed a
profile trace is generated and this can be interpreted by the command \fCgprof\fR.
.IP "-O"
switch on all code optimization within the compiler.
.IP "-Ocse"
turn \fIoff\fR common subexpressions optimization.
This is useful if you are debugging source code as this optimization
does not preserve line number ordering of assignments.
(Perhaps this is a bug?)
.IP "-Oi586"
turn on pentium (i586) processor optimizations, such as pentium instructions
and exploit the pentium pipeline. This can considerably improve the speed of
the generated code. It enables efficient and inteligent pipelining, which understands
cache lines, operand conflicts.
It also aligns data on word and 16 byte boundaries where appropriate.
can be used in conjunction with -Odynamic. This has just been implemented and has
undergone limited testing (used on the compiler and not much else). Only use if
desperate for speed.
This flag tells the compiler to schedule i586 instructions to exploit the pipeline.
.IP "-Oi686"
turn on pentium (i686) processor optimizations. Currently this only enables
the 686 cache line model and i586 optimizations.
.IP "-Ocache"
turn on cache optimizations.
.IP "-Oworst"
tries to generate the worst code for the pipeline. Hopefully only
used for testing out potential speed ups.
.IP "-Ouncalled"
remove all procedures which are never called.
Note that this may have no effect on the performance of the remaining code.
Although it may have a secondary performance improvement as the binary might be smaller
and thus consume less operating system resources.
.IP "-Odynamic"
use the dynamic programming code generator.
Produces much better code, as with most optimization techniques the debugging information
becomes vague. In particular line numbers will not exactly match the code generated.
Only use if you are desperate for speed.
This flag will eventually be enabled with the -O option
once the code generator has been thrashed through many different styles of code.
Do not use this flag with the -16 flag.
.IP "-On"
tells the pipeline optimizer to find the optimal pipeline scheduling ordering for the
next 3*n instructions. Default is -O2. Increasing beyond 4 causes compilation time
to noticably increase.
.IP "-Omemcopy"
use the \fCmemcopy\fR instruction for pushing and popping data from the stack.
This optimization flag is only appropriate for the \fI-apu\fR processor.
The default is not to use the flag.
.IP "-S"
compile the module specified on the command line, only generate an
assembly language file.
.IP "-16"
generate 16 bit code and data.
.IP "-32"
generate 32 bit code and data - the default.
.IP "-gas"
generate GNU GAS assembler syntax.
.IP "-masm"
generate Intel MASM assembler syntax.
.IP "-bas"
generate Bruce Evans assembler syntax. (Currently this only works with
the -16 flag).
.IP "-apu"
generate APU (Abtract Processor Code) machine language code.
.IP "-i386"
generate Intel i386 processor code.
.IP "-dryrun"
show what subcomponents will be invoked by \fIm2f\fR, do not perform
any compilation or linkage.
.IP "-v"
show the subcomponents that are invoked before actually invoking them.
.IP "-o \fIdestinition\fB"
place the output of the compiler into file \fIdestinition\fB.
.IP "-orderfile \fIyourfile\fR"
the initialization order is specified by \fIyourfile\fR.
.IP "-orderonly"
only create the order file, the name of which may have been
specified by the previous \fB-orderfile\fR option.
.IP "-quiet"
do not display any Pass 1, Pass 2, Pass 3, Pass 4 messages, frequently
found on modula-2 compilers.
.IP "-verbose"
if an error occurs the erroneous text is displayed with an arrow indicating
the approximate character in error.
The line number and file is always displayed, but sometimes the compiler
attributes the error to the following line to the error.
The \fB-verbose\fR option is more accurate and more verbose!
.IP "-statistics"
generates quadruple information: number of quadruples generated,
number of quadruples remaining after optimization.
.IP "-students"
checks for novice peculiarities. Note that these warnings are not
the same as error messages. Hardened users of Modula-2 may wish to ignore this
option as it checks to see whether a user has declared a variable name
twice which looks the same and whether a variable looks similar to a keyword.
(Case difference etc). This is of course quite legal Modula-2 but
some first year students might find this option useful?
.IP "-pedantic"
forces the compiler to reject nested \fCWITH\fR
statements referencing the same record type.
Does not allow multiple imports of the same item from a module.
It also checks that: procedure variables are written to before being read;
variables are not only written to but read from; variables are declared and used.
If the compiler encounters a variable being read before written it will terminate
with a message.
It will check that \fCFOR\fR loop indices are not used outside the end of this loop
without being reset.
.IP "-M \fISearchPath\fR"
tells the compiler which directories to search, to find library modules.
Typically one might specify \fB-M "/user/gaius/lib /user/m2/lib"\fR
to indicate that the directory /user/gaius/lib should be searched first,
and later /user/m2/lib should be searched if the required library module
was not found in /user/gaius/lib.
.sp
The module name should be the exact module name, NOT the (possibly truncated)
file name.
.SH "FEATURES"
The compiler generates error messages compatible with the emacs next error
system.
.sp
The compiler gives substantial help in certain annoying modula-2 errors.
For example it will list all the undeclared identifiers at the end of a
compilation, it also lists all exported identifiers which have not been
implemented, it will check parameters declarations carefully to see that
the same types and names are used in the definition module and implementation
module.
Also, provided the \fC-verbose\fR option is used, the compiler
will display any procedure parameters which conflict with the definition.
The compiler will display any procedure parameters which conflict with
declaration and give fairly precise details about where the conflict occurred.
.sp
The compiler allows abstract data types to be \fBANY\fR type, not just
restricted to a pointer type.
It also allows declarations in any order (type declarations
may be in any order the software developper desires, not the way the compiler
desires).
So for example constants and enumerated types may be declared
at the bottom of a module etc.
.SH "Interface to C"
There is an example of how to interface modula-2 in the library directory.
The modula-2 compiler can be configured to push parameters from right to
left (the default) or alternatively from left to right.
To alter the default you need to alter the BOOLEAN constant in
\fCM2Configure.def\fR and \fCmake.rules\fR.
.PP
An example of interfacing modula-2 to C is contained within
the interface file \fBlibc.def\fR and \fBlibc.c\fR.
.SH "SYSTEM MODULE"
The system module for this compiler really exists rather than being
a pseudo module as in many modula-2 compilers.
The compiler internally declares \fCADDRESS, WORD, BYTE, ADR, TSIZE,
SIZE\fR and these are exported by \fCSYSTEM.def\fR.
The module system is actually parsed by the compiler and therefore
one can add functions like \fCIOTRANSFER, TRANSFER, PROCESS,
NEWPROCESS\fR.
.SH "TYPES"
In addition to the base types normally found in Modula-2 compilers
the following types are also implemented: LONGINT and LONGREAL.
These types are 64 bits in length. The module MATH exports trigonometrical
functions such as sin, cos and tan which when compiled produce inline code.
.\" .SH BUGS/LIMITATIONS
.SH AUTHORS/CONTRIBUTORS
Gaius Mulley (gaius@glam.ac.uk) wrote the compiler.
A great debt of thanks goes to:
.sp
.in +.25i
- Duncan White for allowing me to use his Modula-2 file library
(FIO.def, FIO.mod, CIsupport.c)
.br
- Keith Verheyden for both improving and simplifying the buffering
in FIO.mod.
.br
- Stuart Lewis and all his budding first year programmers who
identified so many bad error messages over the last year,
identified so many warning improvements and for finding
bugs/limitations in various library modules.
.br
- Stuart Lewis for adding the Sort and Random library routines.
.in -.25i
.sp
without which the compiler would not be nearly as useful.
Please email \fCgaius@glam.ac.uk\fR with any
queries, suggestions, improvements, grumbles etc.
.SH
AVAILABILITY
.LP
Source is available from all tsx-11 mirror sites and also
from my machine \fCftp floppsie.comp.glam.ac.uk\fR
in directory \fCpub/m2\fR.
