IMPLEMENTATION MODULE M2GenDyn486 ;

(*
    Title      : M2GenDyn486
    Author     : Gaius Mulley
    System     : UNIX (gm2)
    Date       : Tue Nov 19 17:03:42 1996
    Last edit  : Tue Nov 19 17:03:42 1996
    Description: generates a tree for each basic block, then proceeds to walk the tree
                 producing reasonable code for the i586.
*)

FROM SYSTEM IMPORT BITSET ;
FROM ASCII IMPORT nul ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM NumberIO IMPORT WriteCard, WriteHex, IntToStr, CardToStr ;
FROM StdIO IMPORT PushOutput, PopOutput, Write ;
FROM M2Debug IMPORT Assert ;
FROM M2Options IMPORT Optimize586, APUCodeGenerator ;
FROM StrLib IMPORT StrConCat, StrLen, StrCopy ;
FROM Environment IMPORT GetEnvironment ;
FROM FpuIO IMPORT WriteLongReal ;

IMPORT IO ;

FROM SymbolTable IMPORT ModeOfAddr, NulSym, StartScope, EndScope,
                        MakeConstLit,
                        MakeTemporary, PutVar,
                        GetMode, GetStringLength,
                        GetSymName, GetType, GetScopeAuthor,
                        GetVarFather, GetSubrange, GetNth, GetMainModule, GetNthParam, GetString,
                        IsModule, IsTemporary, IsExported,
                        IsConst, IsConstString, IsProcedure, IsConstLit, IsProcType, IsVar, IsSubrange, IsArray, IsUnbounded,
                        IsVarParam, IsProcedureNested,
                        PushSumOfLocalVarSize,
                        PushSumOfParamSize,
                        PushValue, PopValue, PushSize, PushOffset,
                        GetVarReadQuads, GetVarWriteLimitQuads,
                        ForeachInnerModuleDo ;

FROM Lists IMPORT List, InitList, KillList, IsItemInList, IncludeItemIntoList,
                  ForeachItemInListDo, DuplicateList, RemoveItemFromList,
                  NoOfItemsInList, GetItemFromList ;

FROM M2Base IMPORT MixTypes, Integer, LongReal, ActivationPointer, IsMathType ;
FROM M2Inline IMPORT WriteInlineBuffer ;
FROM M2System IMPORT WordSize, Word, Address ;
FROM NameKey IMPORT MakeKey, WriteKey, GetKey ;
FROM M2Lexical IMPORT InternalError, WriteError, WriteErrorFormat1 ;
FROM M2Quads IMPORT QuadOperator, GetNextQuad, GetQuad, DisplayQuad, WriteOperator, IsReferenced ;
FROM M2ALU IMPORT PushCard, PopCard, Add, Sub, Mult, PopInt, PushInt, Less, Equ, PopLongReal ;
FROM M2AsmUtil IMPORT WriteAsmName, DotifyLabel, GetModuleInitName ;
FROM M2Constants IMPORT IsOne, IsZero, IsTwo, MakeNewConstFromValue, IsSame ;
FROM M2Error IMPORT BeginError ;

FROM M2Gen487 IMPORT IsCoprocessorUsed, FlushCoprocessor,
                     CoprocessorBecomes, CoprocessorParam, CoprocessorConvert, CoprocessorNeg,
                     CoprocessorAdd, CoprocessorSub, CoprocessorDiv, CoprocessorMult, CoprocessorMod,
                     CoprocessorXIndr, CoprocessorIndrX,
                     CoprocessorIfEqu, CoprocessorIfNotEqu,
                     CoprocessorIfLess, CoprocessorIfLessEqu,
                     CoprocessorIfGre, CoprocessorIfGreEqu,
                     CoprocessorFunctValue, CoprocessorReturnValue, CoprocessorMath, CoprocessorLineInfo ;

FROM M2GDB IMPORT LineInfo, ProcedureInfo,
                  VariableInfo, BeginInfo, EndInfo ;

FROM M2Pipe586 IMPORT CheckProfiling, WriteConstant,
                      InstRegConstant, InstConstantReg, InstMDConstant,
                      InstMRIConstant, InstMRICardinal, InstMRIReg, InstMRISymConstant,
                      InstMDReg, InstRegMD, InstRegMIO, InstRegMI,
                      InstMIOReg, InstMIReg, InstRegReg, InstRegInt,
                      InstRegHexCard, InstMDOffset, InstRegOffset, InstMDInt,
                      InstRegImmName, InstName, InstRegMISym, InstMISymReg,
                      InstStr,
                      BinInst, BinInstUnit, BinInstReg, BinInstRegI, BinInstConstant,
                      BinInstMD, BinInstHex, Comment,
                      MovePackedStringToIReg, MovePackedStringTo,
                      DeclareLocalVariables,
                      GenerateQuadLabel,
                      GenerateEntryProcedureLabel,
                      GenerateExitProcedureLabel,
                      WritePrologue, WriteEpilogue, FlushPipe ;

(*
FROM M2AS IMPORT CheckProfiling, WriteConstant,
                      InstRegConstant, InstConstantReg, InstMDConstant,
                      InstMRIConstant, InstMRICardinal, InstMRIReg, InstMRISymConstant,
                      InstMDReg, InstRegMD, InstRegMIO, InstRegMI,
                      InstMIOReg, InstMIReg, InstRegReg, InstRegInt,
                      InstRegHexCard, InstMDOffset, InstRegOffset, InstMDInt,
                      InstRegImmName, InstName, InstRegMISym, InstMISymReg,
                      InstStr,
                      BinInst, BinInstUnit, BinInstReg, BinInstRegI, BinInstConstant,
                      BinInstMD, BinInstHex, Comment,
                      MovePackedStringToIReg, MovePackedStringTo,
                      DeclareLocalVariables,
                      GenerateQuadLabel,
                      GenerateEntryProcedureLabel,
                      GenerateExitProcedureLabel,
                      WritePrologue, WriteEpilogue, FlushPipe ;
*)
FROM i486 IMPORT DWORDSIZE, WORDSIZE, BYTESIZE,
                 ModeOfMove, KindOfScope, Register, Instruction,
                 DatumPrefix, GetUnitOfTransfer,
                 GetScope, GetOffset, GetSize, GetUnitPrefix,
                 GetVarTypeSize, EqualSizes, FitsInRegister,
                 MaxDatum,
                 IsRegister, LowReg, HighReg,
                 EAX, EBX, ECX, EDX,
                 ESP, EBP, ESI, EDI ;      (* register symbols *)


CONST
   Indent              =     8 ;  (* Indent spaces per tree level        *)
   MaxCost             = 10000 ;
   MaxStack            =  5000 ;
   MaxParam            =     7 ;
   DebugInst           =  TRUE ;
   MaxNumChars         =    20 ;
   MaxString           =  1024 ;
   MaxWordSize         =    32 ;  (* max bits in a word                  *)
   MaxNoOfItems        =  4096 ;  (* max tree nodes in a basic block     *)
   MaxNoOfTrees        =   500 ;  (* max number of trees per basic block *)
   MaxNoOfRegisters    =     8 ;  (* max number of registers (why so small Intel?) *)
   InsideLeafProcedure = FALSE ;  (* later to become a variable          *)
   NulNode             =     0 ;  (* non existant node                   *)
   BoundsFudge         =     1 ;  (* set to 1 if compiling with -bounds  *)


TYPE
   InstrustionType =  (instmdreg, instregmd, instregreg, instregconst,
                       instmdconst, instbinmd, instbinreg, instbinstr,
                       instmdoffset, instregimmname) ;

   (*
      the following data structures are used to create a labeled tree.
      Choosen not to use pointers as an array is very simple to manage deallocation
      and allocation - although it will consume far more memory.
   *)

   IndexToTree = [0..MaxNoOfTrees] ;   (* debugging and clarity: make the compiler check indices and types *)
   IndexToItem = [0..MaxNoOfItems] ;   (* debugging and clarity: make the compiler check indices and types *)

   SymbolInfo = RECORD
                   Symbol              : CARDINAL ;             (* symboltable symbol                 *)
                   PrerequisiteRegister: CARDINAL ;             (* NulSym = any register              *)
                   IsRegisterUsed      : BOOLEAN ;              (* have we used this register?        *)
                END ;

   BinaryOp = RECORD
                 DestSym: SymbolInfo ;
                 Left,
                 Right  : IndexToItem ;
              END ;

   UnaryOp = RECORD
                 DestSym: SymbolInfo ;
                 Right  : IndexToItem
              END ;

   ConditionOp = RECORD
                    DestQuad: CARDINAL ;
                    Left,
                    Right   : IndexToItem ;
                 END ;

   CombineOp = RECORD
                  DestSym: SymbolInfo ;
                  Tree1,
                  Tree2  : IndexToItem ;
               END ;

   Leaf = RECORD
             DestSym   : SymbolInfo ;
             LabelValue: CARDINAL ;      (* used in the Dragon Book alg. P562 *)
          END ;

   RegisterState = (free, used, hidden, fixed) ;

   Node = RECORD
             LabelValue     : CARDINAL ;      (* used in the Dragon Book alg. P562 *)
             QuadNo         : CARDINAL ;      (* which quad is associated with it  *)
             LineNo         : CARDINAL ;      (* which line no is it associated    *)
             IsSigned       : BOOLEAN ;       (* special case operator mod, div    *)
             IsVarParam     : BOOLEAN ;       (* special case for operator Param   *)
             Op             : QuadOperator ;    (* what kind of node have we here? *)
             CASE Op OF

             DummyOp        : Dummy    : CombineOp |
             XIndrOp,
             IndrXOp,         (* indirection operators which we treat as binary  *)

             (* arithmetic operators *)
             LogicalOrOp,
             LogicalAndOp,
             LogicalXorOp,
             LogicalShiftLeftOp,
             LogicalShiftRightOp,
             DivOp,
             MultOp,
             ModOp,
             AddOp,
             MultOp,
             SubOp              : Binary   : BinaryOp |

             CallOp,
             ParamOp,
             ReturnValueOp,
             FunctValueOp,
             OffsetOp,
             AddrOp,
             BecomesOp          : Unary    : UnaryOp |

             IfInOp,
             IfNotInOp,
             IfEquOp,
             IfNotEquOp,
             IfGreEquOp,
             IfLessEquOp,
             IfLessOp,
             IfGreOp            : Condition: ConditionOp

             END
          END ;

   Item = RECORD
             IsRef : BOOLEAN ;    (* has item been referenced by another item? *)
             IsLeaf: BOOLEAN ;
             CASE IsLeaf OF
             
             FALSE: ItemIsNode: Node |
             TRUE : ItemIsLeaf: Leaf

             END
          END ;

   RegInfo = RECORD
                Reg       : CARDINAL ;          (* register symbol              *)
                State     : RegisterState ;     (* the state of the register    *)
                CleanAlias: List ;              (* symbols whose value is reg   *)
                DirtyAlias: List ;              (* symbols which are dirty.     *)
             END ;


(* %%%FORWARD%%%
PROCEDURE UnHideRepsRegisters ; FORWARD ;
PROCEDURE PaddString (a: ARRAY OF CHAR; DestSize: CARDINAL; i: CARDINAL;
                      VAR b: ARRAY OF CHAR) ; FORWARD ;
PROCEDURE AlignStackPointer ; FORWARD ;
PROCEDURE FlushAllRegisters ; FORWARD ;
PROCEDURE FlushRegister (r: CARDINAL) ; FORWARD ;
PROCEDURE MoveRegisterToTop (r: CARDINAL) ; FORWARD ;
PROCEDURE IsRegisterFree (reg: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE SecondGuessTopRegisterMove (i: IndexToItem) ; FORWARD ;
PROCEDURE IsRegisterValueKnown (reg: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE BuildBinaryNode (q: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ; FORWARD ;
PROCEDURE AlterIndex (parent, from, to: IndexToItem) ; FORWARD ;
PROCEDURE IsDependent (l, r: IndexToItem) : BOOLEAN ; FORWARD ;
PROCEDURE IsVarParamNode (i: IndexToItem) : BOOLEAN ; FORWARD ;
PROCEDURE ResetRegisterValues ; FORWARD ;
PROCEDURE SetRegisterAssociate (reg: CARDINAL; clean, dirty: List) ; FORWARD ;
PROCEDURE IsInThisRegister (reg, sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE IsInRegister (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE GetFromRegister (sym: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE IncludeValueInReg (reg: CARDINAL; sym: CARDINAL) ; FORWARD ;
PROCEDURE AssignRegWithSym (Reg: CARDINAL; Sym: CARDINAL) ; FORWARD ;
PROCEDURE PeepMoveReg2Reg (r1, r2: CARDINAL) ; FORWARD ;
PROCEDURE CopyRegisterLists (r1, r2: CARDINAL) ; FORWARD ;
PROCEDURE TrashRegisterValue (reg: CARDINAL) ; FORWARD ;
PROCEDURE TryToFindOffsetSymbol (i: IndexToItem; Start, End: CARDINAL) : INTEGER ; FORWARD ;
PROCEDURE PushUnbounded (Sym: CARDINAL) ; FORWARD ;
PROCEDURE IncreaseStack (c: CARDINAL) ; FORWARD ;
PROCEDURE IsRegisterByteWordAccessible (reg: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE RemoveAnyUnboundedData ; FORWARD ;
PROCEDURE PrepareToPushUnboundedData ; FORWARD ;
PROCEDURE PushArrayToStack (ArraySym: CARDINAL; parent, i: IndexToItem; Head, Start, End: CARDINAL) ; FORWARD ;
PROCEDURE PushStringToStack (StringSym: CARDINAL) ; FORWARD ;
PROCEDURE BuildAddr (Current: CARDINAL; op1, op3: CARDINAL) ; FORWARD ;
PROCEDURE TryToFindOffset (i: IndexToItem; Start, End: CARDINAL) : INTEGER ; FORWARD ;
PROCEDURE CheckFoldNodes (n, l, r: IndexToItem; Start, End: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE CheckFoldTree (Head, Start, End: CARDINAL; i: IndexToItem) : BOOLEAN ; FORWARD ;
PROCEDURE BlockMove (i: IndexToItem; Head, Start, End: CARDINAL) ; FORWARD ;
PROCEDURE AlterToUnaryNode (n: IndexToItem; op: QuadOperator; r: IndexToItem) ; FORWARD ;
PROCEDURE MoveStringFromToDirect (StringSym: CARDINAL; DestSym: CARDINAL; DestSize: CARDINAL) ; FORWARD ;
PROCEDURE MoveStringFromToIndirect (StringSym: CARDINAL; DestSize: CARDINAL; offset: INTEGER) ; FORWARD ;
PROCEDURE PackString (a: ARRAY OF CHAR; DestSize, i: CARDINAL;
                      VAR b: ARRAY OF CHAR) : CARDINAL ; FORWARD ;
PROCEDURE IsPowerOfTwo (sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE Log2Of (sym: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE Decrement (sym: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE FreeNonDivModRegister () : CARDINAL ; FORWARD ;
PROCEDURE PreReqReg2 (i1, i2: IndexToItem) ; FORWARD ;
PROCEDURE GetRegisterState (reg: CARDINAL) : RegisterState ; FORWARD ;
PROCEDURE MarkTopRegisterInUse ; FORWARD ;
PROCEDURE BuildBecomes (Start: CARDINAL; op: QuadOperator; op1, op3: CARDINAL) ; FORWARD ;
PROCEDURE PutRegisterUsed (i: IndexToItem; IsUsed: BOOLEAN) ; FORWARD ;
PROCEDURE GetRegisterUsed (i: IndexToItem) : BOOLEAN ; FORWARD ;
PROCEDURE GetPrequisiteRegister (i: IndexToItem) : CARDINAL ; FORWARD ;
PROCEDURE WriteRegisterName (Sym: CARDINAL) ; FORWARD ;
PROCEDURE DisplayRegisterStack ; FORWARD ;
PROCEDURE MarkRegister (reg: CARDINAL; state: RegisterState) ; FORWARD ;
PROCEDURE InstNodeReg (i: IndexToItem; op1, op2: IndexToItem; Reg: Register) ; FORWARD ;
PROCEDURE FindLatestItem (Sym: CARDINAL) : IndexToItem ; FORWARD ;
PROCEDURE PreReqReg (i: IndexToItem) ; FORWARD ;
PROCEDURE WriteSymbol (i: IndexToItem) ; FORWARD ;
PROCEDURE FetchAddress (Sym: CARDINAL) ; FORWARD ;
PROCEDURE BuildGoto (Head, Current: CARDINAL; DestQuad: CARDINAL) ; FORWARD ;
PROCEDURE FindDatum2 (op1, op2: CARDINAL) : DatumPrefix ; FORWARD ;
PROCEDURE GetNodeDatum (i: IndexToItem) : DatumPrefix ; FORWARD ;
PROCEDURE IsConditionNode (i: IndexToItem) : BOOLEAN ; FORWARD ;
PROCEDURE FitsIntoRegister (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE OpToInst (op: QuadOperator) : Instruction ; FORWARD ;
PROCEDURE BuildForest (Head, Start, End: CARDINAL) ; FORWARD ;
PROCEDURE AssignValue (i: IndexToItem; v: CARDINAL) ; FORWARD ;
PROCEDURE IsItemNode (i: IndexToItem) : BOOLEAN ; FORWARD ;
PROCEDURE GetDestination (i: IndexToItem) : CARDINAL ; FORWARD ;
PROCEDURE GetItemValue (i: IndexToItem) : CARDINAL ; FORWARD ;
PROCEDURE DisplayTree (i: IndexToItem) ; FORWARD ;
PROCEDURE DisplayLeaf (i: IndexToItem; count: CARDINAL) ; FORWARD ;
PROCEDURE DisplayNode (i: IndexToItem; count: CARDINAL) ; FORWARD ;
PROCEDURE CalculateChildren (i: IndexToItem) : CARDINAL ; FORWARD ;
PROCEDURE LabelTree (i: IndexToItem; IsLeft: BOOLEAN) : CARDINAL ; FORWARD ;
PROCEDURE GetNodeLeft (i: IndexToItem) : IndexToItem ; FORWARD ;
PROCEDURE GetNodeRight (i: IndexToItem) : IndexToItem ; FORWARD ;
PROCEDURE GetNodeOp (i: IndexToItem) : QuadOperator ; FORWARD ;
PROCEDURE IsAbleToCodeQuad (q: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE AhoGenCode (parent, i: IndexToItem; IsLeftLeaf: BOOLEAN; Head, Start, End: CARDINAL) ; FORWARD ;
PROCEDURE FlushForest (Head, Start, End: CARDINAL) ; FORWARD ;
PROCEDURE IsItemLeaf (i: IndexToItem) : BOOLEAN ; FORWARD ;
PROCEDURE IsSignedNode (i: IndexToItem) : BOOLEAN ; FORWARD ;
PROCEDURE InstNodeRegReg (i: IndexToItem; r1, r2: Register) ; FORWARD ;
   %%%FORWARD%%% *)


VAR
   ListOfTrees              : ARRAY IndexToTree OF IndexToItem ;     (* many trees exist in a basic block *)
   ListOfItems              : ARRAY IndexToItem OF Item ;
   LastNode                 : IndexToItem ;                          (* the last node created             *)
   ItemPtr                  : IndexToItem ;                          (* indicates the next available Item *)
   TreePtr                  : IndexToTree ;                          (* indicates the next available Tree *)
   LastFlushQuad            : CARDINAL ;                             (* the last quad when Flush called   *)
   TwoToThePower            : ARRAY [0..MaxWordSize] OF CARDINAL ;   (* power of two constants            *)
   RegisterStack            : ARRAY [0..MaxNoOfRegisters*2] OF CARDINAL ;  (* register list used WalkTree *)
   RegisterStackPtr         : CARDINAL ;                                 (* register stack pointer        *)
   ListOfRegisters          : ARRAY [0..MaxNoOfRegisters] OF RegInfo ;   (* contains state of registers   *)
   InitRegisterPtr          : CARDINAL ;                        (* how many registers have we initialized *)
   CompilingMainModule      : BOOLEAN ;                         (* Determines whether the main module     *)
                                                                (* quadrules are being processed.         *)
   SetFunctionReturnValue   : BOOLEAN ;                         (* have we set the return value yet?      *)
   UsingFunctionReturnValue : BOOLEAN ;                         (* do we need to obtain func return value *)
   UnboundedOnStack         : BOOLEAN ;                         (* have we placed unbounded data on stack *)
   DebugOn                  : BOOLEAN ;                         (* are we debugging this module?          *)
   ConstantSymbol           : CARDINAL ;                        (* constant symbol used in ForeachItemDo  *)
   SameConstantValue        : BOOLEAN ;                         (* return value is the same?              *)
   HideESI,
   HideEDI,                                                     (* are these registers to be hidden       *)
   HideECX                  : BOOLEAN ;                         (* across trees?                          *)
   PreviousScope            ,                                   (* current procedure and previous scope   *)
   CurrentProcedure         : CARDINAL ;                        (* used to implement nested procedures.   *)


(*
   AlterToUnaryNode -
*)

PROCEDURE AlterToUnaryNode (n: IndexToItem; op: QuadOperator; r: IndexToItem) ;
VAR
   t: Item ;
BEGIN
   WITH t DO
      IsLeaf := FALSE ;
      IsRef  := FALSE ;
      WITH ItemIsNode DO
         LabelValue      := 0 ;
         QuadNo          := ListOfItems[n].ItemIsNode.QuadNo ;
         LineNo          := ListOfItems[n].ItemIsNode.LineNo ;
         IsSigned        := ListOfItems[n].ItemIsNode.IsSigned ;
         IsVarParam      := ListOfItems[n].ItemIsNode.IsVarParam ;
         Op              := op ;
         WITH Unary DO
            DestSym := ListOfItems[n].ItemIsNode.Binary.DestSym ;
            Right   := r
         END
      END
   END ;
   ListOfItems[n] := t
END AlterToUnaryNode ;


PROCEDURE Dummy ; BEGIN END Dummy ;


(*
   IsAbleToCode - is the dynamic code generator able to encode these quadruples?
*)

PROCEDURE IsAbleToCode (Start, End: CARDINAL) : BOOLEAN ;
BEGIN
   WHILE (Start#0) AND (Start<End) DO
      IF IsAbleToCodeQuad(Start)
      THEN
         Start := GetNextQuad(Start)
      ELSE
         RETURN( FALSE )
      END
   END ;
   IF Start=0
   THEN
      RETURN( TRUE )
   ELSE
      RETURN( IsAbleToCodeQuad(Start) )
   END
END IsAbleToCode ;


(*
   IsAbleToCodeQuad - returns TRUE if we can encode this quadruple.
*)

PROCEDURE IsAbleToCodeQuad (q: CARDINAL) : BOOLEAN ;
VAR
   Op : QuadOperator ;
   Op1,
   Op2,
   Op3: CARDINAL ;
BEGIN
   GetQuad(q, Op, Op1, Op2, Op3) ;
   CASE Op OF

   NegateOp,
   IndrXOp,
   XIndrOp,
   InlineOp,
   StartModFileOp,
   StartDefFileOp,
   StartOp,
   EndFileOp,
   EndOp,
   FunctValueOp,
   ReturnValueOp,
   ParamOp,
   CallOp,
   BecomesOp,
   OffsetOp,
   AddrOp,
   DivOp,
   ModOp,
   AddOp,
   MultOp,
   SubOp,
   GotoOp,
   KillLocalVarOp,
   NewLocalVarOp,
   ReturnOp,
   BitOp,
   IfInOp,
   IfNotInOp,
   IfEquOp,
   IfNotEquOp,
   IfGreEquOp,
   IfLessEquOp,
   IfLessOp,
   IfGreOp,
   LineNumberOp  :  RETURN( TRUE )

   ELSE
      RETURN( FALSE )
   END
END IsAbleToCodeQuad ;


(*
   CheckQuadLabel - checks to see whether the current quadruple requires a label.
*)

PROCEDURE CheckQuadLabel (Head, Current: CARDINAL) ;
VAR
   Op : QuadOperator ;
   Op1,
   Op2,
   Op3: CARDINAL ;
BEGIN
   GetQuad(Current, Op, Op1, Op2, Op3) ;
   IF IsReferenced(Head, Current) AND (Op#NewLocalVarOp)
   THEN
      FlushCoprocessor ;
      GenerateQuadLabel(Current) ;
      ResetRegisterValues
   END
END CheckQuadLabel ;


(*
   CallGDB - 
*)

PROCEDURE CallGDB ;
BEGIN
   (* debugging stub *)
END CallGDB ;


(*
   GenCode - takes the list of quadruples from Head and generates
             assembly langauge between quadruples Start and End.
*)

PROCEDURE GenCode (Head, Start, End: CARDINAL) ;
BEGIN
   IF DebugOn
   THEN
      PushOutput(IO.Write)   (* debugging purposes *)
   END ;
   LastFlushQuad := Start ;
   CheckQuadLabel(Head, Start) ;
   ResetRegisterValues ;
   BuildForest(Head, Start, End) ;
   FlushForest(Head, Start, End) ;
   IF DebugOn
   THEN
      PopOutput
   END
END GenCode ;


(*
   PushRegister - pushes a register onto the register stack.
*)

PROCEDURE PushRegister (reg: CARDINAL; state: RegisterState) ;
VAR
   i: CARDINAL ;
BEGIN
   (* consistancy check *)
   i := RegisterStackPtr ;
   WHILE i>0 DO
      IF RegisterStack[i-1]=reg
      THEN
         DisplayRegisterStack ;
         InternalError('register already on the register stack', __FILE__, __LINE__)
      END ;
      DEC(i)
   END ;
   (* end of consistancy check *)
   RegisterStack[RegisterStackPtr] := reg ;
   IF RegisterStackPtr=MaxNoOfRegisters
   THEN
      InternalError('RegisterStackPtr is incorrect', __FILE__, __LINE__)
   END ;
   INC(RegisterStackPtr) ;
   MarkRegister(reg, state)
END PushRegister ;


(*
   PopRegister - pops a register from the register stack
*)

PROCEDURE PopRegister (VAR state: RegisterState) : CARDINAL ;
BEGIN
   DEC(RegisterStackPtr) ;
   IF RegisterStackPtr=0
   THEN
      InternalError('RegisterStackPtr is incorrect', __FILE__, __LINE__)
   END ;
   state := GetRegisterState(RegisterStack[RegisterStackPtr]) ;
   RETURN( RegisterStack[RegisterStackPtr] )
END PopRegister ;


(*
   RegisterTop - returns the top register.
*)

PROCEDURE RegisterTop () : CARDINAL ;
BEGIN
   RETURN( RegisterStack[RegisterStackPtr-1] )
END RegisterTop ;


(*
   IsRegisterOnRegisterStack - returns TRUE if the register, r, is free.
*)

PROCEDURE IsRegisterOnRegisterStack (reg: CARDINAL) : BOOLEAN ;
VAR
   i: CARDINAL ;
BEGIN
   i := RegisterStackPtr ;
   WHILE i>0 DO
      IF RegisterStack[i-1]=reg
      THEN
         RETURN( TRUE )
      ELSE
         DEC(i)
      END
   END ;
   RETURN( FALSE )
END IsRegisterOnRegisterStack ;


(*
   RotateRegisters - returns the top register once the registers
                     have been rotated. It will only rotate the
                     registers if all are free on the stack.
*)

PROCEDURE RotateRegisters () : CARDINAL ;
VAR
   i: CARDINAL ;
   t: CARDINAL ;
BEGIN
   i := RegisterStackPtr ;
   WHILE i>0 DO
      IF NOT IsRegisterFree(RegisterStack[i-1])
      THEN
         RETURN( RegisterTop() )
      END ;
      DEC(i)
   END ;
   (* all registers are free, now choose a register which is has a unknown value *)
   i := 0 ;
   WHILE i<RegisterStackPtr DO
      IF NOT IsRegisterValueKnown(RegisterStack[i])
      THEN
         MoveRegisterToTop(RegisterStack[i]) ;
         RETURN( RegisterTop() )
      END ;
      INC(i)
   END ;
   RETURN( RegisterTop() )
END RotateRegisters ;


(*
   MoveRegisterToTop - places register, r, on the top of the free register list.
*)

PROCEDURE MoveRegisterToTop (r: CARDINAL) ;
VAR
   i: CARDINAL ;
   t: CARDINAL ;
BEGIN
   i := RegisterStackPtr ;
   WHILE i>0 DO
      IF RegisterStack[i-1]=r
      THEN
         t := RegisterStack[i-1] ;
         RegisterStack[i-1] := RegisterStack[RegisterStackPtr-1] ;
         RegisterStack[RegisterStackPtr-1] := t ;
         RETURN
      ELSE
         DEC(i)
      END
   END ;
   IF DebugOn
   THEN
      DisplayRegisterStack
   END ;
   InternalError('register allocation algorithm failed', __FILE__, __LINE__)
END MoveRegisterToTop ;


(*
   SwapRegisters - swaps the top two registers on the register stack.
*)

PROCEDURE SwapRegisters ;
VAR
   t: CARDINAL ;
BEGIN
   t := RegisterStack[RegisterStackPtr-1] ;
   RegisterStack[RegisterStackPtr-1] := RegisterStack[RegisterStackPtr-2] ;
   RegisterStack[RegisterStackPtr-2] := t
END SwapRegisters ;


(*
   RemoveStackedRegister - 
*)

PROCEDURE RemoveStackedRegister (reg: CARDINAL) ;
VAR
   i, j: CARDINAL ;
BEGIN
   i := RegisterStackPtr ;
   WHILE i>0 DO
      IF RegisterStack[i-1]=reg
      THEN
         j := i-1 ;
         WHILE j<RegisterStackPtr DO
            RegisterStack[j] := RegisterStack[j+1] ;
            INC(j)
         END ;
         DEC(RegisterStackPtr) ;
         RETURN
      ELSE
         DEC(i)
      END
   END ;
   InternalError('register allocation algorithm failed', __FILE__, __LINE__)
END RemoveStackedRegister ;


(*
   IsTopRegisterUsed - 
*)

PROCEDURE IsTopRegisterUsed () : BOOLEAN ;
BEGIN
   RETURN( GetRegisterState(RegisterStack[RegisterStackPtr-1])=used )
END IsTopRegisterUsed ;


(*
   GetRegisterState - 
*)

PROCEDURE GetRegisterState (reg: CARDINAL) : RegisterState ;
VAR
   i: CARDINAL ;
BEGIN
   i := InitRegisterPtr ;
   WHILE i>0 DO
      IF ListOfRegisters[i-1].Reg=reg
      THEN
         RETURN( ListOfRegisters[i-1].State )
      ELSE
         DEC(i)
      END
   END ;
   InternalError('register allocation algorithm failed', __FILE__, __LINE__)
END GetRegisterState ;


(*
   IsRegisterInUse - 
*)

PROCEDURE IsRegisterInUse (reg: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN( GetRegisterState(reg)=used )
END IsRegisterInUse ;


(*
   IsRegisterHidden - 
*)

PROCEDURE IsRegisterHidden (reg: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN( GetRegisterState(reg)=hidden )
END IsRegisterHidden ;


(*
   IsRegisterFree - 
*)

PROCEDURE IsRegisterFree (reg: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN( GetRegisterState(reg)=free )
END IsRegisterFree ;


(*
   InitRegister - 
*)

PROCEDURE InitRegister (reg: CARDINAL; state: RegisterState) ;
BEGIN
   WITH ListOfRegisters[InitRegisterPtr] DO
      Reg   := reg ;
      State := state ;
      IF CleanAlias#NIL
      THEN
         KillList(CleanAlias)
      END ;
      CleanAlias := NIL ;
      IF DirtyAlias#NIL
      THEN
         KillList(DirtyAlias)
      END ;
      DirtyAlias := NIL
   END ;
   INC(InitRegisterPtr)
END InitRegister ;


(*
   GetRegisterAssociate - 
*)

PROCEDURE GetRegisterAssociate (reg: CARDINAL; VAR clean, dirty: List) ;
VAR
   i: CARDINAL ;
BEGIN
   i := InitRegisterPtr ;
   WHILE i>0 DO
      IF ListOfRegisters[i-1].Reg=reg
      THEN
         clean := DuplicateList(ListOfRegisters[i-1].CleanAlias) ;
         dirty := DuplicateList(ListOfRegisters[i-1].DirtyAlias) ;
         RETURN
      ELSE
         DEC(i)
      END
   END ;
   InternalError('cannot find register', __FILE__, __LINE__)
END GetRegisterAssociate ;


(*
   WriteRegisterState - 
*)

PROCEDURE WriteRegisterState (state: RegisterState) ;
BEGIN
   CASE state OF

   free  :  WriteString('free') |
   used  :  WriteString('used') |
   hidden:  WriteString('hidden') |
   fixed :  WriteString('fixed')

   ELSE
      InternalError('incomplete case clause', __FILE__, __LINE__)
   END
END WriteRegisterState ;


(*
   CheckReadOnlyRegister - checks to see whether the top register is read only
                           and if so it copies it to a writeable register.
*)


PROCEDURE CheckReadOnlyRegister ;
VAR
   state: RegisterState ;
BEGIN
   IF RegisterTop()=EBP
   THEN
      Assert(PopRegister(state)=EBP) ;
      InstRegReg(mov, MaxDatum(), LowReg(RegisterTop()), RegEBP) ;
      MarkTopRegisterInUse
   END
END CheckReadOnlyRegister ;


(*
   MoveByteWordRegisterToTop - returns TRUE if, reg, can be moved to the
                               top of the free register stack. We will,
                               IgnoreValue, of the register.
*)

PROCEDURE MoveByteWordRegisterToTop (reg: CARDINAL;
                                     IgnoreValue: BOOLEAN) : BOOLEAN ;
BEGIN
   IF IgnoreValue
   THEN
      IF IsRegisterFree(reg) AND IsRegisterOnRegisterStack(reg)
      THEN
         MoveRegisterToTop(reg) ;
         RETURN( TRUE )
      END
   ELSE
      IF IsRegisterFree(reg) AND IsRegisterOnRegisterStack(reg) AND
         (NOT IsRegisterValueKnown(reg))
      THEN
         MoveRegisterToTop(reg) ;
         RETURN( TRUE )
      END
   END ;
   RETURN( FALSE )
END MoveByteWordRegisterToTop ;


(*
   CheckUnitAccessRegister - checks to see that a register on the free list if capable of
                             byte and word access.
*)

PROCEDURE CheckUnitAccessRegister (size: CARDINAL) ;
BEGIN
   IF ((size=BYTESIZE) OR (size=WORDSIZE)) AND (NOT IsRegisterByteWordAccessible(RegisterTop()))
   THEN
      (* problem *)
      IF MoveByteWordRegisterToTop(EAX, FALSE) OR
         MoveByteWordRegisterToTop(EBX, FALSE) OR
         MoveByteWordRegisterToTop(ECX, FALSE) OR
         MoveByteWordRegisterToTop(EDX, FALSE) OR
         MoveByteWordRegisterToTop(EAX,  TRUE) OR
         MoveByteWordRegisterToTop(EBX,  TRUE) OR
         MoveByteWordRegisterToTop(ECX,  TRUE) OR
         MoveByteWordRegisterToTop(EDX,  TRUE)
      THEN
         (* whew! *)
      ELSE
         InternalError('register allocation failed ... bug see gaius ... no 8/16 bit registers left', __FILE__, __LINE__)
      END
   END
END CheckUnitAccessRegister ;


(*
   FlushInCaseOfMultiRegOps - flushes registers used as destination operands in multireg operators.
*)

PROCEDURE FlushInCaseOfMultiRegOps (inst: Instruction) ;
BEGIN
   CASE inst OF

   div,
   idiv   :   FlushRegister(EAX) ;
              FlushRegister(EDX) |
   repmovs:   FlushRegister(EDI) ;
              FlushRegister(ESI) ;
              FlushRegister(ECX)

   ELSE
   END
END FlushInCaseOfMultiRegOps ;


(*
   CheckForMultiRegOps - checks for multiple register operators
*)

PROCEDURE CheckForMultiRegOps (inst: Instruction) ;
BEGIN
   CASE inst OF

   div,
   idiv   :   TrashRegisterValue(EAX) ;
              TrashRegisterValue(EDX) |
   repmovs:   TrashRegisterValue(EDI) ;
              TrashRegisterValue(ESI) ;
              TrashRegisterValue(ECX) ;
              IncludeValueInReg(ECX, MakeConstLit(MakeKey('0')))

   ELSE
   END
END CheckForMultiRegOps ;


(*
   InstConstNodeReg - produces an instruction   op(i)   #op1, (op2=Reg)
*)

PROCEDURE InstConstNodeReg (i: IndexToItem; op1, op2: IndexToItem; Reg: Register) ;
VAR
   reg,
   Sym : CARDINAL ;
   Inst: Instruction ;
BEGIN
   Sym := GetDestination(op1) ;
   reg := HighReg(Reg) ;
   IF IsItemLeaf(i)
   THEN
      Inst := mov
   ELSE
      Inst := OpToInst(GetNodeOp(i))
   END ;
   FlushInCaseOfMultiRegOps(Inst) ;
   IF IsConstString(Sym) AND (GetStringLength(Sym)#1)
   THEN
      InternalError('cannot place a string with more than 1 character in it into a register', __FILE__, __LINE__)
   ELSIF IsConst(Sym)
   THEN
      IF IsConditionNode(i)
      THEN
         FlushAllRegisters ;   (* we compare and then jump *)
         (* must not operate on full register size as higher components maybe garbage *)
         IF IsZero(Sym)
         THEN
            (* test reg is faster than cmp reg, $0 *)
            InstRegReg(test, GetUnitPrefix(GetSize(GetDestination(op2))), Reg, Reg)
         ELSE
            IF IsInRegister(Sym)
            THEN
               IF DebugOn
               THEN
                  Comment('peep: no need to emit load constant, use register ')
               END ;
               InstRegReg(Inst, GetUnitPrefix(GetSize(GetDestination(op2))), Reg, LowReg(GetFromRegister(Sym)))
            ELSE
               InstRegConstant(Inst, GetUnitPrefix(GetSize(GetDestination(op2))), Reg, Sym)
            END ;
            IF (GetNodeOp(i)=IfInOp) OR (GetNodeOp(i)=IfNotInOp)
            THEN
               InstRegReg(test, GetUnitPrefix(GetSize(GetDestination(op2))), Reg, Reg)
            END
         END
      ELSE
         IF (Inst=mov) AND IsInRegister(Sym)
         THEN
            IF IsInThisRegister(HighReg(Reg), Sym)
            THEN
               IF DebugOn
               THEN
                  WriteString('# peep: found symbol already inside register ') ; WriteRegisterName(HighReg(Reg)) ; WriteLn
               END
            ELSE
               FlushRegister(reg) ;
               InstRegReg(mov, MaxDatum(), Reg, LowReg(GetFromRegister(Sym))) ;
               AssignRegWithSym(reg, Sym)
            END
         ELSIF (Inst=mov) AND IsZero(Sym)
         THEN
            FlushRegister(reg) ;
            InstRegReg(xor, MaxDatum(), Reg, Reg) ;
            AssignRegWithSym(reg, Sym)
         ELSIF Inst=div
         THEN
            BinInstConstant(div, MaxDatum(), Sym)
         ELSIF IsOne(Sym)
         THEN
            IF Inst=add
            THEN
               FlushRegister(reg) ;
               BinInstReg(inc, MaxDatum(), Reg) ;
               TrashRegisterValue(reg)
            ELSIF Inst=sub
            THEN
               FlushRegister(reg) ;
               BinInstReg(dec, MaxDatum(), Reg) ;
               TrashRegisterValue(reg)
            ELSE
               (* remember we know that Sym is NOT inside reg *)
               FlushRegister(reg) ;  (* flush because a different value is being placed inside reg *)
               InstRegConstant(Inst, MaxDatum(), Reg, Sym) ;
               IF Inst=mov
               THEN
                  AssignRegWithSym(reg, Sym)
               ELSE
                  TrashRegisterValue(reg)
               END
            END
         ELSE
            (* remember we know that Sym is NOT inside reg *)
            FlushRegister(reg) ;  (* flush because a different value is being placed inside reg *)
            InstRegConstant(Inst, MaxDatum(), Reg, Sym) ;
            IF Inst=mov
            THEN
               AssignRegWithSym(reg, Sym)
            ELSE
               TrashRegisterValue(reg)
            END
         END
      END
   END ;
   CheckForMultiRegOps(Inst)
END InstConstNodeReg ;


(*
   InstNodeReg - produces an instruction which operates on  op(i)   op1, (op2=Reg).
*)

PROCEDURE InstNodeReg (i: IndexToItem; op1, op2: IndexToItem; Reg: Register) ;
VAR
   Inst: Instruction ;
   Size,
   reg,
   Sym : CARDINAL ;
BEGIN
   Sym := GetDestination(op1) ;
   reg := HighReg(Reg) ;
   IF IsConst(Sym)
   THEN
      InstConstNodeReg(i, op1, op2, Reg)
   ELSE
      IF IsItemLeaf(i)
      THEN
         Inst := mov
      ELSE
         Inst := OpToInst(GetNodeOp(i))
      END ;
      IF IsRegister(Sym)
      THEN
         IF (Inst=mov) AND (NOT IsTopRegisterUsed()) AND IsRegisterOnRegisterStack(Sym)
         THEN
            IF DebugOn
            THEN
               DisplayRegisterStack ;
               WriteString('# no need to copy from ') ; WriteSymbol(op1) ;
               WriteString(' to ') ; WriteSymbol(op2) ; WriteString(' via register ') ;
               WriteRegisterName(reg) ; WriteString(' we move ') ;
               WriteRegisterName(Sym) ; WriteString(' to top of stack') ; WriteLn ;
               DisplayRegisterStack ;
            END ;
            Assert(IsRegisterOnRegisterStack(reg)) ;
            MoveRegisterToTop(Sym) ;
            reg := Sym
         ELSE
            InstNodeRegReg(i, LowReg(Sym), Reg)
         END
      ELSIF IsProcedure(Sym)
      THEN
         InstRegImmName(mov, MaxDatum(), Reg, Sym)
      ELSIF IsInRegister(Sym)
      THEN
         IF (Inst=mov) AND IsInThisRegister(HighReg(Reg), Sym)
         THEN
            IF DebugOn
            THEN
               WriteString('# no need to copy from ') ; WriteKey(GetSymName(Sym)) ;
               WriteString(' to ') ; WriteRegisterName(HighReg(Reg)) ; WriteLn
            END
         ELSE
            InstNodeRegReg(i, LowReg(GetFromRegister(Sym)), Reg)
         END
      ELSE
         Size := GetSize(Sym) ;
         Assert(Size<=WordSize) ;
         IF IsItemNode(i) AND ((GetNodeOp(i)=DivOp) OR (GetNodeOp(i)=ModOp))
         THEN
            IF IsSignedNode(i)
            THEN
               FlushInCaseOfMultiRegOps(idiv) ;
               BinInst(cdq) ;
               BinInstMD(idiv, GetUnitPrefix(GetUnitOfTransfer(Size)), GetOffset(Sym), 0, GetScope(Sym), Sym) ;
               CheckForMultiRegOps(idiv)
            ELSE
               FlushInCaseOfMultiRegOps(div) ;
               InstRegReg(xor, MaxDatum(), RegEDX, RegEDX) ;   (* edx := 0 *)
               BinInstMD(div, GetUnitPrefix(GetUnitOfTransfer(Size)), GetOffset(Sym), 0, GetScope(Sym), Sym) ;
               CheckForMultiRegOps(div)
            END ;
            Assert(Reg=RegEAX)
         ELSE
            IF Inst=mov
            THEN
               IF IsInThisRegister(reg, Sym)
               THEN
                  IF DebugOn
                  THEN
                     WriteString('# peep: no need to emit: ') ;
                     InstRegMD(Inst, GetUnitPrefix(GetUnitOfTransfer(Size)), Reg, GetOffset(Sym), 0, GetScope(Sym), Sym)
                  END
               ELSIF IsInRegister(Sym)
               THEN
                  PeepMoveReg2Reg(GetFromRegister(Sym), reg)
               ELSE
                  FlushRegister(reg) ;  (* flush because a different value is being placed inside reg *)
                  InstRegMD(Inst, GetUnitPrefix(GetUnitOfTransfer(Size)), Reg, GetOffset(Sym), 0, GetScope(Sym), Sym) ;
                  AssignRegWithSym(reg, Sym)
               END
            ELSE
               IF IsConditionNode(i)
               THEN
                  FlushAllRegisters    (* we compare and then jump *)
               END ;
               FlushRegister(reg) ;  (* flush because a different value is being placed inside reg *)
               InstRegMD(Inst, GetUnitPrefix(GetUnitOfTransfer(Size)), Reg, GetOffset(Sym), 0, GetScope(Sym), Sym) ;
               TrashRegisterValue(reg) ;
               IF IsItemNode(i) AND ((GetNodeOp(i)=IfInOp) OR (GetNodeOp(i)=IfNotInOp))
               THEN
                  InstRegReg(test, GetUnitPrefix(GetUnitOfTransfer(Size)), Reg, Reg)
               END
            END
         END
      END
   END ;
   MarkRegister(reg, used)
(*   MarkTopRegisterInUse *)
END InstNodeReg ;


(*
   InstNodeRegReg - generate an operation of the form   op   r1, r2
*)

PROCEDURE InstNodeRegReg (i: IndexToItem; r1, r2: Register) ;
VAR
   Inst: Instruction ;
BEGIN
   IF IsItemLeaf(i)
   THEN
      Inst := mov
   ELSE
      Inst := OpToInst(GetNodeOp(i))
   END ;
   IF IsConditionNode(i)
   THEN
      FlushAllRegisters ;   (* we compare and then jump *)
   END ;
   IF IsItemNode(i) AND ((GetNodeOp(i)=DivOp) OR (GetNodeOp(i)=ModOp))
   THEN
      IF IsSignedNode(i)
      THEN
         FlushInCaseOfMultiRegOps(idiv) ;
         BinInst(cdq) ;
         BinInstReg(idiv, GetNodeDatum(i), r1) ;
         CheckForMultiRegOps(idiv)
      ELSE
         FlushInCaseOfMultiRegOps(div) ;
         InstRegReg(xor, MaxDatum(), RegEDX, RegEDX) ;   (* edx := 0 *)
         BinInstReg(div, GetNodeDatum(i), r1) ;
         CheckForMultiRegOps(div)
      END ;
      Assert(r2=RegEAX)
   ELSE
      IF Inst=mov
      THEN
         IF r1=r2
         THEN
            IF DebugOn
            THEN
               WriteString('# no need to emit ') ;
               InstRegReg(Inst, MaxDatum(), r2, r1)
            END
         ELSE
            PeepMoveReg2Reg(HighReg(r1), HighReg(r2))
         END
      ELSE
         FlushRegister(HighReg(r2)) ;  (* flush because a different value is being placed inside r2 *)
         InstRegReg(Inst, GetNodeDatum(i), r2, r1) ;
         TrashRegisterValue(HighReg(r2)) ;
         IF IsItemNode(i) AND ((GetNodeOp(i)=IfInOp) OR (GetNodeOp(i)=IfNotInOp))
         THEN
            InstRegReg(test, GetNodeDatum(i), r2, r2)
         END
      END
   END
END InstNodeRegReg ;


(*
   InstRegSym - places register contents into a symbol Sym.
*)

PROCEDURE InstRegSym (i: Instruction; Reg: Register; Sym: CARDINAL) ;
VAR
   Size: CARDINAL ;
BEGIN
   IF IsConst(Sym)
   THEN
      InternalError('cannot use a constant as a destination', __FILE__, __LINE__)
   ELSE
      Size := GetSize(Sym) ;
      Assert(Size<=WordSize) ;
      (* Correct since we ignore LeftValue mode and treat as RightValue *)
      InstMDReg(i, GetUnitPrefix(GetUnitOfTransfer(Size)), GetOffset(Sym), 0, Reg, GetScope(Sym), Sym) ;
      IF i=mov
      THEN
         AssignRegWithSym(HighReg(Reg), Sym)
      ELSE
         TrashRegisterValue(HighReg(Reg)) ;   (* <---- ? *)
         InternalError('did we really need the instruction above?', __FILE__, __LINE__)
      END
   END
END InstRegSym ;


(*
   OpToInst - converts a quadruple operator into a machine dependant instruction.
*)

PROCEDURE OpToInst (op: QuadOperator) : Instruction ;
BEGIN
   CASE op OF

   NegateOp           :  RETURN( neg ) |
   CallOp             :  RETURN( call ) |
   ParamOp            :  RETURN( push ) |
   AddOp              :  RETURN( add ) |
   SubOp              :  RETURN( sub ) |
   LogicalAndOp       :  RETURN( and ) |
   LogicalOrOp        :  RETURN( or ) |
   LogicalXorOp       :  RETURN( xor ) |
   LogicalShiftLeftOp :  RETURN( shl ) |
   LogicalShiftRightOp:  RETURN( shr ) |
   MultOp             :  RETURN( mult ) |
   ModOp,
   DivOp              :  RETURN( div ) |
   BecomesOp          :  RETURN( mov ) |
   ReturnValueOp      :  RETURN( mov ) |
   FunctValueOp       :  RETURN( mov ) |
   AddrOp             :  RETURN( mov ) |
   OffsetOp           :  RETURN( mov ) |
   GotoOp             :  RETURN( jmp ) |
   IfInOp,
   IfNotInOp          :  RETURN( and ) |
   IfEquOp,
   IfNotEquOp,
   IfGreEquOp,
   IfLessEquOp,
   IfLessOp,
   IfGreOp            :  RETURN( cmp )

   ELSE
      InternalError('quadruple not yet completed', __FILE__,  __LINE__)
   END
END OpToInst ;


(*
   IsConditionNode - returns TRUE if, i, is a condition node.
*)

PROCEDURE IsConditionNode (i: IndexToItem) : BOOLEAN ;
BEGIN
   IF IsItemNode(i)
   THEN
      CASE ListOfItems[i].ItemIsNode.Op OF

      IfInOp,
      IfNotInOp,
      IfEquOp,
      IfNotEquOp,
      IfGreEquOp,
      IfLessEquOp,
      IfLessOp,
      IfGreOp    : RETURN( TRUE )

      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END IsConditionNode ;


(*
   IsUnaryNode - returns TRUE if, i, is a unary node.
*)

PROCEDURE IsUnaryNode (i: IndexToItem) : BOOLEAN ;
BEGIN
   IF IsItemNode(i)
   THEN
      CASE ListOfItems[i].ItemIsNode.Op OF

      BitOp,
      ConvertOp,
      CallOp,
      ParamOp,
      OffsetOp,
      BecomesOp,
      ReturnValueOp,
      FunctValueOp,
      AddrOp,
      NegateOp     : RETURN( TRUE )

      ELSE
         RETURN( FALSE )    (* keeps p2c happy *)
      END
   END ;
   RETURN( FALSE )
END IsUnaryNode ;


(*
   IsBinaryNode - returns TRUE if, i, is a binary node.
*)

PROCEDURE IsBinaryNode (i: IndexToItem) : BOOLEAN ;
BEGIN
   IF IsItemNode(i)
   THEN
      CASE ListOfItems[i].ItemIsNode.Op OF

      UnboundedOp,
      XIndrOp,
      IndrXOp,
      LogicalAndOp,
      LogicalOrOp,
      LogicalXorOp,
      LogicalShiftLeftOp,
      LogicalShiftRightOp,
      AddOp,
      MultOp,
      DivOp,
      ModOp,
      SubOp              : RETURN( TRUE )

      ELSE
         RETURN( FALSE )    (* keeps p2c happy *)
      END
   END ;
   RETURN( FALSE )
END IsBinaryNode ;


(*
   GetNodeQuad - returns the quadruple that produced the node, i.
*)

PROCEDURE GetNodeQuad (i: IndexToItem) : CARDINAL ;
BEGIN
   Assert(NOT ListOfItems[i].IsLeaf) ;
   RETURN( ListOfItems[i].ItemIsNode.QuadNo )
END GetNodeQuad ;


(*
   IsSymNeeded - returns TRUE if the symbol, Sym, is required in future quadruples.
                 If the Sym is not needed in the future then we can optimize the
                 register flush.
*)

PROCEDURE IsSymNeeded (Sym: CARDINAL; Start, End: CARDINAL) : BOOLEAN ;
VAR
   ReadStart, ReadEnd: CARDINAL ;
BEGIN
   IF IsTemporary(Sym)
   THEN
      GetVarReadQuads(Sym, ReadStart, ReadEnd) ;
      IF (ReadEnd>End) OR (ReadStart<Start)
      THEN
         (* read outside of our basic block, we must flush symbol *)
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   END ;
   RETURN( IsVar(Sym) )
END IsSymNeeded ;


(*
   IsUsedLaterOn - is the symbol, sym, used later on from node, i.
*)

PROCEDURE IsUsedLaterOn (sym: CARDINAL; i: IndexToItem) : BOOLEAN ;
VAR
   j: IndexToItem ;
BEGIN
   j := ItemPtr-1 ;   (* ItemPtr points to the next free (uninitialized node) *)
   WHILE j>0 DO
      IF (GetDestination(j)=sym) AND IsItemLeaf(j)
      THEN
         RETURN( j>i )
      END ;
      DEC(j)
   END ;
   RETURN( FALSE )
END IsUsedLaterOn ;


(*
   WriteSymbolValue - writes register, r, to sym.
*)

PROCEDURE WriteSymbolValue (Sym: CARDINAL; r: CARDINAL) ;
VAR
   Size: CARDINAL ;
BEGIN
   IF DebugOn
   THEN
      WriteString('# flushing dirty symbol ') ; WriteRegisterName(r) ;
      WriteString(' -> ') ; WriteKey(GetSymName(Sym)) ; WriteLn
   END ;
   IF IsConst(Sym)
   THEN
      InternalError('cannot use a constant as a destination', __FILE__, __LINE__)
   ELSE
      Size := GetSize(Sym) ;
      Assert(Size<=WordSize) ;
      (* Correct since we ignore LeftValue mode and treat as RightValue *)
      InstMDReg(mov, GetUnitPrefix(GetUnitOfTransfer(Size)), GetOffset(Sym), 0, LowReg(r), GetScope(Sym), Sym)
   END
END WriteSymbolValue ;


(*
   FlushRegister - flushes all dirty symbols belonging to register, reg.
                   Normally called when reg is about to be written by an operator.
*)

PROCEDURE FlushRegister (r: CARDINAL) ;
VAR
   i, Sym: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<InitRegisterPtr DO
      WITH ListOfRegisters[i] DO
         IF Reg=r
         THEN
            IF DirtyAlias#NIL
            THEN
               WHILE NoOfItemsInList(DirtyAlias)>0 DO
                  Sym := GetItemFromList(DirtyAlias, 1) ;
                  WriteSymbolValue(Sym, ListOfRegisters[i].Reg) ;
                  RemoveItemFromList(DirtyAlias, Sym)
               END ;
               KillList(DirtyAlias) ;
               DirtyAlias := NIL
            END ;
            RETURN
         END
      END ;
      INC(i)
   END
END FlushRegister ;


(*
   FlushAllRegisters - flushes all registers dirty symbols.
*)

PROCEDURE FlushAllRegisters ;
VAR
   Sym: CARDINAL ;
   i  : CARDINAL ;
BEGIN
   IF DebugOn
   THEN
      DisplayRegisterStack
   END ;
   i := 0 ;
   WHILE i<InitRegisterPtr DO
      WITH ListOfRegisters[i] DO
         IF DirtyAlias#NIL
         THEN
            WHILE NoOfItemsInList(DirtyAlias)>0 DO
               Sym := GetItemFromList(DirtyAlias, 1) ;
               WriteSymbolValue(Sym, ListOfRegisters[i].Reg) ;
               RemoveItemFromList(DirtyAlias, Sym)
            END ;
            KillList(DirtyAlias) ;
            DirtyAlias := NIL
         END
      END ;
      INC(i)
   END
END FlushAllRegisters ;


(*
   TrashSymbolValue - removes all aliases to symbol, sym.
*)

PROCEDURE TrashSymbolValue (Sym: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<InitRegisterPtr DO
      WITH ListOfRegisters[i] DO
         IF CleanAlias#NIL
         THEN
            RemoveItemFromList(CleanAlias, Sym)
         END
      END ;
      INC(i)
   END
END TrashSymbolValue ;


(*
   TrashAllRegisterValues - 
*)

PROCEDURE TrashAllRegisterValues ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<InitRegisterPtr DO
      WITH ListOfRegisters[i] DO
         IF CleanAlias#NIL
         THEN
            KillList(CleanAlias)
         END ;
         InitList(CleanAlias)
      END ;
      INC(i)
   END
END TrashAllRegisterValues ;


(*
   TrashRegisterValue - removes all aliases value for register, reg.
                        The register, reg, has been trashed by an operator, thus
                        all previous aliases for this register are null and void.
*)

PROCEDURE TrashRegisterValue (reg: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<InitRegisterPtr DO
      WITH ListOfRegisters[i] DO
         IF Reg=reg
         THEN
            IF CleanAlias#NIL
            THEN
               KillList(CleanAlias)
            END ;
            InitList(CleanAlias) ;
            IF DirtyAlias#NIL
            THEN
               IF NoOfItemsInList(DirtyAlias)=0
               THEN
                  KillList(DirtyAlias)
               ELSE
                  Comment('hmm dirty symbols have not been flushed - check register list below') ;
                  DisplayRegisterStack ;
                  InternalError('dirty symbols have not been flushed', __FILE__, __LINE__)
               END
            END ;
            InitList(DirtyAlias) ;
            RETURN
         END
      END ;
      INC(i)
   END
END TrashRegisterValue ;


(*
   CopyRegisterLists - copy register alias lists from r1 to r2.
*)

PROCEDURE CopyRegisterLists (r1, r2: CARDINAL) ;
VAR
   i, j: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<InitRegisterPtr DO
      WITH ListOfRegisters[i] DO
         IF Reg=r2
         THEN
            IF CleanAlias#NIL
            THEN
               KillList(CleanAlias)
            END ;
            CleanAlias := NIL ;
            IF DirtyAlias#NIL
            THEN
               IF NoOfItemsInList(DirtyAlias)=0
               THEN
                  KillList(DirtyAlias)
               ELSE
                  Comment('expecting dirty symbols to have been flushed') ;
                  DisplayRegisterStack ;
                  InternalError('expecting dirty symbols to have been flushed', __FILE__, __LINE__)
               END
            END ;
            DirtyAlias := NIL ;
            j := 0 ;
            WHILE j<InitRegisterPtr DO
               IF ListOfRegisters[j].Reg=r1
               THEN
                  IF ListOfRegisters[j].CleanAlias=NIL
                  THEN
                     InitList(ListOfRegisters[j].CleanAlias)
                  END ;
                  CleanAlias := DuplicateList(ListOfRegisters[j].CleanAlias) ;
                  (*
                   * we do NOT copy dirty lists as this causes inefficiency, as more registers are
                   * likely to be needed to be flushed.
                   *)
                  RETURN   (* all done now *)
               END ;
               INC(j)
            END
         END
      END ;
      INC(i)
   END
END CopyRegisterLists ;


(*
   PeepMoveReg2Reg - emits an instruction mov r1, r2 providing that the value of r1 is not
                     already inside r2.
*)

PROCEDURE PeepMoveReg2Reg (r1, r2: CARDINAL) ;
BEGIN
   IF r1#r2
   THEN
      IF (GetFromRegister(r2)=NulSym) OR (GetFromRegister(r1)=NulSym)
      THEN
         FlushRegister(r2) ;
         InstRegReg(mov, MaxDatum(), LowReg(r2), LowReg(r1)) ;
         CopyRegisterLists(r1, r2)
      ELSE
         IF IsSame(GetFromRegister(r2), GetFromRegister(r1))
         THEN
            IF DebugOn
            THEN
               WriteString('# peep: no need to emit: ') ; InstRegReg(mov, MaxDatum(), LowReg(r2), LowReg(r1))
            END
         ELSE
            FlushRegister(r2) ;
            InstRegReg(mov, MaxDatum(), LowReg(r2), LowReg(r1)) ;
            CopyRegisterLists(r1, r2)
         END
      END
   END
END PeepMoveReg2Reg ;


(*
   PeepPushReg - pushes a register to the stack, it returns the old register associate.
*)

PROCEDURE PeepPushReg (r: CARDINAL; VAR clean, dirty: List) ;
BEGIN
   BinInstReg(push, MaxDatum(), LowReg(r)) ;
   GetRegisterAssociate(r, clean, dirty)
END PeepPushReg ;


(*
   PeepPopReg - pops a register from the stack, it assigns the new register alias.
*)

PROCEDURE PeepPopReg (r: CARDINAL; clean, dirty: List) ;
BEGIN
   BinInstReg(pop, MaxDatum(), LowReg(r)) ;
   SetRegisterAssociate(r, clean, dirty)
END PeepPopReg ;


(*
   AssignRegWithSym - assigns a symbol to a register, it trashes the previous values within the
                      register, reg.
*)

PROCEDURE AssignRegWithSym (Reg: CARDINAL; Sym: CARDINAL) ;
BEGIN
   IF NOT IsInThisRegister(Reg, Sym)
   THEN
      (* assignment destroys previous alias list *)
      TrashRegisterValue(Reg) ;
      IncludeValueInReg(Reg, Sym)
   END
END AssignRegWithSym ;


(*
   IncludeValueInReg - includes symbol, sym, in the alias list of register, reg.
*)

PROCEDURE IncludeValueInReg (reg: CARDINAL; sym: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<InitRegisterPtr DO
      WITH ListOfRegisters[i] DO
         IF Reg=reg
         THEN
            IF CleanAlias=NIL
            THEN
               InitList(CleanAlias)
            END ;
            IncludeItemIntoList(CleanAlias, sym) ;
            RETURN
         END
      END ;
      INC(i)
   END ;
   InternalError('register not found', __FILE__, __LINE__)
END IncludeValueInReg ;


(*
   IncludeDirtyValueInReg - includes dirty symbol, sym, in the alias list of register, reg.
*)

PROCEDURE IncludeDirtyValueInReg (reg: CARDINAL; sym: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<InitRegisterPtr DO
      WITH ListOfRegisters[i] DO
         IF Reg=reg
         THEN
            IF DirtyAlias=NIL
            THEN
               InitList(DirtyAlias)
            END ;
            IncludeItemIntoList(DirtyAlias, sym) ;
            RETURN
         END
      END ;
      INC(i)
   END ;
   InternalError('register not found', __FILE__, __LINE__)
END IncludeDirtyValueInReg ;


(*
   PeepMoveConst2Reg - moves a constant symbol, sym, to register, r.
*)

PROCEDURE PeepMoveConst2Reg (sym: CARDINAL; r: CARDINAL) ;
BEGIN
   IF GetFromRegister(sym)=NulSym
   THEN
      IF IsZero(sym)
      THEN
         InstRegReg(xor, MaxDatum(), LowReg(r), LowReg(r))
      ELSE
         InstRegConstant(mov, MaxDatum(), LowReg(r), sym)
      END ;
      IncludeValueInReg(r, sym)
   ELSE
      IF DebugOn
      THEN
         Comment('peep: found constant inside register')
      END ;
      InstRegReg(mov, MaxDatum(), LowReg(r), LowReg(GetFromRegister(sym)))
   END
END PeepMoveConst2Reg ;


(*
   CheckSameConstantValue - assigns SameConstantValue to TRUE if the symbol, Sym, contains
                            the same value as ConstantSymbol.
*)

PROCEDURE CheckSameConstantValue (Sym: CARDINAL) ;
BEGIN
   IF IsSame(ConstantSymbol, Sym)
   THEN
      SameConstantValue := TRUE
   END
END CheckSameConstantValue ;


(*
   IsItemInRegisterList - returns TRUE if symbol, sym, is contained within list, l.
                          NOTE that we also check for same constant values.
*)

PROCEDURE IsItemInRegisterList (l: List; sym: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsItemInList(l, sym)
   THEN
      RETURN( TRUE )    (* quick decision *)
   ELSIF IsConst(sym)
   THEN
      (* must now check for constants with same value *)
      ConstantSymbol := sym ;
      SameConstantValue := FALSE ;
      ForeachItemInListDo(l, CheckSameConstantValue) ;
      RETURN( SameConstantValue )
   ELSE
      RETURN( FALSE )  (* no good *)
   END
END IsItemInRegisterList ;


(*
   GetFromRegister - attempts to return a register which contains the symbol, sym.
*)

PROCEDURE GetFromRegister (sym: CARDINAL) : CARDINAL ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<InitRegisterPtr DO
      WITH ListOfRegisters[i] DO
         IF sym=Reg
         THEN
            RETURN( Reg )   (* sym is a register *)
         ELSIF CleanAlias=NIL
         THEN
            InitList(CleanAlias)
         ELSIF IsItemInRegisterList(CleanAlias, sym)
         THEN
            RETURN( Reg )
         END
      END ;
      INC(i)
   END ;
   RETURN( NulSym)
END GetFromRegister ;


(*
   IsInRegister - returns TRUE if symbol, Sym, is currently held inside a register.
*)

PROCEDURE IsInRegister (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN( GetFromRegister(Sym)#NulSym )
END IsInRegister ;


(*
   IsInThisRegister - returns TRUE if symbol, sym, is held inside register, reg.
*)

PROCEDURE IsInThisRegister (reg, sym: CARDINAL) : BOOLEAN ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<InitRegisterPtr DO
      WITH ListOfRegisters[i] DO
         IF CleanAlias=NIL
         THEN
            InitList(CleanAlias)
         ELSIF Reg=reg
         THEN
            RETURN( IsItemInRegisterList(CleanAlias, sym) )
         END
      END ;
      INC(i)
   END
END IsInThisRegister ;


(*
   SetRegisterAssociate - 
*)

PROCEDURE SetRegisterAssociate (reg: CARDINAL; clean, dirty: List) ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<InitRegisterPtr DO
      WITH ListOfRegisters[i] DO
         IF Reg=reg
         THEN
            IF CleanAlias#NIL
            THEN
               KillList(CleanAlias)
            END ;
            CleanAlias := clean ;
            IF DirtyAlias#NIL
            THEN
               KillList(DirtyAlias)
            END ;
            DirtyAlias := dirty ;
            RETURN ;
         END
      END ;
      INC(i)
   END ;
   InternalError('cannot find register', __FILE__, __LINE__)
END SetRegisterAssociate ;


(*
   IsRegisterUsedByParent - 
*)

PROCEDURE IsRegisterUsedByParent (r: CARDINAL; parent: IndexToItem) : BOOLEAN ;
BEGIN
   RETURN(
           ((GetNodeOp(parent)=DivOp) OR (GetNodeOp(parent)=ModOp)) AND
           ((r=EAX) OR (r=EDX))
         )
END IsRegisterUsedByParent ;


(*
   IsLastTimeWeWrite - returns TRUE if this is the last time that a write to this symbol is
                       performed, within the quadruple range: Start..End.
                       It ignores a:

                          ParamOp quadruple which writes to a symbol (VAR parameter)
                                  as we never use *this variables* value to compute
                                  *this variables* address.
                          AddrOp  quadruple whose op3=sym

                       ie we err on the side of caution.
*)

PROCEDURE IsLastTimeWeWrite (sym: CARDINAL; Start, End: CARDINAL) : BOOLEAN ;
VAR
   WriteStart,
   WriteEnd     : CARDINAL ;
   op           : QuadOperator ;
   op1, op2, op3: CARDINAL ;
BEGIN
   op := DummyOp ;
   REPEAT
      GetVarWriteLimitQuads(sym, Start, End, WriteStart, WriteEnd) ;
      IF WriteEnd>0
      THEN
         GetQuad(WriteEnd, op, op1, op2, op3) ;
         IF (op=ParamOp) OR ((op=AddrOp) AND (op3=sym))
         THEN
            End := WriteEnd-1
         END
      END
   UNTIL (op#ParamOp) AND ((op#AddrOp) OR (op3#sym)) ;
   RETURN( WriteEnd=Start )
END IsLastTimeWeWrite ;


(*
   CheckNodeNeeded - checks to see whether we should write register, reg, back to
                     the destiniation symbol at node, i. Given a basic block: Start..End.
*)

PROCEDURE CheckNodeNeeded (i: IndexToItem; r: CARDINAL; Start, End: CARDINAL) ;
VAR
   reg: Register ;
BEGIN
   IF (NOT IsConditionNode(i)) AND (GetNodeOp(i)#ParamOp)
   THEN
      reg := LowReg(r) ;
      IF IsRegister(GetDestination(i))
      THEN
         IF GetDestination(i)#r
         THEN
            IF DebugOn
            THEN
               WriteString('# specific destination register needs to be written to: ') ; WriteRegisterName(GetDestination(i)) ;
               WriteLn ;
               DisplayRegisterStack
            END ;
            PeepMoveReg2Reg(r, GetDestination(i)) ;
            IF IsRegisterOnRegisterStack(r)
            THEN
               MarkRegister(r, free)
            END ;
            IF (GetDestination(i)#ESP) AND
               IsRegisterInUse(GetDestination(i))
            THEN
               DisplayRegisterStack ;
               InternalError('problems as the specific register is used', __FILE__, __LINE__)
            END
         END
      ELSE
         IncludeValueInReg(r, GetDestination(i)) ;
         IF DebugOn
         THEN
            WriteString('# register ') ; WriteRegisterName(r) ; WriteString(' contains symbol value, is it ') ; WriteKey(GetSymName(GetDestination(i))) ; WriteString(' needed? ') ;
         END ;
         IF IsUsedLaterOn(GetDestination(i), i)
         THEN
            IF DebugOn
            THEN
               WriteString('yes because it is used later on in a different tree') ; WriteLn ;
               WriteString('# delay ') ; WriteRegisterName(r) ; WriteString(' -> ') ; WriteSymbol(i) ; WriteLn ;
               Dummy
            END ;
            (* this is the last time we can write to the symbol *)
            IncludeDirtyValueInReg(r, GetDestination(i)) ;
            IF DebugOn
            THEN
               WriteString('# register is now free (last time before different tree): ') ; WriteRegisterName(r) ; WriteLn
            END ;
            MarkRegister(r, free)
         ELSIF IsSymNeeded(GetDestination(i), Start, End)
         THEN
            IF DebugOn
            THEN
               WriteString(' yes, flush now? ')
            END ;
            (* needed, but is this the last time we come across symbol GetDestination? *)
            IF IsLastTimeWeWrite(GetDestination(i), GetNodeQuad(i), End)
            THEN
               IF DebugOn
               THEN
                  WriteString('yes') ; WriteLn ;
                  WriteString('# delay ') ; WriteRegisterName(r) ; WriteString(' -> ') ; WriteSymbol(i) ; WriteLn
               END ;
               (* this is the last time we can write to the symbol *)
               IncludeDirtyValueInReg(r, GetDestination(i)) ;
               IF DebugOn
               THEN
                  WriteString('# register is now free (last time we write to symbol): ') ; WriteRegisterName(r) ; WriteLn
               END ;
               MarkRegister(r, free)
            ELSE
               IF DebugOn
               THEN
                  WriteString('no') ; WriteLn
               END
            END
         ELSE
            IF DebugOn
            THEN
               WriteString('no') ; WriteLn
            END
         END
      END
   END
END CheckNodeNeeded ;


(*
   CreateJump - creates a jump for an instruction and quadruple.
*)

PROCEDURE CreateJump (i: Instruction; q: CARDINAL) ;
VAR
   Label: ARRAY [0..MaxNumChars] OF CHAR ;
BEGIN
   CardToStr(q, 0, Label) ;
   StrConCat('L', Label, Label) ;
   DotifyLabel(Label) ;
   InstStr(i, Label) ;
   ResetRegisterValues
END CreateJump ;


(*
   CheckForConditional - checks to see if, i, indexes to a conditional node.
                         If so then it generates one of six conditional jumps.
*)

PROCEDURE CheckForConditional (i: IndexToItem) ;
BEGIN
   IF IsConditionNode(i)
   THEN
      WITH ListOfItems[i].ItemIsNode DO
         CASE Op OF

         IfInOp     :  CreateJump(jne, Condition.DestQuad) |
         IfNotInOp  :  CreateJump(je, Condition.DestQuad) |
         IfEquOp    :  CreateJump(je, Condition.DestQuad) |
         IfNotEquOp :  CreateJump(jne, Condition.DestQuad) |
         IfGreEquOp :  CreateJump(jge, Condition.DestQuad) |
         IfLessEquOp:  CreateJump(jle, Condition.DestQuad) |
         IfLessOp   :  CreateJump(jl, Condition.DestQuad) |
         IfGreOp    :  CreateJump(jg, Condition.DestQuad)

         ELSE
            InternalError('CASE not expecting this quad op', __FILE__, __LINE__)
         END
      END
   END
END CheckForConditional ;


(*
   CheckForLineNo - checks node, i, for a reference to a source code line number.
*)

PROCEDURE CheckForLineNo (i: IndexToItem) ;
BEGIN
   Assert(IsItemNode(i)) ;
   WITH ListOfItems[i].ItemIsNode DO
      IF (LineNo#0) AND (NOT APUCodeGenerator)
      THEN
         LineInfo(LineNo)
      END
   END
END CheckForLineNo ;


(*
   IsRegisterByteWordAccessible - returns TRUE if register can access byte or word entities (8, 16 bits)
*)

PROCEDURE IsRegisterByteWordAccessible (reg: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN( (reg=EAX) OR (reg=EBX) OR (reg=ECX) OR (reg=EDX) )
END IsRegisterByteWordAccessible ;


(*
   BecomesViaReg - perform a block becomes using registers.
*)

PROCEDURE BecomesViaReg (op1, op2: CARDINAL) ;
VAR
   s1, s2,
   unit,
   size,
   count,
   j     : CARDINAL ;
   go,
   offset: INTEGER ;
BEGIN
   size   := GetSize(op1) ;
   go     := GetOffset(op2) ;
   offset := 0 ;
   WHILE size#0 DO
      count := 1 ;
      s1    := size ;
      WHILE (count<=RegisterStackPtr) AND (s1#0) DO
         unit := GetUnitOfTransfer(s1) ;
         IF (unit=WordSize) OR IsRegisterByteWordAccessible(RegisterStack[RegisterStackPtr-count])
         THEN
            FlushRegister(RegisterStack[RegisterStackPtr-count]) ;
            InstRegMD(mov, GetUnitPrefix(unit), LowReg(RegisterStack[RegisterStackPtr-count]), offset, go, GetScope(op2), op2) ; ;
            TrashRegisterValue(RegisterStack[RegisterStackPtr-count]) ;
            INC(offset, unit) ;
            DEC(s1, unit)
         END ;
         INC(count)
      END ;
      go     := GetOffset(op1) ;
      offset := 0 ;
      j      := 1 ;
      s2     := size ;
      WHILE (j<=count) AND (s2#0) DO
         unit := GetUnitOfTransfer(s2) ;
         IF (unit=WordSize) OR IsRegisterByteWordAccessible(RegisterStack[RegisterStackPtr-j])
         THEN
            FlushRegister(RegisterStack[RegisterStackPtr-j]) ;
            InstMDReg(mov, GetUnitPrefix(unit), offset, go, LowReg(RegisterStack[RegisterStackPtr-j]), GetScope(op1), op1) ;
            TrashRegisterValue(RegisterStack[RegisterStackPtr-j]) ;
            INC(offset, unit) ;
            DEC(s2, unit)
         END ;
         INC(j)
      END ;
      Assert((s1=s2) AND (j=count)) ;
      DEC(size, offset)
   END
END BecomesViaReg ;


(*
   IsBaseRegister - returns TRUE if the value of, i, is the base register.
*)

PROCEDURE IsBaseRegister (i: IndexToItem; Head, Start, End: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsItemLeaf(i) AND (GetDestination(i)=EBP)
   THEN
      RETURN( TRUE )
   ELSE
      IF IsItemNode(i) AND (GetNodeOp(i)=BecomesOp) AND IsBaseRegister(GetNodeRight(i), Head, Start, End)
      THEN
         CheckNodeNeeded(i, EBP, Start, End) ;
         RETURN( TRUE )
      END
   END ;
   RETURN( FALSE )
END IsBaseRegister ;


(*
   TryToFindIndexRegister - trys to return the %ebp register if at all possible, otherwise
                            it returns the calculated index register.
*)

PROCEDURE TryToFindIndexRegister (parent, i: IndexToItem; Head, Start, End: CARDINAL) : CARDINAL ;
VAR
   state: RegisterState ;
BEGIN
   IF IsBaseRegister(i, Head, Start, End)
   THEN
      IF DebugOn
      THEN
         Comment('can optimizize the use of the base register')
      END ;
      RETURN( EBP )
   ELSE
      IF GetPrequisiteRegister(i)=NulSym
      THEN
         IF IsItemLeaf(i) AND
            IsInRegister(GetDestination(i)) AND
            IsRegisterFree(GetFromRegister(GetDestination(i))) AND
            IsRegisterOnRegisterStack(GetFromRegister(GetDestination(i)))
         THEN
            (* can optimize, index, register *)
            MoveRegisterToTop(GetFromRegister(GetDestination(i)))
         ELSE
            AhoGenCode(parent, i, FALSE, Head, Start, End)    (* op3 inside TopRegister *)
         END
      ELSE
         PreReqReg(i) ;
         AhoGenCode(parent, i, FALSE, Head, Start, End)    (* op3 inside TopRegister *)
      END ;
      RETURN( PopRegister(state) )     (* op3 in reg (the source address) *)
   END
END TryToFindIndexRegister ;


(*
   TidyUpBase - if the regsister, reg, is not EBP then push the register and state.
*)

PROCEDURE TidyUpBase (reg: CARDINAL; state: RegisterState) ;
BEGIN
   IF (reg#EBP) AND (reg#NulSym)
   THEN
      PushRegister(reg, state)
   END
END TidyUpBase ;


(*
   IndrXMoveViaRegProcedure - performs a = *b where b is a procedure variable.
*)

PROCEDURE IndrXMoveViaRegProcedure (i: IndexToItem; evalreg: CARDINAL; Head, Start, End: CARDINAL) ;
VAR
   unit,
   size,
   count,
   last,
   reg   : CARDINAL ;
   go,
   dooff,
   offset: INTEGER ;
BEGIN
   IF DebugOn
   THEN
      DisplayRegisterStack
   END ;
   IF evalreg=NulSym
   THEN
      dooff  := TryToFindOffset(GetNodeRight(i), Start, End) ;
      reg    := TryToFindIndexRegister(i, GetNodeRight(i), Head, Start, End)
   ELSE
      dooff  := 0 ;
      reg    := evalreg    (* has already been evaluated, thus make use of this value *)
   END ;
   IF DebugOn
   THEN
      DisplayRegisterStack
   END ;
   size   := GetSize(GetDestination(GetNodeLeft(i))) ;
   go     := GetOffset(GetDestination(i)) ;
   offset := 0 ;
   count  := 1 ;
   WHILE size#0 DO
      unit := GetUnitOfTransfer(size) ;
      InstRegMIO(mov, GetUnitPrefix(unit), LowReg(RegisterStack[RegisterStackPtr-count]), LowReg(reg), dooff+offset) ; 
      INC(count) ;
      INC(offset, unit) ;
      DEC(size, unit)
   END ;
   size   := GetSize(GetDestination(GetNodeLeft(i))) ;
   offset := 0 ;
   count  := 1 ;
   last   := NulSym ;
   WHILE size#0 DO
      unit := GetUnitOfTransfer(size) ;
      FlushRegister(RegisterStack[RegisterStackPtr-count]) ;
      InstMIOReg(mov, GetUnitPrefix(unit), RegEBP, go+offset, LowReg(RegisterStack[RegisterStackPtr-count])) ;
      Assert(last#RegisterStack[RegisterStackPtr-count]) ;
      last := RegisterStack[RegisterStackPtr-count] ;
      TrashRegisterValue(RegisterStack[RegisterStackPtr-count]) ;
      INC(count) ;
      INC(offset, unit) ;
      DEC(size, unit)
   END ;
   TidyUpBase(reg, free)
END IndrXMoveViaRegProcedure ;


(*
   IndrXMoveViaRegGlobal - performs a = *b where b is a global variable.
*)

PROCEDURE IndrXMoveViaRegGlobal (i: IndexToItem; evalreg: CARDINAL; Head, Start, End: CARDINAL) ;
VAR
   unit,
   size,
   count,
   dest,
   offsym,
   last,
   reg   : CARDINAL ;
   go,
   dooff,
   offset: INTEGER ;
BEGIN
   IF evalreg=NulSym
   THEN
      dooff  := TryToFindOffset(GetNodeRight(i), Start, End) ;
      offsym := TryToFindOffsetSymbol(GetNodeRight(i), Start, End) ;
      reg    := TryToFindIndexRegister(i, GetNodeRight(i), Head, Start, End) ;    (* op3 inside TopRegister *)
   ELSE
      dooff  := 0 ;
      offsym := NulSym ;
      reg    := evalreg    (* has already been evaluated, thus make use of this value *)
   END ;
   size   := GetSize(GetDestination(GetNodeLeft(i))) ;
   dest   := GetDestination(i) ;
   go     := GetOffset(GetDestination(GetNodeRight(i))) ;
   offset := 0 ;
   count  := 1 ;
   last   := NulSym ;
   WHILE size#0 DO
      unit := GetUnitOfTransfer(size) ;
      IF offsym=NulSym
      THEN
         InstRegMIO(mov, GetUnitPrefix(unit), LowReg(RegisterStack[RegisterStackPtr-count]), LowReg(reg), dooff+offset)
      ELSE
         InstRegMISym(mov, GetUnitPrefix(unit), LowReg(RegisterStack[RegisterStackPtr-count]), LowReg(reg), dooff+offset, offsym)
      END ;
      Assert(last#RegisterStack[RegisterStackPtr-count]) ;
      last := RegisterStack[RegisterStackPtr-count] ;
      INC(count) ;
      INC(offset, unit) ;
      DEC(size, unit)
   END ;
   size   := GetSize(GetDestination(GetNodeLeft(i))) ;
   offset := 0 ;
   count  := 1 ;
   WHILE size#0 DO
      unit := GetUnitOfTransfer(size) ;
      FlushRegister(RegisterStack[RegisterStackPtr-count]) ;
      InstMDReg(mov, GetUnitPrefix(unit), go, offset, LowReg(RegisterStack[RegisterStackPtr-count]), Global, dest) ;
      TrashRegisterValue(RegisterStack[RegisterStackPtr-count]) ;
      INC(count) ;
      INC(offset, unit) ;
      DEC(size, unit)
   END ;
   TidyUpBase(reg, free)
END IndrXMoveViaRegGlobal ;


(*
   IndrXMoveViaReg - performs:     a = *b   using free registers
                     and tries to produce cachable instruction sequences
*)

PROCEDURE IndrXMoveViaReg (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   state : RegisterState ;
   reg,
   src   : CARDINAL ;
   offset: INTEGER ;
   offsym: CARDINAL ;
BEGIN
   IF GetSize(GetDestination(GetNodeLeft(i)))<=WordSize
   THEN
      offset := TryToFindOffset(GetNodeRight(i), Start, End) ;
      offsym := TryToFindOffsetSymbol(GetNodeRight(i), Start, End) ;
      src    := TryToFindIndexRegister(i, GetNodeRight(i), Head, Start, End) ;    (* op3 inside TopRegister *)
      CheckUnitAccessRegister(GetSize(GetDestination(GetNodeLeft(i)))) ;
      FlushRegister(RegisterTop()) ;
      IF offsym=NulSym
      THEN
         InstRegMIO(mov, GetUnitPrefix(GetSize(GetDestination(GetNodeLeft(i)))), LowReg(RegisterTop()), LowReg(src), offset)
      ELSE
         InstRegMISym(mov, GetUnitPrefix(GetSize(GetDestination(GetNodeLeft(i)))), LowReg(RegisterTop()), LowReg(src), offset, offsym)
      END ;
      TrashRegisterValue(RegisterTop()) ;
      IF src#EBP
      THEN
         PushRegister(src, free) ;
         SwapRegisters
      END ;
      MarkTopRegisterInUse ;
      CheckNodeNeeded(i, RegisterTop(), Start, End)
   ELSE
      (* should we flush registers here? *)
      IF IsItemNode(GetNodeRight(i))
      THEN
         PreReqReg(GetNodeRight(i)) ;
         AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
         reg := PopRegister(state)    (* register is pushed inside IndrXMove...... *)
      ELSE
         reg := NulSym
      END ;
      IF IsProcedure(GetVarFather(GetDestination(GetNodeRight(i))))
      THEN
         IndrXMoveViaRegProcedure(i, reg, Head, Start, End)
      ELSE
         IndrXMoveViaRegGlobal(i, reg, Head, Start, End)
      END
   END
END IndrXMoveViaReg ;


(*
   BinaryIndrX - handle the indirection operator:   op1 := *op3
*)

PROCEDURE BinaryIndrX (i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   CheckForLineNo(i) ;
   IF GetDestination(GetNodeRight(i))=ESI
   THEN
      BlockMove(i, Head, Start, End) ;
      (*
         after indirection the cached register values might be wrong and sadly
         we have no idea which one might be affected. It is really anoying to
         reset the values, but we cannot afford to gamble... always sacrifice
         speed for safe code...
      *)
      ResetRegisterValues
   ELSE
      IndrXMoveViaReg(i, Head, Start, End)
   END
END BinaryIndrX ;


(*
   XIndrMoveViaRegProcedure - performs *a = b where b is a procedure variable.
*)

PROCEDURE XIndrMoveViaRegProcedure (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   unit,
   size,
   count,
   offsym,
   last,
   reg   : CARDINAL ;
   go,
   dooff,
   offset: INTEGER ;
BEGIN
   dooff  := TryToFindOffset(GetNodeLeft(i), Start, End) ;
   offsym := TryToFindOffsetSymbol(GetNodeLeft(i), Start, End) ;
   reg    := TryToFindIndexRegister(i, GetNodeLeft(i), Head, Start, End) ;    (* op2 inside TopRegister *)
   (* op2 (left) in reg (the destination address) *)
   size   := GetSize(GetDestination(i)) ;
   go     := GetOffset(GetDestination(GetNodeRight(i))) ;
   offset := 0 ;
   count  := 1 ;
   FlushAllRegisters ;
   WHILE size#0 DO
      unit := GetUnitOfTransfer(size) ;
      FlushRegister(RegisterStack[RegisterStackPtr-count]) ;
      InstRegMIO(mov, GetUnitPrefix(unit), LowReg(RegisterStack[RegisterStackPtr-count]), RegEBP, go+offset) ;
      TrashRegisterValue(RegisterStack[RegisterStackPtr-count]) ;
      INC(count) ;
      INC(offset, unit) ;
      DEC(size, unit)
   END ;
   offset := 0 ;
   size   := GetSize(GetDestination(i)) ;
   count  := 1 ;
   last   := NulSym ;
   WHILE size#0 DO
      unit := GetUnitOfTransfer(size) ;
      IF offsym=NulSym
      THEN
         InstMIOReg(mov, GetUnitPrefix(unit), LowReg(reg), dooff+offset, LowReg(RegisterStack[RegisterStackPtr-count]))
      ELSE
         InstMISymReg(mov, GetUnitPrefix(unit), LowReg(RegisterStack[RegisterStackPtr-count]), LowReg(reg), dooff+offset, offsym)
      END ;
      Assert(last#RegisterStack[RegisterStackPtr-count]) ;
      last := RegisterStack[RegisterStackPtr-count] ;
      INC(count) ;
      INC(offset, unit) ;
      DEC(size, unit)
   END ;
   TidyUpBase(reg, free)
END XIndrMoveViaRegProcedure ;


(*
   XIndrMoveViaRegGlobal - performs *a = b where b is a global variable.
*)

PROCEDURE XIndrMoveViaRegGlobal (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   unit,
   size,
   count,
   dest,
   offsym,
   last,
   reg   : CARDINAL ;
   dooff,
   go,
   offset: INTEGER ;
BEGIN
   dooff  := TryToFindOffset(GetNodeLeft(i), Start, End) ;
   offsym := TryToFindOffsetSymbol(GetNodeLeft(i), Start, End) ;
   reg    := TryToFindIndexRegister(i, GetNodeLeft(i), Head, Start, End) ;    (* op2 inside TopRegister *)
   (* op2 (left) in reg (the destination address) *)
   size   := GetSize(GetDestination(i)) ;
   dest   := GetDestination(GetNodeRight(i)) ;
   go     := GetOffset(GetDestination(GetNodeRight(i))) ;
   offset := 0 ;
   count  := 1 ;
   WHILE size#0 DO
      unit := GetUnitOfTransfer(size) ;
      FlushRegister(RegisterStack[RegisterStackPtr-count]) ;
      InstRegMD(mov, GetUnitPrefix(unit), LowReg(RegisterStack[RegisterStackPtr-count]), go, offset, Global, dest) ;
      TrashRegisterValue(RegisterStack[RegisterStackPtr-count]) ;
      INC(count) ;
      INC(offset, unit) ;
      DEC(size, unit)
   END ;
   offset := 0 ;
   size   := GetSize(GetDestination(i)) ;
   count  := 1 ;
   last   := NulSym ;
   FlushAllRegisters ;
   WHILE size#0 DO
      unit := GetUnitOfTransfer(size) ;
      IF offsym=NulSym
      THEN
         InstMIOReg(mov, GetUnitPrefix(unit), LowReg(reg), dooff+offset, LowReg(RegisterStack[RegisterStackPtr-count]))
      ELSE
         InstMISymReg(mov, GetUnitPrefix(unit), LowReg(RegisterStack[RegisterStackPtr-count]), LowReg(reg), dooff+offset, offsym)
      END ;
      Assert(last#RegisterStack[RegisterStackPtr-count]) ;
      last := RegisterStack[RegisterStackPtr-count] ;
      INC(count) ;
      INC(offset, unit) ;
      DEC(size, unit)
   END ;
   TidyUpBase(reg, free)
END XIndrMoveViaRegGlobal ;


(*
   XIndrMoveViaReg - performs:     *a = b   using free registers
                     and tries to produce cachable instruction sequences
*)

PROCEDURE XIndrMoveViaReg (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   top,
   index,
   offsym,
   src   : CARDINAL ;
   state : RegisterState ;
   dooff : INTEGER ;
BEGIN
   IF GetSize(GetDestination(i))<=WordSize
   THEN
      IF IsItemLeaf(GetNodeRight(i)) AND IsConst(GetDestination(GetNodeRight(i)))
      THEN
         dooff := TryToFindOffset(GetNodeLeft(i), Start, End) ;
         offsym := TryToFindOffsetSymbol(GetNodeLeft(i), Start, End) ;
         index := TryToFindIndexRegister(i, GetNodeLeft(i), Head, Start, End) ;    (* op2 inside TopRegister *)
         PushSize(GetDestination(i)) ;
         FlushAllRegisters ;
         IF offsym=NulSym
         THEN
            InstMRIConstant(mov, GetUnitPrefix(PopCard()), dooff, 0, LowReg(index), GetDestination(GetNodeRight(i)))
         ELSE
            InstMRISymConstant(mov, GetUnitPrefix(PopCard()), offsym, dooff, LowReg(index), GetDestination(GetNodeRight(i)))
         END ;
         TidyUpBase(index, free)
      ELSE
         dooff  := TryToFindOffset(GetNodeLeft(i), Start, End) ;
         offsym := TryToFindOffsetSymbol(GetNodeLeft(i), Start, End) ;

         CheckUnitAccessRegister(GetSize(GetDestination(i))) ;

         IF IsItemLeaf(GetNodeRight(i)) AND
            (GetPrequisiteRegister(GetNodeRight(i))=NulSym) AND
            IsInRegister(GetDestination(GetNodeRight(i))) AND
            IsRegisterFree(GetFromRegister(GetDestination(GetNodeRight(i))))
         THEN
            (* can optimize, src, register *)
            MoveRegisterToTop(GetFromRegister(GetDestination(GetNodeRight(i))))
         ELSE
            PreReqReg(GetNodeRight(i)) ;
            AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End)     (* op3 inside TopRegister *)
         END ;

         (* must make sure that we have a register that can manipulate bytes should GetSize demand it *)
         IF (GetSize(GetDestination(i))<DWORDSIZE) AND (NOT IsRegisterByteWordAccessible(RegisterTop()))
         THEN
            IF DebugOn
            THEN
               Comment('need a top register that can access bytes')
            END ;
            src := PopRegister(state) ;
            CheckUnitAccessRegister(GetSize(GetDestination(i))) ;
            top := RegisterTop() ;
            PeepMoveReg2Reg(src, top) ;
            PushRegister(src, free) ;
            MoveRegisterToTop(top)
         END ;

         src := PopRegister(state) ;

         (* GetNodeLeft(i) ->  RegisterTop() *)
         index  := TryToFindIndexRegister(i, GetNodeLeft(i), Head, Start, End) ;    (* op2 inside TopRegister *)
         (* src -> [index] *)

         FlushAllRegisters ;
         IF offsym=NulSym
         THEN
            InstMRIReg(mov, GetUnitPrefix(GetSize(GetDestination(i))), dooff, 0, LowReg(index), LowReg(src))
         ELSE
            InstMISymReg(mov, GetUnitPrefix(GetSize(GetDestination(i))), LowReg(src), LowReg(index), dooff, offsym)
         END ;
         TidyUpBase(index, free) ;
         PushRegister(src, free)
      END
   ELSE
      (* should we flush registers here? *)
      IF IsItemNode(GetNodeRight(i))
      THEN
         Assert(
                (GetNodeOp(GetNodeRight(i))=BecomesOp) OR
                (GetNodeOp(GetNodeRight(i))=XIndrOp) OR
                (GetNodeOp(GetNodeRight(i))=IndrXOp)
               ) ;
         PreReqReg(GetNodeRight(i)) ;
         AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End)
      END ;
      (* we flush registers in both XIndrMoveViaRegProcedure and XIndrMoveViaRegGlobal *)
      IF IsProcedure(GetVarFather(GetDestination(GetNodeRight(i))))
      THEN
         XIndrMoveViaRegProcedure(i, Head, Start, End)
      ELSE
         XIndrMoveViaRegGlobal(i, Head, Start, End)
      END
   END
END XIndrMoveViaReg ;


(*
   MoveStringFromToIndirect - moves a string, StringSym, to an indirect destination.
*)

PROCEDURE MoveStringFromToIndirect (StringSym: CARDINAL; DestSize: CARDINAL; offset: INTEGER) ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
   b: ARRAY [0..3] OF CHAR ;
   Pack,
   i,
   String,
   SourceSize: CARDINAL ;
BEGIN
   (* Make sure the string is the same size as DestSym *)
   String := GetString(StringSym) ;
   SourceSize := GetStringLength(StringSym) ;
   GetKey(String, a) ;
   i := 0 ;
   WHILE i<SourceSize DO
      Pack := PackString(a, DestSize, i, b) ;
      MovePackedStringToIReg(LowReg(RegisterTop()), INTEGER(i)+offset, b, Pack) ;
      INC(i, Pack)
   END ;
   IF i>DestSize
   THEN
      WriteError('string is too large for the variable')
   END ;
   (* Now fill in the remaining space with nul characters *)
   WHILE i<DestSize DO
      Pack := PackString('', DestSize, i, b) ;
      MovePackedStringToIReg(LowReg(RegisterTop()), INTEGER(i)+offset, b, Pack) ;
      INC(i, Pack)
   END
END MoveStringFromToIndirect ;


(*
   Memcopy - perform the reps on the number of bytes held on the ALU stack.
*)

PROCEDURE Memcopy ;
VAR
   nBytes,
   nWords: CARDINAL ;
BEGIN
   FlushInCaseOfMultiRegOps(repmovs) ;
   nBytes := PopCard() ;
   nWords := nBytes DIV WordSize ;
   IF nWords>0
   THEN
      DEC(nBytes, nWords * WordSize) ;
      IF NOT IsRegisterFree(ECX)
      THEN
         DisplayRegisterStack ;
         InternalError('%ecx should be free', __FILE__, __LINE__)
      END ;
      InstRegInt(mov, MaxDatum(), RegECX, INTEGER(nWords)) ;
      BinInstUnit(repmovs, MaxDatum(), nWords)
   END ;
   IF nBytes>0
   THEN
      InstRegInt(mov, MaxDatum(), RegECX, INTEGER(nBytes)) ;
      BinInstUnit(repmovs, PByte, nBytes)
   END ;
   UnHideRepsRegisters ;
   CheckForMultiRegOps(repmovs)
END Memcopy ;


(*
   BlockMove - moves a block of data using the reps instruction.
*)

PROCEDURE BlockMove (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   dest, src: CARDINAL ;
   state    : RegisterState ;
BEGIN
   IF NOT IsRegisterFree(ECX)
   THEN
      DisplayRegisterStack ;
      InternalError('%ecx is in use and we will need to use it...', __FILE__, __LINE__)
   END ;
   PreReqReg(GetNodeRight(i)) ;
   AhoGenCode(i, GetNodeRight(i), TRUE, Head, Start, End) ;    (* op3 inside TopRegister *)
   IF RegisterTop()#ESI
   THEN
      IF DebugOn
      THEN
         DisplayRegisterStack
      END
   END ;
   (* ESI may not be on top but will have been written to *)
   MoveRegisterToTop(ESI) ;
   Assert(RegisterTop()=ESI) ;
   src := PopRegister(state) ;
   PreReqReg(GetNodeLeft(i)) ;
   AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End) ;
   (* EDI may not be on top but will have been written to *)
   MoveRegisterToTop(EDI) ;
   Assert(RegisterTop()=EDI) ;
   dest := PopRegister(state) ;
   Assert(IsConst(GetDestination(i))) ;
   PushValue(GetDestination(i)) ;
   FlushAllRegisters ;
   Memcopy ;
   PushRegister(dest, free) ;
   PushRegister(src, free)
END BlockMove ;


(*
   BinaryXIndr - handle the indirection operator:   *op1 := op3
                 Note that:  the type is in GetDestination(i) (GetSize(type) = no of bytes to be copied)
                             op1 is in GetNodeLeft(i)
                             op3 is in GetNodeRight(i)
*)

PROCEDURE BinaryXIndr (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
(*   offsym: CARDINAL ; *)
   dooff : INTEGER ;
BEGIN
   CheckForLineNo(i) ;
   IF GetDestination(GetNodeRight(i))=ESI
   THEN
      BlockMove(i, Head, Start, End)
   ELSIF IsConstString(GetDestination(GetNodeRight(i)))
   THEN
      PreReqReg(GetNodeLeft(i)) ;
      IF IsItemLeaf(GetNodeRight(i)) AND IsConst(GetDestination(GetNodeRight(i)))
      THEN
         dooff  := TryToFindOffset(GetNodeLeft(i), Start, End)
         (* offsym := TryToFindOffsetSymbol(GetNodeLeft(i), Start, End) *)
      END ;
      AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End) ;
      (* RegisterTop is the address where we must place the constant string *)
      FlushAllRegisters ;
      MoveStringFromToIndirect(GetDestination(GetNodeRight(i)), GetSize(GetDestination(i)), dooff)
   ELSE
      XIndrMoveViaReg(i, Head, Start, End)
   END ;
   (*
      after indirection the cached register values might be wrong and sadly
      we have no idea which one might be affected. It is really anoying to
      reset the values, but we cannot afford to gamble... always sacrifice
      speed for safe code...
   *)
   ResetRegisterValues
END BinaryXIndr ;


(*
   SubtreeRequires - returns TRUE if the subtree, i, requires register, reg.
*)

PROCEDURE SubtreeRequires (i: IndexToItem; reg: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsItemNode(i)
   THEN
      CASE ListOfItems[i].ItemIsNode.Op OF

      DivOp,
      ModOp,
      MultOp       :  IF (reg=EAX) OR (reg=EDX)
                      THEN
                         RETURN( TRUE )
                      END |
      FunctValueOp :  IF reg=EAX
                      THEN
                         RETURN( TRUE )
                      END

      ELSE
         (* keeps p2c happy *)
      END ;
      IF IsBinaryNode(i)
      THEN
         RETURN(
                SubtreeRequires(GetNodeRight(i), reg) OR
                SubtreeRequires(GetNodeLeft(i), reg)
               )
      ELSIF IsUnaryNode(i)
      THEN
         RETURN( SubtreeRequires(ListOfItems[i].ItemIsNode.Unary.Right, reg) )
      END
   END ;
   RETURN( FALSE )
END SubtreeRequires ;


(*
   LikelyToReturn - returns TRUE if the node, i, is likely to return
                    its value in register, reg.
*)

PROCEDURE LikelyToReturn (i: IndexToItem; reg: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsItemNode(i)
   THEN
      CASE ListOfItems[i].ItemIsNode.Op OF

      ModOp       :  RETURN( reg=EDX ) |
      DivOp,
      MultOp      :  RETURN( reg=EAX ) |
      FunctValueOp:  RETURN( reg=EAX )

      ELSE
      END
   END ;
   RETURN( RegisterTop()=reg )
END LikelyToReturn ;


(*
   MultBinaryCase1 - 
*)

PROCEDURE MultBinaryCase1 (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   state             : RegisterState ;
   oldClean, oldDirty: List ;
BEGIN
   (* right node requires no registers *)
   IF IsRegisterOnRegisterStack(EAX) AND IsRegisterFree(EAX)
   THEN
      IF DebugOn
      THEN
         WriteString('# yippee we can utilize the register we desire: ') ;
         WriteRegisterName(EAX) ; WriteLn
      END ;
      MoveRegisterToTop(EAX)
   END ;
   AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End) ;
   IF RegisterTop()=EAX
   THEN
      InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX)
   ELSE
      IF DebugOn
      THEN
         Comment('bother result has come back in a different register')
      END ;
      IF IsRegisterOnRegisterStack(EAX) AND IsRegisterFree(EAX)
      THEN
         PeepMoveReg2Reg(RegisterTop(), EAX) ;
         InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX) ;
         PushRegister(PopRegister(state), free) ;
         MoveRegisterToTop(EAX)
      ELSE
         PeepPushReg(EAX, oldClean, oldDirty) ;
         PeepMoveReg2Reg(RegisterTop(), EAX) ;
         InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX) ;
         PeepMoveReg2Reg(EAX, RegisterTop()) ;
         PeepPopReg(EAX, oldClean, oldDirty)
      END
   END
END MultBinaryCase1 ;


(*
   MultBinaryCase3 - traverse the IsLeft branch then the opposite branch
                     and then perform the multiply.
*)

PROCEDURE MultBinaryCase3 (IsLeft: BOOLEAN; i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   oldClean,
   oldDirty  : List ;
   reg1, reg2: CARDINAL ;
   state     : RegisterState ;
BEGIN
   IF DebugOn
   THEN
      DisplayRegisterStack
   END ;
   reg1 := FreeNonDivModRegister() ;
   IF reg1#NulSym
   THEN
      PushRegister(reg1, free)    (* excellent we use this register *)
   END ;
   IF IsLeft
   THEN
      AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End)
   ELSE
      AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End)
   END ;
   reg1 := PopRegister(state) ;
   IF reg1=EAX
   THEN
      IF DebugOn
      THEN
         Comment('bother subtree has used %eax')
      END ;
      reg1 := FreeNonDivModRegister() ;
      IF reg1=NulSym
      THEN
         IF IsLeft
         THEN
            AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End)
         ELSE
            AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End)
         END
      ELSE
         PeepMoveReg2Reg(EAX, reg1) ;
         PushRegister(EAX, free) ;
         IF IsLeft
         THEN
            AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End)
         ELSE
            AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End)
         END ;
         reg2 := PopRegister(state) ;
         IF reg2=EAX
         THEN
            InstNodeRegReg(i, LowReg(reg1), RegEAX) ;
            PushRegister(EAX, used)   (* result in %eax *)
         ELSE
            IF DebugOn
            THEN
               Comment('bother tree did not evaluate into %eax (perhaps this could be improved)')
            END ;
            IF IsRegisterFree(EAX)
            THEN
               PeepMoveReg2Reg(reg2, EAX) ;
               InstNodeRegReg(i, LowReg(reg1), RegEAX) ;
               PushRegister(reg2, free) ;
               MoveRegisterToTop(EAX)
            ELSE
               PeepPushReg(EAX, oldClean, oldDirty) ;
               PeepMoveReg2Reg(reg2, EAX) ;
               InstNodeRegReg(i, LowReg(reg1), RegEAX) ;
               PeepMoveReg2Reg(EAX, reg2) ;
               PeepPopReg(EAX, oldClean, oldDirty)
            END
         END
      END
   ELSE
      (* ************************ *)
      IF DebugOn
      THEN
         Comment('WARNING WARNING') ;
         DisplayRegisterStack ;
         WriteString('# good we have not used %eax, we will make sure %eax is on top, ')
      END ;
      IF IsRegisterFree(EAX) AND IsRegisterOnRegisterStack(EAX)
      THEN
         IF RegisterTop()=EAX
         THEN
            IF DebugOn
            THEN
               WriteString('fine %eax is already on top') ; WriteLn
            END
         ELSE
            MoveRegisterToTop(EAX) ;
            IF DebugOn
            THEN
               WriteString('we will move %eax to the top') ; WriteLn ;
               DisplayRegisterStack
            END
         END
      ELSE
         IF DebugOn
         THEN
            WriteString('%eax is not free after all - strange?') ; WriteLn
         END
      END ;
      (* ************************* *)
      IF IsLeft
      THEN
         IF (reg1=EDX) AND SubtreeRequires(GetNodeRight(i), EDX)
         THEN
            reg2 := FreeNonDivModRegister() ;
            IF reg2#NulSym
            THEN
               IF DebugOn
               THEN
                  Comment('not used %eax, but we have used %edx - which we will copy to a spare register')
               END ;
               PushRegister(reg1, free) ;
               PeepMoveReg2Reg(EDX, reg2) ;
               PushRegister(reg2, used) ;
               reg1 := reg2
            END
         END ;
         AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End)
      ELSE
         IF (reg1=EDX) AND SubtreeRequires(GetNodeLeft(i), EDX)
         THEN
            reg2 := FreeNonDivModRegister() ;
            IF reg2#NulSym
            THEN
               IF DebugOn
               THEN
                  Comment('not used %eax, but we have used %edx - which we will copy to a spare register')
               END ;
               PushRegister(reg1, free) ;
               PeepMoveReg2Reg(EDX, reg2) ;
               reg1 := reg2
            END
         END ;
         AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End)
      END ;
      IF RegisterTop()=EAX
      THEN
         InstNodeRegReg(i, LowReg(reg1), RegEAX) ;
         PushRegister(reg1, free) ;
         (* these two lines were absent and  were replaced by PushRegister(EAX, used) *)
         MoveRegisterToTop(EAX) ;
         MarkTopRegisterInUse    (* result in %eax *) 
      ELSE
         IF DebugOn
         THEN
            Comment('bother result has come back in a different register')
         END ;
         IF IsRegisterOnRegisterStack(EAX) AND IsRegisterFree(EAX)
         THEN
            PeepMoveReg2Reg(RegisterTop(), EAX) ;
            InstNodeRegReg(i, LowReg(reg1), RegEAX) ;
            PushRegister(PopRegister(state), free) ;
            PushRegister(reg1, free) ;
            MoveRegisterToTop(EAX)
         ELSE
            PushRegister(reg1, free) ;
            SwapRegisters ;
            PeepPushReg(EAX, oldClean, oldDirty) ;
            PeepMoveReg2Reg(RegisterTop(), EAX) ;
            InstNodeRegReg(i, LowReg(reg1), RegEAX) ;
            PeepMoveReg2Reg(EAX, RegisterTop()) ;
            PeepPopReg(EAX, oldClean, oldDirty)
         END
      END
   END
END MultBinaryCase3 ;


(*
   MultBinaryCase2 - handles the case where both subtrees require at least one
                     register. It tests for EAX, EDX requirements and free register
                     availability to decide which subtree to explore first.
*)

PROCEDURE MultBinaryCase2 (i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   IF SubtreeRequires(GetNodeLeft(i), EAX) AND
      SubtreeRequires(GetNodeRight(i), EAX)
   THEN
      IF NOT LikelyToReturn(GetNodeLeft(i), EAX)
      THEN
         MultBinaryCase3(TRUE, i, Head, Start, End)     (* left first  *)
      ELSIF NOT LikelyToReturn(GetNodeRight(i), EAX)
      THEN
         MultBinaryCase3(FALSE, i, Head, Start, End)    (* right first *)
      ELSE
         IF DebugOn
         THEN
            Comment('bother both subtrees use %eax')
         END ;
         IF GetItemValue(GetNodeRight(i))>GetItemValue(GetNodeLeft(i))
         THEN
            MultBinaryCase3(FALSE, i, Head, Start, End)    (* right first *)
         ELSE
            MultBinaryCase3(TRUE, i, Head, Start, End)     (* left first  *)
         END
      END
   ELSIF SubtreeRequires(GetNodeLeft(i), EAX)
   THEN
      IF DebugOn
      THEN
         Comment('explore right side as it does not need %eax')
      END ;
      (* the right hand side does not need EAX, explore this first *)
      MultBinaryCase3(FALSE, i, Head, Start, End)
   ELSIF SubtreeRequires(GetNodeRight(i), EAX)
   THEN
      IF DebugOn
      THEN
         Comment('explore left side as it does not need %eax')
      END ;
      MultBinaryCase3(TRUE, i, Head, Start, End)
   ELSE
      IF DebugOn
      THEN
         Comment('neither require %eax')
      END ;
      (* perhaps we should use a different case *)
      IF GetItemValue(GetNodeRight(i))>GetItemValue(GetNodeLeft(i))
      THEN
         MultBinaryCase3(FALSE, i, Head, Start, End)    (* right first *)
      ELSE
         MultBinaryCase3(TRUE, i, Head, Start, End)     (* left first  *)
      END
   END
END MultBinaryCase2 ;


(*
   BinaryMult - 
*)

PROCEDURE BinaryMult (i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   CheckForLineNo(i) ;
   IF GetItemValue(GetNodeRight(i))=0
   THEN
      MultBinaryCase1(i, Head, Start, End)
   ELSE
      (* right requires at least one register *)
      MultBinaryCase2(i, Head, Start, End)
   END ;
   MarkTopRegisterInUse ;
   CheckNodeNeeded(i, RegisterTop(), Start, End) ;
   CheckForConditional(i)
END BinaryMult ;


(*
   FreeNonDivModRegister - finds a free register which is not associated with division or modulus.
*)

PROCEDURE FreeNonDivModRegister () : CARDINAL ;
VAR
   reg: CARDINAL ;
   i  : CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<RegisterStackPtr DO
      reg := RegisterStack[i] ;
      IF (reg=ECX) OR (reg=EBX) OR (reg=ESI) OR
         (reg=EDI) OR (reg=EBP)
      THEN
         RemoveStackedRegister(reg) ;
         RETURN( reg )
      END ;
      INC(i)
   END ;
   RETURN( NulSym )
END FreeNonDivModRegister ;


(*
   BinaryDivCase1 - handle the divide using a variable and the EDX:EAX pair.
*)

PROCEDURE BinaryDivCase1 (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   oldeaxClean,
   oldeaxDirty,
   oldedxClean,
   oldedxDirty: List ;
BEGIN
   IF DebugOn
   THEN
      Comment('divide through variable')
   END ;
   IF (RegisterTop()=EDX) AND (RegisterStackPtr>1)
   THEN
      SwapRegisters (* no point in putting EDX on the top of the free list as we will overwrite it here *)
   END ;
   PreReqReg(GetNodeLeft(i)) ;
   AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End) ;
   IF RegisterTop()=EAX
   THEN
      InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX)
   ELSE
      PeepPushReg(EAX, oldeaxClean, oldeaxDirty) ;
      PeepMoveReg2Reg(RegisterTop(), EAX) ;
      IF (NOT IsRegisterFree(EDX)) AND (RegisterTop()#EDX)
      THEN
         PeepPushReg(EDX, oldedxClean, oldedxDirty) ;
         InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX) ;
         PeepMoveReg2Reg(EAX, RegisterTop()) ;
         PeepPopReg(EDX, oldedxClean, oldedxDirty)
      ELSE
         InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX) ;
         PeepMoveReg2Reg(EAX, RegisterTop())
      END ;
      PeepPopReg(EAX, oldeaxClean, oldeaxDirty)
   END
END BinaryDivCase1 ;


(*
   BinaryDivCase2 - handle the divide using a register and the EDX:EAX pair.
*)

PROCEDURE BinaryDivCase2 (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   oldeaxClean,
   oldeaxDirty,
   oldedxClean,
   oldedxDirty: List ;
   reg        : CARDINAL ;
   state      : RegisterState ;
BEGIN
   IF DebugOn
   THEN
      Comment('divide through register')
   END ;
   reg := PopRegister(state) ;
   PreReqReg(GetNodeLeft(i)) ;
   AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End) ;
   IF RegisterTop()=EAX
   THEN
      InstNodeRegReg(i, LowReg(reg), RegEAX) ;
      PushRegister(reg, free) ;   (* and back again for the result *)
      SwapRegisters               (* return result in EAX          *)
   ELSIF IsRegisterFree(EAX) AND IsRegisterOnRegisterStack(EAX)
   THEN
      PeepMoveReg2Reg(RegisterTop(), EAX) ;
      IF RegisterTop()=EDX
      THEN
         (* EDX is being used as the top register, hence no need to save it *)
         InstNodeRegReg(i, LowReg(reg), RegEAX) ;
         MarkRegister(EDX, free) ;
         MoveRegisterToTop(EAX) ;
         MarkTopRegisterInUse
      ELSE
         IF IsRegisterFree(EDX)
         THEN
            InstNodeRegReg(i, LowReg(reg), RegEAX) ;
            PeepMoveReg2Reg(EAX, RegisterTop())
         ELSE
            PeepPushReg(EDX, oldedxClean, oldedxDirty) ;
            InstNodeRegReg(i, LowReg(reg), RegEAX) ;
            PeepMoveReg2Reg(EAX, RegisterTop()) ;
            PeepPopReg(EDX, oldedxClean, oldedxDirty)
         END
      END
   ELSE
      PeepPushReg(EAX, oldeaxClean, oldeaxDirty) ;
      PeepMoveReg2Reg(RegisterTop(), EAX) ;
      IF (NOT IsRegisterFree(EDX)) AND (RegisterTop()#EDX)
      THEN
         PeepPushReg(EDX, oldedxClean, oldedxDirty) ;
         InstNodeRegReg(i, LowReg(reg), RegEAX) ;
         PushRegister(reg, free) ;   (* and back again for the result *)
         PeepMoveReg2Reg(EAX, RegisterTop()) ;
         PeepPopReg(EDX, oldedxClean, oldedxDirty)
      ELSE
         InstNodeRegReg(i, LowReg(reg), RegEAX) ;
         PushRegister(reg, free) ;   (* and back again for the result *)
         PeepMoveReg2Reg(EAX, RegisterTop())
      END ;
      PeepPopReg(EAX, oldeaxClean, oldeaxDirty)
   END
END BinaryDivCase2 ;


(*
   PreReqDivModReg - 
*)

PROCEDURE PreReqDivModReg (i: IndexToItem) ;
VAR
   reg: CARDINAL ;
BEGIN
   IF (GetPrequisiteRegister(i)=NulSym) AND (NOT IsTopRegisterUsed())
   THEN
      reg := FreeNonDivModRegister() ;
      IF reg#NulSym
      THEN
         PushRegister(reg, free)
      END
   ELSE
      PreReqReg(i)
   END
END PreReqDivModReg ;


(*
   BinaryDiv - 
*)

PROCEDURE BinaryDiv (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   reg  : CARDINAL ;
   state: RegisterState ;
BEGIN
   CheckForLineNo(i) ;
   (* sadly the div instruction does not work with a constant in the right hand side *)
   IF (GetItemValue(GetNodeRight(i))=0) AND (NOT IsConst(GetDestination(GetNodeRight(i))))
   THEN
      (* right requires no registers *)
      BinaryDivCase1(i, Head, Start, End)
   ELSE
      (* right requires at least one register *)
      PreReqDivModReg(GetNodeRight(i)) ;
      AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
      IF (RegisterTop()=EAX) OR (RegisterTop()=EDX)
      THEN
         (* bother we are using a register that will be required elsewhere to perform this operation *)
         reg := FreeNonDivModRegister() ;
         IF reg#NulSym
         THEN
            IF DebugOn
            THEN
               Comment('able to use a spare register for intemediate division result')
            END ;
            PeepMoveReg2Reg(RegisterTop(), reg) ;
            PushRegister(PopRegister(state), free) ;
            PushRegister(reg, used) ;
            BinaryDivCase2(i, Head, Start, End)
         ELSE
            InstRegSym(mov, LowReg(RegisterTop()), GetDestination(GetNodeRight(i))) ;     (* save to a temporary *)
            BinaryDivCase1(i, Head, Start, End)
         END
      ELSE
         BinaryDivCase2(i, Head, Start, End)
      END
   END ;
   MarkTopRegisterInUse ;
   CheckNodeNeeded(i, RegisterTop(), Start, End) ;
   CheckForConditional(i)
END BinaryDiv ;


(*
   BinaryModCase1 - handle the modulus using a variable and the EDX:EAX pair.
*)

PROCEDURE BinaryModCase1 (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   oldedxClean,
   oldedxDirty,
   oldeaxClean,
   oldeaxDirty: List ;
BEGIN
   IF DebugOn
   THEN
      Comment('modulus through variable')
   END ;
   IF (RegisterTop()=EDX) AND (RegisterStackPtr>1)
   THEN
      SwapRegisters (* no point in putting EDX on the top of the free list as we will overwrite it here *)
   END ;
   PreReqReg(GetNodeLeft(i)) ;
   AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End) ;
   IF DebugOn
   THEN
      DisplayRegisterStack
   END ;
   IF RegisterTop()=EAX
   THEN
      IF IsRegisterFree(EDX)
      THEN
         InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX) ;
         MarkRegister(EAX, free) ;
         MoveRegisterToTop(EDX)
      ELSE
         PeepPushReg(EDX, oldedxClean, oldedxDirty) ;
         InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX) ;
         PeepMoveReg2Reg(EDX, EAX) ;
         PeepPopReg(EDX, oldedxClean, oldedxDirty)
      END
   ELSIF IsRegisterFree(EAX)
   THEN
      PeepMoveReg2Reg(RegisterTop(), EAX) ;
      IF RegisterTop()=EDX
      THEN
         (* EDX is being used as the top register, hence no need to save it *)
         InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX)
      ELSE
         IF IsRegisterFree(EDX)
         THEN
            InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX) ;
            PeepMoveReg2Reg(EDX, RegisterTop())
         ELSE
            PeepPushReg(EDX, oldedxClean, oldedxDirty) ;
            InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX) ;
            PeepMoveReg2Reg(EDX, RegisterTop()) ;
            PeepPopReg(EDX, oldedxClean, oldedxDirty)
         END
      END
   ELSE
      PeepPushReg(EAX, oldeaxClean, oldeaxDirty) ;
      PeepMoveReg2Reg(RegisterTop(), EAX) ;
      IF IsRegisterFree(EDX)
      THEN
         InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX) ;
         PeepMoveReg2Reg(EDX, RegisterTop())
      ELSE
         PeepPushReg(EDX, oldedxClean, oldedxDirty) ;
         InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), RegEAX) ;
         PeepMoveReg2Reg(EDX, RegisterTop()) ;
         PeepPopReg(EDX, oldedxClean, oldedxDirty)
      END ;
      PeepPopReg(EAX, oldeaxClean, oldeaxDirty)
   END
END BinaryModCase1 ;


(*
   BinaryModCase2 - handle the modulus using a register and the EDX:EAX pair.
*)

PROCEDURE BinaryModCase2 (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   oldedxClean,
   oldedxDirty,
   oldeaxClean,
   oldeaxDirty: List ;
   reg        : CARDINAL ;
   state      : RegisterState ;
BEGIN
   IF DebugOn
   THEN
      Comment('modulus through register')
   END ;
   reg := PopRegister(state) ;
   PreReqReg(GetNodeLeft(i)) ;
   AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End) ;
   IF RegisterTop()=EAX
   THEN
      IF IsRegisterFree(EDX)
      THEN
         InstNodeRegReg(i, LowReg(reg), RegEAX) ;
         MarkRegister(EAX, free) ;
         PushRegister(reg, free) ;   (* and back again for the result *)
         MoveRegisterToTop(EDX)
      ELSE
         PeepPushReg(EDX, oldedxClean, oldedxDirty) ;
         InstNodeRegReg(i, LowReg(reg), RegEAX) ;
         PushRegister(PopRegister(state), free) ;
         PushRegister(reg, free) ;   (* and back again for the result *)
         PeepMoveReg2Reg(EDX, reg) ;
         PeepPopReg(EDX, oldedxClean, oldedxDirty)
      END
   ELSE
      PeepPushReg(EAX, oldeaxClean, oldeaxDirty) ;
      PeepMoveReg2Reg(RegisterTop(), EAX) ;
      IF IsRegisterFree(EDX)
      THEN
         InstNodeRegReg(i, LowReg(reg), RegEAX) ;
         PushRegister(reg, free) ;   (* and back again for the result *)
         PeepMoveReg2Reg(EDX, RegisterTop())
      ELSE
         PeepPushReg(EDX, oldedxClean, oldedxDirty) ;
         InstNodeRegReg(i, LowReg(reg), RegEAX) ;
         PushRegister(reg, free) ;   (* and back again for the result *)
         PeepMoveReg2Reg(EDX, RegisterTop()) ;
         PeepPopReg(EDX, oldedxClean, oldedxDirty)
      END ;
      PeepPopReg(EAX, oldeaxClean, oldeaxDirty)
   END
END BinaryModCase2 ;


(*
   BinaryMod - 
*)

PROCEDURE BinaryMod (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   reg  : CARDINAL ;
   state: RegisterState ;
BEGIN
   CheckForLineNo(i) ;
   (* sadly the divl instructions will not work with a constant operand *)
   IF (GetItemValue(GetNodeRight(i))=0) AND (NOT IsConst(GetDestination(GetNodeRight(i))))
   THEN
      (* right requires no registers *)
      BinaryModCase1(i, Head, Start, End)
   ELSE
      (* right requires at least one register *)
      PreReqDivModReg(GetNodeRight(i)) ;
      AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
      IF (RegisterTop()=EAX) OR (RegisterTop()=EDX)
      THEN
         (* bother we are using a register that will be required elsewhere to perform this operation *)
         reg := FreeNonDivModRegister() ;
         IF reg#NulSym
         THEN
            IF DebugOn
            THEN
               Comment('able to use a spare register for intemediate division result')
            END ;
            PeepMoveReg2Reg(RegisterTop(), reg) ;
            PushRegister(PopRegister(state), free) ;
            PushRegister(reg, used) ;
            BinaryModCase2(i, Head, Start, End)
         ELSE
            InstRegSym(mov, LowReg(RegisterTop()), GetDestination(GetNodeRight(i))) ;     (* save to a temporary *)
            BinaryModCase1(i, Head, Start, End)
         END
      ELSE
         BinaryModCase2(i, Head, Start, End)
      END
   END ;
   MarkTopRegisterInUse ;
   CheckNodeNeeded(i, RegisterTop(), Start, End) ;
   CheckForConditional(i)
END BinaryMod ;


(*
   MarkTopRegisterInUse - 
*)

PROCEDURE MarkTopRegisterInUse ;
VAR
   state: RegisterState ;
   reg  : CARDINAL ;
BEGIN
   reg := PopRegister(state) ;
   PushRegister(reg, used)   (* now in use *)
END MarkTopRegisterInUse ;


(*
   MarkRegister - mark a register, reg, to the state, state.
*)

PROCEDURE MarkRegister (reg: CARDINAL; state: RegisterState) ;
VAR
   i: CARDINAL ;
BEGIN
   i := InitRegisterPtr ;
   WHILE i>0 DO
      IF ListOfRegisters[i-1].Reg=reg
      THEN
         IF state#ListOfRegisters[i-1].State
         THEN
            ListOfRegisters[i-1].State := state
         END ;
         RETURN
      ELSE
         DEC(i)
      END
   END
END MarkRegister ;


(*
   BinaryUnbounded - evaluate the unbounded node.
*)

PROCEDURE BinaryUnbounded (parent, i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   array,
   op3,
   reg   : CARDINAL ;
   offset: INTEGER ;
BEGIN
   CheckForLineNo(i) ;
   offset := TryToFindOffset(GetNodeLeft(i), Start, End) ;
   array := GetDestination(i) ;
   IF RegisterTop()=EBX
   THEN
      (* do not want to use a register used for saving the stack pointer *)
      SwapRegisters
   END ;
   IF NOT IsConstString(array)
   THEN
      (*
         we will be using reps edi, esi, ecx  therefore we must check we
         are not going to use these registers to hold the address of the
         unbounded descriptor.
      *)
      IF (RegisterTop()=ECX) OR (RegisterTop()=EDI) OR (RegisterTop()=ESI)
      THEN
         IF DebugOn
         THEN
            WriteString('# we have a potential problem we must move ') ;
            WriteRegisterName(RegisterTop()) ; WriteString(' out of the way of a pending reps') ; WriteLn
         END ;
         IF IsRegisterFree(EDX)
         THEN
            IF DebugOn
            THEN
               Comment('moving %edx to the top')
            END ;
            MoveRegisterToTop(EDX)
         ELSE
            IF DebugOn
            THEN
               Comment('we have real problems - as edx is not available')
            END ;
            InternalError('need to find a way of saving RegisterTop and restoring it at end', __FILE__, __LINE__)
         END
      END
   END ;
      
   reg := TryToFindIndexRegister(parent, GetNodeLeft(i), Head, Start, End) ;
   IF reg=EBX
   THEN
      InternalError('not going to work as the indirection is held in ebx and also the unbounded data start is in ebx', __FILE__, __LINE__)
   END ;
   
   PrepareToPushUnboundedData ;                 (* we will be using %ebx ...                  *)

   op3 := GetDestination(GetNodeRight(i)) ;
   (*
      Set op1 to the start of the copy of op3

      push op3 onto the stack - therefore op1 := esp
   *)
   IF IsConstString(array)
   THEN
      PushStringToStack(op3) ;
      (*
         set op2 to the start of the copy of op3
         Remember 80*86 stack grows from HIGH -> LOW memory.
         Therefore op2 := esp
         Last entity on the stack is the first element of the unbounded array.
      *)
      InstMIOReg(mov, MaxDatum(), LowReg(reg), offset, RegESP) ;
      TidyUpBase(reg, free)
   ELSE
      (* all the below use reps, so we must align the stack pointer after copying esp *)
      IF IsArray(GetType(array))
      THEN
         PushArrayToStack(array, parent, i, Head, Start, End)
      ELSIF IsUnbounded(GetType(array))
      THEN
         PushUnbounded(op3)
      ELSE
         (*
            we should only execute this if we are passing an object to an ARRAY OF BYTE or ARRAY OF WORD
            parameter. Parameter checking will have been done in Pass 3 and we have no way of double
            checking - so here we go...
         *)
         PushArrayToStack(array, parent, i, Head, Start, End)
      END ;
      (*
         set op2 to the start of the copy of op3
         Remember 80*86 stack grows from HIGH -> LOW memory.
         Therefore op2 := esp
         Last entity on the stack is the first element of the unbounded array.
      *)
      InstMIOReg(mov, MaxDatum(), LowReg(reg), offset, RegESP) ;
      TidyUpBase(reg, free) ;
      AlignStackPointer
   END
END BinaryUnbounded ;


(*
   UnaryConvert - 
*)

PROCEDURE UnaryConvert (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   op3,
   s1, s3: CARDINAL ;
   reg   : CARDINAL ;
   state : RegisterState ;
BEGIN
   op3 := GetDestination(GetNodeRight(i)) ;
   s1 := GetVarTypeSize(GetDestination(i)) ;
   IF IsConst(op3)
   THEN
      s3 := s1
   ELSE
      s3 := GetVarTypeSize(op3)
   END ;
   IF (s1<=WordSize) AND (s3<=WordSize)
   THEN
      AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
      CheckForLineNo(i) ;
      reg := PopRegister(state) ;
      FlushRegister(reg) ;
      IF s1>s3
      THEN
         IF s3=BYTESIZE
         THEN
            InstRegHexCard(and, MaxDatum(), LowReg(reg), 0FFH)
         ELSIF s3=WORDSIZE
         THEN
            InstRegHexCard(and, MaxDatum(), LowReg(reg), 0FFFFH)
         ELSE
            WriteError('peculiar type size - check this out!')
         END
      END ;
      TrashRegisterValue(reg) ;
      PushRegister(reg, used) ;
      CheckNodeNeeded(i, RegisterTop(), Start, End) ;
      CheckForConditional(i)
   ELSE
      WriteErrorFormat1('implementation restriction - can only CONVERT types which have sizes <= %d bytes', WordSize)
      (* Alter this when time permits *)
   END
END UnaryConvert ;


(*
   UnaryBit - builds a bit op:   op1 := 1 << op3
              using   sal  op3, cl

              must use ecx register, but it executes 3x faster than the bit set instruction...
*)

PROCEDURE UnaryBit (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   oldecxClean,
   oldecxDirty: List ;
   reg        : CARDINAL ;
   state      : RegisterState ;
BEGIN
   IF (GetItemValue(GetNodeRight(i))=0) AND IsConst(GetDestination(GetNodeRight(i)))
   THEN
      InternalError('should have folded constants in M2Optimize', __FILE__, __LINE__)
   ELSE
      CheckForLineNo(i) ;
      PreReqReg(GetNodeRight(i)) ;
      AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
      IF RegisterTop()=ECX
      THEN
         IF DebugOn
         THEN
            Comment('good ecx contains the shift value')
         END ;
         reg := PopRegister(state) ;
         PeepMoveConst2Reg(MakeConstLit(MakeKey('1')), RegisterTop()) ;
         FlushRegister(RegisterTop()) ;
         InstRegReg(sal, GetUnitPrefix(GetVarTypeSize(GetDestination(i))), LowReg(RegisterTop()), RegECX) ;
         TrashRegisterValue(RegisterTop()) ;
         PushRegister(reg, free) ;  (* and free up ecx *)
         SwapRegisters              (* result on top *)
      ELSE
         PeepPushReg(ECX, oldecxClean, oldecxDirty) ;
         PeepMoveReg2Reg(RegisterTop(), ECX) ;
         PeepMoveConst2Reg(MakeConstLit(MakeKey('1')), RegisterTop()) ;
         FlushRegister(RegisterTop()) ;
         InstRegReg(sal, GetUnitPrefix(GetVarTypeSize(GetDestination(i))), LowReg(RegisterTop()), RegECX) ;
         TrashRegisterValue(RegisterTop()) ;
         PeepPopReg(ECX, oldecxClean, oldecxDirty)
      END
   END ;
   MarkTopRegisterInUse ;
   CheckNodeNeeded(i, RegisterTop(), Start, End) ;
   CheckForConditional(i)
END UnaryBit ;


(*
   GenBinaryCase1 - called when the right branch requires NO registers to evaluate.
                    This is the general case 1.
*)

PROCEDURE GenBinaryCase1 (i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   CheckForLineNo(i) ;
   (* compute the left branch into a register and issue an instruction:   <op  reg, right> *)
   IF GetItemValue(GetNodeLeft(i))=0
   THEN
      (* left takes NO registers either, special case optimization:  <op   left, right>  *)
      InternalError('finish this code gaius', __FILE__, __LINE__)
   ELSE
      PreReqReg(GetNodeLeft(i)) ;
      AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End) ;
      InstNodeReg(i, GetNodeRight(i), GetNodeLeft(i), LowReg(RegisterTop())) ;
      MarkTopRegisterInUse ;
      CheckNodeNeeded(i, RegisterTop(), Start, End) ;
      CheckForConditional(i)
   END
END GenBinaryCase1 ;


(*
   GenBinaryCase2 - 
*)

PROCEDURE GenBinaryCase2 (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   reg  : CARDINAL ;
   state: RegisterState ;
BEGIN
   SwapRegisters ;
   PreReqReg2(GetNodeLeft(i), GetNodeRight(i)) ;
   IF DebugOn
   THEN
      Comment('case 2 performed both prereq and the stack is as follows') ;
      DisplayRegisterStack
   END ;
   AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
   reg := PopRegister(state) ;  (* Right(i) is evaluated into reg *)
   AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End) ;
   CheckForLineNo(i) ;
   InstNodeRegReg(i, LowReg(reg), LowReg(RegisterTop())) ;
   MarkTopRegisterInUse ;
   CheckNodeNeeded(i, RegisterTop(), Start, End) ;
   (*
      it would seem that we should free register, reg, as it will no longer
      be on top after the swap below.
   *)
   FlushRegister(reg) ;         (* an addition with PushRegister(reg, free) *)
   CheckForConditional(i) ;
   PushRegister(reg, free) ;    (* was    PushRegister(reg, GetRegisterState(reg)) ;  (gaius) *)
   SwapRegisters
END GenBinaryCase2 ;


(*
   GenBinaryCase3 - 
*)

PROCEDURE GenBinaryCase3 (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   reg  : CARDINAL ;
   state: RegisterState ;
BEGIN
   PreReqReg2(GetNodeRight(i), GetNodeLeft(i)) ;
   IF DebugOn
   THEN
      Comment('case 3')
   END ;
   AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End) ;
   reg := PopRegister(state) ;  (* Left(i) is evaluated into reg *)
   AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
   CheckForLineNo(i) ;
   InstNodeRegReg(i, LowReg(RegisterTop()), LowReg(reg)) ;
   PushRegister(reg, used) ;
   CheckNodeNeeded(i, reg, Start, End) ;
   CheckForConditional(i)
END GenBinaryCase3 ;


(*
   GenBinaryCase4 - 
*)

PROCEDURE GenBinaryCase4 (i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
   InstRegSym(mov, LowReg(RegisterTop()), GetDestination(i)) ;   (* save to a temporary *)
   MarkRegister(RegisterTop(), free) ;                           (* no longer using top register *)
   PreReqReg(GetNodeLeft(i)) ;
   AhoGenCode(i, GetNodeLeft(i), TRUE, Head, Start, End) ;
   CheckForLineNo(i) ;
   InstNodeReg(i, i, GetNodeLeft(i), LowReg(RegisterTop())) ;   (* operate on temporary and reg *)
   MarkTopRegisterInUse ;
   CheckNodeNeeded(i, RegisterTop(), Start, End) ;
   CheckForConditional(i)
END GenBinaryCase4 ;


(*
   GeneralBinaryCases - enumerates the cases outlined in the dragon book on P564.
                        We optimize P564 slightly by checking to see whether a leaf is a constant.
*)

PROCEDURE GeneralBinaryCases (i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   IF GetItemValue(GetNodeRight(i))=0
   THEN
      (* case 1 *)
      GenBinaryCase1(i, Head, Start, End)
   ELSIF (1<=GetItemValue(GetNodeLeft(i))) AND
         (GetItemValue(GetNodeLeft(i))<GetItemValue(GetNodeRight(i))) AND
         (GetItemValue(GetNodeLeft(i))<RegisterStackPtr)
   THEN
      (* case 2 *)
      GenBinaryCase2(i, Head, Start, End)
   ELSIF ((1<=GetItemValue(GetNodeRight(i))) AND
          (GetItemValue(GetNodeRight(i))<=GetItemValue(GetNodeLeft(i))) AND
          (GetItemValue(GetNodeRight(i))<RegisterStackPtr)) OR
         IsConditionNode(i)
   THEN
      (* case 3 *)
      GenBinaryCase3(i, Head, Start, End)
   ELSE
      (* case 4   both labels >= the number of registers available *)
      GenBinaryCase4(i, Head, Start, End)
   END
END GeneralBinaryCases ;


(*
   BinaryCases - check for special binary cases and the default to the general binary
                 algorithm given in the Dragon book.
*)

PROCEDURE BinaryCases (parent, i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   (* i must be an interior node *)
   CASE GetNodeOp(i) OF

   IndrXOp    :  BinaryIndrX(i, Head, Start, End) |
   XIndrOp    :  BinaryXIndr(i, Head, Start, End) |
   DivOp      :  BinaryDiv(i, Head, Start, End) |
   ModOp      :  BinaryMod(i, Head, Start, End) |
   MultOp     :  BinaryMult(i, Head, Start, End) |
   UnboundedOp:  BinaryUnbounded(parent, i, Head, Start, End)

   ELSE
      GeneralBinaryCases(i, Head, Start, End)
   END
END BinaryCases ;


(*
   UnaryBecomes - special case unary operator: becomes
*)

PROCEDURE UnaryBecomes (parent, i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   IF IsConstString(GetDestination(GetNodeRight(i))) AND (GetStringLength(GetDestination(GetNodeRight(i)))>1)
   THEN
      CheckForLineNo(i) ;
      MoveStringFromToDirect(GetDestination(GetNodeRight(i)), GetDestination(i), GetVarTypeSize(GetDestination(i)))
   ELSIF (GetItemValue(ListOfItems[i].ItemIsNode.Unary.Right)=0) AND 
         IsConst(GetDestination(ListOfItems[i].ItemIsNode.Unary.Right))
   THEN
      CheckForLineNo(i) ;
      IF IsRegister(GetDestination(i))
      THEN
         PeepMoveConst2Reg(GetDestination(ListOfItems[i].ItemIsNode.Unary.Right), GetDestination(i))
      ELSE
         InstMDConstant(mov, GetUnitPrefix(GetUnitOfTransfer(GetSize(GetDestination(i)))),
                        0, GetOffset(GetDestination(i)), GetScope(GetDestination(i)),
                        GetDestination(i), GetDestination(ListOfItems[i].ItemIsNode.Unary.Right)) ;
         (* now we need to remove GetDestination(i) from all the register values *)
         TrashSymbolValue(GetDestination(i))
      END
   ELSE
      IF IsConst(GetDestination(GetNodeRight(i))) OR (GetSize(GetDestination(GetNodeRight(i)))<=WordSize)
      THEN
         (* these moves must go through a register *)

         SecondGuessTopRegisterMove(i) ;

         PreReqReg(GetNodeRight(i)) ;
         AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
         CheckForLineNo(i) ;
         CheckNodeNeeded(i, RegisterTop(), Start, End)
      ELSE
         IF IsItemNode(GetNodeRight(i))
         THEN
            PreReqReg(GetNodeRight(i)) ;
            AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End)
         END ;
         CheckForLineNo(i) ;
         BecomesViaReg(GetDestination(i), GetDestination(GetNodeRight(i)))     (* straight forward block copy *)
      END
   END
END UnaryBecomes ;


(*
   UnaryReturnValue - special case unary operator: ReturnValueOp
*)

PROCEDURE UnaryReturnValue (i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   (* all other moves must go through a register *)
   IF IsRegisterOnRegisterStack(EAX)
   THEN
      MoveRegisterToTop(EAX)
   END ;
   AhoGenCode(i, ListOfItems[i].ItemIsNode.Unary.Right, FALSE, Head, Start, End) ;
   CheckForLineNo(i) ;
   IF RegisterTop()#EAX
   THEN
      IF IsRegisterOnRegisterStack(EAX)
      THEN
         PeepMoveReg2Reg(RegisterTop(), EAX)
      ELSE
         InternalError('hmm, we lost EAX somewhere in the register allocator', __FILE__, __LINE__)
      END
   END ;
   IF SetFunctionReturnValue
   THEN
      InternalError('cannot set the return value twice...', __FILE__, __LINE__)
   END ;
   SetFunctionReturnValue := TRUE
END UnaryReturnValue ;


(*
   UnaryFunctValue - special case unary operator: FunctValueOp.
                     This node will be the first in a basic block therefore
                     we can alter the register order...
*)

PROCEDURE UnaryFunctValue (i: IndexToItem; Start, End: CARDINAL) ;
BEGIN
   RemoveAnyUnboundedData ;
   CheckForLineNo(i) ;
   Assert(GetDestination(ListOfItems[i].ItemIsNode.Unary.Right)=EAX) ;
   Assert(UsingFunctionReturnValue) ;
   UsingFunctionReturnValue := FALSE ;  (* no longer using return value *)
   PushRegister(EAX, used) ;   (* now make return result available *)
   CheckNodeNeeded(i, RegisterTop(), Start, End)
END UnaryFunctValue ;


(*
   FetchAddress - places the address of symbol, sym, into the top register.
*)

PROCEDURE FetchAddress (Sym: CARDINAL) ;
BEGIN
   FlushRegister(RegisterTop()) ;
   IF IsProcedure(GetVarFather(Sym))
   THEN
      IF GetOffset(Sym)=0
      THEN
         (* so unlikely - it may even be impossible? *)
         InstRegReg(mov, MaxDatum(), LowReg(RegisterTop()), RegEBP)
      ELSE
         InstRegMD(lea, MaxDatum(), LowReg(RegisterTop()), GetOffset(Sym), 0, GetScope(Sym), Sym)
      END
(*
      IF GetOffset(Sym)=1
      THEN
         BinInstReg(inc, MaxDatum(), LowReg(RegisterTop()))
      ELSIF GetOffset(Sym)=-1
      THEN
         BinInstReg(dec, MaxDatum(), LowReg(RegisterTop()))
      ELSIF GetOffset(Sym)>1
      THEN
         InstRegInt(add, MaxDatum(), LowReg(RegisterTop()), GetOffset(Sym))
      ELSIF GetOffset(Sym)<1
      THEN
         InstRegInt(sub, MaxDatum(), LowReg(RegisterTop()), -GetOffset(Sym))
      END
*)
   ELSE
      InstRegOffset(mov, MaxDatum(), LowReg(RegisterTop()), Sym)
   END ;
   TrashRegisterValue(RegisterTop())
END FetchAddress ;


(*
   UnaryAddr - special case unary operator: Addr
*)

PROCEDURE UnaryAddr (i: IndexToItem; Start, End: CARDINAL) ;
VAR
   Sym: CARDINAL ;
BEGIN
   CheckForLineNo(i) ;
   Sym := GetDestination(ListOfItems[i].ItemIsNode.Unary.Right) ;
   IF IsConst(Sym)
   THEN
      WriteError('not expecting to be asked to find the address of a constant')
   ELSE
      FlushRegister(RegisterTop()) ;
      IF IsProcedure(GetVarFather(Sym))
      THEN
         InstRegMD(lea, MaxDatum(), LowReg(RegisterTop()), GetOffset(Sym), 0, GetScope(Sym), Sym)
      ELSE
         InstRegOffset(mov, MaxDatum(), LowReg(RegisterTop()), Sym)
      END ;
      TrashRegisterValue(RegisterTop()) ;
      MarkTopRegisterInUse
   END ;
   CheckNodeNeeded(i, RegisterTop(), Start, End)
END UnaryAddr ;


(*
   UnaryCall - handles the indirect procedure call
*)

PROCEDURE UnaryCall (i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   Proc : CARDINAL ;
   reg  : CARDINAL ;
   state: RegisterState ;
BEGIN
   CheckForLineNo(i) ;
   AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
   FlushAllRegisters ;

   IF GetMode(GetDestination(GetNodeRight(i)))=LeftValue
   THEN
      reg := PopRegister(state) ;
      InstRegMIO(mov, MaxDatum(), LowReg(RegisterTop()), LowReg(reg), 0) ;
      BinInstRegI(call, MaxDatum(), LowReg(RegisterTop()))
   ELSE
      BinInstRegI(call, MaxDatum(), LowReg(RegisterTop()))
   END ;

   ResetRegisterValues ;
   Proc := GetType(GetDestination(GetNodeRight(i))) ;
   Assert(IsProcType(Proc)) ;
   IF GetType(Proc)=NulSym   (* Is it a function *)
   THEN
      (* Procedure *)
      IF UnboundedOnStack
      THEN
         RemoveAnyUnboundedData
      ELSE
         PushSumOfParamSize(Proc) ;
         IncreaseStack(PopCard())
      END
   ELSE
      (* Function *)
      PushSumOfParamSize(Proc) ;
      IncreaseStack(PopCard())
   END
END UnaryCall ;


(*
   GeneralUnaryCases - handle the register sized general unary operators.
*)

PROCEDURE GeneralUnaryCases (i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   (* not a BecomesOp so we can utilize constant or memory operand more efficiently *)
   IF GetItemValue(ListOfItems[i].ItemIsNode.Unary.Right)=0
   THEN
      CheckForLineNo(i) ;
      IF IsConst(GetDestination(ListOfItems[i].ItemIsNode.Unary.Right))
      THEN
         BinInstConstant(OpToInst(GetNodeOp(i)),
                         GetUnitPrefix(GetUnitOfTransfer(GetSize(GetDestination(i)))),
                         GetDestination(ListOfItems[i].ItemIsNode.Unary.Right))
      ELSE
         BinInstMD(OpToInst(GetNodeOp(i)),
                   GetUnitPrefix(GetUnitOfTransfer(GetSize(GetDestination(i)))),
                   GetOffset(GetDestination(ListOfItems[i].ItemIsNode.Unary.Right)), 0,
                   GetScope(GetDestination(ListOfItems[i].ItemIsNode.Unary.Right)),
                   GetDestination(ListOfItems[i].ItemIsNode.Unary.Right))
      END ;
      (* now we need to remove GetDestination(i) from all the register values *)
      TrashSymbolValue(GetDestination(i))
   (* we now look for special case of pushing a known register *)
   ELSIF (OpToInst(GetNodeOp(i))=push) AND
         IsItemLeaf(ListOfItems[i].ItemIsNode.Unary.Right) AND
         IsInRegister(GetDestination(ListOfItems[i].ItemIsNode.Unary.Right)) AND
         ((GetSize(GetDestination(i))=WordSize) OR
          IsRegisterByteWordAccessible(GetFromRegister(GetDestination(ListOfItems[i].ItemIsNode.Unary.Right))))
   THEN
      IF DebugOn
      THEN
         Comment('found special case push known reg value')
      END ;
      CheckForLineNo(i) ;
      BinInstReg(OpToInst(GetNodeOp(i)),
                 GetUnitPrefix(GetUnitOfTransfer(GetSize(GetDestination(i)))),
                 LowReg(GetFromRegister(GetDestination(ListOfItems[i].ItemIsNode.Unary.Right)))) ;
      CheckNodeNeeded(i, GetFromRegister(GetDestination(ListOfItems[i].ItemIsNode.Unary.Right)), Start, End)
   ELSE
      (* operand held in a register *)
      AhoGenCode(i, ListOfItems[i].ItemIsNode.Unary.Right, FALSE, Head, Start, End) ;
      CheckForLineNo(i) ;
      IF OpToInst(GetNodeOp(i))#push
      THEN
         FlushRegister(RegisterTop())
      END ;
      BinInstReg(OpToInst(GetNodeOp(i)),
                 GetUnitPrefix(GetUnitOfTransfer(GetSize(GetDestination(i)))),
                 LowReg(RegisterTop())) ;
      IF OpToInst(GetNodeOp(i))#push
      THEN
         TrashRegisterValue(RegisterTop())
      END ;
      MarkTopRegisterInUse ;
      CheckNodeNeeded(i, RegisterTop(), Start, End)
   END
END GeneralUnaryCases ;


(*
   SecondGuessTopRegisterMove -  check to see whether a register already
                                 contains the value we want. If so and it
                                 is free then we put it on top of the stack.
                                 Basically we try and second guess the
                                 register to use.
*)

PROCEDURE SecondGuessTopRegisterMove (i: IndexToItem) ;
BEGIN
   IF IsItemLeaf(GetNodeRight(i)) AND
      IsInRegister(GetDestination(GetNodeRight(i))) AND
      IsRegisterFree(GetFromRegister(GetDestination(GetNodeRight(i)))) AND
      IsRegisterOnRegisterStack(GetFromRegister(GetDestination(GetNodeRight(i))))
   THEN
      IF DebugOn
      THEN
         Comment('priming the becomes with a register with same value')
      END ;
      MoveRegisterToTop(GetFromRegister(GetDestination(GetNodeRight(i))))
   END
END SecondGuessTopRegisterMove ;


(*
   UnaryParam - handles the param operator.
*)

PROCEDURE UnaryParam (parent, i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   offset         : INTEGER ;
   Unit, Size, Sym: CARDINAL ;
BEGIN
   IF IsVarParamNode(i)
   THEN
      (* these parameters will be passed by a single wordsize address.
         Firstly check to see whether we have a register that contains
         the correct value *)
      IF IsItemLeaf(GetNodeRight(i)) AND
         IsInRegister(GetDestination(GetNodeRight(i)))
      THEN
         IF DebugOn
         THEN
            Comment('found special case push known reg value')
         END ;
         CheckForLineNo(i) ;
         BinInstReg(push,
                    MaxDatum(),
                    LowReg(GetFromRegister(GetDestination(GetNodeRight(i)))))
      ELSE
         PreReqReg(GetNodeRight(i)) ;
         AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
         CheckForLineNo(i) ;
         BinInstReg(push, MaxDatum(), LowReg(RegisterTop()))
      END
(*
   ELSIF GetSize(GetDestination(i))=BYTESIZE
   THEN
      (* 80*86 cannot push a byte *)
      BinInstReg(dec, MaxDatum(), RegESP) ;

      SecondGuessTopRegisterMove(i) ;

      PreReqReg(GetNodeRight(i)) ;
      CheckUnitAccessRegister(BYTESIZE) ;
      AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
      CheckForLineNo(i) ;
      InstMRIReg(mov, GetUnitPrefix(BYTESIZE), 0, 0, RegESP, LowReg(RegisterTop()))
*)
   ELSIF GetSize(GetDestination(i))<WordSize
   THEN
      SecondGuessTopRegisterMove(i) ;

      PreReqReg(GetNodeRight(i)) ;
      AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
      CheckForLineNo(i) ;
      BinInstReg(push, GetUnitPrefix(DWORDSIZE), LowReg(RegisterTop()))
   ELSIF GetSize(GetDestination(i))=WordSize
   THEN
      GeneralUnaryCases(i, Head, Start, End)
   ELSE
      IF IsItemNode(GetNodeRight(i))
      THEN
         AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End)
      END ;
      CheckForLineNo(i) ;
      Size := GetSize(GetDestination(i)) ;
      Sym  := GetDestination(GetNodeRight(i)) ;
      offset := GetOffset(GetDestination(GetNodeRight(i))) ;
      WHILE Size>0 DO
         Unit := GetUnitOfTransfer(Size) ;
         DEC(Size, Unit) ;
         BinInstMD(push, GetUnitPrefix(Unit), offset, INTEGER(Size), GetScope(Sym), Sym)
      END
   END
END UnaryParam ;


(*
   UnaryCases - handles the unary cases.
*)

PROCEDURE UnaryCases (parent, i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   CASE GetNodeOp(i) OF

   (* special case unary ops *)
   BecomesOp    :  UnaryBecomes(parent, i, Head, Start, End) |
   AddrOp       :  UnaryAddr(i, Start, End) |
   CallOp       :  UnaryCall(i, Head, Start, End) |
   ReturnValueOp:  UnaryReturnValue(i, Head, Start, End) |
   FunctValueOp :  UnaryFunctValue(i, Start, End) |
   ParamOp      :  UnaryParam(parent, i, Head, Start, End) |
   ConvertOp    :  UnaryConvert(i, Head, Start, End) |
   BitOp        :  UnaryBit(i, Head, Start, End)

   ELSE
      GeneralUnaryCases(i, Head, Start, End)
   END
END UnaryCases ;


(*
   PreReqReg - tries to place the correct prerequesite register on top of the free list.
               Note that it does not have to succeed, in which case the specific operator
               node in the tree will have to sort the problem out. PreReqReg simply
               tries to utilize sensible registers wherever possible, hopefully producing
               better code for simple statement sequences. (Less register moves etc).
*)

PROCEDURE PreReqReg (i: IndexToItem) ;
VAR
   reg: CARDINAL ;
BEGIN
   reg := GetPrequisiteRegister(i) ;
   IF reg#NulSym
   THEN
      Assert(IsRegister(reg)) ;
      IF IsRegisterOnRegisterStack(reg) AND IsRegisterFree(reg)
      THEN
         WriteString('# yippee (PreReqReg) we can utilize the register we desire: ') ;
         WriteRegisterName(reg) ; WriteLn ;
         MoveRegisterToTop(reg)
      ELSE
         WriteString('# sadly we cannot utilize the register we desire: ') ; WriteRegisterName(reg) ; WriteLn
      END
   END
END PreReqReg ;


(*
   PreReqReg2 - ensures that the correct prerequesite register is available on the free list.
*)

PROCEDURE PreReqReg2 (i1, i2: IndexToItem) ;
VAR
   reg1, reg2: CARDINAL ;
BEGIN
   reg1 := GetPrequisiteRegister(i1) ;
   reg2 := GetPrequisiteRegister(i2) ;
   IF reg1=NulSym
   THEN
      IF reg2#NulSym
      THEN
         WriteString('# double prerequisite only requires second prereq') ; WriteRegisterName(reg2) ; WriteLn ;
         PreReqReg(i2)
      END
   ELSE
      PreReqReg(i1) ;
      IF reg2=NulSym
      THEN
         IF DebugOn
         THEN
            Comment('caught double prerequisite')
         END ;
         (* problems as we will use reg1 for evaluating the second subtree and not the first! *)
         SwapRegisters
      END
   END
END PreReqReg2 ;


(*
   IsDummyNode - returns TRUE if a dummy node is encountered.
*)

PROCEDURE IsDummyNode (i: IndexToItem) : BOOLEAN ;
BEGIN
   RETURN( IsItemNode(i) AND (GetNodeOp(i)=DummyOp) )
END IsDummyNode ;


(*
   DummyCases - process the dummy case.
*)

PROCEDURE DummyCases (i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   Assert(IsDummyNode(i)) ;
   WITH ListOfItems[i].ItemIsNode.Dummy DO
      PreReqReg(Tree1) ; AhoGenCode(i, Tree1, FALSE, Head, Start, End) ;
      PreReqReg(Tree2) ; AhoGenCode(i, Tree2, FALSE, Head, Start, End)
   END
END DummyCases ;


(*
   CheckRegisterDestination - checks to see whether, i, is the top node and
                              whether the destination is a register. In which
                              case this register is removed from the available list.
*)

PROCEDURE CheckRegisterDestination (parent, i: IndexToItem) ;
VAR
   reg: CARDINAL ;
BEGIN
   IF parent=NulNode
   THEN
      reg := GetDestination(i) ;
      IF IsRegister(reg)
      THEN
         IF IsRegisterOnRegisterStack(reg)
         THEN
            IF DebugOn
            THEN
               WriteString('# hiding register ') ; WriteRegisterName(reg) ; WriteLn
            END ;
            IF reg=EDI
            THEN
               RemoveStackedRegister(reg) ;
               MarkRegister(reg, hidden) ;
               HideEDI := TRUE
            ELSIF reg=ESI
            THEN
               RemoveStackedRegister(reg) ;
               MarkRegister(reg, hidden) ;
               HideESI := TRUE
            ELSIF reg=ECX
            THEN
               RemoveStackedRegister(reg) ;
               MarkRegister(reg, hidden) ;
               HideECX := TRUE
            ELSE
               InternalError('do not know how to hide this register', __FILE__, __LINE__)
            END
         ELSE
            InternalError('register should be on the register stack', __FILE__, __LINE__)
         END
      END
   END
END CheckRegisterDestination ;


(*
   AhoGenCode - algorithm taken from P564 of the Dragon Book.
*)

PROCEDURE AhoGenCode (parent, i: IndexToItem; IsLeftLeaf: BOOLEAN;
                      Head, Start, End: CARDINAL) ;
BEGIN
   IF (GetDestination(i)#NulSym) AND IsVar(GetDestination(i))
   THEN
      CheckUnitAccessRegister(GetSize(GetDestination(i)))
   END ;
   IF IsLeftLeaf AND IsItemLeaf(i)
   THEN
      InstNodeReg(i, i, i, LowReg(RegisterTop()))
   ELSIF IsBinaryNode(i) OR IsConditionNode(i)
   THEN
      BinaryCases(parent, i, Head, Start, End)
   ELSIF IsDummyNode(i)
   THEN
      DummyCases(i, Head, Start, End)
   ELSE
      IF IsItemLeaf(i)
      THEN
         (*
            this looks like bad code being generated, but in fact we sometimes
            must put constants and leaves into a register as we have no mem -> mem instructions
         *)
         InstNodeReg(i, i, i, LowReg(RegisterTop()))
      ELSIF IsUnaryNode(i)
      THEN
         UnaryCases(parent, i, Head, Start, End)
      END
   END ;
   CheckRegisterDestination(parent, i)
END AhoGenCode ;


(*
   DisplaySym - 
*)

PROCEDURE DisplaySym (Sym: CARDINAL) ;
BEGIN
   IF Sym=NulSym
   THEN
      WriteString(' <nul> ')
   ELSE
      WriteKey(GetSymName(Sym))
   END ;
   WriteString(', ')
END DisplaySym ;


(*
   DisplayRegisterStack - 
*)

PROCEDURE DisplayRegisterStack ;
VAR
   i: CARDINAL ;
BEGIN
   WriteString('# Reg stack [') ;
   i := RegisterStackPtr ;
   WHILE i>0 DO
      WriteRegisterName(RegisterStack[i-1]) ; WriteString(' | ') ;
      DEC(i)
   END ;
   Write(']') ; WriteLn ;
   WriteString('# Reg clean [') ;
   i := InitRegisterPtr ;
   WHILE i>0 DO
      WITH ListOfRegisters[i-1] DO
         WriteRegisterName(Reg) ; WriteString(' ') ;
         WriteRegisterState(State) ; WriteString(' <') ;
         IF CleanAlias#NIL
         THEN
            ForeachItemInListDo(CleanAlias, DisplaySym)
         END ;
         WriteString('> | ')
      END ;
      DEC(i)
   END ;
   WriteString(']') ; WriteLn ;
   WriteString('# Reg dirty [') ;
   i := InitRegisterPtr ;
   WHILE i>0 DO
      WITH ListOfRegisters[i-1] DO
         WriteRegisterName(Reg) ; WriteString(' ') ;
         WriteRegisterState(State) ; WriteString(' <') ;
         IF DirtyAlias#NIL
         THEN
            ForeachItemInListDo(DirtyAlias, DisplaySym)
         END ;
         WriteString('> | ')
      END ;
      DEC(i)
   END ;
   WriteString(']') ; WriteLn
END DisplayRegisterStack ;


(*
   ResetRegisterValues - reset all known values of registers to NulSym, unknown.
*)

PROCEDURE ResetRegisterValues ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<InitRegisterPtr DO
      WITH ListOfRegisters[i] DO
         IF (DirtyAlias#NIL) AND (NoOfItemsInList(DirtyAlias)>0)
         THEN
            DisplayRegisterStack ;
            WriteString('# problem with register ') ; WriteRegisterName(Reg) ; WriteLn ;
            InternalError('register is dirty and we are resetting all registers, need to flush all regs before here', __FILE__, __LINE__)
         END
      END ;
      INC(i)
   END ;
   InitRegisterPtr := 0 ;
   InitRegister(ESP, used) ;   (* ESP is included in the list as well *)
   IF HideEDI
   THEN
      InitRegister(EDI, hidden)
   ELSE
      InitRegister(EDI, free)
   END ;
   IF UnboundedOnStack
   THEN
      InitRegister(EBX, hidden)
   ELSE
      InitRegister(EBX, free)
   END ;
   IF HideECX
   THEN
      InitRegister(ECX, hidden)
   ELSE
      InitRegister(ECX, free)
   END ;
   IF HideESI
   THEN
      InitRegister(ESI, hidden)
   ELSE
      InitRegister(ESI, free)
   END ;
   InitRegister(EDX, free) ;
   IF UsingFunctionReturnValue OR SetFunctionReturnValue
   THEN
      InitRegister(EAX, hidden)
   ELSE
      InitRegister(EAX, free)
   END ;
   IF InsideLeafProcedure
   THEN
      InitRegister(EBP, free)       (* can use EBP if in a leaf procedure (later optimization) *)
   ELSE
      InitRegister(EBP, hidden)
   END
END ResetRegisterValues ;


(*
   IsRegisterValueKnown - returns TRUE if a register, reg, value is
                          currently known.
*)

PROCEDURE IsRegisterValueKnown (reg: CARDINAL) : BOOLEAN ;
VAR
   clean, dirty: List ;
BEGIN
   GetRegisterAssociate(reg, clean, dirty) ;
   RETURN( (clean#NIL) AND (NoOfItemsInList(clean)>0) )
END IsRegisterValueKnown ;


(*
   SwapRegister - swaps registers in position, i, j, on the register stack.
*)

PROCEDURE SwapRegister (i, j: CARDINAL) ;
VAR
   regtmp: CARDINAL ;
BEGIN
   regtmp := RegisterStack[i] ;
   RegisterStack[i] := RegisterStack[j] ;
   RegisterStack[j] := regtmp
END SwapRegister ;


(*
   ReOrderStackedRegisters - we attempt to leave registers which have
                             known values in them, alone, for as long
                             as possible. We do this by moving registers
                             with no value in them to the top of the stack.
*)

PROCEDURE ReOrderStackedRegisters ;
VAR
   regtmp: CARDINAL ;
   bot,
   i     : CARDINAL ;
BEGIN
   i   := RegisterStackPtr ;
   bot := 1 ;
   WHILE (i>0) AND (i#bot-1) DO
      (* find lowest register which has no known value *)
      IF IsRegisterValueKnown(RegisterStack[bot-1])
      THEN
         IF DebugOn
         THEN
            WriteString('# we know about register ') ; WriteRegisterName(RegisterStack[bot-1]) ; WriteLn
         END ;
         INC(bot)
      ELSIF IsRegisterValueKnown(RegisterStack[i-1])
      THEN
         (* bot is guarenteed to index an register without a value *)
         IF DebugOn
         THEN
            WriteString('# moving register out of immediate use ' ) ; WriteRegisterName(RegisterStack[i-1]) ;
            WriteString(' swapping with ') ; WriteRegisterName(RegisterStack[bot-1]) ; WriteLn
         END ;
         (* swap with bot and increment bot *)
         SwapRegister(i-1, bot-1) ;
         INC(bot)
      ELSE
         DEC(i)
      END
   END ;
   IF DebugOn
   THEN
      WriteString('# inside ReOrderStackedRegisters') ; WriteLn ;
      DisplayRegisterStack
   END
END ReOrderStackedRegisters ;


(*
   WalkTree - initializes the register stack and calls AhoGenCode algorithm.
*)

PROCEDURE WalkTree (i: IndexToItem; Head, Start, End: CARDINAL) ;
BEGIN
   RegisterStackPtr := 0 ;
   PushRegister(EDI, free) ;
   IF NOT UnboundedOnStack
   THEN
      PushRegister(EBX, free)
   END ;
   PushRegister(ECX, free) ;
   PushRegister(ESI, free) ;
   PushRegister(EDX, free) ;
   IF (NOT UsingFunctionReturnValue) AND (NOT SetFunctionReturnValue)
   THEN
      PushRegister(EAX, free)    (* eax is free *)
   END ;
   IF InsideLeafProcedure
   THEN
      PushRegister(EBP, free)    (* can use EBP if in a leaf procedure (later optimization) *)
   END ;
   ReOrderStackedRegisters ;
   PreReqReg(i) ;
   AhoGenCode(NulNode, i, FALSE, Head, Start, End)
END WalkTree ;


(*
   FlushForest - flush the forest of trees, we must flush the call node last.
*)

PROCEDURE FlushForest (Head, Start, End: CARDINAL) ;
VAR
   i         : IndexToTree ;
   SeenCall,
   SeenReturn: BOOLEAN ;
BEGIN
   SetFunctionReturnValue := FALSE ;
   (* label forest, and reduce tree *)
   i := 1 ;   
   WHILE i<=TreePtr DO
      IF ListOfTrees[i]#0
      THEN
         IF DebugOn
         THEN
            WriteString('# ' ) ; DisplayTree(ListOfTrees[i]) ; WriteLn
         END ;
         WHILE CheckFoldTree(Head, Start, End, ListOfTrees[i]) DO
            IF DebugOn
            THEN
               WriteString('# ' ) ; DisplayTree(ListOfTrees[i]) ; WriteLn
            END ;
         END ;
         AssignValue(ListOfTrees[i], LabelTree(ListOfTrees[i], FALSE)) ;
         IF DebugOn
         THEN
            WriteString('# ') ; DisplayTree(ListOfTrees[i]) ; WriteLn
         END
      END ;
      INC(i)
   END ;
   SeenCall   := FALSE ;
   SeenReturn := FALSE ;
   i := 1 ;   
   WHILE i<=TreePtr DO
      IF ListOfTrees[i]#0
      THEN
         IF DebugOn
         THEN
            DisplayRegisterStack ;
            Comment('going to code') ;
            WriteString('# ') ; DisplayTree(ListOfTrees[i]) ; WriteLn
         END ;
         IF SeenCall
         THEN
            InternalError('already coded the call before this tree', __FILE__, __LINE__)
         END ;
         (* emit non call trees *)
         IF NOT SeenCall
         THEN
            SeenCall   := GetNodeOp(ListOfTrees[i])=CallOp
         END ;
         IF NOT SeenReturn
         THEN
            SeenReturn := GetNodeOp(ListOfTrees[i])=ReturnValueOp
         END ;
         IF IsRegister(GetDestination(ListOfTrees[i]))
         THEN
            IF DebugOn
            THEN
               WriteString('# need to hide register after this tree is evaluated: ') ; WriteRegisterName(GetDestination(ListOfTrees[i])) ;
               WriteLn
            END
         END ;
         WalkTree(ListOfTrees[i], Head, Start, End) ;
         ListOfTrees[i] := 0
      END ;
      INC(i)
   END ;
   ItemPtr       := 1 ;
   TreePtr       := 0 ;  (* now get rid of the forest *)
   LastNode      := 0 ;
   LastFlushQuad := End ;
   FlushAllRegisters
END FlushForest ;


(*
   Max - returns the largest of two numbers, a, and, b.
*)

PROCEDURE Max (a, b: CARDINAL) : CARDINAL ;
BEGIN
   IF a>b
   THEN
      RETURN( a )
   ELSE
      RETURN( b )
   END
END Max ;


(*
   Min - returns the smallest of two numbers, a, and, b.
*)

PROCEDURE Min (a, b: CARDINAL) : CARDINAL ;
BEGIN
   IF a<b
   THEN
      RETURN( a )
   ELSE
      RETURN( b )
   END
END Min ;


(*
   GetNodeLeft - returns the Left branch of node, i.
*)

PROCEDURE GetNodeLeft (i: IndexToItem) : IndexToItem ;
BEGIN
   CASE ListOfItems[i].IsLeaf OF

   TRUE :  InternalError('expecting a node not a leaf', __FILE__, __LINE__) |
   FALSE:  IF IsBinaryNode(i)
           THEN
              RETURN( ListOfItems[i].ItemIsNode.Binary.Left )
           ELSIF IsConditionNode(i)
           THEN
              RETURN( ListOfItems[i].ItemIsNode.Condition.Left )
           ELSIF IsUnaryNode(i)
           THEN
              InternalError('unary nodes do not have a left branch', __FILE__, __LINE__)
           ELSE
              InternalError('unexpected node type', __FILE__, __LINE__)
           END

   ELSE
      InternalError('unexpected value in CASE', __FILE__, __LINE__)
   END
END GetNodeLeft ;


(*
   GetNodeRight - returns the Right branch of node, i.
*)

PROCEDURE GetNodeRight (i: IndexToItem) : IndexToItem ;
BEGIN
   CASE ListOfItems[i].IsLeaf OF

   TRUE :  InternalError('expecting a node not a leaf', __FILE__, __LINE__) |
   FALSE:  IF IsBinaryNode(i)
           THEN
              RETURN( ListOfItems[i].ItemIsNode.Binary.Right )
           ELSIF IsConditionNode(i)
           THEN
              RETURN( ListOfItems[i].ItemIsNode.Condition.Right )
           ELSIF IsUnaryNode(i)
           THEN
              RETURN( ListOfItems[i].ItemIsNode.Unary.Right )
           ELSE
              InternalError('unexpected node type', __FILE__, __LINE__)
           END

   ELSE
      InternalError('unexpected value in CASE', __FILE__, __LINE__)
   END
END GetNodeRight ;


(*
   GetNodeOp - returns the operator of node, i.
*)

PROCEDURE GetNodeOp (i: IndexToItem) : QuadOperator ;
BEGIN
   CASE ListOfItems[i].IsLeaf OF

   TRUE :  InternalError('expecting a node not a leaf', __FILE__, __LINE__) |
   FALSE:  RETURN( ListOfItems[i].ItemIsNode.Op )

   ELSE
      InternalError('unexpected value in CASE', __FILE__, __LINE__)
   END
END GetNodeOp ;


(*
   GetItemValue - returns the value for item, i.
*)

PROCEDURE GetItemValue (i: IndexToItem) : CARDINAL ;
BEGIN
   CASE ListOfItems[i].IsLeaf OF

   FALSE:  RETURN( ListOfItems[i].ItemIsNode.LabelValue ) |
   TRUE :  RETURN( ListOfItems[i].ItemIsLeaf.LabelValue )

   ELSE
      InternalError('unexpected value in CASE', __FILE__, __LINE__)
   END
END GetItemValue ;


(*
   IsItemLeaf - 
*)

PROCEDURE IsItemLeaf (i: IndexToItem) : BOOLEAN ;
BEGIN
   RETURN( ListOfItems[i].IsLeaf )
END IsItemLeaf ;


(*
   IsItemNode - 
*)

PROCEDURE IsItemNode (i: IndexToItem) : BOOLEAN ;
BEGIN
   RETURN( NOT ListOfItems[i].IsLeaf )
END IsItemNode ;


(*
   AssignValue - assigns a value, v, to a node or leaf, i.
*)

PROCEDURE AssignValue (i: IndexToItem; v: CARDINAL) ;
BEGIN
   WITH ListOfItems[i] DO
      CASE IsLeaf OF

      FALSE:  ItemIsNode.LabelValue := v |
      TRUE :  ItemIsLeaf.LabelValue := v

      ELSE
         InternalError('should never reach here', __FILE__, __LINE__)
      END
   END
END AssignValue ;


(*
   LabelTree - algorithm presented on P562 of the Dragon Book.
*)

PROCEDURE LabelTree (i: IndexToItem; IsLeft: BOOLEAN) : CARDINAL ;
VAR
   v: CARDINAL ;
BEGIN
   IF IsItemLeaf(i)
   THEN
      IF IsLeft OR IsRegister(GetDestination(i))
      THEN
         AssignValue(i, 1) ;
         RETURN( 1 )
      ELSE
         AssignValue(i, 0) ;
         RETURN( 0 )
      END
   ELSE
      (* i is an interior node *)
      v := CalculateChildren(i) ;
      AssignValue(i, v) ;
      RETURN( v )
   END
END LabelTree ;


(*
   NoOfRegsOpNeeds - returns the number of registers that a node operation requires.
*)

PROCEDURE NoOfRegsOpNeeds (i: IndexToItem) : CARDINAL ;
BEGIN
   CASE GetNodeOp(i) OF

   DivOp,
   ModOp    :  RETURN( 2 ) |      (* eax and edx *)
   BecomesOp:  IF GetDestination(GetNodeRight(i))=ESI
               THEN 
                  RETURN( 3 )
               ELSE
                  RETURN( ((GetSize(GetDestination(i)))+WordSize) DIV WordSize )
               END |
   IndrXOp  :  IF GetDestination(GetNodeRight(i))=ESI
               THEN
                  RETURN( 3 )
               ELSE
                  RETURN( ((GetSize(GetDestination(GetNodeLeft(i))))+WordSize) DIV WordSize )
               END |
   XIndrOp  :  IF GetDestination(GetNodeRight(i))=ESI
               THEN
                  RETURN( 3 )
               ELSE
                  RETURN( ((GetSize(GetDestination(i)))+WordSize) DIV WordSize )
               END

   ELSE
      RETURN( 1 )
   END
END NoOfRegsOpNeeds ;


(*
   CalculateBinaryLabelValue - 
*)

PROCEDURE CalculateBinaryLabelValue (i: IndexToItem) : CARDINAL ;
VAR
   NoOfRegs,
   lv, rv  : CARDINAL ;
BEGIN
   Assert(NOT ListOfItems[i].IsLeaf) ;
   WITH ListOfItems[i].ItemIsNode.Binary DO
      NoOfRegs := NoOfRegsOpNeeds(i) ;
      lv := LabelTree(Left, TRUE) ;
      CASE GetNodeOp(i) OF

      DivOp,
      ModOp      : rv := LabelTree(Right, TRUE)      (* no constant operands optimizations with div/mod *)

      ELSE
         rv := LabelTree(Right, FALSE)
      END ;
      IF Right=Left
      THEN
         AssignValue(i, lv+NoOfRegs) ;
         RETURN( lv+NoOfRegs )
      ELSE
         AssignValue(i, Max(lv, rv)+NoOfRegs) ;
         RETURN( Max(lv, rv)+NoOfRegs )
      END
   END
END CalculateBinaryLabelValue ;


(*
   CalculateUnaryLabelValue - 
*)

PROCEDURE CalculateUnaryLabelValue (i: IndexToItem) : CARDINAL ;
BEGIN
   Assert(NOT ListOfItems[i].IsLeaf) ;
   WITH ListOfItems[i].ItemIsNode.Unary DO
      IF (GetNodeOp(i)=BecomesOp) AND
         IsConstString(GetDestination(Right)) AND (GetStringLength(GetDestination(Right))>1)
      THEN
         RETURN( 0 )    (* can move constant strings into memory without any registers *)
      ELSE
         RETURN( LabelTree(Right, TRUE) )
      END
   END
END CalculateUnaryLabelValue ;


(*
   CalculateConditionLabelValue - 
*)

PROCEDURE CalculateConditionLabelValue (i: IndexToItem) : CARDINAL ;
VAR
   lv, rv: CARDINAL ;
BEGIN
   Assert(NOT ListOfItems[i].IsLeaf) ;
   WITH ListOfItems[i].ItemIsNode.Condition DO
      lv := LabelTree(Right, FALSE) ;
      rv := LabelTree(Left, TRUE) ;
      IF Right=Left
      THEN
         AssignValue(i, lv+1) ;
         RETURN( lv+1 )
      ELSE
         AssignValue(i, Max(lv, rv)) ;
         RETURN( Max(lv, rv) )
      END
   END
END CalculateConditionLabelValue ;


(*
   CalculateDummyLabelValue - returns the label value for node, i.
*)

PROCEDURE CalculateDummyLabelValue (i: IndexToItem) : CARDINAL ;
VAR
   tv1, tv2: CARDINAL ;
BEGIN
   Assert(IsDummyNode(i)) ;
   WITH ListOfItems[i].ItemIsNode.Dummy DO
      tv1 := LabelTree(Tree1, FALSE) ;
      tv2 := LabelTree(Tree2, FALSE) ;
      AssignValue(i, Max(tv1, tv2)) ;
      RETURN( Max(tv1, tv2) )
   END
END CalculateDummyLabelValue ;


(*
   CalculateChildren - calculates the label value for all of node, i, children.
*)

PROCEDURE CalculateChildren (i: IndexToItem) : CARDINAL ;
BEGIN
   Assert(NOT ListOfItems[i].IsLeaf) ;
   WITH ListOfItems[i].ItemIsNode DO
      CASE Op OF

      DummyOp      : RETURN( CalculateDummyLabelValue(i) )

      ELSE
         IF IsBinaryNode(i)
         THEN
            RETURN( CalculateBinaryLabelValue(i) )
         ELSIF IsUnaryNode(i)
         THEN
            RETURN( CalculateUnaryLabelValue(i) )
         ELSIF IsConditionNode(i)
         THEN
            RETURN( CalculateConditionLabelValue(i) )
         ELSE
            InternalError('unexpected operator', __FILE__, __LINE__)
         END
      END
   END
END CalculateChildren ;


(*
   WriteNSpaces - writes, n, spaces.
*)

PROCEDURE WriteNSpaces (n: CARDINAL) ;
BEGIN
   WriteString('# ') ;
   WHILE n>0 DO
      Write(' ') ;
      DEC(n)
   END
END WriteNSpaces ;


(*
   DisplayTreeIndent - displays the subtree with count spaces.
*)

PROCEDURE DisplayTreeIndent (i: IndexToItem; count: CARDINAL) ;
BEGIN
   IF i>0
   THEN
      WITH ListOfItems[i] DO
         CASE IsLeaf OF

         FALSE:  DisplayNode(i, count) |
         TRUE :  DisplayLeaf(i, count)

         ELSE
            InternalError('should never reach here', __FILE__, __LINE__)
         END
      END
   END
END DisplayTreeIndent ;


(*
   DisplayTree - 
*)

PROCEDURE DisplayTree (i: IndexToItem) ;
BEGIN
   DisplayTreeIndent(i, 0)
END DisplayTree ;


(*
   DisplayNode - displays node, i.
*)

PROCEDURE DisplayNode (i: IndexToItem; count: CARDINAL) ;
BEGIN
   WriteLn ;
   WriteNSpaces(count) ;
   INC(count, Indent) ;
   Assert(NOT ListOfItems[i].IsLeaf) ;
   WITH ListOfItems[i].ItemIsNode DO
      WriteString(' [') ; WriteCard(CARDINAL(i), 0) ;
      WriteString(']  { (') ; WriteCard(LabelValue, 0) ; Write(')') ;
      (* special cases *)
      CASE Op OF

      XIndrOp      : WITH Binary DO
                        WriteKey(GetSymName(GetDestination(i))) ; WriteString(' <- ') ; WriteOperator(Op) ; DisplayTreeIndent(Left, count) ; DisplayTreeIndent(Right, count)
                     END |

      IndrXOp      : WITH Binary DO
                        WriteSymbol(i) ; WriteString(' <- ') ; WriteOperator(Op) ; WriteKey(GetSymName(GetDestination(Left))) ; DisplayTreeIndent(Right, count)
                     END |

      DummyOp      : WITH Dummy DO
                        WriteSymbol(i) ; DisplayTreeIndent(Tree1, count) ; DisplayTreeIndent(Tree2, count) ;
                     END |

      UnboundedOp  : WITH Binary DO
                        WriteOperator(Op) ; DisplayTreeIndent(Left, count) ; DisplayTreeIndent(Right, count)
                     END

      ELSE
         (* general cases now *)
         IF IsBinaryNode(i)
         THEN
            WITH Binary DO
               WriteSymbol(i) ; WriteString(' <- ') ; WriteOperator(Op) ; DisplayTreeIndent(Left, count) ; DisplayTreeIndent(Right, count)
            END
         ELSIF IsUnaryNode(i)
         THEN
            WITH Unary DO
               IF GetDestination(i)=NulSym
               THEN
                  Assert(GetNodeOp(i)=CallOp)       (* indirect procedure call *)
               ELSIF IsVar(GetDestination(i))
               THEN
                  WriteSymbol(i)
               ELSE
                  WriteKey(GetSymName(GetDestination(i)))
               END ;
               WriteOperator(Op) ; DisplayTreeIndent(Right, count)
            END
         ELSIF IsConditionNode(i)
         THEN
            WITH Condition DO
               WriteCard(DestQuad, 6) ; WriteOperator(Op) ; DisplayTreeIndent(Left, count) ; DisplayTreeIndent(Right, count)
            END
         ELSE
            InternalError('unexpected operator', __FILE__, __LINE__)
         END
      END ;
      WriteString(' }, ')
   END
END DisplayNode ;


(*
   DisplayLeaf - displays the leaf, i.
*)

PROCEDURE DisplayLeaf (i: IndexToItem; count: CARDINAL) ;
BEGIN
   WriteLn ;
   WriteNSpaces(count) ;
   Assert(ListOfItems[i].IsLeaf) ;
   WITH ListOfItems[i].ItemIsLeaf DO
      WriteString(' [') ; WriteCard(CARDINAL(i), 0) ; WriteString('] ') ;
      WriteSymbol(i) ; Write('(') ; WriteCard(LabelValue, 0) ; WriteString('), ')
   END
END DisplayLeaf ;


(*
   NewNode - returns the next free node.
*)

PROCEDURE NewNode () : IndexToItem ;
VAR
   n: IndexToItem ;
BEGIN
   IF ItemPtr=MaxNoOfItems
   THEN
      InternalError('exceeded maximum no of nodes (increase MaxNoOfItems)', __FILE__, __LINE__)
   ELSE
      n := ItemPtr ;
      INC(ItemPtr) ;
      RETURN( n )
   END
END NewNode ;


(*
   NewLeaf - returns the next free leaf.
*)

PROCEDURE NewLeaf () : IndexToItem ;
VAR
   n: IndexToItem ;
BEGIN
   IF ItemPtr=MaxNoOfItems
   THEN
      InternalError('exceeded maximum no of leaves (increase MaxNoOfItems)', __FILE__, __LINE__)
   ELSE
      n := ItemPtr ;
      INC(ItemPtr) ;
      RETURN( n )
   END
END NewLeaf ;


(*
   PutVarParamNode - 
*)

PROCEDURE PutVarParamNode (i: IndexToItem; isvarparam: BOOLEAN) ;
BEGIN
   Assert(IsItemNode(i)) ;
   ListOfItems[i].ItemIsNode.IsVarParam := isvarparam
END PutVarParamNode ;


(*
   IsVarParamNode - 
*)

PROCEDURE IsVarParamNode (i: IndexToItem) : BOOLEAN ;
BEGIN
   Assert(IsItemNode(i)) ;
   RETURN( ListOfItems[i].ItemIsNode.IsVarParam )
END IsVarParamNode ;


(*
   PutSignedNode - makes a node, signed.
*)

PROCEDURE PutSignedNode (i: IndexToItem; signed: BOOLEAN) ;
BEGIN
   Assert(IsItemNode(i)) ;
   ListOfItems[i].ItemIsNode.IsSigned := signed
END PutSignedNode ;


(*
   IsSignedNode - returns TRUE if the node, i, is signed.
*)

PROCEDURE IsSignedNode (i: IndexToItem) : BOOLEAN ;
BEGIN
   Assert(IsItemNode(i)) ;
   RETURN( ListOfItems[i].ItemIsNode.IsSigned )
END IsSignedNode ;


(*
   GetDestinationLeaf - returns the leaf symbol.
*)

PROCEDURE GetDestinationLeaf (i: IndexToItem) : CARDINAL ;
BEGIN
   Assert(ListOfItems[i].IsLeaf) ;
   RETURN( ListOfItems[i].ItemIsLeaf.DestSym.Symbol )
END GetDestinationLeaf ;


(*
   GetDestinationNode - returns the destination symbol from a node, i.
*)

PROCEDURE GetDestinationNode (i: IndexToItem) : CARDINAL ;
BEGIN
   Assert(NOT ListOfItems[i].IsLeaf) ;
   WITH ListOfItems[i].ItemIsNode DO
      CASE Op OF

      (* conditional operators have no destination symbol *)
      IfInOp,
      IfNotInOp,
      IfEquOp,
      IfNotEquOp,
      IfGreEquOp,
      IfLessEquOp,
      IfLessOp,
      IfGreOp            : RETURN( NulSym ) |

      (* arithmetic operators *)
      LogicalShiftLeftOp,
      LogicalShiftRightOp,
      LogicalOrOp,
      LogicalAndOp,
      LogicalXorOp,
      DivOp,
      ModOp,
      AddOp,
      MultOp,
      SubOp,
      UnboundedOp        : RETURN( Binary.DestSym.Symbol ) |

      BitOp,
      CallOp,
      NegateOp,
      ConvertOp,
      XIndrOp,
      IndrXOp,
      ParamOp,
      FunctValueOp,
      ReturnValueOp,
      AddrOp,
      BecomesOp          : RETURN( Unary.DestSym.Symbol ) |

      DummyOp            : RETURN( Dummy.DestSym.Symbol )

      ELSE
         InternalError('unexpected operator in node', __FILE__, __LINE__)
      END
   END
END GetDestinationNode ;


(*
   GetDestination - returns the destination symbol for this leaf or node.
*)

PROCEDURE GetDestination (i: IndexToItem) : CARDINAL ;
BEGIN
   WITH ListOfItems[i] DO
      CASE IsLeaf OF

      FALSE: RETURN( GetDestinationNode(i) ) |
      TRUE : RETURN( GetDestinationLeaf(i) )

      ELSE
         InternalError('should never reach this clause', __FILE__, __LINE__)
      END
   END
END GetDestination ;


(*
   GetPrequisiteRegister - 
*)

PROCEDURE GetPrequisiteRegister (i: IndexToItem) : CARDINAL ;
BEGIN
   IF IsItemLeaf(i)
   THEN
      RETURN( ListOfItems[i].ItemIsLeaf.DestSym.PrerequisiteRegister )
   ELSIF IsItemNode(i)
   THEN
      IF IsBinaryNode(i)
      THEN
         RETURN( ListOfItems[i].ItemIsNode.Binary.DestSym.PrerequisiteRegister )
      ELSIF IsUnaryNode(i)
      THEN
         RETURN( ListOfItems[i].ItemIsNode.Unary.DestSym.PrerequisiteRegister )
      END
   END ;
   RETURN( NulSym )   (* any register will do *)
END GetPrequisiteRegister ;


(*
   PutPrerequisiteRegister - 
*)

PROCEDURE PutPrerequisiteRegister (i: IndexToItem; reg: CARDINAL) ;
BEGIN
   IF IsItemLeaf(i)
   THEN
      ListOfItems[i].ItemIsLeaf.DestSym.PrerequisiteRegister := reg
   ELSIF IsItemNode(i)
   THEN
      IF IsBinaryNode(i)
      THEN
         ListOfItems[i].ItemIsNode.Binary.DestSym.PrerequisiteRegister := reg
      ELSIF IsUnaryNode(i)
      THEN
         ListOfItems[i].ItemIsNode.Unary.DestSym.PrerequisiteRegister := reg
      ELSE
         InternalError('node type not expected', __FILE__, __LINE__)
      END
   ELSE
      InternalError('node type not expected', __FILE__, __LINE__)
   END
END PutPrerequisiteRegister ;


(*
   PutRegisterUsed - 
*)

PROCEDURE PutRegisterUsed (i: IndexToItem; IsUsed: BOOLEAN) ;
BEGIN
   IF IsItemLeaf(i)
   THEN
      ListOfItems[i].ItemIsLeaf.DestSym.IsRegisterUsed := IsUsed
   ELSIF IsItemNode(i)
   THEN
      IF IsBinaryNode(i)
      THEN
         ListOfItems[i].ItemIsNode.Binary.DestSym.IsRegisterUsed := IsUsed
      ELSIF IsUnaryNode(i)
      THEN
         ListOfItems[i].ItemIsNode.Unary.DestSym.IsRegisterUsed := IsUsed
      ELSE
         InternalError('node type not expected', __FILE__, __LINE__)
      END
   ELSE
      InternalError('node type not expected', __FILE__, __LINE__)
   END
END PutRegisterUsed ;


(*
   GetRegisterUsed - 
*)

PROCEDURE GetRegisterUsed (i: IndexToItem) : BOOLEAN ;
BEGIN
   IF IsItemLeaf(i)
   THEN
      RETURN( ListOfItems[i].ItemIsLeaf.DestSym.IsRegisterUsed )
   ELSIF IsItemNode(i)
   THEN
      IF IsBinaryNode(i)
      THEN
         RETURN( ListOfItems[i].ItemIsNode.Binary.DestSym.IsRegisterUsed )
      ELSIF IsUnaryNode(i)
      THEN
         RETURN( ListOfItems[i].ItemIsNode.Unary.DestSym.IsRegisterUsed )
      ELSE
         InternalError('node type not expected', __FILE__, __LINE__)
      END
   ELSE
      InternalError('node type not expected', __FILE__, __LINE__)
   END
END GetRegisterUsed ;


(*
   PutDestination - give node, i, the destination symbol, sym.
*)

PROCEDURE PutDestination (i: IndexToItem; sym: CARDINAL) ;
BEGIN
   IF IsItemLeaf(i)
   THEN
      ListOfItems[i].ItemIsLeaf.DestSym.Symbol := sym
   ELSIF IsItemNode(i)
   THEN
      IF IsBinaryNode(i)
      THEN
         ListOfItems[i].ItemIsNode.Binary.DestSym.Symbol := sym
      ELSIF IsUnaryNode(i)
      THEN
         ListOfItems[i].ItemIsNode.Unary.DestSym.Symbol := sym
      ELSIF IsDummyNode(i)
      THEN
         ListOfItems[i].ItemIsNode.Dummy.DestSym.Symbol := sym
      ELSE
         InternalError('node type not expected', __FILE__, __LINE__)
      END
   ELSE
      InternalError('node type not expected', __FILE__, __LINE__)
   END
END PutDestination ;


(*
   InitSymbol - initializes a SymbolInfo record
*)

PROCEDURE InitSymbol (VAR s: SymbolInfo) ;
BEGIN
   WITH s DO
      Symbol               := NulSym ;
      PrerequisiteRegister := NulSym ;
      IsRegisterUsed       := FALSE
   END
END InitSymbol ;


(*
   InitDestination - give node, i, the destination symbol, sym.
*)

PROCEDURE InitDestination (i: IndexToItem; sym: CARDINAL) ;
BEGIN
   IF IsItemLeaf(i)
   THEN
      InitSymbol(ListOfItems[i].ItemIsLeaf.DestSym)
   ELSIF IsItemNode(i)
   THEN
      IF IsBinaryNode(i)
      THEN
         InitSymbol(ListOfItems[i].ItemIsNode.Binary.DestSym)
      ELSIF IsUnaryNode(i)
      THEN
         InitSymbol(ListOfItems[i].ItemIsNode.Unary.DestSym)
      ELSIF IsDummyNode(i)
      THEN
         InitSymbol(ListOfItems[i].ItemIsNode.Dummy.DestSym)
      ELSE
         InternalError('node type not expected', __FILE__, __LINE__)
      END
   ELSE
      InternalError('node type not expected', __FILE__, __LINE__)
   END ;
   PutDestination(i, sym)
END InitDestination ;


(*
   GetNodeDatum - returns the size of operation at node, i.
*)

PROCEDURE GetNodeDatum (i: IndexToItem) : DatumPrefix ;
BEGIN
   Assert(NOT ListOfItems[i].IsLeaf) ;
   WITH ListOfItems[i].ItemIsNode DO
      IF IsBinaryNode(i) OR IsUnaryNode(i)
      THEN
         RETURN( GetUnitPrefix(GetSize(GetDestination(i))) )
      ELSIF IsConditionNode(i)
      THEN
         RETURN( FindDatum2(GetDestination(Condition.Left), GetDestination(Condition.Right)) )
      ELSE
         InternalError('unexpected type of node', __FILE__, __LINE__)
      END
   END
END GetNodeDatum ;


(*
   MakeNewLeaf - returns a new leaf containing symbol, Sym.
*)

PROCEDURE MakeNewLeaf (Sym: CARDINAL) : IndexToItem ;
VAR
   l: IndexToItem ;
BEGIN
   l := ItemPtr ;
   WITH ListOfItems[l] DO
      IsRef  := FALSE ;
      IsLeaf := TRUE ;
      WITH ItemIsLeaf DO
         InitDestination(l, Sym) ;
         LabelValue := 0
      END
   END ;
   INC(ItemPtr) ;
   RETURN( l )
END MakeNewLeaf ;


(*
   FindUnreferencedItem - returns the node or leaf where the symbol, Sym, was last a Destination.
*)

PROCEDURE FindUnreferencedItem (Sym: CARDINAL) : IndexToItem ;
VAR
   i: IndexToItem ;
BEGIN
   i := ItemPtr-1 ;   (* ItemPtr points to the next free (uninitialized node) *)
   WHILE i>0 DO
      IF (GetDestination(i)=Sym) AND (NOT ListOfItems[i].IsRef)
      THEN
         ListOfItems[i].IsRef := TRUE ;
         RETURN( i )
      END ;
      DEC(i)
   END ;
   i := MakeNewLeaf(Sym) ;
   ListOfItems[i].IsRef := TRUE ;
   RETURN( i )
END FindUnreferencedItem ;


(*
   FindLatestItem - returns the latest node or leaf where the symbol, Sym, was last a Destination.
*)

PROCEDURE FindLatestItem (Sym: CARDINAL) : IndexToItem ;
VAR
   i: IndexToItem ;
BEGIN
   i := ItemPtr-1 ;   (* ItemPtr points to the next free (uninitialized node) *)
   WHILE i>0 DO
      IF GetDestination(i)=Sym
      THEN
         RETURN( i )
      END ;
      DEC(i)
   END ;
   RETURN( MakeNewLeaf(Sym) )
END FindLatestItem ;


(*
   DisplayForest - 
*)

PROCEDURE DisplayForest ;
VAR
   i: IndexToTree ;
BEGIN
   i := 1 ;
   WHILE i<=TreePtr DO
      IF ListOfTrees[i]#0
      THEN
         DisplayTree(ListOfTrees[i])
      END ;
      INC(i)
   END ;
END DisplayForest ;


(*
   IsIndirectNode - returns TRUE if item, i, is a node which is an indirect operator.
*)

PROCEDURE IsIndirectNode (i: IndexToItem) : BOOLEAN ;
BEGIN
   RETURN( IsItemNode(i) AND ((GetNodeOp(i)=XIndrOp) OR (GetNodeOp(i)=IndrXOp)) )
END IsIndirectNode ;


(*
   IsIndirectNodeUsedInTree - returns TRUE if the tree, i, uses an indirect operator.
*)

PROCEDURE IsIndirectNodeUsedInTree (i: IndexToItem) : BOOLEAN ;
BEGIN
   IF IsBinaryNode(i) OR IsConditionNode(i)
   THEN
      RETURN(
             IsIndirectNode(i) OR
             IsIndirectNodeUsedInTree(GetNodeLeft(i)) OR
             IsIndirectNodeUsedInTree(GetNodeRight(i))
            )
   ELSIF IsUnaryNode(i)
   THEN
      RETURN( IsIndirectNodeUsedInTree(GetNodeRight(i)) )
   ELSE
      RETURN( FALSE )    (* must be a leaf *)
   END
END IsIndirectNodeUsedInTree ;


(*
   IsFunctNodeUsedInTree - returns TRUE if the tree, i, uses an funct node operator.
*)

PROCEDURE IsFunctNodeUsedInTree (i: IndexToItem) : BOOLEAN ;
BEGIN
   IF IsBinaryNode(i) OR IsConditionNode(i)
   THEN
      RETURN(
             IsFunctNodeUsedInTree(GetNodeLeft(i)) OR
             IsFunctNodeUsedInTree(GetNodeRight(i))
            )
   ELSIF IsUnaryNode(i)
   THEN
      RETURN(
             (GetNodeOp(i)=FunctValueOp) OR
             IsFunctNodeUsedInTree(GetNodeRight(i))
            )
   ELSE
      RETURN( FALSE )    (* must be a leaf *)
   END
END IsFunctNodeUsedInTree ;


(*
   IsDependentTree - returns TRUE if a destination symbol in tree, l, is a leaf symbol in tree, r.
*)

PROCEDURE IsDependentTree (l, r: IndexToItem) : BOOLEAN ;
BEGIN
   IF GetDestination(l)=GetDestination(r)
   THEN
      IF IsItemNode(l) OR IsItemNode(r)
      THEN
         WriteString('# clash with symbol ') ; WriteSymbol(l) ; WriteLn
      END ;
      RETURN( IsItemNode(l) OR IsItemNode(r) )
   ELSE
      IF IsBinaryNode(l) OR IsConditionNode(l)
      THEN
         RETURN( IsDependent(GetNodeLeft(l), r) OR IsDependent(GetNodeRight(l), r) )
      ELSIF IsUnaryNode(l)
      THEN
         RETURN( IsDependent(GetNodeRight(l), r) )
      ELSE
         RETURN( FALSE )    (* must be a leaf *)
      END
   END
END IsDependentTree ;


(*
   IsDependent - returns TRUE if a destination symbol in tree, l, is a leaf symbol in tree, r,
                 or visa versa.
*)

PROCEDURE IsDependent (l, r: IndexToItem) : BOOLEAN ;
BEGIN
   IF (l=NulNode) OR (r=NulNode)
   THEN
      RETURN( FALSE )
   ELSE
      RETURN( IsDependentTree(l, r) OR IsDependentTree(r, l) )
   END
END IsDependent ;


(*
   IsTreeMoveable - returns TRUE if tree, i, can be moved to, j.
*)

PROCEDURE IsTreeMoveable (i, j: IndexToTree) : BOOLEAN ;
VAR
   k      : IndexToTree ;
   EAXUsed: BOOLEAN ;
BEGIN
   EAXUsed := IsFunctNodeUsedInTree(ListOfTrees[i]) ;
   k := i ;
   IF i<j
   THEN
      INC(k)
   ELSIF i>j
   THEN
      DEC(k)
   END ;
   WHILE k#j DO
      IF (ListOfTrees[k]#NulNode) AND
         (IsDependent(ListOfTrees[k], ListOfTrees[i]) OR
          IsIndirectNodeUsedInTree(ListOfTrees[k]) OR
          (EAXUsed AND SubtreeRequires(ListOfTrees[k], EAX)))
      THEN
         IF DebugOn
         THEN
            WriteString('# not moving tree as ') ;
            IF IsDependent(ListOfTrees[k], ListOfTrees[i])
            THEN
               WriteString('dependant found')
            ELSIF IsIndirectNodeUsedInTree(ListOfTrees[k])
            THEN
               WriteString('indirect operator found')
            ELSIF EAXUsed AND SubtreeRequires(ListOfTrees[k], EAX)
            THEN
               WriteString('EAX used in subsequent tree')
            ELSE
               InternalError('inconsistant debugging statement', __FILE__, __LINE__)
            END ;
            WriteLn
         END ;
         RETURN( FALSE )
      END ;
      IF i<j
      THEN
         INC(k)
      ELSE
         DEC(k)
      END
   END ;
   RETURN( TRUE )
END IsTreeMoveable ;


(*
   GraftNewNodeOntoTree - grafts node, i, onto tree, j, and makes sure new leaves are used
                          instead of nodes for t1 and t2. (Must preserve ordering).
*)

PROCEDURE GraftNewNodeOntoTree (i: IndexToItem; j: IndexToTree; t1, t2: IndexToItem) ;
VAR
   newleaf: IndexToItem ;
BEGIN
   ListOfTrees[j] := i ;
   IF (t1#0) AND IsItemNode(t1)
   THEN
      (* we must use a different leaf symbol for, t1, to keep trees separate *)
      newleaf := MakeNewLeaf(GetDestination(t1)) ;
      ListOfItems[newleaf].IsRef := TRUE ;
      AlterIndex(i, t1, newleaf)
   END ;
   IF (t2#0) AND IsItemNode(t2)
   THEN
      (* we must use a different leaf symbol for, t2, to keep trees separate *)
      newleaf := MakeNewLeaf(GetDestination(t2)) ;
      ListOfItems[newleaf].IsRef := TRUE ;
      AlterIndex(i, t2, newleaf)
   END
END GraftNewNodeOntoTree ;


(*
   GraftNodeOntoTree - grafts node, i, onto tree, j, and checks for duplicate branch, t.
*)

PROCEDURE GraftNodeOntoTree (i: IndexToItem; j: IndexToTree;
                             t: IndexToItem; Dependent: BOOLEAN) ;
VAR
   k      : IndexToTree ;
   newleaf: IndexToItem ;
BEGIN
   ListOfTrees[j] := i ;
   IF (t#0) AND IsItemNode(t)
   THEN
      IF NOT Dependent
      THEN
         (*
            not yet dependent so we will check all trees between, t..ListOfTrees[j] to
            make sure that when we move, t, it does not violate any dependancy.
         *)
         k := j-1 ;
         WHILE (k#0) AND (ListOfTrees[k]#t) AND (NOT Dependent) DO
            IF (ListOfTrees[k]#0) AND IsDependent(ListOfTrees[k], t)
            THEN
               IF DebugOn
               THEN
                  WriteString('# found dependency when trying to graft branch') ; WriteLn
               END ;
               Dependent := TRUE
            END ;
            DEC(k)
         END
      END ;
      IF Dependent
      THEN
         (* we must use a different leaf symbol for, r, to keep trees separate *)
         newleaf := MakeNewLeaf(GetDestination(t)) ;
         ListOfItems[newleaf].IsRef := TRUE ;
         AlterIndex(i, t, newleaf)
      ELSE
         DEC(j) ;
         WHILE j>0 DO
            (* check to see that a previous tree defined by, r, is removed *)
            IF (ListOfTrees[j]#0) AND (ListOfTrees[j]=t) AND (GetNodeOp(ListOfTrees[j])#XIndrOp)
            THEN
               ListOfTrees[j] := 0 ;
               RETURN     (* nothing left to do *)
            END ;
            DEC(j) ;
         END
      END
   END
END GraftNodeOntoTree ;


(*
   OptimizeParameter - returns TRUE if we are able to optimize parameter node, i.
                       It checks backwards through all the trees to see whether
                       we can move a tree upto this new parameter. Parameter order
                       must be retained (of course) and also we must check
                       to see that dependants are not violated and no indirections
                       are violated.
*)

PROCEDURE OptimizeParameter (i, l: IndexToItem) : BOOLEAN ;
VAR
   k: IndexToTree ;
BEGIN
   k := TreePtr ;
   WHILE k>0 DO
      DEC(k) ;
      IF (ListOfTrees[k]#NulNode) AND (ListOfTrees[k]=l) AND IsTreeMoveable(k, TreePtr)
      THEN
         IF DebugOn
         THEN
            Comment('optimizing paramater')
         END ;
         IF TreePtr=MaxNoOfTrees
         THEN
            InternalError('increase MaxNoOfTrees', __FILE__, __LINE__)
         END ;
         INC(TreePtr) ;
         GraftNodeOntoTree(i, k, NulNode, FALSE) ;    (* parameters are unary therefore l is not dependant on r *)
         ListOfTrees[TreePtr] := ListOfTrees[k] ;     (* move tree into new position *)
         ListOfTrees[k] := NulNode ;
         RETURN( TRUE )
      END
   END ;
   RETURN( FALSE )
END OptimizeParameter ;


(*
   AddTree - given an IndexToItem, i, make sure that this node is known as a tree.
*)

PROCEDURE AddTree (i, l, r: IndexToItem) ;
VAR
   j        : IndexToTree ;
   Dependent: BOOLEAN ;
BEGIN
   Dependent := IsDependent(l, r) ;
   IF Dependent AND DebugOn
   THEN
      WriteString('# found left right dependency') ; WriteLn ;
      WriteString('# Left  = ') ; DisplayTree(l) ; WriteLn ;
      WriteString('# Right = ') ; DisplayTree(r) ; WriteLn ;
      WriteString('# =========================================') ; WriteLn
   END ;
   (*
      we must preserve ordering of the following nodes:
         -   parameter ordering, so if we are adding a ParamOp we must not
             add it before a previous ParamOp.
         -   conditional branch must be at the head of the final tree.
         -   procedure calls must be at the head of the final tree.
             (indirect procedure calls might not naturally fall there, but
              must be forced)
         -   a tree with an indirect operator must not be moved, but can have previous
             trees added under it.

      note these rules are added inside the loop, as they are easier to maintain.
      we could simplify the expressions via De Morgan's law but it would be difficult
      to maintain and add new rules...
   *)
   j := TreePtr ;
   (* do not care about dependence on the last tree, as order is preserved *)
   WHILE (j>=1) AND ((NOT Dependent) OR (j=TreePtr)) DO
      IF ListOfTrees[j]#NulNode
      THEN
         (* check to make sure that tree, j, is not going to clash with r, and, l. *)
         IF ((r#ListOfTrees[j]) AND IsDependent(r, ListOfTrees[j])) OR
            ((l#ListOfTrees[j]) AND IsDependent(l, ListOfTrees[j]))
         THEN
            (* yes this earlier tree does clash *)
            IF DebugOn
            THEN
               Comment('yes this earlier tree does clash, we must not add to an earlier tree')
            END ;
            j := 1
         ELSIF (GetNodeOp(i)=CallOp) AND (j<TreePtr)
         THEN
            (* a call (indirect and direct) *must* be at the end of the forest *)
            j := 1
         ELSIF IsConditionNode(i) AND (j<TreePtr)
         THEN
            (* only allowed to add condition nodes at end of forest - exit loop *)
            j := 1
         ELSIF (GetNodeOp(i)=ReturnValueOp) AND (j<TreePtr)
         THEN
            (* funct values operators must be at the end of the forest - exit loop *)
            j := 1
         ELSIF (GetNodeOp(ListOfTrees[j])=ParamOp) AND (GetNodeOp(i)=ParamOp)
         THEN
            (* we have now found a ParamOp at, j, therefore we must not try and add, i, at this node
               or before. Abort trying to combine, i, with a tree <=j.

               However if there are no dependants in trees j-1..1 then we can consider moving
               a tree <j up to the position, i. Providing no tree j+1..i-1 uses anything calculated
               in j.

            *)
            IF OptimizeParameter(i, l)
            THEN
               i := 0
            END ;
            j := 1
         ELSIF (j#TreePtr) AND IsIndirectNodeUsedInTree(ListOfTrees[j])
         THEN
            (*
               j, is not the last tree and indirection occuring in tree, j,
               thus we cannot move this tree.
            *)
            j := 1
         ELSIF (GetNodeOp(ListOfTrees[j])#XIndrOp) AND (GetNodeOp(i)#AddrOp)
         THEN
            (* XIndrOp have a type as their destination, thus must not try and combine such trees *)
(* was
            IF (l=ListOfTrees[j]) AND (l#0) AND (r#0) AND (NOT IsIndirectNodeUsedInTree(r))
            THEN
               GraftNodeOntoTree(i, j, r, Dependent) ;
               i := 0
            ELSIF (r=ListOfTrees[j]) AND (r#0) AND (l#0) AND (NOT IsIndirectNodeUsedInTree(l))
            THEN
               GraftNodeOntoTree(i, j, l, Dependent) ;
               i := 0
            END
*)
            IF (l=ListOfTrees[j]) AND (l#0) AND ((r=0) OR (NOT IsIndirectNodeUsedInTree(r)))
            THEN
               GraftNodeOntoTree(i, j, r, Dependent) ;
               i := 0 ;
               j := 1
            ELSIF (r=ListOfTrees[j]) AND (r#0) AND ((l=0) OR (NOT IsIndirectNodeUsedInTree(l)))
            THEN
               GraftNodeOntoTree(i, j, l, Dependent) ;
               i := 0 ;
               j := 1
            END
         END
      END ;
      DEC(j)
   END ;
   IF i#0
   THEN
      (* no tree found which contains, i, so (for now we create a new tree) *)
      IF FALSE (* Optimize586 AND IsMutuallyExclusive(i) *)
      THEN
         (*
            with the 586 we desire parallel computation of trees and subexpressions.
            So here we add the node to the last tree using a DummyOp
         *)
      ELSE
         IF TreePtr=MaxNoOfTrees
         THEN
            InternalError('increase MaxNoOfTrees', __FILE__, __LINE__)
         END ;
         INC(TreePtr) ;
         GraftNewNodeOntoTree(i, TreePtr, r, l)
      END
   END ;
   IF DebugOn
   THEN
      Comment('forest now looks like this') ;
      DisplayForest ;
      Comment('end of forest')
   END
END AddTree ;


(*
   FindDatum2 - returns the datum prefix for an operation.
*)

PROCEDURE FindDatum2 (op1, op2: CARDINAL) : DatumPrefix ;
BEGIN
   IF IsConst(op1) AND IsConst(op2)
   THEN
      InternalError('improve Should have folded constants', __FILE__, __LINE__) ;
      RETURN( MaxDatum() )    (* may if we want to remove the internal error *)
   ELSE
      IF IsConst(op1)
      THEN
         RETURN( GetUnitPrefix(GetSize(op2)) )
      ELSE
         RETURN( GetUnitPrefix(GetSize(op1)) )
      END
   END
END FindDatum2 ;


(*
   BuildConditionNode - adds a new condition node to the tree.
*)

PROCEDURE BuildConditionNode (q: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
VAR
   l, r, n: IndexToItem ;
BEGIN
   IF NOT EqualSizes(op1, op2)
   THEN
      WriteError('size of data must be the same for comparison')
   END ;
   (* Testing for size - word size *)
   IF (IsConst(op1) OR FitsIntoRegister(op1)) AND
      (IsConst(op2) OR FitsIntoRegister(op2))
   THEN
      l := FindUnreferencedItem(op1) ;
      r := FindUnreferencedItem(op2) ;
      n := NewNode() ;
      WITH ListOfItems[n] DO
         IsLeaf := FALSE ;
         IsRef  := FALSE ;
         WITH ItemIsNode DO
            LabelValue      := 0 ;
            QuadNo          := q ;
            LineNo          := 0 ;   (* no line number yet *)
            IsSigned        := FALSE ;
            IsVarParam      := FALSE ;
            Op              := op ;
            WITH Condition DO
               DestQuad := op3 ;         
               Left     := l ;
               Right    := r ;
               AddTree(n, Left, Right)       (* last node to be created will be the head of the tree *)
            END
         END
      END ;
      LastNode := n
   END
END BuildConditionNode ;


(*
   BuildCondition - 
*)

PROCEDURE BuildCondition (q: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsMathType(GetType(op1)) OR IsMathType(GetType(op2))
   THEN
      CASE op OF

      IfEquOp    :   CoprocessorIfEqu(op1, op2) ;
                     CreateJump(je, op3) |
      IfNotEquOp :   CoprocessorIfNotEqu(op1, op2) ;
                     CreateJump(jne, op3) |
      IfLessOp   :   CoprocessorIfLess(op1, op2) ;
                     CreateJump(jae, op3) |
      IfLessEquOp:   CoprocessorIfLessEqu(op1, op2) ;
                     CreateJump(jne, op3) |
      IfGreOp    :   CoprocessorIfGre(op1, op2) ;
                     CreateJump(jne, op3) |
      IfGreEquOp :   CoprocessorIfGreEqu(op1, op2) ;
                     CreateJump(jne, op3)

      ELSE
         InternalError('unknown condition found', __FILE__, __LINE__)
      END
   ELSE
      BuildConditionNode(q, op, op1, op2, op3)
   END
END BuildCondition ;


(*
   CheckFoldAdd - 
*)

PROCEDURE CheckFoldAdd (n, l, r: IndexToItem; Start, End: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsItemLeaf(l) AND IsConst(GetDestination(l))
   THEN
      IF IsItemLeaf(r) AND IsConst(GetDestination(r))
      THEN
         IF DebugOn
         THEN
            Comment('folding add of two leaves')
         END ;
         PushValue(GetDestination(l)) ;
         PushValue(GetDestination(r)) ;
         Add ;
         AlterToUnaryNode(n, BecomesOp, MakeNewLeaf(MakeNewConstFromValue())) ;
         RETURN( TRUE )
      ELSIF IsItemNode(r) AND
            (NOT IsUsedLaterOn(GetDestination(r), r)) AND
            (NOT IsSymNeeded(GetDestination(r), Start, End))
      THEN
         IF GetNodeOp(r)=AddOp
         THEN
            IF IsConst(GetDestination(GetNodeLeft(r)))
            THEN
               IF DebugOn
               THEN
                  Comment('folding add of two leaves on different branches') ;
                  WriteString('# ') ; WriteSymbol(GetNodeLeft(r)) ; WriteString(' + ') ; WriteSymbol(l) ; WriteLn
               END ;
               PushValue(GetDestination(GetNodeLeft(r))) ;
               PushValue(GetDestination(l)) ;
               Add ;
               AlterToUnaryNode(r, BecomesOp, GetNodeRight(r)) ;
               ListOfItems[l].ItemIsLeaf.DestSym.Symbol := MakeNewConstFromValue() ;
               RETURN( TRUE )
            ELSIF IsConst(GetDestination(GetNodeRight(r)))
            THEN
               IF DebugOn
               THEN
                  Comment('folding add of two leaves on different branches') ;
                  WriteString('# ') ; WriteSymbol(GetNodeRight(r)) ; WriteString(' + ') ; WriteSymbol(l) ; WriteLn
               END ;
               PushValue(GetDestination(GetNodeRight(r))) ;
               PushValue(GetDestination(l)) ;
               Add ;
               AlterToUnaryNode(r, BecomesOp, GetNodeLeft(r)) ;
               ListOfItems[l].ItemIsLeaf.DestSym.Symbol := MakeNewConstFromValue() ;
               RETURN( TRUE )
            END
         END
      END
   END ;
   RETURN( FALSE )
END CheckFoldAdd ;


(*
   AlterIndex - alters node parent index from index, from, to index, to.
*)

PROCEDURE AlterIndex (parent, from, to: IndexToItem) ;
BEGIN
   Assert(IsItemNode(parent)) ;
   IF IsBinaryNode(parent)
   THEN
      IF GetNodeRight(parent)=from
      THEN
         ListOfItems[parent].ItemIsNode.Binary.Right := to
      ELSIF GetNodeLeft(parent)=from
      THEN
         ListOfItems[parent].ItemIsNode.Binary.Left := to
      ELSE
         InternalError('unexpected index', __FILE__, __LINE__)
      END
   ELSIF IsUnaryNode(parent)
   THEN
      IF GetNodeRight(parent)=from
      THEN
         ListOfItems[parent].ItemIsNode.Unary.Right := to
      ELSE
         InternalError('unexpected index', __FILE__, __LINE__)
      END
   ELSIF IsConditionNode(parent)
   THEN
      IF GetNodeRight(parent)=from
      THEN
         ListOfItems[parent].ItemIsNode.Condition.Right := to
      ELSIF GetNodeLeft(parent)=from
      THEN
         ListOfItems[parent].ItemIsNode.Condition.Left := to
      ELSE
         InternalError('unexpected index', __FILE__, __LINE__)
      END
   ELSE
      InternalError('unexpected type of node', __FILE__, __LINE__)
   END
END AlterIndex ;


(*
   CheckFoldBecomes - remove BecomesOp nodes which are not needed. We do this so that other
                      semi peephole optimizations (base register and offset optimizations)
                      can work in more cases.
*)

PROCEDURE CheckFoldBecomes (parent, n: IndexToItem; Start, End: CARDINAL) : BOOLEAN ;
VAR
   Sym: CARDINAL ;
BEGIN
   IF IsItemNode(n) AND (GetNodeOp(n)=BecomesOp) AND
      (NOT IsUsedLaterOn(GetDestination(n), n)) AND
      (NOT IsSymNeeded(GetDestination(n), Start, End))
   THEN
      Sym := GetDestination(GetNodeRight(n)) ;
      IF (IsConst(Sym) AND ((NOT IsConstString(Sym)) OR (GetStringLength(Sym)=1))) OR
         (IsVar(Sym) AND (GetSize(Sym)=GetSize(GetDestination(n))))
      THEN
         IF DebugOn
         THEN
            Comment('we can remove a redundant becomes')
         END ;
         (* now need to alter parent reference to the right hand side of becomes *)
         AlterIndex(parent, n, GetNodeRight(n)) ;
         RETURN( TRUE )
      END
   END ;
   RETURN( FALSE )
END CheckFoldBecomes ;


(*
   CheckFoldTree - scans the subtree for constants which can be folded.
*)

PROCEDURE CheckFoldTree (Head, Start, End: CARDINAL; i: IndexToItem) : BOOLEAN ;
VAR
   folded: BOOLEAN ;
BEGIN
   folded := FALSE ;
   IF IsItemNode(i)
   THEN
      IF IsBinaryNode(i) OR IsConditionNode(i)
      THEN
         IF CheckFoldTree(Head, Start, End, GetNodeLeft(i))
         THEN
            folded := TRUE
         END ;
         IF CheckFoldTree(Head, Start, End, GetNodeRight(i))
         THEN
            folded := TRUE
         END ;
         IF CheckFoldNodes(i, GetNodeLeft(i), GetNodeRight(i), Start, End)
         THEN
            folded := TRUE
         END ;
         (* now it might have been changed because of the above optimizations *)
         IF (IsBinaryNode(i) OR IsConditionNode(i)) AND
            CheckFoldBecomes(i, GetNodeLeft(i), Start, End)
         THEN
            folded := TRUE
         END ;
         IF CheckFoldBecomes(i, GetNodeRight(i), Start, End)
         THEN
            folded := TRUE
         END
      ELSIF IsUnaryNode(i)
      THEN
         IF CheckFoldBecomes(i, GetNodeRight(i), Start, End)
         THEN
            folded := TRUE
         END ;
         IF CheckFoldTree(Head, Start, End, GetNodeRight(i))
         THEN
            folded := TRUE
         END
      END
   END ;
   RETURN( folded )
END CheckFoldTree ;


(*
   CheckFoldNodes - checks to see whether two constants can be folded
                    given a subtree:
                           n
                         /   \
                        l     r
*)

PROCEDURE CheckFoldNodes (n, l, r: IndexToItem; Start, End: CARDINAL) : BOOLEAN ;
VAR
   folded: BOOLEAN ;
BEGIN
   folded := FALSE ;
   IF GetNodeOp(n)=AddOp
   THEN
      folded := CheckFoldAdd(n, l, r, Start, End) ;   (* add is commutative *)
      (* we might have changed the node operator to becomes *)
      IF GetNodeOp(n)=AddOp
      THEN
         IF CheckFoldAdd(n, r, l, Start, End)
         THEN
            folded := TRUE
         END
      END
   END ;
   RETURN( folded )
END CheckFoldNodes ;


(*
   BuildSubNode - checks to see whether we should use the FPU code generator or gencode.
*)

PROCEDURE BuildSubNode (q: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsCoprocessorUsed(op1, op2, op3)
   THEN
      CoprocessorSub(op1, op2, op3)
   ELSE
      BuildBinaryNode(q, op, op1, op2, op3)
   END
END BuildSubNode ;


(*
   BuildAddNode - checks to see whether we should use the FPU code generator or gencode.
*)

PROCEDURE BuildAddNode (q: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsCoprocessorUsed(op1, op2, op3)
   THEN
      CoprocessorAdd(op1, op2, op3)
   ELSE
      BuildBinaryNode(q, op, op1, op2, op3)
   END
END BuildAddNode ;


(*
   BuildBinaryNode - adds a new binary node to the tree.
*)

PROCEDURE BuildBinaryNode (q: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
VAR
   l, r, n: IndexToItem ;
BEGIN
   l := FindUnreferencedItem(op2) ;
   r := FindUnreferencedItem(op3) ;
   n := NewNode() ;
   WITH ListOfItems[n] DO
      IsLeaf := FALSE ;
      IsRef  := FALSE ;
      WITH ItemIsNode DO
         LabelValue := 0 ;
         QuadNo     := q ;
         LineNo     := 0 ;   (* no line number yet *)
         IsSigned   := FALSE ;
         IsVarParam := FALSE ;
         Op         := op ;
         WITH Binary DO
            InitDestination(n, op1) ;
            Left    := l ;
            Right   := r ;
            AddTree(n, Left, Right)       (* last node to be created will be the head of the tree *)
         END
      END
   END ;
   LastNode := n
END BuildBinaryNode ;


(*
   NodeWillBeWritten - returns TRUE if node, i, destination symbol will be written to once processed.
*)

PROCEDURE NodeWillBeWritten (i: IndexToItem; Start, End: CARDINAL) : BOOLEAN ;
VAR
   WriteStart, WriteEnd: CARDINAL ;
BEGIN
   IF IsUsedLaterOn(GetDestination(i), i)
   THEN
      RETURN( TRUE )
   ELSE
      IF IsSymNeeded(GetDestination(i), Start, End)
      THEN
         GetVarWriteLimitQuads(GetDestination(i), GetNodeQuad(i), End, WriteStart, WriteEnd) ;
         RETURN( WriteEnd=GetNodeQuad(i) )
      ELSE
         RETURN( FALSE )
      END
   END
END NodeWillBeWritten ;


(*
   TryToFindOffset - looks for a + or - node beneth, i, which
                     has a constant.
*)

PROCEDURE TryToFindOffset (i: IndexToItem; Start, End: CARDINAL) : INTEGER ;
BEGIN
   IF IsItemNode(i) AND (NOT NodeWillBeWritten(i, Start, End))
   THEN
      IF GetNodeOp(i)=AddOp
      THEN
         IF IsConst(GetDestination(GetNodeRight(i)))
         THEN
            IF DebugOn
            THEN
               Comment('found offset')
            END ;
            PushValue(GetDestination(GetNodeRight(i))) ;
            AlterToUnaryNode(i, BecomesOp, GetNodeLeft(i)) ;
            RETURN( PopInt() )
         ELSIF IsConst(GetDestination(GetNodeLeft(i)))
         THEN
            IF DebugOn
            THEN
               Comment('found offset')
            END ;
            PushValue(GetDestination(GetNodeLeft(i))) ;
            AlterToUnaryNode(i, BecomesOp, GetNodeRight(i)) ;
            RETURN( PopInt() )
         END
      ELSIF GetNodeOp(i)=SubOp
      THEN
         IF IsConst(GetDestination(GetNodeRight(i)))
         THEN
            IF DebugOn
            THEN
               Comment('found offset')
            END ;
            PushCard(0) ;
            PushValue(GetDestination(GetNodeRight(i))) ;
            Sub ;
            AlterToUnaryNode(i, BecomesOp, GetNodeLeft(i)) ;
            RETURN( PopInt() )
         END
      END
   END ;
   RETURN( 0 )
END TryToFindOffset ;


(*
   TryToFindOffsetSymbol - looks for a + or - node beneth, i, which
                           exists [& globalsymbol]
*)

PROCEDURE TryToFindOffsetSymbol (i: IndexToItem; Start, End: CARDINAL) : INTEGER ;
VAR
   j  : IndexToItem ;
   Sym: CARDINAL ;
BEGIN
   IF IsItemNode(i) AND (NOT NodeWillBeWritten(i, Start, End))
   THEN
      IF GetNodeOp(i)=AddOp
      THEN
         j := GetNodeLeft(i) ;
         IF IsItemNode(j) AND (NOT NodeWillBeWritten(j, Start, End)) AND (GetNodeOp(j)=AddrOp) AND
            IsVar(GetDestination(GetNodeRight(j))) AND (NOT IsProcedure(GetVarFather(GetDestination(GetNodeRight(j)))))
         THEN
            (* trying to obtain the address of a global variable - a canditate for optimization *)
            Sym := GetDestination(GetNodeRight(j)) ;
            AlterToUnaryNode(i, BecomesOp, GetNodeRight(i)) ;
            IF DebugOn
            THEN
               Comment('found offset symbol')
            END ;
            Assert(Sym#NulSym) ;
            RETURN( Sym )
         END ;
         j := GetNodeRight(i) ;
         IF IsItemNode(j) AND (NOT NodeWillBeWritten(j, Start, End)) AND (GetNodeOp(j)=AddrOp) AND
            IsVar(GetDestination(GetNodeLeft(j))) AND (NOT IsProcedure(GetVarFather(GetDestination(GetNodeLeft(j)))))
         THEN
            (* trying to obtain the address of a global variable - a canditate for optimization *)
            Sym := GetDestination(GetNodeLeft(j)) ;
            AlterToUnaryNode(i, BecomesOp, GetNodeLeft(i)) ;
            IF DebugOn
            THEN
               Comment('found offset symbol')
            END ;
            Assert(Sym#NulSym) ;
            RETURN( Sym )
         END
      END
   END ;
   RETURN( NulSym )
END TryToFindOffsetSymbol ;


(* AlterToUnaryNode *)


(*
   BuildUnaryNode - adds a new unary node to the tree.
*)

PROCEDURE BuildUnaryNode (q: CARDINAL; op: QuadOperator; op1, op2: CARDINAL) ;
VAR
   n: IndexToItem ;
BEGIN
   n := NewNode() ;
   WITH ListOfItems[n] DO
      IsLeaf := FALSE ;
      IsRef  := FALSE ;
      WITH ItemIsNode DO
         LabelValue      := 0 ;
         QuadNo          := q ;
         LineNo          := 0 ;   (* no line number yet *)
         IsSigned        := FALSE ;
         IsVarParam      := FALSE ;
         Op              := op ;
         WITH Unary DO
            InitDestination(n, op1) ;
            Right   := FindUnreferencedItem(op2) ;
            AddTree(n, Right, 0)       (* last node to be created will be the head of the tree *)
         END
      END
   END ;
   LastNode := n
END BuildUnaryNode ;


(*
   BuildStartModFile - StartModFileOp is a quadruple which has the following
                       format:

                       StartModFileOp  _  _  ModuleSym

                       Its function is to reset the source file to another
                       file, hence all line numbers emmitted with the
                       generated code will be relative to this source file.
*)

PROCEDURE BuildStartModFile (Head, Current: CARDINAL; op1, op3: CARDINAL) ;
BEGIN
   CompilingMainModule := GetMainModule()=op3
END BuildStartModFile ;


(*
   BuildStartDefFile - StartDefFileOp is a quadruple which has the following
                       format:

                       StartDefFileOp  _  _  ModuleSym

                       Its function is to reset the source file to another
                       file, hence all line numbers emmitted with the
                       generated code will be relative to this source file.
                       We use it in the code generation phase to set a boolean
                       indicating whether we have reached the main module yet.
*)

PROCEDURE BuildStartDefFile ;
BEGIN
   CompilingMainModule := FALSE
END BuildStartDefFile ;


(*
   BuildInline - InlineOp is a quadruple which has the following format:

                 InlineOp   NulSym  NulSym  Buf

                 The Inline Buffer, Buf, is written to standard output.
*)

PROCEDURE BuildInline (Head, Current: CARDINAL; op3: CARDINAL) ;
BEGIN
   (*
      Must Flush the out standing instructions in the code buffer
      otherwise the Inline code will be emmitted in the incorrect
      order.
   *)
   FlushCoprocessor ;
   FlushForest(Head, LastFlushQuad, Current) ;
   FlushPipe(0) ;
   ResetRegisterValues ;
   IF NOT APUCodeGenerator
   THEN
      (*
         we only want one copy of the inline assembly, use this method
         if we are not generating APU code.
      *)
      WriteInlineBuffer(op3)
   END
END BuildInline ;


(*
   BuildLineNode - 
*)

PROCEDURE BuildLineNode (line: CARDINAL) ;
BEGIN
   IF LastNode>0
   THEN
      ListOfItems[LastNode].ItemIsNode.LineNo := line
   END
END BuildLineNode ;


(*
   DecreaseStack - decreases the runtime stack by c, using the least expensive method.
*)

PROCEDURE DecreaseStack (c: CARDINAL) ;
BEGIN
   IF c#0
   THEN
      FlushRegister(ESP)
   END ;
   IF c=1
   THEN
      BinInstReg(dec, MaxDatum(), RegESP)
   ELSIF c=2
   THEN
      BinInstReg(push, GetUnitPrefix(2), RegEAX)
   ELSIF c=WordSize
   THEN
      BinInstReg(push, GetUnitPrefix(WordSize), RegEAX)
   ELSIF c>0
   THEN
      InstRegInt(sub, MaxDatum(), RegESP, VAL(INTEGER, c))
   END ;
   IF c#0
   THEN
      TrashRegisterValue(ESP)
   END
END DecreaseStack ;


(*
   IncreaseStack - increases the runtime stack by c.
*)

PROCEDURE IncreaseStack (c: CARDINAL) ;
BEGIN
   IF c#0
   THEN
      FlushRegister(ESP)
   END ;
   IF c=1
   THEN
      BinInstReg(inc, MaxDatum(), RegESP)
   ELSIF c>0
   THEN
      InstRegInt(add, MaxDatum(), RegESP, VAL(INTEGER, c))
   END ;
   IF c#0
   THEN
      TrashRegisterValue(ESP)
   END
END IncreaseStack ;


(*
   BuildNewLocalVar - Builds a new frame on the stack to contain the procedure
                      local variables.
*)

PROCEDURE BuildNewLocalVar (Head, Current: CARDINAL; op1, op2, op3: CARDINAL) ;
BEGIN
   PreviousScope    := op2 ;
   CurrentProcedure := op3 ;
   FlushForest(Head, LastFlushQuad, Current) ;
   ResetRegisterValues ;

   IF NOT APUCodeGenerator
   THEN
      ProcedureInfo(op3, op1)     (* Generate GDB information *)
   END ;
   GenerateEntryProcedureLabel(Current, op3) ;
   IF NOT APUCodeGenerator
   THEN
      LineInfo(op1) ;             (* Generate GDB information *)
   END ;

   (* Create new frame *)

   BinInstReg(push, MaxDatum(), RegEBP) ;
   InstRegReg(mov, MaxDatum(), RegEBP, RegESP) ;

   IF IsProcedureNested(CurrentProcedure)
   THEN
      BinInstReg(push, MaxDatum(), RegECX)
   END ;

   IF DebugOn
   THEN
      DeclareLocalVariables(op3)   (* only emits comments about stack frame *)
   END ;

   (* Make local variable space *)

   PushSumOfLocalVarSize(op3) ;
   DecreaseStack(PopCard()) ;

   IF NOT APUCodeGenerator
   THEN
      CheckProfiling
   END ;

   (* Push Registers *)

   BinInstReg(push, MaxDatum(), RegEBX) ;
   BinInstReg(push, MaxDatum(), RegEDI) ;
   BinInstReg(push, MaxDatum(), RegESI) ;

   IF NOT APUCodeGenerator
   THEN
      (* Generate debugging trace infomation *)
      
      BeginInfo
   END

END BuildNewLocalVar ;


(*
   BuildKillLocalVar - destroy activation record and return.
*)

PROCEDURE BuildKillLocalVar (Head, Current: CARDINAL; op3: CARDINAL) ;
BEGIN
   FlushForest(Head, LastFlushQuad, Current) ;

   IF NOT APUCodeGenerator
   THEN
      (* Generate debugging trace infomation *)

      EndInfo
   END ;

   (* Pop Registers *)

   BinInstReg(pop, MaxDatum(), RegESI) ;
   BinInstReg(pop, MaxDatum(), RegEDI) ;
   BinInstReg(pop, MaxDatum(), RegEBX) ;

   (* Kill local variable space *)

   PushSumOfLocalVarSize(op3) ;
   IncreaseStack(PopCard()) ;

   (* Kill activation record - (call frame) *)
   IF IsProcedureNested(CurrentProcedure)
   THEN
      BinInstReg(pop, MaxDatum(), RegECX)
   END ;

   BinInst(leave) ;
   BinInst(ret) ;

   ResetRegisterValues ;

   GenerateExitProcedureLabel(op3) ;

   CurrentProcedure := NulSym ;
   PreviousScope    := NulSym
END BuildKillLocalVar ;


(*
   BuildOffsetNode - creates a unary node contains  op1 offset op3
*)

PROCEDURE BuildOffsetNode (Start: CARDINAL; op1, op3: CARDINAL) ;
BEGIN
   IF IsConst(op1)
   THEN
      InternalError('offset should be calculated when folding constants', __FILE__, __LINE__)
   ELSE
      PushOffset(op3) ;
      BuildUnaryNode(Start, BecomesOp, op1, MakeNewConstFromValue())
   END
END BuildOffsetNode ;


(*
   CallInnerInit - produce a call to inner module initialization routine.
*)

PROCEDURE CallInnerInit (Sym: CARDINAL) ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   GetKey(GetModuleInitName(Sym), a) ;
   InstStr(call, a)
END CallInnerInit ;


(*
   BuildStart - emits starting code before the main BEGIN END of the
                current module.
*)

PROCEDURE BuildStart (Head, Current: CARDINAL; op1, op3: CARDINAL) ;
BEGIN
   FlushForest(Head, LastFlushQuad, Current) ;
   IF CompilingMainModule
   THEN
      WritePrologue(op3, Current, op1) ;
      ForeachInnerModuleDo(op3, CallInnerInit)
   END
END BuildStart ;


(*
   BuildEnd - emits terminating code after the main BEGIN END of the
              current module.
*)

PROCEDURE BuildEnd (Head, Current: CARDINAL; op3: CARDINAL) ;
BEGIN
   FlushForest(Head, LastFlushQuad, Current) ;
   WriteEpilogue(op3) ;
   IF CompilingMainModule
   THEN
      FlushCoprocessor
   END
END BuildEnd ;


(*
   AlignStackPointer - aligns the stack pointer on a sensible boundary.
*)

PROCEDURE AlignStackPointer ;
BEGIN
   IF WordSize=DWORDSIZE
   THEN
      InstRegHexCard(and, MaxDatum(), RegESP, 0FFFFFFF8H) ;
      TrashRegisterValue(ESP)
   END
END AlignStackPointer ;


(*
   BuildParamNode - builds a parameter node. We ignore param #0  (the return pseudo param)
*)

PROCEDURE BuildParamNode (q: CARDINAL; op1, op2, op3: CARDINAL) ;
BEGIN
   IF op1#0
   THEN
      IF IsVarParam(op2, op1)
      THEN
         IF IsConst(op3)
         THEN
            WriteError('cannot pass a constant as a VAR parameter')
         ELSE
            (* ignore LeftAddr and RightAddr, but must be careful about size of Operand3 *)
            (* SIZE(operand3) will normally be TSIZE(ADDRESS) but NOT when it is unbounded *)
            BuildUnaryNode(q, ParamOp, GetType(GetNthParam(op2, op1)), op3) ;
            IF NOT IsUnbounded(GetType(GetNthParam(op2, op1)))
            THEN
               (*
                  we only use the var boolean when we have items which need to be referenced by a single address.
                  remember that unbounded parameters are passed in the same way whether they are VAR or NON VAR.
               *)
               PutVarParamNode(LastNode, TRUE)
            END
         END
      ELSE
         IF IsConst(op3)
         THEN
            IF IsMathType(GetType(GetNthParam(op2, op1)))
            THEN
               InternalError('should have handled FPU parameter constant before this function was called', __FILE__, __LINE__)
            ELSE
               BuildUnaryNode(q, ParamOp, GetType(GetNthParam(op2, op1)), op3)
            END
         ELSE
            (* ignore LeftAddr and RightAddr for ParamOp *)
            (*
               we must use GetSize(GetType(GetNth ... as we might want to pass a procedure as a parameter.
               (a procedure does not know about its size)
            *)
            IF GetSize(GetType(GetNthParam(op2, op1)))>WordSize*3
            THEN
               (* use block copy *)
               PushSize(GetType(GetNthParam(op2, op1))) ;
               BuildBinaryNode(q, SubOp, ESP, ESP,
                               MakeNewConstFromValue()) ;
               PutPrerequisiteRegister(FindLatestItem(ESP), ESP) ;
               BuildUnaryNode(q, BecomesOp, EDI, ESP) ;
               PutPrerequisiteRegister(FindLatestItem(EDI), EDI) ;
               BuildAddr(q, ESI, op3) ;
               PutPrerequisiteRegister(FindLatestItem(ESI), ESI) ;
               PushSize(GetType(GetNthParam(op2, op1))) ;
               BuildBinaryNode(q, XIndrOp, MakeNewConstFromValue(),
                               EDI, ESI) ;
               (* here we flush because expression likely to become too complex later on as
                  we have tied up EDI, ESI, ECX *)
               FlushForest(q, LastFlushQuad, q)
            ELSE
               BuildUnaryNode(q, ParamOp, GetType(GetNthParam(op2, op1)), op3)
            END
         END
      END
   END
END BuildParamNode ;


(*
   BuildParam - 
*)

PROCEDURE BuildParam (q: CARDINAL; op1, op2, op3: CARDINAL) ;
BEGIN
   IF (op1#0) AND IsConst(op3) AND IsMathType(GetType(GetNthParam(op2, op1)))
   THEN
      CoprocessorParam(op3, GetType(GetNthParam(op2, op1)))
   ELSE
      BuildParamNode(q, op1, op2, op3)
   END      
END BuildParam ;


(*
   RemoveParameters - removes the parameters for procedure, proc.
*)

PROCEDURE RemoveParameters (proc: CARDINAL) ;
BEGIN
   IF UnboundedOnStack
   THEN
      RemoveAnyUnboundedData
   ELSE
      PushSumOfParamSize(proc) ;
      IF GetType(proc)=NulSym   (* Is it a function *)
      THEN
         (* Procedure *)
         IncreaseStack(PopCard())
      ELSE
         (* Function - assumes return argument held in %eax *)
         IncreaseStack(PopCard())
      END
   END
END RemoveParameters ;


(*
   CheckNestedProcedure - checks whether we are about to call a nested procedure.
                          If the procedure is nested we must set up the activation
                          record pointer:

                            o    if we are the statically defined parent scope then we
                                 pass our frame pointer in ecx
                            o    otherwise we find out the previous statically defined parent
                                 and pass this in ecx.
*)

PROCEDURE CheckNestedProcedure (Sym: CARDINAL) ;
VAR
   BackScope: CARDINAL ;
BEGIN
   IF IsProcedureNested(Sym)
   THEN
      (*
         following Advanced Compiler Design and Implementation by Steven S.Muchnick p114

         (i)   if the procedure being called is nested directly within its caller, its
               static link points to its callers frame.
         (ii)  if the procedure is at the same level of nesting as its caller, then its
               static link is the same as the callers static link
         (iii) if the procedure being called is n levels higher than the caller in the
               nesting structure, then its static link can be determined by following, n,
               static links back from the callers static link and copying the static link
               found there.
      *)
      IF GetScopeAuthor(Sym)=CurrentProcedure
      THEN
         (* found (i) *)
         InstRegReg(mov, MaxDatum(), RegECX, RegEBP)
      ELSIF GetScopeAuthor(Sym)=PreviousScope
      THEN
         (* found (ii) *)
         InstRegMD(mov, MaxDatum(), RegECX, GetOffset(ActivationPointer), 0,
                   GetScope(ActivationPointer), ActivationPointer)
      ELSE
         (* must be (iii) *)
         InstRegMD(mov, MaxDatum(), RegECX, GetOffset(ActivationPointer), 0,
                   GetScope(ActivationPointer), ActivationPointer) ;
         BackScope := PreviousScope ;
         WHILE BackScope#GetScopeAuthor(Sym) DO
            PushOffset(ActivationPointer) ;
            InstRegMIO(mov, MaxDatum(), RegECX, RegECX, PopInt()) ;
            BackScope := GetScopeAuthor(BackScope)
         END
      END
   END
END CheckNestedProcedure ;


(*
   BuildCallNode - builds a call to subroutine node.
*)

PROCEDURE BuildCallNode (Head, Current: CARDINAL; op3: CARDINAL) ;
BEGIN
   IF IsProcedure(op3)
   THEN
      FlushForest(Head, LastFlushQuad, Current) ;
      FlushAllRegisters ;
      CheckNestedProcedure(op3) ;
      InstName(call, MaxDatum(), op3) ;
      RemoveParameters(op3)
   ELSIF IsProcType(GetType(op3))
   THEN
      (* maybe we need to compute op3 *)
      BuildUnaryNode(Current, CallOp, NulSym, op3) ;   (* op1 is meaningless *)
      FlushForest(Head, LastFlushQuad, Current) ;
      (* no need to remove parameters here as it is done in UnaryCall *)
   ELSE
      InternalError('expecting procedure or procedure variable', __FILE__, __LINE__)
   END
END BuildCallNode ;


(*
   BuildFunctValueNode - builds a unary node which will obtain return value.
*)

PROCEDURE BuildFunctValueNode (Current: CARDINAL; op1: CARDINAL) ;
BEGIN
   IF UsingFunctionReturnValue
   THEN
      InternalError('cannot obtain the return value from a function twice in basic block', __FILE__, __LINE__)
   ELSE
      UsingFunctionReturnValue := TRUE ;
      BuildUnaryNode(Current, FunctValueOp, op1, EAX)
   END
END BuildFunctValueNode ;


(*
   BuildAddr - 
*)

PROCEDURE BuildAddr (Current: CARDINAL; op1, op3: CARDINAL) ;
BEGIN
   BuildUnaryNode(Current, AddrOp, op1, op3)
(* was
   IF IsProcedure(GetVarFather(op3))
   THEN
      (* ok we can calculate the address using add, sub, and mov *)
      BuildUnaryNode(0, BecomesOp, op1, EBP) ;
      PushOffset(op3) ;
      BuildBinaryNode(Current, AddOp, op1, op1, MakeNewConstFromValue())
   ELSE
      (*
         perhaps we can do better than this by utilizing immediate with operator?
         for now we simply place the symbols address inside a free register.
      *)
      BuildUnaryNode(Current, AddrOp, op1, op3)
   END
*)
END BuildAddr ;


(*
   BuildDummyNode - builds a dummy node whose duty is to make sure GenCode
                    traverses all op1, op2, and op3.
*)

PROCEDURE BuildDummyNode (q: CARDINAL; op1, op2, op3: CARDINAL) ;
VAR
   t1, t2, n: IndexToItem ;
BEGIN
   t1 := FindUnreferencedItem(op2) ;
   t2 := FindUnreferencedItem(op3) ;
   n  := NewNode() ;
   WITH ListOfItems[n] DO
      IsLeaf := FALSE ;
      IsRef  := FALSE ;
      WITH ItemIsNode DO
         LabelValue := 0 ;
         QuadNo     := q ;
         LineNo     := 0 ;
         IsSigned   := FALSE ;
         IsVarParam := FALSE ;
         Op         := DummyOp ;
         WITH Dummy DO
            InitDestination(n, op1) ;
            Tree1 := t1 ;
            Tree2 := t2
         END
      END
   END ;
   AddTree(n, t1, t2)
END BuildDummyNode ;


(*
   BuildIndrXNode - builds an indirection node. It decides whether to use memcopy (reps)
                    or register moves.
*)

PROCEDURE BuildIndrXNode (Start: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   IF GetSize(op2)>3*WordSize
   THEN
      (* requires more than three registers, hence we choose reps *)
      BuildBecomes(0, BecomesOp, ESI, op3) ;
      PutPrerequisiteRegister(FindLatestItem(ESI), ESI) ;
      BuildAddr(Start, EDI, op1) ;
      PutPrerequisiteRegister(FindLatestItem(EDI), EDI) ;
      PushSize(op2) ;
      BuildBinaryNode(Start, op, MakeNewConstFromValue(), EDI, ESI) ;
      (* here we flush because expression likely to become too complex later on as
         we have tied up EDI, ESI, ECX *)
      FlushForest(Start, LastFlushQuad, Start)
   ELSE
      (* use standard mov and registers *)
      BuildBinaryNode(Start, op, op1, op2, op3)
   END
END BuildIndrXNode ;


(*
   BuildIndrX - 
*)

PROCEDURE BuildIndrX (Start: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsMathType(op2)
   THEN
      CoprocessorIndrX(op1, op2, op3)
   ELSE
      BuildIndrXNode(Start, op, op1, op2, op3)
   END
END BuildIndrX ;


(*
   BuildXIndrNode - builds in indirection node. It decides whether to use memcopy (reps)
                    or register moves.   ( *op1 := op3 ) with TSIZE(op2) bytes
*)

PROCEDURE BuildXIndrNode (Start: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsConstString(op3)
   THEN
      (* moving a string indirectly - catch this later *)
      BuildBinaryNode(Start, op, op2, op1, op3)
   ELSIF GetSize(op2)>3*WordSize
   THEN
      (* requires more than three registers, hence we choose reps *)
      BuildBecomes(0, BecomesOp, EDI, op1) ;
      PutPrerequisiteRegister(FindLatestItem(EDI), EDI) ;
      BuildAddr(Start, ESI, op3) ;
      PutPrerequisiteRegister(FindLatestItem(ESI), ESI) ;
      PushSize(op2) ;
      BuildBinaryNode(Start, op, MakeNewConstFromValue(), EDI, ESI) ;
      (* here we flush because expression likely to become too complex later on as
         we have tied up EDI, ESI, ECX *)
      FlushForest(Start, LastFlushQuad, Start)
   ELSE
      (* use standard mov and registers *)
      BuildBinaryNode(Start, op, op2, op1, op3)
   END
END BuildXIndrNode ;


(*
   BuildXIndr - 
*)

PROCEDURE BuildXIndr (Start: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsMathType(op2)
   THEN
      CoprocessorXIndr(op1, op2, op3)
   ELSE
      BuildXIndrNode(Start, op, op1, op2, op3)
   END
END BuildXIndr ;


(*
   NeedsSignedOperator - returns TRUE if the binary operator using operands,
                         op2 op3, requires its signed equivalent.
*)

PROCEDURE NeedsSignedOperator (op2, op3: CARDINAL) : BOOLEAN ;
VAR
   Mix, Low, High: CARDINAL ;
BEGIN
   IF IsConst(op2) AND IsConst(op3)
   THEN
      InternalError('should have folded constants', __FILE__, __LINE__)
   ELSE
      IF IsConst(op2)
      THEN
         PushValue(op2) ;
         PushCard(0) ;
         IF Less()
         THEN
            RETURN( TRUE )
         END ;
         Mix := GetType(op3)
      ELSIF IsConst(op3)
      THEN
         PushValue(op3) ;
         PushCard(0) ;
         IF Less()
         THEN
            RETURN( TRUE )
         END ;
         Mix := GetType(op2)
      ELSE
         Mix := MixTypes(GetType(op2), GetType(op3))
      END ;
      IF Mix=Integer
      THEN
         RETURN( TRUE )
      ELSIF (Mix#NulSym) AND IsSubrange(Mix)
      THEN
         GetSubrange(Mix, High, Low) ;
         PushValue(Low) ;
         PushCard(0) ;
         RETURN( Less() )
      ELSE
         RETURN( FALSE )
      END
   END
END NeedsSignedOperator ;


(*
   BuildDivNode - builds the division node, it checks for signed and unsigned division.
*)

PROCEDURE BuildDivNode (Start: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   (* eax = op2 *)
   IF (NOT NeedsSignedOperator(op2, op3)) AND IsConst(op3) AND IsPowerOfTwo(op3)
   THEN
      BuildBinaryNode(Start, LogicalShiftRightOp, op1, op2, Log2Of(op3))
   ELSE
      BuildBinaryNode(Start, op, op1, op2, op3) ;
      PutPrerequisiteRegister(FindLatestItem(op2), EAX) ;
      PutSignedNode(LastNode, NeedsSignedOperator(op2, op3))
   END
END BuildDivNode ;


(*
   BuildMultNode - builds the multiplication node, it checks for the shift left optimization.
*)

PROCEDURE BuildMultNode (Start: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   IF NOT NeedsSignedOperator(op2, op3)
   THEN
      IF IsConst(op3) AND IsPowerOfTwo(op3)
      THEN
         BuildBinaryNode(Start, LogicalShiftLeftOp, op1, op2, Log2Of(op3))
      ELSIF IsConst(op2) AND IsPowerOfTwo(op2)
      THEN
         BuildBinaryNode(Start, LogicalShiftLeftOp, op1, op3, Log2Of(op2))
      ELSE
         (* cannot use shift as no constant or no direct power of two *)
         BuildBinaryNode(Start, op, op1, op2, op3)
      END
   ELSE
      (* cannot use shift as the result is signed *)
      BuildBinaryNode(Start, op, op1, op2, op3)
   END
END BuildMultNode ;


(*
   BuildModNode - builds the modulus node, it checks for signed and unsigned division.
*)

PROCEDURE BuildModNode (Start: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   IF (NOT NeedsSignedOperator(op2, op3)) AND IsConst(op3) AND IsPowerOfTwo(op3)
   THEN
      BuildBinaryNode(Start, LogicalAndOp, op1, op2, Decrement(op3))
   ELSE
      BuildBinaryNode(Start, op, op1, op2, op3) ;
      PutPrerequisiteRegister(FindLatestItem(op2), EAX) ;
      PutSignedNode(LastNode, NeedsSignedOperator(op2, op3))
   END
END BuildModNode ;


(*
   BuildDiv - checks to see whether we should use the FPU code generator or gencode.
*)

PROCEDURE BuildDiv (q: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsCoprocessorUsed(op1, op2, op3)
   THEN
      CoprocessorDiv(op1, op2, op3)
   ELSE
      BuildDivNode(q, op, op1, op2, op3)
   END
END BuildDiv ;


(*
   BuildMod - checks to see whether we should use the FPU code generator or gencode.
*)

PROCEDURE BuildMod (q: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsCoprocessorUsed(op1, op2, op3)
   THEN
      CoprocessorMod(op1, op2, op3)
   ELSE
      BuildModNode(q, op, op1, op2, op3)
   END
END BuildMod ;


(*
   BuildMult - checks to see whether we should use the FPU code generator or gencode.
*)

PROCEDURE BuildMult (q: CARDINAL; op: QuadOperator; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsCoprocessorUsed(op1, op2, op3)
   THEN
      CoprocessorMult(op1, op2, op3)
   ELSE
      BuildMultNode(q, op, op1, op2, op3)
   END
END BuildMult ;


(*
   PackString - packs a string 'a' into another string 'b',
                starting at position i in string 'a'.
                The number of characters packed into b are returned.
                String b is assummed to contain 4 elements.
*)

PROCEDURE PackString (a: ARRAY OF CHAR; DestSize, i: CARDINAL;
                      VAR b: ARRAY OF CHAR) : CARDINAL ;
VAR
   SourceSize,
   Pack, j   : CARDINAL ;
BEGIN
   SourceSize := StrLen(a) ;
   IF (DestSize>SourceSize) AND
      (((SourceSize-i=3) AND (WordSize=DWORDSIZE)) OR (SourceSize-i=1))
   THEN
      (* Padd out remaining boundary byte with the nul character *)
      Pack := SourceSize-i+1
   ELSE
      Pack := GetUnitOfTransfer(DestSize-i)
   END ;
   FOR j := 0 TO Pack-1 DO
      IF i+j>=SourceSize
      THEN
         b[j] := nul
      ELSE
         b[j] := a[i+j]
      END
   END ;
   IF Pack<HIGH(b)
   THEN
      b[Pack] := nul     (* May need to terminate the string, b *)
   END ;
   RETURN( Pack )
END PackString ;


(*
   MoveStringFromToDirect - moves a string, StringSym, to a direct destination.
*)

PROCEDURE MoveStringFromToDirect (StringSym: CARDINAL; DestSym: CARDINAL; DestSize: CARDINAL) ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
   b: ARRAY [0..3] OF CHAR ;
   Pack,
   i,
   String,
   SourceSize: CARDINAL ;
BEGIN
   (* Make sure the string is the same size as DestSym *)
   String := GetString(StringSym) ;
   SourceSize := GetStringLength(StringSym) ;
   GetKey(String, a) ;
   i := 0 ;
   WHILE i<SourceSize DO
      Pack := PackString(a, DestSize, i, b) ;
      MovePackedStringTo(DestSym, INTEGER(i), b, Pack) ;
      INC(i, Pack)
   END ;
   IF i>DestSize
   THEN
      WriteError('string is too large for the variable')
   END ;
   (* Now fill in the remaining space with nul characters *)
   WHILE i<DestSize DO
      Pack := PackString('', DestSize, i, b) ;
      MovePackedStringTo(DestSym, INTEGER(i), b, Pack) ;
      INC(i, Pack)
   END
END MoveStringFromToDirect ;


(*
   BuildBecomesNode - creates a becomes node in the tree.
*)

PROCEDURE BuildBecomesNode (Start: CARDINAL; op: QuadOperator; op1, op3: CARDINAL) ;
VAR
   Size1, Size3: CARDINAL ;
BEGIN
   IF IsConst(op1) AND IsConst(op3)
   THEN
      InternalError('constants should have been resolved before code generation', __FILE__, __LINE__)
   ELSIF IsConstString(op3)
   THEN
      BuildUnaryNode(Start, op, op1, op3)
   ELSE
      IF GetMode(op1)=LeftValue
      THEN
         PushSize(Address) ;
         Size1 := PopCard()
      ELSE
         Size1 := GetVarTypeSize(op1)
      END ;
      IF IsConst(op3) OR IsProcedure(op3)
      THEN
         Size3 := Size1
      ELSE
         IF GetMode(op3)=LeftValue
         THEN
            PushSize(Address) ;
            Size3 := PopCard()
         ELSE
            Size3 := GetVarTypeSize(op3)
         END
      END ;
      IF (IsMathType(GetType(op1)) OR IsMathType(GetType(op3))) AND
         ((GetMode(op1)#LeftValue) AND (GetMode(op3)#LeftValue))
      THEN
         (* coprocessor used and we are not calculating addresses *)
         CoprocessorBecomes(op1, op3)
      ELSIF (Size1#Size3) AND (NOT (IsMathType(GetType(op1)) AND (IsMathType(GetType(op3)))))
      THEN
         BeginError ;
         WriteLn ;
         WriteKey(GetSymName(op1)) ; WriteString(' := ') ; WriteKey(GetSymName(op3)) ; WriteLn ;
         WriteString('Type of ') ; WriteKey(GetSymName(op1)) ;
         WriteString(' is ') ; WriteKey(GetSymName(GetType(op1))) ; WriteLn ;
         WriteString('Type of ') ; WriteKey(GetSymName(op3)) ;
         WriteString(' is ') ; WriteKey(GetSymName(GetType(op3))) ; WriteLn ;
         WriteError('different sizes in :=')
      ELSE
         IF Size1>3*WordSize
         THEN
            (* requires more than three registers, hence we choose reps *)
            BuildAddr(Start, EDI, op1) ;
            PutPrerequisiteRegister(FindLatestItem(EDI), EDI) ;
            BuildAddr(Start, ESI, op3) ;
            PutPrerequisiteRegister(FindLatestItem(ESI), ESI) ;
            PushCard(Size1) ;
            BuildBinaryNode(Start, IndrXOp, MakeNewConstFromValue(),
                            EDI, ESI) ;  (* block moves are the same *)
            (* here we flush because expression likely to become too complex later on as
               we have tied up EDI, ESI, ECX *)
            FlushForest(Start, LastFlushQuad, Start)
         ELSE
            BuildUnaryNode(Start, op, op1, op3)
         END
      END
   END
END BuildBecomesNode ;


(*
   BuildBecomes - 
*)

PROCEDURE BuildBecomes (Start: CARDINAL; op: QuadOperator; op1, op3: CARDINAL) ;
BEGIN
   IF (IsMathType(GetType(op1)) OR IsMathType(GetType(op3))) AND
      ((GetMode(op1)#LeftValue) AND (GetMode(op3)#LeftValue))
   THEN
      (* coprocessor used and we are not calculating addresses *)
      CoprocessorBecomes(op1, op3)
   ELSE
      BuildBecomesNode(Start, op, op1, op3)
   END
END BuildBecomes ;


(*
   BuildCoerce - coerce op3 to type op2 placing the result into
                 op1.
                 Coerce will NOT alter the machine representation
                 of op3 to comply with TYPE op2.
                 Therefore it INSISTS that under all circumstances that the
                 type sizes of op1 and op3 are the same.
                 CONVERT will perform machine manipulation to change variable
                 types, coerce does no such thing.
*)

PROCEDURE BuildCoerce (Start: CARDINAL; op1, op3: CARDINAL) ;
BEGIN
   (*
      Must not call GetVarTypeSize if sym is a constant.
      Constants can assume any size from BYTESIZE .. WordSize
   *)
   IF IsProcedure(op3)
   THEN
      IF GetVarTypeSize(op1)=WordSize
      THEN
         BuildUnaryNode(Start, BecomesOp, op1, op3)
      ELSE
         WriteError('procedure address can only be stored in a word size operand')
      END
   ELSIF IsConst(op3) OR (GetVarTypeSize(op1)=GetVarTypeSize(op3))
   THEN
      BuildUnaryNode(Start, BecomesOp, op1, op3)
   ELSE
      WriteError('TYPE Coersion can only be achieved with types of the same size')
      (* Alter this when time permits *)
   END
END BuildCoerce ;


(*
   PrepareToPushUnboundedData - prepares to push some unbounded data.
                                If this has not been called before
                                then Register RC := SP.
                                The value of SP is before any unbounded data
                                has been pushed - therefore take a copy of SP
                                into a reserve register.
*)

PROCEDURE PrepareToPushUnboundedData ;
BEGIN
   IF NOT UnboundedOnStack
   THEN
      (*
         Not done yet -
         so do this now and never again before the procedure call.
      *)
      IF (NOT IsRegisterHidden(EBX)) AND (NOT IsRegisterFree(EBX))
      THEN
         DisplayRegisterStack ;
         InternalError('expected register %ebx to be hidden for pushing unbounded data, need to hide %ebx before this procedure is called', __FILE__, __LINE__)
      END ;
      UnboundedOnStack := TRUE ;
      PeepMoveReg2Reg(ESP, EBX)     (* RegEBX := old RegESP *)
   END
END PrepareToPushUnboundedData ;


(*
   RemoveAnyUnboundedData - attempts to remove any undounded data.
                            Unbounded data will be on the stack if the
                            boolean, UnboundedOnStack, is true.
*)

PROCEDURE RemoveAnyUnboundedData ;
BEGIN
   IF UnboundedOnStack
   THEN
      UnboundedOnStack := FALSE ;
      PeepMoveReg2Reg(EBX, ESP) ;   (* RegESP := RegEBX *)
      PushRegister(EBX, free)
   END
END RemoveAnyUnboundedData ;


(*
   GenerateCommentString - generates a comment string, b, from a string, a,
                           which contains, n, characters.
*)

PROCEDURE GenerateCommentString (a: ARRAY OF CHAR; n: CARDINAL;
                                 VAR b: ARRAY OF CHAR) ;
VAR
   i: CARDINAL ;
   c: ARRAY [0..1] OF CHAR ;
BEGIN
   StrCopy('"', b) ;
   FOR i := 0 TO n DO
      IF a[i]=nul
      THEN
         StrConCat(b, '<nul>', b)
      ELSE
         c[0] := a[i] ;
         c[1] := nul ;
         StrConCat(b, c, b)
      END
   END ;
   StrConCat(b, '"   ( ', b) ;
   FOR i := n TO 0 BY -1 DO
      IF a[i]=nul
      THEN
         StrConCat(b, '<nul>', b)
      ELSE
         c[0] := a[i] ;
         c[1] := nul ;
         StrConCat(b, c, b)
      END
   END ;
   StrConCat(b, ' )', b)
END GenerateCommentString ;


(*
   PushCardinal - pushes a CARDINAL to the stack.
*)

PROCEDURE PushCardinal (c: CARDINAL) ;
BEGIN
   BinInstHex(push, MaxDatum(), c)
END PushCardinal ;


(*
   PushString - pushes a 1..4 byte string to the stack. The 80386 cannot
                push an 8 bit entity as 8 bits only so we manufacture
                this instruction.
*)

PROCEDURE PushString (Pack: CARDINAL; a: ARRAY OF CHAR) ;
VAR
   i,
   Word: CARDINAL ;
   b   : ARRAY [0..MaxString] OF CHAR ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(a)
      THEN
         INC(Word, ORD(a[i]))
      END
   END ;
   GenerateCommentString(a, Pack-1, b) ;
   Comment(b) ;
   PushCardinal(Word)
END PushString ;


(*
   PushStringToStack - moves a string, StringSym, onto the stack.
                       The string is terminated with a nul character.
*)

PROCEDURE PushStringToStack (StringSym: CARDINAL) ;
VAR
   a     : ARRAY [0..MaxString] OF CHAR ;
   b     : ARRAY [0..3] OF CHAR ;
   i,
   String,
   Length: CARDINAL ;
BEGIN
   String := GetString(StringSym) ;
   Length := GetStringLength(StringSym)+1 ;  (* +1 to include nul *)
   GetKey(String, a) ;
   i := Length ;
   IF i MOD WordSize#0
   THEN
      i := i+WordSize-(i MOD WordSize)
   END ;
   WHILE i>=WordSize DO
      DEC(i, WordSize) ;
      PaddString(a, Length, i, b) ;
      PushString(WordSize, b)
   END
END PushStringToStack ;


(*
   PaddString - padd a string 'a' into another string 'b',
                starting at position i in string 'a'.
                The number of characters packed into b are returned.
                String b is assummed to contain 4 elements and we
                will use all of them.
*)

PROCEDURE PaddString (a: ARRAY OF CHAR; DestSize: CARDINAL; i: CARDINAL;
                      VAR b: ARRAY OF CHAR) ;
VAR
   SourceSize, j: CARDINAL ;
BEGIN
   SourceSize := StrLen(a) ;
   FOR j := 0 TO WordSize-1 DO
      IF i+j>=SourceSize
      THEN
         b[j] := nul
      ELSE
         b[j] := a[i+j]
      END
   END ;
   IF WordSize<HIGH(b)
   THEN
      b[WordSize] := nul     (* May need to terminate the string, b *)
   END
END PaddString ;


(*
   PushArrayToStack - moves an array, ArraySym, onto the stack.
                      Pushed in reverse order and therefore the start
                      of the array must be the top of the stack.
                      80*86 stack grows from HIGH -> LOW.
                      These instructions must be emitted now as they
                      alter the stack. We cannot allow gencode to reorder
                      this statement.
*)

PROCEDURE PushArrayToStack (ArraySym: CARDINAL; parent, i: IndexToItem; Head, Start, End: CARDINAL) ;
VAR
   Size: CARDINAL ;
BEGIN
   CASE GetMode(ArraySym) OF

   RightValue: Size := GetSize(ArraySym) |
   LeftValue : Size := GetVarTypeSize(ArraySym) ;

   ELSE
      InternalError('not expecting to reach here', __FILE__, __LINE__)
   END ;
   IF NOT (IsRegisterFree(EDI) AND IsRegisterFree(ESI) AND
           IsRegisterFree(ECX))
   THEN
      IF DebugOn
      THEN
         DisplayRegisterStack
      END ;
      InternalError('expecting edi ecx and esi to be free', __FILE__, __LINE__)
   END ;
   (* align parameter if required *)
   PushCard(Size) ;
   DecreaseStack(PopCard()) ;

   PeepMoveReg2Reg(ESP, EDI) ;
   MoveRegisterToTop(ESI) ;
   MarkRegister(ESI, fixed) ;
   AhoGenCode(i, GetNodeRight(i), FALSE, Head, Start, End) ;
(*
   IF RegisterTop()#ESI
   THEN
      InternalError('expected ESI at the top of the register stack for unbounded parameter', __FILE__, __LINE__)
   END ;
*)
   PushCard(Size) ;
   Memcopy
END PushArrayToStack ;


(*
   PushUnbounded - pushes an Unbounded Array Sym data onto the stack.
*)

PROCEDURE PushUnbounded (Sym: CARDINAL) ;
BEGIN
   FlushRegister(ECX) ;
   FlushRegister(ESP) ;
   FlushRegister(ESI) ;
   FlushRegister(EDI) ;

   (*
      esi := ArrayAddress
      ecx := ArrayHigh + 2     because size of array -1 = high
                               possible null on array of char also 1
   *)
   InstRegMD(mov, MaxDatum(), RegESI, GetOffset(Sym), 0, GetScope(Sym), Sym) ;   (* Record field offset = 0 *)
   InstRegMD(mov, MaxDatum(), RegECX, GetOffset(Sym), INTEGER(WordSize), GetScope(Sym), Sym) ;

   (*
      INC(ecx, 1+1)  because:  Size of Array -1 = HIGH
                               Possible null on array of char also 1
   *)
   InstRegInt(add, MaxDatum(), RegECX, 1+1) ;
   (*
      sub  esp, ecx,
      mov  edi, esp
      rep movs          ;  while ecx>0 do
                        ;     mem[edi] := mem[esi]
                        ;     inc(edi)
                        ;     inc(esi)
                        ;     dec(ecx)
                        ;  end
   *)
   InstRegReg(sub, MaxDatum(), RegESP, RegECX) ;
   InstRegReg(mov, MaxDatum(), RegEDI, RegESP) ;
   TrashRegisterValue(ECX) ;
   IncludeValueInReg(ECX, MakeConstLit(MakeKey('0'))) ;
   TrashRegisterValue(ESP) ;
   TrashRegisterValue(ESI) ;
   TrashRegisterValue(EDI) ;
   BinInstUnit(repmovs, PByte, 0) ;   (* no idea how long string is *)
   UnHideRepsRegisters
END PushUnbounded ;


(*
   BuildUnbounded - codes the creation of an unbounded parameter
                    variable.
                    Pushes op3 onto the stack and places the start address into op2.
                    Note that there is no op1 and thus we flush after unbounded
                    has been encoded as the start address might be going through
                    an indirect pointer.
*)

PROCEDURE BuildUnbounded (Head, Start: CARDINAL; op2, op3: CARDINAL) ;
BEGIN
   IF IsConstString(op3)
   THEN
      BuildBinaryNode(Start, UnboundedOp, op3, op2, op3)
   ELSIF IsArray(GetType(op3))
   THEN
      CASE GetMode(op3) OF

      LeftValue : BuildBecomes(0, BecomesOp, ESI, op3) ;
                  PutPrerequisiteRegister(FindLatestItem(ESI), ESI) ;
                  BuildBinaryNode(Start, UnboundedOp, op3, op2, ESI) |
      RightValue: BuildAddr(Start, ESI, op3) ;
                  PutPrerequisiteRegister(FindLatestItem(ESI), ESI) ;
                  BuildBinaryNode(Start, UnboundedOp, op3, op2, ESI)

      ELSE
         InternalError('not expecting to reach here', __FILE__, __LINE__)
      END
   ELSE
      BuildBinaryNode(Start, UnboundedOp, op3, op2, op3)
   END ;
   FlushForest(Head, LastFlushQuad, Start)
END BuildUnbounded ;


(*
   BuildConvertNode - converts op3 into type op2 placing value into op1.
*)

PROCEDURE BuildConvertNode (q: CARDINAL; op1, op3: CARDINAL) ;
VAR
   s1: CARDINAL ;
BEGIN
   s1 := GetVarTypeSize(op1) ;
   IF IsConst(op3)
   THEN
      IF s1>WordSize
      THEN
         WriteErrorFormat1('implementation restriction - can only CONVERT constant types to a type whose size <= %d bytes', WordSize)
      END ;
      (* Constants assume any size up to WordSize *)
      BuildUnaryNode(q, BecomesOp, op1, op3)
   ELSE
      BuildUnaryNode(q, ConvertOp, op1, op3)
   END
END BuildConvertNode ;


(*
   BuildConvert - 
*)

PROCEDURE BuildConvert (q: CARDINAL; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsMathType(GetType(op1)) OR IsMathType(op2) OR
      IsMathType(GetType(op3))
   THEN
      CoprocessorConvert(op1, op2, op3)
   ELSE
      BuildConvertNode(q, op1, op3)
   END
END BuildConvert ;


(*
   BuildNegateNode - 
*)

PROCEDURE BuildNegateNode (q: CARDINAL; op1, op3: CARDINAL) ;
BEGIN
   IF IsConst(op3)
   THEN
      IF GetSize(GetType(op1)) > WordSize
      THEN
         WriteErrorFormat1('cannot negate a variable larger than register size (%d bytes)', WordSize)
      END ;
      PushCard(0) ;
      PushValue(op3) ;
      Sub ;
      BuildUnaryNode(q, BecomesOp, op1, MakeNewConstFromValue())
   ELSE
      BuildUnaryNode(q, NegateOp, op1, op3)
   END
END BuildNegateNode ;


(*
   BuildNegate - 
*)

PROCEDURE BuildNegate (q: CARDINAL; op1, op3: CARDINAL) ;
BEGIN
   IF IsConst(op1)
   THEN
      InternalError('constants should have been folded before now', __FILE__, __LINE__)
   ELSIF IsCoprocessorUsed(op1, op1, op3)
   THEN
      CoprocessorNeg(op1, op3)
   ELSE
      BuildNegateNode(q, op1, op3)
   END
END BuildNegate ;


(*
   BuildReturnValue - 
*)

PROCEDURE BuildReturnValue (q: CARDINAL; op1, op3: CARDINAL) ;
BEGIN
   IF IsMathType(GetType(op3))
   THEN
      CoprocessorReturnValue(op1, op3)
   ELSE
      BuildUnaryNode(q, ReturnValueOp, op3, op1)    (* op3, op1 is correct as op3 is the function op1 = value *)
   END
END BuildReturnValue ;


(*
   BuildFunctValue - 
*)

PROCEDURE BuildFunctValue (q: CARDINAL; op1: CARDINAL) ;
BEGIN
   IF IsMathType(GetType(op1))
   THEN
      CoprocessorFunctValue(op1)
   ELSE
      BuildFunctValueNode(q, op1)
   END
END BuildFunctValue ;


(*
   BuildBit - builds a node:  op1 := 1 << op3
*)

PROCEDURE BuildBit (q: CARDINAL; op1, op3: CARDINAL) ;
BEGIN
   BuildUnaryNode(q, BitOp, op1, op3) ;
   PutPrerequisiteRegister(FindLatestItem(op3), ECX)
END BuildBit ;


(*
   GoingToUseCoprocessor - returns TRUE if we are about to use the coprocessor.
*)

PROCEDURE GoingToUseCoprocessor (op: QuadOperator; op1, op2, op3: CARDINAL) : BOOLEAN ;
BEGIN
   CASE op OF

   AddOp        :  RETURN( IsCoprocessorUsed(op1, op2, op3) ) |
   SubOp        :  RETURN( IsCoprocessorUsed(op1, op2, op3) ) |
   MultOp       :  RETURN( IsCoprocessorUsed(op1, op2, op3) ) |
   DivOp        :  RETURN( IsCoprocessorUsed(op1, op2, op3) ) |
   ModOp        :  RETURN( IsCoprocessorUsed(op1, op2, op3) ) |
   NegateOp     :  RETURN( IsCoprocessorUsed(op1, op1, op3) ) |
   BecomesOp    :  RETURN( IsMathType(GetType(op1)) OR IsMathType(GetType(op3)) ) |
   ConvertOp    :  RETURN( IsMathType(GetType(op1)) OR IsMathType(op2) OR
                           IsMathType(GetType(op3)) ) |
   IfEquOp      :  RETURN( IsMathType(GetType(op1)) OR IsMathType(GetType(op2)) ) |
   IfNotEquOp   :  RETURN( IsMathType(GetType(op1)) OR IsMathType(GetType(op2)) ) |
   IfLessOp     :  RETURN( IsMathType(GetType(op1)) OR IsMathType(GetType(op2)) ) |
   IfLessEquOp  :  RETURN( IsMathType(GetType(op1)) OR IsMathType(GetType(op2)) ) |
   IfGreOp      :  RETURN( IsMathType(GetType(op1)) OR IsMathType(GetType(op2)) ) |
   IfGreEquOp   :  RETURN( IsMathType(GetType(op1)) OR IsMathType(GetType(op2)) ) |
   FunctValueOp :  RETURN( IsMathType(GetType(op1)) ) |
   ReturnValueOp:  RETURN( IsMathType(GetType(op3)) ) |
   XIndrOp      :  RETURN( IsMathType(op2) ) |
   IndrXOp      :  RETURN( IsMathType(op2) ) |
   ParamOp      :  RETURN( IsConst(op3) AND IsMathType(GetType(GetNthParam(op2, op1))) ) |
   MathOp       :  RETURN( TRUE ) |
   LineNumberOp :  RETURN( TRUE )

   ELSE
      RETURN( FALSE )
   END
END GoingToUseCoprocessor ;


(*
   BuildForest - builds a list of trees for the quadruples, Start..End.
*)

PROCEDURE BuildForest (Head, Start, End: CARDINAL) ;
VAR
   Previous     : CARDINAL ;
   op           : QuadOperator ;
   op1, op2, op3: CARDINAL ;
   FPUstate     : BOOLEAN ;
BEGIN
   TreePtr  := 0 ;
   ItemPtr  := 1 ;   (* no nodes/leaves yet *)
   LastNode := 0 ;
   FPUstate := FALSE ;
   Previous := Start ;
   WHILE (Start#0) AND (Start<=End) DO
      GetQuad(Start, op, op1, op2, op3) ;
      IF op=LineNumberOp
      THEN
         IF FPUstate
         THEN
            CoprocessorLineInfo(op3)
         ELSE
            BuildLineNode(op3)
         END
      ELSIF GoingToUseCoprocessor(op, op1, op2, op3)
      THEN
         FPUstate := TRUE ;
         FlushForest(Head, LastFlushQuad, Previous) ;  (* flush the integer gencode instructions *)
         ResetRegisterValues ;
         FlushPipe(0)
      ELSE
         FPUstate := FALSE ;
         FlushCoprocessor                              (* flush the coprocessor instructions     *)
      END ;
      CASE op OF

      IndrXOp           : BuildIndrX(Start, op, op1, op2, op3) |
      ModOp             : BuildMod(Start, op, op1, op2, op3) |
      DivOp             : BuildDiv(Start, op, op1, op2, op3) |
      MultOp            : BuildMult(Start, op, op1, op2, op3) |
      BecomesOp         : BuildBecomes(Start, op, op1, op3) |
      XIndrOp           : BuildXIndr(Start, op, op1, op2, op3) |
      LogicalOrOp,
      LogicalAndOp,
      LogicalXorOp      : BuildBinaryNode(Start, op, op1, op2, op3) |
      AddOp             : BuildAddNode(Start, op, op1, op2, op3) |
      SubOp             : BuildSubNode(Start, op, op1, op2, op3) |
      OffsetOp          : BuildOffsetNode(Start, op1, op3) |
      ReturnValueOp     : BuildReturnValue(Start, op1, op3) |
      FunctValueOp      : BuildFunctValue(Start, op1) |    (* return result in EAX, which must be moved to op1       *)
      CallOp            : BuildCallNode(Head, Start, op3) |
      ParamOp           : BuildParam(Start, op1, op2, op3) |
      AddrOp            : BuildAddr(Start, op1, op3) |
      GotoOp            : BuildGoto(Head, Start, op3) |
      NewLocalVarOp     : BuildNewLocalVar(Head, Start, op1, op2, op3) |
      KillLocalVarOp    : BuildKillLocalVar(Head, Start, op3) |
      ReturnOp          : (* nothing to do on the 80x86 *) |
      CoerceOp          : BuildCoerce(Start, op1, op3) |
      UnboundedOp       : BuildUnbounded(Head, Start, op1, op3) |
      StartModFileOp    : BuildStartModFile(Head, Start, op1, op3) |
      StartDefFileOp    : BuildStartDefFile |
      StartOp           : BuildStart(Head, Start, op1, op3) |
      EndFileOp         : (* nothing to do on the 80x86 *) |
      EndOp             : BuildEnd(Head, Start, op3) |
      InlineOp          : BuildInline(Head, Start, op3) |
      ConvertOp         : BuildConvert(Start, op1, op2, op3) |
      NegateOp          : BuildNegate(Start, op1, op3) |
      MathOp            : CoprocessorMath(op1, op2, op3) |
      BitOp             : BuildBit(Start, op1, op3) |

      IfInOp,
      IfNotInOp,
      IfEquOp,
      IfNotEquOp,
      IfGreEquOp,
      IfLessEquOp,
      IfLessOp,
      IfGreOp           : BuildCondition(Start, op, op1, op2, op3) |
      LineNumberOp      : (* caught above *)

      ELSE
         InternalError('do not understand this quadruple yet', __FILE__, __LINE__)
      END ;
      Previous := Start ;
      Start := GetNextQuad(Start)
   END ;
   FlushCoprocessor
END BuildForest ;


(*
   BuildGoto - 
*)

PROCEDURE BuildGoto (Head, Current: CARDINAL; DestQuad: CARDINAL) ;
VAR
   Label: ARRAY [0..MaxNumChars] OF CHAR ;
BEGIN
   FlushForest(Head, LastFlushQuad, Current) ;
   CardToStr(DestQuad, 0, Label) ;
   StrConCat('L', Label, Label) ;
   DotifyLabel(Label) ;
   InstStr(jmp, Label) ;
   ResetRegisterValues
END BuildGoto ;


(*
   WriteRegisterName - 
*)

PROCEDURE WriteRegisterName (Sym: CARDINAL) ;
BEGIN
   IF Sym=EAX
   THEN
      WriteString('%eax')
   ELSIF Sym=EBX
   THEN
      WriteString('%ebx')
   ELSIF Sym=ECX
   THEN
      WriteString('%ecx')
   ELSIF Sym=EDX
   THEN
      WriteString('%edx')
   ELSIF Sym=ESP
   THEN
      WriteString('%esp')
   ELSIF Sym=EBP
   THEN
      WriteString('%ebp')
   ELSIF Sym=EDI
   THEN
      WriteString('%edi')
   ELSIF Sym=ESI
   THEN
      WriteString('%esi')
   ELSE
      InternalError('no such register', __FILE__, __LINE__)
   END
END WriteRegisterName ;


(*
   WriteSymbol - 
*)

PROCEDURE WriteSymbol (i: IndexToItem) ;
VAR
   Sym: CARDINAL ;
BEGIN
   Sym := GetDestination(i) ;
   IF IsRegister(Sym)
   THEN
      WriteRegisterName(Sym)
   ELSE
      IF IsConst(Sym)
      THEN
         Write('#') ;
         IF IsConstLit(Sym)
         THEN
            WriteKey(GetSymName(Sym))
         ELSIF GetType(Sym)=LongReal
         THEN
            PushValue(Sym) ;
            WriteLongReal(PopLongReal(), 24, 12)
         ELSE
            WriteKey(GetSymName(Sym))
         END
      ELSIF IsVar(Sym) OR IsProcedure(Sym)
      THEN
         WriteAsmName(Sym)
      ELSE
         WriteKey(GetSymName(Sym))
      END
   END ;
   IF GetPrequisiteRegister(i)#NulSym
   THEN
      WriteString(' "') ;
      WriteRegisterName(GetPrequisiteRegister(i)) ;
      WriteString('" ')
   END
END WriteSymbol ;


(*
   FitsIntoRegister - returns TRUE if a symbol, Sym, fill fit into a register.
*)

PROCEDURE FitsIntoRegister (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsConst(Sym)
   THEN
      IF IsConstString(Sym)
      THEN
         RETURN( FALSE )   (* GetStringLength(Sym)<=WordSize      when we feel adventerous *)
      ELSE
         (* we assume all constants can fit into a register, this will NOT work for LONGREAL *)
         RETURN( TRUE )
      END
   ELSE
      RETURN( GetSize(Sym)<=WordSize )
   END
END FitsIntoRegister ;


(*
   NoOfRegisters - returns the number of registers for this processor.
*)

PROCEDURE NoOfRegisters () : CARDINAL ;
BEGIN
   RETURN( 4 )
END NoOfRegisters ;


(*
   Log2Of - returns a constant symbol whose value is log2 of constant, sym.
*)

PROCEDURE Log2Of (sym: CARDINAL) : CARDINAL ;
VAR
   i: CARDINAL ;
   a: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IF IsConst(sym)
   THEN
      FOR i := 0 TO WordSize*8 - 1 - BoundsFudge DO
         PushValue(sym) ;
         PushValue(TwoToThePower[i]) ;
         IF Equ()
         THEN
            CardToStr(i, 0, a) ;
            RETURN( MakeConstLit(MakeKey(a)) )
         END
      END
   END ;
   InternalError('expecting constant to be a power of two', __FILE__, __LINE__)
END Log2Of ;


(*
   IsPowerOfTwo - returns TRUE if sym is a power of two.
*)

PROCEDURE IsPowerOfTwo (sym: CARDINAL) : BOOLEAN ;
VAR
   i: CARDINAL ;
BEGIN
   IF IsConst(sym)
   THEN
      FOR i := 0 TO WordSize*8 - 1 - BoundsFudge DO
         PushValue(sym) ;
         PushValue(TwoToThePower[i]) ;
         IF Equ()
         THEN
            RETURN( TRUE )
         END
      END
   END ;
   RETURN( FALSE )
END IsPowerOfTwo ;


(*
   Decrement - returns a constant whose value is one less than, sym.
*)

PROCEDURE Decrement (sym: CARDINAL) : CARDINAL ;
VAR
   i: CARDINAL ;
   a: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IF IsConst(sym)
   THEN
      PushValue(sym) ;
      PushCard(1) ;
      Sub ;
      i := PopCard() ;
      CardToStr(i, 0, a) ;
      RETURN( MakeConstLit(MakeKey(a)) )
   ELSE
      InternalError('expecting a constant', __FILE__, __LINE__)
   END
END Decrement ;


(*
   DefineConstants - 
*)

PROCEDURE DefineConstants ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
   i: CARDINAL ;
   b: BITSET ;
BEGIN
   FOR i := 0 TO WordSize*8 - 1 - BoundsFudge DO
      b := {} ;
      INCL(b, i) ;
      IntToStr(INTEGER(b), 0, a) ;
      TwoToThePower[i] := MakeConstLit(MakeKey(a)) ;
   END
END DefineConstants ;


(*
   InitRegisters - 
*)

PROCEDURE InitRegisters ;
VAR
   i: CARDINAL ;
BEGIN
   FOR i := 0 TO MaxNoOfRegisters DO
      ListOfRegisters[i].CleanAlias := NIL ;
      ListOfRegisters[i].DirtyAlias := NIL
   END
END InitRegisters ;


(*
   UnHideRepsRegisters - makes EDI, ESI and ECX visible across all trees.
*)

PROCEDURE UnHideRepsRegisters ;
BEGIN
   HideEDI := FALSE ;
   HideESI := FALSE ;
   HideECX := FALSE
END UnHideRepsRegisters ;


(*
   Init - 
*)

PROCEDURE Init ;
VAR
   a: ARRAY [0..10] OF CHAR ;
BEGIN
   DebugOn                  := GetEnvironment('DEBUGDYN', a) ;
   CompilingMainModule      := FALSE ;
   UsingFunctionReturnValue := FALSE ;
   UnboundedOnStack         := FALSE ;

   PreviousScope            := NulSym ;
   CurrentProcedure         := NulSym ;

   (* define some useful constants *)
   DefineConstants ;
   UnHideRepsRegisters ;
   InitRegisters

END Init ;


BEGIN
   Init
END M2GenDyn486.
(*
 * Local variables:
 *  compile-command: "../bin1/m2f -quiet -g -verbose -M \"../libs ../gm2s -M ../libs 386 .\" -o M2GenDyn486.o M2GenDyn486.mod"
 * End:
 *)
